/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.constants;

import java.io.ObjectStreamException;
import java.rmi.RemoteException;
import java.rmi.NoSuchObjectException;
import java.rmi.MarshalException;
import java.rmi.UnmarshalException;
import java.rmi.ServerException;
import java.rmi.ServerError;

/**
 * Various constants useful in processing exceptions
 */
public class ThrowableConstants {
    /**
     * Value returned by <code>retryable</code> to indicate that the
     * passed <code>Throwable</code> does not provide any new
     * information on the state of the object that threw it.
     * @see #retryable
     */
    final static public int INDEFINITE = 0;

    /**
     * Value returned by <code>retryable</code> to indicate that the
     * passed <code>Throwable</code> implies that retrying the
     * operation that threw the <code>Throwable</code> with the same
     * arguments and the same expected return value would not be
     * fruitful.  
     * @see #retryable
     */
    final static public int BAD_INVOCATION = 1;

    /**
     * Value returned by <code>retryable</code> to indicate that the
     * passed <code>Throwable</code> implies that any further
     * operations on the object that threw the <code>Throwable</code>
     * would not be fruitful.  
     * @see #retryable
     */
    final static public int BAD_OBJECT = 2;

    /**
     * Value returned by <code>retryable</code> to indicate that the
     * passed <code>Throwable</code> was of a type that could not be
     * classified.
     * @see #retryable
     */
    final static public int UNCATEGORIZED = 3;

    /**
     * Attempt to classify the passed <code>Throwable</code> in terms of 
     * what it implies about the probability of success of future operations
     * on the object that threw the exception.
     *
     * @return <code>INDEFINITE</code>, <code>BAD_INVOCATION</code>,
     * or <code>BAD_OBJECT</code> if the exception is a
     * <code>RuntimeException</code>, <code>Error</code>, or
     * <code>java.rmi.RemoteException</code> depending on the details of 
     * the <code>Throwable</code>.  Otherwise return <code>UNCATEGORIZED</code>
     * @throws NullPointerException if the passed <code>Throwable</code> is
     * <code>null</code>
     */
    public static int retryable(Throwable t) {
	if (t == null) 
	    throw new NullPointerException("Must pass a non-null Throwable");

	if (t instanceof RuntimeException)
	    return BAD_OBJECT;

	if (t instanceof Error) {
	    if ((t instanceof OutOfMemoryError) ||
		(t instanceof LinkageError))
	    {
		return INDEFINITE;
	    }

	    return BAD_OBJECT;
	}

	if (t instanceof RemoteException) {
	    final RemoteException re = (RemoteException)t;

	    if (re instanceof NoSuchObjectException) 
		return BAD_OBJECT;

	    final Throwable detail = re.detail;
	    if (detail == null)
		return INDEFINITE;
	    	    
	    if (re instanceof MarshalException || 
		re instanceof UnmarshalException) 
	    {
		if (detail instanceof ObjectStreamException)
		    return BAD_INVOCATION;

		return INDEFINITE;
	    }

	    if (re instanceof ServerException) {
		return retryable(detail);
	    }

	    if (re instanceof ServerError) {
		return retryable(detail);
	    }

	    return INDEFINITE;
	}
	    
	return UNCATEGORIZED;

    }
}
