/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.example.books;

import com.sun.jini.example.spaceApplet.*;

import java.applet.*;
import java.awt.*;
import java.rmi.*;
import java.rmi.server.*;

import net.jini.space.*;
import net.jini.core.entry.*;
import net.jini.core.event.*;
import net.jini.core.lease.*;
import net.jini.core.transaction.*;

/**
 * A book seller applet that lets one read or take requests for bids
 * from a JavaSpace.  If an ISBN is specified, the templates are
 * <code>FullRFB</code> objects; otherwise they are simply
 * <code>RFB</code> objects.
 */
public class BookSeller extends RFBApplet
    implements RemoteEventListener, TxnCommune
{
    /**
     * @serial
     */

    private Checkbox		ifExistsToggle;

    /**
     * @serial
     */

    private Checkbox		ifNullTmplToggle;

    /**
     * @serial
     */

    private TimeoutField	timeout;

    public BookSeller() throws RemoteException {
        super();
    }

    /**
     * Initialize the extra GUI stuff that we need to be a seller.
     * Invokes <code>super.init()</code>
     */
    public void init() {
	setForeground(Color.black);
	setBackground(new Color(0xff, 0xff, 0xee));

	super.init();
	markFailure(false);

	buttonPanel().add(new ReadButton(this));
	buttonPanel().add(new TakeButton(this));
	buttonPanel().add(new NotifyButton(this));
	ifExistsToggle = new Checkbox("if exists", false);
	ifNullTmplToggle = new Checkbox("match any", false);
	buttonPanel().add(ifExistsToggle);
	buttonPanel().add(ifNullTmplToggle);

	timeout = new TimeoutField();
	borderPanel().add("South", timeout);

	setLeaseList();
    }

    /**
     * Implement the read button, creating a template of the
     * appropriate type and contents.
     */
    void read() {
	try {
	    Entry response;
	    Entry templ = template();
	    Transaction txn = currentTxn();
	    long lease = timeout();
	    if (ifNullTmplToggle.getState()) {
                if (ifExistsToggle.getState())
                    response = space().readIfExists(null, txn, lease);
                else
                    response = space().read(null, txn, lease);
	    } else {
	        if (ifExistsToggle.getState())
		    response = space().readIfExists(templ, txn, lease);
	        else
		    response = space().read(templ, txn, lease);
	    }
	    display(response);
	} catch (Exception e) {
	    if (e instanceof RuntimeException)
		throw (RuntimeException) e;
	    show(e);
	}
    }


    /**
     * Implement the take button, creating a template of the
     * appropriate type and contents.
     */
    void take() {
	try {
	    Entry response;
	    Entry templ = template();
	    Transaction txn = currentTxn();
	    long lease = timeout();
	    if (ifExistsToggle.getState())
		response = space().takeIfExists(templ, txn, lease);
	    else
		response = space().take(templ, txn, lease);
	    display(response);
	} catch (Exception e) {
	    if (e instanceof RuntimeException)
		throw (RuntimeException) e;
	    show(e);
	}
    }

    /**
     * Implement the notify button, creating a template of the
     * appropriate type and contents.
     */
    void requestNotify() {
	try {
	    Entry tmpl = template();
	    EventRegistration reg =
		space().notify(tmpl, currentTxn(), this, timeout(), null);
	    getLeaseList().addLease(reg.getID() + ": " + tmpl,
				    reg.getLease());
	} catch (TransactionException e) {
	    show(e);
	} catch (RemoteException e) {
	    show(e);
	}
    }

    public void notify(RemoteEvent ev) {
	msg("Event " + ev.getID() + "(" + ev.getSequenceNumber() + ")");
    }

    /**
     * Display a given RFB object in the fields, marking failure
     * if the RFB is <code>null</code>.
     */
    private void display(Entry entry) {
	if (entry == null)
	    markFailure(true);
	else {
	    RFB rfb = (RFB) entry;
	    markFailure(false);
	    order().setText(rfb.orderNumber.toString());
	    title().setText(rfb.title);
	    count().setText(rfb.count.toString());
	    if (rfb instanceof FullRFB)
		isbn().setText(((FullRFB) rfb).isbn);
	    else
		isbn().setText("<NONE>");
	}
    }

    // inherit doc comment
    void clear() {
	markFailure(false);
	super.clear();
    }

    private long timeout() {
	return timeout.value();
    }

    private static class TimeoutField extends Panel {
	TextField field = new TextField("NO_WAIT", 15);

	TimeoutField() {
	    add(new Label("Timeout:"));
	    add(field);
	}

	long value() {
	    long val;
	    String txt = field.getText();
	    try {
		val = Long.parseLong(txt);
	    } catch (NumberFormatException e) {
		if (Character.toLowerCase(txt.charAt(0)) == 'f') {
		    val = 0;
		    field.setText("FOREVER");
		} else {
		    val = JavaSpace.NO_WAIT;
		    field.setText("NO_WAIT");
		}
	    }
	    return val;
	}
    }

    public static void main(String[] args) throws Throwable {
	main(new BookSeller());
    }
}
