/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.example.books;

import com.sun.jini.example.spaceApplet.*;
import net.jini.core.transaction.*;
import net.jini.core.transaction.server.*;
import net.jini.core.lease.*;

import java.awt.*;
import java.awt.List;	// works around some conflict in JDK 1.2 beta 2
import java.awt.event.ActionEvent;
import java.rmi.*;
import java.rmi.registry.*;
import java.rmi.server.*;
import java.util.*;

/**
 * An abstract class for writing applets that manage RFB objects.
 * @see BookBuyer
 * @see BookSeller
 * @see RFB
 * @see FullRFB
 */
public abstract class RFBApplet extends SpaceApplet implements TxnCommune {
    /**
     * @serial
     */
    private Label	order;			// display of order num

    /**
     * @serial
     */
    private TextField	isbn, title, count;	// fields for entries

    /**
     * @serial
     */
    private Panel	buttonPanel;		// the command buttons

    /**
     * @serial
     */
    private Panel	borderPanel;		// contains buttons & additions

    /**
     * @serial
     */
    private TextArea	msg;			// messages to the user

    /**
     * @serial
     */
    private List	txnList;		// list of known txns

    /**
     * @serial
     */
    private int		nextTxn = 0;		// next Txn # to assign

    /**
     * @serial
     */
    private Hashtable	txns = new Hashtable();	// all Txns stored here

    /**
     * @serial
     */
    private LeaseList	leaselist = null;	// list of transaction leases

    /**
     * @serial
     */
    private TxnCommune	executive = null;	// collective's excutive

    /**
     * @serial
     */
    private Hashtable	members = null;		// collective's members

    /**
     * @serial
     */
    private String	myName;			// my member name

    /**
     * @serial
     */
    private Color	bg = null;		// background color

    /**
     * @serial
     */
    private Color	fg = null;		// foreground color
    private static final boolean DEBUG = false;

    /** Return the object used to store the order */
    public Label order() { return order; }
    /** Return the object used for the ISBN number */
    public TextField isbn() { return isbn; }
    /** Return the object used for the title */
    public TextField title() { return title; }
    /** Return the object used for the count of desired copies */
    public TextField count() { return count; }
    /** Return the object used for the command buttons */
    public Panel buttonPanel() { return buttonPanel; }

    public RFBApplet() throws RemoteException {
	UnicastRemoteObject.exportObject(this);
    }

    /**
     * Set up for this applet.  Must be invoked by subtypes
     */
    public void init() {
	super.init();
	joinCommune();
	GridBagConstraints left = new GridBagConstraints();
	GridBagConstraints right = new GridBagConstraints();

	right.anchor = GridBagConstraints.WEST;
	right.fill = GridBagConstraints.HORIZONTAL;
	right.gridwidth = GridBagConstraints.REMAINDER;
	right.weightx = 1.0;

	add(new Label("Order:", Label.RIGHT), left);
	add(order = new Label("", Label.LEFT), right);

	add(new Label("ISBN:", Label.RIGHT), left);
	add(isbn = new TextField(30), right);

	add(new Label("Title:", Label.RIGHT), left);
	add(title = new TextField(30), right);

	add(new Label("Count:", Label.RIGHT), left);
	add(count = new TextField(30), right);

	buttonPanel = new Panel();
	buttonPanel.setLayout(new FlowLayout(FlowLayout.CENTER));
	buttonPanel.add(new ActionButton("Clear") {
	    public void actionPerformed(ActionEvent e) { invalidate(); clear(); }
	});
	borderPanel = new Panel();
	borderPanel.setLayout(new BorderLayout());
	borderPanel.add("North", buttonPanel);
	add(borderPanel, right);

	add(txnPanel(), right);

	add(msg = new TextArea(5, 30), right);

	clear();	// start out in the normal initial state
    }

    public void setLeaseList() {
        GridBagConstraints c = new GridBagConstraints();
        c.fill = GridBagConstraints.BOTH;
        c.gridwidth = GridBagConstraints.REMAINDER;
        c.weightx = 1.0;
        c.weighty = 1.0;

	leaselist = new LeaseList(10, this);
	add(leaselist, c);
    }

    public LeaseList getLeaseList() {
	return leaselist;
    }

    public long getLeaseValue() {
        long lease =
	    (leaselist == null ? Long.MAX_VALUE : leaselist.leaseTime());
	return lease;
    }

    public Panel borderPanel() {
	return borderPanel;
    }

    public void destroy() {
	super.destroy();
	if (members == null) {	// the executive hangs out as long as it can
	    try {
		executive.delete(this);
	    } catch (RemoteException e) {
		;		// well, we tried...
	    }
	}
    }

    private class BorderedPanel extends Panel {
	private final int gap, width;

	BorderedPanel(int gap, int width) {
	    this.gap = gap;
	    this.width = width;
	}

	public Insets getInsets() {
	    return new Insets(2 * gap + width, 2 * gap + width,
			      2 * gap + width, 2 * gap + width);
	}

	public void paint(Graphics g) {
	    Dimension sz = getSize();
	    for (int i = 0; i < width; i++) {
		g.drawRect(gap + i, gap + i,
			   sz.width - 2 * gap - 2 * i,
			   sz.height - 2 * gap - 2 * i);
	    }
	    super.paint(g);
	}
    }

    private Component txnPanel() {
	Panel p = new BorderedPanel(3, 1);
	p.setLayout(new GridBagLayout());

	GridBagConstraints list = new GridBagConstraints();
	GridBagConstraints right = new GridBagConstraints();

	list.weightx = 1.0;
	list.weighty = 1.0;
	list.gridheight = GridBagConstraints.REMAINDER;
	list.fill = GridBagConstraints.BOTH;
	list.insets.left = list.insets.right = 3;

	right.anchor = GridBagConstraints.WEST;
	right.fill = GridBagConstraints.HORIZONTAL;
	right.gridwidth = GridBagConstraints.REMAINDER;
	right.weightx = 0.0;

	p.add(new Label("Transactions", Label.CENTER), right);

	p.add(txnList = new List(), list);

	p.add(new ActionButton("New") {
	    public void actionPerformed(ActionEvent e) { newTxn(); }
	}, right);
	p.add(new ActionButton("Commit") {
	    public void actionPerformed(ActionEvent e) { commitTxn(); }
	}, right);
	p.add(new ActionButton("Abort") {
	    public void actionPerformed(ActionEvent e) { abortTxn(); }
	}, right);
	p.add(new ActionButton("None") {
	    public void actionPerformed(ActionEvent e) { noTxn(); }
	}, right);

	right.fill = GridBagConstraints.BOTH;
	right.gridheight = GridBagConstraints.REMAINDER;

	return p;
    }

    void addTxnListener(java.awt.event.ItemListener listener) {
	txnList.addItemListener(listener);
    }

    synchronized void newTxn() {
	try {
	    TransactionManager mgr = mgr();
	    Transaction.Created trc =
		TransactionFactory.create(mgr, getLeaseValue());
	    Transaction txn = trc.transaction;
	    executive.add(txn, myName + " " + nextTxn++);
	} catch (LeaseDeniedException lde) {
	    show(lde);
	} catch (RemoteException e) {
	    show(e);
	}
    }

    void addTxn(Transaction txn, String name) {
	synchronized (txnList) {
	    txnList.add(name);
	    txns.put(name, txn);
	    txnList.select(txnList.getItemCount() - 1);
	}
    }

    void deleteTxn(String name) {
	synchronized (txnList) {
	    String[] items = txnList.getItems();
	    for (int i = 0; i < items.length; i++) {
		if (items[i].equals(name)) {
		    txnList.remove(i);
		    return;
		}
	    }
	}
	System.out.println("Transaction named \"" + name + "\" not found!");
    }

    synchronized void noTxn() {
	synchronized (txnList) {
	    txnList.deselect(txnList.getSelectedIndex());
	}
    }

    synchronized void commitTxn() {
	String txnName = txnList.getSelectedItem();
	Transaction txn = (Transaction) txns.get(txnName);
	if (txn == null)
	    return;
	try {
	    txn.commit();  //do not wait any amount of time
	    executive.delete(txnName);
	} catch (TransactionException e) {
	    show(e);
	} catch (RemoteException e) {
	    show(e);
	}
    }

    synchronized void abortTxn() {
	String txnName = txnList.getSelectedItem();
	Transaction txn = (Transaction) txns.get(txnName);
	if (txn == null)
	    return;
	try {
	    txn.abort(); // don't wait any amount of time
	    executive.delete(txnName);
	} catch (RemoteException e) {
	    show(e);
	} catch (TransactionException e) {
	    show(e);
	}
    }

    synchronized Transaction currentTxn() {
	String txnName = txnList.getSelectedItem();
	if (txnName == null)
	    return null;
	return (Transaction) txns.get(txnName);
    }

    /**
     * Clear out the fields to their default states.  Subtypes should
     * invoke <code>super.clear()</code>.
     */
    void clear() {
	isbn.setText("");
	title.setText("");
	count.setText("");
	isbn.requestFocus();
    }

    /**
     * Return the template object that is appropriate for the contents
     * of the fields.
     */
    RFB template() {
	if(DEBUG)
	    System.out.println("RFBApplet: tempate");
	boolean full = !isbn().getText().equals("");
	RFB tmpl = (full ? new FullRFB() : new RFB());

	tmpl.title = title().getText();
	if (tmpl.title.equals(""))
	    tmpl.title = null;

	String cnt = count().getText();
	if (cnt != null && !cnt.equals(""))
	    tmpl.count = new Integer(cnt);

	if (full) {
	    FullRFB fulltmpl = (FullRFB) tmpl;
	    fulltmpl.isbn = isbn().getText();
	    if (fulltmpl.isbn.equals(""))
		fulltmpl.isbn = null;
	}

	if(DEBUG)
	    System.out.println("RFBApplet: template: tmpl= " + tmpl);

	return tmpl;
    }

    /**
     * Display the given message to the user
     */
    //!! Maybe this should be a scrolling text area that people write to
    public void msg(String str) {
	msg.append(str);
	msg.append("\n");
    }

    void show(Throwable t) {
	t.printStackTrace(System.out);
	msg(t.toString());
	markFailure(true);
    }

    /**
     * Mark the failure of the query.
     * @see #display
     */
    void markFailure(boolean failed) {
	if (bg == null) {
	    bg = getBackground();
	    fg = getForeground();
	}
	if (failed) {
	    setForeground(bg);
	    setBackground(fg);
	    order().setText("null");
	} else {
	    setForeground(fg);
	    setBackground(bg);
	}
    }

    /**
     * A "Clear the RFB fields" action.
     */
    private class ClearButton extends ActionButton {
	ClearButton(RFBApplet mgr) {
	    super("Clear");
	}

	public final void actionPerformed(java.awt.event.ActionEvent e) {
	    clear();
	}
    }

    // ------------------------------------------------------------
    // 			Commune stuff
    // ------------------------------------------------------------

    private void joinCommune() {
	final String communeName = "RFB TxnCommune";

	Throwable exc = null;
	try {
	    try {
		Naming.bind(communeName, this);
		executive = this;
		members = new Hashtable();
		myName = myType();
		members.put(myName, this);
		System.out.println(myType() + " is executive");
	    } catch (AlreadyBoundException e) {
		executive = (TxnCommune) Naming.lookup(communeName);
		members = null;
		myName = executive.add(this, myType());
	    }
	} catch (NotBoundException e) {
	    exc = e;
	} catch (java.net.MalformedURLException e) {
	    exc = e;
	} catch (RemoteException e) {
	    exc = e;
	} finally {
	    if (exc != null)
		exc.printStackTrace();
	}

	if (DEBUG)
	    System.out.println("RFBApplet: joinCommune: executive = " + executive);
    }

    public synchronized String add(TxnCommune member, String nameBase) {
	if (members == null)
	    throw new IllegalArgumentException("member sent to non-executive");
	String name = nameBase;
	for (int i = 0; members.containsKey(name); i++)
	    name = nameBase + "_" + i;
	members.put(name, member);
	return name;
    }

    public void delete(TxnCommune member) {
	if (members == null)
	    throw new IllegalArgumentException("deletion from non-executive");
	members.remove(member);
    }

    public void add(Transaction tr, String name) {
	addTxn(tr, name);
	if (members != null) {
	    Enumeration e = members.elements();
	    while (e.hasMoreElements()) {
		TxnCommune member = (TxnCommune) e.nextElement();
		if (member == this)
		    continue;
		try {
		    member.add(tr, name);
		} catch (RemoteException ex) {
		    delete(member);
		    ex.printStackTrace();
		}
	    }
	}
    }

    public void delete(String name) {
	deleteTxn(name);
	if (members != null) {
	    Enumeration e = members.elements();
	    while (e.hasMoreElements()) {
		TxnCommune member = (TxnCommune) e.nextElement();
		if (member == this)
		    continue;
		try {
		    member.delete(name);
		} catch (RemoteException ex) {
		    ex.printStackTrace();
		}
	    }
	}
    }
}
