/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.example.books;

import com.sun.jini.example.spaceApplet.*;

import java.awt.*;
import java.awt.event.*;
import java.rmi.*;
import java.util.Vector;
import java.util.Date;
import java.text.DateFormat;

import net.jini.core.entry.*;
import com.sun.jini.outrigger.*;

/**
 * A panel that watches the contents of a JavaSpace.
 * It creates a thread that regularly queries the contents of the space,
 * updating things that have changed.  This will not scale to large contents.
 *
 * <b>Note:</b> currently this requires deseriailizing the <code>Entry</code>
 * objects, which means loading the classes into the server process.
 */
public class Watcher extends RFBApplet implements Runnable, ItemListener {
    /**
     * @serial
     */
    private List		contents;

    /**
     * @serial
     */
    private Thread		thread;

    /**
     * @serial
     */
    private long		timeout = 5000;

    /**
     * @serial
     */
    private Entry		tmpl = new RFB();

    /**
     * @serial
     */
    private Vector		entries = new Vector();

    /**
     * @serial
     */
    private boolean		watching = true;	// user wants watching

    /**
     * @serial
     */
    private boolean		doUpdates = false;	// show the updates

    /**
     * @serial
     */
    private DateFormat		dateFmt;

    /**
     * @serial
     */
    private Label		countLabel;

    /**
     * @serial
     */
    private boolean		finished = false;
    private static final boolean DEBUG = false;

    /**
     * @serial
     */
    private Checkbox            matchAnyToggle;


    public Watcher() throws RemoteException { }

    /**
     * After invoking <code>super.init</code> set up the additional state
     * needed for administering the space.
     */
    public void init() {
	super.init();
	waitForAdmin();

	buttonPanel().add(new UpdateButton());
	buttonPanel().add(new WatchButton());
	buttonPanel().add(new UseButton());
	buttonPanel().add(new DestroyButton());
        matchAnyToggle = new Checkbox("match any", false);
	buttonPanel().add(matchAnyToggle);

	GridBagConstraints line = new GridBagConstraints();

	line.fill = GridBagConstraints.HORIZONTAL;
	line.gridwidth = GridBagConstraints.REMAINDER;

	line.fill = GridBagConstraints.HORIZONTAL;
	line.weighty = 0.0;
	countLabel = new Label("", Label.LEFT);
	add(countLabel, line);

	line.fill = GridBagConstraints.BOTH;
	line.weighty = 1.0;
	contents = new List();
	contents.setMultipleMode(true);
	add(contents, line);

	dateFmt = DateFormat.getInstance();

	addTxnListener(this);
	setLeaseList();
    }

    /**
     * How long do we wait between auto-updates?
     */
    public synchronized long timeout() {
	return timeout;
    }

    /**
     * Set how long we wait between auto-updates.
     */
    public synchronized void timeout(long newTimeout) {
	timeout = newTimeout;
    }

    public void stop() {
	doUpdates = false;
    }

    /**
     * Get the auto-update thread running.
     */
    public void start() {
	super.start();
	if (thread == null) {
	    thread = new Thread(this);
	    thread.setPriority(Thread.MIN_PRIORITY);
	    thread.start();
	}
	doUpdates = true;
    }

    /**
     * Repetitively update the display, pausing inbetween.
     * @see #update
     */
    public void run() {
	while (!finished) {
	    if (DEBUG) {
		System.out.println("Watcher: run: watching= " + watching + " doUpdates= " + doUpdates);
	    }

	    if (watching && doUpdates)
		update();
	    pause();
	}
    }

    /**
     * Make a "best effort" attempt to update the display to the current
     * state.
     */
    public synchronized void update() {
	if (DEBUG)
	    System.out.println("Watcher: update");
	synchronized (contents) {
	    try {
	        if (DEBUG) {
		   System.out.println("Watcher: update: admin= " + admin());
		   System.out.println("Watcher: update: tmpl= " + tmpl);
                }
		AdminIterator iter = admin().contents(tmpl, currentTxn(), 10);
		Entry cur;
		int origCnt = contents.getItemCount();
		int pos = 0;

		entries.removeAllElements();

		while ((cur = iter.next()) != null) {
		    String n;
		    entries.addElement(cur);
		    n = cur.toString();
		    if (DEBUG)
			System.out.println("Watcher: update: item= " + n);
			    
		    if (pos >= origCnt)
			contents.add(n);
		    else {
			String o = contents.getItem(pos);
			if (!n.equals(o))
			    contents.replaceItem(n, pos);
		    }
		    
		    pos++;
		}

		if (pos < origCnt) {
		    for (int i = origCnt - 1; i >= pos; i--)
			contents.remove(i);
		}
		if (pos == 1)
		    countLabel.setText("1  entry");
		else
		    countLabel.setText(String.valueOf(pos) + " entries");

	    } catch (UnusableEntryException e) {
		show(e);
		return;	// just try again next time
	    } catch (RemoteException e) {
		return;	// just try again next time
	    } catch (net.jini.core.transaction.TransactionException e) {
		return;	// just try again next time
	    }
	}
    }

    /**
     * Wait for the next update time.
     */
    private void pause() {
	try {
	    long t;
	    synchronized (this) {
		t = timeout;
	    }
	    Thread.sleep(t);
	} catch (InterruptedException e) {
	    return;
	}
    }

    public void itemStateChanged(ItemEvent evt) {
	forceUpdate();
    }

    /**
     * Force an update now, either by making the thread cycle or just
     * doing an <code>update</code>.
     */
    private void forceUpdate() {
	if (DEBUG)
	    System.out.println("Watcher: forceUpdate");
	if (!watching)
	    update();
	else
	    thread.interrupt();
    }

    /**
     * Set the template for the iterator.
     */
    private void setTemplate() {
	if (matchAnyToggle.getState())
	    tmpl = null;
	else
	    tmpl = template();
	forceUpdate();
    }

    // purposefully inherit doc comment from supertype
    public void destroy() {
	super.destroy();
	if (thread != null) {
	    finished = true;
	    thread.interrupt();
	}
    }

    /**
     * Start/stop watching
     */
    void watch(boolean start) {
	if (start != watching)
	    watching = start;
    }

    /**
     * Return whether we are currently watching.
     */
    boolean watching() {
	return watching;
    }

    /**
     * Superclass for buttons we add.
     */
    private abstract class WatcherButton extends TimedButton {
	/**
	 * Create a WatcherButton for the given <code>watcher</code>, with the
	 * given <code>label</code>.
	 */
	WatcherButton(String label) {
	    super(label, Watcher.this);
	}
    }

    /**
     * A button that forces an update now.
     */
    private class UpdateButton extends WatcherButton {
	UpdateButton() { super("Update"); }
	public void pushed() { setTemplate(); Watcher.this.forceUpdate(); }
    }

    /**
     * A button that toggles whether to auto-update the list of entries.
     */
    private final static String START_LABEL = "Watch";	// no statics in inner
    private final static String STOP_LABEL  = "Stop";	// ... sigh ...

    private class WatchButton extends WatcherButton {

	WatchButton() { super(watching() ? STOP_LABEL : START_LABEL); }

	public void pushed() {
	    setTemplate();
	    boolean start = getLabel().equals(START_LABEL);
	    watch(start);
	    setLabel(start ? STOP_LABEL : START_LABEL);
	}
    }

    /**
     * A button that gets us to start using the template described in
     * the fields.
     */
    private class UseButton extends WatcherButton {
	UseButton() { super("Use"); }
	public void pushed() { setTemplate(); }
    }

    /**
     * A button to destroy the server.
     */
    private class DestroyButton extends WatcherButton {
	DestroyButton() { super("Destroy"); }
	public void pushed() {
	    try {
		admin().destroy();
	    } catch (RemoteException e) {
		Watcher.this.show(e);
	    }
	}
    }

    public static void main(String[] args) throws Throwable {
	main(new Watcher());
    }
}
