/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.example.genericWorker;

import com.sun.jini.example.spaceApplet.*;
import net.jini.core.entry.*;
import net.jini.core.transaction.*;
import net.jini.core.lease.*;
import com.sun.jini.collection.WeakTable;

import java.rmi.RemoteException;
import java.awt.*;
import java.awt.event.*;
import java.net.*;

import java.util.Hashtable;

import sun.misc.Ref;

/**
 * A generic worker class that takes <code>TaskEntry</code> entries,
 * invokes their <code>execute</code> methods, and writes any resulting
 * entries back into the space.
 *
 * @see TaskEntry
 */
public class GenericWorker extends SpaceApplet implements Runnable {
    private TextArea	history;
    private Label	msg;
    private Thread	thread;
    private boolean	pause;
    private boolean	first = true;
    private WeakTable	cache = new WeakTable();

    public void init() {
        super.init();

	// find out who I am
        InetAddress myAddr = null;
        String myHost;
        try {
            myAddr = InetAddress.getLocalHost();
            myHost = myAddr.getHostName();
        } catch (java.net.UnknownHostException e) {
            if (myAddr != null)
                myHost = myAddr.toString();
            else
                myHost = "localhost";
        }
        System.out.println("I am running on " + myHost + " and talking to " +
			   "the space named " + getSpaceName());

	GridBagConstraints left = new GridBagConstraints();
	GridBagConstraints right = new GridBagConstraints();
	right.anchor = GridBagConstraints.WEST;
	right.fill = GridBagConstraints.HORIZONTAL;
	right.gridwidth = GridBagConstraints.REMAINDER;
	right.weightx = 1.0;
	right.weighty = 0.0;

        String whoAmI = myHost + "->" + getSpaceName();
        add(new Label(whoAmI, Label.CENTER), right);
	add(new StartButton(), left);
	add(new StopButton(), left);
	add(new CancelButton(), left);
	add(new Label(), right);

	// message display
	msg = new Label("");
	add(msg, right);

	// work history scrolling region
	right.weighty = 1.0;
	history = new TextArea();
	history.setEditable(false);
	add(history, right);
    }

    /** Start reading requests. */
    private class StartButton extends ActionButton {
	StartButton() { super("Start"); }
	public void actionPerformed(ActionEvent e) { start(); }
    }

    /** Stop reading requests. */
    private class StopButton extends ActionButton {
	StopButton() { super("Stop"); }
	public void actionPerformed(ActionEvent e) { stop(); }
    }

    /** Cancel current task */
    private class CancelButton extends ActionButton {
	CancelButton() { super("Cancel"); }
	public void actionPerformed(ActionEvent e) { cancelTask(); }
    }

    public synchronized void start() {
	if (thread == null) {
	    thread = new Thread(this);
	    thread.start();
	}
	pause = false;	// unpause the thread
	notifyAll();	// let them know
    }

    public synchronized void stop() {
	if (thread == null)
	    return;
	pause = true;	// pause the thread
    }

    public synchronized void cancelTask() {
	if (thread == null)
	    return;
	if (!pause)
	    thread.interrupt();	// request an interrupt -- it's all we can do
    }

    public synchronized void run() {
	Entry taskTmpl = new TaskEntry();	// any TaskEntry will do
	for (;;) {
	    while (pause) {			// if we're being asked to pause
		msg("pausing");
		try {
		    wait();			// wait until we're asked not to
		} catch (InterruptedException e) {
		    continue;
		}
	    }

	    msg("getting task");
	    try {
		TaskEntry task =
		    (TaskEntry) space().take(taskTmpl, null, Long.MAX_VALUE);
		msg("");
		showEntry("in:  ", task);
		task.verbose(false);
		Entry result = task.execute(cache);// execute the task
		showEntry("out: ", result);
		if (result != null)		   // write any result back
		    space().write(result, null, task.responseLeaseTime());
	    } catch (RemoteException e) {
		show(e);
	    } catch (TransactionException e) {
		show(e);
	    } catch (UnusableEntryException e) {
		show(e);
	    } catch (InterruptedException e) {
		msg("Task cancelled");
	    }
	}
    }

    private static class Leaving {
	private Class cl;
	private String key;
	private Leaving(Class cl, String key) {
	    if (key.length() >= 128)
		throw new IllegalArgumentException("key string must be < 128");
	    this.cl = cl;
	    this.key = key;
	}

	public int hashCode() {
	    return cl.hashCode() ^ key.hashCode();
	}
	public boolean equals(Object thatObj) {
	    try {
		Leaving that = (Leaving)thatObj;
		return (cl.equals(that.cl) && key.equals(that.key));
	    } catch (ClassCastException e) {
		return false;
	    }
	}
    }

    /**
     * Show the entry in the history log.
     */
    private void showEntry(String prefix, Entry task) {
	if (!first)
	    history.append("\n");

	//StringBuffer buf = new StringBuffer();
	history.append(prefix);
	String cname = task.getClass().getName();
	int i = cname.lastIndexOf('.');
	if (i >= 0)
	    cname = cname.substring(i + 1);
	history.append(cname);
	history.append(": ");
	history.append(task.toString());
	//history.append(buf.toString());
	//System.out.println(buf);
	first = false;
    }

    /**
     * Display the given message.
     */
    public void msg(String txt) {
	/*
	if (!txt.equals(""))
	    System.out.println("MSG: " + txt);
	*/
	msg.setText(txt);
    }

    /**
     * Show an exception.
     */
    void show(Throwable t) {
	t.printStackTrace(System.out);
	msg(t.toString());
    }

    public static void main(String[] args) throws Throwable {
	main(new GenericWorker());
    }
}
