/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.example.spaceApplet;

import net.jini.space.*;
import net.jini.admin.Administrable;
import com.sun.jini.outrigger.TransientSpace;
import com.sun.jini.mahout.Locator;
import com.sun.jini.outrigger.Finder;
import com.sun.jini.outrigger.JavaSpaceAdmin;
import net.jini.core.transaction.server.TransactionManager;

import java.applet.*;
import java.awt.*;
import java.rmi.*;
import java.rmi.registry.*;
import javax.swing.JFrame;

/**
 * An abstract class for writing applets that interact with a JavaSpace.
 *
 * @author Ken Arnold
 */
public abstract class SpaceApplet extends Applet {
    /**
     * @serial
     */

    private GridBagLayout	gridBag = new GridBagLayout();

    /**
     * @serial
     */
    private JavaSpace		space;			// the space being used

    /**
     * @serial
     */

    private JavaSpaceAdmin	admin;			// admin being used

    /**
     * @serial
     */
    private TransactionManager	mgr;			// txn mgr being used

    /**
     * @serial
     */
    private String		myTypeName = null;

    /**
     * @serial
     */

    private boolean		asApplet = true;	// run as real applet

    /**
     * The name of the space we are looking for
     * @serial
     */
    private String spaceName;

    /**
     * The name the application wishes to call itself.
     * This field is only needed for as long as we support or
     * warn about looking up services in RMI registry
     * @serial
     */
    private String appName;

    private static boolean DEBUG = false;

    /**
     * Display the given message to the user.
     */
    public abstract void msg(String str);

    /**
     * Return the space being used by this applet.  <code>null</code>
     * means it hasn't yet appeared.  Call <code>waitForSpace</code> at
     * least once before using this method.
     *
     * @see #waitForSpace()
     */
    public JavaSpace space() { return space; }

    /**
     * Return the transaction manager being used by this applet.
     *
     * @see #space()
     */
    public TransactionManager mgr() { return mgr; }

    /**
     * Return the administrator being used by this applet.  <code>null</code>
     * means it hasn't yet appeared.  Call <code>waitForAdmin</code> at
     * least once before using this method.
     *
     * @see #waitForAdmin()
     */
    public JavaSpaceAdmin admin() { return admin; }

    /**
     * Simple no-arg constructor
     * <p>
     * This constructor is only needed for as long as we support or
     * warn about looking up services in RMI registry
     */
    public SpaceApplet() {
	try {
	    final String className = (this.getClass()).getName();
	    final int lastDot = className.lastIndexOf(".");
	    appName = className.substring(lastDot + 1);
	} catch (RuntimeException e) {
	    appName = "SpaceApplet";
	}
    }

    /**
     * Set up for this applet.  Must be invoked by subtypes
     */
    public void init() {
	space = waitForSpace();
	mgr = waitForMgr();
	setLayout(gridBag);
    }

    public void start() {
	try {
	    Dimension pref = getPreferredSize();
	    Dimension size = getSize();
	    if (size.width < pref.width || size.height < pref.height ||
	        Boolean.getBoolean("com.sun.jini.example.spaceApplet.SpaceApplet.reportIdealSize"))
	    {
		System.out.println("prefered size for " + myType() + ": "
		    + pref.width + " x " + pref.height);
		System.out.println("actual size for " + myType() + ": "
		    + size.width + " x " + size.height);
	    }
	} catch (SecurityException e) {		// properties may be protected
	    e.printStackTrace(System.err);
	}
    }

    /**
     * Set the name used for this application/applet in warning messages
     * <p>
     * This method is only needed for as long as we support or
     * warn about looking up services in RMI registry
     * <p>
     * @param Name we want this application to have
     */
    public void setApplicationName(String name) {
	appName = name;
    }

    /**
     * Print a warning message about RMI registry use
     * <p>
     * This method is only needed for as long as we support or
     * warn about looking up services in RMI registry
     * <p>
     * @param Server that is being found in the RMI registry
     */
    private void registryWarning(String server) {
	System.err.println(
	"\n" + 
 	"**Warning**: You are using the com.sun.jini.use.registry property\n" + 
 	"in order to have " + appName + " find a " + server + " server\n" +
 	"in an RMI registry instead of a Jini(TM) technology lookup\n" +
        "service. Support for this functionality is going to be removed\n" +
	"in future versions of the JavaSpaces(TM) Technology Kit (JSTK).\n");
    }


    /**
     * Wait for the JavaSpace object to appear, and then return it.
     */
    public JavaSpace waitForSpace() {
	// transient space and persistent space use the same name
	Locator locator = null;
	Finder finder = null;
	
	if (System.getProperty("com.sun.jini.use.registry") != null) {
	    locator = new com.sun.jini.mahout.RegistryLocator();
	    finder = new com.sun.jini.outrigger.RegistryFinder();

	    // This call is only needed for as long as we support or
	    // warn about looking up services in RMI registry
	    registryWarning("JavaSpaces(TM)");
	} else {
	    locator = new com.sun.jini.outrigger.DiscoveryLocator();
	    finder = new com.sun.jini.outrigger.LookupFinder();
	}

	if (DEBUG) {
	    System.out.println("SpaceApplet: waitForSpace: finder = " + finder);
	    System.out.println("SpaceApplet: waitForSpace: locator = " + locator);
	}
	space = (JavaSpace)finder.find(locator, getSpaceName());

	if (DEBUG) {
	    if (space != null)
	        System.out.println("SpaceApplet: waitForSpace: " +
				"space = " + space);
	}

	return space;
    }

    /**
     * Return the name of the space we are looking for
     */
    public String getSpaceName() {
	if (spaceName == null) {
	    spaceName = System.getProperty("outrigger.spacename");
	}

	return spaceName;
    }

    /**
     * Wait for the administrative object to appear, and then return it.
     */
    public JavaSpaceAdmin waitForAdmin() {
	// transient space and persistent space use the same name
	do {
	    try {
		if (space != null) {
		    admin = (JavaSpaceAdmin)((Administrable)space).getAdmin();
		    if (DEBUG) {
			System.out.println("SpaceApplet: waitForAdmin" +
					   " admin = " + admin);
		    }
		}
	    } catch (RemoteException e) {
		e.printStackTrace();
	    }
	} while (admin == null);

        if (DEBUG) {
            if (space != null)
                System.out.println("SpaceApplet: waitForAdmin: " +  
                                "admin = " + admin);
        }
 
	return admin;
    }

    /**
     * Wait for the transaction manager object to appear, and then return it.
     */
    public TransactionManager waitForMgr() {
	Locator locator = null;
	Finder finder = null;
	
	if (System.getProperty("com.sun.jini.use.registry") != null) {
	    locator = new com.sun.jini.mahout.RegistryLocator();
	    finder = new com.sun.jini.outrigger.RegistryFinder();

	    // This call is only needed for as long as we support or
	    // warn about looking up services in RMI registry
	    registryWarning("Jini(TM) Transaction Manager");
	} else {
	    locator = new com.sun.jini.outrigger.DiscoveryLocator();
	    finder = new com.sun.jini.outrigger.LookupFinder();
	}

        mgr = (TransactionManager)finder.find(locator, 
	    com.sun.jini.mahalo.TxnManagerImpl.DEFAULT_NAME);
        return mgr;
    }

    protected String myType() {
	if (myTypeName == null) {
	    String full = getClass().getName();
	    myTypeName = full.substring(full.lastIndexOf('.') + 1);
	}
	return myTypeName;
    }

    /**
     * Start up the applet in its own frame.
     */
    public static void main(SpaceApplet applet) {
	applet.asApplet = false;
	System.setSecurityManager(new RMISecurityManager());
	JFrame frame = new JFrame();
	frame.setTitle(applet.getClass().getName());
	frame.getContentPane().add(applet, BorderLayout.CENTER);
	applet.init();
	applet.start();
	frame.setVisible(true);
	Dimension pref = applet.getPreferredSize();
	frame.setSize(pref); // prefered size known only after visibility (sigh)
	frame.validate();
    }
}
