/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.fiddler;

import com.sun.jini.fiddler.FiddlerAdmin;

import com.sun.jini.mahout.KillVMAdmin;
import com.sun.jini.proxy.UUID;

import net.jini.core.entry.Entry;
import net.jini.core.discovery.LookupLocator;

import java.io.IOException;
import java.io.Serializable;
import java.rmi.RemoteException;

/**
 * This class is a proxy providing access to the methods of an implementation
 * of the lookup discovery service which allow the administration of the
 * service. Clients only see instances of this class via the FiddlerAdmin
 * interface.
 */
class FiddlerAdminProxy implements FiddlerAdmin, KillVMAdmin, Serializable {

    private static final long serialVersionUID = -9083313206374117425L;

    /**
     * The reference through which communication occurs between the
     * client-side and the server-side of the lookup discovery service
     *
     * @serial
     */
    private final Fiddler server;
    /** 
     * The unique identifier assigned to the current instance of this 
     * proxy class by the lookup discovery service. This ID is used to
     * determine equality between proxies.
     *
     * @serial
     */
    private final UUID proxyID;

    /**
     * Constructs a new instance of FiddlerAdminProxy.
     *
     * @param server   reference to the server object through which 
     *                 communication occurs between the client-side and
     *                 server-side of the lookup discovery service
     * @param proxyID  the unique identifier assigned by the lookup
     *                 discovery serviceto the current instance of this proxy
     */
    public FiddlerAdminProxy(Fiddler server, UUID proxyID) {
	this.server  = server;
	this.proxyID = proxyID;
    }

    /* *** Methods of com.sun.jini.fiddler.FiddlerAdmin *** */

    /**
     * Changes the least upper bound applied to all lease durations granted
     * by the lookup discovery service.
     * <p>
     * This method is a mechanism for an entity with the appropriate
     * privileges to administratively change the value of the least upper
     * bound that will be applied by the Fiddler implementation of the lookup
     * discovery service when determining the duration to assign to the lease
     * on a requested registration.
     *
     * @param newBound <code>long</code> value representing the new least
     *        upper bound (in milliseconds) on the set of all possible
     *        lease durations that may be granted
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         bound value may or may not have been changed successfully.
     *
     * @see com.sun.jini.fiddler.FiddlerAdmin#setLeaseDurationBound
     */
    public void setLeaseDurationBound(long newBound) throws RemoteException {
        server.setLeaseDurationBound(newBound);
    }
    /**
     * Retrieves the least upper bound applied to all lease durations granted
     * by the lookup discovery service.
     *
     * @return <code>long</code> value representing the current least
     *         upper bound (in milliseconds) on the set of all possible
     *         lease durations that may be granted
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see com.sun.jini.fiddler.FiddlerAdmin#setLeaseDurationBound
     */
    public long getLeaseDurationBound() throws RemoteException {
        return server.getLeaseDurationBound();
    }

    /**
     * Change the weight factor applied by the lookup discovery service
     * to the snapshot size during the test to determine whether or not
     * to take a "snapshot" of the system state.
     *
     * @param weight weight factor for snapshot size
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         weight factor may or may not have been changed successfully.
     *
     * @see com.sun.jini.fiddler.FiddlerAdmin#setSnapshotWeight
     */
    public void setSnapshotWeight(float weight) throws RemoteException {
        server.setSnapshotWeight(weight);
    }

    /**
     * Retrieve the weight factor applied by the lookup discovery service
     * to the snapshot size during the test to determine whether or not to
     * take a "snapshot" of the system state.
     * 
     * @return float value corresponding to the weight factor for snapshot
     *         size
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see com.sun.jini.fiddler.FiddlerAdmin#getSnapshotWeight
     */
    public float getSnapshotWeight() throws RemoteException {
        return server.getSnapshotWeight();
    }

    /**
     * Change the value of the size threshold of the snapshot; which is
     * employed by the lookup discovery service in the test to determine
     * whether or not to take a "snapshot" of the system state.
     *
     * @param threshold size threshold for taking a snapshot
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         threshold may or may not have been changed successfully.
     *
     * @see com.sun.jini.fiddler.FiddlerAdmin#setLogToSnapshotThreshold
     */
    public void setLogToSnapshotThreshold(int threshold) throws RemoteException
    {
        server.setLogToSnapshotThreshold(threshold);
    }

    /**
     * Retrieve the value of the size threshold of the snapshot; which is
     * employed by the lookup discovery service in the test to determine
     * whether or not to take a "snapshot" of the system state.
     * 
     * @return int value corresponding to the size threshold of the snapshot
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see com.sun.jini.fiddler.FiddlerAdmin#getLogToSnapshotThreshold
     */
    public int getLogToSnapshotThreshold() throws RemoteException {
        return server.getLogToSnapshotThreshold();
    }

    /**
     * Changes the value of the <code>com.sun.jini.fiddler.server.debug</code>
     * system property. The value of this property configures the debug
     * facility of the fiddler implementation of the lookup discovery service
     * to write debug information (to a file or standard output) selectively;
     * depending on the values the debug property is given through this
     * method.
     *
     * @param newValues <code>String</code> consisting of a comma-separated set
     *                  of  valid debug property values
     *
     * @return <code>boolean</code> value that indicates whether or not the
     *         debug property was successfully set
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         debug property may or may not have been changed successfully.
     *
     * @see com.sun.jini.fiddler.FiddlerAdmin#setDebugProperty
     */
    public boolean setDebugProperty(String newValues) throws RemoteException {
	return server.setDebugProperty(newValues);
    }

    /**
     * Returns the current value of the system property used to configure the
     * debug facility of the fiddler implementation of the lookup discovery
     * service (<code>com.sun.jini.fiddler.server.debug</code>).
     *
     * @return <code>String</code> consisting of a comma-separated set
     *         of valid debug property values
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see com.sun.jini.fiddler.FiddlerAdmin#setDebugProperty
     * @see com.sun.jini.fiddler.FiddlerAdmin#getDebugProperty
     */
    public String getDebugProperty() throws RemoteException {
	return server.getDebugProperty();
    }

    /* *** Methods of net.jini.admin.JoinAdmin *** */

    /** 
     * Get the current attribute sets for the lookup discovery service. 
     * 
     * @return array of net.jini.core.entry.Entry containing the current
     *         attribute sets for the lookup discovery service
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see net.jini.admin.JoinAdmin#getLookupAttributes
     */
    public Entry[] getLookupAttributes() throws RemoteException {
	return server.getLookupAttributes();
    }

    /** 
     * Add attribute sets to the current set of attributes associated
     * with the lookup discovery service. The resulting set will be used
     * for all future registrations with lookup services. The new attribute
     * sets are also added to the lookup discovery service's attributes
     * on each lookup service with which the lookup discovery service
     * is currently registered.
     *
     * @param  attrSets array of net.jini.core.entry.Entry containing the
     *         attribute sets to add
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         attributes may or may not have been added successfully.
     *
     * @see net.jini.admin.JoinAdmin#addLookupAttributes
     */
    public void addLookupAttributes(Entry[] attrSets) throws RemoteException {
	server.addLookupAttributes(attrSets);
    }

    /** 
     * Modify the current set of attributes associated with the lookup
     * discovery service. The resulting set will be used for all future
     * registrations with lookup services. The same modifications are 
     * also made to the lookup discovery service's attributes on each
     * lookup service with which the lookup discovery service is currently
     * registered.
     *
     * @param  attrSetTemplates  array of net.jini.core.entry.Entry containing
     *         the templates for matching attribute sets
     * @param  attrSets array of net.jini.core.entry.Entry containing the
     *         modifications to make to matching sets
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         attributes may or may not have been modified successfully.
     *
     * @see net.jini.admin.JoinAdmin#modifyLookupAttributes
     */
    public void modifyLookupAttributes(Entry[] attrSetTemplates,
				       Entry[] attrSets)
	throws RemoteException
    {
	server.modifyLookupAttributes(attrSetTemplates, attrSets);
    }

    /** 
     * Get the names of the groups whose members are lookup services the
     * lookup discovery services wishes to register with (join).
     * 
     * @return String array containing the names of the groups whose members
     *         are lookup services the lookup discovery service wishes to
     *         join.
     * <p>
     *         If the array returned is empty, the lookup discovery service
     *         is configured to join no groups. If null is returned, the
     *         lookup discovery service is configured to join all groups.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see net.jini.admin.JoinAdmin#getLookupGroups
     */
    public String[] getLookupGroups() throws RemoteException {
	return server.getLookupGroups();
    }

    /** 
     * Add new names to the set consisting of the names of groups whose
     * members are lookup services the lookup discovery service wishes
     * to register with (join). Any lookup services belonging to the
     * new groups that the lookup discovery service has not yet registered
     * with, will be discovered and joined.
     *
     * @param  String array containing the names of the groups to add
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         group names may or may not have been added successfully.
     *
     * @see net.jini.admin.JoinAdmin#addLookupGroups
     */
    public void addLookupGroups(String[] groups) throws RemoteException {
	server.addLookupGroups(groups);
    }

    /** 
     * Remove a set of group names from lookup discovery service's managed
     * set of groups (the set consisting of the names of groups whose
     * members are lookup services the lookup discovery service wishes
     * to join). Any leases granted to the lookup discovery service by
     * lookup services that are not members of the groups whose names 
     * remain in the managed set will be cancelled at those lookup services.
     *
     * @param  String array containing the names of the groups to remove
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         group names may or may not have been removed successfully.
     *
     * @see net.jini.admin.JoinAdmin#removeLookupGroups
     */
    public void removeLookupGroups(String[] groups) throws RemoteException {
	server.removeLookupGroups(groups);
    }

    /** 
     * Replace the lookup discovery service's managed set of groups with a
     * new set of group names. Any leases granted to the lookup discovery
     * service by lookup services that are not members of the groups whose
     * names are in the new managed set will be cancelled at those lookup
     * services. Lookup services that are members of groups reflected in
     * the new managed set will be discovered and joined.
     *
     * @param  String array containing the names of the new groups
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         group names may or may not have been replaced successfully.
     *
     * @see net.jini.admin.JoinAdmin#setLookupGroups
     */
    public void setLookupGroups(String[] groups) throws RemoteException {
	server.setLookupGroups(groups);
    }

    /** 
     * Get the lookup discovery service's managed set of locators. The
     * managed set of locators is the set of LookupLocator objects
     * corresponding to the specific lookup services with which the lookup
     * discovery service wishes to register (join).
     * 
     * @return array of objects of type net.jini.core.discovery.LookupLocator,
     *         each of which corresponds to a specific lookup service the
     *         lookup discovery service wishes to join.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see net.jini.admin.JoinAdmin#getLookupLocators
     */
    public LookupLocator[] getLookupLocators() throws RemoteException {
	return server.getLookupLocators();
    }

    /** 
     * Add a set of LookupLocator objects to the lookup discovery service's
     * managed set of locators. The managed set of locators is the set of
     * LookupLocator objects corresponding to the specific lookup services
     * with which the lookup discovery service wishes to register (join).
     * <p>
     * Any lookup services corresponding to the new locators that the lookup
     * discovery service has not yet joined, will be discovered and joined.
     *
     * @param  array of net.jini.core.discovery.LookupLocator objects to add
     *         to the managed set of locators
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         new locators may or may not have been added successfully.
     *
     * @see net.jini.admin.JoinAdmin#addLookupLocators
     */
    public void addLookupLocators(LookupLocator[] locators)
	throws RemoteException
    {
	server.addLookupLocators(locators);
    }

    /** 
     * Remove a set of LookupLocator objects from the lookup discovery
     * service's managed set of locators. The managed set of locators is the
     * set of LookupLocator objects corresponding to the specific lookup
     * services with which the lookup discovery service wishes to register
     * (join).
     * <p>
     * Note that any leases granted to the lookup discovery service by
     * lookup services that do not correspond to any of the locators
     * remaining in the managed set will be cancelled at those lookup
     * services.
     *
     * @param  array of net.jini.core.discovery.LookupLocator objects to
     *         remove from the managed set of locators
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         new locators may or may not have been removed successfully.
     *
     * @see net.jini.admin.JoinAdmin#removeLookupLocators
     */
    public void removeLookupLocators(LookupLocator[] locators)
	throws RemoteException
    {
	server.removeLookupLocators(locators);
    }

    /** 
     * Replace the lookup discovery service's managed set of locators with
     * a new set of locators. The managed set of locators is the set of
     * LookupLocator objects corresponding to the specific lookup services
     * with which the lookup discovery service wishes to register (join).
     * <p>
     * Note that any leases granted to the lookup discovery service by
     * lookup services whose corresponding locator is removed from the
     * managed setwill be cancelled at those lookup services. The lookup
     * services corresponding to the new locators in the managed set
     * will be discovered and joined.
     *
     * @param  array of net.jini.core.discovery.LookupLocator objects with
     *         which to replace the current managed set of locators
     *         remove from the managed set of locators
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         locators in the managed set may or may not have been replaced
     *         successfully.
     *
     * @see net.jini.admin.JoinAdmin#setLookupLocators
     */
    public void setLookupLocators(LookupLocator[] locators)
	throws RemoteException
    {
	server.setLookupLocators(locators);
    }

    /* *** Methods of com.sun.jini.admin.StorageLocationAdmin *** */

    /**
     * Returns the location of the lookup discovery service's persistent
     * storage. Typically returns a directory pathname unless the lookup
     * discovery service specifies otherwise.
     * 
     * @return String that contains the current location of the lookup
     *         discovery service's persistent storage.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service.
     *
     * @see com.sun.jini.admin.StorageLocationAdmin#getStorageLocation
     */
    public String getStorageLocation() throws RemoteException {
        return server.getStorageLocation();
    }

    /**
     * Sets the location of the lookup discovery service's persistent
     * storage, moving all current persistent storage from the current
     * location to the specified new location.
     * 
     * @param String that contains the new location of the lookup discovery
     *         service's persistent storage.
     *
     * @throws java.io.IOException typically, this exception occurs when
     *         there is a failure while moving the persistent storage
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         new location for the lookup discovery service's persistent
     *         storage may or may not have been replaced successfully.
     *
     * @see com.sun.jini.admin.StorageLocationAdmin#getStorageLocation
     */
    public void setStorageLocation(String location)
                                   throws IOException, RemoteException {
        server.setStorageLocation(location);
    }

    /* *** Methods of com.sun.jini.admin.DestroyAdmin *** */

    /**
     * Destroy the lookup discovery service, if possible, including its
     * persistent storage. This method will typically spawn a separate
     * thread to do the actual work asynchronously, so a successful
     * return from this method usually does not mean that the service
     * has been destroyed.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         lookup discovery service. When this exception does occur, the
     *         lookup discovery service may or may not have been successfully
     *         destroyed.
     *
     * @see com.sun.jini.admin.DestroyAdmin#destroy
     */
    public void destroy() throws RemoteException {
	server.destroy();
    }

    /**
     * This method is intended to allow for the simulation of a system
     * crash by exiting the virtual machine (VM) in which the current 
     * instance of the service is running. No clean up is performed
     * (just as no clean up would be performed in a real system crash).
     * That is, unlike the <code>destroy</code> method of the
     * <code>DestroyAdmin</code> interface, no threads are terminated and
     * the storage locations in which this service's persistence state is
     * logged (the "log files") are not removed.
     * <p>
     * This method should be used to test the persistent state recovery 
     * mechanism of the service. As such, it should only be invoked on
     * versions of the service which actually persist their state; that is,
     * it should not be invoked on versions of the service which are 
     * typically referred to as "transient". If the service is activatable
     * and "restartable", the activation daemon (RMID) will automatically
     * restart the service after the "crash". If the service is activatable
     * but not "restartable", then RMID will restart the service upon the
     * first invocation of any remote method of the service. When such a
     * service is restarted, it typically attempts to recover any state
     * it may have persisted before the "crash".
     * <p>
     * Note that because calls to this method are asynchronous, certain
     * requirements are imposed on the implementation of the entity that
     * invokes this method.  When an entity invokes this method, it must
     * delay a certain amount time before attempting to again interact with
     * the service. This delay is necessary to allow the service to be
     * re-activated and to recover its persisted state. If the entity 
     * attempts to interact with the service before it has completely
     * restarted and recovered, a <code>RemoteException</code> may result.
     * <p>
     * This method returns a <code>long</code> value that represents the
     * amount of time the client-side entity should delay before attempting
     * to interact with the service again. This time value is only a rough
     * estimate, and may need to be adjusted in the service's backend if
     * the entity experiences <code>RemoteException</code>s on a regular
     * basis when the entity attempts to interact with the service after
     * invoking this method.
     *
     * @return <code>long</code> value that represents an estimate of the
     *         amount of time (in milliseconds) the entity that invokes this
     *         method should delay before attempting to interact with the
     *         service again.
     *                      
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         server. When this exception does occur, the VM of the service
     *         may or may not have been exited successfully.
     *
     * @see com.sun.jini.mahout.KillVMAdmin#killVM
     * @see com.sun.jini.fiddler.FiddlerImpl#killVM
     */
    public long killVM() throws RemoteException {
	return server.killVM();
    }

    /* *** HashCode and Equals for this class *** */

    /** 
     * For any instance of this class, returns the hashcode value generated
     * by the hashCode method of the proxy ID associated with the current
     * instance of this proxy.
     *
     * @return an int value representing the hashcode of the object on
     *         which this method is invoked.
     */
    public int hashCode() {
	return proxyID.hashCode();
    }

    /** 
     * For any instance of this class, indicates whether the object input
     * to this method is equal to the instance; where equality of proxies
     * to a lookup discovery service is defined by equality between
     * the unique identifiers assigned by the lookup discovery service
     * to the respective proxies.
     *
     * @param obj reference to the object that is to be compared to the
     *            object on which this method is invoked.
     *
     * @return true if the object input is equal to the object on which
     *         this method is invoked; false otherwise.
     */
    public boolean equals(Object obj) {
	return (obj instanceof FiddlerAdminProxy &&
		proxyID.equals(((FiddlerAdminProxy)obj).proxyID));
    }
}
