/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.fiddler;

import com.sun.jini.fiddler.Fiddler;
import com.sun.jini.proxy.UUID;

import net.jini.admin.Administrable;
import net.jini.discovery.LookupDiscoveryService;
import net.jini.discovery.LookupDiscoveryRegistration;

import net.jini.core.discovery.LookupLocator;
import net.jini.core.event.RemoteEventListener;

import java.io.Serializable;
import java.rmi.RemoteException;
import java.rmi.MarshalledObject;
import java.security.AccessController;
import java.security.PrivilegedAction;

/**
 * This class is a proxy for a lookup discovery service. Clients only see
 * instances of this class via the LookupDiscoveryService interface (and
 * the FiddlerAdmin interface if needed).
 */
class FiddlerProxy implements Administrable,LookupDiscoveryService,Serializable
{
    private static final long serialVersionUID = 6090476386551900771L;
    /**
     * The reference through which communication occurs between the
     * client-side and the server-side of the lookup discovery service
     *
     * @serial
     */
    private final Fiddler server;
    /** 
     * The unique identifier assigned to the current instance of this 
     * proxy class by the lookup discovery service. This ID is used to
     * determine equality between proxies.
     *
     * @serial
     */
    private final UUID proxyID;

    /**
     * Constructs a new instance of FiddlerProxy.
     *
     * @param server  reference to the server object through which 
     *                communication occurs between the client-side and
     *                server-side of the lookup discovery service
     * @param proxyID the unique identifier assigned by the lookup discovery
     *                service to the current instance of this proxy
     */
    public FiddlerProxy(Fiddler server, UUID proxyID) {
	this.server  = server;
	this.proxyID = proxyID;
    }

    /* From net.jini.admin.Administrable */
    /** 
     * Returns a proxy object through which the lookup discovery service
     * for which the object on which this method is invoked serves as
     * proxy may be administered
     *
     * @return a proxy object through which the lookup discovery service
     *         may be administered.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         server.
     *
     * @see net.jini.Administrable
     */
    public Object getAdmin() throws RemoteException {
        return server.getAdmin();
    }

    /* From net.jini.discovery.LookupDiscoveryService */
    /**
     * Registers with the lookup discovery service. When a client invokes
     * this method, it requests that the lookup discovery service perform
     * discovery processing on its behalf.
     *
     * @param groups        String array, none of whose elements may be null,
     *                      consisting of zero or more names of groups to
     *                      which lookup services to discover belong.
     *                      A null value or an empty array
     *                      (net.jini.discovery.LookupDiscovery.ALL_GROUPS
     *                      or net.jini.discovery.LookupDiscovery.NO_GROUPS)
     *                      are both acceptable.
     *                      
     * @param locators      array of zero or more non-null LookupLocator
     *                      objects, each corresponding to a specific lookup
     *                      service to discover. If either the empty array
     *                      or null is passed to this argument, then no
     *                      locator discovery will be performed for the
     *                      associated registration.
     *                      
     * @param listener      a non-null instance of RemoteEventListener. This 
     *                      argument specifies the entity that will receive
     *                      events notifying the registration that a lookup
     *                      service of interest has been discovered. A 
     *                      non-null value must be passed to this argument,
     *                      otherwise a NullPointerException will be thrown
     *                      and the registration.
     *                      
     * @param handback      null or an instance of MarshalledObject. This
     *                      argument specifies an object that will be 
     *                      included in the notification event that the
     *                      lookup discovery service sends to the registered
     *                      listener.
     *                      
     * @param leaseDuration long value representing the amount of time (in
     *                      milliseconds) for which the resources of the
     *                      lookup discovery service are being requested.
     *                      
     * @return an instance of the LookupDiscoveryRegistration interface.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         server. When this exception does occur, the registration may
     *         or may not have completed successfully.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         null is input to the listener parameter.
     *
     * @see net.jini.discovery.LookupDiscoveryService
     */
    public LookupDiscoveryRegistration register(String[] groups,
                                                LookupLocator[] locators,
                                                RemoteEventListener listener,
                                                MarshalledObject handback,
                                                long leaseDuration)
                                                        throws RemoteException
    {
	return server.register(groups,
                               locators,
                               listener,
                               handback,
                               leaseDuration);
    }

    /** 
     * For any instance of this class, returns the hashcode value generated
     * by the hashCode method of the proxy ID associated with the current
     * instance of this proxy.
     *
     * @return an int value representing the hashcode of the object on
     *         which this method is invoked.
     */
    public int hashCode() {
	return proxyID.hashCode();
    }

    /** 
     * For any instance of this class, indicates whether the object input
     * to this method is equal to the instance;  where equality of proxies
     * to a lookup discovery service is defined by equality between
     * the unique identifiers assigned by the lookup discovery service
     * to the respective proxies.
     *
     * @param obj reference to the object that is to be compared to the
     *            object on which this method is invoked.
     *
     * @return true if the object input is equal to the object on which
     *         this method is invoked; false otherwise.
     */
    public boolean equals(Object obj) {
	return (obj instanceof FiddlerProxy &&
		proxyID.equals(((FiddlerProxy)obj).proxyID));
    }
}
