/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */

package com.sun.jini.lease;

import net.jini.core.lease.*;
import java.util.*;
import java.rmi.RemoteException;  
/**
* A lease renewal manager.  The operations are appropriately synchronized
* for concurrent operation.
*
* @deprecated This class was experimental; as of 1.1Alpha, it has been
*	replaced by {@link net.jini.lease.LeaseRenewalManager}. This class  
*	will be removed in a future release, so please update your
*       code, and refer to the javadoc for 
*       <tt>net.jini.lease.LeaseRenewalManager</tt> for details on 
*       <code>LeaseRenewalManager</code>.
* 
* @see net.jini.lease.LeaseRenewalManager
* @author Zane Pan 
*/

public class LeaseRenewalManager {
    private static final long renewbatchTimeGap =5*60*1000;
    private static final long RENEW_RTT = 1000 * 10;
    private static class Entry implements Comparable {
	// Since the cnt only gets modified in the constructor, it does
	// not need to be synchronized.
	private static int cnt = 0;
	private long endTime;
	public Entry(Lease lease, long expiration, LeaseListener listener) {
	    this.endTime =  lease.getExpiration();
	    this.lease = lease;
	    this.expiration = expiration;
	    this.listener = listener;
	    calcRenew();
	    id = cnt++;
	}
	
	/** create a fake entry for tailMap */
	public Entry(long renew) {
	    this.renew = renew + renewbatchTimeGap;
	    id = Integer.MAX_VALUE;
	}

	public int id;
	public Lease lease;
	public long expiration;
	public LeaseListener listener;
	public long renew;
	private RemoteException ex = null;
	public long getRenewDuration(long now) {
	    if(expiration == Lease.FOREVER || expiration ==Lease.ANY )
		return expiration;
	    else
		return (expiration - now);

	    
	}
	
	/** calculate the renew time for the lease entry */
	public void calcRenew() {
	    calcRenew(System.currentTimeMillis());
	}
	public void calcRenew(long now) {
	    this.ex = null;
	    this.endTime = lease.getExpiration();	    
	    if (expiration != Lease.ANY && expiration <= endTime) {
		renew = Long.MAX_VALUE; // no need for renew anymore
		return;
	    } 

	    long delta = endTime - now;
	    if(delta <= RENEW_RTT ) {
		// we may not have enough time to do the renew, but 
		// try to renew it as soon as possible.
		renew = now;
		return;
	    } else if(delta <= RENEW_RTT*2)
		delta = RENEW_RTT;
	    else if(delta <= RENEW_RTT*8)
		delta /=2;
	    else if (delta <= 1000 * 60 * 60 * 24 * 7)
		delta /= 8;
	    else if (delta <= 1000 * 60 * 60 * 24 * 14)
		delta = 1000 * 60 * 60 * 24;
	    else
		delta = 1000 * 60 * 60 * 24 * 3;
	    renew = endTime - delta;
	
	}
	
	/** call the method if renew fails, it re-calculates the renew time */
	public void delayRenew(RemoteException ex) {
	    this.ex = ex;
	    long delta = endTime - renew;
	    if(delta <= RENEW_RTT )
		return;
	    else if (delta <= RENEW_RTT*3)
		 delta = RENEW_RTT;
	    else if (delta <= 1000 * 60 * 60)
		delta /= 3;
	    else if (delta <= 1000 * 60 * 60 * 24)
		delta = 1000 * 60 * 30;
	    else if (delta <= 1000 * 60 * 60 * 24 * 7)
		delta = 1000 * 60 * 60 * 3;
	    else
		delta = 1000 * 60 * 60 * 8;
	    renew += delta;
	}
	
	/** 
	 *  Used by the treeMap to sort the entry by renew. The first key 
	 *  has the longest renew 
         */
	public int compareTo(Object obj) {
	    if (this == obj)
		return 0;

	    Entry e = (Entry)obj;
	    if(renew == e.renew ) {
		if(id < e.id )
		    return 1;
		else
		    return -1;
	    }
	    if (renew < e.renew)
		return 1;
	    return -1;
	}

	public boolean canBatch(Entry e) {
	    return this.lease.canBatch(e.lease);
	} 

	public LeaseMap createLeaseMap(long now) {
	    return this.lease.createLeaseMap(getRenewDuration(now));
	}
	
    }
    
    /** a TreeMap to store Entry */
    private final TreeMap leases = new TreeMap();
    /** the renew thread */
    private RenewThread renewer = null;
    /** the time for the renewThread to wake up */
    private long minRenew = Long.MAX_VALUE;

    /** Create a manager, initially managing no leases. */
    public LeaseRenewalManager() {

    }

    /**
     * Create a manager, initially managing a single lease.
     * Equivalent to calling the no-parameter constructor and
     * then calling renewUntil(lease, expiration, listener).
     *
     * @param lease the lease to renew
     * @param expiration the desired absolute expiration of the lease,
     * with ANY allowed
     * @param listener the event listener for this lease, or null
     */
    public LeaseRenewalManager(Lease lease, long expiration,
			       LeaseListener listener) 
    {
	addLease(lease, expiration, listener);
    }
    
 
    /** 
     * Add a lease to the managed set managed, renewed for a specific
     * duration relative to now.  Equivalent to calling
     * renewUntil(lease, duration + System.currentTimeMillis(), listener).
     *
     * @param lease the lease to renew
     * @param duration the desired duration of the lease.  An absolute
     * expiration is computed by adding System.currentTimeMillis().
     * @param listener the event listener for this lease, or null
     */
    public void renewFor(Lease lease, long duration, LeaseListener listener) {
	long curTime = System.currentTimeMillis();
	if( duration < Lease.FOREVER - curTime  )// check overflow.
	   duration += curTime;
	else
	   duration = Lease.FOREVER;  
	addLease(lease, duration, listener);
    }


   /** 
     * Add a lease to the managed set, renewed until a specific absolute
     * expiration.  If the lease is already managed, it is first removed,
     * and is then added.  The lease will remain in the managed set until
     * either the desired expiration is reached, or the listener is called,
     * or an explicit remove or cancel is done.  If the desired expiration
     * is earlier than the current actual expiration, the actual expiration
     * is not changed.  If the desired expiration is ANY, then renewals
     * will be made forever with a requested duration of ANY.
     *
     * @param lease the lease to renew
     * @param expiration the desired absolute expiration of the lease,
     * with ANY allowed
     * @param listener the event listener for this lease, or null
     */
    public void renewUntil(Lease lease, 
			   long expiration, 
			   LeaseListener listener) 
    {
	addLease(lease, expiration, listener);

    }
   
    /** 
     * add a lease for the LeaseManager to manager.
     *
     * @param lease the lease object 
     * @param expiration the absolute expiration time of the lease
     * @param LeaseListener When lease expires before the duration, the  
     * leaseExpired method of the listener gets called. If lease renew
     * throws the LeaseDenied exception, the leaseDenied method gets
     * called. If listener is null, no notification method gets called.
     */
    private synchronized void addLease(Lease lease,long expiration,
				      LeaseListener listener) 
    {
	Entry e = findEntryDo(lease);
	if(e != null ) 
	    leases.remove(e);
	addEntry (new Entry(lease, expiration, listener));
    }
        
    /**
     * Return the current desired absolute expiration (which may be UNKNOWN)
     * for the lease.
     *
     * @param lease the lease
     */
    public synchronized long getExpiration(Lease lease)
	throws UnknownLeaseException 
    {
	Entry e = findEntry(lease);
	return e.expiration;
    }

    /** 
     * Change the desired absolute expiration for a lease that is already
     * managed.  If the lease is already managed, equivalent to calling
     * renewUntil(lease, expiration, listener) using the existing listener
     * associated with the lease.
     *
     * @param lease the lease
     * @param expiration the new desired absolute expiration for the lease
     */
    public synchronized void setExpiration(Lease lease, long expiration)
	throws UnknownLeaseException 
    {
	Entry e = findEntry(lease);
	long oldExpiration = e.expiration;
	e.expiration = expiration;
	if( oldExpiration >= expiration)
	    return;

	leases.remove(e);
	e.calcRenew();
	addEntry(e);
    }

    /**
     * Remove a lease from the managed set, and cancel it (only if it was
     * in the managed set).  Even if an exception is thrown from the cancel,
     * the lease will have been removed.
     *
     * @param lease the lease
     */
    public void cancel(Lease lease)
	throws UnknownLeaseException, java.rmi.RemoteException
    {
	Entry e;
	synchronized(this) {
	    e = findEntry(lease);
	    leases.remove(e);
	    calcMinRenew();
	}
	e.lease.cancel();
    }

    /** add an lease Entry to the treeMap */
    private void addEntry(Entry ne) {
	leases.put(ne, ne);
	calcMinRenew();
    }

    /**
     * Remove a lease from the managed set.
     *
     * @param lease the lease
     */
    public synchronized void remove(Lease lease) throws UnknownLeaseException {
	Entry e = findEntry(lease);
	leases.remove(e);
	calcMinRenew();

    }


    /** Remove all leases from the managed set. */
    public synchronized void clear() {
	leases.clear();
	calcMinRenew();

    }

    /** calculate the time for the renewThread to wake up */
    private void calcMinRenew() {
 	long renew = Long.MAX_VALUE;
	long now = System.currentTimeMillis();
	Iterator iter = leases.values().iterator();
	while(iter.hasNext()) {
	    Entry e = (Entry)iter.next();
	    if(e.renew >= e.endTime ) {
		if(now >=e.endTime)
		     iter.remove();
		continue;
	    }
	    if (e.endTime <= now ) {
		//lease has expired and no listener, just remove it.
		if(e.listener == null ) {
		    iter.remove();
		    continue;
		}
	    }
	    
	    if (!canBatch(e)) {
		if (e.renew < renew - RENEW_RTT ) {
		     renew = e.renew;
		} else { 
		    renew -= RENEW_RTT;
		}
	    }
	}

	if(renewer == null && renew < Long.MAX_VALUE )
	    createRenewThread();
	else 
	    if(minRenew > renew || renew ==  Long.MAX_VALUE  ) 
		this.notifyAll();	    
	minRenew = renew;

    }
    
    /**
     * Return true if e can be batched with another entry that expires
     * between e.renew - renewbatchTimeGap and e.renew.
     */
    private boolean canBatch(Entry e) {
	Iterator iter = leases.tailMap(e).values().iterator();
	iter.next(); // skip e itself
	while (iter.hasNext()) {
	    Entry be = (Entry)iter.next();
	    if (e.renew - be.renew > renewbatchTimeGap)
		break;
	    if (e.canBatch(be))
		return true;
	}
	return false;
    }

    /** find a lease entry in the treeMap */
    private Entry findEntry(Lease lease) throws UnknownLeaseException {
	Entry e = findEntryDo(lease);
	if(e != null )
	    return e;
	else
	    throw new UnknownLeaseException();
    }

    private Entry findEntryDo(Lease lease) {
	Iterator iter = leases.values().iterator();
	while(iter.hasNext()) {
	    Entry e = (Entry)iter.next();
	    if (e.lease.equals(lease))
		return e;
	}
	return null;
    }
    
    /** renew a lease in the entry */
    private void renewEntry(Entry e) {
	try {
	    e.lease.renew(e.getRenewDuration(System.currentTimeMillis()));
	    synchronized(this) {
		if(leases.remove(e) != null ){
		    e.calcRenew();
		    addEntry(e);
		}
	    }
	} catch (java.rmi.RemoteException ex) {
		synchronized(this) {
		    if(leases.remove(e) != null) {
			e.delayRenew(ex);
			addEntry(e);
		    }
		}
	} catch (Exception ex) {
	    boolean b;
	    synchronized(this) {
		b = leases.remove(e) != null && (e.listener != null);
		calcMinRenew();
	    }
	    if(b)
		e.listener.notify(
		    new LeaseRenewalEvent(this, e.lease, e.expiration, ex));
	}
    }

    private synchronized void batchDelayRenew(ArrayList bList, RemoteException ex) {
	Iterator iter = bList.iterator();
	while(iter.hasNext()) {
	    Entry e = (Entry)iter.next();
	    if(leases.remove(e) != null) {
		e.delayRenew(ex);
		leases.put(e, e);
	    }
	}	
	calcMinRenew();	
    }
    
    private void updateAfterRenew(LeaseMap batchLeaseMap,
					       Map exMap, List bList)
    {
	ArrayList al = new ArrayList(1);
	synchronized(this) {
	    Iterator iter = bList.iterator();
	    while(iter.hasNext()) {
		Entry e = (Entry)iter.next();
		if( leases.remove(e) == null)
		    continue;

		if( exMap != null && exMap.containsKey(e.lease)) {
		    al.add(e);

		} else {
		    e.calcRenew();
		    leases.put(e, e);
		}
	    }
	    calcMinRenew();
	}
	for(int i=0; i<al.size(); i++) {
	    Entry e = (Entry)al.get(i);
	    if(e.listener != null ) 
		e.listener.notify( new LeaseRenewalEvent(this, e.lease, 
							 e.expiration,
				   (LeaseException)exMap.get(e.lease)));
	    
	}
    }

    private void batchRenew(ArrayList bList) {
	LeaseMap batchLeaseMap = createBatchLeaseMap(bList);
	try {
	    batchLeaseMap.renewAll();
	    updateAfterRenew(batchLeaseMap, null, bList);
	} catch (LeaseMapException ex) {
	    updateAfterRenew(batchLeaseMap, ex.exceptionMap, bList);
	} catch (java.rmi.RemoteException ex) {
	    batchDelayRenew(bList, ex);
	}	
    }

    /** create a batch set for batch renew */
    private ArrayList createBatchList(Entry e) {
	Entry fakeEntry = new Entry(e.renew);
	SortedMap lMap = leases.tailMap(fakeEntry);
	Iterator iter = lMap.values().iterator();
	ArrayList batchList = new ArrayList(1);
	while(iter.hasNext()) {
	    Entry be = (Entry)iter.next();
	    if (e.canBatch(be)) {
		batchList.add(be);
	    }
	}
	return batchList;
    }

    /** create a batch leaseMap for batch renew */
    private LeaseMap createBatchLeaseMap(ArrayList bList) {
	Iterator iter = bList.iterator();
	long now = System.currentTimeMillis();
	LeaseMap batchLeaseMap = ((Entry)iter.next()).createLeaseMap(now);
	while(iter.hasNext()) {
	    Entry e = (Entry)iter.next();
	    batchLeaseMap.put(e.lease, new Long(e.getRenewDuration(now)));
	}
	return batchLeaseMap;
    }


    /** create the renewThread */
    private void createRenewThread() {
	renewer = new RenewThread();
	renewer.start();
    }

    private class RenewThread extends Thread {

	public RenewThread() {
	    super("Renew lease");
	    setDaemon(true);
	}

        public void run() {
	    try {
		while (true) {
		    Entry e;
		    ArrayList bList=null;
		    boolean bExpired = false;
		    synchronized (LeaseRenewalManager.this) {
			/**
			 * minRenew < now can happen when we try to renew
                         * a lease that is going to expire soon. An examples
			 * of this case is that a lease fails to renew the 
			 * several times, then the delayRenew calculates
			 * the new renew time to current time. 
			 */
			long now = System.currentTimeMillis();
			while (minRenew > now ) {
			    //System.out.println("sleep:." + (minRenew-now));
			    if (minRenew == Long.MAX_VALUE) {
				renewer = null;
				//System.out.println("thread exit");
				return;
			    }
			    LeaseRenewalManager.this.wait(minRenew - now);
			    now = System.currentTimeMillis();
			}
			e = (Entry)leases.lastKey();
			bExpired = e.endTime <= now;
			if(bExpired) {
			    leases.remove(e);
			    calcMinRenew();
			} else
			    bList = createBatchList(e);
		    }
		    if(bExpired) {
			if(e.listener != null ) 
			    e.listener.notify(
				      new LeaseRenewalEvent(LeaseRenewalManager.this, 
							    e.lease, e.expiration,
							    e.ex));
		    } else {
			if(bList.size() > 1)
			    batchRenew(bList);
			else
			    renewEntry(e);
		    }
		}
	    } catch (InterruptedException ex) {
	    }
	    
	}
    }
    /** a debug function to display the lease info in the treeMap */
    /*************************************************************
    public  void printRenewTime() {
    	Iterator iter = leases.values().iterator();
	System.out.println("++++++++++++++++++++++++++++++++++++++");
	while(iter.hasNext()) {
	    Entry e = (Entry)iter.next();
	    System.out.println("renew:" + ( e.renew - 5000000) + 
		     "  expiration:" + (e.expiration - 5000000)+
		     " id:" + e.id);
	}
	System.out.println("--------------------------------------");
    }
    ***************************************************************/
}
