/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.lease.landlord;

import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.net.MalformedURLException;
import java.rmi.server.RMIClassLoader;
import net.jini.core.lease.Lease;
import net.jini.core.lease.LeaseDeniedException;

/**
 * Creates <code>LandlordLeases</code> without bringing
 * <code>LandlordLease</code> in to the classpath.  Useful if you
 * don't want to clobber the codebase of annotation of incoming
 * <code>LandlordLease</code>s.  In order for this class to work
 * <code>LandlordLease.Factory</code>, and the classes it depends on
 * must be accessible from the codebase specified by the
 * <code>java.rmi.server.codebase</code> property.
 */
public class CodebaseLeaseFactory implements LandlordLeaseFactory,
    LandlordLeaseAccessor
{
    /**
     * Ref to the factory that actually creates the objects, this 
     * particular one is also implements LandlordLeaseAccessor
     */
    private LandlordLeaseAccessor factory;

    /** Fully qualified class name of the leases we are granting */
    final static private String LEASE_CLASS_NAME = 
	"com.sun.jini.lease.landlord.LandlordLease";

    /** 
     * Create a <code>CodebaseLeaseFactory</code>.
     * @throws MalformedURLException if the value of the
     * <code>java.rmi.server.codebase</code> is not a well formed URL.
     * @throws ClassNotFoundException if
     * <code>LandlordLease.Factory</code> or one of the classes it
     * depends can not be loaded from the
     * <code>java.rmi.server.codebase</code>
     * @throws InstantiationException if an instance of
     * <code>LandlordLease.Factory</code> can not be created.
     */
    public CodebaseLeaseFactory() 
	throws MalformedURLException, ClassNotFoundException, 
	       InstantiationException
    {
	Class factoryClass = 
	    RMIClassLoader.loadClass(
                System.getProperty("java.rmi.server.codebase"),
		LEASE_CLASS_NAME + "$Factory");

	try {
	    factory = (LandlordLeaseAccessor)factoryClass.newInstance();
	} catch (IllegalAccessException e) {
	    // this should never happen
	    throw new InternalError(e.getMessage());
	}
    }

    // purposefully inherit doc comment from supertype
    public Lease newLease(Object cookie, Landlord landlord, long expiration)
        throws LeaseDeniedException
    {
	return ((LandlordLeaseFactory)factory).newLease(cookie, landlord,
							expiration);
    }

    /**
     * Utility to check to see if the passed lease is an instance of
     * of LandlordLease (by class name, not by class loader).
     * @return <code>null</code> if th passed lease is not a LandlordLease,
     * othewise return the lease's class object
     */
    final static private Class sameName(Lease lease) {
	final Class lc = lease.getClass();
	return LEASE_CLASS_NAME.equals(lc.getName())?lc:null;
    }

    // purposefully inherit doc comment from supertype
    public boolean canAccess(Lease lease) {
	if (factory.canAccess(lease))
	    return true;

	// Otherwise, we need to check to see if its the Landlord lease
	// class with a diffrent codebase
	return null != sameName(lease);
    }

    /**
     * Used to find the setExpiration method
     */
    final static private Class[] SET_EXP_ARG_TYPE = new Class[]{Long.class};

    /**
     * Utility function to re-throw the contents of InvocationTargetExceptions,
     * Assumes that the Method object being invoked does not throw any
     * checked exceptions
     */
    final static private void throwContent(InvocationTargetException e) {
	final Throwable t = e.getTargetException();
	if (t instanceof Error) 
	    throw (Error)t;
	else if (t instanceof RuntimeException) 
	    throw (RuntimeException)t;
	else 
	    // This should never happen
	    throwInternalError(t);
    }

    /**
     * Utility method to throw a InternalError when we get exceptions
     * that should never occur
     */
    final static private void throwInternalError(Throwable t) {
	System.err.println("Got an exception that should never occur");
	t.printStackTrace();
	throw new InternalError("This exception should never happen:" + t);
    }

    /**
     * Update the expiration time of the passed lease to reflect a new
     * expiration time.
     * <p>
     * In general this method requires the
     * <code>java.lang.reflect.ReflectPermission
     * suppressAccessChecks</code> permission.  This permission is
     * required if this method is invoked on a
     * <code>LandlordLease</code> who's codebase annotation is not the
     * same as the <code>java.rmi.server.codebase</code>
     * <p>
     * @param lease The lease to be updated
     * @param expiration The new expiration for the lease in milliseconds
     *                   since the beginning of the epoch
     * @throws IllegalArgumentException if <code>canAccess(lease)</code>
     *         would return <code>false</code> 
     */
    public void setExpiration(Lease lease, long expiration) {
	Class lc;

	if (factory.canAccess(lease)) {
	    factory.setExpiration(lease, expiration);
	} else if (null != (lc = sameName(lease))) {
	    // Put on the hip boots
	    try {
		final Method setExp = lc.getDeclaredMethod("setExpiration",
		    SET_EXP_ARG_TYPE);
		setExp.setAccessible(true);
		setExp.invoke(lease, new Object[]{new Long(expiration)});
	    } catch (NoSuchMethodException e) {
		throwInternalError(e);
	    } catch (IllegalAccessException e) {
		throwInternalError(e);
	    } catch (InvocationTargetException e) {
		throwContent(e);
	    }
	} else {
	    throw new IllegalArgumentException("Can not access this type " +
					       "of lease");
	}
    }

    /**
     * Used to find the setExpiration method
     */
    final static private Class[] NO_ARGS = new Class[0];


    /**
     * Return the landlord associated with the passed lease.
     * <p>
     * In general this method requires the
     * <code>java.lang.reflect.ReflectPermission
     * suppressAccessChecks</code> permission.  This permission is
     * required if this method is invoked on a
     * <code>LandlordLease</code> who's codebase annotation is not the
     * same as the <code>java.rmi.server.codebase</code>
     * <p>
     * @param lease The lease who's Landlord is to be returned.
     * @return The <code>Landlord</code> of the passed lease.
     * @throws IllegalArgumentException if <code>canAccess(lease)</code>
     *         would return <code>false</code>
     */
    public Landlord getLandlord(Lease lease) {
	Class lc;

	if (factory.canAccess(lease)) {
	    return factory.getLandlord(lease);
	} else if (null != (lc = sameName(lease))) {
	    try {
		final Method getLandlord = lc.getDeclaredMethod("landlord",
		    NO_ARGS);
		getLandlord.setAccessible(true);
		return (Landlord)getLandlord.invoke(lease, NO_ARGS);
	    } catch (NoSuchMethodException e) {
		throwInternalError(e);
	    } catch (IllegalAccessException e) {
		throwInternalError(e);
	    } catch (InvocationTargetException e) {
		throwContent(e);
	    }
	} else {
	    throw new IllegalArgumentException("Can not access this type " +
					       "of lease");
	}
	
	// We will never get here, make the complier happy
	return null;
    }

    
    /**
     * Return the cookie associated with the passed lease.
     * <p>
     * In general this method requires the
     * <code>java.lang.reflect.ReflectPermission
     * suppressAccessChecks</code> permission.  This permission is
     * required if this method is invoked on a
     * <code>LandlordLease</code> who's codebase annotation is not the
     * same as the <code>java.rmi.server.codebase</code>
     * <p>
     * @param lease The lease who's cookie is to be returned.
     * @return The cookie of the passed lease.
     * @throws IllegalArgumentException if <code>canAccess(lease)</code>
     *         would return <code>false</code>
     */
    public Object getCookie(Lease lease) {
	Class lc;

	if (factory.canAccess(lease)) {
	    return factory.getCookie(lease);
	} else if (null != (lc = sameName(lease))) {
	    // Put on the hip boots
	    try {
		final Method getCookie = lc.getDeclaredMethod("cookie",
		    NO_ARGS);
		getCookie.setAccessible(true);
		return getCookie.invoke(lease, NO_ARGS);
	    } catch (NoSuchMethodException e) {
		throwInternalError(e);
	    } catch (IllegalAccessException e) {
		throwInternalError(e);
	    } catch (InvocationTargetException e) {
		throwContent(e);
	    }
	} else {
	    throw new IllegalArgumentException("Can not access this type " +
					       "of lease");
	}

	// We will never get here, make the complier happy
	return null;
    }
}

