/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.lease.landlord;

import java.io.Serializable; 
import java.rmi.Remote;
import java.rmi.RemoteException;
import net.jini.core.lease.Lease;
import net.jini.core.lease.LeaseDeniedException;
import net.jini.core.lease.UnknownLeaseException;

/** 
 * Interface that owners of leased resources must meet in order to work
 * with the <code>LandlordLease</code> implementation of
 * <code>Lease</code> interface.
 * @see net.jini.core.lease.Lease
 * @see LandlordLease 
 */
public interface Landlord extends Remote {
    /**
     * Called by the lease when its <code>renew</code> method is called.
     * Renews the lease that is associated with the given <code>cookie</code>.
     *
     * @param cookie Associated with the lease when it was created.  This 
     *               object should uniquely identify one lease granted
     *               by this <code>Landlord</code>
     * @param duration
     *               The duration argument passed to the
     *               <code>Lease.renew()</code> call
     * @return The new duration the lease should have 
     */
    public long renew(Object cookie, long duration)
	throws LeaseDeniedException, UnknownLeaseException, RemoteException;
       
    /**
     * Called by the lease when its <code>cancel</code> method is called
     * Cancels the lease that is associated with the given <code>cookie</code>.
     *
     * @param cookie Associated with the lease when it was created. This
     *               object should uniquely identify one lease granted
     *               by this <code>Landlord</code>
     */
    public void cancel(Object cookie)
	throws UnknownLeaseException, RemoteException;

    /**
     * Called by the lease map when its <code>renewAll</code> method is called.
     * Should renew the lease that is associated with each element of 
     * <code>cookies</code>
     *
     * @param cookies An array of objects, each uniquely identifying
     *                a lease granted by this <code>Landlord</code>.
     * @param durations
     *                An array of longs, each representing an a
     *                duration in milliseconds that the client
     *                wants the lease associated with the object
     *                from the corresponding element of
     *                <code>cookies</code> renewed for.
     * @return A RenewResults object that contains the new
     *         duration of each lease that was successfully renewed or
     *         the exception encountered for each lease that could not
     *         be renewed.
     */
    public RenewResults renewAll(Object[] cookies, long[] durations)
	throws RemoteException;
       
    /**
     * Called by the lease map when its <code>cancelAll</code> method
     * is called.  Should cancel the lease that is associated with each
     * element of <code>cookies</code>
     * 
     * @param cookies An array of objects, each uniquely identifying
     *                a lease granted by this <code>Landlord</code>.
     * @return If all the leases specified in the <code>cookies</code>
     *         could be cancelled return <code>null</code>.  Otherwise,
     *         return a <code>Map</code> that for each failed cancel
     *         attempt maps the corresponding cookie object to an
     *         exception describing the failure.  
     */
    public java.util.Map cancelAll(Object[] cookie) throws RemoteException;

    /** 
     * Simple class that holds return values of
     * the <code>Landlord.renewAll</code> method.
     *
     * @see Landlord#renewAll
     */
    public class RenewResults implements java.io.Serializable {
	static final long serialVersionUID = 5053579718775268824L;

	/**
	 * For each cookie passed to <code>renewAll</code>,
	 * <code>granted[i]</code> is the granted lease time, or -1 if the
	 * renewal for that least generated an exception.  If there was
	 * an exception, the exception is held in <code>denied</code>.
	 *
	 * @see #denied
	 * @serial
	 */
	public long[] granted;

	/**
	 * The <code>i</code><sup><i>th</i></sup> -1 in <code>granted</code>
	 * was denied because of <code>denied[i]</code>.  If nothing was denied,
	 * this field is <code>null</code>.
	 *
	 * @serial
	 */
	public Exception[] denied;

	/**
	 * Create a <code>RenewResults</code> object setting the field
	 * <code>granted</code> to the passed value, and <code>denied</code>
	 * to <code>null</code>.
	 *
	 * @param granted	The value for the field <code>granted</code>
	 */
	public RenewResults(long[] granted) {
	    this(granted, null);
	}

	/**
	 * Create a <code>RenewResults</code> object setting the field
	 * <code>granted</code> and <code>denied</code> fields to the
	 * passed values.
	 *
	 * @param granted	The value for the field <code>granted</code>
	 * @param denied	The value for the field <code>denied</code>
	 */
	public RenewResults(long[] granted, Exception[] denied) {
	    this.granted = granted;
	    this.denied = denied;
	}
    }
}
