/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.lease.landlord;

import net.jini.core.lease.*;
import com.sun.jini.lease.AbstractLeaseMap;

import java.rmi.RemoteException;
import java.util.Map;
import java.util.Iterator;
import java.util.HashMap;
import java.util.ConcurrentModificationException;

/**
 * Implementation of <code>LeaseMap</code> for <code>LandlordLease</code>.
 *
 * @see LandlordLease
 * @see net.jini.core.lease.LeaseMap
 */
public class LandlordLeaseMap extends AbstractLeaseMap {
    static final long serialVersionUID = -5754145801478146770L;

    /**
     * The landlord which this map will talk to.
     *
     * @serial
     */
    private Landlord landlord;

    /**
     * Create a new <code>LandlordLeaseMap</code>.
     * @param landlord Owner of the resource <code>lease</code> is for.
     * @param lease    First lease to be placed in the map.  It is
     *                 assumed that <code>canContainKey(lease)</code>
     *                 is <code>true</code>.  Must work with the 
     *                 landlord protocol.
     * @param duration The duration the lease should be renewed for if 
     *                 <code>renewAll</code> is called.
     */
    LandlordLeaseMap(Landlord landlord, Lease lease, long duration) {
	super(lease, duration);
	this.landlord = landlord;
    }

    // inherit doc comment
    public boolean canContainKey(Object key) {
	if (key instanceof LandlordLease)
	    return landlord.equals(((LandlordLease) key).landlord());
	else
	    System.out.println("map can't contain " + key.getClass().getName());
	return false;
    }

    // inherit doc comment
    public void cancelAll() throws LeaseMapException, RemoteException {
	Object[] cookies = new Object[size()];
	LandlordLease[] leases = new LandlordLease[cookies.length];
	Iterator it = keySet().iterator();
	for (int i = 0; it.hasNext(); i++) {
	    LandlordLease lease = (LandlordLease) it.next();
	    leases[i] = lease;
	    cookies[i] = lease.cookie();
	}

	    final Map rslt = landlord.cancelAll(cookies);

	    if (rslt == null) {
		// Everything worked out, normal return
		return;
	    } else {
		// Some the leases could not be canceled, generate a
		// LeaseMapException

		// translate the cookie->exception map into a
		// lease->exception map

		int origSize = rslt.size();
		for (int i = 0; i < cookies.length; i++) {
		    Object exception = rslt.remove(cookies[i]); // harmless if
		                                                // not in map
		    if (exception != null) {		   // if it was in map
			rslt.put(leases[i], exception);	   // put back as lease
			remove(leases[i]);		   // remove from map
		    }
		}

		if (origSize != rslt.size())	// some cookie wasn't found
		    throw new ConcurrentModificationException();

		throw new LeaseMapException(
		    "Failure canceling one or more leases", rslt);
	    }
    }


    // inherit doc comment
    public void renewAll() throws LeaseMapException, RemoteException {
	Object[] cookies = new Object[size()];
	long[] extensions = new long[cookies.length];
	LandlordLease[] leases = new LandlordLease[cookies.length];
	Iterator it = keySet().iterator();
	for (int i = 0; it.hasNext(); i++) {
	    LandlordLease lease = (LandlordLease) it.next();
	    leases[i] = lease;
	    cookies[i] = lease.cookie();
	    extensions[i] = ((Long) get(lease)).longValue();
	}

	long now = System.currentTimeMillis();
	Landlord.RenewResults results = landlord.renewAll(cookies, extensions);

	Map bad = null;
	int d = 0;
	for (int i = 0; i < cookies.length; i++) {
	    if (results.granted[i] != -1) {
		long newExp = now + results.granted[i];
		if (newExp < 0) // Overflow, set to Long.MAX_VALUE
		    newExp = Long.MAX_VALUE;
		    
		leases[i].setExpiration(newExp);
	    } else {
		if (bad == null) {
		    bad = new HashMap(results.denied.length +
				      results.denied.length / 2);
		}
		Object badTime = remove(leases[i]);	// remove from this map
		if (badTime == null)			// better be in there
		    throw new ConcurrentModificationException();
		bad.put(leases[i], results.denied[d++]);// add to "bad" map
	    }
	}

	if (bad != null)
	    throw new LeaseMapException("renewing", bad);
    }
}
