/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.lease.landlord;

import java.util.Map;
import java.util.ArrayList;

import net.jini.core.lease.LeaseException;
import net.jini.core.lease.UnknownLeaseException;
import com.sun.jini.lease.landlord.Landlord.RenewResults;


/**
 * Static methods useful for implementing landlords.
 */
public class LandlordUtil {
    /**
     * Call <code>landlord.renew()</code> for each object in 
     * <code>cookie[]</code>, passing <code>cookie[i]</code> and
     * <code>durations[i]</code>.  Gather the results and 
     * exceptions into a <code>RenewResults</code> object.
     *
     * @param landlord   A grantor of leases.
     * @param cookies    An array of objects, each uniquely
     *                   identifying a lease granted by <code>landlord</code>.
     * @param durations An array of longs, each representing a
     *                  duration in milliseconds that the client
     *                  wants the lease associated with the object
     *                  from the corresponding element of
     *                  <code>cookies</code> renewed for.
     * @return A <code>RenewResults</code> object that contains the new
     *         duration of each lease that was successfully renewed or
     *         the exception encountered for each lease that could not
     *         be renewed.
     */ 
    public static RenewResults renewAll(LocalLandlord landlord,
					Object[] cookies, long[] durations) 
    {
	final int count = cookies.length;
	final long[] granted = new long[count];
	final ArrayList exceptions = new ArrayList(count);

	for (int i = 0; i < count; i++) {
	    try {
		granted[i] = landlord.renew(cookies[i], durations[i]);
	    } catch (LeaseException le) {
		// Set flag for client-side handling
		granted[i] = -1;
		exceptions.add(le);
	    }
	}

	if(exceptions.size() == 0)  {
	    return new Landlord.RenewResults(granted);
	} else {
	    // Note: Can't just cast exceptions.toArray() to Exception[]
	    final Exception[] es = new Exception[exceptions.size()];
	    return new RenewResults(granted,
                (Exception[])exceptions.toArray(es));
	}
    }

    /**
     * Call <code>landlord.cancel()</code> for each object in 
     * <code>cookies[]</code>, passing <code>cookies[i]</code>.
     *
     * @param landlord   A grantor of leases.
     * @param cookies An array of objects, each uniquely
     *                identifying a lease granted by <code>landlord</code>.
     * @return If no exceptions are encountered, return
     *         <code>null</code>, otherwise, return a Map that 
     *         for each failed <code>cancel()</code> call maps
     *         the passed cookie object to the exception that 
     *         was generated.
     */
    public static Map cancelAll(LocalLandlord landlord, Object[] cookies) {
	final int count = cookies.length;

	Map map = null;
	for (int i = 0; i < count; i++) {
	    try {
		landlord.cancel(cookies[i]);
	    } catch (UnknownLeaseException e) {
		if (map == null)
		    map = new java.util.HashMap();
		map.put(cookies[i], e);
	    }
	}
	
	return map;
    }
}
