/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.lease.landlord;

import net.jini.core.lease.Lease;
import net.jini.core.lease.LeaseDeniedException;

/** 
 * Provides an implementation of LeasePolicy that grants and renews
 * leases based on constant values for maximum and default lease
 * durations.
 *
 * When an instance of the class is created a
 * <code>LeaseManager</code> may be associated with it.  The manager
 * will be notified when a Lease is created or renewed.
 *
 * @see Landlord
 * @see LeasedResource
 * @see LeaseManager
 * @see net.jini.core.lease.Lease 
 */
public class LeaseDurationPolicy implements LeasePolicy {
    final private long maximum;       // Maximum lease or renewal we will grant
    final private long defaultLength; // Default lease or renewal length
    final private Landlord landlord;
    final private LeaseManager mgr;
    final private LandlordLeaseFactory factory;
    /**
     * Create a <code>LeaseDurationPolicy</code> that will handout the
     * leases with the specified maximum and default durations.
     * Leases with a requested duration less than maximum will be
     * granted exactly.
     *
     * @param maximum  The length in milliseconds of the longest lease
     *                 this  manager should handout.
     * @param defaultLength
     *                 The length in milliseconds of lease this
     *                 manager should handout if the client expresses
     *                 no preference.
     * @param landlord The owner of the leased resources being managed
     *                 by this manager.
     * @param mgr      <code>LeaseManager</code> that will be notified
     *                 of lease creations and renewals.  May be <code>
     *                 null</code> 
     * @param factory  Factory that will be used to generate new leases. 
     *                 If <code>null</code> a <code>LandlordLease</code> 
     *                 factory will be used.
     * @exception IllegalArgumentException thrown if
     *            <code>defaultLength</code> and <code>maximum</code>
     *            are not larger than 0 or if <code>landlord</code> is
     * <code>null</code>.  
     */
    public LeaseDurationPolicy(long maximum, long defaultLength, 
        Landlord landlord, LeaseManager mgr, LandlordLeaseFactory factory) 
    {
	if (!(maximum > 0))
	    throw new IllegalArgumentException("LeaseDurationPolicy:" +
	        "maximum lease time must be larger than 0, passed:" + maximum);

	if (!(defaultLength > 0))
	    throw new IllegalArgumentException("LeaseDurationPolicy:" +
	        "default lease lease time must be larger than 0 passed:" +
		 defaultLength);

	if (landlord == null)
	    throw new IllegalArgumentException("LeaseDurationPolicy:" +
	        "must provide a Landlord");

	this.maximum = maximum;
	this.defaultLength = defaultLength;
	this.landlord = landlord;
	this.mgr = mgr;

	if (factory == null) {
	    try {
		final Class factoryClass = 
		    Class.forName(
                        "com.sun.jini.lease.landlord.LandlordLease$Factory");
		factory = (LandlordLeaseFactory)factoryClass.newInstance();
	    } catch (ClassNotFoundException e) {
		// re-throw as a NoClassDefFoundError
		throw new NoClassDefFoundError(e.getMessage());
	    } catch (IllegalAccessException e) {
		// This should never happen, if this happens something
		// very bad is going on. Rethrow as an InternalError
		throw new InternalError(e.getMessage());
	    } catch (InstantiationException e) {
		// This should never happen, if this happens something
		// very bad is going on. Rethrow as an InternalError
		throw new InternalError(e.getMessage());
	    }
	}

	this.factory = factory;
    }

    /**
     * Factory method that returns lease objects for the specified
     * resource. The lease will call the <code>Landlord</code> object
     * passed to the manager's constructor to handle renewal and
     * cancel requests.  The cookie passed to these calls will be the
     * same as the cookie returned by
     * <code>resource.getCookie()</code>.  The duration of the lease
     * will be affected by the requested duration, and the default and
     * maximum parameters passed to the constructor.
     *
     * @param resource The resource the lease is for. 
     *                 <code>resource.setExpiration</code> will be called 
     *                 to initialize the expiration time of the resource.
     * @param requestedDuration
     *                 The length of lease the client wants.
     * @return A lease for the passed resource
     * @exception IllegalArgumentException thrown if requestedDuration
     * is less than 0 and not equal to Lease.ANYLENGTH or Lease.FOREVER. 
     */
    public Lease leaseFor(LeasedResource resource, long requestedDuration)
	throws LeaseDeniedException
    {
	final long now = currentTime();
	final long expiration =
	    calcExpiration(calculateDuration(resource, requestedDuration), 
			   now);	
	final long duration = expiration - now;

	final Lease rslt = 
	    factory.newLease(resource.getCookie(), landlord, expiration);

	resource.setExpiration(expiration);

	if (mgr != null)
	    mgr.register(resource, duration);

	return rslt;
    }

    /**
     * Renews the expiration of the passed resource.  The duration of
     * renewal is based on the requested duration
     * and the default and maximum parameters passed to the constructor.
     * @param resource The resource having its lease renewed.
     *                 <code>resource.setExpiration</code> will be called
     *                 to update the expiration time of the resource.
     * @param requestedDuration 
     *                 The new duration the client wants
     * @return The new duration the lease should have.
     * @exception IllegalArgumentException thrown if requestedDuration
     * is less than 0 and not equal to <code>Lease.ANYLENGTH</code> or
     * <code>Lease.FOREVER</code>.  
     */
    public long renew(LeasedResource resource, long requestedDuration) {
	final long now = currentTime();

	// The new expiration we would grant independent of the
	// current expiration time 
	long newExpiration =
	    calcExpiration(calculateDuration(resource, requestedDuration), 
			   now);	

	// Do not change the expiration time to be sooner that it is
	// now unless the client explicitly asked for it.  If they did
	// ask for a sooner expiration give them exactly what they
	// asked for.

	// The current expiration of the lease
	final long oldExpiration = resource.getExpiration();

	if (oldExpiration > newExpiration) {
	    // need to make sure they asked for it

	    if (requestedDuration == Lease.ANY) {
		// they did not explicitly ask for a sooner
		// expiration, give them at least what they had
		newExpiration = oldExpiration; 
	    } else {
		// The new expiration time they are requesting
		// (possably cliped to the end of the epoch) this is
		// not affected by what we are willing to grant.
		// [Note we know at this point that requestedDuration
		// is >= 0]
		final long requestedExpiration = calcExpiration(
		    requestedDuration, now);
		
		if (requestedExpiration > oldExpiration) {
		    // they made a request for something longer, give
		    // them at least what they had
		    newExpiration = oldExpiration; 
		} else {
		    // They made a request for something shorter,
		    // give them what they at least what they want
		    newExpiration = requestedExpiration;
		}
	    }
	}

	// Now that we finally have decided on a new expiration,
	// calculate the actual duration and update the manager and
	// resource.
	final long duration = newExpiration - now;

	resource.setExpiration(newExpiration);

	if (mgr != null)
	    mgr.renewed(resource, duration, oldExpiration);

	return duration;
    }

    /**
     * Returns <code>true</code> if the lease on the passed resource
     * has not expired yet.
     */
    public boolean ensureCurrent(LeasedResource resource) {
	return resource.getExpiration() > currentTime();
    }

    /**
     * Returns the duration this policy is willing to grant for the
     * passed resource at this time.  The duration actually granted
     * will be shorter if the duration extends pass the end of the
     * epoch.  Must return a positive number.
     * <p>
     * Note the duration returned by this method will be shorter than
     * the final duration granted if the requested duration extends
     * past the current expiration, and duration return by this 
     * method is before the current expiration.
     *     
     * @param resource          The resource having a lease granted
     *                          or renewed
     * @param requestedDuration The duration the client wants 
     * @exception IllegalArgumentException thrown if requestedDuration
     * is less than 0 and not equal to <code>Lease.ANYLENGTH</code> or
     * <code>Lease.FOREVER</code>.  */
    protected long calculateDuration(LeasedResource resource, 
        long requestedDuration) 
    {
	if (requestedDuration == Lease.FOREVER)
	    requestedDuration = Long.MAX_VALUE;
	else if (requestedDuration == Lease.ANY)
	    requestedDuration = defaultLength;
	else if (requestedDuration < 0)
	    throw new 
		IllegalArgumentException("Negative lease duration " +
		    "requested");

	return Math.min(requestedDuration, maximum);
    }

    /**
     * Method that provides some notion of the current time in milliseconds
     * since the beginning of the epoch. Default implementation
     * calls System.currentTimeMillis()
     */
    protected long currentTime() {
	return System.currentTimeMillis();
    }

    /**
     * Calculates an expiration based on the passed time and a request
     * duration. Will clip the expiration to the end of the epoch
     * @param preferredDuration The duration the policy wants to grant.
     * @param now               The current time in milliseconds since
     *                          the beginning of the epoch
     * @return The new expiration time for the lease in milliseconds since 
     *         the beginning of the epoch.
     */
    private long calcExpiration(long preferredDuration, long now)
    {
	long expiration = now + preferredDuration;
	// Any addition of two positive longs is guaranteed to be
	// negative if it overflowed
	if (expiration < 0) {
	    // Not enough time for the duration we want, set to end of
	    // epoch 
	    return Long.MAX_VALUE;
	}

	return expiration;
    }
}
