/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.mahalo;

import com.sun.jini.mahalo.*;
import com.sun.jini.thread.*;

import java.util.*;

/**
 * A <code>Job</code> manages the division of work for a problem
 * whose solution is obtained by assembling partial results to
 * original problem.
 *
 * @author Gary Holness
 */
public abstract class Job {
    private TaskManager pool;
    private int pending = -1;
    Object[] results;
    int[] attempts;
    private Map tasks = new HashMap();  //used to maintain account
					//of the tasks for which
					//the job is responsible

    private static final boolean DEBUG = false;

    /**
     * Create the <code>Job</code> object giving it the
     * <code>TaskManager</code> responsible for the pool of
     * threads which perform the necessary work.
     *
     * @param pool the <code>TaskManager</code> which provides the threads
     */
    public Job(TaskManager pool) {
	this.pool = pool;
    }


    /**
     * Used by a task to do a piece of work and record the
     * number of attempts.
     *
     * @param who The task which is performing the work
     * @param param A parameter used in performing the work
     */
    boolean performWork(TaskManager.Task who, Object param)
        throws JobException
    {
	if (DEBUG)
	    System.out.println("Job: performWork");

	Integer tmp = null;
	
	synchronized (tasks) {
	    tmp = (Integer)tasks.get(who);
	}

	if (tmp == null)
	    throw new UnknownTaskException();

	int rank = tmp.intValue();

	synchronized (attempts) {
	    attempts[rank]++;
	}

	Object result = doWork(who, param);
	if (result == null)
	    return false;

	try {
	    reportDone(who, result);
	} catch (UnknownTaskException e) {
	} catch (PartialResultException e) {
	} catch (JobException e) {
	}

	return true;
    }


    /**
     * Given a <code>TaskManager.Task</code>, this method
     * returns the current number of attempts it has made.
     * 
     * @param who The task for which the number of attempts
     *		  is inquired
     */
    int attempt(TaskManager.Task who) throws JobException {
	Integer tmp = null;

	synchronized(tasks) {
	    tmp = (Integer)tasks.get(who);
	}

	if (tmp == null)
	    throw new UnknownTaskException();

	int rank = tmp.intValue();

	synchronized(attempts)  {
	    return attempts[rank];
	}
    }



    /**
     * The work performed is implemented here.
     * A null return value indicates failure
     * while a non-null return value indicates
     * success and contains the result.
     * 
     * @param who The task performing the work
     * @param param A parameter used to do the work
     *
     */
    abstract Object doWork(TaskManager.Task who, Object param)
	throws JobException;

    /**
     * Create the tasks required to compute all of the
     * <code>PartialResult</code> objects necessary for the
     * solution to the original problem. 
     */
    abstract TaskManager.Task[] createTasks();


    /**
     * Schedules tasks for execution
     */
    public void scheduleTasks() {
	TaskManager.Task[] tmp = createTasks();

	if (tmp != null) {
	    if (DEBUG)
	        System.out.println("Job: scheduleTasks: there are " +
				    tmp.length + " tasks");

	    results = new Object[tmp.length];
	    attempts = new int[tmp.length];
	    setPending(tmp.length);

	    for (int i = 0; i < tmp.length; i++) {

		//Record the position if each
		//task for later use when assembling
		//the partial results

		synchronized(tasks) {
		    tasks.put(tmp[i],new Integer(i));
		    pool.add(tmp[i]);
		    if (DEBUG)
			System.out.println("Job: scheduleTasks: added " +
				tmp[i] + " to the thread pool");

		    attempts[i] = 0;
		}
	    }
	}
    }


    private synchronized void awaitPending(long waitFor) {
	if (pending < 0)
	    return;

	if (pending == 0)
	    return;

	try {
	    if (DEBUG) {
		System.out.println("Job: awaitPending: pending = " + pending);
	        System.out.println("Job: awaitPending: waiting...");
	    }

	    if (waitFor == Long.MAX_VALUE) {
		while (pending > 0) {
		    wait();
		    if (DEBUG)
		        System.out.println("Job: awaitPending: I'm awake");
		}
	    } else {
		//When waiting for a given amount of time,
		//if notified, make sure that the desired
		//wait time has actually transpired.

		long start = System.currentTimeMillis();
		long curr = start;

		while ((pending > 0) && ((curr - start) < waitFor)) {
                    wait(waitFor - (curr - start));
		    curr = System.currentTimeMillis();
		}
	    }
        } catch (InterruptedException ie) {
        }
    }

    private synchronized void setPending(int num) {
        pending = num;

	if (pending <= 0) {
	    if (DEBUG)
	        System.out.println("Job: setPending: notifying on set " +
				    "pending = " + pending);
	    notifyAll();

	    if (DEBUG)
	        System.out.println("Job: setPending: did notifyAll");
	}
    }

    private synchronized void decrementPending() {
	pending--;

	if (pending <= 0) {
	    if (DEBUG)
	        System.out.println("Job: decrementPending: notifying on " +
					"decrement pending = " + pending);
	    notifyAll();

	    if (DEBUG)
	        System.out.println("Job: decrementPending: did notifyAll");
	}
    }


    /**
     * Returns a reference to the <code>TaskManager</code> which
     * supplies the threads used to executed tasks created by
     * this <code>Job</code>
     */
    protected TaskManager getPool() {
	return pool;
    }


    /*
     * Tasks which perform work on behalf of the <code>Job</code>
     * report in that they are done using this method.
     */
    private void reportDone(TaskManager.Task who, Object param)
	throws JobException
    {
	if (DEBUG)
	    System.out.println("Job: reportDone");

	if (param == null)
	    throw new NullPointerException("param must be non-null");

	if (who == null)
	    throw new NullPointerException("task must be non-null");

	Integer position = null;
	
	synchronized(tasks) {
	    position = (Integer) tasks.get(who);
	}

	if (position == null) 
	    throw new UnknownTaskException();

	synchronized(results) {
	    if (results[position.intValue()] == null) {
		if (DEBUG)
		    System.out.println("Job: reportDone: who = " + who +
					" param = " + param);
	        results[position.intValue()] = param;
	        decrementPending();
	    } else {
	        throw new PartialResultException("result already set");
	    }
	}
    }


    /**
     * Check to see if the <code>Job</code> execution has
     * completed.
     *
     * @param waitFor	The amount of time the caller is willing
     *			to wait for the completion status to arrive.
     */
    public boolean isCompleted(long waitFor) throws JobException {
	//If nothing has started, the
	//task could not have completed.
	//Less than zero means initial value
	//and greater than zero means there
	//are outstanding tasks. In each of
	//these cases, the Job is not done.

	awaitPending(waitFor);

	synchronized(this) {
	    if (pending == 0)
		return true;

	    if (pending < 0)
		throw new JobNotStartedException("No jobs started");

	    return false;
	} 
    }


    /**
     * Generate the solution to the original problem.
     * The subclass decides how it computes the final
     * outcome.
     */
    abstract Object computeResult() throws JobException;


    /**
     * Halt all of the work being performed  by
     * the <code>Job</code>
     */
    public void stop() {
	Set s = tasks.keySet();
	Object[] vals = s.toArray();

	//Remove and interrupt all tasks

	for (int i = 0; i < vals.length; i++) {
	    TaskManager.Task t = (TaskManager.Task) vals[i];
	    pool.remove(t);
	}

	//Erase record of tasks, results and the
	//counting mechanism

	tasks = new HashMap();
	setPending(-1);
	results = null;
    }
}
