/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.mahalo;

import java.io.IOException;
import com.sun.jini.mahout.ExecUtilBase;

/**
 * Utility class for starting Mahalo Transaction Managers
 */
public class MahaloExecUtil extends ExecUtilBase {

    /**
     * The fully specified path of the executable jar file for starting mahalo 
     */
    String mahaloJarFile = null;

    /** 
     * Change codebase, classpath, and/or policyFilePath the
     * appropriate <code>com.sun.jini.exec.mahalo.*</code> property has
     * been set
     */
    private void constructorUtil() {
	final String mahaloCodebase = 
	    System.getProperty("com.sun.jini.exec.mahalo.codebase");

	if (codebase == null && mahaloCodebase == null)
	    throw new IllegalStateException("MahaloExecUtil:Must specify " +
	         "a codebase using com.sun.jini.exec.mahalo.codebase or " +
		 "com.sun.jini.exec.codebase");
	
	if (mahaloCodebase != null) codebase = mahaloCodebase;


	mahaloJarFile = 
	    System.getProperty("com.sun.jini.exec.mahalo.class.path");

	if (classpath == null && mahaloJarFile == null)
	    throw new IllegalStateException("MahaloExecUtil:Must specify " +
	         "a classpath using com.sun.jini.exec.mahalo.class.path or " +
		 "com.sun.jini.exec.class.path\n" +
		 "Note: currently the class path is really the name of the\n" +
		 "      executable jar file (mahalo.jar)");
	
	if (mahaloJarFile == null) 
	    mahaloJarFile = classpath;

	final String mahaloPolicy = 
	    System.getProperty("com.sun.jini.exec.mahalo.policy");

	if (policyFilePath == null && mahaloPolicy == null)
	    throw new IllegalStateException("MahaloExecUtil:Must specify " +
	         "a policy file using com.sun.jini.exec.mahalo.policy or " +
		 "com.sun.jini.exec.policy");
	
	if (mahaloPolicy != null) policyFilePath = mahaloPolicy;
    }

    /**
     * Append <code>cmd</code> with approprate string to set the
     * transaction manager's name to <code>name</code>.
     * <code>name</code> may be <code>null</code>.
     */
    private void addName(String name) {
	if (name == null) return;

	cmd += "-Dcom.sun.jini.mahalo.managerName=" + name + " ";
    }

    /**
     * Set <code>cmd</code>
     * @param useRegistry If <code>true</code> setup the service to bind into
     *                    RMI registry.  Otherwise use Jini lookup
     * @param port        Port to find the RMI registry on. Ignored if 
     *                    <code>useRegistry</code> is <code>false</code>.
     * @param groups      Lookup groups the service should bind into. Ignored
     *                    if <code>useRegistry</code> is <code>true</code>.
     * @param name        Name to register the service under.
     */
    private void buildCmd(boolean useRegistry, int port, String groups,
			  String name)
    {
	cmd += "-Djava.security.policy=" + policyFilePath  + " ";

	if (useRegistry) {
	    cmd += "-Dcom.sun.jini.use.registry=yes ";
	    if (port >= 0)
		cmd += "-Dcom.sun.jini.rmiRegistryPort=" + port + " ";
	}

	/* we might want to reinstate the line below on a conditional
           basis sometime in the future so I am leaving it in but commented
	   out. */
	//cmd += "-Djava.class.path=" + classpath + " ";

	addName(name);

	cmd += "-jar " + mahaloJarFile + " " + codebase + " " + 
	    policyFilePath + " " + persistencePath + " " + groups;
    }

    /**
     * Build an MahaloExecUtil for a Transaction Manager that will bind itself
     * into the registry
     *
     * @param port   Registry port to bind to 
     * @param name   The transaction manager's name
     */
    public MahaloExecUtil(int port, String name) throws IOException 
    {
	super("Mahalo");
	constructorUtil();
	buildCmd(true, port, "none", name);
    }

    /**
     * Build an MahaloExecUtil for a Transaction Manager that will bind itself
     * into a lookup service
     *
     * @param groups Lookup group list
     * @param name   The transaction manager's name
     */
    public MahaloExecUtil(String groups, String name) throws IOException 
    {
	super("Mahalo");
	constructorUtil();
	buildCmd(false, -1, groups, name);
    }
}



