/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.mahalo;

import net.jini.core.transaction.*;
import net.jini.core.transaction.server.*;

import com.sun.jini.mahalo.*;
import com.sun.jini.mahalo.log.*;
import com.sun.jini.thread.*;

import java.rmi.RemoteException;

/**
 * An implementation of a <code>com.sun.jini.mahalo.Job</code> which
 * interacts with a set of
 * <code>net.jini.core.transaction.server.TransactionParticipant</code>s
 * to inform them to vote.
 *
 * @see com.sun.jini.mahalo.Job
 * @see com.sun.jini.mahalo.ParticipantTask
 * @see net.jini.core.transaction.Transaction
 * @see net.jini.core.transaction.server.TransactionParticipant
 *
 * @author Gary Holness
 */
public class PrepareJob extends Job implements TransactionConstants {
    ServerTransaction tr;
    ClientLog log;
    ParticipantHandle[] handles;
    int maxtries = 5;
    private static final boolean DEBUG = false;

    /**
     * Constructs an <code>PrepareJob</code>
     *
     *
     * @param tr The <code>Transaction</code> whose participants
     *           will be instructed to vote
     *
     * @param pool The <code>TaskManager</code> which provides the
     *             threads used for interacting with participants.
     *
     * @param log  The <code>ClientLog</code> used for recording
     *             recovery data.
     *
     * @param handles The array of participants which will be contacted
     *                and informed to vote
     *
     * @see com.sun.jini.thread.TaskManager
     * @see com.sun.jini.mahalo.log.ClientLog
     * @see net.jini.core.transaction.server.TransactionParticipant
     */
    public PrepareJob(Transaction tr, TaskManager pool,
		      ClientLog log,
		      ParticipantHandle[] handles) {
	super(pool);

	if (log == null)
	    throw new IllegalArgumentException("PrepareJob: PrepareJob: " +
							"log is null");

	this.log = log;

	if (!(tr instanceof ServerTransaction))
	    throw new IllegalArgumentException("PrepareJob: PrepareJob: " +
					"must be a ServerTransaction");

	this.tr =  (ServerTransaction) tr;

	if (handles == null)
	    throw new IllegalArgumentException("PrepareJob: PrepareJob: " +
					"must have participants");

	if (handles.length == 0)
	    throw new IllegalArgumentException("PrepareJob: PrepareJob: " +
					"must have participants");

	this.handles = handles;
    }


    /**
     * The work to be performed by each <code>TaskManager.Task</code>
     * is provided by the <code>Job</code> that creates it.
     * The work performed by a task belonging to the AbortJob
     * contacts a participant, instructs it to vote and
     * log appropriately.
     *
     * @param who The task performing the work
     *
     * @param param A parameter, of the task's choosing, useful
     *              in performing work.
     *
     * @see com.sun.jini.mahalo.Job
     * @see com.sun.jini.thread.TaskManager.Task
     */
    Object doWork(TaskManager.Task who, Object param) {
	ParticipantHandle handle = (ParticipantHandle)param;
	TransactionParticipant par = null;

        //check if a vote already exists because it was
        //recovered from the log. In this situation,
        //we do not need to log this info since it
        //exists in the log which was used for recovery...
 
	if (DEBUG)
	    System.out.println("PrepareJob: " + handle + " is preparing");

        int vote = handle.getPrepState();
 
        switch (vote) {
            case COMMITTED:
            case NOTCHANGED:
            case ABORTED:
            case PREPARED:
		return new Integer(vote);
        }
 
        //...otherwise, explicitly instruct the participant to
        //prepare after unpacking it and checking against the
        //max retry threashold
 
        if (par == null)
            par = handle.getParticipant();
 
        //If you have exhausted the max retry threashold
        //stop, so that no further attempts are made.
 
	try {
            if (attempt(who) > maxtries) {
	        return new Integer(ABORTED);
            }
	} catch (JobException je) {
	    return null;
	}
 
 
        //At this point, if participant is null, there
        //must be an error unpacking, so retry later
        if (par == null)
            return null;
 
 
        //Here we actually need to ask the participant to
        //prepare.  Note the RemoteException causes a
        //retry. Here we only log info for the cases
        //where a final outcome is available.
 
        Object response = null;
 
        try {
            vote = par.prepare(tr.mgr, tr.id);
            response = new Integer(vote);
        } catch (TransactionException bte) {
            vote = ABORTED;
            response = new Integer(vote);
        } catch (RemoteException re) {
        } catch (RuntimeException rte) {
	    vote = ABORTED;
	    response = new Integer(vote);
	}

        if (response != null) {
	    handle.setPrepState(vote);
            try {
                log.write( new PrepareRecord(handle, vote));
            } catch (com.sun.jini.mahalo.log.LogException le) {
                //the full package name used to disambiguate
                //the LogException
            }

	    return response;
        }
 
	return null;
    }


    /**
     * Creates the <code>TaskManager.Task</code>s necessary to
     * inform participants to vote.
     */
    TaskManager.Task[] createTasks() {
	TaskManager.Task[] tmp = new TaskManager.Task[handles.length];

	for (int i = 0; i < handles.length; i++) {
	    tmp[i] = new ParticipantTask(getPool(), this, handles[i]);
	}

	return tmp;
    }


    /**
     * Gathers partial results submitted by tasks and produces
     * a single outcome.
     *
     * @see com.sun.jini.mahalo.Job
     */
    Object computeResult() throws JobException {
	try {
	    if (!isCompleted(0))
	        throw new ResultNotReadyException("Cannot compute result " +
					"since there are jobs pending");
	} catch (JobNotStartedException jnse) {
	    throw new ResultNotReadyException("Cannot compute result since" +
					   " jobs were not created");
	}

	int prepstate = NOTCHANGED;
	int tmp = 0;

	if (DEBUG)
	    System.out.println("PrepareJob: computeResult: " + results.length +
				" participants prepared");
	checkresults:
	for (int i = 0; i < results.length; i++) {
	    tmp = ((Integer)results[i]).intValue();

	    switch(tmp) {
	      case NOTCHANGED:
		//Does not affect the prepstate
	        break;

	      case ABORTED:
		//Causes all further checks to end
		//while marking ABORTED. A single abort
		//aborts the whole transaction.

		prepstate = ABORTED;
		break checkresults;

	      case PREPARED:
		//changes the state to PREPARED only
		//if currently NOTCHANGED
		if (prepstate == NOTCHANGED)
		    prepstate = PREPARED;
		break;
	    }
	}

	return new Integer(prepstate);
    }
}
