/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.mahalo;

import com.sun.jini.mahalo.log.*;
import net.jini.core.entry.*;

import net.jini.core.transaction.*;
import net.jini.core.transaction.server.*;
import net.jini.core.event.*;
import net.jini.core.lease.*;
import net.jini.lookup.entry.*;

import net.jini.admin.JoinAdmin;
import net.jini.admin.Administrable;
import com.sun.jini.admin.DestroyAdmin;

import com.sun.jini.debug.Debug;

import com.sun.jini.lease.landlord.LeasePolicy;
import com.sun.jini.lease.landlord.Landlord;
import com.sun.jini.lease.landlord.LocalLandlord;
import com.sun.jini.lease.landlord.LandlordUtil;

import com.sun.jini.thread.TaskManager;

import com.sun.jini.mahout.RegistryAdmin;
import com.sun.jini.mahout.KillVMAdmin;
import com.sun.jini.mahout.ParseUtils;
import com.sun.jini.mahout.PersistentServiceDesc;
import com.sun.jini.mahout.BasicServiceDesc;
import com.sun.jini.mahout.RegistryLocator;
import com.sun.jini.mahout.binder.Binder;
import com.sun.jini.mahout.binder.BinderSetup;
import com.sun.jini.mahout.binder.BinderFactory;
import com.sun.jini.mahout.binder.DirectoryAdmin;

import net.jini.core.discovery.LookupLocator;

import java.util.*;
import java.lang.reflect.*;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.io.Serializable;
import java.security.SecureRandom;

import java.net.URL;

import java.rmi.*;
import java.rmi.activation.*;
import java.rmi.activation.ActivationGroupDesc.CommandEnvironment;
import java.rmi.server.*;
import java.rmi.registry.*;

/**
 * An implementation of the Jini Transaction Specification.
 *
 * @author Gary Holness
 * @author Ken Arnold
 */
public class TxnManagerImpl extends RemoteServer
    implements TxnManager, TransactionManager, LeaseExpirationMgr.Expirer,
	       LogRecovery, TxnSettler, com.sun.jini.constants.TimeConstants,
               LocalLandlord
{
    /**
     * @serial
     */
    private LogManager logmgr;

    /* Default tuning parameters for thread pool */
    /* Retrieve values from properties.          */

    private transient int settlerthreads = 150;
    private transient long settlertimeout = 1000 * 15;
    private transient float settlerload = 1.0f;


    private transient int taskthreads = 50;
    private transient long tasktimeout = 1000 * 15;
    private transient float taskload = 3.0f;


    /* Its important here to schedule SettlerTasks on a */
    /* different TaskManager from what is given to      */
    /* TxnManagerTransaction objects.  Tasks on a given */
    /* TaskManager which create Tasks cannot be on the  */
    /* same TaskManager as their child Tasks.		*/

    private transient TaskManager settlerpool;
    private transient TaskManager taskpool;

    private transient Map       txns;

    /** The ID-setting thread that keeps the ID as an approximate time stamp */
    private transient Thread            idThread;

    private transient Vector		unsettledtxns;
    private transient Thread		settleThread;

    /**
     * @serial
     */
    private String logpath;

    /**
     * @serial
     */
    private ActivationID activationID;

    private static final boolean DEBUG = false;

    static final String DEBUG_PROP = "com.sun.jini.mahalo.debug";

    /**
     * Utility for outputting debug messages.
     */
    private static final Debug debug = new Debug(TxnManagerImpl.DEBUG_PROP);

    /**
     * Print writer for startup related messages.
     */
    private static final PrintWriter dbgStartup = debug.getWriter("startup");

    /** The generator for our IDs. */
    private static transient SecureRandom idGen = new SecureRandom();

    /** The buffer for generating IDs. */
    private static transient final byte[] idGenBuf = new byte[8];

    /**
     * @serial
     */
    private final LeasePolicy txnLeasePolicy =
        new com.sun.jini.lease.landlord.LeaseDurationPolicy(
            10 * 365 * DAYS,    // Maximum lease is 10 years
	    1 * MINUTES,        // Default lease is 1 minute
	    this,		// We handle our own leases
	    // Want to aggressively expire leases
	    new LeaseExpirationMgr(this),
	    null		// default lease implementation
	);

    /**
     * Object we use to lock binder and admin
     */
    transient private Object adminLock = new Object();

    /**
     * Flag which is set upon a destroy request. If the this flag
     * flag is true, then binder creation is locked out.
     * This variable is guarded by <code>adminLock</code>.
     */
    private boolean destroyInProgress = false;

    /**
     * @serial
     */
    private Binder binder;

    /**
     * @serial
     */
    private Object adminProxy;

    /**
     * Cache of our activatable reference.
     */
    private TransactionManager proxyRef = null;

    /**
     * Constructs a transaction manager.
     *
     * @param activationID activation ID passed in by the activation daemon.
     *
     * @param data state data needed to re-activate a transaction manager.
     */
    public TxnManagerImpl(ActivationID activationID, MarshalledObject data)
	throws RemoteException, IOException, ClassNotFoundException
    {
	super();

        // Persistent data path
        logpath = (String)data.get();

        // Check path for validity, and create directory
        com.sun.jini.system.FileSystem.ensureDir(logpath);

	this.activationID = activationID;
	proxyRef = 
	    (TransactionManager)Activatable.exportObject(this, activationID, 0);

	init();
    }

    void init() throws RemoteException {
        txns = Collections.synchronizedMap(new HashMap());
	settlerpool = new TaskManager(settlerthreads,
					settlertimeout,settlerload);

	taskpool = new TaskManager(taskthreads, tasktimeout, taskload);

        if (DEBUG)
            System.out.println("creating TxnManagerImpl");

	if (logpath == null)
	    logmgr = new MultiLogManager(this);
	else
	    logmgr = new MultiLogManager(this, logpath);

        try {
            unsettledtxns = new Vector();
            System.out.println("TxnManagerImpl: recovering state...");
            logmgr.recover();
            System.out.println("TxnManagerImpl: ...state recovered");

            settleThread = new Thread() {
                public void run() {
                    settleTxns();
                };
            };
            settleThread.start();
        } catch (LogException le) {
            System.out.println("TxnManagerImpl: unable to recover state");
            le.printStackTrace();
            throw new RemoteException("TxnManagerImpl: unable to recover " +
                                                                "state");
        }

	/*
	 * With SecureRandom, the first ID requires generation of a
	 * secure seed, which can take several seconds.  We do it here
	 * so it doesn't affect the first call's time.  (I tried doing
	 * this in a separate thread so some of the startup would occur
	 * during the roundtrip back the client, but it didn't help
	 * much and this is simpler.)
	 */
	nextID();
    }


    //TransactionManager interface method

    public TransactionManager.Created create(long lease)
        throws LeaseDeniedException
    {
	TxnManagerTransaction txntr = null;

	long tid = nextID();

	if (DEBUG) {
	    System.out.println("TxnManagerImpl: create: " +
	                       "creating TxnManagerTransaction");
	    System.out.println("TxnManagerImpl: create: manager =" +
	                       manager() + ", TID = " + tid);
	}


	if (DEBUG)
	   System.out.println("TxnManagerImpl: create: generating " +
			    "TxnManagerLease for lease = " + lease);

        txntr = new TxnManagerTransaction(manager(), logmgr, tid,
							taskpool, this);
	Lease txnmgrlease = txnLeasePolicy.leaseFor(txntr, lease);

	if (DEBUG)
	    System.out.println("TransactionManagerImpl: create: " +
	                       "created new TxnManagerTransaction");

	Transaction tr = txntr.getTransaction();
	ServerTransaction str = null;

        try {
	    str = serverTransaction(tr);
	    if (DEBUG)
	        System.out.println("TxnManagerImpl: create: " +
	                           "recording the TxnManagerTransaction");

            txns.put(new Long(str.id), txntr);

            if (DEBUG)
                System.out.println("TxnManagerImpl: create: recorded " +
				   "TxnManagerTransaction= " + txntr);


        } catch(Exception e) {
	    System.out.println("TxnManagerImpl: create: problem: "
	                       + e.getMessage());
	    e.printStackTrace();
	    throw new RuntimeException("TxnManagerImpl: create: unable " +
						"to create transaction");
	}


	if (DEBUG)
	    System.out.println("TxnManagerImpl: create: returning Transaction= "
	                       + str + "with id= " + str.id );

        return new TransactionManager.Created(str.id, txnmgrlease);
    }

    public void
        join(long id, TransactionParticipant part, long crashCount)
        throws UnknownTransactionException, CannotJoinException,
	       CrashCountException, RemoteException
    {
	if (DEBUG)
	  System.out.println("TxnManagerImpl: join: crashCount = " +
								crashCount);

        TxnManagerTransaction txntr =
		(TxnManagerTransaction) txns.get(new Long(id));

	if (txntr == null)
	    throw new UnknownTransactionException("unknown transaction");

	txntr.join(part, crashCount);

    }


    public int getState(long id)
        throws UnknownTransactionException
    {
        TxnManagerTransaction txntr =
	        (TxnManagerTransaction) txns.get(new Long(id));

	if (txntr == null)
	    throw new UnknownTransactionException("unknown transaction");

        return txntr.getState();
    }


    public void commit(long id)
        throws UnknownTransactionException, CannotCommitException {
        try {
            commit(id, 0);
        } catch(TimeoutExpiredException tee) {
	    //This exception is swallowed because the
	    //commit with no timeout only schedules a
	    //roll-forward to happen
        }
    }

    public void commit(long id, long waitFor)
        throws UnknownTransactionException, CannotCommitException,
	       TimeoutExpiredException
    {
        //!! No early return when not synchronous

	TxnManagerTransaction txntr =
		(TxnManagerTransaction) txns.get(new Long(id));

	if (DEBUG)
	    System.out.println("TxnManagerImpl: commit: retrieved " +
				"TxnManagerTransaction = " + txntr);

	if (txntr == null)
	    throw new CannotCommitException();

	txntr.commit(waitFor);
        txns.remove(new Long(id));

	if (DEBUG)
	    System.out.println("TxnManagerImpl: commit: " +
	                       "committed transaction(" + id + ")");
    }


    public void abort(long id)
	throws UnknownTransactionException, CannotAbortException
    {
        try {
            abort(id, 0);
        } catch(TimeoutExpiredException tee) {
	    //Swallow this exception because we only want to
	    //schedule a settler task
        }
    }

    public void abort(long id, long waitFor)
        throws UnknownTransactionException, CannotAbortException,
	       TimeoutExpiredException
    {
        try {
            //!! Multi-participants not supported
            //!! No early return when not synchronous


            // At this point, ask the Participants associated
	    // with the Transaction to prepare

	    TxnManagerTransaction txntr = 
	            (TxnManagerTransaction) txns.get(new Long(id));

	    if (DEBUG)
	        System.out.println("TxnManagerImpl: abort: retrieved " +
				      "TxnManagerTransaction = " + txntr);
	    if (txntr == null)
		throw new CannotAbortException();

	    txntr.abort(waitFor);
	    txns.remove(new Long(id));

        } catch (NullPointerException npe) {
            System.out.println("TxnManagerImpl: abort: " + npe.getMessage());
            npe.printStackTrace();
            throw npe;
        }

	if (DEBUG)
            System.out.println("TxnManagerImpl: abort: " +
                               "aborted transaction(" + id + ")");
    }



    //Satisfies the LogRecovery interface so that the
    //TransactionManager can recover it's non-transient
    //state in the face of process failure.

    /**
     *  This method recovers state changes resulting from
     *  committing a transaction.  This re-creates the
     *  internal representation of the transaction.
     * 
     * @param cookie the transaction's ID
     *
     * @param rec the <code>LogRecord</code>
     */
    public void recover(long cookie, LogRecord rec) throws LogException {
	TxnManagerTransaction tmt = enterTMT(cookie);
	TxnLogRecord trec = (TxnLogRecord) rec;
	trec.recover(tmt);
    }


    /**
     * Informs the transaction manager to attempt to
     * settle a given transaction.
     *
     * @param tid the transaction's ID
     */
    public synchronized void noteUnsettledTxn(long tid) {
	unsettledtxns.add(new Long(tid));

	notifyAll();

	if (DEBUG)
	    System.out.println("TxnManagerImpl: noteUnsettledTxn: " +
				"noting tid(" + tid + ")");
    }

    private synchronized void settleTxns() {
	ClientLog log = null;

	if (DEBUG) {
	    System.out.println("TxnManagerImpl: settleTxns: " +
				"settling unsettled transactions...");
	    System.out.println("TxnManagerImpl: settleTxns: " +
				"there are " + unsettledtxns.size());
	}

	int numtxns = 0;
	Long first = null;
	long tid = 0;

	while (true) {
		numtxns = unsettledtxns.size();

	    if (numtxns == 0) {
		try {
		    if (DEBUG)
		        System.out.println("TxnManagerImpl: settleTxns: " +
								"waiting");
		    wait();

		    if (DEBUG)
		        System.out.println("TxnManagerImpl: settleTxns: " +
								"waited");
		} catch (InterruptedException ie) {
		}
		continue;
	    }

	    first = null;

	    first = (Long) unsettledtxns.firstElement();
	    tid = first.longValue();

	    SettlerTask task = new SettlerTask(settlerpool,
					this, tid);
	    settlerpool.add(task);
	    unsettledtxns.remove(first);

            if (DEBUG)
                System.out.println("TxnManagerImpl: " +
                    "settleTxns: added SettlerTask for tid(" + tid + ")");
	}
    }


    //TransactionParticipant interface go here
    //when I implement nested transactions



    /**
     *  Method from <code>TxnManager</code> which produces
     *  a <code>Transaction</code> from its ID.
     *
     * @param id the ID
     *
     * @see net.jini.core.transaction.Transaction
     * @see com.sun.jini.mahalo.TxnManager
     */
    public Transaction getTransaction(long id)
        throws UnknownTransactionException {

	if (DEBUG)
	    System.out.println("TxnManagerImpl: getTransaction(" + id + ")");

        if (id == ((long)-1))
            return null;

        // First consult the hashtable for the Object
        // containing all actions performed under a
        // particular transaction

        TxnManagerTransaction txntr =
	        (TxnManagerTransaction) txns.get(new Long(id));

	if (txntr == null)
	    throw new UnknownTransactionException("unknown transaction");

        Transaction tn = (Transaction) txntr.getTransaction();
        ServerTransaction tr = serverTransaction(tn);

        if (tr == null)
           throw new UnknownTransactionException(
					     "TxnManagerImpl: getTransaction: "
                                             + "unable to find transaction(" +
					     id + ")");
        if (!tr.mgr.equals(this))
            throw new UnknownTransactionException("wrong manager (" + tr.mgr +
						  " instead of " + this + ")");

	if (DEBUG)
	    System.out.println("TxnManagerImpl: getTransaction got " +
				"transaction(" + id + ")");

        return tr;
    }


    /**
     * Requests the renewal of  a lease on a <code>Transaction</code>.
     *
     * @param cookie identifies the leased resouce
     *
     * @param extension requested lease extension
     *
     * @see net.jini.core.lease.Lease
     * @see com.sun.jini.lease.landlord.LeasedResource
     * @see com.sun.jini.lease.landlord.LeaseManager
     */
    public long renew(Object cookie, long extension)
	        throws UnknownLeaseException, LeaseDeniedException
    {
	//The cookie is the unique identifier which
	//distinguishes the resources offered by
	//the grantor.
	//The grantor is the transaction manager
	//and the leased resources offered by it
	//are the transactions.  Thus, the unique
	//identifier or cookie is the transaction id

	TxnManagerTransaction txntr =
	        (TxnManagerTransaction)txns.get(cookie);

	if (txntr == null)
	    throw new UnknownLeaseException();

	// synchronize on the resource so there is not a race condition
	// between renew and expiration
	synchronized (txntr) {
	    if (!txnLeasePolicy.ensureCurrent(txntr))
		throw new UnknownLeaseException("Lease already expired");

	    final long newduration = txnLeasePolicy.renew(txntr, extension);
	    if (DEBUG)
		System.out.println("TxnManagerImpl: renew: Lease(" + cookie +
				   ")" + " = " + newduration);
	    return newduration;
	}
    }

    /**
     * Cancels the lease on a <code>Transaction</code>.
     *
     * @param cookie identifies the leased resouce
     *
     * @see net.jini.core.lease.Lease
     * @see com.sun.jini.lease.landlord.LeasedResource
     * @see com.sun.jini.lease.landlord.LeaseManager
     */
    public void cancel(Object cookie) throws UnknownLeaseException {
	//The cookie is the transaction id

	TxnManagerTransaction txntr =
	        (TxnManagerTransaction)txns.get(cookie);

	if (txntr == null)
	    throw new UnknownLeaseException();

	int state = txntr.getState();

	if (state == ACTIVE) {

	    synchronized (txntr) {
	        txntr.setExpiration(0);	// Mark as done
	    }

	    try {
	        abort(((Long)cookie).longValue());
	    } catch (TransactionException e) {
	        throw new
		    UnknownLeaseException("When canceling abort threw:" +
	        e.getClass().getName() + ":" + e.getLocalizedMessage());
	    }
	}

	if (DEBUG)
	    System.out.println("TxnManagerImpl: cancel:  canceled Lease(" +
				cookie + ")");
    }

    /**
     * Bulk renewal request of leases on <code>Transaction</code>s.
     *
     * @param cookies identifies the leased resouces
     *
     * @param extensions requested lease extensions
     *
     * @see net.jini.core.lease.Lease
     * @see com.sun.jini.lease.landlord.LeasedResource
     * @see com.sun.jini.lease.landlord.LeaseManager
     */
    public Landlord.RenewResults renewAll(Object[] cookies, long[] extensions) {
	return LandlordUtil.renewAll(this, cookies, extensions);
    }


    /**
     * Bulk cancel of leases on <code>Transaction</code>s.
     *
     * @param cookies identifies the leased resouces
     *
     * @see net.jini.core.lease.Lease
     * @see com.sun.jini.lease.landlord.LeasedResource
     * @see com.sun.jini.lease.landlord.LeaseManager
     */
    public Map cancelAll(Object[] cookies) {
	return LandlordUtil.cancelAll(this, cookies);
    }

    // local methods

    /**
     * gets the next available transaction ID.
     */
    static long nextID() {
	long id;
	synchronized (idGen) {
	    do {
		id = 0;
		idGen.nextBytes(idGenBuf);
		for (int i = 0; i < 8; i++)
		    id = (id << 8) | (idGenBuf[i] & 0xFF);
	    } while (id == 0);				// skip flag value
	}
	return id;
    }



    private ServerTransaction serverTransaction(Transaction baseTr)
        throws UnknownTransactionException
    {
        try {
            return (ServerTransaction) baseTr;
        } catch (ClassCastException e) {
            throw new UnknownTransactionException("unexpected transaction type");
        }
    }


    /**
     * Returns a reference to the <code>TransactionManager</code>
     * interface.
     *
     * @see net.jini.core.transaction.server.TransactionManager
     */
    public TransactionManager manager() {
        return (TransactionManager)this;
    }


    private TxnManagerTransaction enterTMT(long cookie) {
	Long key = new Long(cookie);
	TxnManagerTransaction tmt =
	        (TxnManagerTransaction) txns.get(key);

	if (tmt == null) {
	    tmt = new TxnManagerTransaction(this, logmgr, cookie,
						    taskpool, this);
	    noteUnsettledTxn(cookie);
	}

	txns.put(key, tmt);

	return tmt;
    }

    //***********************************************************
    // Admin

    // Methods required by DestroyAdmin

    /**
     * Cleans up and exits the transaction manager.
     */
    public void destroy() {
        // Lock out binder creation attempts
        synchronized (adminLock) {
            destroyInProgress = true;
        }

	Exception failed = null;

	// Stop being an activatable (don't want to come back do we?)
	try {
	    ActivationGroupID grp = ActivationGroup.currentGroupID();
	    ActivationSystem sys = ActivationGroup.getSystem();
	    sys.unregisterGroup(grp);
	} catch (ActivationException e) {
	    failed = e;
	} catch (RemoteException e) {
	    failed = e;
	}
	if (failed != null) {
	    System.err.println("destroy unregistering:");
	    failed.printStackTrace(System.err);
	    System.err.println("destroy continuing");
	}

        try {
            // binder creation should be locked out at this point
            // so we don't need the adminLock here.
            if (binder != null)
                binder.destroy();
        } catch (Exception t) {
            System.err.println("Binder destruction:");
            t.printStackTrace(System.err);
            System.err.println("destroy continuing");
        }


	// Remove persistent store- ask LogManager to clean
	// itself up, then clean up the persistence path.

	MultiLogManagerAdmin logadmin =
	    (MultiLogManagerAdmin) logmgr.getAdmin();
	
	logadmin.destroy();

	try {
	    com.sun.jini.system.FileSystem.destroy(new File(logpath), true);
	} catch (IOException e) {
	    System.err.println("destroying " + logpath);
	    e.printStackTrace();
	    System.err.println("destroy continuing");
	}

	killVM(); // go away
    }

    // Methods required by KillVMAdmin
    final private static long KILL_DELAY = 10;

    /**
     * Exits the transaction manager's VM.
     */
    public long killVM() {
	/*
	 * We have to do this in a separate thread or the remote
	 * system will always get a RemoteException when the server to
	 * which it is talking (this one) dies in the middle of the
	 * call.  So we start up a new thread that sleeps for
	 * KILL_DELAY milliseconds and then shuts down the VM.  This
	 * lets the thread running killVM return before the
	 * denoument.
	 */
	new Thread() {
	    public void run() {
		try {
		    sleep(KILL_DELAY);
		} catch (InterruptedException e) {
		}
		System.exit(0);
	    }
	}.start();
	// We have the client sleep longer than KILL_DELAY so we have a buffer
	// for variegates of thread scheduling, note 100x is an empirically
	// arrived at multiplier.
	return KILL_DELAY * 100;
    }

    /**
     * Registers the transaction manager with the
     * appropriate directory service
     */
    public void advertise() throws IOException { 
        synchronized (adminLock) {
	    getBinder().advertise();
        }
    }

    // Methods required by Administrable

    /**
     * Returns the administration object for the
     * transaction manager.
     */
    public Object getAdmin() {
	synchronized (adminLock) {
	    if (adminProxy == null)
	        adminProxy = 
                    new AdminProxy(this, getBinder().getDirectoryAdmin());
	    return adminProxy;
	}
    }

    /**
     * Utility method to get the binder, creating if necessary
     * assumes adminLock is held
     */
    private Binder getBinder() {
        if (destroyInProgress)
            return null;
        if (binder == null) {
            // We need to get the binder
            try {
                binder = createBinder();
            } catch (Exception e) {
                throw new TxnManagerInternalException(
                        "Trouble creating binder admin: " + e);
            }
        }
        return binder;
    }

    /**
     * Look to see if we have already created a binder, if not
     * create one and log it to file system
     */
    private Binder createBinder()
        throws IOException, ClassNotFoundException, ActivationException
    {
        // have we already created a binder in a previous
        // VM?
        final File logFile = new File(logpath, DEFAULT_NAME + ".binder");
        BinderFactory factory;

        if (logFile.exists()) {
	     if (dbgStartup != null)
	         dbgStartup.println("TxnManagerImpl::createBinder() - "
		     + " reading from " + DEFAULT_NAME + ".binder");
            // Get binder factory from log
            final FileInputStream fis = new FileInputStream(logFile);
            final ObjectInputStream ois = new ObjectInputStream(fis);
            factory = (BinderFactory)ois.readObject();
            ois.close();
        } else {
            // Create the binder builder, get the factory from it
            // and write it to the file system
	     if (dbgStartup != null)
	         dbgStartup.println("TxnManagerImpl::createBinder() - "
		     + " creating " + DEFAULT_NAME + ".binder");
            final ActivationGroupID gid = ActivationGroup.currentGroupID();
            final BinderSetup builder =
                new com.sun.jini.mahout.binder.LookupBinderSetup(
                    null, logpath, null, gid, attributesFor());
	     if (dbgStartup != null)
	         dbgStartup.println("TxnManagerImpl::createBinder() - "
		     + " setting service " + proxyRef);
            builder.setService(proxyRef);

            final FileOutputStream fos = new FileOutputStream(logFile);
            final ObjectOutputStream oos = new ObjectOutputStream(fos);
            factory = builder.getBinderFactory();

	     if (dbgStartup != null)
	         dbgStartup.println("TxnManagerImpl::createBinder() - "
		     + " writing factory ref " + factory);
            oos.writeObject(factory);
            oos.close();
        }

	if (dbgStartup != null)
	    dbgStartup.println("TxnManagerImpl::createBinder() - "
	        + " returning " + factory.get());
        return factory.get();
    }

    // AdminProxy class
    private static class AdminProxy implements JoinAdmin, DestroyAdmin,
	Serializable, KillVMAdmin, RegistryAdmin
    {
	static final long serialVersionUID = 4135648781811608262L;

	/**
	 * @serial
	 */
	final private TxnManager txnMgr;

	/**
	 * @serial
	 */
	final DirectoryAdmin     dirAdmin;

	AdminProxy(TxnManager txnMgr, DirectoryAdmin dirAdmin) {
	    this.txnMgr = txnMgr;
	    this.dirAdmin = dirAdmin;
	}

	public void destroy() throws RemoteException {
	    txnMgr.destroy();
	}

	// Methods required by KillVMAdmin
	public long killVM() throws RemoteException {
	    return txnMgr.killVM();
	}

        // --------------------------------------------------
        //              RegistryAdmin
        // --------------------------------------------------
	public void enableRegistryBinding(String name, int port) 
	    throws IOException 
	{
	    dirAdmin.enableRegistryBinding(name, port);
	}

        // --------------------------------------------------
        //              JoinAdmin
        // --------------------------------------------------
	public Entry[] getLookupAttributes() throws RemoteException {
	    return dirAdmin.getLookupAttributes();
	}

	public void addLookupAttributes(Entry[] attrSets)
	    throws RemoteException
	{
	    dirAdmin.addLookupAttributes(attrSets);
	}

	public void modifyLookupAttributes(Entry[] attrSetTemplates,
					   Entry[] attrSets)
	    throws RemoteException
	{
	    dirAdmin.modifyLookupAttributes(attrSetTemplates, attrSets);
	}

	public String[] getLookupGroups() throws RemoteException {
	    return dirAdmin.getLookupGroups();
	}

	public void addLookupGroups(String[] groups) throws RemoteException {
	    dirAdmin.addLookupGroups(groups);
	}

	public void removeLookupGroups(String[] groups)
	    throws RemoteException
	{
	    dirAdmin.removeLookupGroups(groups);
	}

	public void setLookupGroups(String[] groups) throws RemoteException {
	    dirAdmin.setLookupGroups(groups);
	}

	public LookupLocator[] getLookupLocators() throws RemoteException {
	    return dirAdmin.getLookupLocators();
	}

	public void addLookupLocators(LookupLocator[] locators)
	    throws RemoteException
	{
	    dirAdmin.addLookupLocators(locators);
	}

	public void removeLookupLocators(LookupLocator[] locators)
	    throws RemoteException
	{
	    dirAdmin.removeLookupLocators(locators);
	}

	public void setLookupLocators(LookupLocator[] locators)
	    throws RemoteException
	{
	    dirAdmin.setLookupLocators(locators);
	}
    }

    //***********************************************************
    // Startup

    /**
     * Create the service owned attributes for an Mahalo server
     */
    private static Entry[] attributesFor() {
	final Entry info = new ServiceInfo("Transaction Manager", 
	    "Sun Microsystems, Inc.",  "Sun Microsystems, Inc.", 
	    com.sun.jini.constants.VersionConstants.SERVER_VERSION,
	    "", "");
	
	final Entry type = 
	    new com.sun.jini.lookup.entry.BasicServiceType("Transaction Manager");

	return new Entry[]{info, type};
    }
}
