/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.mahout;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import com.sun.jini.system.FileSystem;

/**
 * Base class for utility classes that start services.  Provides
 * defaults for classpath, codebase, and security policy file path.
 * Generates a unique name for the persistence file or directory for the
 * service.  Handles details of spawning the process and forwarding
 * the output and error streams of the sub-process.
 * <p>
 * Sub-classes need to set the value of <code>cmd</code> field, and
 * should check additional properties for overrides for the values of
 * the classpath, codebase, and security policy file.  
 */
public class ExecUtilBase {
    /**
     * Codebase for service. Initialized with value of
     * <code>com.sun.jini.exec.codebase</code> property.
     */
    protected String codebase;

    /**
     * Classpath for service. Initialized with value of
     * <code>com.sun.jini.exec.class.path</code> property.
     */
    protected String classpath;

    /**
     * Path for service's persistence directory, initialized to a
     * unique value.
     */
    protected String persistencePath;

    /**
     * Command required to start service should be placed here
     */
    protected String cmd;


    /**
     * Security policy filename for service. Initialized with value of
     * <code>com.sun.jini.exec.policy</code> property.
     */
    protected String policyFilePath;

    /** Sub-process of spawned service, valid after <code>exec()</code> call */
    private Process proc;
    
    /** Service name, used for debuging  */
    private String app;

    /** Init codebase, classpath, and policyFilePath from system properties */
    private void constructorUtil() {
	codebase = System.getProperty("com.sun.jini.exec.codebase");
	classpath = System.getProperty("com.sun.jini.exec.class.path");
	policyFilePath = System.getProperty("com.sun.jini.exec.policy");
    }

    /**
     * Create a <code>ExecUtilBase</code>
     * @param app Type of service being started, used for debugging messages
     */
    protected ExecUtilBase(String app) throws IOException {
	this.app = app;
	final File temp = File.createTempFile(app + "Log", null);
	persistencePath = temp.getAbsolutePath();
	temp.delete();

	constructorUtil();

	cmd = System.getProperty("java.home") + File.separator + "bin" + 
	    File.separator + "java ";
    }

    /**
     * Spawn the service
     */
    public Process exec() throws IOException {
	proc = Runtime.getRuntime().exec(cmd);
	return proc;
    }

    /**
     * Attach stdout and stderr to the subprocess
     */
    public void bindOutput() throws IOException {
	bindOutput(System.out, System.err);
    }

    /**
     * Attach the designated streams to the subprocess
     */
    public void bindOutput(OutputStream out, OutputStream err) 
	throws IOException
    {
	Thread outThread = 
	    new Thread(new Dumper(proc.getInputStream(), out, false),
		       "Out Forwarding Thread:" + app);

	Thread errThread = 
	    new Thread(new Dumper(proc.getErrorStream(), err, false),
		       "Err Forwarding Thread:" + app);
	
	outThread.start();
	errThread.start();
    }

    /**
     * Return the name of the persisistence directory created for this
     * service.
     */
    public String persistencePath() {return persistencePath;}

    /**
     * Delete the persisistence directory created for this service.
     */
    public void deletePersistence() throws IOException {
	FileSystem.destroy(new File(persistencePath), true);
    }

    // purposefully inherit doc comment from supertype
    public String toString() {
	return getClass().getName() + ":" + cmd;
    }

    /**
     * If the service is activatable wait for the set-up code to finish,
     * otherwise return immediately
     */
    public void waitForSetup() throws InterruptedException {
	proc.waitFor();
    }

    /** Utility class to forward bytes from one stream to another */
    private class Dumper implements Runnable {
	private final InputStream in;
	private final OutputStream out;
	private final boolean flush;

	/**
	 * Create a new <code>Dumper</code>
	 * @param in The stream to take bytes from
	 * @param out The stream to dump the bytes to
	 * @param flust If true flush out after each write.
	 */
	Dumper(InputStream in, OutputStream out, boolean flush) {
	    this.in = in;
	    this.out = out;
	    this.flush = flush;
	}

	/**
	 * Read a bunch of bytes from the associated
 	 * <code>InputStream</code> and write them to the associated
	 * <code>OutputStream<code>, repeat until the read returns
	 * EOF, or until there is an <code>IOException</code>
	 */
	public void run() {
	    byte[] buf = new byte[4096];
	    try {
		while (true) {
		    final int bytesRead = in.read(buf);
		    if (bytesRead > 0) {
			out.write(buf, 0, bytesRead);
			if (flush) out.flush();
		    } else if (bytesRead < 0)
			break;
		}
	    } catch (IOException e) {
		e.printStackTrace();
	    } finally {
		try {
		    in.close();
		} catch (IOException e) {
		    // Just trying to be nice
		}
	    }
	}
    }
}
