/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.mahout;

import java.io.File;
import java.net.URL;
import java.net.MalformedURLException;
import java.util.Properties;
import java.util.StringTokenizer;

/**
 * Utilites for parsing command line arguments
 */
public class ParseUtils {
    /**
     * Parse a passed string into an array of strings.  The passed
     * string will be parsed as a list of one or more strings
     * separated by commas or white space.  If any of the elements are
     * the string "public" it will converted into "" (the empty
     * string) in the returned array.  If the passed string equals
     * "none" an empty array will be returned.
     *
     * @exception Throws IllegalArgumentException if <code>list</code>
     * has more than one element and one of the elements is the string
     * "none".
     */
    static public String[] extractGroups(String list) {
	String groups[];

	if (list.equals("none")) {
	    groups = new String[0];
	} else {
	    final StringTokenizer st = new StringTokenizer(list, " \t\n\r\f,");
	    groups = new String[st.countTokens()];
	    for (int k = 0; st.hasMoreTokens(); k++) {
		String group = st.nextToken();

		if (group.equals("public"))
		    group = "";
		if (group.equals("none"))
		    throw new IllegalArgumentException("If \"none\" is used " +
                        "it must be only element in list");
		groups[k] = group;
	    }
	}
	return groups;
    }

    private static final int NUMARGS = 3;

    /**
     * Top level parsing routine.  Turns an argument array into a
     * <code>BasicServiceDesc</code>.  The elements are parsed using
     * the following template:
     * <p>
     * codebase policyfile dbdir [group,...] [javaprog] [javaoptions]
     * <p>
     * The classpath element of the <code>BasicServiceDesc</code> will
     * be taken from the value of <code>java.class.path</code>
     * property of the current VM.
     * <p>
     * The codebase must be a sequence of properly formatted URLs.  URLs
     * must end with ".jar", ".zip", or "/".
     * <p>
     * The policyfile must be an absolute path to an existing file.
     * <p>
     * The dbdir must be an absolute path to a non existent
     * file/directory.
     * <p>
     * The group list will be parsed by <code>extractGroups()</code>,
     * if present it must not start with
     * <code>File.separatorChar</code>.  If not PR sent will default to {""}
     * <p>
     * javaprog if present must be an absolute path to an existing file.
     * <p>
     * javaoptions must be a list of strings that being with "-".
     * <p>
     */
    static public BasicServiceDesc persistentParse(String[] args) {
	if (args.length < NUMARGS) {
	    System.out.println("codebase policyfile dbdir [group,...] [javaprog] [javaoptions]");
	    System.out.println("    use \"none\" to specify no groups, \"public\" to specify group \"\"");
	    System.out.println("    groups defaults to \"public\"");
	    return null;
	}

	String classpath = System.getProperty("java.class.path");

	String codebase = args[0];
	String policy = args[1];
	String dbdir = args[2];
	boolean error = false;
	StringTokenizer st = new StringTokenizer(classpath,
						 File.pathSeparator);

	st = new StringTokenizer(codebase);
	for (int i = 0; st.hasMoreTokens(); i++) {
	    String url = st.nextToken();
	    if (!url.endsWith("/") &&
		!url.endsWith(".jar") && !url.endsWith(".zip"))
	    {
		System.out.println("codebase URL missing trailing slash: " +
				   url);
		error = true;
	    }
	    try {
		new URL(url);
	    } catch (MalformedURLException e) {
		System.out.println("codebase contains malformed URL: " + url);
		error = true;
	    }
	}
	File pf = new File(policy);
	if (!pf.exists()) {
	    System.out.println("policy file " + policy + " does not exist");
	    error = true;
	}
	File df = new File(dbdir);
	if (df.exists()) {
	    System.out.println("dbdir " + dbdir +
			       " exists, specify a different path");
	    error = true;
	}
	String[] groups = {""};
	String prog = null;
	Properties props = null;

	int i = NUMARGS;
	int j = NUMARGS;
	if (i < args.length &&
	    !args[i].startsWith("-") &&
	    args[i].indexOf(File.separatorChar) < 0)
	{
	    try {
		groups = extractGroups(args[i]);
	    } catch (IllegalArgumentException e) {
		error = true;
		System.out.println("Error parsing group list:" +
				   e.getMessage());
	    }
	    i++;
	}

	if (i < args.length && !args[i].startsWith("-")) {
	    prog = args[i];
	    i++;
	    File jf = new File(prog);
	    if (!jf.exists() || jf.isDirectory() || !jf.isAbsolute()) {
		System.out.println(
			       "java program " + prog +
			       " must be absolute path to existing file");
		error = true;
	    }
	}
	for ( ; i < args.length; i++) {
	    if (!args[i].startsWith("-") &&
		(j == NUMARGS || (!args[j-1].equals("-classpath") &&
				  !args[j-1].equals("-cp"))))
	    {
		System.out.println("java option " + args[i] +
				   " must start with -");
		error = true;
	    } else if (args[i].startsWith("-D")) {
		if (props == null)
		    props = new Properties();
		int k = args[i].indexOf('=');
		if (k < 0)
		    props.put(args[i].substring(2), "");
		else
		    props.put(args[i].substring(2, k),
			      args[i].substring(k + 1));
	    } else {
		args[j++] = args[i];
	    }
	}

	if (error)
	    return null;

	String[] opts = null;
	if (j > NUMARGS) {
	    opts = new String[j - NUMARGS];
	    System.arraycopy(args, NUMARGS, opts, 0, opts.length);
	}

	return new BasicServiceDesc(classpath, codebase, policy, dbdir, 
				    groups, prog, opts, props);
    }
}

