/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.mahout;

import com.sun.jini.admin.DestroyAdmin;
import com.sun.jini.mahout.RegistryAdmin;
import com.sun.jini.mahout.RegistryLocator;

import java.io.IOException;
import java.rmi.RemoteException;
import java.text.MessageFormat;
import java.util.ResourceBundle;
import java.util.MissingResourceException;

import net.jini.admin.Administrable;
import net.jini.admin.JoinAdmin;
import net.jini.core.entry.Entry;
import net.jini.core.discovery.LookupLocator;
import net.jini.lookup.entry.Name;

/** 
 * This class provides utilities 
 */
public class ServiceConfig {

    /** Fully-qualified package name that, together with a resource item
     *  name, identifies the resource bundle from which to retrieve 
     *  resource information (which can be localized)
     */
    private static final String resourceBundleName = 
	"com.sun.jini.mahout.resources.mahout";

    /** The set of resources used to enable localization */
    private ResourceBundle resources;

    /**
     * Constructs a new instance of this utility class, initializing the
     * information about the resource bundle that is used to display usage
     * and error information.
     */
    public ServiceConfig() {
	resources = ResourceBundle.getBundle(resourceBundleName);
    }//end constructor

    /**
     * Perform service specific configuration.  This method adds a
     * <code>Name</code> attribute to the service's lookup registration.
     * This process also has the side-effect of activating the service.
     * It also binds the the service into the RMI registry,
     * if explicitly requested. This code is shared by the outrigger
     * and mahalo implementations.
     */
    public static void config(Object service,  String serviceName,
                               String nameProp, String defaultName) 
        throws IOException 
    {
        // Get JoinAdmin interface and add name attribute
        final JoinAdmin jadmin = (JoinAdmin)((Administrable)service).getAdmin();
        String name = System.getProperty(nameProp, defaultName);
        jadmin.addLookupAttributes(new Entry[]{new Name(name)});

        // Add an rmiregistry binding if requested
        if (System.getProperty("com.sun.jini.use.registry") != null) {
            // Print warning message that this feature is going away.
            ServiceConfig sc = new ServiceConfig();
            System.err.println(sc.getString("rmiregistry.dprecation", 
                                            serviceName));

            // Enable registry binding
            final RegistryAdmin radmin = (RegistryAdmin)jadmin;
            radmin.enableRegistryBinding(name, RegistryLocator.calcPort());
        } else {
            final String[] groups = jadmin.getLookupGroups();
            final LookupLocator[] locators = jadmin.getLookupLocators();
            if ((groups != null && groups.length == 0) &&
                 locators.length == 0) {
                 ServiceConfig sc = new ServiceConfig();
		 final DestroyAdmin dadmin = (DestroyAdmin)jadmin;
		 try { 
		     dadmin.destroy();
		 } catch (RemoteException re) {
		     System.err.println(sc.getString("admin.destroy.trouble"));
		     re.printStackTrace();
		 }
                 throw new
                     IllegalStateException(sc.getString("binder.nobinding"));
            }
        }


    }

    /**
     * Displays the resource String corresponding to a particular key, but
     * with a given value substituted in the appropriate location in the
     * String.
     *
     * @param resourceKey a non-null case-sensitive String containing the
     *                    suffix identifier that, together with the
     *                    resourceName, make up the resource whose String
     *                    is to be printed.
     * @param val         the value to substitute in the resource String
     *                    before displaying
     */
    private String getString(String resourceKey, String val) {
        String fmt = getString(resourceKey);
        if (fmt == null)
            fmt = "no text found: \"" + resourceKey + "\" {0}";
        return MessageFormat.format(fmt, new String[]{val});
    }//end print

    /**
     * Retrieves a String resource corresponding to a particular key from the
     * resource bundle. If the resource bundle does not yet exist, this
     * method will first load and instantiate the resource bundle.
     *
     * @param resourceKey a non-null case-sensitive String containing the
     *                    suffix identifier that, together with the
     *                    resourceName, make up the resource whose String
     *                    is to be retrieved.
     *
     * @return a non-null String containing the value of the desired resource
     */
    private String getString(String resourceKey) {
	try {
	    return resources.getString(resourceKey);
	} catch (MissingResourceException e) {
	    return null;
	}
    }//end getString

}//end class ServiceConfig
