/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.mahout.binder;

import java.io.IOException;
import net.jini.core.discovery.LookupLocator;
import net.jini.core.lookup.ServiceID;
import net.jini.core.entry.Entry;
import net.jini.discovery.LookupDiscoveryManager;
import net.jini.lookup.ServiceIDListener;
import net.jini.lookup.JoinManager;

/**

 * <code>JoinAdminState</code> abstracts the implementation of
 * <code>RemoteJoinAdminImpl</code>'s store.  
 * <p>
 * Because an activatable <code>RemoteJoinAdminImpl</code> needs to
 * create it activation descriptor before the service object is
 * available; persistent implementations should provide a companion
 * object that is:
 * <ul>
 * <li>Serializable 
 * <p>
 * <li>Provides a method to initialize the state's store with the
 * service object and initial list of attributes.  This method is used
 * after the companion object is created and incorporated into the
 * <code>RemoteJoinAdminImpl</code> activation descriptor.
 * <p>
 * <li>A method that will yield a new <code>JoinAdminState</code>
 * initialized from the store.  This method is called when the
 * <code>RemoteJoinAdminImpl</code> is activated. 
 * <p>
 * </ul>
 * <code>JoinAdminState</code> stores the following data:
 * <ul>
 * <li>An object representing the service to be handed out to clients.
 * Once set this object does not change.  Usually this will be passed
 * in to the constructor or during the setup processes.  It should be
 * returned by the <code>getService()</code> method and used to
 * initialized the <code>JoinManager</code> in
 * <code>startNewJoinManager()</code> method.
 * <p>
 * <li>A list of attributes.  It should be possible to provide an
 * initial list of attributes when the service object is provided.
 * After that each call to <code>updateAttributes()</code> should
 * update the list from the current state of the
 * <code>JoinManager</code> returned from
 * <code>startNewJoinManager()</code>.  When the
 * <code>JoinManager</code> is created it should be initialized using
 * this list.
 * <p>
 * <li>A list of lookup groups.  The list should start empty.  Each
 * call to <code>updateGroups()</code> should update the list from the
 * current state of the <code>JoinManager</code> returned from
 * <code>startNewJoinManager()</code>.  When the
 * <code>JoinManager</code> is created it should be initialized using
 * this list.
 * <p>
 * <li>A list of lookup locators.  The list should start empty.  Each
 * call to <code>updateLookupLocators()</code> should update the list
 * from the current state of the <code>JoinManager</code> returned
 * from <code>startNewJoinManager()</code>.  When the
 * <code>JoinManager</code> is created it should be initialized using
 * this list.
 * <p>
 * <li>A lookup service ID. When creating its first
 * <code>JoinManager</code> the state should register as a
 * <code>ServiceIDListener</code> and record the service id when it
 * become available.  Subsequent <code>JoinManager</code>s be
 * initialized with this id.
 * <p>
 * <li>A port number and service name.  These can be set once by the
 * <code>recordRegistryBindingEnable()</code> method and should be
 * returned by the <code>getPort()</code> and <code>getName()</code>
 * and methods respectively.
 * <p>
 * </ul>
 */
interface JoinAdminState {
    /** Return the service RemoteJoinAdminImpl is working for  */
    public Object getService();

    /** Return the lookup service ID of the service */
    public ServiceID getServiceID();

    /////////////////////////
    // Lookup related state

    /** 
     * Create a <code>LookupDiscoveryManager</code> to manage to find
     * lookup servers. If called more than once within a given VM
     * should return the result of the previous call.
     * <p>
     * If the implementation makes persistence guarantees the
     * <code>LookupDiscoveryManager</code> should be created with
     * lookup locator, and lookup group lists from its store.  
     */
    public LookupDiscoveryManager startNewLookupDiscoveryManager() 
	throws IOException;

    /** 
     * Create a <code>JoinManager</code> to manage the relationship
     * between the service and the Jini lookup services.  If called
     * more than once within a given VM should return the result of
     * the previous call.
     * <p>
     * If the implementation makes persistence guarantees the
     * <code>JoinManager</code> should be created with the attributes
     * and if available the service id from its store.  
     * <p>
     * Result is undefined if called before
     * <code>startNewLookupDiscoveryManager()</code> has been.  @see
     * JoinAdminState#startNewLookupDiscoveryManager
     */
    public JoinManager startNewJoinManager() throws IOException;

    /**
     * Update the state's store with the current lookup locator list.
     * Should get the current list from
     * <code>LookupDiscoveryManager</code> returned by
     * <code>startNewLookupDiscoveryManager()</code>.  Result is
     * undefined if called before
     * <code>startNewLookupDiscoveryManager()</code> has been.
     * @see JoinAdminState#startNewLookupDiscoveryManager 
     */
    public void updateLookupLocators() throws IOException;

    /**
     * Update the state's store with the current lookup group list.
     * Should get the current list from <code>LookupDiscoveryManager</code>
     * returned by <code>startNewLookupDiscoveryManager()</code>.  Result is
     * undefined if called before <code>startNewLookupDiscoveryManager()</code>
     * has been.
     * @see JoinAdminState#startNewLookupDiscoveryManager 
     */
    public void updateGroups() throws IOException;

    /**
     * Update the state's store with the current lookup attribute
     * list.  Should get the current list from
     * <code>JoinManager</code> returned by
     * <code>startNewJoinManager()</code>.  Result is undefined if
     * called before <code>startNewJoinManager</code> has been.
     * @see JoinAdminState#startNewJoinManager 
     */
    public void updateAttributes() throws IOException;

    /////////////////////////
    // Registry related state

    /**
     * Record in the state's persistent store that binding into the
     * RMI registry is enabled under the specified name.  The registry
     * should be found on the specified port.
     * @throws IllegalStateException if registry binding has already been
     * enabled.
     */
    public void recordRegistryBindingEnable(String name, int port) 
	throws IOException;

    /**
     * Return true if registry binding is enabled.
     */
    public boolean isRegistryBindingEnabled();

    /**
     * Returns the port set by <code>recordRegistryBindingEnable()</code>.
     * @throws IllegalStateException if registry binding has  not been
     * enabled.
     */
    public int getPort();

    /**
     * Returns the name set by <code>recordRegistryBindingEnable()</code>.
     * @throws IllegalStateException if registry binding has  not been
     * enabled.
     */
    public String getName();

    /**
     * Destroy the state's persistent store.
     */
    public boolean destroy();
}

