/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */


package com.sun.jini.mahout.binder;
 
import java.util.Map;
import java.rmi.Remote;
import java.rmi.RemoteException;


/**
 * Allows clients of local activatable objects to find them.  
 * <p>
 * Having multiple activatable objects in the same VM is non-trivial
 * if they want to talk to each other directly instead of using RMI.
 * Because the activation system handles construction of all
 * activatable objects there is no well defined way for activatable
 * objects in the same VM to find each other.  There is also no safe
 * way to handle cyclic graphs of activatable objects that need direct
 * references to each other. 
 * <p>
 * The problem of circular references can't be solved in general,
 * however trees of activatable objects can work.  
 * <p>
 * First, all non-root nodes call
 * LocalActivatableObjectTable.register() in each constructor so their
 * parents can find them after activation. The first argument is the
 * RemoteStub to themselves, the second is themselves.  These non-root
 * nodes must implement LocalActivatableObjectTable.Findable().
 * <p>
 * Second, when a non-leaf nodes is setup it gets the RemoteStubs for
 * each of its activatable children and stores them in its persistent
 * store. When a non-leaf node is activated it calls
 * LocalActivatableObjectTable.lookup() to get the actual reference to
 * each of its children.  
 */
public class LocalActivatableObjectTable {
    /**
     * Map from remote stubs to local implementations
     */    
    final private static Map table = new java.util.HashMap();

    /**
     * Interface objects must implement if they are going to register with 
     * LocalActivatableObjectTable
     */
    public interface Findable extends Remote {
	/** 
	 * The implementation of this method should do nothing.  It
	 * simply gives <code>LocalActivatableObjectTable</code> a way
	 * to ensure the object has been activated and its constructor
	 * is finished.
	 */
  	public void activate() throws RemoteException;
    }

    /**
     * Used by activatable objects who's actual implementations need
     * to be found by other activatable objects in the same activation
     * group
     *
     * @param key Is the objects remote stub
     * @param value is the object itself
     * @exception Throws IllegalArgumentException if key or value is
     * <code>null</code> 
     */
    static public void register(Findable key, Object value) throws 
       IllegalArgumentException 
    {
	if (key == null) 
	    throw new IllegalArgumentException("key is null");

	if (value == null) 
	    throw new IllegalArgumentException("value is null");

	synchronized (table) {
	    table.put(key, value);
	}
    }

    /**
     * Used by clients of activatable objects that need the actual
     * object, not just the remote stub.
     * @param key the activatable object's remote stub.
     * @exception Throws IllegalArgumentException if key is
     * <code>null</code> or if key is not a <code>RemoteStub</code>
     */
    static public Object lookup(Findable key) throws
        IllegalArgumentException, RemoteException 
    {
	// lookup() invokes activate() on the key and then looks up
	// the key in the table. Invoking activate() does two things:
	//     Ensures that key will be in the table
	//     Ensures that key's constructor has finished

	// Don't grab the lock here since that would keep key's
	// register from happening
	key.activate();
	synchronized (table) {
	    return table.get(key);
	}
    }
}

