/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.mahout.binder;

import java.io.IOException;
import net.jini.core.discovery.LookupLocator;
import net.jini.core.lookup.ServiceID;
import net.jini.core.entry.Entry;

import net.jini.discovery.DiscoveryGroupManagement;
import net.jini.discovery.LookupDiscoveryManager;
import net.jini.lookup.JoinManager;
import net.jini.lookup.ServiceIDListener;

/**
 * Implementation of JoinAdminState that does not persist its data.  Used by
 * RemoteJoinAdminImpl when it is working with a non-activatable service.
 */
class TransientJoinAdminState implements JoinAdminState {
    /** The service RemoteJoinAdminImpl is working for */
    private Object  service;

    /** 
     * The list of attributes we were passed in at construction, used only to 
     * initialize the <code>JoinManager</code>.
     */
    private Entry[] attributes;

    /** 
     * <code>JoinManager</code> we created for the
     * <code>RemoteJoinAdminImpl</code>. 
     */
    private JoinManager mgr;

    /**
     * <code>LookupDiscoveryManager</code> we created for the
     * <code>RemoteJoinAdminImpl</code>. 
     */
    private LookupDiscoveryManager ldm;

    /**
     * True if binding into the RMI registry is enabled.
     */
    private boolean registryEnabled = false;

    /** 
     * Value of port argument of
     * <code>recordRegistryBindingEnable()</code> call. Valid if
     * <code>registryEnabled</code> is true.  
     */
    private int port;

    /** 
     * Value of name argument of
     * <code>recordRegistryBindingEnable()</code> call. Valid if
     * <code>registryEnabled</code> is true.  
     */
    private String name;

    /**
     * Create a <code>TransientJoinAdminState</code> object.
     * @param service Service that we are binding into the jini
     *                lookup/RMI registry 
     * @param attributes Initial set of lookup attributes.
     */
    TransientJoinAdminState(Object service, Entry[] attributes) {
	this.service = service;
	this.attributes = attributes;
    }

    // Inherit doc comment from super interface
    public LookupDiscoveryManager startNewLookupDiscoveryManager() 
	throws IOException 
    {
	if (ldm == null) {
	    // Create LookupDiscoveryManager
	    try {
		ldm = new LookupDiscoveryManager(
                    DiscoveryGroupManagement.NO_GROUPS, new LookupLocator[0],
		    null);
	    } catch (IOException e) {
		System.err.println("JoinAdmin:Problem starting discovery");
		e.printStackTrace();
		throw new IOException("JoinAdmin: Problem starting " +
				      "discovery:" + e.getLocalizedMessage());  
	    }
	}
	return ldm;
    }

    // Inherit doc comment from super interface
    public JoinManager startNewJoinManager() throws IOException {
	if (ldm == null) 
	    throw new IllegalStateException("JoinAdmin:Called " +
		"startNewJoinManager before startNewLookupDiscoveryManager");

	if (mgr == null) {
	    mgr = new JoinManager(service, attributes, 
				  (ServiceIDListener)null, ldm, null);
	}

	return mgr;
    }

    // Inherit doc comment from super interface
    public Object getService() {return service;}

    // Inherit doc comment from super interface
    public ServiceID getServiceID() {return null;}

    // Inherit doc comment from super interface
    public void updateLookupLocators() {}

    // Inherit doc comment from super interface
    public void updateGroups() {}

    // Inherit doc comment from super interface
    public void updateAttributes() {}

    // Inherit doc comment from super interface
    public boolean destroy() {return true;}


    // Inherit doc comment from super interface
    public void recordRegistryBindingEnable(String name, int port) {
	if (isRegistryBindingEnabled())
	    throw new 
		IllegalStateException("Registry binding already enabled");
	registryEnabled = true;
	this.port = port;
	this.name = name;
    }

    // Inherit doc comment from super interface
    public boolean isRegistryBindingEnabled() {
	return registryEnabled;
    }

    // Inherit doc comment from super interface
    public int getPort() {
	if (!isRegistryBindingEnabled())
	    throw new IllegalStateException("Registry binding not enabled");
	return port;
    }

    // Inherit doc comment from super interface
    public String getName() {
	if (!isRegistryBindingEnabled())
	    throw new IllegalStateException("Registry binding not enabled");
	return name;
    }
}


