/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.mercury;

import com.sun.jini.debug.Debug;
import com.sun.jini.proxy.UUID; 

import java.io.File;
import java.io.FileNotFoundException;
import java.io.InterruptedIOException;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Serializable;
import java.util.NoSuchElementException;

import net.jini.core.event.RemoteEvent;

/**
 * Class that implements the interface for an <tt>EventLog</tt>. 
 * This class encapsulates the details of reading/writing events from/to
 * some underlying persistence mechanism.
 *
 * This class makes certain assumptions. First, the <tt>next</tt> and
 * <tt>remove</tt> methods are intended to be called in pairs. If 
 * <tt>remove</tt> is not called, then subsequent calls to <tt>next</tt> 
 * will attempt to return the same object. Calling <tt>remove</tt> 
 * essentially advances the read pointer to the next object, if any. 
 * Second, if any <tt>IOExceptions</tt> are encountered during the reading
 * or writing of an event the associated read/write pointer is advanced
 * past the offending event. This means that events can be lost if I/O 
 * errors are encountered.
 *
 * @since Jini1.1
 */

/*
Implementation details:

Event state information is maintained by a control log file 
which maintains the current read and write counts/offsets.  This file
is updated whenever next(), add() or remove() is called.  This file used to 
initialize the state of the EventLog, if it exists, during the 
initialization process.

Events are logged into log files that maintain a fixed # of objects
per file.  The writing process adds events to a log file until this limit
is reached. A new log file is then created to handle the next set of events. 
The reading process will delete a log file once all the contained events 
have been successfully processed.  This serves as the garbage 
collection mechanism. Note that the actual number of events per log is a 
user-configurable environmental parameter via the 
com.sun.jini.mercury.eventsPerLog property (default = 10).

This class also employs the services of a StreamPool, which limits the
number of concurrently open file descriptors. This limit is also a user
configurable value via the com.sun.jini.mercury.streamPoolSize property 
(default = 10).

Event state is kept separate from the service's registration state in 
order to keep the event logging implementation as flexible as possible.
*/

class EventLog {

    //
    // Class fields
    //

    /** Debug utility class */
    private static final Debug debug = new Debug(MailboxBackEnd.DEBUG_PROP);

    /** <tt>PrintWriter,</tt> used for log-related debugging messages */
    private static final PrintWriter dbgLogs = debug.getWriter("logs"); 

    /** Size of control data file: 4 longs * 8 bytes per long */
    private static final int CTLBLOCK_LEN = 8 * 4;

    /** File suffix for the control file */
    private static final String CTLFILE_SUFFIX = "ctl";

    /** File suffix for the log files */
    private static final String LOGFILE_SUFFIX = "log";

    /** Default number of events per log file, if not overriden by the user */
    private static final long DEFAULT_EVENTS_PER_LOGFILE = 10L;

    /** Property string for the number of events per log file */
    static final String eventsPerLogProperty = 
        "com.sun.jini.mercury.eventsPerLog"; 

    /** 
     * Maximum number of event objects per log file. The
     * default is used unless overridden by <tt>eventsPerLogProperty</tt>.
     */
    // Should be final but the compiler complains.
    private static /*final*/ long eventsPerLogFile = 0;
    static { 
        try {
            eventsPerLogFile = Long.getLong(eventsPerLogProperty, 
                DEFAULT_EVENTS_PER_LOGFILE).longValue();
	} catch (SecurityException se) {
	    // Couldn't read the property, most likely.
	    // Fall through and use the default.
	}
	if (eventsPerLogFile < 1) 
	    eventsPerLogFile = DEFAULT_EVENTS_PER_LOGFILE;
	if (dbgLogs != null)
	    dbgLogs.println(eventsPerLogProperty + ": " + eventsPerLogFile);
    } 

    /** Default size for the stream pool, if not overriden by the user */
    static final int DEFAULT_STREAM_POOL_SIZE = 10;

    /** Property string for stream pool size limit */
    static final String streamPoolSizeProperty = 
        "com.sun.jini.mercury.streamPoolSize"; 

    /** 
     * Maximum limit for the number of concurrent <tt>LogStream</tt>s
     * that this pool will concurrently manage. The
     * default is used unless overridden by <tt>streamPoolSizeProperty</tt>.
     */
    private static /*final*/ int maxPoolSize = 10; 
    
    /** 
     * The <tt>StreamPool</tt> reference to be used for all 
     * instances of <tt>EventLog</tt>.
     */
    // Should be final but the compiler complains.
    private static /*final*/ StreamPool streamPool;

    static { 
        try {
            maxPoolSize = Integer.getInteger(streamPoolSizeProperty, 
                DEFAULT_STREAM_POOL_SIZE).intValue();
	} catch (SecurityException se) {
	    // Couldn't read the property, most likely.
	    // Fall through and use the default.
	}
	if (maxPoolSize < 1) 
	    maxPoolSize = DEFAULT_STREAM_POOL_SIZE;;
	if (dbgLogs != null)
	    dbgLogs.println(streamPoolSizeProperty + ": " + maxPoolSize);

	streamPool = new StreamPool(maxPoolSize);
    }

    //
    // Object fields
    //

    /** The associated <tt>UUID</tt> for this <tt>EventLog</tt>. */
    private UUID uuid = null;

    /** The current number of written events. */
    private long wcount = 0;

    /** The current number of read events. */
    private long rcount = 0;

    /** The current write offset into the current "write" log. */
    private long wpos = 0;

    /** 
     * The read offset into the current "read" log of the 
     * "last read" object.
     */
    private long rpos = 0;

    /** 
     * The read offset of the next event. This gets updated to
     * <tt>rpos</tt> once <tt>remove</tt> is called (indicating
     * that the last event read was succcessful).
     */
    private long nextReadPos = 0;

    /** The <tt>EventReader</tt> used to retrieve events. */
    private EventReader eventReader;

    /** The <tt>EventWriter</tt> used to store events */
    private EventWriter eventWriter;

    /** The <tt>File</tt> object of the event persitence directory */
    private File logDir;

    /** 
     * The <tt>File</tt> object that will maintain the control data for
     * for this <tt>EventLog</tt>.
     */
    private File controlFile;

    /** The in memory buffer that holds the control data */
    private byte[] ctlbuf = new byte[CTLBLOCK_LEN];

    /** 
     * Flag that is used to determine whether or not this object 
     * has been initialized. 
     */
    private boolean initialized = false;

    /** 
     * Flag that is used to determine whether or not this object 
     * has been closed. 
     */
    private boolean closed = false;

    // TODO (FCS) - turn off for final delivery
    private static final boolean debugState = true;

    /**
     * Simple constructor that takes a <tt>UUID</tt> argument
     * and a <tt>File</tt> argument.  These arguments are simply
     * assigned to the appropriate internal fields. 
     *
     * @exception IllegalArgumentException if any of the arguments are null
     */
    EventLog(UUID uuid, File logDir) {
        if (logDir == null || uuid == null) 
            throw new IllegalArgumentException("Arguments cannot be null");
        this.uuid = uuid;
        this.logDir = logDir;

        if (dbgLogs != null)
            dbgLogs.println("EventLog::EventLog for: " + uuid);
    }

    /**
     * Initializes the state of this <tt>EventLog</tt> object.
     * The required functionality can potentially throw an 
     * <tt>IOException</tt> and has therefore been separated from
     * from object construction.
     *
     * @exception IOException if an I/O error occurs
     */
    void init() throws IOException {

        if (initialized)
            throw new InternalMailboxException(
		"Trying to re-initialize control data "
                + "for: " + uuid);

        try {
            if (!logDir.exists()) // Create log directory if it doesn't exist
                logDir.mkdirs();

            if (!logDir.isDirectory()) // Verify that logDir is a directory
                throw new FileNotFoundException(logDir.toString()
                                  + " is not a directory");

            controlFile = getControlFile();
            if (controlFile.isFile()) { // Recover state from existing file
                if (dbgLogs != null)
                    dbgLogs.println("EventLog::init() recovering data for " 
                        + uuid);
                readControlFile();
	    } else { 
                if (dbgLogs != null)
                    dbgLogs.println("EventLog::init() default initialization " 
                        + uuid);
	    }

	    eventReader = new EventReader();
	    eventWriter = new EventWriter();

	} finally {
	}

        printControlData(dbgLogs, "After EventLog::init");
	initialized = true;

	if (debugState)
	    assertInvariants();
    }

    /**
     * Writes the given <tt>RemoteEvent</tt> to the underlying
     * storage mechanism, if possible. If an <tt>IOException</tt>
     * occurs, then the write cannot be guaranteed.
     *
     * @exception IOException if an I/O error occurs
     */
    void add(RemoteEvent event) throws IOException {

        stateCheck();

	if (debugState)
	    assertInvariants();

	long logNum = getLogNum(wcount);
	File log = getLogFile(logNum);
	LogOutputStream out = null;
        try {
            // Get output stream
	    out = streamPool.getLogOutputStream(log, wpos);

            // Write the event
	    eventWriter.write(event, out);

	    // Update the control data
	    wpos = out.getOffset();
	    ++wcount;
	} catch (IOException ioe) {
	    // We'll get interrupted when asked to shutdown.
	    // In this case, we can skip the call to nextWriteLog.
	    if (ioe instanceof InterruptedIOException) {
                if (dbgLogs != null)
                    dbgLogs.println("EventLog::add() interrupted "); 
	    } else {
                if (dbgLogs != null)
                    dbgLogs.println("EventLog::add() received IOExcepion " 
			+ "... skipping to next write log"); 
	        nextWriteLog(); 
	    }
            if (dbgLogs != null)
	        ioe.printStackTrace(dbgLogs);
	    throw (IOException)ioe.fillInStackTrace();
	} finally {
	    // Always make sure to release any obtained resources
	    if (out != null) {
	        if (getLogNum(wcount) == logNum) {
	            streamPool.releaseLogStream(out);
		} else { 
		    // Done writing to this particular log file, so
		    // remove it.
	            streamPool.removeLogStream(out);
	            wpos = 0;
		}
	    }
	    // Write control data before returning
            // TODO (FCS) - handle potential IOException
            writeControlFile();
	}

        printControlData(dbgLogs, "EventLog::add");

	if (debugState) {
	    assertInvariants();
	}
    }

    /**
     * Return the next <tt>RemoteEvent</tt> to be read. Note that
     * <tt>next</tt> is meant to be used in conjunction with 
     * <tt>remove</tt>. Subsequent calls to <tt>next</tt> will
     * return the same event until <tt>remove</tt> is called, which
     * actually updates the read pointer to the next event (indicating
     * that the previously read event was successfully processed).
     *
     * @exception IOException if an I/O error occurs
     *
     * @exception ClassNotFoundException if a class for the serialized
     *                object could not be found
     *
     * @exception NoSuchObjectException if no event is available 
     */
    RemoteEvent next() throws IOException, ClassNotFoundException {
        boolean IOExceptionCaught = false;

        stateCheck();

	if (debugState) {
	    assertInvariants();
	}

        // Check if empty
	if (isEmpty()) 
	    throw new NoSuchElementException();

        long logNum = getLogNum(rcount);
        File log = getLogFile(logNum);
        LogInputStream in = null;
        RemoteEvent evt = null;
        try {
            // get the input stream
            in = streamPool.getLogInputStream(log, rpos);
            // read the event
            evt = eventReader.read(in); 
            // update the control data
            nextReadPos = in.getOffset();
            // Don't increment "real" read count until event is delivered
            // indicated by a call to remove().
        } catch (IOException ie) {
	    // We'll get interrupted when asked to shutdown.
	    // In this case, we can skip the call to nextReadLog.
	    if (ie instanceof InterruptedIOException) {
                if (dbgLogs != null)
                    dbgLogs.println("EventLog::next() interrupted "); 
	    } else {
                if (dbgLogs != null)
                    dbgLogs.println("EventLog::next() received IOExcepion " 
			+ "... skipping to next read log"); 
	        nextReadLog();
                IOExceptionCaught = true; 
	    }
            if (dbgLogs != null)
                ie.printStackTrace(dbgLogs);
	    throw (IOException)ie.fillInStackTrace();
        } catch (ClassNotFoundException cnfe) {
            // Note that the read offset is still valid since the
            // MarshalledObject extraction should always succceed.
            // It's just that the RemoteEvent within it could not
            // be reconstituted. Therefore, just skip to the next
            // MarshalledObject in the stream.
            nextReadPos = in.getOffset();
	    throw (ClassNotFoundException)cnfe.fillInStackTrace();
	} finally {
	    // If an IOException occurs then the eventReader is in 
	    // a bad state. Therefore, we get rid of our existing
	    // reference and create a fresh one. The underlying
	    // stream is also removed from from the pool, forcing
	    // a fresh copy to be returned on the next "get" request.
	    if (IOExceptionCaught) {
	        //TODO (FCS) add close to interface
	        //eventReader.close();
	        eventReader = new EventReader();
	        if (in != null)
	            streamPool.removeLogStream(in);
		// Remove associated log file since we are skiping over it
		if (!log.delete())
		    if (dbgLogs != null)
		        dbgLogs.println("Had trouble deleting " + log);
	    } else {
	        if (in != null)
	            streamPool.releaseLogStream(in);
	    }

            // TODO (FCS) - handle potential IOException
            writeControlFile();

	}

        printControlData(dbgLogs, "After Event::next");

	if (debugState) {
	    assertInvariants();
	}

        return evt;
    }
    
    /**
     * Returns <tt>true</tt> if this log contains no events and
     * false otherwise.
     */
    boolean isEmpty() throws IOException {
        stateCheck();
        return !(rcount < wcount);
    }

    /**
     * Effectively removes the last read event from the log.
     * It does this by advancing the read pointer to the 
     * next availble event, if any.
     *
     * @exception NoSuchElementException if no events are availble
     */
    void remove() throws IOException {
        stateCheck();

	if (debugState) {
	    assertInvariants();
	}

	// Get current log number
        long logNum = getLogNum(rcount);

	// Update read counts
	if (rcount < wcount) {
	    ++rcount;
	    rpos = nextReadPos;
	} else {
	    throw new NoSuchElementException();
	}

	// Remove old log if we've advanced to a new
	// log file (ie done reading the old log file).
	if (getLogNum(rcount) != logNum) {
	    File log = getLogFile(logNum);
	    LogInputStream in = null;
	    try {
	        in = streamPool.getLogInputStream(log, rpos);
	        streamPool.removeLogStream(in);
	    } finally {
		if (!log.delete())
		    if (dbgLogs != null)
		        dbgLogs.println("Had trouble deleting " + log);
	        rpos = nextReadPos = 0;
	    }
	}

	// Store control data
        writeControlFile();

        // Verify that state remains intact
	assertInvariants();


        printControlData(dbgLogs, "After Event::remove");

    }

    /**
     * Close this log and release any associated runtime resources.
     */
    void close() throws IOException {
        stateCheck();
	if (debugState) {
	    assertInvariants();
	}

	// Close read log
	long logNum = getLogNum(rcount);
	File log = getLogFile(logNum);
	LogStream strm = null;
	try {
	    strm = streamPool.getLogInputStream(log, rpos);
	    streamPool.removeLogStream(strm);
	} catch (FileNotFoundException fnfe) {
	    // Can happen if close is called just after you've finished
	    // reading a log. That is, the readCount is advanced to a new
	    // log which might not exist if a write hasn't taken place.
	    // TODO (FCS) - put a check to verify this condition and throw an
	    // exception otherwise?
	} catch (IOException ioe) {
	    // catch, but ignore so as not to skip the following
	    // code.
	}

	// Close write log
	logNum = getLogNum(wcount);
	log = getLogFile(logNum);
	try {
	    strm = streamPool.getLogOutputStream(log, wpos);
	    streamPool.removeLogStream(strm);
	} catch (IOException ioe) {
	    // catch, but ignore so as not to skip the following
	    // code.
	}
	
	// Close control log
	try {
	    strm = streamPool.getControlLog(controlFile);
	    streamPool.removeLogStream(strm);
	} catch (IOException ioe) {
	    // catch, but ignore so as not to skip the following
	    // code.
	}

        //TODO (FCS) - flag that an IOException was caught and rethrow to
        // indicate that there was a problem.

	closed = true;

	if (dbgLogs != null) 
	    dbgLogs.println("EventLog::close for " + uuid);
    }

    /**
     * Delete associated storage resources for this log.
     * 
     * @exception IOException if an IO error occurs
     */
    void delete() throws IOException {
        if (!closed)
            throw new IOException("Cannot delete log until it is closed");
        removeDir(logDir);
    }

    /**
     * Attempt to delete the associated event log persistence directory.
     */
    private static void removeDir(File logDir) {
        if (logDir == null || !logDir.isDirectory())
            return;

        // Get the contents of this directory
        String[] contents = logDir.list();
	if (contents == null)  // Can happen if there was an IO error
	    return;

        File entry = null;
	if (dbgLogs != null)
	    dbgLogs.println("Deleting contents of: " + logDir);
        for (int i=0; i < contents.length; i++) {
            entry = new File(logDir, contents[i]);
            if (entry.isDirectory()) {
                removeDir(entry);
	    } else {
                if(!entry.delete()) {
                    if (dbgLogs != null)
                        dbgLogs.println("Had trouble deleteing file: " + 
                            entry);
		} else {
                    if (dbgLogs != null)
                        dbgLogs.println("Deleted file: " + entry);
		}
	    }
	}

        if (!logDir.delete())
            if (dbgLogs != null)
                dbgLogs.println("Had trouble deleteing directory: " + 
                    logDir);
	    else
                if (dbgLogs != null)
                    dbgLogs.println("Deleted directory: " + logDir);
    }

    /**
     * Advance the "read" state to the next available log.
     */
    private void nextReadLog() {
	if (debugState)
	    assertInvariants();

        // Normalize event count in case we were called due to a problem with
        // one of the log files
        long remainder = rcount % eventsPerLogFile; 
        rcount += (eventsPerLogFile - remainder);
	rpos = 0;

        // Ensure that the writeLog data doesn't get behind
	if (verifyInvariants() == false) {
	    nextWriteLog();
	}

        // Assert that state is still valid
	if (debugState)
	    assertInvariants();

        printControlData(dbgLogs, "EventLog::nextReadLog");
    }

    /**
     * Advance the "write" state to the next available log.
     */
    private void nextWriteLog() {
        // Don't check invariants at the top since we can get called
        // from readNextLog() with invalid state.

        // Normalize event count in case we were called due to a problem with
        // one of the log files
        long remainder = wcount % eventsPerLogFile; 
        wcount += (eventsPerLogFile - remainder);
	wpos = 0;

        printControlData(dbgLogs, "EventLog::nextWriteLog");

	if (debugState)
	    assertInvariants();
    }

    /**
     * Pass-through method to allow unit test access. 
     * This is intended for debugging purposes only.
     */
    synchronized void dumpControlData(PrintWriter pw, String msg) {
        printControlData(pw, msg);
    }
    
    /**
     * Output state information to the given <tt>PrintWriter</tt>.
     * This is intended for debugging purposes only.
     */
    private void printControlData(PrintWriter pw, String msg) {
	if (pw != null) {
            pw.println(msg);
            pw.println("ID:         " + uuid);
            pw.println("ReadCount:  " + rcount);
            pw.println("ReadPos:    " + rpos);
            pw.println("NextReadPos:" + nextReadPos);
            pw.println("WriteCount: " + wcount);
            pw.println("WritePos:   " + wpos);
	}
    }

    /**
     * Write state information to the underlying store.
     */
    private void writeControlFile() throws IOException {
	packLong(wcount, ctlbuf, 0);
	packLong(rcount, ctlbuf, 8);
	packLong(wpos, ctlbuf, 16);
	packLong(rpos, ctlbuf, 24);
	
	ControlLog ctl = streamPool.getControlLog(controlFile);
	ctl.seek(0L);
	ctl.write(ctlbuf);
	ctl.getFD().sync();
	streamPool.releaseLogStream(ctl);
    }
    
    /**
     * Read state information from the underlying store.
     */
    private void readControlFile() throws IOException {
	ControlLog ctl = streamPool.getControlLog(controlFile);
	ctl.seek(0L);
	ctl.readFully(ctlbuf);
	streamPool.releaseLogStream(ctl);

	wcount = unpackLong(ctlbuf, 0);
	rcount = unpackLong(ctlbuf, 8);
	wpos = unpackLong(ctlbuf, 16);
	rpos = unpackLong(ctlbuf, 24);
    }
    
    /**
     * Utility method for packing a <tt>long</tt> into a <tt>byte</tt> array.
     */
    private static void packLong(long val, byte[] b, int off) {
	b[off++] = (byte) (val >>> 56);
	b[off++] = (byte) (val >>> 48);
	b[off++] = (byte) (val >>> 40);
	b[off++] = (byte) (val >>> 32);
	b[off++] = (byte) (val >>> 24);
	b[off++] = (byte) (val >>> 16);
	b[off++] = (byte) (val >>> 8);
	b[off++] = (byte) (val >>> 0);
    }
    
    /**
     * Utility method for unpacking a <tt>long</tt> from a <tt>byte</tt> array.
     */
    private static long unpackLong(byte[] b, int off) {
	return ((b[off + 0] & 0xFFL) << 56) +
	    ((b[off + 1] & 0xFFL) << 48) +
	    ((b[off + 2] & 0xFFL) << 40) +
	    ((b[off + 3] & 0xFFL) << 32) +
	    ((b[off + 4] & 0xFFL) << 24) +
	    ((b[off + 5] & 0xFFL) << 16) +
	    ((b[off + 6] & 0xFFL) << 8) +
	    ((b[off + 7] & 0xFFL) << 0);
    }
    
    /**
     * Utility method for returning the <tt>File</tt> associated with the
     * given <tt>lognum</tt>.
     */
    private File getLogFile(long lognum) {
	return new File(logDir, lognum + "." + LOGFILE_SUFFIX).getAbsoluteFile();
    }
    
    /**
     * Utility method for returning the <tt>File</tt> that contains the
     * state information for this log.
     */
    private File getControlFile() {
	return new File(logDir, "log." + CTLFILE_SUFFIX).getAbsoluteFile();
    }

    /**
     * Utility method for returning the log file number for the given
     * (event) <tt>count</tt>.
     */
    private static long getLogNum(long count) {
	return (count / eventsPerLogFile); 
    }


    /**
     * Asserts that the log is in a valid state.
     * 
     * @exception IOException if the log is in an invalid state
     */
    private void stateCheck() throws IOException {
        if (!initialized)
            throw new IOException("Trying to use an uninitialized "
                + "control data object for: " + uuid);
	if (closed)
	    throw new IOException("Attempt to access closed log file for : "
	        + uuid);
    }

    /**
     * Utility method for checking if the object invariants are valid.
     */
    private boolean verifyInvariants() {
        if ((wcount < rcount) ||
            ((getLogNum(wcount) == getLogNum(rcount)) && (wpos < rpos))) {
            return false;
	}
	return true;
    }

    /**
     * Utility method for asserting that the object invariants are valid.
     *
     * @exception InternalMailboxException if the invariants aren't valid
     */
    private void assertInvariants() {
        if (verifyInvariants() == false) {
            printControlData(dbgLogs, "Dumping invalid state for " + uuid);
            throw new InternalMailboxException("Invalid state for " + uuid);
	}
    }
}
