/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.mercury;

import com.sun.jini.proxy.UUID;

import java.io.File;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Collections;
import java.util.HashMap;
import java.util.NoSuchElementException;
import java.util.LinkedList;
import java.util.List;

import net.jini.core.event.RemoteEvent;
 
/**
 * This class serves as a factory for generating <tt>EventLogIterator</tt>
 * objects. The iterator objects are cached so that subsequent calls 
 * for the same iterator return the same object.
 *
 * @since Jini1.1
 */
public class EventLogFactory {

    /** <tt>Map</tt> that contains references to generated iterators */
    private HashMap iterators = new HashMap();

    /** 
     * Method to return the iterator object for the designated 
     * <tt>UUID</tt>. The <tt>File</tt> argument determines where the
     * persistence store will be maintained and is only used for the
     * first instance of the iterator. (Note that upon recovery
     * from a restart/crash the <tt>UUID</tt> will already exist so the
     * <tt>logPath</tt> arg will not be used.)
     * Subsequent calls for the same <tt>UUID</tt> will return the 
     * original object.
     */
    public EventLogIterator iterator(UUID uuid, File logPath) {

        // Try to get reference from cache
        EventLogIteratorImpl eli = (EventLogIteratorImpl)iterators.get(uuid);
        if (eli == null) { // doesn't exist, so create one
            eli = new EventLogIteratorImpl(uuid, logPath);
	    try { 
		eli.init(); // initialize the iterator
	    } catch (IOException ioe) {
		// ignore ... the next usage of this
		// object will throw IOException
	    }
            iterators.put(uuid, eli); // add to cache
	}
            
	return eli;
    }

    /**
     * Remove the <TT>EventLogIterator</TT> associated with the given
     * <TT>UUID</TT>. This is (presumabley) called to flush an existing
     * <TT>EventLogIterator</TT> whose storage location has changed.
     * This way, the next call to <TT>iterator</TT> will produce a new
     * object instead of returning the cached version.
     */
    public void remove(UUID uuid) {
        iterators.remove(uuid); // remove from cache, if it's there
    }

    /**
     * Private class which implements the <tt>EventLogIterator</tt>
     * interface. This class delegates to an <tt>EventLog</tt> for most
     * of its functionality.
     */
    private static class EventLogIteratorImpl 
	implements EventLogIterator
    {
      
        /** The associated <tt>UUID</tt> for this iterator */
        private final UUID uuid;

        /** The associated <tt>EventLog</tt> for this iterator */
        private EventLog log = null; 

        /** 
         * Simple constructor that assigns the <tt>UUID</tt>
         * field to the appropriate internal field and creates
         * an <tt>EventLog</tt> object using the provided
         * <tt>File</tt> argument.
         */
        EventLogIteratorImpl(UUID id, File logPath) {
            uuid = id;
	    log = new EventLog(id, logPath);
        }

        // Inherit documentation from supertype
        public void init() throws IOException {
	    log.init();
	}

        // Inherit documentation from supertype
        public void add(RemoteEvent evt) throws IOException, 
		IllegalArgumentException 
	{
            log.add(evt);
	}
            
        // Inherit documentation from supertype
        public boolean hasNext()  throws IOException {
            return !log.isEmpty();
        }

        // Inherit documentation from supertype
        public RemoteEvent next() throws IOException, 
		ClassNotFoundException, NoSuchElementException 
	{
            return log.next();
        }

        // Inherit documentation from supertype
        public void remove() throws IOException, IllegalStateException {
            log.remove();
        }

        // Inherit documentation from supertype
        public void destroy() throws IOException {
            log.close();
	    log.delete();
        }
    }
}


