/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.mercury;

import com.sun.jini.proxy.ThrowThis;
import com.sun.jini.proxy.UUID;

import java.io.Serializable;
import java.rmi.RemoteException;

import net.jini.core.event.RemoteEventListener;
import net.jini.core.event.RemoteEvent;
import net.jini.core.event.UnknownEventException;

/**
 * The <code>ListenerProxy</code> class implements the 
 * <code>RemoteEventListener</code> interface.
 * Instances of this class are provided as the event "forwarding" 
 * target to clients of the mailbox service.  
 *
 * This class works in conjunction with the <code>ListenerProxyRep</code>
 * class in order to facilitate smart proxy functionality. 
 * Prior to the Java 2 SDK, version 1.2.2, RMI semantics would attempt 
 * to substitute the associated
 * "stub" class for any Remote object being marshalled.  This behavior would
 * not allow the passing of state information (i.e. <code>registrationID</code>
 * for <code>notify</code> invocations.  
 *
 * This class, instead, overrides the writeReplace method, which allows 
 * this class to substitute a class of its choice into the marshalling 
 * stream before RMI semantics are applied. In this case, a ListenerProxyRep 
 * (a non-Remote interface) is substituted for this class in the marshalling 
 * stream. The ListenerProxyRep, in turn, overrides its readResolve method in 
 * order to reconstitute a ListenerProxy object when it is unmarshalled.
 *
 * This "trick" essentially allows us to pass a smart proxy through
 * third party objects while maintaining the RemoteEventListener interface
 * contract (which could not occur using default RMI semantics).
 *
 * @since Jini1.1
 */

class ListenerProxy implements RemoteEventListener, Serializable {

    /**
     * The reference to the event mailbox service implementation
     *
     * @serial
     */
    private final MailboxBackEnd server;

    /**
     * The proxy's <code>UUID</code>
     *
     * @serial
     */
    private final UUID registrationID;

    /** Simple constructor */
    ListenerProxy(MailboxBackEnd ref, UUID regID) {
        server = ref;
        registrationID = regID;
    }

    // documentation inherited from supertype
    public void notify(RemoteEvent theEvent) 
	throws UnknownEventException, RemoteException 
    {
	try {
            server.notify(registrationID, theEvent);
	} catch (ThrowThis tt) {
	    tt.throwRemoteException();
	}

    }

    /** 
     * Called by <tt>ObjectOutputStream</tt> to allow 
     * this class to substitute a <tt>ListenerProxyRep</tt> into
     * the serialization stream during marshalling.
     */
    private Object writeReplace() {
	return new ListenerProxyRep(server, registrationID);
    }

    // documentation inherited from supertype
    public int hashCode() {
	return registrationID.hashCode();
    }

    /** Proxies for registrations with the same ID are considered equal. */
    public boolean equals(Object o) {
        if (this == o)
            return true;

        if (o == null)
            return false;

        if (o.getClass() != getClass())
            return false;

        ListenerProxy lp = (ListenerProxy)o;
	return registrationID.equals(lp.registrationID);
    }

    /** Returns the unique registration ID associated with this listener */
    public UUID getID() {
	return registrationID;
    }
}


