/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.mercury;


import com.sun.jini.constants.TimeConstants;
import com.sun.jini.debug.Debug;
import com.sun.jini.lease.landlord.LeasedResource;
import com.sun.jini.lease.landlord.LeaseDurationPolicy;
import com.sun.jini.lease.landlord.Landlord.RenewResults;
import com.sun.jini.lease.landlord.LocalLandlord;
import com.sun.jini.lease.landlord.LandlordUtil;
import com.sun.jini.lookup.entry.BasicServiceType;
import com.sun.jini.lookup.entry.LookupAttributes;
import com.sun.jini.mahout.BasicServiceDesc; 
import com.sun.jini.mahout.ParseUtils; 
import com.sun.jini.proxy.ThrowThis;
import com.sun.jini.proxy.UUIDFactory;
import com.sun.jini.proxy.UUID;
import com.sun.jini.reliableLog.ReliableLog;
import com.sun.jini.reliableLog.LogException;
import com.sun.jini.reliableLog.LogHandler;
import com.sun.jini.system.FileSystem;
import com.sun.jini.thread.ReadersWriter;
import com.sun.jini.thread.ReadersWriter.ConcurrentLockException;
import com.sun.jini.thread.RetryTask;
import com.sun.jini.thread.TaskManager;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.Serializable;
import java.rmi.activation.Activatable;
import java.rmi.activation.ActivationException;
import java.rmi.activation.ActivationGroup;
import java.rmi.activation.ActivationGroupDesc;
import java.rmi.activation.ActivationGroupDesc.CommandEnvironment;
import java.rmi.activation.ActivationGroupID;
import java.rmi.activation.ActivationDesc;
import java.rmi.activation.ActivationID;
import java.rmi.activation.ActivationSystem;
import java.rmi.activation.UnknownObjectException;
import java.rmi.MarshalledObject;
import java.rmi.NoSuchObjectException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.ServerException;
import java.rmi.server.RemoteObject;
import java.rmi.server.UnicastRemoteObject;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Random;
import java.util.TreeMap;

import net.jini.admin.Administrable;
import net.jini.admin.JoinAdmin;
import net.jini.core.discovery.LookupLocator;
import net.jini.core.entry.Entry;
import net.jini.core.event.RemoteEventListener;
import net.jini.core.event.RemoteEvent;
import net.jini.core.event.UnknownEventException;
import net.jini.core.lease.Lease;
import net.jini.core.lease.LeaseException;
import net.jini.core.lease.LeaseDeniedException;
import net.jini.core.lease.UnknownLeaseException;
import net.jini.core.lookup.ServiceID;
import net.jini.discovery.LookupDiscoveryManager;
import net.jini.event.EventMailbox;
import net.jini.event.MailboxRegistration;
import net.jini.lookup.entry.ServiceInfo;
import net.jini.lookup.JoinManager;
import net.jini.lookup.ServiceIDListener;
import net.jini.discovery.LookupDiscovery;

/**
 * <tt>MailboxImpl</tt> implements the server side of the event 
 * mailbox service.
 * <p>
 * Client-side proxies make the appropriate transformation from client
 * requests to the methods of this class.
 *
 * @since Jini1.1
 */

/*
Implementation Details:

The runtime state of this service is basically maintained in the following
data structures:
   * regByID - contains the UUID to Registration mapping
   * regByExpiration - contains a time sorted ordering of 
       Registration objects
   * pendingReg - contains a list of UUIDs that have been
       enabled for event delivery
   * activeReg - contains a list of UUIDs that have an
       event delivery task in progress

Note that the latter 3 data structures can be rebuilt from the first, so only
regByID is actually stored to and retrieved from disk. 

See init() for exact details of how transient state gets rebuilt.
See takeSnapshot() for exact details of what gets stored.
See recoverSnapshot() for exact details of what gets retrieved.
*/

public class MailboxImpl implements MailboxBackEnd, TimeConstants {

    /** ServiceInfo product value */
    private static final String PRODUCT = "EventMailbox";
    /** ServiceInfo manufacturer value */
    private static final String MANUFACTURER = "Sun Microsystems, Inc.";
    /** ServiceInfo vendor value */
    private static final String VENDOR = MANUFACTURER;
    /** ServiceInfo version value */
    private static final String VERSION = 
	com.sun.jini.constants.VersionConstants.SERVER_VERSION;

    /** The primary proxy to the current instance of this class */
    private MailboxProxy mailboxProxy;

    /** Concurrent object (lock) to control read and write access */
    private final ReadersWriter concurrentObj = new ReadersWriter();

    /** Map from <code>UUID</code> to <code>ServiceRegistration</code> */
    // HashMap is unsynchronized, but we are performing external
    // synchronization via the <code>concurrentObj</code> field. 
    private HashMap regByID = new HashMap(); 

    /**
     * Identity map of <tt>ServiceRegistration</tt>, ordered by lease 
     * expiration. This is a parallel data structure to <code>regByID</code>.
     */
    // TreeMap is unsynchronized, but we are performing external
    // synchronization via the <code>concurrentObj</code> field. 
    private final TreeMap regByExpiration = new TreeMap();

    /** 
     * List of <tt>ServiceRegistration</tt>s that have event 
     * delivery enabled, but don't have any event delivery tasks 
     * currently scheduled.
     */
    // Using an ArrayList for random access performance. 
    // ArrayList is unsynchronized, but we are performing external
    // synchronization via the <code>concurrentObj</code> field. 
    private List pendingReg = new ArrayList();

    /**
     * Map of <tt>ServiceRegistration</tt>s that have event 
     * delivery enabled and have event delivery tasks currently 
     * scheduled.
     */
    // HashMap is unsynchronized, but we are performing external
    // synchronization via the <code>concurrentObj</code> field. 
    private Map activeReg = new HashMap();

    /** Reliable log to hold registration state information */
    // Note that event state is kept separately
    private ReliableLog log;

    /** Log format version */
    private static final int LOG_VERSION = 1;

    /** Flag indicating whether system is in a state of recovery */
    private boolean inRecovery;

    /** Current number of records in the Log File since the last snapshot */
    private int logFileSize = 0;

    /** Log File must contain this many records before a snapshot is allowed */
    // TODO (FCS)- allow this to be a user configurable parameter
    private int logToSnapshotThresh = 50;

    /** Object on which the snapshot-taking thread will synchronize */
    private final Object snapshotNotifier = new Object();

    /** Snapshot-taking thread */
    private Thread snapshotter = null;

    /** Name of persistence directory */
    private String logDirname = null;

    /** ServiceID returned from the lookup registration process */
    private ServiceID serviceID = null;

    /** Our activation ID */
    private ActivationID activationID = null;

    /** <code>UUID</code> generator. */
    private final UUIDFactory uuidFactory = new UUIDFactory();

    /** <code>EventLogIterator</code> generator */
    private final EventLogFactory eventLogFactory = new EventLogFactory();

    /** 
     * <code>UUID</code> that is shared among service proxies generated 
     * from this service instance (e.g. <tt>MailboxProxy</tt>,
     * <tt>MailboxAdminProxy</tt>, and <tt>ListenerProxy</tt>).
     * This value is used for determining equality among proxy objects.  
     * Service proxy objects with the same <code>UUID</code> are 
     * presumed to be equal.
     */
    private UUID serviceProxyID;
  
    /** <code>LeaseDurationPolicy</code> for this service */
    private LeaseDurationPolicy leasePolicy = new LeaseDurationPolicy(
	    3 * HOURS,     // Maximum lease length
	    1 * HOURS,     // Default lease length 
	    this,          // Landlord reference (managing own leases) 
	    null,          // Lease manager reference (none used)
	    null);         // LandlordLeaseFactory reference (use default) 

    /** LocalLandlord to use with LandlordUtil calls */
    private LocalLandlordAdaptor localLandlord = new LocalLandlordAdaptor();

    /** Manager for joining lookup services */
    private JoinManager joiner = null;

    /** 
     * DiscoveryManager for joining lookup services. 
     * This can always be obtained from the JoinManager, so
     * this is just used as a shortcut.
     */
    private LookupDiscoveryManager lookupDiscMgr = null;

    /** The attributes to use when joining lookup services */
    private Entry[] lookupAttrs = new Entry[] { 
	    new ServiceInfo(PRODUCT, MANUFACTURER, VENDOR, VERSION, "", ""),
            new BasicServiceType("Event Mailbox")
    };

    /** 
     * The lookup groups we should join. 
     * Default is to join no groups. 
     */
    private String[] lookupGroups = LookupDiscovery.NO_GROUPS;

    /** 
     * The lookup locators we should join 
     * Default is to join with no locators. 
     */
    private LookupLocator[] lookupLocators = new LookupLocator[0];

    /** Event delivery thread */
    private Notifier notifier = null;

    /** Object for coordinating actions with the notification thread */
    private final Object eventNotifier = new Object();

    /** Registration expirer thread */
    private ExpirationThread expirer = null;

    /** Earliest expiration time of any registration */
    private long minRegExpiration = Long.MAX_VALUE;

    /** Object for coordinating actions with the expire thread */
    private final Object expirationNotifier = new Object();
 
    /** Debug utility class */
    private static final Debug debug = new Debug(MailboxBackEnd.DEBUG_PROP);

    /** <tt>PrintWriter</tt> handle for leasing related messages */
    private static final PrintWriter dbgLeases = debug.getWriter("leases");

    /** <tt>PrintWriter</tt> handle for event delivery related messages */
    private static final PrintWriter dbgDelivery = debug.getWriter("delivery");

    /** <tt>PrintWriter</tt> handle for administrative related messages */
    private static final PrintWriter dbgAdmin = debug.getWriter("admin");

    /** <tt>PrintWriter</tt> handle for initialization related messages */
    private static final PrintWriter dbgInit = debug.getWriter("init");

    /** <tt>PrintWriter</tt> handle for event reception related messages */
    private static final PrintWriter dbgReceive = debug.getWriter("receive");

    /** <tt>PrintWriter</tt> handle for expiration related messages */
    private static final PrintWriter dbgExpiration = 
	debug.getWriter("expiration");

    /** <tt>PrintWriter</tt> handle for state recovery related messages */
    private static final PrintWriter dbgRecovery = debug.getWriter("recovery");

    /** <tt>PrintWriter</tt> handle for event storage related messages */
    private static final PrintWriter dbgLogs = debug.getWriter("logs"); 

    /** 
     * Constructor for creating transient (i.e. non-activatable) service 
     * instances. 
     * State information is still logged to persistent storage.
     * This method is only intended for debugging purposes at this time.
     *
     * @param log directory where persistent state is maintained
     */
    public MailboxImpl(String log) throws IOException {
        UnicastRemoteObject.exportObject(this);
	if (dbgInit != null) {
	    dbgInit.println("MailboxImpl::MailboxImpl(log=" + log + ")");
	}
        logDirname = log;
        init();
    }

    // inherit javadoc from supertype
    public UUID getProxyID() {
	return serviceProxyID;
    }

    /** Initialization common to both activatable and transient instances. */
    private void init() throws IOException {
	if (dbgInit != null) {
	    dbgInit.println("MailboxImpl::init() ReliableLog --> " 
	        + logDirname);
	}
        log = new ReliableLog(logDirname, new LocalLogHandler());

	if (dbgInit != null) {
	    dbgInit.println("MailboxImpl::init() recovering persistent state");
	}
        inRecovery = true;
        log.recover();
        inRecovery = false;

				      
        // Would like to use serviceID instead, but don't want to 
        // delay construction in order to wait for the first 
        // lookup registration, which might not occur in a timely 
        // fashion.
	if (serviceProxyID == null) {
	    serviceProxyID = uuidFactory.newUUID();
        }

	if (dbgInit != null) {
	    dbgInit.println("MailboxImpl::init() taking snapshot");
	}

	// Take snapshot of current state.
        log.snapshot();

        // Reconstruct any transient state, if necessary.
        rebuildTransientState();

        // Create the proxy that will be registered in the lookup service
        mailboxProxy = 
	    new MailboxProxy((MailboxBackEnd)RemoteObject.toStub(this),
                             serviceProxyID);

        /* Register this service with any configured lookup services */
	if (serviceID == null) { // First instance ... need service id
	    if (dbgInit != null) {
	        dbgInit.println("MailboxImpl::init() creating JoinMgr w/o serviceID");
	    }
	    joiner = new JoinManager(
		mailboxProxy,              // service object 
		lookupAttrs,               // service attributes
		new SrvcIDListener(),      // ServiceIDListener reference 
		null,			   // DiscoveryManagement ref - default
		null);                     // LeaseRenewalManager reference
		 		           // - using default manager
	} else { // Rejoin with (recovered) state information
	    if (dbgInit != null) {
	        dbgInit.println("MailboxImpl::init() creating JoinMgr w/ serviceID");
	    }
	    joiner = new JoinManager(
		mailboxProxy,              // service object
		lookupAttrs,               // service attributes
		serviceID,                 // Service ID
		null,            	   // DiscoveryManagement ref - default
		null);                     // LeaseRenewalManager reference
		 		           // - using default manager
        }

        // Get shorthand reference to the discovery manager
        lookupDiscMgr = (LookupDiscoveryManager)joiner.getDiscoveryManager();

        // Start threads
	notifier = new Notifier();
	expirer = new ExpirationThread();
	snapshotter = new SnapshotThread();

        if (dbgInit != null) {
	    dbgInit.println("MailboxImpl::init() done");
        }
    } // End init()

    // Rebuilds internal data structures after a restart.
    private void rebuildTransientState() {

	// Reconstruct regByExpiration and pendingReg data structures,
	// if necessary.
	if (!regByID.isEmpty()) {
	    if (dbgInit != null) {
	        dbgInit.println("Rebuilding transient state ...");
	    }
	    Collection regs = regByID.values();
	    Iterator iter = regs.iterator();
	    ServiceRegistration reg = null;
	    UUID uuid = null;
            EventLogIterator eli = null; 
	    while (iter.hasNext()) {
	        reg = (ServiceRegistration)iter.next(); // get Reg
	        uuid = (UUID)reg.getCookie();           // get its UUID
	        if (dbgInit != null) {
	            dbgInit.println("Checking reg : " + reg);
		}
	        // Check if registration is still current
	        if (leasePolicy.ensureCurrent(reg)) {
	            if (dbgInit != null) {
	                dbgInit.println("Reinitializing iterator ...");
		    }
		    // regenerate an EventLogIterator for this Reg
		    // Note that event state is maintained separately
		    // through the event log mechanism.
                    eli = eventLogFactory.iterator(uuid, 
		        getEventLogPath(logDirname, uuid)); 
		    reg.setIterator(eli);
	            if (dbgInit != null) {
	                dbgInit.println("Adding registration to "
	                    + "expiration watchlist");
		    }
		    // Put Reg into time sorted collection
	            regByExpiration.put(reg, reg);

	            // Check if registration needs to be added to the
	            // pending list. Note, we could have processed
	            // an "enabled" log record during recovery, so 
	            // only add it if it's not already there.
	            // We don't need to check activeReg since the
	            // the notifier hasn't kicked in yet. Don't call
	            // enableRegistration() since it clears the "unknown
	            // events" list which we want to maintain.
	            if (reg.getEventTarget() != null &&
	                !pendingReg.contains(uuid))
		    {
	                if (dbgInit != null) {
	                    dbgInit.println("Adding registration to "
	                        + "pending task list");
			}
        	        pendingReg.add(uuid);
	            }
	        } else {
	            // Registration has expired, so remove it.
	            removeRegistration(uuid, reg);
	            if (dbgInit != null) {
	                dbgInit.println("Removing expired registration: ");
		    }
	        }
	    }
	}
        
    }

    ///////////////////////////
    // MailboxBackEnd Methods
    ///////////////////////////

    // This method's javadoc is inherited from an interface of this class
    public MailboxRegistration register(long leaseDuration) 
	throws RemoteException
    {
	concurrentObj.writeLock();
	try {
	    return registerDo(leaseDuration);
        } finally {
	    concurrentObj.writeUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public void enableDelivery(UUID uuid, RemoteEventListener target) 
	throws RemoteException, ThrowThis 
    {
	concurrentObj.writeLock();
	try {
	    enableDeliveryDo(uuid, target);
        } finally {
	    concurrentObj.writeUnlock();
	}
    }
	
    // This method's javadoc is inherited from an interface of this class
    public void disableDelivery(UUID uuid) 
	throws RemoteException, ThrowThis 
    {
	concurrentObj.writeLock();
	try {
	    disableDeliveryDo(uuid);
        } finally {
	    concurrentObj.writeUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public void notify(UUID registrationID, RemoteEvent theEvent) 
	throws UnknownEventException, RemoteException, ThrowThis
    {
	concurrentObj.writeLock();
	try {
	    notifyDo(registrationID, theEvent);
        } finally {
	    concurrentObj.writeUnlock();
	}
    }


    ////////////////////////
    // Administrable Methods
    ////////////////////////

    // This method's javadoc is inherited from an interface of this class
    public Object getAdmin() throws RemoteException {
        if (dbgAdmin != null) {
            dbgAdmin.println("getAdmin called");
        }
	return (new MailboxAdminProxy(this, serviceProxyID));
    }


    ////////////////////////
    // Landlord Methods
    ////////////////////////

    // inherit javadoc from parent
    public long renew(Object cookie, long extension)
	    throws LeaseDeniedException, UnknownLeaseException, RemoteException 
    {
	concurrentObj.writeLock();
	try {
	    return renewDo(cookie, extension);
        } finally {
	    concurrentObj.writeUnlock();
	}
    }

    // inherit javadoc from parent
    public void cancel(Object cookie)
	    throws UnknownLeaseException, RemoteException 
    {
	concurrentObj.writeLock();
	try {
	    cancelDo(cookie);
        } finally {
	    concurrentObj.writeUnlock();
	}
    }

    // inherit javadoc from parent
    public RenewResults renewAll(Object[] cookie, long[] extension)
	    throws RemoteException 
    {
	concurrentObj.writeLock();
	try {
	    return renewAllDo(cookie, extension);
        } finally {
	    concurrentObj.writeUnlock();
	}
    }

    // inherit javadoc from parent
    public Map cancelAll(Object[] cookie) throws RemoteException 
    {
	concurrentObj.writeLock();
	try {
	    return cancelAllDo(cookie);
        } finally {
	    concurrentObj.writeUnlock();
	}
    }

    //////////////////////////////////
    // KillVMAdmin methods
    //////////////////////////////////
    /**
     * Sleep time for <code>killVM</code>.
     *
     * @see killVM
     */
    private final static long KILL_DELAY = 500; // milliseconds

    // inherit doc comment
    public long killVM() {
        if (dbgAdmin != null) {
            dbgAdmin.println("killVM called");
        }
	/*
	 * We have to do this in a separate thread or the remote
	 * system will always get a RemoteException when the server
	 * to which it is talking (this one) dies in the middle of
	 * the call.  So we start up a new thread that sleeps for 
	 * KILL_DELAY and then shuts down the VM.  This lets
	 * the invoking thread return before the "end" occurs.
	 */
	new Thread() {
	    public void run() {
		try {
		    sleep(KILL_DELAY);
		} catch (InterruptedException e) {
		    ;	// oh, well -- we tried
		}
		System.exit(0);
	    }
	}.start();
	return KILL_DELAY * 6;
    }

    //////////////////////////////////
    // DestroyAdmin methods
    //////////////////////////////////

    // This method's javadoc is inherited from an interface of this class
    public void destroy() {
        if (dbgAdmin != null) {
            dbgAdmin.println("destroy called");
        }
	(new DestroyThread()).start();
    }


    //////////////////////////////////
    // JoinAdmin methods
    //////////////////////////////////

    // This method's javadoc is inherited from an interface of this class
    public Entry[] getLookupAttributes() throws RemoteException {
        if (dbgAdmin != null) {
            dbgAdmin.println("getLookupAttributes called");
        }
	concurrentObj.readLock();
	try {
	    return lookupAttrs;
	} finally {
	    concurrentObj.readUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public void addLookupAttributes(Entry[] attrSets) throws RemoteException {
        if (dbgAdmin != null) {
            dbgAdmin.println("addLookupAttributes called");
        }
	concurrentObj.writeLock();
	try {
	    // delegate functionality to JoinManager
	    joiner.addAttributes(attrSets, true);
	    lookupAttrs = joiner.getAttributes();
	    // Log the modification
	    addLogRecord(new AttrsAddedLogObj(attrSets));
	} finally {
	    concurrentObj.writeUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public void modifyLookupAttributes(Entry[] attrSetTemplates,
				       Entry[] attrSets)
	throws RemoteException
    {
        if (dbgAdmin != null) {
            dbgAdmin.println("modifyLookupAttributes called");
        }
	concurrentObj.writeLock();
	try {
	    // delegate functionality to JoinManager
	    joiner.modifyAttributes(attrSetTemplates, attrSets, true);
	    lookupAttrs = joiner.getAttributes();
	    // Log the modification
	    addLogRecord(new AttrsModifiedLogObj(attrSetTemplates, attrSets));
	} finally {
	    concurrentObj.writeUnlock();
	}
    }


    // This method's javadoc is inherited from an interface of this class
    public String[] getLookupGroups() throws RemoteException {
        if (dbgAdmin != null) {
            dbgAdmin.println("getLookupGroups called");
	    dumpGroups(lookupGroups, dbgAdmin);
        }
	concurrentObj.readLock();
	try {
	    return lookupGroups;
	} finally {
	    concurrentObj.readUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public void addLookupGroups(String[] groups) throws RemoteException {
        if (dbgAdmin != null) {
            dbgAdmin.println("addLookupGroups called");
	    dumpGroups(groups, dbgAdmin);
        }
	concurrentObj.writeLock();
	try {
	    try {
	        // delegate functionality to discovery manager
		lookupDiscMgr.addGroups(groups);
	    } catch (IOException e) {
		throw new RuntimeException(e.toString());
	    }
	    lookupGroups = lookupDiscMgr.getGroups();
	    // Log the modification
            addLogRecord(new LookupGroupsChangedLogObj(lookupGroups));
	} finally {
	    concurrentObj.writeUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public void removeLookupGroups(String[] groups) throws RemoteException {
        if (dbgAdmin != null) {
            dbgAdmin.println("removeLookupGroups called");
	    dumpGroups(groups, dbgAdmin);
        }
	concurrentObj.writeLock();
	try {
	    // delegate functionality to discovery manager
	    lookupDiscMgr.removeGroups(groups);
	    lookupGroups = lookupDiscMgr.getGroups();
	    // Log the modification
            addLogRecord(new LookupGroupsChangedLogObj(lookupGroups));
	} finally {
	    concurrentObj.writeUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public void setLookupGroups(String[] groups) throws RemoteException {
        if (dbgAdmin != null) {
            dbgAdmin.println("setLookupGroups called");
	    dumpGroups(groups, dbgAdmin);
        }
	concurrentObj.writeLock();
	try {
	    try {
	        // delegate functionality to discovery manager
		lookupDiscMgr.setGroups(groups);
	    } catch (IOException e) {
		throw new RuntimeException(e.toString());
	    }
	    lookupGroups = lookupDiscMgr.getGroups();
	    // Log the modification
            addLogRecord(new LookupGroupsChangedLogObj(lookupGroups));
	} finally {
	    concurrentObj.writeUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public LookupLocator[] getLookupLocators() throws RemoteException {
        if (dbgAdmin != null) {
            dbgAdmin.println("getLookupLocators called");
        }
	concurrentObj.readLock();
	try {
	    return lookupLocators;
	} finally {
	    concurrentObj.readUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public void addLookupLocators(LookupLocator[] locators)
	throws RemoteException
    {
        if (dbgAdmin != null) {
            dbgAdmin.println("addLookupLocators called");
        }
	concurrentObj.writeLock();
	try {
	    // delegate functionality to discovery manager
	    lookupDiscMgr.addLocators(locators);
	    lookupLocators = lookupDiscMgr.getLocators();
	    // Log the modification
	    addLogRecord(new LookupLocatorsChangedLogObj(lookupLocators));
	} finally {
	    concurrentObj.writeUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public void removeLookupLocators(LookupLocator[] locators)
	throws RemoteException
    {
        if (dbgAdmin != null) {
            dbgAdmin.println("removeLookupLocators called");
        }
	concurrentObj.writeLock();
	try {
	    // delegate functionality to discovery manager
	    lookupDiscMgr.removeLocators(locators);
	    lookupLocators = lookupDiscMgr.getLocators();
	    // Log the modification
	    addLogRecord(new LookupLocatorsChangedLogObj(lookupLocators));
	} finally {
	    concurrentObj.writeUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public void setLookupLocators(LookupLocator[] locators)
	throws RemoteException
    {
        if (dbgAdmin != null) {
            dbgAdmin.println("setLookupLocators called");
        }
	concurrentObj.writeLock();
	try {
	    // delegate functionality to discovery manager
	    lookupDiscMgr.setLocators(locators);
	    lookupLocators = lookupDiscMgr.getLocators();
	    // Log the modification
	    addLogRecord(new LookupLocatorsChangedLogObj(lookupLocators));
	} finally {
	    concurrentObj.writeUnlock();
	}
    }


    //////////////////////////////////
    // StorageAdmin Methods
    //////////////////////////////////

    // This method's javadoc is inherited from an interface of this class
    public String getStorageLocation() throws RemoteException {
        if (dbgAdmin != null) {
            dbgAdmin.println("getStorageLocation called");
        }
	concurrentObj.readLock();
	try {
	    return logDirname;
	} finally {
	    concurrentObj.readUnlock();
	}
    }

    // This method's javadoc is inherited from an interface of this class
    public void setStorageLocation(String location)
	throws IOException, RemoteException
    {
        if (dbgAdmin != null) {
            dbgAdmin.println("setStorageLocation called with " + location);
        }

	// Check to see if any work is needed
	// TODO (FCS)- check for absolute pathname?
	if (location.equals(logDirname)) {
	    return;
	}

	concurrentObj.writeLock();
	try {
	    // Move storage location.
	    rename(logDirname, location);

	    // Create new log file to pointing to the new location
 	    ReliableLog newLog = new ReliableLog(location,
						 new LocalLogHandler());
						   
	    // Take a fresh snapshot
	    newLog.snapshot();
	    this.log = newLog;

	    // Update local state
	    logDirname = location;

	    // Reset iterators to point to new storage location
	    Iterator iter = regByID.keySet().iterator();
	    UUID uuid = null;
	    ServiceRegistration reg = null;
	    EventLogIterator eli = null;
	    while (iter.hasNext()) {
		try {
		    uuid = (UUID)iter.next();
		    reg = getServiceRegistration(uuid);
		    eventLogFactory.remove(uuid); // remove cached copy
		    eli = eventLogFactory.iterator(uuid,
		        getEventLogPath(logDirname, uuid));
		    reg.setIterator(eli);
                    if (dbgAdmin != null) {
                        dbgAdmin.println("reset iterator for " + uuid
			   + " to look in " + logDirname);
                    }
		} catch (ThrowThis tt) {
		   // Not a valid Registration -- ignore it.
		}
	    }

	    // Note: logDirname doesn't have to be persisted here
	    // since the activation system is doing it for us.
	    if (activationID != null) {
		try {
		    ActivationSystem sys = ActivationGroup.getSystem();
		    ActivationDesc desc = sys.getActivationDesc(activationID);
		    desc = new ActivationDesc(desc.getGroupID(),
					      desc.getClassName(),
					      desc.getLocation(),
					      new MarshalledObject(location),
					      desc.getRestartMode());
		    sys.setActivationDesc(activationID, desc);
		} catch (ActivationException e) {
		    throw new LogException("activation system problem", e);
		}
	    }
	} finally {
	    concurrentObj.writeUnlock();
	}
    }

    /**
     * Utility method to rename (i.e. move) a file.
     * This method takes a source and destination string. The source
     * file must exist and be readable. The destiantion file must not
     * exist. If either of the previous conditions aren't met then
     * an IOException will be thrown.
     */
    private static void rename(String from, String to) 
        throws IOException 
    {
        // Quick checks to see if we need to proceed.
        if (from == null || to == null)
            throw new IllegalArgumentException(
                "Source and destination arguments must be non-null");

        // Create files and validate
	File src = new File(from);
	File dest = new File(to);
        if (!src.exists() || !src.canRead())
            throw new IOException("Source file is inaccessible: " + src);
	if (dest.exists())
            throw new IOException("Destination file exists: " + dest);

        // Get absolute path names so that equals() works as expected.
        File absSrc = new File(src.getAbsolutePath());
        File absDest = new File(dest.getAbsolutePath());

        // Check to see if source and destination files are the same
        // If so, then this is most likely a programming error, so 
        // don't just ignore it.
        if (absSrc.equals(absDest)) 
            throw new IOException("Source and destination are the same.");

        if (dbgAdmin != null) {
            dbgAdmin.println("Moving file from " + absSrc
		+ " ---> " + absDest);
        }

	// Move the file/directory.
	// We know that dest doesn't exist at this point, so we can
	// assume that the target destination is the same file type as our
	// source object (eg file -> file or dir -> dir).
	if (absSrc.isFile()) {
            if (!absSrc.renameTo(absDest)) { 
		// rename won't copy across filesystems
		// so we need to do it ourselves
		BufferedInputStream in = null;
		BufferedOutputStream out = null;
		int read; 
		try {
		    in = new BufferedInputStream(
			     new FileInputStream(absSrc));
		    out = new BufferedOutputStream(
			     new FileOutputStream(absDest));
		    while((read = in.read()) > -1) {
			out.write(read); 
		    }
		    // Delete source file after its copied
		    if (!absSrc.delete()) 
                        if (dbgAdmin != null) 
                            dbgAdmin.println("Failure deleting " + absSrc);
		} catch (IOException ioe) {
                    if (dbgAdmin != null) {
                        dbgAdmin.println("Failed to rename " + absSrc
		            + " --> " + absDest);
		    }
		} finally {
		    try {
		        if (in != null)
			    in.close();
		    } catch (IOException io1) {}
		    try {
		        if (out != null)
		            out.close();
		    } catch (IOException io2) {}
		}
	    }
	} else { // it's a directory
	    // Renaming directories doesn't always work across filesystems.
	    // Therefore, we attempt to rename it first and if that doesn't
	    // work, we then try to recursively call ourself on the
	    // subtree.
            if (!absSrc.renameTo(absDest)) { // Rename failed, so move files

	        // Ensure destination dir exists before 
		// renaming subtree components
	        FileSystem.ensureDir(absDest.toString());
		if(!absDest.exists())
                    if (dbgAdmin != null) 
                        dbgAdmin.println("Failure creating destination dir: "
			    + absDest);

	        // List subtree components and recurse
	        String[] names = absSrc.list();
	        for (int i = 0; i < names.length; i++) {
	            rename(absSrc + File.separator + names[i], 
		           absDest + File.separator + names[i]);
	        }

	        // Delete directory node which was not 
		// renamed from the source tree
	        if (!absSrc.delete())
                    if (dbgAdmin != null) 
                        dbgAdmin.println(
			    "Failure deleting directory: " + absSrc);
	    }
	}
    }


    /**
     * Utility method that adds the provided EventID
     * to the given registration's "unknown" event list.
     * If the mailbox receives another event, on this registration's
     * behalf, and its EventID is on this list, the serivce will 
     * throw a UnknownEventException back to the event sender.
     */
    private void addUnknownEvent(UUID regID, EventID evid) {
	concurrentObj.writeLock();
	try {
	    //Ensure that the registration is still valid
	    ServiceRegistration reg = (ServiceRegistration)regByID.get(regID);
	    if(reg == null || !leasePolicy.ensureCurrent(reg)) {
		return; // nothing to do ... registration is gone
	    }
            if (dbgDelivery != null) {
                dbgDelivery.println("addUnknownEvent called for: " + reg);
            }

	    // Add EventID to the list
	    reg.getUnknownEvents().put(evid, evid);
	    // Log this event
	    addLogRecord(new UnknownEventExceptionLogObj(regID, evid));

            if (dbgDelivery != null) {
                dbgDelivery.println("UnknownEvents size: " 
                    + reg.getUnknownEvents().size());
            }
	} finally {
	    concurrentObj.writeUnlock();
	}
    }

    /**
     * Utility method that tries to obtain the ServiceRegistration
     * object associated with the given UUID.  If there is
     * no associated ServiceRegistration object, or it has expired,
     * this method will throw a "wrapped" NoSuchObjectException.
     * Note: If the NoSuchObjectException was not wrapped, then
     * the RMI system would wrap into a ServerException (since it's 
     * a RemoteEvent).  The proxies for this service know how to unwrap
     * the ThrowThis object and re-throw the NoSuchObjectException back on
     * the client-side.
     */
    private ServiceRegistration getServiceRegistration(UUID regID) 
        throws ThrowThis
    {
        ServiceRegistration reg = (ServiceRegistration)regByID.get(regID); 
	if(reg == null) { // either expired or never existed
	    throw new ThrowThis( 
		new NoSuchObjectException("Not managing requested " +
	            "registration object"));
	} else if(!leasePolicy.ensureCurrent(reg)) { // check validity
	    throw new ThrowThis( 
		new NoSuchObjectException("Requested registration object" +
		    "has expired"));
	}
	// Must be a valid registration at this point
	return reg;
    }

    /**
     * Utilty method that consolidates registration removal activities.
     * It's called when registration's are either expired or cancelled.
     */
    private void removeRegistration(UUID regID, ServiceRegistration reg) {
	if (dbgLeases != null) {
	    dbgLeases.println("Removing --> " + reg);
	}
 
        // Remove Reg from data structures, if present
        regByID.remove(regID);
	regByExpiration.remove(reg);
	boolean exists = pendingReg.remove(regID);
	NotifyTask task = (NotifyTask)activeReg.remove(regID);
	if (task != null) { // cancel active task, if any
	    task.cancel();
	    if (dbgLeases != null) {
	        dbgLeases.println("Cancelling active notification task for "
	            + regID);
	    }
	}

	// Delete any associated resources
	try {
	    if (dbgLogs != null) 
	        dbgLogs.println("Removing logs for " + reg);
	    EventLogIterator iter = reg.iterator();
	    if (iter != null) // iter is null when recovering state
		iter.destroy();
	} catch (IOException ioe) {
	    if (dbgLogs != null)
	        dbgLogs.println("Trouble removing logs for " + reg);
	    // Did the best we could ... continue.
	}


	// Sanity check
	if (exists && task != null) {
	    if (dbgLeases != null) {
	        dbgLeases.println("ERROR: Registration was found "
	            + "on both the active and pending lists");
	    }
	    // TODO (FCS)- throw internal exception
	}
    }

    /**
     * Utility method that returns the associated File object
     * for the given UUID's persistence directory 
     */
    private static File getEventLogPath(String parent, UUID uuid) {
        return new File(parent, uuid.toString());
    }

    /** Actual implementation of the registration process. */
    private Registration registerDo(long duration) {
        // Create new ServiceRegistration object
        UUID uuid = uuidFactory.newUUID(); 
        EventLogIterator iter = 
	    eventLogFactory.iterator(uuid, getEventLogPath(logDirname, uuid)); 
        ServiceRegistration reg = new ServiceRegistration(uuid, iter);
        Lease l = null;
        try {
            l = leasePolicy.leaseFor(reg, duration);
	} catch (LeaseDeniedException lde) {
	    // Should never happen in our implementation.
	    // Re-throw as an internal exception.
	    throw new InternalMailboxException("Registration lease was denied", 
		                               lde);
	}

        // Add registration to internal data structures 
        addRegistration(reg);

        // log a record of this event
        addLogRecord(new RegistrationLogObj(reg));

        // Check expiration and notify expiration thread to 
        // wake up earlier than scheduled, if needed.
	if (reg.getExpiration() < minRegExpiration) {
	    minRegExpiration = reg.getExpiration();
            if (dbgExpiration != null) {
                dbgExpiration.println("Notifying expiration thread");
            }
	    concurrentObj.waiterNotify(expirationNotifier);
	}

	if(dbgLeases != null) {
	    dbgLeases.println("Generated new lease for: ");
	    reg.dumpInfo(dbgLeases);
	}

        // Return client-side proxy for the ServiceRegistration
	return new Registration(uuid, this, l);
    }

    /** 
     * Utility method used to add a registration to our state information. 
     * This method is invoked from <code>registerDo()</code> as well as
     * <code> RegistrationLogObj.apply()</code>.  
     */
    private void addRegistration(ServiceRegistration reg) {
	regByID.put(reg.getCookie(), reg);
	regByExpiration.put(reg, reg);
    }

    /** Performs the actual registration renewal logic */
    private long renewDo(Object cookie, long extension)
        throws UnknownLeaseException 
    {
        ServiceRegistration reg = null;
	long expiration = 0;
        
	if(dbgLeases != null) {
	    dbgLeases.println("Attempting to renew " + cookie 
	        + "'s lease for " + extension/1000 + " sec"); 
	}
	// Get registration of interest
	try { 
            // Note: the following method will throw a ThrowThis exception
            // if the registration is invalid (i.e. expired or non-existent)
	    reg = getServiceRegistration((UUID)cookie);
	    // delegate renewal to policy object
	    expiration = leasePolicy.renew(reg, extension);
	    // Log this event
	    addLogRecord(new RegistrationRenewedLogObj((UUID)cookie, 
	                                               reg.getExpiration()));

	    // remove then add to cause a resort to occur
	    regByExpiration.remove(reg);
	    regByExpiration.put(reg, reg);
	} catch (ThrowThis tt) {
	    // Registration doesn't exist or has expired
	    if(dbgLeases != null) {
	        dbgLeases.println("Lease for " + cookie + " was NOT renewed");
	    }
	    throw new UnknownLeaseException("Not managing requested lease");
	}

	if(dbgLeases != null) {
	    dbgLeases.println("Lease for " + cookie + " was renewed");
	    reg.dumpInfo(dbgLeases);
	}

	return expiration;
    }
    
    
    /** Performs the actual registration cancellation logic */
    private void cancelDo(Object cookie) throws UnknownLeaseException {
        ServiceRegistration reg = null;

    	if(dbgLeases != null) {
    	    dbgLeases.println("Attempting to cancel lease for: " + cookie); 
    	}

	try { 
            // Note: the following method will throw a ThrowThis exception
            // if the registration is invalid (i.e. expired or non-existent)
	    reg = getServiceRegistration((UUID)cookie);
	    // Remove registration from internal data structures
            removeRegistration((UUID)cookie, reg);
            // Log this event
	    addLogRecord(new RegistrationCancelledLogObj((UUID)cookie));
	    // Notify expiration thread in case we deleted the registration
	    // nearest to expiration, so it can recompute minRegExpiration.
            if (reg.getExpiration() == minRegExpiration) {
                if (dbgExpiration != null) {
                    dbgExpiration.println("Notifying expiration thread");
                }
	        concurrentObj.waiterNotify(expirationNotifier);
	    }
	} catch (ThrowThis tt) {
	    // Registration doesn't exist or has expired
	    throw new UnknownLeaseException("Not managing requested lease");
	}
    	if(dbgLeases != null) {
    	    dbgLeases.println("Lease cancelled for: " + cookie); 
    	}
    }
    
    /** Performs the actual renewAll logic */
    private RenewResults renewAllDo(Object[] cookie, long[] extension)
	    throws RemoteException 
    {
	int count = cookie.length;

	if(dbgLeases != null) {
	    dbgLeases.println("Attempting to renew a batch of " +
			      count + " leases"); 
	}

	// Delegate functionality to Landlord utility
	final RenewResults rslt = LandlordUtil.renewAll(localLandlord,
							cookie,	extension);

	if(rslt.denied == null)  {
	    if(dbgLeases != null) {
	        dbgLeases.println("Batch renew totally successful");
	    }
	} else {
	    if(dbgLeases != null) {
	        dbgLeases.println("Batch renew contained exceptions");
	    }
	}

	return rslt;
    }	     

    /** Performs the actual cancelAll logic */
    private Map cancelAllDo(Object[] cookie) throws RemoteException {
	int count = cookie.length;
	if(dbgLeases != null) {
	    dbgLeases.println("Attempting to cancel a batch of " +
		count + " leases"); 
	}

	// Delegate functionality to Landlord utility
	return LandlordUtil.cancelAll(localLandlord, cookie);
    }
	
    /** Performs the actual enable delivery logic */
    private void enableDeliveryDo(UUID uuid, RemoteEventListener target) 
	throws ThrowThis 
    {
        // Proxy already checks for target == null, but just in case ...
        if (target == null) {
            //TODO (FCS)- Throw internal exception
        }

	// Check for the resubmission of one of our own listeners
	if (target instanceof ListenerProxy) {
	    UUID id = ((ListenerProxy)target).getID(); 
	    if (regByID.get(id) != null) {
	        if (dbgDelivery != null) {
	            dbgDelivery.println("Disallowing resubmission of one of" 
		       +  " this service's own listeners: " + id); 
	        }
	        throw new IllegalArgumentException("Cannot submit a " +
		    "listener that was provided by the same EventMailbox " +
		    "service");
	    }
	}

	// Enable the listener for the given UUID
	enableRegistration(uuid, target);
	// Log this event
	addLogRecord(new RegistrationEnabledLogObj(uuid, target));

	// Notify the notifier thread of new "enabled" entry
	concurrentObj.waiterNotify(eventNotifier);

	if (dbgDelivery != null) {
	    dbgDelivery.println("Enabling delivery for " + uuid +
		" to " + target);
	}
    }

    /** 
     * Utility method that associates the given listener with the
     * associated registration object. 
     */
    private void enableRegistration(UUID uuid, RemoteEventListener target) 
	throws ThrowThis 
    {
        // Note: the following method will throw a ThrowThis exception
        // if the registration is invalid (i.e. expired or non-existent)
        ServiceRegistration reg = getServiceRegistration(uuid); 
        // Set the event target
	reg.setEventTarget(target); 

	// Clear unknownEvent collection regardless of whether the same target 
	// is provided. We'll assume that an active re-registration means 
	// that the client's listener has changed (for the better) somehow.
	reg.getUnknownEvents().clear();

        // Add this registration to set of enabled registrations
        // if it's not already there. Note that an enabled registration
        // can be in either the pendingReg or activeReg data structures
        // at any one time.
	if (!pendingReg.contains(uuid) && !activeReg.containsKey(uuid))
	    pendingReg.add(uuid);
    }

    /**
     * Performs the actual logic for disabling event delivery
     * for a particular registration.
     */
    private void disableDeliveryDo(UUID uuid) throws ThrowThis {
        // Disable event delivery for the associated registration
        disableRegistration(uuid);
        // Log this event
        addLogRecord(new RegistrationDisabledLogObj(uuid)); 
    }
        
    /**
     * Utility method that disables event delivery for the
     * registration associated with the given <tt>UUID</tt>
     */
    private void disableRegistration(UUID uuid) throws ThrowThis {
	if (dbgDelivery != null) {
	    dbgDelivery.println("Attempting to disable delivery for " + uuid);
	}

        // get associated registration object
        // Note: the following method will throw a ThrowThis exception
        // if the registration is invalid (i.e. expired or non-existent)
        ServiceRegistration reg = getServiceRegistration(uuid); 

        // cache pre-disabled value for debugging purposes
        RemoteEventListener rel = reg.getEventTarget();

        // Set delivery target to null
	reg.setEventTarget(null); 

	// Remove registration from pending and active lists
	boolean inPending = pendingReg.remove(uuid);  
	NotifyTask task = (NotifyTask)activeReg.remove(uuid);
	if (task != null) { // cancel active task, if any
	    task.cancel();
	    if (dbgDelivery != null) {
	        dbgDelivery.println("Cancelling active notification task for " 
	            + uuid);
	    }
	}

        // If rel != null then the registration was previously enabled.
        // If so, then the its UUID should have appeared in either
        // the active or pending data structures. This can happen 
        // during recovery but otherwise there's a problem.
	if (rel != null && !inPending && task == null && !inRecovery) {
	    if (dbgDelivery != null) {
	        dbgDelivery.println("*** Registration was not found "
	            + "on the active or pending lists");
	    }
	    // TODO (FCS)- throw internal exception
	}

	if (dbgDelivery != null) {
	    dbgDelivery.println("Disabled delivery for " + uuid);
	}
    }

    /**
     * Performs the actual logic of handling received events and 
     * storing them.
     */
    private void notifyDo(UUID registrationID, RemoteEvent theEvent) 
	throws UnknownEventException, RemoteException, ThrowThis
    {
        if (dbgReceive != null) {
            dbgReceive.println("Received event for " + registrationID);
	    dumpEvent(theEvent, dbgReceive);
	}

        // Note: the following method will throw a ThrowThis exception
        // if the registration is invalid (i.e. expired or non-existent)
        ServiceRegistration reg = getServiceRegistration(registrationID); 

	// Check if we need to propagate an UnknownEventException back to
	// the event sender.
	// This occurs whenever the mailbox service previously received
	// an UnknownEventException while delivering an event with
	// the same eventID for the same registration,listener pair.
	EventID evtid = new EventID(theEvent);
	if (reg.getUnknownEvents().containsKey(evtid)) {
            if (dbgReceive != null) {
                dbgReceive.println("Discarding event -- it's unknown");
	    }
	    // TODO (FCS)- store exception state from event target and retransmit 
	    // to event sender.
	    throw new UnknownEventException();
	}

        // Store event with appropriate registration
	try {
	    reg.iterator().add(theEvent);
	} catch (IOException ioe) {
	    // TODO (FCS)- pick better recovery algorithm
	    // Assumptions: 
	    //    - that any IOExceptions are transient in nature,
	    //      so keep trying.
	    //    - event logging state is still valid in the face
	    //      of an IOException
	}

        // Notify the notifier if this registration is enabled
	if (reg.getEventTarget() != null) {
	    concurrentObj.waiterNotify(eventNotifier);
	}

        if (dbgReceive != null) {
	    reg.dumpInfo(dbgReceive);
	}
	
    }

    /**
     * Utility method for setting the service's ID obtained from
     * the lookup registration process.
     */
    private void setServiceID(ServiceID id) {
	if (dbgInit != null) {
	    dbgInit.println("serviceID being set to " + id);
	}
        serviceID = id;
    }

    ///////////////////////////
    // Inner Classes
    //////////////////////////

    /** 
     * Class which handles the callback of a service ID assignment
     * by the JoinManager
     */
    private class SrvcIDListener implements ServiceIDListener
    {
        /** Simple constructor */
        public SrvcIDListener() {
            super();
        }

        /** 
         * The JoinManager will invoke this method when it receives a
         * valid ServiceID from a lookup service.
         */
        public void serviceIDNotify(ServiceID id) {
            // Set the ID
            setServiceID(id);
            // Log this event
            addLogRecord(new ServiceIDLogObj(id));
        }
    }//end class SrvcIDListener



    /**
     * The notifier thread.  This thread is responsible for delivering events
     * to client supplied notification targets when their associated 
     * registrations are enabled. 
     * <p>
     * The constructor calls <code>start()</code> so there is no need 
     * to explicitly start this thread.
     */
    // Implementation details:
    // Event delivery is done on a per registration basis.  
    // A TaskManager is used
    // to pool event delivery tasks, but there will only be one active task for
    // any one registration. This allows us to ensure "in order" delivery of 
    // events. 
    private class Notifier extends Thread implements TimeConstants {
        /** 
         * <code>TaskManager</code> that will be handling the notification tasks 
         */
        private TaskManager	taskManager;	
    
        /** 
         * Random number generator that will be used for implementing
         * a simple load balancing scheme. Seed it with the current time.
         */
        private Random rand = new Random(System.currentTimeMillis());
    
        /** Time to wait between notification checks */
        private final static long	PAUSE_TIME = 5000; // 5 seconds
    
        /**
         * Simple constructor.
         */
        Notifier() {
    	    super("Notifier");
    	    taskManager = new TaskManager();
    	    start();
        }
    
        /**
         * Schedule delivery tasks for any enabled registrations.
         */
        public void run() {
	    try {
		concurrentObj.writeLock();
	    } catch (ConcurrentLockException e) {
		return;
	    }

	    try {
	        while (!isInterrupted()) { 
    	            int count = pendingReg.size();
    	            if (dbgDelivery != null) {
		        dbgDelivery.println(getName() + " checking " + count 
		            + " possible registrations");
		    }
    	            // don't need to check for size() > 0 (below) since
    	            // we hold the lock. 
    	            // TODO: optimize loop to ignore indices w/o available events
    	            // TODO: optimize loop to skip nextInt if only one choice
    	            while (count-- > 0) { 
    	                int index = rand.nextInt(pendingReg.size());
    	                UUID uuid = (UUID)pendingReg.get(index);
    	                ServiceRegistration reg = null;
        	        try { 
                            // Note: the following method will throw a 
                            // ThrowThis exception if the registration is 
                            // invalid (i.e. expired or non-existent)
                            reg = getServiceRegistration(uuid); 
                            // Check if the registration has any events to 
                            // be delivered
    	                    if (reg.iterator().hasNext()) {
    	                        if (dbgDelivery != null) {
		                    dbgDelivery.println(getName() 
		                        + " Scheduling"
		                        + " delivery task for " + reg); 
				}
				// Create and schedule a event delivery task
    	                        NotifyTask t = new NotifyTask(taskManager, uuid);
    	                        taskManager.add(t);
    	                        // Put registration onto active list
    	                        activeReg.put(uuid, t);
    	                        // Remove registration from pending list
    	                        pendingReg.remove(index);
    		            }
			} catch (ThrowThis tt) {
			    // Invalid registration ... skip
    	                    if (dbgDelivery != null) {
				dbgDelivery.println(getName()
				    + " invalid registration for " + uuid);
			    }

			} catch (IOException ioe) {
			    // Could not access registration state  ... skip
    	                    if (dbgDelivery != null) {
				dbgDelivery.println(getName()
				    + " inaccessible registration data for " 
				    + uuid);
			    }
			}
    		    }
		    try {
                        if (dbgDelivery != null ) {
                            dbgDelivery.println(getName()
                                + " delayed until " 
                                + new Date(PAUSE_TIME + 
                                           System.currentTimeMillis()));
			}
			// There is a potential for missing "notify" calls
			// so we can't just use a large timeout value.
			concurrentObj.writerWait(eventNotifier, PAUSE_TIME);
		    } catch (ConcurrentLockException e) {
			return;
		    }
		}

    	    } finally {
    	        concurrentObj.writeUnlock();
		// If exiting because of a thread interrupt,
		// it means we are shutting down so try to clean
		// up after ourselves.
		if (isInterrupted()) {
                    if (dbgDelivery != null ) {
                        dbgDelivery.println(getName() 
			    + " terminating taskManager");
		    }
    	            taskManager.terminate();
		} 
                if (dbgDelivery != null ) {
                    dbgDelivery.println(getName()
		       + " exiting ...");
		}
    	    }
        }
    }
    
    /*
     * Static variables for NotifyTask. Unfortunately, we can't have 
     * static references in an inner class (oterwise they'd be in
     * NotifyTask itself).
     */
    /** The maximum time to maintain a notification task */
    private static final long	MAX_TIME = 1 * HOURS;

    /** The maximum mnumber of times to retry event delivery */
    private final static int	MAX_ATTEMPTS = 5;	// max times to retry
    
    /**
     * A task that represents an event notification task for a particular
     * registration. A NotifyTask will be returned to the task pool after
     * one of the following has occured (which ever comes first):
     * <ul>
     * <li> The event was successfully delivered
     * <li> The task expires after <tt>MAX_TIME</tt> has elapsed
     * <li> The task quits after <tt>MAX_ATTEMPTS</tt> unsuccessful delivery 
     *      attempts
     * <li> The event could not be successfully retrieved from the 
     *      registration's event log iterator.
     * </ul>
     * 
     */
    class NotifyTask extends RetryTask {
        /** The UUID of the associated registration */
        private UUID	regID;

    	/**
    	 * Create an object to represent an event notification task.
    	 */
    	NotifyTask(TaskManager tm, UUID regID) {
    	    super(tm);
    	    this.regID = regID;
    	}

    	/**
    	 * This utility function tries to obtain the next event for the
    	 * given registration. If it finds one, the event is returned. 
    	 * Otherwise null is returned.
    	 */
    	// Note that next() will return the same event on subsequent calls
    	// unless remove() is called (indicating sucessful delivery).
    	private RemoteEvent getNextEvent(ServiceRegistration reg) {
    	    RemoteEvent evt = null;
	    try {
    	        if (reg.iterator().hasNext()) {
		    try {
		        evt =  (RemoteEvent)reg.iterator().next();
		    } catch (IOException ioe) {
		        //just return null in this case.
		    } catch (ClassNotFoundException cnfe) {
		        //just return null in this case.
		    }
		}
	    } catch (IOException ioe) {
		//just return null in this case.
	    }
	    return evt;
    	}

    	/**
    	 * This utility function atempts to remove the event after
    	 * from the associated event registration's storage after
    	 * 1) a successful delivery attempt
    	 * 2) task has expired
    	 * 3) the event type caused an earlier UnknownEventException
    	 * 4) the event delivery resulted in an UnknownEventException
    	 */
    	private void deleteNextEvent(ServiceRegistration reg) {
	    try {
    	        if (reg.iterator().hasNext()) {
		    try {
		        reg.iterator().remove();
		    } catch (IOException ioe) {
	                if (dbgDelivery != null) {
	                    dbgDelivery.println( getClass().getName() + 
			        " could not deleteNextEvent for" + reg);
			}
		    }
	        } 
	    } catch (IOException ioe) {
	        if (dbgDelivery != null) {
	            dbgDelivery.println(getClass().getName() 
			+ " could not delete event becuase state info"
			+ " for " + reg + " was inaccessible");
		}
	    }
    	}    

    	/**
    	 * This utility function attempts to disable the given registration
    	 * because there was a problem with the provided listener.
    	 */
    	private boolean disableRegistration(UUID regID, RemoteEventListener l) {
    	    boolean disabled = true;
	    concurrentObj.writeLock();
	    try {
	        try {
                    // Note: the following method will throw a ThrowThis exception
                    // if the registration is invalid (i.e. expired or non-existent)
	            ServiceRegistration reg = getServiceRegistration(regID); 

	            // Check to see if the listener has changed since 
	            // the initial call to this method. If so, 
	            // assume that the new listener doesn't need to be
	            // disabled.  If not, continue with the disabling
	            // of the registration.
		    RemoteEventListener currentListener = reg.getEventTarget();
		    if (currentListener != null && currentListener.equals(l)) {
		        disableDeliveryDo(regID);
		    } else {
	                if (dbgDelivery != null) {
	                    dbgDelivery.println("Disabling registration for " 
	                        + regID + " skipped due to listener change.");
	                }
	                disabled = false;
		    }
		} catch (ThrowThis tt) {
		    // registration is gone ... ignore
		}
	    } finally {
	        concurrentObj.writeUnlock();
	    }
	    return disabled;
    	}  

    	/**
    	 * Try to notify the target.  Return <code>true</code> if the
    	 * notification was successful. Return <code>false</code> otherwise.
    	 * If we return false, this task will be rescheduled at some
    	 * point in the future.
    	 */
    	public boolean tryOnce() {
    
    	    boolean succeeded = false;           // attempt this task again?      
    	    boolean deleteEvent = false;         // delete the retrieved event? 
    	    boolean doNotify = true;             // attempt notification?
    	    ServiceRegistration reg = null;      // 
    	    RemoteEventListener listener = null; // event delivery target
	    RemoteEvent ev = null;               // event to deliver

    	    if (dbgDelivery != null) {
		dbgDelivery.println("Attempting event delivery for: " + regID
		    + " at " + new java.util.Date(System.currentTimeMillis()));
	    }
    
            // Put an upper time limit on how long to keep
            // this task on the active list.  This prevents
            // slow/bad event recipients from monopolizing
            // a thread in the thread pool. This will not 
            // prevent a "malicious" event recipient from 
            // hanging up this thread if it never returns
            // from the notify() call.
	    long curTime = System.currentTimeMillis();
    	    if (curTime - startTime() > MAX_TIME) {
		succeeded = true;
		deleteEvent = false;
    	        doNotify = false;
    	        if (dbgDelivery != null) {
		    dbgDelivery.println("Cancelling delivery due to time " +
		        "limit expiration.");
	        }
	    } else {
                // Get required delivery information
                concurrentObj.readLock();
        	try {
        	    try { 
                        // Note: the following method will throw a 
                        // ThrowThis exception if the registration 
                        // is invalid (i.e. expired or non-existent)
                        reg = getServiceRegistration(regID); 
        		listener = reg.getEventTarget();
        		ev = getNextEvent(reg);
    
                        // Check to see if either the listener or the event 
                        // is null. If so, skip the notify() call.
                        // If the event is not null, make an 
                        // additional check to see if the event is on the 
                        // unknown event list for this registration.
                        // If so, then skip the notify() call as well.
        		if (listener == null || ev == null) {
        		    succeeded = true;    // don't try again
        		    // Note that if getNextEvent returned null
        		    // then there was a problem reading or extracting
        		    // the event. If so, the EventLog mechanism
        		    // wil have already advanced past the offending
        		    // event, so we don't have to remove it.
        		    deleteEvent = false; 
        		    doNotify = false;    // skip notify
    	                    if (dbgDelivery != null) {
		                dbgDelivery.println("Cancelling delivery " +
		                   " because of null listener or event");
	                    }
			} else if (ev != null &&
			           reg.getUnknownEvents().containsKey(
			               new EventID(ev))) {
			    // If this event type caused an 
			    // UnknownEventException in the past, then
			    // delete it since it likely do so again.
        		    succeeded = true;    // don't try again
        		    deleteEvent = true;  // delete event
        		    doNotify = false;    // skip notify
    	                    if (dbgDelivery != null) {
		                dbgDelivery.println("Cancelling delivery " +
		                   " because of unknown event");
	                    }
        		}
		    } catch (ThrowThis tt) { // reg was not valid ... skip it
    		        succeeded = true;        // don't try again
    		        deleteEvent = false;     // don't remove event
        		doNotify = false;        // skip notify
    	                if (dbgDelivery != null) {
		            dbgDelivery.println("Cancelling delivery " +
		                " because of unknown registration");
	                }
    		    }
        	} finally {
                    concurrentObj.readUnlock();
        	}
	    } // end else
    	    
    	    // Important - don't hold any locks during a remote invocation
    	    if (doNotify) {
    	        if (dbgDelivery != null) {
		    dbgDelivery.println("Delivering evt: " + ev);
		    dbgDelivery.println("\tID#:  " + ev.getID());
		    dbgDelivery.println("\tSeq#: " + ev.getSequenceNumber());
	        }
		try {
		    // Notify target listener and note a successful delivery
		    listener.notify(ev);
		    succeeded = true;
    		    deleteEvent = true;
    	            if (dbgDelivery != null) {
		        dbgDelivery.println("Delivery was successful");
		    }
		} catch (UnknownEventException e) {
		    // Target wasn't expecting this event, so prevent
		    // future notifications of the same event type.
    	            if (dbgDelivery != null) {
		        dbgDelivery.println("Caught UnknownEventException " +
		            " during notify");
	            }
	            // Add offending event type to this registration's
	            // "unknown" list and move on to the next task.
		    addUnknownEvent(regID, new EventID(ev));
		    succeeded = true;
		    deleteEvent = true;
	        } catch (RemoteException re) {
	            if (((re instanceof ServerException) &&
	                 (re.detail != null && 
	                  re.detail instanceof NoSuchObjectException)) ||
	                re instanceof NoSuchObjectException) 
		    {
		        // Definite remote exception means there is
		        // no possibility that a retry attempt will
		        // succeed (ie NoSuchObjectException).
    	                if (dbgDelivery != null) {
		            dbgDelivery.println("Caught a definite " +
		                " RemoteException during notify");
	                }
	                // Disable event delivery for this particuler
	                // target listener and move onto the next task
		        disableRegistration(regID, listener);
		        succeeded = true;
		        deleteEvent = false;
		    } else {
    	                if (dbgDelivery != null) {
		            dbgDelivery.println("Caught an indefinite " +
		                " RemoteException during notify");
	                }
		        // Indefinite remote exception means there is
		        // possibility that a retry attempt will succeed
		        succeeded = false;
		        deleteEvent = false;
		    }
		} // end catch    
	    } // end if
    
            // If we still aren't successful after MAX_ATTEMPTS
            // then give up
    	    if (!succeeded && attempt() > MAX_ATTEMPTS) {
    	        if (dbgDelivery != null) {
		    dbgDelivery.println("Maximum delivery attempts reached");
		}
		succeeded = true;		
		deleteEvent = true;
	    }
    
            // Check if data structures need to be modified
            // If so, make the change(s) and notify the notifier thread.
            if (succeeded || deleteEvent) {
                concurrentObj.writeLock();
        	try {
    		    try { 
                        // Note: the following method will throw a ThrowThis 
                        // exception if the registration is invalid 
                        // (i.e. expired or non-existent)
                        reg = getServiceRegistration(regID); 
                        if (succeeded) { 
                            // If we are still enabled, then move this reg
                            // from active --> pending list
                            if (reg.getEventTarget() != null) { 
    	                        if (dbgDelivery != null) {
		                    dbgDelivery.println("Putting task back "
		                        + "onto pending list");
		                }
        	                activeReg.remove(regID);
        	                pendingReg.add(regID);
			    } else { // reg is disabled 
    	                        if (dbgDelivery != null) {
		                    dbgDelivery.println("Removing task ...");
				}
				// Note: disabling delivery should
				// have taken care of this for us.
        	                if (activeReg.remove(regID) != null ||
        	                    pendingReg.remove(regID) == true  ) {
    	                            if (dbgDelivery != null) {
		                        dbgDelivery.println("ERROR: Found "
		                            + "pending/active task for a "
		                            + "disabled registration"); 
				    }
				    // TODO (FCS)- throw internal exception
				}
			    }
    		        }

        	        if (deleteEvent) {
    	                    if (dbgDelivery != null) {
		                dbgDelivery.println("Deleting event ...");
			    }
        	            deleteNextEvent(reg); 
    		        }
    		    } catch (ThrowThis tt) { 
    		        // Registration is gone ... nothing to do
    		    }
        	} finally {
                    concurrentObj.writeUnlock();
        	}
    	        if (dbgDelivery != null) {
		    dbgDelivery.println("Waking up notifier");
		}
	        concurrentObj.waiterNotify(eventNotifier);
	    }

    	    return succeeded;
    	}
    
        /**
         * Return <code>true</code> if this task needs to runAfter any
         * of the tasks in the provided list and <code>false</code>
         * otherwise.  The notifier thread ensures that there should only
         * be one task per registration, so this method just returns
         * false.
         */
    	public boolean runAfter(java.util.List list, int max) {
    	    return false;
       	}
    }


    /**
     * Termination thread code.  We do this in a separate thread to
     * avoid deadlock, because Activatable.inactive will block until
     * in-progress RMI calls are finished.
     */
    private class DestroyThread extends Thread {

	/** Create a non-daemon thread */
	public DestroyThread() {
	    super("DestroyThread");
	    /* override inheritance from RMI daemon thread */
	    setDaemon(false);
	}

	public void run() {
	    if (dbgAdmin != null) {
		dbgAdmin.println("DestroyThread started ...");
	    }

	    /* must unregister before unexport */
	    if (activationID != null) {
		try {
		    Activatable.unregister(activationID);
		} catch (RemoteException e) {
		    /* give up until we can at least unregister */
		    return;
		} catch (ActivationException e) {
		}
	    }
	    try {
	        /* wait for any pending operations to complete */
		while (!UnicastRemoteObject.unexportObject(MailboxImpl.this,
							   false))
		{
		    Thread.yield();
		}
	    } catch (NoSuchObjectException e) {
	    }

	    /* all daemons must terminate before deleting persistent store */
	    if (dbgAdmin != null) {
		dbgAdmin.println("Terminating JoinManager ...");
	    }
	    joiner.terminate();

	    if (dbgAdmin != null) {
		dbgAdmin.println("Interrupting Notifier ...");
	    }
	    notifier.interrupt();
	    
	    if (dbgAdmin != null) {
		dbgAdmin.println("Interrupting Expirer ...");
	    }
	    expirer.interrupt();

	    if (dbgAdmin != null) {
		dbgAdmin.println("Interrupting Snapshotter ...");
	    }
	    snapshotter.interrupt();

	    try {
	        if (dbgAdmin != null) {
		    dbgAdmin.println("Waiting for Notifier ...");
	        }
	        notifier.join();

	        if (dbgAdmin != null) {
		    dbgAdmin.println("Waiting for Expirer ...");
	        }
	        expirer.join();	        

	        if (dbgAdmin != null) {
		    dbgAdmin.println("Waiting for Snapshotter ...");
	        }
	        snapshotter.join();
	    } catch (InterruptedException e) {
	    }


	    // We don't need to acquire any locks at this point 
	    // since there shouldn't be any other threads modifying data.
	    // We don't check for expiration since any regs in the
	    // table that haven't been explicity expired probably
	    // still have associated resources that need to be freed.
	    Iterator iter = regByID.values().iterator();
	    ServiceRegistration reg = null;
	    EventLogIterator logIter = null;
	    while (iter.hasNext()) {
                reg = (ServiceRegistration)iter.next();
                try {
                    if (dbgLogs != null)
                        dbgLogs.println("Destroying logs for -> " + reg);
                    reg.iterator().destroy();
	            logIter = reg.iterator();
	            if(logIter != null) 
		        logIter.destroy();
                } catch (IOException ioe) {
                    if (dbgLogs != null)
                        dbgLogs.println("Destroy unsuccessful for -> " 
                            + reg);
		    // Did the best we could ... continue.
                }
	    }

            // This call actually tries to delete the persistence
	    // directory, so it has to be last.
	    log.deletePersistentStore();

	    if (activationID != null) {
		/* inactive will set current group ID to null */
		ActivationGroupID gid = ActivationGroup.currentGroupID();
		try {
		    while (!Activatable.inactive(activationID)) {
			Thread.yield();
		    }
		} catch (RemoteException e) {
		} catch (ActivationException e) {
		}
		try {
		    ActivationGroup.getSystem().unregisterGroup(gid);
		} catch (RemoteException e) {
		} catch (ActivationException e) {
		}
	    }
	    if (dbgAdmin != null) {
		dbgAdmin.println("DestroyThread finished ...");
		//dbgAdmin.println("DestroyThread sleeping 10s...");
		//try{ Thread.sleep(10000); } catch (Exception e) {}
	    }
	}
    }

    /** Registration expiration thread code */
    private class ExpirationThread extends Thread implements TimeConstants {

	/** Create a daemon thread and start it */
	public ExpirationThread() {
	    super("ExpirationThread");
	    setDaemon(true);
            if (dbgExpiration != null ) {
                dbgExpiration.println(getName() + " started ...");
            }
            start();
	}

	/** 
	 * Loop over time-sorted collection and remove any entries
	 * that have expired.
	 */
	public void run() {
            ServiceRegistration reg = null;
            long delay = 0;

	    try {
		concurrentObj.writeLock();
	    } catch (ConcurrentLockException e) {
		return;
	    }

	    try {
	        while (!isInterrupted()) {
	            // Get current time
		    long now = System.currentTimeMillis();
                    if (dbgExpiration != null ) {
                        dbgExpiration.println(getName() 
                                + " Checking regs @ " + new Date(now));
		    }
		    // loop over time-sorted collection (by expiration) 
		    while (!regByExpiration.isEmpty()) {
		        // Collection ordered by expiration time
		        // (earliest -> latest)
                        reg = (ServiceRegistration)regByExpiration.firstKey(); 
			minRegExpiration = reg.getExpiration();
                        if (dbgExpiration != null ) {
                            dbgExpiration.println(getName() 
                                + " Checking " + reg 
                                + " which expires @ " 
                                + new Date(minRegExpiration));
                        }

			// If earliest one is in the future
			// then break out and wait until then
			if (minRegExpiration > now) 
			    break; 
			// Otherwise remove expired registration
                        removeRegistration((UUID)reg.getCookie(), reg);
		    }
		    try {
		        // Calculate delay
		        delay = minRegExpiration - now;
		        delay = (delay >= 0) ? delay : Long.MAX_VALUE;
                        if (dbgExpiration != null ) {
                            dbgExpiration.println(getName()
                                + " delayed until " + new Date(delay + now));
			}
			// Wait for delay (or until we get an asynchronous
			// notification from some other thread).
			// This call also releases the lock. 
			concurrentObj.writerWait(expirationNotifier, delay);
		    } catch (ConcurrentLockException e) {
			return;
		    }

		}
	    } finally {
	        concurrentObj.writeUnlock();
                if (dbgExpiration != null ) {
                    dbgExpiration.println(getName() + " exiting ...");
	        }
            }

	}
    } // End ExpirationThread


    /** 
     * Adaptor class implementation of LocalLandlord.  We use this
     * adaptor class with LandlordUtil because we want to only acquire
     * the concurrentObj.writeLock() once per batch lease operation.
     */
    private class LocalLandlordAdaptor implements LocalLandlord {
	// inherit javadoc from supertype
	public long renew(Object cookie, long extension)
	    throws LeaseDeniedException, UnknownLeaseException 
	{
	    return renewDo(cookie, extension);
	}
       
	// inherit javadoc from supertype
	public void cancel(Object cookie) throws UnknownLeaseException {
	    cancelDo(cookie);
	}
    }


    ////////////////////////////////
    // Persistence Methods/Classes
    ////////////////////////////////

    /**
     * Interface defining the method(s) that must be implemented by each of
     * the concrete LogObj classes. This allows for the definition of
     * object-dependent invocations of the appropriate implementation of
     * the method(s) declared in this interface.  
     */
    private static interface LogRecord extends Serializable {
	void apply(MailboxImpl mb);
    }

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * the serviceID notification from the lookup service occurs (which
     * should only happen once per instantiation).
     */
    private static class ServiceIDLogObj implements LogRecord {
	// TODO (FCS)- generate serialVersionUID

	/**
	 * The service ID.
	 *
	 * @serial
	 */
	private ServiceID id;

	/** Simple constructor */
        public ServiceIDLogObj(ServiceID id) {
            this.id = id;
        }

	/**
	 * Sets the mailbox's service ID to this object's stored value.
	 */
	public void apply(MailboxImpl mb) {
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Applying a " + getClass().getName());
	    }
            mb.setServiceID(id);
	}
    }

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * a new registration is created.
     */
    private static class RegistrationLogObj implements LogRecord {

	// TODO (FCS)- generate serialVersionUID

	/**
	 * The registration.
	 *
	 * @serial
	 */
	private ServiceRegistration reg;

	/** Simple constructor */
	public RegistrationLogObj(ServiceRegistration reg) {
	    this.reg = reg;
	}

	/**
	 * Adds the stored registration object to the mailbox's state.
	 */
	public void apply(MailboxImpl mb) {
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Applying a " + getClass().getName());
	        reg.dumpInfo(dbgRecovery);
	    }
	    mb.addRegistration(reg);
	}
    }

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * a registration is enabled.
     */
    private static class RegistrationEnabledLogObj implements LogRecord {

	// TODO (FCS)- generate serialVersionUID

	/**
	 * The registration ID.
	 *
	 * @serial
	 */
	private UUID regID;

	/**
	 * The registration ID.
	 *
	 * @serial
	 */
	private transient RemoteEventListener target;

        /** Simple constructor */
        public RegistrationEnabledLogObj(UUID id, RemoteEventListener l) {
            regID = id;
            target = l;
        }

	/**
	 * @serialData RemoteEventListener as a MarshalledObject
	 */
	private void writeObject(ObjectOutputStream stream)
	    throws IOException
	{
	    stream.defaultWriteObject();
	    stream.writeObject(new MarshalledObject(target));
	}

	/**
	 * Unmarshals the event listener.
	 */
	private void readObject(ObjectInputStream stream)
	    throws IOException, ClassNotFoundException
	{
	    stream.defaultReadObject();
	    MarshalledObject mo = (MarshalledObject)stream.readObject();
	    try {
		target = (RemoteEventListener)mo.get();
	    } catch (Throwable e) {
		if (e instanceof Error &&
		    !(e instanceof LinkageError ||
		      e instanceof OutOfMemoryError ||
		      e instanceof StackOverflowError))
		    throw (Error)e;
	    }
	}

	/**
	 * Enables the stored RemoteEventListener object with the
	 * associated, stored registration.
	 */
	public void apply(MailboxImpl mb) {
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Applying a " + getClass().getName());
	        dbgRecovery.println("\t" + regID);
	        dbgRecovery.println("\t" + target);
	    }
	    try {
	        if (target != null) {
	            mb.enableRegistration(regID, target);
		} else {
	            if (dbgRecovery != null) {
	                dbgRecovery.println(getClass().getName()
	                    + " cancelled due to null target");
		    }
		}
		    
	    } catch (ThrowThis tt) {
	        // TODO (FCS)- throw internal exception since this shouldn't happen
	    }
	}
    }

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * a registration is disabled.
     */
    private static class RegistrationDisabledLogObj implements LogRecord {

	// TODO (FCS)- generate serialVersionUID

	/**
	 * The registration ID.
	 *
	 * @serial
	 */
	private UUID regID;

	/** Simple constructor */
        public RegistrationDisabledLogObj(UUID id) {
            regID = id;
        }

	/**
	 * Disables the associated, stored registration.
	 */
	public void apply(MailboxImpl mb) {
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Applying a " + getClass().getName());
	        dbgRecovery.println("\tregID: " + regID);
	    }
	    try {
	        mb.disableRegistration(regID);
	    } catch (ThrowThis tt) {
	        // TODO (FCS)- throw internal exception since this shouldn't happen
	    }
	}
    }

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * the set of groups to join is changed.
     */
    private static class LookupGroupsChangedLogObj implements LogRecord {

	// TODO (FCS)- generate serialVersionUID

	/**
	 * The new groups to join.
	 *
	 * @serial
	 */
	private String[] groups;

	/** Simple constructor */
	public LookupGroupsChangedLogObj(String[] groups) {
	    this.groups = groups;
	}

	/**
	 * Sets the value of <code>lookupGroups</code> to the provided
         * set.
	 */
	public void apply(MailboxImpl mb) {
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Applying a " + getClass().getName());
	        mb.dumpGroups(groups, dbgRecovery);
    	    }
	    mb.lookupGroups = groups;
	}
    }

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * the set of locators of lookup services to join is changed.
     */
    private static class LookupLocatorsChangedLogObj implements LogRecord {

	// TODO (FCS)- generate serialVersionUID

	/**
	 * The new locators to join.
	 *
	 * @serial
	 */
	private LookupLocator[] locators;

	/** Simple constructor */
	public LookupLocatorsChangedLogObj(LookupLocator[] locators) {
	    this.locators = locators;
	}

	/**
	 * Sets <code>lookupLocators</code> to value of the provided
	 * set.
	 */
	public void apply(MailboxImpl mb) {
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Applying a " + getClass().getName());
	        mb.dumpLocators(locators, dbgRecovery);
	    }
	    mb.lookupLocators = locators;
	}
    }

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * the set of lookup attributes is augmented.
     */
    private static class AttrsAddedLogObj implements LogRecord {
	// TODO (FCS)- generate serialVersionUID

        /** 
         * The attributes to be added to each lookup service with which
         * this service is registered, written out in marshalled form.
         *
         * @serial
         */
        private MarshalledObject[] marshalledAttrs;

        /** Constructs this class and stores the attributes that were added */
        public AttrsAddedLogObj(Entry[] attrs) {
            this.marshalledAttrs = marshalAttributes(attrs);
        }

        /** 
         * Modifies this service's state by adding (after unmarshalling) the
         * elements of marshalledAttrs to the service's existing set of
         * attributes.
         *
         *  @see MailboxImpl.LocalLogHandler#applyUpdate
         */
        public void apply(MailboxImpl mb) {
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Applying a " + getClass().getName());
	    }
            Entry[] attrs = unmarshalAttributes(marshalledAttrs);
            mb.lookupAttrs = LookupAttributes.add(mb.lookupAttrs,attrs);
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Added the following attributes:");
	        dumpAttrs(attrs, dbgRecovery);
	    }
        }
    }

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * the set of lookup attributes is modified.
     */
    private static class AttrsModifiedLogObj implements LogRecord {
	// TODO (FCS)- generate serialVersionUID

	/** 
	 * The set of attribute templates used to select the attributes (from
         * the service's existing set of attributes) that were modified,
         * written out in marshalled form.
         * 
	 * @serial
	 */
        private MarshalledObject[] marshalledAttrTmpls;

        /** 
         * The attributes with which this service's existing attributes 
         * were modified, written out in marshalled form.
         *
         *  @serial
         */
        private MarshalledObject[] marshalledModAttrs;

        /** 
         * Constructs this class and stores the modified attributes 
         * along with the corresponding set of templates.
         */
        public AttrsModifiedLogObj(Entry[] attrTmpls, Entry[] modAttrs) {
            this.marshalledAttrTmpls = marshalAttributes(attrTmpls);
            this.marshalledModAttrs = marshalAttributes(modAttrs);
        }

        /** 
         * Modifies this service's state by modifying (after unmarshalling)
         * the service's existing attributes according to the contents of
         * marshalledAttrTmpls and marshalledModAttrs.
         *
         *  @see MailboxImpl.LocalLogHandler#applyUpdate
         */
        public void apply(MailboxImpl mb) {
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Applying a " + getClass().getName());
	        dbgRecovery.println("Attributes (before):");
	        dumpAttrs(mb.lookupAttrs, dbgRecovery);
	    }
            Entry[] attrTmpls = unmarshalAttributes
                                            (marshalledAttrTmpls);
            Entry[] modAttrs  = unmarshalAttributes
                                            (marshalledModAttrs);
            mb.lookupAttrs = LookupAttributes.modify(mb.lookupAttrs,
                                                     attrTmpls, modAttrs);
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Attributes (after):");
	        dumpAttrs(mb.lookupAttrs, dbgRecovery);
	    }
        }
    }

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * a registration is renewed.
     */
    private static class RegistrationRenewedLogObj implements LogRecord {

	// TODO (FCS)- generate serialVersionUID

	/**
	 * The unique id of the associated ServiceRegistration. 
	 *
	 * @serial
	 */
	private UUID regID;

	/**
	 * The absolute time granted for this renewal.
	 *
	 * @serial
	 */
	private long expirationTime;

	/** Simple constructor */
	public RegistrationRenewedLogObj(UUID id, long time) {
	    regID = id;
	    expirationTime = time;
	}

	/**
	 * Sets the expiration time of the associated registration.
	 */
	public void apply(MailboxImpl mb) {
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Applying a " + getClass().getName());
	        dbgRecovery.println("Reg: " + regID + " will expire at " 
	            + new Date(expirationTime));
	    }

            // Note: The registration might be expired at this point
            // but we can't be sure until the recovery process is
            // complete (eg there might be another renewal log for this
            // registration later on in the recovery process).
            ServiceRegistration reg = 
                (ServiceRegistration)mb.regByID.get(regID);
            if (reg != null) { 
                reg.setExpiration(expirationTime);
	    } else {
	        if (dbgRecovery != null) {
	            dbgRecovery.println("*** Registration not renewed "
	                + "- not found");
		}
		// TODO (FCS)- throw internal exception
	    }
	}
    }

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * a registration is cancelled.
     */
    private static class RegistrationCancelledLogObj implements LogRecord {

	// TODO (FCS)- generate serialVersionUID

	/**
	 * The unique id of the associated ServiceRegistration. 
	 *
	 * @serial
	 */
	private UUID regID;

	/** Simple constructor */
	public RegistrationCancelledLogObj(UUID id) {
	    regID = id;
	}

	/**
	 * Sets the expiration time of the associated registration.
	 */
	public void apply(MailboxImpl mb) {
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Applying a " + getClass().getName());
	        dbgRecovery.println("Cancelling Reg: " + regID);
	    }

            ServiceRegistration reg = 
                (ServiceRegistration)mb.regByID.get(regID);
            mb.removeRegistration(regID, reg);
	}
    }

    /**
     * LogObj class whose instances are recorded to the log file whenever
     * an UnknownEventException is received during event notification. 
     * The event's EventID is associated with the given registration.
     */
    private static class UnknownEventExceptionLogObj implements LogRecord {

	// TODO (FCS)- generate serialVersionUID

	/**
	 * The unique id of the associated ServiceRegistration. 
	 *
	 * @serial
	 */
	private UUID regID;

	/**
	 * The internal representation of the remote event which
	 * caused the UnknownEventException.
	 *
	 * @serial
	 */
	private EventID evtID;

        /** Simple constructor */
        public UnknownEventExceptionLogObj(UUID id, EventID eid) {
            regID = id;
            evtID = eid;
        }

	/**
	 * Adds the given <code>EventID</code> to the set of unknownEvents
	 * for the associated registration.
	 */
	public void apply(MailboxImpl mb) {
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Applying a " + getClass().getName());
	        dbgRecovery.println("Adding: " + evtID + " to " + regID);
	    }
	    ServiceRegistration reg = 
	        (ServiceRegistration)mb.regByID.get(regID);
	    reg.getUnknownEvents().put(evtID, evtID);
	}
	
    }

    /**
     * Marshals each element of the <code>Entry[]</code> array parameter.
     * This method is <code>static</code> so that it may called from
     * the <code>static</code> <code>LogRecord</code> classes when a set
     * of attributes is being logged to persistent storage.
     *
     * @param attrs       <code>Entry[]</code> array consisting of the
     *                    attributes to marshal
     * @return array of <code>MarshalledObject[]</code>, where each element
     *         corresponds to an attribute in marshalled form 
     */
    private static MarshalledObject[] marshalAttributes(Entry[] attrs) {
        if(attrs == null) 
            return new MarshalledObject[0];

        ArrayList marshalledAttrs = new ArrayList();
        for(int i=0; i<attrs.length; i++) {
            /* 
             * Marshalling errors should not prevent the service from
             * working, so just ignore them.
             */
            try {
                marshalledAttrs.add(new MarshalledObject(attrs[i]));
            } catch(Throwable e) {
	        if (dbgRecovery != null) {
	            dbgRecovery.println("Error while marshalling attribute["
	                + i + "] (" + attrs[i] + "): ");
		    e.printStackTrace(dbgRecovery);
	        }
            }
        }
        return ((MarshalledObject[])(marshalledAttrs.toArray
                             (new MarshalledObject[marshalledAttrs.size()])));
    }

    /**
     * Unmarshals each element of the <code>MarshalledObject</code> array
     * parameter. This method is <code>static</code> so that it may called
     * from the <code>static</code> <code>LogRecord</code> classes when a
     * set of attributes is being recovered from persistent storage.
     *
     * @param marshalledAttrs <code>MarshalledObject</code> array consisting
     *                        of the attributes to unmarshal
     * @return array of <code>Entry[]</code>, where each element corresponds
     *         to an attribute that was successfully unmarshalled
     */
    private static Entry[] unmarshalAttributes( 
	MarshalledObject[] marshalledAttrs)
    {
        if(marshalledAttrs == null) 
            return new Entry[0];

        ArrayList attrs = new ArrayList();
        for(int i=0; i<marshalledAttrs.length; i++) {
            /* 
             * Unmarshalling errors should not prevent the service from
             * working, so just ignore them.
             */
            try {
                attrs.add( (Entry)( marshalledAttrs[i].get() ) );
            } catch(Throwable e) {
	        if (dbgRecovery != null) {
	            dbgRecovery.println("Error while unmarshalling attribute["
	                + i + "]");
		    e.printStackTrace(dbgRecovery);
	        }
            }
        }
        return ((Entry[])(attrs.toArray(new Entry[attrs.size()])));
    }

    /**
     * Add a state-change record to persistent storage.
     * <p>
     * Whenever a significant change occurs to the Mercury's state, this
     * method is invoked to record that change in a file called a log file.
     * Each record written to the log file is an object reflecting both
     * the data used and the ACTIONS taken to make one change to the
     * Mercury's state at a particular point in time. If the number of
     * records contained in the log file exceeds the pre-defined threshold,
     * a snapshot of the current state of the Mercury will be recorded.
     * <p>
     * Whenever one of the following state changes occurs, this method
     * will be invoked with the appropriate implementation of the
     * LogRecord interface as the input argument.
     * <ul>
     * <li> registration creation
     * <li> registration renewal
     * <li> registration expired/cancellation
     * <li> registration enabled
     * <li> registration disabled
     * <li> An UnknownEventException is received during event delivery
     * <li> service's set of lookup attributes is modified
     * <li> service's set of lookup groups is modified
     * <li> service's set of lookup locators is modified
     * <li> service's service ID is assigned
     * </ul>
     * 
     * @see MailboxImpl.LocalLogHandler
     */
    private void addLogRecord(LogRecord rec) {
	try {
	    if (dbgRecovery != null) {
	        dbgRecovery.println("Adding log record: " + rec);
	    }
	    log.update(rec, true);
	    if (++logFileSize >= logToSnapshotThresh) {
	        if (dbgRecovery != null) {
	            dbgRecovery.println("Notifying snapshot thread");
	        }
                concurrentObj.waiterNotify(snapshotNotifier);
	    }
	} catch (Exception e) {
	    /* if log updating fails, then one of the following must be done:
	     *   -- output the problem to a separate file and exit()
	     *   -- output the problem to a separate file and continue
	     *   -- set an "I have a problem" attribute & send notification
	     * XXX this issue will be addressed at a later time
	     */
	    e.printStackTrace();
	}
    }


    /**
     * Write the current state of the Mercury to persistent storage.
     * <p>
     * A 'snapshot' of the Mercury's current state is represented by
     * the data contained in certain fields of the Mercury. That data
     * represents many changes -- over time -- to the Mercury's state.
     * This method will record that data to a file referred to as the
     * snapshot file.
     * <p>
     * The data written by this method to the snapshot file -- as well as
     * the format of the file -- is shown below:
     * <ul>
     * <li> our class name
     * <li> log format version number
     * <li> our service ID
     * <li> current proxy ID
     * <li> current configuration parameters
     * <li> current service attributes
     * <li> current registration state
     * </ul>
     * Each data item is written to the snapshot file in serialized form.
     *
     * Note that event state is kept separately and maintained by the 
     * event logging mechanism.
     * 
     * @see Mailbox.LocalLogHandler
     */

    private void takeSnapshot(OutputStream  out) throws IOException {
	if (dbgRecovery != null) {
	    dbgRecovery.println("Taking snapshot ...");
	}
	ObjectOutputStream stream = new ObjectOutputStream(out);

	stream.writeUTF(MailboxImpl.class.getName());
	stream.writeInt(LOG_VERSION);
	stream.writeObject(serviceID);
	stream.writeObject(serviceProxyID);
	stream.writeInt(logToSnapshotThresh);
	stream.writeObject(lookupGroups);
	stream.writeObject(lookupLocators);
        stream.writeObject(marshalAttributes(lookupAttrs));
	stream.writeObject(regByID);
	stream.flush();
	// Note: logDirname doesn't have to be persisted here
	// since the activation system is doing it for us.
    }

    /**
     * Retrieve the contents of the snapshot file and reconstitute the 'base'
     * state of the Mercury from the retrieved data.
     * <p>
     * The data retrieved by this method from the snapshot file is shown
     * below:
     * <ul>
     * <li> our class name
     * <li> log format version number
     * <li> our service ID
     * <li> our proxy ID
     * <li> our configuration parameters
     * <li> our service attributes
     * <li> our registration state
     * </ul>
     * During recovery, the state of the Mailbox at the time of a crash
     * or outage is re-constructed by first reconstituting the 'base state'
     * from the snapshot file; and then modifying that base state according
     * to the records retrieved from the log file. This method is invoked to
     * perform the first step in that reconstruction. 
     * 
     * @see Mailbox.LocalLogHandler
     */
    private void recoverSnapshot(InputStream in)
	throws IOException, ClassNotFoundException
    {
	if (dbgRecovery != null) {
	    dbgRecovery.println("Recovering snapshot ...");
	}
	int i = 0;
	ObjectInputStream stream = new ObjectInputStream(in);
	if (!MailboxImpl.class.getName().equals(stream.readUTF()))
	    throw new IOException("log from wrong implementation");
	if (stream.readInt() != LOG_VERSION)
	    throw new IOException("wrong log format version");
	serviceID = (ServiceID)stream.readObject();
	serviceProxyID = (UUID)stream.readObject();
	logToSnapshotThresh = stream.readInt();
	lookupGroups = (String[])stream.readObject();
	lookupLocators = (LookupLocator[])stream.readObject();
        MarshalledObject[] marshalledAttrs
                                    = (MarshalledObject[])stream.readObject();
	lookupAttrs = unmarshalAttributes(marshalledAttrs);
	regByID = (HashMap)stream.readObject();
	if (dbgRecovery != null) {
	    dbgRecovery.println("serviceID:           " + serviceID);
	    dbgRecovery.println("logToSnapshotThresh: " + logToSnapshotThresh);
	    dbgRecovery.println("lookupGroups:        ");
	    dumpGroups(lookupGroups, dbgRecovery);
	    dbgRecovery.println("lookupLocators:      ");
	    dumpLocators(lookupLocators, dbgRecovery);
	    dbgRecovery.println("lookupAttributes:    ");
	    dumpAttrs(lookupAttrs, dbgRecovery);
	    Collection regs = regByID.values();
	    Iterator iter = regs.iterator();
	    while (iter.hasNext()) {
	        ServiceRegistration reg = (ServiceRegistration)iter.next();
	        reg.dumpInfo(dbgRecovery);
	    }
	}
    }


    /**
     * Handler class for the persistent storage facility.
     * <p>
     * At any point during processing in the mailbox, there will exist
     * both a 'snapshot' of the mailbox's state and a set of records
     * detailing each significant change that has occurred to the state
     * since the snapshot was taken. The snapshot information and the
     * incremental change information will be stored in separate files
     * called, respectively, the snapshot file and the log file. Together,
     * these files are used to recover the state of the mailbox after a
     * crash or a network outage (or if the mailbox or its ActivationGroup
     * is un-registered and then re-registered through the activation daemon).
     * <p>
     * This class contains the methods that are used to record and recover
     * the snapshot of the mailbox's state; as well as the method used to
     * apply the state changes that were recorded in the log file.
     * <p>
     * When the ReliableLog class is instantiated, a new instance of this
     * class is passed to its constructor so that the methods of this
     * class may be invoked by the methods defined in the ReliableLog.
     * Because this class extends the LogHandler class associated with
     * the ReliableLog class, this class must provide implementations of
     * the abstract methods declared in the LogHandler. Also, some of the
     * methods defined in this class override the methods of the LogHandler
     * in order to customize the handling of snapshot creation and
     * retrieval.
     * <p>
     * Each significant change to the mailbox's state is written to the
     * log file as an individual record (when addLogRecord() is invoked).
     * After the number of records logged exceeds a pre-defined threshold,
     * a snapshot of the state is recorded by invoking -- through the
     * ReliableLog and its LogHandler -- the snapshot() method defined in
     * this class. After the snapshot is taken, the log file is cleared
     * and the incremental log process starts over.
     * <p>
     * The contents of the snapshot file reflect the DATA contained in
     * the fields making up the current state of the mailbox. That data
     * represents many changes -- over time -- to the mailbox's state.
     * On the other hand, each record written to the log file is an object
     * that reflects both the data used and the ACTIONS taken to make one
     * change to the mailbox's state at a particular point in time.
     * <p>
     * During recovery, the state of the mailbox at the time of a crash
     * or outage is re-constructed by first retrieving the 'base' state from
     * the snapshot file; and then modifying that base state according to
     * the records retrieved from the log file. The reconstruction of the
     * base state is achieved by invoking the recover() method defined in
     * this class. The modifications recorded in the log file are then
     * applied to the base state by invoking the applyUpdate() method
     * defined in this class. Both recover() and applyUpdate() are invoked
     * through the ReliableLog and its associated Loghandler.
     * <p>
     * NOTE: The following lines must be added to the mailbox's policy file
     * <pre>
     *     permission java.io.FilePermission "dirname",   "read,write,delete";
     *     permission java.io.FilePermission "dirname/-", "read,write,delete";
     * </pre>
     *     where 'dirname' is the name of the directory path (relative or
     *     absolute) where the snapshot and log file will be maintained.
     */
    private class LocalLogHandler extends LogHandler {
	
	/** Simple constructor */
	public LocalLogHandler() { }

	/* Overrides snapshot() defined in ReliableLog's LogHandler class. */
	public void snapshot(OutputStream out) throws IOException {
	    takeSnapshot(out);
	}

	/* Overrides recover() defined in ReliableLog's LogHandler class. */
	public void recover(InputStream in)
	    throws IOException, ClassNotFoundException
	{
	    recoverSnapshot(in);
	}

	/**
	 * Required method implementing the abstract applyUpdate()
	 * defined in ReliableLog's associated LogHandler class.
	 * <p>
	 * During state recovery, the recover() method defined in the
	 * ReliableLog class is invoked. That method invokes the method
	 * recoverUpdates() which invokes the method readUpdates(). Both
	 * of those methods are defined in ReliableLog. The method
	 * readUpdates() retrieves a record from the log file and then
	 * invokes this method.
	 * <p>
	 * This method invokes the version of the method apply() that
	 * corresponds to the particular type of 'log record' object
	 * that is input as the first argument. The log record object and its
	 * corresponding apply() method are defined in one of the so-called
	 * LogObj classes. Any instance of one the LogObj classes is an
	 * implementation of the LogRecord interface. The particular
	 * implementation that is input to this method is dependent on the
	 * type of record that was originally logged. The apply() method
	 * will then modify the state of the mailbox in a way dictated
	 * by the type of record that was retrieved.
	 */
	public void applyUpdate(Object logRecObj) {
	    ((LogRecord)logRecObj).apply(MailboxImpl.this);
	}
    }

    /**
     * Snapshot-taking thread. 
     * <p>
     * A snapshot is taken when -- after writing a new record to the 
     * log file -- it is determined that the size of the log file has 
     * exceeded a certain threshold. The code which adds the new record 
     * to the log file and which, in turn, decides that a snapshot
     * must be taken is "wrapped" in a writer mutex. That is, synchronization
     * of processing is achieved in the mailbox through a "reader/writer"
     * mutex construct. This construct allows only one writer at any one
     * time; but allows an unlimited number of simultaneous readers as
     * long as no writer has locked the mutex. During steady-state, it is
     * anticipated that there will be far more readers in use 
     * than writers. Since the process of
     * taking a snapshot can be time-consuming, if the whole snapshot-taking
     * process occupies that single writer mutex, then a significant number
     * of read requests will be un-necessarily blocked; possibly resulting
     * in an unacceptable degradation in response time. 
     * <p>
     * It is for the above reason that the process of taking a snapshot is
     * performed in a separate thread. The thread waits on the monitor
     * belonging to the snapshotNotifier instance until it is notified
     * (or "signaled") that a snapshot must be taken. The notification
     * is sent by another thread, created by the mailbox, which
     * determines when the conditions are right for a snapshot. The
     * notification takes the form of an interrupt indicating that the
     * snapshot monitor is available. Although the interrupt is sent 
     * while the writer mutex is locked, the act of sending the notification
     * is less time-consuming than the act of taking the snapshot itself.
     * When the thread receives a notification, it awakens and requests a
     * lock on the reader mutex (this is all done in the readerWait() method).
     * Because a reader -- not a writer -- mutex is locked, read-only
     * processes still have access to the system state performed; 
     * and the reader mutex prevents changes to the data while
     * the snapshot is in progress.  
     * <p>
     * Note that the current snapshot is guaranteed to complete before the
     * next snapshot request is received. This is because even though
     * the act of taking a snapshot can be viewed as a writer process, 
     * the fact that the next snapshot notification will be wrapped in a
     * writer mutex, combined with the fact that a writer mutex can not
     * be locked while a reader mutex is locked, allows the snapshot to
     * be treated as a reader process.
     */
    private class SnapshotThread extends Thread {

	/** Create a daemon thread */
	public SnapshotThread() {
	    super("SnapshotThread");
	    setDaemon(true);
	    start();
	}

	public void run() {
	    try {
		concurrentObj.readLock();
	    } catch (ConcurrentLockException e) {
		return;
	    }
	    try {
		while (!isInterrupted()) {
		    try {
			concurrentObj.readerWait(snapshotNotifier,
						 Long.MAX_VALUE);
                        try {
                            // Take snapshot
                            log.snapshot();
                            // Reset log count
		            logFileSize = 0;
	                } catch (InterruptedIOException e) {
			    return;
	                } catch (Exception e) {
			    if (e instanceof LogException &&
				((LogException)e).detail instanceof
				InterruptedIOException)
				return;
	                    /* XXX
	                     * if taking the snapshot fails for any reason,
                             * then one of the following must be done:
	                     *   -- output the problem to a file and exit
	                     *   -- output the problem to a file and continue
	                     *   -- set an "I have a problem" attribute and
                             *      then send a notification
	                     * this issue will be addressed at a later time
	                     */
			    e.printStackTrace();
		        }
		    } catch (ConcurrentLockException e) {
			return;
		    }
		}
	    } finally {
		concurrentObj.readUnlock();
	    }
	}
    }


    ///////////////////////
    // Activation Methods
    ///////////////////////

    /**
     * If for some reason we can't activate, we dump a message to the log 
     * and unregister with activation.
     * @param m message to dump to <code>System.err</code>
     * @param e If non-<code>null</code> this exception is dumped to
     * <code>System.err</code> as well 
     */
    private void alertAndUnregister(String m, Throwable e) {
	System.err.println("Fatal error during activation:" + m);
	if (e != null)
	    e.printStackTrace();
	
	Exception failed = null;

	System.err.println("Attempting to unregister with activation system");
	try {
	    ActivationGroupID grp = ActivationGroup.currentGroupID();
	    ActivationSystem sys = ActivationGroup.getSystem();
	    sys.unregisterGroup(grp);
	} catch (ActivationException ee) {
	    failed = ee;
	} catch (RemoteException ee) {
	    failed = ee;
	}

	if (failed != null) {
	    System.err.println("Failed to unregister with activation system");
	    failed.printStackTrace();
	} else {
	    System.err.println("Succeeded in unregistering");
	}
    }


    /**
     * Activation constructor
     *
     * @param activationID activation ID passed in by the activation daemon.
     * @param data state data needed to re-activate a Mercury server
     */
    public MailboxImpl(ActivationID activationID, MarshalledObject data) 
	throws IOException, ClassNotFoundException
    {
	boolean runtimeThrown = false;
	if (dbgInit != null) {
	    dbgInit.println("MailboxImpl is being activated");
	}
	this.activationID = activationID;
	Remote ourStub = null;

	try {
	    try {
	        if (dbgInit != null) {
	            dbgInit.println("MailboxImpl is being exported");
	        }
		ourStub = Activatable.exportObject(this, activationID, 0);
	    } catch (RemoteException e) {
		 alertAndUnregister("Failure exporting object", e);
		 throw e;
	    }

	    // Get Log Dir Name
	    try {
                logDirname = (String)data.get();
	        if (dbgInit != null) {
	            dbgInit.println("Obtained persistence path: " + logDirname);
	        }
	    } catch (IOException e) {
		alertAndUnregister("Failure recovering activation data", e);
		throw e;
	    } catch (ClassNotFoundException e) {
		alertAndUnregister("Failure recovering activation data", e);
		throw e;	  
	    } catch (ClassCastException e) {
		alertAndUnregister("Failure recovering activation data", e);
		runtimeThrown = true;
		throw e;
	    }

	    // Initialize state
	    init();

	} catch (RuntimeException e) {
	    if (runtimeThrown == false) { 
		alertAndUnregister("Unexpected exception during activation",
				   e);
	    }

	    throw e;
	} catch (Error e) {
	    alertAndUnregister("Unexpected exception during activation", e);
	    throw e;
	}	  
    }


    /** Utility method for displaying lookup group attributes */
    private static void dumpGroups(String[] grps, PrintWriter pw) {
	if (grps == null)
	    pw.println("<ALL_GROUPS>");
	else if (grps.length == 0)
	    pw.println("<NO_GROUPS>");
	else {
	    for (int i = 0; i < grps.length; i++) {
		 pw.println("\t" + grps[i]);
	    }
	}
    }

    /** Utility method for displaying lookup locator attributes */
    private static void dumpLocators(LookupLocator[] locs, PrintWriter pw) {
	if (locs == null)
	    pw.println("<null>");
	else if (locs.length == 0)
	    pw.println("<EMPTY>");
	else {
	    for (int i = 0; i < locs.length; i++) {
		 pw.println("\t" + locs[i]);
	    }
	}
    }

    /** Utility method for displaying lookup service attributes */
    private static void dumpAttrs(Entry[] attrs, PrintWriter pw) {
	if (attrs.length == 0)
	    pw.println("<EMPTY>");
	else {
	    for (int i = 0; i < attrs.length; i++) {
		 pw.println("\t" + attrs[i]);
	    }
	}
    }

    /** Utility method for displaying relavent contents of a RemoteEvent */
    private static void dumpEvent(RemoteEvent re, PrintWriter pw) {
	pw.println("RemoteEvent:");
	pw.println("\tID:   " + re.getID());
	pw.println("\tSeq#: " + re.getSequenceNumber());
    }
}
