/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.mercury;

import java.io.Serializable;
import java.rmi.RemoteException;

import com.sun.jini.proxy.ThrowThis;
import com.sun.jini.proxy.UUID;

import net.jini.event.MailboxRegistration;
import net.jini.core.event.RemoteEventListener;
import net.jini.core.lease.Lease;

/**
 * The <tt>Registration</tt> class is the client-side proxy
 * returned to event mailbox clients as the result of the 
 * registration process. It implements the <tt>MailboxRegistration</tt>
 * interface and delegates functionality to the mailbox service
 * where necessary.
 *
 * @since Jini1.1
 */ 
class Registration implements MailboxRegistration, Serializable {

    /** Unique identifier for this registration */
    private final UUID registrationID;

    /** Reference to service implementation */
    private final MailboxBackEnd mailbox;

    /** Reference to service provided RemoteEventListener implementation */
    private final ListenerProxy listener;

    /** The service's registration lease */ 
    private final Lease lease;

    /** Convenience constructor */
    public Registration(UUID id, MailboxBackEnd srv, Lease l) {
        if (id == null || srv == null || l == null)
            throw new IllegalArgumentException("Cannot accept null arguments");
        registrationID = id;
        mailbox = srv;
        listener = new ListenerProxy(srv, id);
        lease = l;
    }

    // inherit javadoc from supertype
    public Lease getLease() {
	return lease;
    }
    
    // inherit javadoc from supertype
    public RemoteEventListener getListener() {
	return (RemoteEventListener)listener;
    }

    // inherit javadoc from supertype
    public void enableDelivery(RemoteEventListener target) 
	throws RemoteException
    {
	if (target == null) { // same as disableDelivery
	    try {
                mailbox.disableDelivery(registrationID);
	    } catch (ThrowThis tt) {
	        tt.throwRemoteException();
	    }
	} else if ((target instanceof ListenerProxy) &&
	           (listener.equals((ListenerProxy)target))) {
	    // Can't re-submit a mailbox-provided listener
	    throw new IllegalArgumentException("Cannot provide an " +
		"a target that was provided by the EventMailbox service");
	} else { // OK to make the call, now
	    try {
	        mailbox.enableDelivery(registrationID, target);
	    } catch (ThrowThis tt) { 
	        tt.throwRemoteException();
	    }
	}
    }

    // inherit javadoc from supertype
    public void disableDelivery() throws RemoteException {
	try { 
            mailbox.disableDelivery(registrationID);
	} catch (ThrowThis tt) { 
	    tt.throwRemoteException();
	}
    }

    /** Proxies with the same registrationID have the same hash code. */
    public int hashCode() {
        return registrationID.hashCode();
    }

    /** Proxies with the same registrationID are considered equal. */
    public boolean equals(Object o) {
        if (this == o)
            return true;

        if (o == null)
            return false;

        if (o.getClass() != getClass())
            return false;

        Registration reg = (Registration)o;
        return registrationID.equals(reg.registrationID);
    }
}
