/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.norm;

import java.util.Map;
import java.util.Set;
import java.util.Collection;
import java.util.Iterator;
import java.util.AbstractMap;
import java.util.AbstractSet;
import java.util.AbstractCollection;

import java.rmi.RemoteException;

import net.jini.core.lease.Lease;
import net.jini.core.lease.LeaseMap;
import net.jini.core.lease.LeaseException;
import net.jini.core.lease.UnknownLeaseException;
import net.jini.core.lease.LeaseDeniedException;
import net.jini.core.lease.LeaseMapException;

/**
 * Class that wraps LeaseMap created by client Leases.  Provides hooks
 * for synchronization and data associated with each client lease
 * while allowing us to use <code>LeaseRenewalManager</code>.  Objects 
 * of this class are returned by <code>createLeaseMap</code> calls made on
 * <code>ClientLeaseWrapper</code> objects that are not deformed.
 * <p>
 * This class only allows as keys ClientLeaseWrappers that are non-deformed,
 * and who's leases come from compatable codebases. Internally the 
 * Mapping from ClientLeaseWrappers to longs is held in two Maps.
 * The first is a LeaseMap that created by the first client lease assocated
 * with the first ClientLeaseWrapper added to the set.
 * The 2nd is a Map from client leases to the ClientLeaseWrapper that 
 * wraps them.
 * <p>
 * @see ClientLeaseWrapper
 */
class ClientLeaseMapWrapper extends AbstractMap implements LeaseMap {
    /** Lease Map created by the client lease */
    private LeaseMap clientLeaseMap;

    /** Map from client leases to wrappers */
    private Map wrapperMap = new java.util.HashMap();

    /**
     * Retain inital wrapper so we canBatch() can check for codebase
     * compatibility before delegating to clientLeaseMap
     */
    private ClientLeaseWrapper example;

    /**
     * Create a ClientLeaseMapWrapper object that will hold
     * the specified client Lease
     * @param wrapper a Wrapper for the lease that wants to be renewed.
     * @param duration The duration to assocate with wrapper
     * @throws IllegalArgumentException if wrapper is deformed
     */
    ClientLeaseMapWrapper(ClientLeaseWrapper wrapper, long duration) 
    {
	final Lease clientLease = wrapper.getClientLease();
	if (clientLease == null) 
	    throw new IllegalArgumentException("Wrapper can not be deformed");

	clientLeaseMap = clientLease.createLeaseMap(duration);
	wrapperMap.put(clientLease, wrapper);
	example = wrapper;
    }

    // inherit javadoc
    public void cancelAll() {
	throw new UnsupportedOperationException(
	     "ClientLeaseMapWrapper.cancelAll:" + 
	     "LRS should not being canceling client leases ");
    }

    /**
     * For each lease in the map, call failedRenewal
     */
    private void applyException(Throwable t) {
	for (Iterator i=wrapperMap.values().iterator(); i.hasNext(); ) {
	    final ClientLeaseWrapper clw = (ClientLeaseWrapper)i.next();
	    clw.failedRenewal(t);
	}	
    }

    // inherit javadoc
    // This method assumes that none of the leases in the map are
    // also being renewed by some other thread.
    public void renewAll() throws LeaseMapException, RemoteException {
	LeaseMapException lme = null;

	try {
	    clientLeaseMap.renewAll();
	} catch (LeaseMapException e) {
	    lme = e;
	} catch (RemoteException e) {
	    applyException(e);
	    throw e;
	} catch (Error e) {
	    applyException(e);
	    throw e;
	} catch (RuntimeException e) {
	    applyException(e);
	    throw e;
	}

	// For each Lease still in the map we need to update the wrapper
	for (Iterator i=clientLeaseMap.keySet().iterator(); i.hasNext(); ) {
	    final Lease cl = (Lease)i.next();
	    final ClientLeaseWrapper clw = 
		(ClientLeaseWrapper)wrapperMap.get(cl);

	    clw.successfulRenewal();
	}

	// For each Lease in the LeaseMapException we have to update
	// the wrapper.  Also create a LeaseMapException that maps
	// the client leases to the exceptions
	if (lme == null) 
	    return;

	final Map exceptionMap = lme.exceptionMap;
	if (lme != null) {
	    final Map newExceptionMap = 
		new java.util.HashMap(exceptionMap.size());

	    for (Iterator i=exceptionMap.entrySet().iterator(); i.hasNext(); ) {
		final Map.Entry e = (Map.Entry)i.next();
		final Lease cl = (Lease)e.getKey();
		final Throwable t = (Throwable)e.getValue(); 
		final ClientLeaseWrapper clw = 
		    (ClientLeaseWrapper)wrapperMap.get(cl);

		clw.failedRenewal(t);
		newExceptionMap.put(clw, t);
	    }

	    throw new LeaseMapException(lme.getMessage(), newExceptionMap);
	}	    
    }

    /**
     * If key is an undeformed ClientLeaseWrapper who's lease comes
     * from a compatible codebase return the client lease it wraps,
     * otherwise return null.
     */
    private Lease extractClientLease(Object key) {
	// We only batch with ClientLeaseWrappers
	if (!(key instanceof ClientLeaseWrapper))
	    return null;

	final ClientLeaseWrapper clw = (ClientLeaseWrapper)key;

	// Are the leases already in the map from the same codebase
	// as the candidate? 
	if (!(example.isCodebaseMatch(clw)))
	    return null;

	// Is it deformed?
	return clw.getClientLease();	
    }

    // inherit javadoc
    public boolean canContainKey(Object key) {
	final Lease cl = extractClientLease(key);

	if (cl == null)
	    return false;

	// Let the map we are wrapping decide
	return clientLeaseMap.canContainKey(cl);
    }

    /**
     * Check that the key is valid for this map, if it is return the 
     * client lease, if not throw IllegalArgumentException
     */
    private Lease checkKey(Object key) {
	final Lease cl = extractClientLease(key);

	if (cl != null)
	    return cl;

	throw new IllegalArgumentException(
					"key is not valid for this LeaseMap");
    }

    /** Check that the value is a Long */
    private static void checkValue(Object value) {
	if (!(value instanceof Long))
	    throw new IllegalArgumentException("value is not a Long");
    }

    // inherit javadoc
    public boolean containsKey(Object key) {
	final Lease cl = checkKey(key);	
	return clientLeaseMap.containsKey(cl);
    }

    // inherit javadoc
    public boolean containsValue(Object value) {
	checkValue(value);
	return clientLeaseMap.containsValue(value);
    }

    // inherit javadoc
    public Object get(Object key) {
	final Lease cl = checkKey(key);	
	return clientLeaseMap.get(cl);
    }

    // inherit javadoc
    public Object put(Object key, Object value) {
	final Lease cl = checkKey(key);
	checkValue(value);
	
	// At this point we know key is a ClientLeaseWrapper

	/*
	 * Since there is a 1:1 mapping between wrappers and client
	 * leases, and once we get going we never have two wrapper
	 * objects that != but are .equals() if key is already in this
	 * map wrapperMap.put() will replace the existing copy of 
	 * key with a == copy, this gives us the key non-replacement
	 * semantics Map.put() should have.
	 */
	wrapperMap.put(cl, key);
	return clientLeaseMap.put(cl, value);
    }

    // inherit javadoc
    public Object remove(Object key) {
	final Lease cl = checkKey(key);

	wrapperMap.remove(cl);
	return clientLeaseMap.remove(cl);
    }

    // inherit javadoc
    public void putAll(Map m) {
	Iterator iter = m.entrySet().iterator();
	while (iter.hasNext()) {
	    Map.Entry e = (Map.Entry)iter.next();
	    put(e.getKey(), e.getValue());
	}
    }

    // inherit javadoc
    public void clear() {
	clientLeaseMap.clear();
	wrapperMap.clear();
    }

    // inherit javadoc
    public boolean equals(Object o) {
	return clientLeaseMap.equals(o); // XXX should sameDestination matter?
    }

    // inherit javadoc
    public int hashCode() {
	return clientLeaseMap.hashCode();
    }

    // inherit javadoc
    public Set entrySet() {
	return new EntrySet();
    }

    /*
     * Classes that we use to implement entrySet()
     */

    /**
     * An implemention of Set backed by the ClientMapWrapper's mappings
     * (which are from wrapperMap's values to clientLeaseMap)
     */
    private final class EntrySet extends AbstractSet {
	// inherit javadoc
	public Iterator iterator() {
	    return new EntryIterator();
	}

	/**
	 * If the pased object is a Map.Entry that is in the
	 * ClientMapWrapper return the client lease assocated with it,
	 * otherwise return null
	 */
	private Lease getClientLease(Object o) {
	    if (!(o instanceof Map.Entry))
		return null;

	    final Map.Entry e = (Map.Entry)o;
	    final Object eValue = e.getValue();

	    if (!(e.getKey() instanceof ClientLeaseWrapper) ||
		!(eValue instanceof Long) ||
		(eValue == null))
		return null;

	    final ClientLeaseWrapper clw = (ClientLeaseWrapper)e.getKey();
	    // Note if clw is deformed this call will return null, this is
	    // the right thing since a deforemed lease can't be in this map
	    return clw.getClientLease();
	}

	// inherit javadoc
	public boolean contains(Object o) {
	    final Lease cl = getClientLease(o);

	    if (cl == null)
		return false;

	    final Object eValue = ((Map.Entry)o).getValue();
	    final Object value = clientLeaseMap.get(cl);
	    if (value == null)
		return false;

	    return value.equals(eValue);
	}

	// inherit javadoc
	public boolean remove(Object o) {
	    final Lease cl = getClientLease(o);

	    if (cl == null)
		return false;

	    final Object eValue = ((Map.Entry)o).getValue();
	    final Object value = clientLeaseMap.get(cl);
	    if (value == null || !value.equals(eValue))
		return false;

	    
	    // Use cl to remove the data from the clientLeaseMap
	    clientLeaseMap.remove(cl);
	    wrapperMap.remove(cl);

	    return true;
	}

	// inherit javadoc
	public int size() {
	    return clientLeaseMap.size();
	}

	// inherit javadoc
	public void clear() {
	    wrapperMap.clear();
	    clientLeaseMap.clear();
	}
    }

    /** Our implemention of Map.Entry */
    private final class Entry implements Map.Entry {
	/** The key */
	private final ClientLeaseWrapper key;

	public Entry(ClientLeaseWrapper key) {
	    this.key = key;
	}

	public Object getKey() {
	    return key;
	}

	public Object getValue() {
	    return clientLeaseMap.get(key.getClientLease());
	}

	public Object setValue(Object value) {
	    checkValue(value);
	    return clientLeaseMap.put(key.getClientLease(), value);
	}

	public boolean equals(Object o) {
	    if (o instanceof Entry) {
		final Entry that = (Entry)o;
		return that.key.equals(key);
	    }

	    return false;
	}

	public int hashCode() {
	    return key.hashCode();
	}
    }

    /**
     * An implemention of Iterator backed by the ClientMapWrapper's mappings
     * (which are from wrapperMap's values to clientLeaseMap)
     */
    private final class EntryIterator implements Iterator {
	/** itertor over the wrapperMap values */
	private final Iterator iter;

	/** Lease associated with the last value returned by next() */
	private Lease last;

	public EntryIterator() {
	    iter = wrapperMap.values().iterator();
	}

	public boolean hasNext() {
	    return iter.hasNext();
	}

	public Object next() {
	    final Map.Entry e = (Map.Entry)iter.next();	    	    
	    last = (Lease)e.getKey();		    
	    return new Entry((ClientLeaseWrapper)e.getValue());
	}

	public void remove() {
	    // Use last to remove the data from the clientLeaseMap
	    clientLeaseMap.remove(last);
	    // use iter.remove() to remove from wrapperMap so we don't
	    // get a concurrent access exception
	    iter.remove();
	}
    }
}

