/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.norm;

import java.util.Set;
import java.util.Map;
import java.util.Iterator;
import net.jini.core.lease.Lease;

/**
 * <code>DeformedLeaseList</code> provides a list for keeping track of
 * of client lease that could not be fully recovered during log
 * recovery.  Unless otherwise noted all the methods in the class are
 * not thread safe.
 */
class DeformedLeaseList {
    /** 
     * A list of all the leases we are tracking.
     */
    private Set leases = null;

    /**
     * Add a wrapped client lease to the list of deformed client leases.
     * This method is not thread safe.
     * @param clw A deformed client lease
     */
    void add(ClientLeaseWrapper clw) {
	if (leases == null)
	    leases =  new java.util.HashSet();
	leases.add(clw);
    }

    /**
     * Remove a lease from the list.
     * This method is not thread safe.
     * @param clw A deformed client lease
     */
    void remove(ClientLeaseWrapper clw) {
	if (leases != null)
	    leases.remove(clw);

	if (leases.isEmpty()) 
	    leases = null;
    }

    /**
     * Query the list to see if the pasted client lease is in this
     * list of deformed leases.  If it is returned the existing client
     * lease wrapper for this client lease.  During the query any
     * client leases wrapper (including the target if found) which are
     * discovered to no longer be deformed will be inserted into the
     * passed table using their associated client lease as a key and removed
     * from the list of deformed leases. This method is not thread safe.
     * @param box An EqualBox that wraps a client lease which may be
     *           referenced by a deformed client lease wrapper
     * @param table A table mapping client leases to client lease wrappers.
     *           Any client lease wrappers encountered during the 
     *           query that are no longer deformed will be placed in this table.
     *           It is assumed that no other thread is trying to access
     *           this table
     * @return If it could be found the client lease wrapper that goes with
     *         <code>cl</code>.  <code>null</code> otherwise.
     */
    ClientLeaseWrapper query(EqualBox box, Map table) {
	// Recheck now that we have the lock
	if (leases == null) 
	    return null;
	
	try {
	    ClientLeaseWrapper clw;  // our result;
	    
	    for (Iterator i = leases.iterator(); i.hasNext(); ) {
		clw = (ClientLeaseWrapper)i.next();
		final EqualBox candidate = clw.getEqualBox();
		
		if (candidate != null) {
		    // We've got a live one!, remove it from the
		    // deformed list and place it in the table
		    table.put(box, clw);
		    i.remove();

		    // Is it the one we are looking for?
		    if (candidate.equals(box)) {
			return clw;
		    }
		} 
	    }
	} finally {
	    // Can we GC the leases set?
	    if (leases.isEmpty()) 
		leases = null;
	}

	return null;
    }

    /**
     * Number of leases in the list.
     */
    public int size() {
	final Set temp = leases;
	if (temp == null)
	    return 0;

	return temp.size();
    }
}

