/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.norm;

import java.util.Map;
import java.rmi.RemoteException;
import java.lang.ref.WeakReference;

import net.jini.core.lease.UnknownLeaseException;

import com.sun.jini.thread.WakeupManager;
import com.sun.jini.collection.WeakTable;
import com.sun.jini.lease.landlord.LeaseManager;
import com.sun.jini.lease.landlord.LeasedResource;


/**
 * Lease Mgr implementation that aggressively expires leases as they
 * expiration times occur.  Also schedules and manages expiration
 * warning events. 
 * <p>
 * Note, unlike Mahalo's <code>LeaseExpirationManager</code> (which this
 * was seeded from) we make no attempt to make it generic because
 * of the need to schedule expiration warning events.
 *
 * @see com.sun.jini.lease.landlord.LeaseManager
 * @author John W.F. McClain 
 */
class LeaseExpirationMgr implements LeaseManager, WeakTable.KeyGCHandler {
    /** 
     * Map of sets to task tickets.
     * <p>
     * A Note on Synchronization
     * <p>
     * Whenever we operate on the <code>ticketMap</code> we hold
     * the lock on the key being used.  This is necessary because
     * expiration and warning sender tasks need to remove tickets from
     * the map but at the same time a renewal may be updating the map
     * to associate the set with a new ticket.  If we don't synchronize
     * there is a small window where a task could remove the ticket
     * for its replacement.
     */
    private WeakTable		ticketMap = new WeakTable(this); 

    /** Ref to the main server object has all the top level methods */
    private NormServerImpl	server;

    /** Queue of tasks order my time */
    private WakeupManager	runQueue = new WakeupManager();

    /**
     * Create a <code>LeaseExpirationMgr</code> to aggressively expire
     * the leases of the passed <code>NormServerImpl</code>
     */
    LeaseExpirationMgr(NormServerImpl server) {
	this.server = server;
    }

    
    // purposefully inherit doc comment from supertype
    public void register(LeasedResource resource, long duration) {
	// Need to synchronize because schedule manipulates 
	// ticketMap.
	synchronized (resource) {
	    schedule(resource);
	}
    }

    // purposefully inherit doc comment from supertype
    public void renewed(LeasedResource resource, long duration, 
			long oldExpiration) {
	reschedule(resource);
    }

    /**
     * The actually work of the <code>renewed</code> method, broken into
     * a separate method so it can be called from <code>NormServerImpl</code>
     * <p>
     * This method assumes the lock on <code>set</code> is owned
     * by the current thread.
     * @param resource The set for which tasks have to be re-scheduled
     */
    void reschedule(LeasedResource resource) {
	// Remove the old event.  This method is only called
	// (indirectly) from NormServerImpl.renew() so we know that we
	// own the lock on resource
	runQueue.cancel(
	    (WakeupManager.Ticket)ticketMap.remove(resource)
	);
	// Schedule the new event
	schedule(resource);	
    }

    /** 
     * Schedule a leased resource to be reaped in the future. Called
     * when a resource gets a lease, a lease is renewed, and during 
     * log recovery
     * <p>
     * This method assumes the lock on <code>resource</code> is owned
     * by the current thread.
     */
    void schedule(LeasedResource resource) {
	WakeupManager.Ticket ticket;
	final LeaseSet set = (LeaseSet)resource;
	MgrTask task;

	if (set.haveWarningRegistration()) {
	    task = new SendWarning(set);	    
	    ticket = runQueue.schedule(set.getWarningTime(), task);	    
	} else {
	    task = new Expiration(set);
	    // The cancels in general have to block on snapshot so
	    // place them in a separate thread from the WakeupManager
	    ticket = runQueue.schedule(set.getExpiration(), task,
				       new WakeupManager.ThreadDesc());
	}
	
	// No window here because the tasks only use the ticket
	// after the acquire the lock on their set, but we still own
	// the lock on the set.
	task.setTicket(ticket);
	ticketMap.getOrAdd(set, ticket);
    }

    // purposefully inherit doc comment from supertype
    // Called when LeaseResource we are tracking is garbage collected
    public void keyGC(Object value) {
	final WakeupManager.Ticket ticket = (WakeupManager.Ticket)value;
	runQueue.cancel(ticket);
    }


    /**
     * Utility base class for our tasks, manly provides the 
     * the proper locking for manipulating the ticketMap.
     */
    private abstract class MgrTask implements Runnable {
	/** Resource this task is to operate on */
	protected final WeakReference resourceRef;
	
	/** Ticket for this task */
	private WakeupManager.Ticket ticket;

	/** 
	 * Simple constructor
	 * @param set The set this task is to operate on
	 */
	protected MgrTask(LeaseSet set) {
	    resourceRef = new WeakReference(set);
	}

	/**
	 * Set the ticket associated with this task
	 */
	private void setTicket(WakeupManager.Ticket ticket) {
	    this.ticket = ticket;
	}

	/**
	 * Removes this task's ticket from the ticket map iff
	 * this tasks ticket is in the map
	 * Returns this <code>LeaseSet</code> task is to operate on
	 * or <code>null</code> if this task should stop
	 */
	protected LeaseSet removeOurTicket() {
	    final LeaseSet set = (LeaseSet)resourceRef.get();

	    if (set != null) {
		synchronized (set) {
		    final WakeupManager.Ticket currentTicket = 
			(WakeupManager.Ticket)ticketMap.get(set);
		    if (ticket.equals(currentTicket)) {
			ticketMap.remove(set);
		    } else {
			// Someone removed us after we were committed
			// to run we should stop
			return null;
		    }
		}
	    }

	    return set;
	}

	// purposefully inherit doc comment from supertype
	public abstract void run();
    }
    /**
     * Objects that do the actually expiration the resource in question, stuck
     * in <code>WakeupManager</code>
     */
    private class Expiration extends MgrTask {
	/**
	 * Create a <code>Expiration</code> task for the passed resource
	 * @param set The set this task is to operate on
	 */
	private Expiration(LeaseSet set) {
	    super(set);
	}

	// purposefully inherit doc comment from supertype
	public void run() {
	    final LeaseSet set = removeOurTicket();
	    if (set == null)
		// Already gone
		return;
	    
	    server.expireIfTime(set);
	    // Note we don't care if is actually time or not, if it is not
	    // the task will be re-scheduled by the renewal.
	}
    }

    /**
     * Objects that do the schedule the warning events, also
     * schedules an expiration task.
     */
    private class SendWarning extends MgrTask {
	/**
	 * Create a <code>SendWarning</code> task for the passed resource
	 * @param set The set this task is to operate on
	 */
	private SendWarning(LeaseSet set) {
	    super(set);
	}

	// purposefully inherit doc comment from supertype
	public void run() {
	    final LeaseSet s = (LeaseSet)resourceRef.get();
	    if (s == null) 
		// set is gone, no work to do
		return;

	    // By holding this lock we prevent other threads from
	    // scheduling new tasks for this set...if we have been replaced
	    // we will return before scheduling a new task, if we have not
	    // been we will schedule the new task and it can be cleanly 
	    // removed by any renew that is happening at the same time
	    synchronized (s) {
		final LeaseSet set = removeOurTicket();
		if (set == null)
		    // set is gone, or our task was replaced, no work to do
		    return;

		// Send event
		server.sendWarningEvent(set);

		// Schedule expiration task
		final MgrTask task = new Expiration(set);
		final WakeupManager.Ticket newTicket =
		    runQueue.schedule(set.getExpiration(), task,
				       new WakeupManager.ThreadDesc());
		task.setTicket(newTicket);
		ticketMap.getOrAdd(set, newTicket);
	    }
	}
    }
}


