/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.norm;

import java.lang.reflect.Field;
import java.util.Set;
import java.util.List;
import java.util.Iterator;

import java.io.Serializable;
import java.io.IOException;
import java.rmi.MarshalledObject;

import net.jini.core.event.RemoteEvent;
import net.jini.core.event.EventRegistration;
import net.jini.core.event.RemoteEventListener;

import net.jini.lease.LeaseRenewalSet;
import net.jini.lease.ExpirationWarningEvent;

import com.sun.jini.lease.landlord.LeasedResource;
import com.sun.jini.proxy.UUID;

import com.sun.jini.norm.event.EventType;
import com.sun.jini.norm.event.EventTypeGenerator;
import com.sun.jini.norm.event.EventFactory;
import com.sun.jini.norm.event.SendMonitor;

/**
 * Norm's internal representation of LeaseRenewalSets.  Unless otherwise
 * noted the methods of this class assume synchronization being handled
 * by the caller.
 */
class LeaseSet implements Serializable, LeasedResource {
    /**
     * Expiration time of the set
     * @serial
     */
    private long expiration;
  
    /**
     * ID that uniquely identifies this set
     * @serial
     */
    private UUID ID;

    /**
     * A collection of the client leases in this set (in wrapped form).
     * @serial
     */ 
    private Set leases = new java.util.HashSet();

    /**
     * Time before <code>expiration</code> to send expiration warning events
     */
    private long minWarning = NormServer.NO_LISTENER;

    /**
     * The event type for expiration warning events
     * @serial
     */
    private EventType warningEventType;

    /**
     * The event type for failure events
     * @serial
     */
    private EventType failureEventType;

    /**
     * <code>PersistentStore</code> that changes should be loged to.
     */
    private transient PersistentStore store;

    /**
     * The <code>NormServerImpl</code> are attached to
     */
    private transient NormServerImpl normServerImpl;

    // Constructors and state restoration
    /**
     * Simple constructor.  Note expiration will be set when we allocate
     * a lease for this set.
     * @param ID for this set
     * @param generator object set can use to create new event type objects
     * @param store <code>PersistentStore<code> that changes should be loged
     *              to.
     * @param normServerImpl The <code>normServerImpl</code> that 
     *              created this set.
     */
    LeaseSet(UUID ID, EventTypeGenerator generator, PersistentStore store,
	     NormServerImpl normServerImpl)
    {
	this.store = store;
	this.normServerImpl = normServerImpl;
	this.ID = ID;

	final SendMonitor sendMonitor = normServerImpl.newSendMonitor(this);	

	try {
	    warningEventType = 
		generator.newEventType(null, null, sendMonitor,
		    LeaseRenewalSet.EXPIRATION_WARNING_EVENT_ID);
	    failureEventType =
		generator.newEventType(null, null, sendMonitor,
 	            LeaseRenewalSet.RENEWAL_FAILURE_EVENT_ID);
	} catch (IOException e) {
	    // Because we are passing null for the listener we will
	    // never get an exception
	}
    }

    /** 
     * Restore the transient state of the set that can't be restored 
     * automatically after a log recovery.  Also drop any client
     * leases who's memebership expiration is before now.
     * @param generator  Event type generator associated with this set's events
     * @param store      <code>PersistentStore<code> that changes should be 
     *                   loged to.
     * @param normServerImpl The <code>normServerImpl</code> that 
     *              created this set.
     * @param now A conservative estimate of the current time
     * @return The an iterator over the set of client leaese
     */
    Iterator restoreTransientState(EventTypeGenerator  generator, 
				   PersistentStore     store,
				   NormServerImpl      normServerImpl,
				   long                now)

    {
	this.normServerImpl = normServerImpl;
	this.store = store;

	final SendMonitor sendMonitor = normServerImpl.newSendMonitor(this);	
	warningEventType.restoreTransientState(generator, sendMonitor);
	failureEventType.restoreTransientState(generator, sendMonitor);

	// Run through leases, dropping leases that have passed
	// due membership expirations
	final Iterator i = leases.iterator();
	while (i.hasNext()) {
	    final ClientLeaseWrapper clw = (ClientLeaseWrapper)i.next();
	    if (clw.getMembershipExpiration() < now)
		i.remove();
	}

	return leases.iterator();
    }

    /**
     * Utility to method to replace a client lease 
     */
    private void replace(ClientLeaseWrapper clw) {
	leases.remove(clw);
	leases.add(clw);
    }

    /**
     * Add or update the specified wrapped client lease to the set.
     * @param clw Lease to be added
     * @param membershipExpiration When should the lease be removed from the
     *				   set
     */
    void update(ClientLeaseWrapper clw) {
	replace(clw);
	final Object u = new UpdateClientLease(this, clw);
	store.update(u);
    }

    /**
     * Remove the specified wrapped client lease from the set
     * @param cwl Lease to removed
     * @return False if the lease has already been removed, or 
     * if logicaly the lease is no longer in the set (e.g. its
     * membership expiration has been reached).
     */
    boolean remove(ClientLeaseWrapper clw) {
	if (!leases.remove(clw)) 
	    // if we are here the lease must have already been 
	    // removed by some other thread, don't bother logging
	    return false;

	// Event if the membership expiration was expired we still
	// want to update the log
	final Object u = new RemoveClientLease(this, clw);
	store.update(u);

	return (clw.getMembershipExpiration() > System.currentTimeMillis());
    }

    /**
     * Remove the specified wrapped client lease from the set if
     * its membership expiration has been reached.
     * @param clw The wrapper that we might want to remove
     * @return true if the lease has already been removed, or 
     * if it is membership expired, otherwise return false
     */
    boolean removeIfMembershipExpired(ClientLeaseWrapper clw) {
	if (!leases.contains(clw)) {
	    return true;
	}

	if (System.currentTimeMillis() <= clw.getMembershipExpiration()) {
	    // still here and it should be
	    return false;
	}

	// If we here then the lease's membership expiration has passed
	// and it should be removed
	leases.remove(clw);
	final Object u = new RemoveClientLease(this, clw);
	store.update(u);

	return true;
    }

    /**
     * Destroy a lease set
     * @return a <code>Set</code> with all of the sets WrappedClientLeases
     */
    Set destroy() {
	setExpiration(-1);
	final Object u = new CancelLeaseSet(getUUID());
	store.update(u);
	return leases;
    }

    /** 
     * Cleanup all the stale data assocated with this set. Called during 
     * snapshot before the set is writen out so none of the changes need
     * to be loged.
     * @param now A conservative estimate of the current time
     * @return If one or more client leases are dropped from the set an
     * iterator enumerating those leases.  Otherwise null
     */
    Iterator cleanup(long now) {
	final Iterator i = leases.iterator();
	List deadLeases = null;

	while (i.hasNext()) {
	    final ClientLeaseWrapper clw = (ClientLeaseWrapper)i.next();

	    if (now > clw.getMembershipExpiration()) {
		// Drop lease
		i.remove();

		// Add to deadList
		if (deadLeases == null) 
		    deadLeases = new java.util.LinkedList();

		deadLeases.add(clw);
	    }
	}

	if (deadLeases == null)
	    return null;

	return deadLeases.iterator();
    }

    /**
     * Return an array of leases in marshalled form.  For each lease we have
     * in unmarshalled form serilize using the duration format, for each
     * lease we can't unmarshall just use the MarshalledObject we have on
     * hand.  If the set is empty return null
     */
    MarshalledObject[] getLeases() {
	final long now = System.currentTimeMillis();
	final Iterator i = leases.iterator();
	final List l = new java.util.ArrayList(leases.size());

	while (i.hasNext()) {
	    final ClientLeaseWrapper clw = (ClientLeaseWrapper)i.next();

	    // Check to make sure the leases membership expiration has not 
	    // passed
	    if (now > clw.getMembershipExpiration())
		continue;

	    l.add(clw.getMarshalledClientLease());
	}

	if (l.isEmpty())
	    return null;
	else
	    return 
		(MarshalledObject[])l.toArray(new MarshalledObject[l.size()]);
    }

    /** 
     * Set/update/clear the expiration warning listener
     * @param listener   The new listener
     * @param minWarning How long be for the lease on the set expires
     *                   should the event be sent. 
     * @param handback   The new handback
     * @return If <code>listener</code> is non-<code>null</code> return
     * an <code>EventRegistration</code> otherwise return <code>null</code>
     * @throws IOException if listener can not be serialized 
     */
    EventRegistration setExpirationWarningListener(
	RemoteEventListener listener, 
	long                minWarning,
	MarshalledObject    handback)
        throws IOException     					       
    {
	this.minWarning = minWarning;
	warningEventType.setListener(listener, handback);

	final Object u = new WarningEventRegistration(this);
	store.update(u);

	if (listener == null) 
	    return null;
	
	final SetProxy proxy = newSetProxy();

	return new EventRegistration(warningEventType.getEventID(),
				     proxy, 
				     proxy.getRenewalSetLease(), 
				     warningEventType.getLastSequenceNumber());
    }

    /**
     * Set/update/clear the renewal failure listener
     * @param listener The new listener
     * @param handback The new handback
     * @return If <code>listener</code> is non-<code>null</code> return
     * an <code>EventRegistration</code> otherwise return <code>null</code>
     * @throws IOException if listener can not be serialized 
     */
    EventRegistration setRenewalFailureListener(RemoteEventListener listener, 
						MarshalledObject    handback)
        throws IOException     					       
    {
	failureEventType.setListener(listener, handback);

	final Object u = new FailureEventRegistration(this);
	store.update(u);

	if (listener == null) 
	    return null;

	final SetProxy proxy = newSetProxy();

	return new EventRegistration(failureEventType.getEventID(),
				     proxy, 
				     proxy.getRenewalSetLease(), 
				     failureEventType.getLastSequenceNumber());
    }

    /**
     * Handle failures to renew a lease by removing the lease from the set
     * and if needed schedule sending an event.
     * @param clw The wrapped client lease for the lease that could not
     *            be removed.
     */
    void renewalFailure(ClientLeaseWrapper clw) {
	if (!leases.remove(clw))
	    // if we are here the lease must have already been 
	    // removed by some other thread, don't bother logging or 
	    // sending an event
	    return;

	Object u;
	try {
	    final long seqNum = 
		failureEventType.sendEvent(
		    clw.newFailureFactory(newSetProxy()));
	    u = new RenewalFailure(this, clw, seqNum);
	} catch (IOException e) {
	    // Can't form event just drop it
	    // $$$ is this the right thing to do?  The only  jars
	    u = new RemoveClientLease(this, clw);
	}

	store.update(u);
    }

    /**
     * Send an expiration warning event for this set
     */
    void sendWarningEvent() {
	warningEventType.sendEvent(new WarningFactory(this));
    }

    /**
     * Nested class that implements <code>EventFactory</code> that
     * generates <code>ExpirationWarningEvent</code>s.
     */
    private class WarningFactory implements EventFactory {
	/** The source for the event */
	final private SetProxy proxy;

	/**
	 * Create a new WarningFactory 
	 * @param set The set generating this event
	 */
	WarningFactory(LeaseSet set) {
	    proxy = set.newSetProxy();
	}

	// Inherit java doc from super type
	public RemoteEvent createEvent(long             eventID, 
				       long             seqNum, 
				       MarshalledObject handback) 
	{
	    return new ExpirationWarningEvent(proxy, seqNum, handback);
	}	
    }

    /**
     * Create a new SetProxy for this set that has a lease with the
     * current expiration     
     */
    private SetProxy newSetProxy() {
	return normServerImpl.newSetProxy(this); 
    }


    /**
     * Return true if there is a non-<code>null</code> listener registered
     * for the expiration warning event.
     * <p>
     * Note, this method assumes the current thread owns the set's lock
     */
    boolean haveWarningRegistration() {
	return warningEventType.haveListener();
    }

    /**
     * Return the absolute time when a expiration warning should be sent.
     */
    long getWarningTime() {
	return expiration - minWarning;
    }

    /**
     * Log the renewal of a client lease.
     * @param clw the wrapper for the client lease that was renewed
     */
    void logRenewal(ClientLeaseWrapper clw) {
	if (!leases.contains(clw))
	    // Some other thread must have removed this lease from
	    // the set after renewal, don't bother logging change
	    return;

	final Object u = new UpdateClientLease(this, clw);
	store.update(u);
    }

    // Methods need to meet contract of LeasedResource	
    // Inherit java doc from super type
    public void setExpiration(long newExpiration) {
	expiration = newExpiration;
    }

    // Inherit java doc from super type
    public long getExpiration() {
	return expiration;
    }

    // Inherit java doc from super type
    public Object getCookie() {
	return ID;
    }

    /**
     * Return the <code>UUID</code> for this set.
     */
    UUID getUUID() {
	return ID;
    }

    /**
     * If the passed registrationNumber number matches the 
     * current registrationNumber for the passed event 
     * clear the current registration and persist the change
     */
    void definiteException(EventType type, RemoteEvent ev,
			   long registrationNumber) 
    {
	final boolean changed = 
	    type.clearListenerIfSequenceMatch(registrationNumber);
	if (changed) {
	    // Need to log the change
	    Object u;
	    if (ev instanceof ExpirationWarningEvent)
		u = new WarningEventRegistration(LeaseSet.this);
	    else 
		u = new FailureEventRegistration(LeaseSet.this);
	    store.update(u);
	}
    }
	
    // All the inner classes we use to log changes to a lease set

    /**
     * Class used to log changes to the set expiration time.  
     */
    static class ChangeSetExpiration extends LeaseSetOperation {
	/**
	 * Updated expiration time
	 * @serial
	 */
	private long expiration;

	/**
	 * Simple constructor
	 * @param set that changed
	 */
	ChangeSetExpiration(LeaseSet set) {
	    super(set.getUUID());
	    expiration = set.expiration;
	}
	
	// Inherit java doc from super type
	void apply(LeaseSet set) {
	    set.setExpiration(expiration);
	}
    }

    /**
     * Class used to log adding or updating a client lease to the set
     */
    private static class UpdateClientLease extends LeaseSetOperation {
	/**
	 * Wrapped version of client lease
	 * @serial
	 */
	private ClientLeaseWrapper clw;

	/**
	 * Simple constructor
	 * @param set that changed
	 * @param clw Wrapped client lease
	 */
	private UpdateClientLease(LeaseSet set, ClientLeaseWrapper clw) {
	    super(set.getUUID());
	    this.clw = clw;
	}

	// Inherit java doc from super type
	void apply(LeaseSet set) {
	    set.replace(clw);
	}
    }

    /**
     * Class used to log the removal of a client lease from the set
     */
    private static class RemoveClientLease extends LeaseSetOperation {
	/**
	 * Client lease to be removed
	 * @serial
	 */
	private ClientLeaseWrapper clw;

	/**
	 * Simple constructor
	 * @param set that changed
	 * @param clw Wrapped client lease
	 */
	private RemoveClientLease(LeaseSet set, ClientLeaseWrapper clw) {
	    super(set.getUUID());
	    this.clw = clw;
	}

	
	// Inherit java doc from super type
	void apply(LeaseSet set) {
	    set.leases.remove(clw);
	}
    }

    /**
     * Class used to log a renewal failure
     */
    private static class RenewalFailure extends RemoveClientLease {
	/**
	 * Event ID of the corresponding renewal failure event (if any)
	 * @serial
	 */
	private long evID;

	/**
	 * Simple constructor
	 * @param set that changed
	 * @param clw Wrapped client lease
	 * @param evID event ID of the renewal event that was sent
	 *             to mark this renewal failure
	 */
	private RenewalFailure(LeaseSet set, ClientLeaseWrapper clw,
			       long evID)
	{
	    super(set, clw);
	    this.evID = evID;
	}
	
	// Inherit java doc from super type
	void apply(LeaseSet set) {
	    super.apply(set);
	    set.failureEventType.setLastSequenceNumber(evID);
	}
    }

    /**
     * Class used to log changes to warning event registrations
     */
    private static class WarningEventRegistration extends LeaseSetOperation {
	/**
	 * Warning time assocated with warning event registration
	 * @serial
	 */
	private long warningTime;

	/**
	 * EventType object that resulted from registration change
	 * @serial
	 */
	private EventType registration;

	/**
	 * Simple constructor
	 * @param set that changed
	 */
	private WarningEventRegistration(LeaseSet set) {
	    super(set.getUUID());
	    warningTime = set.minWarning;
	    registration = set.warningEventType;
	}

	
	// Inherit java doc from super type
	void apply(LeaseSet set) throws StoreException {
	    set.minWarning = warningTime;
	    set.warningEventType = registration;
	}
    }

    /**
     * Class used to log changes to failure event registrations
     */
    private static class FailureEventRegistration extends LeaseSetOperation {
	/**
	 * EventType object that resulted from registration change
	 * @serial
	 */
	private EventType registration;

	/**
	 * Simple constructor
	 * @param set that changed
	 */
	private FailureEventRegistration(LeaseSet set) {
	    super(set.getUUID());
	    registration = set.failureEventType;
	}

	
	// Inherit java doc from super type
	void apply(LeaseSet set) throws StoreException {
	    set.failureEventType = registration;
	}
    }
}
