/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.norm;

import java.util.Set;
import java.util.Map;
import java.util.Iterator;

import net.jini.core.lease.Lease;

/**
 * Maps client leases to the wrapper objects that we use internally.
 * The actually mapping is from <code>EqualBox</code>s to wrappers.
 * This lets us control how the <code>HashMap</code> we use performs
 * comparisons.  We expose this detail to the client so we don't have
 * to constantly create new <code>EqualBox</code>s.
 * <p>
 * In internally we keep two tables of client leases, one a map of
 * client lease (wrapped in <code>EqualBoxe</code>s) to client lease
 * wrappers, the other is a list of client lease wrappers who's leases
 * could not be unpacked (aka "deformed wrappers") when they were
 * added (and thus could not be put in the map).  This bifurcation is
 * not exposed to the client.  
 * <p>
 * Unless othewise noted the methods of this class are not thread safe
 * @see EqualBox
 */
class LeaseTable {
    /**
     * The list we store deformed wrappers in
     */
    final private DeformedLeaseList deformedLeases = new DeformedLeaseList();

    /**
     * The map we use to map EqualBoxes to wrappers
     */
    final private Map leaseTable = new java.util.HashMap();

    /**
     * Find the client lease wrapper assocated with the passed lease.
     * @param clientLease The lease we need the wrapper for
     * @return Returns the wrapper assocated with the passed lease, or
     *         <code>null</code> if we don't know about this lease
     */
    ClientLeaseWrapper get(Lease clientLease) {
	return get(new EqualBox(clientLease));
    }

    /**
     * Find the client lease wrapper assocated with the passed lease.
     * @param clientLease The lease we need the wrapper for wrapped in a 
     *         <code>EqualBox</code>
     * @return Returns the wrapper assocated with the passed lease, or
     *         <code>null</code> if we don't know about this lease
     */
    ClientLeaseWrapper get(EqualBox clientLease) {
	final ClientLeaseWrapper clw = 
	    (ClientLeaseWrapper)leaseTable.get(clientLease);
	if (clw == null) {
	    // Could it be that this was a lease that
	    // could not be recovered? Check the deformed list
	    return deformedLeases.query(clientLease, leaseTable);
	} else {
	    return clw;
	}
    }

    /**
     * Add a mapping from lease wrapper to client lease.
     * Gets client lease from wrapper.
     * @param wrapper Client lease wrapper, and client lease
     *                to add to table     
     */
    void put(ClientLeaseWrapper clw) {
	final EqualBox box = clw.getEqualBox();
	
	if (box == null) {
	    // Deformed, put in deformed list
	    deformedLeases.add(clw);
	} else {
	    // Not deformed, add to list
	    leaseTable.put(box, clw);
	}
    }

    /**
     * Remote a lease from the table
     * @param wrapper Client lease wrapper for the lease we want to remove
     */
    void remove(ClientLeaseWrapper clw) {
	final EqualBox equalBox = clw.getEqualBox();

	if (equalBox != null) {
	    // Not deformed, we need to remove both the client
	    // lease and the wrapper from the leaseTable
	    leaseTable.remove(equalBox);
	} else {
	    // deformed, we need to remove the lease from
	    // the deformed list
	    deformedLeases.remove(clw);
	}
    }

    /**
     * Number of client leases in the table
     */
    int size() {
	return leaseTable.size() + deformedLeases.size();
    }

    /**
     * Return the number of deformed leases in the table 
     */
    int deformedCount() {
	return deformedLeases.size();
    }
}


