/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.norm;

import java.util.Map;
import java.util.Set;
import java.util.List;
import java.util.Iterator;
import java.util.Collection;
import java.util.Collections;
import java.util.Properties;

import java.io.InputStream;
import java.io.OutputStream;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.io.PrintWriter;

import java.rmi.MarshalledObject;
import java.rmi.RemoteException;
import java.rmi.NoSuchObjectException;

import java.rmi.activation.Activatable;
import java.rmi.activation.ActivationID;
import java.rmi.activation.ActivationSystem;
import java.rmi.activation.ActivationGroup;
import java.rmi.activation.ActivationGroupID;
import java.rmi.activation.ActivationException;

import net.jini.admin.Administrable;
import net.jini.admin.JoinAdmin;

import net.jini.lookup.entry.ServiceInfo;

import net.jini.core.entry.Entry;
import net.jini.core.discovery.LookupLocator;

import net.jini.core.lease.Lease;
import net.jini.core.lease.UnknownLeaseException;
import net.jini.core.lease.LeaseDeniedException;

import net.jini.core.event.EventRegistration;
import net.jini.core.event.RemoteEvent;
import net.jini.core.event.RemoteEventListener;
import net.jini.lease.LeaseRenewalService;
import net.jini.lease.LeaseRenewalSet;
import net.jini.lease.LeaseRenewalManager;
import net.jini.lease.LeaseRenewalEvent;

import com.sun.jini.lease.landlord.LeasePolicy;
import com.sun.jini.lease.landlord.LeaseDurationPolicy;
import com.sun.jini.lease.landlord.LandlordLeaseFactory;
import com.sun.jini.lease.landlord.Landlord.RenewResults;
import com.sun.jini.lease.landlord.CodebaseLeaseFactory;
import com.sun.jini.lease.landlord.LocalLandlord;
import com.sun.jini.lease.landlord.LandlordUtil;

import com.sun.jini.proxy.UUID;
import com.sun.jini.proxy.UUIDFactory;
import com.sun.jini.proxy.ThrowThis;

import com.sun.jini.lookup.entry.BasicServiceType;

import com.sun.jini.reliableLog.LogHandler;
import com.sun.jini.reliableLog.LogException;

import com.sun.jini.mahout.KillVMAdmin;
import com.sun.jini.mahout.ParseUtils;
import com.sun.jini.mahout.PersistentServiceDesc;

import com.sun.jini.constants.VersionConstants;
import com.sun.jini.constants.ThrowableConstants;

import com.sun.jini.debug.Debug;

import com.sun.jini.norm.event.EventType;
import com.sun.jini.norm.event.EventTypeGenerator;
import com.sun.jini.norm.event.SendMonitor;
import com.sun.jini.norm.lookup.JoinState;

/**
 * Implementation of NormServer.
 */
public class NormServerImpl implements NormServer, LocalLandlord {
    /** Current version of log format */
    private static final int CURRENT_LOG_VERSION = 1;

    /** Our activation ID */
    private ActivationID activationID;

    /** The name of our log dir */
    private String logDir;

    /** Our JoinManager */
    private JoinState joinState;

    /** Map of UUIDs to LeaseSets */
    private Map setTable = Collections.synchronizedMap(new java.util.HashMap());

    /** 
     * A table of all the ClientLeaseWrappers we have.  Can be used to find
     * the wrapper associated with a given client lease.
     * <p>
     * Queries can be submitted using the client lease directly, or by
     * wrapping the lease in a EqualBox.
     * <p>
     * The EqualBoxes are used because using the client leases
     * directly as keys opens up a security hole where one client
     * lease could capture another when they get compared by the Map.
     * EqualBoxes wrap client leases and provide and equals method
     * that will only allow the underlying at each other
     * if they are from the same codebase.
     * <p>
     * We maintain the invariant that any client lease that is in a
     * set is also in the leaseTable. 
     */
    private LeaseTable leaseTable = new LeaseTable();

    /** Lease renewal manager we that actually renews the leases */
    private LeaseRenewalManager lrm = new LeaseRenewalManager();

    /** Object the expires sets and generates expiration warning events */
    private LeaseExpirationMgr expMgr = new LeaseExpirationMgr(this);

    /** Factory/Accessor object we are using for our leases */
    private CodebaseLeaseFactory leaseFactory;

    /** Policy we use for granting and renewing renewal set leases */
    private LeasePolicy setLeasePolicy;

    /** Our persistant store */
    private PersistentStore store;

    /** Factory we use to create UUIDs */
    private UUIDFactory setIDFactory = new UUIDFactory();

    /** Factory we use to create ClientLeaseWrapper IDs */
    private UIDGenerator idGen = new UIDGenerator() ;

    /** List of leases that have been renewed by not persisted */
    private List renewedList = new java.util.LinkedList();

    /**
     * Thread that pulls wrapped client leases off the renewedList and logs
     * them to disk
     */
    private RenewLogThread renewLogger;

    /** Object use to generate new event types */
    private EventTypeGenerator generator = new EventTypeGenerator();

    /**
     * Object that transfers events from the renewal manager to 
     * us we can remove dead lease and send events
     */
    private FailureNotifer failureNotifer;

    /** Log File must contain this many records before snapshot allowed */
    private int logToSnapshotThresh = 200;

    /** Weight factor applied to snapshotSize when deciding to take snapshot */
    private float snapshotWt = 10;

    /** Our remote stub */
    private NormServer ourStub = null;

    /** Thread the performs snapshots when signaled */
    private SnapshotThread snapshotter;

    // Debuging message support
    /** Simple object supporting selective debug output to files or display */
    private Debug debugObj = new Debug("com.sun.jini.norm.debug");

    /** Debug PrintWriter for Client lease codebase info */
    private PrintWriter debugClientLeaseCodebase = 
	debugObj.getWriter("clientLeaseCodebase");

    /** Debug PrintWriter for snapshot stats */
    private PrintWriter debugSnapshot = 
	debugObj.getWriter("snapshot");

    /** Debug PrintWriter for renewal failure */
    private PrintWriter debugRenewalFailure = 
	debugObj.getWriter("renewalFailure");

    /** Debug PrintWriter for loggin a renewal */
    private PrintWriter debugRenewals = 
	debugObj.getWriter("renewals");

    ////////////////////////////////
    // Methods defined in NormServer

	
    // Inherit java doc from super type
    public void renewFor(UUID id, Lease leaseToRenew,
			 long membershipDuration, long renewDuration) 
	throws ThrowThis
    {
	// Ensure that the lease was not granted by this server
	if (leaseFactory.canAccess(leaseToRenew)) 
	    if (ourStub.equals(leaseFactory.getLandlord(leaseToRenew)))
		throw new IllegalArgumentException("Can't add a leases " + 
		    "granted by a LeaseRenewalService to a set created by " +
		    "that service.");
	
	// If we are told to dump the codbase of the lease we are adding
	final PrintWriter writer = debugClientLeaseCodebase;
	if (writer != null) {
	    final Class lc = leaseToRenew.getClass();
	    final String annotation = 
		java.rmi.server.RMIClassLoader.getClassAnnotation(lc);
	    writer.println("Adding lease of class " + 
                lc.getName() + " from " + annotation);
	}

	// Lookup the set
	final LeaseSet set = getSet(id);

	// Try to add the lease to the set
	final EqualBox equalBox = new EqualBox(leaseToRenew);
	LeaseSet conflict = add(set, leaseToRenew, equalBox, 
				membershipDuration, renewDuration); 
	while (conflict != null) {
	    // If there is a conflict we have to make sure that the
	    // confliciting set is up-to-date and try adding again

	    if (resolveConflict(conflict, equalBox)) {
		// There was a change, try the add again
		conflict = add(set, leaseToRenew, equalBox, 
			       membershipDuration, renewDuration);
	    } else {
		// Conflict was not resolved, throw IAE
		throw new IllegalArgumentException(
	            "NormServerImpl.renewFor:Lease already in " +
		    "another set created by this renewal service");
	    }
	}
    }

       
    /**
     * Try to add the passed lease to the passed set. If we can't
     * because the lease may be in another return the conflicting set.
     * @param set The LeaseSet to add the leaseToRenew too.
     * @param leaseToRenew The lease the client wants managed
     * @param equalBox An EqualBox that wraps leaseToRenew
     * @param membershipDuration The lenght of time the client
     *        wants the lease managed for
     * @param renewDuration The length of time the client want the 
     *        lease renewed for each time it is renewed.
     * @return If there was a conflict with another set the conflicting
     *         set, otherwise null.
     * @throws ThrowThis if the set no longer exists.
     */
    // $$$ This method is a bit long, unfortunately the best candidate
    // block for pulling out into a method (the manipulation of the leaseTable)
    // needs to return two values....
    private LeaseSet add(LeaseSet set, Lease leaseToRenew, EqualBox equalBox,
			 long membershipDuration, long renewDuration) 
	throws ThrowThis
    {
	try {
	    store.acquireMuatorLock();	    
	    synchronized (set) {
		ensureCurrent(set);

		// Make sure the lease is not owned by some other set
		ClientLeaseWrapper clw;
		synchronized (leaseTable) {
		    final long now = System.currentTimeMillis();
		    clw = leaseTable.get(equalBox);

		    if (clw == null) {
			// We don't know about this lease, create a new
			// wrapper and place it in the table

			try {
			    // Until the lease gets removed we always
			    // what it serialized using the absolute
			    // format 
			    leaseToRenew.setSerialFormat(Lease.ABSOLUTE);
			    clw = new ClientLeaseWrapper(leaseToRenew,
			        idGen.newID(), renewedList, set.getUUID(),
				equalBox, membershipDuration, renewDuration,
				now);
			} catch (IOException e) {
			    throw new IllegalArgumentException(
			       "NormServerImpl.renewFor:Handed lease that " +
			       "can't be marshalled");
			}

			leaseTable.put(clw);
		    } else {
			// We know about this lease is in set, or another?
			final LeaseSet other = clw.getLeaseSet(setTable);
			if (other == set) {
			    // This one, update its stats
			    clw.update(membershipDuration, renewDuration, now);
			} else {
			    // Already in another set signal a conflict
			    return other;
			}
		    }
		}
		 
		set.update(clw);
		lrm.renewUntil(clw, clw.getMembershipExpiration(), 
			       clw.getRenewDuration(), failureNotifer);

		// $$$ What if this lease was just dropped by the LRM
		// (say because there was some problem renewing the
		// lease)?  Are we going to lose the lease because
		// we will shortly process a renewalFailureEvent?
		// Is it a problem if we do? Prosuably re-adding it
		// is not going to fix the underlying problem...
		return null;
	    }
	} finally {
	    store.releaseMuatorLock();
	}    
    }


    /**
     * We think the passed lease is in the passed set, make sure really is.
     * It may not be because its membership expiration has passed, or
     * because the set has expired.
     * 
     * @param set The LeaseSet we think lease is in
     * @param lease The lease we want to ensure is in set wrapped in a
     *              EqualBox
     * @return true if lease is no longer in set and false otherwise
     */
    private boolean resolveConflict(LeaseSet set, EqualBox lease) {
	try {
	    store.acquireMuatorLock();	    
	    synchronized (set) {
		if (!setLeasePolicy.ensureCurrent(set)) {
		    // Set has expired, cleanup and (since this resolves the
		    // the conflict return true.
		    removeSet(set);
		    return true;
		}

		// The set is still good, make sure the lease's membership
		// expiration has not expired.
		synchronized (leaseTable) {
		    final ClientLeaseWrapper clw = leaseTable.get(lease);
		    if (clw == null) 
			// Must have just been droped, conflict
			// resolved, return true
			return true;

		    final boolean removed = set.removeIfMembershipExpired(clw);
		    if (removed) {
			// Lease is membership expired and is no
			// longer in set, need to cleanup lease table
			// and LRM as well
			try {
			    lrm.remove(clw);
			} catch (UnknownLeaseException e) {
			    // This can happen if there was some
			    // problem renewing the lease or its LRM
			    // expiration just ran out. Since we are
			    // removing the lease anyway ignore.
			}

			leaseTable.remove(clw);
		    }

		    return removed;
		}
	    }
	} finally {
	    store.releaseMuatorLock();
	}    
    }

    // Inherit java doc from super type
    public Lease remove(UUID id, Lease leaseToRemove) throws ThrowThis {
	final LeaseSet set = getSet(id);

	// The most uptodate ref to the lease we have
	Lease rslt = null;
	try {
	    store.acquireMuatorLock();	    
	    synchronized (set) {
		ensureCurrent(set);
		
		synchronized (leaseTable) {
		    final ClientLeaseWrapper clw 
			= leaseTable.get(leaseToRemove);

		    // Down to set.remove() could be outside of of the
		    // sync block...just seems like a waste to grab
		    // the leaseTable lock, let it go, do a little bit
		    // of work, and grab it again...the removal from
		    // the table does have to be after the lease is
		    // remove from the set/LRM [Otherwise another
		    // thread could add the lease to another set and
		    // add to the LRM but then have it removed by this
		    // thread.]  [Of course it may be more than just a
		    // little bit of work, set.remove() does need to
		    // flush an object to disk...if we do acquire 
		    // leaseTable's lock twice we need to do a 2nd get
		    // and make sure that the clw is in set.

		    if (clw == null)
			// Lease must have been removed already
			return null;

		    try {
			lrm.remove(clw);
		    } catch (UnknownLeaseException e) {
			// This can happen if there was some problem
			// renewing the lease or its LRM expiration just
			// ran out. Since we are removing the lease anyway
			// ignore. 
		    }

		    final boolean present = set.remove(clw);	
		    leaseTable.remove(clw);

		    // Only return a non-null result if the removed lease
		    // had not be removed already
		    if (present)
			// At this point we can assume clw is no
			// longer deformed
			rslt = clw.getClientLease();
		}
	    }
	} finally {
	    store.releaseMuatorLock();
	}    

	if (rslt == null) 
	    return null;

	// Whenever we serialize a lease we have to make sure that
	// its serial form will stay the same during the
	// serialization.  Since we have removed it from the set we
	// don't have to worry about this lease being serialized to disk
	// any more so changing the serial format here should be safe.
	rslt.setSerialFormat(Lease.DURATION);
	return rslt;
    }

    // Inherit java doc from super type
    public MarshalledObject[] getLeases(UUID id) throws ThrowThis {
	final LeaseSet set = getSet(id);

	// We are not modifying the set so we don't need the mutator lock
	// $$$ Do we need a reader lock, or is the lock on the set enough?
	// $$$ Need to make sure we really don't mutate any persistent state
	// or have issues with serializing the leases.
	synchronized (set) {
	    ensureCurrent(set);
	    return set.getLeases();
	}
    }


    // Inherit java doc from super type
    public EventRegistration setExpirationWarningListener(
				 UUID		     id,
				 RemoteEventListener listener, 
				 long		     minWarning, 
				 MarshalledObject    handback)
	throws ThrowThis 
    {
	final LeaseSet set = getSet(id);

	try {
	    store.acquireMuatorLock();	    
	    synchronized (set) {
		ensureCurrent(set);

		try {
		    final boolean haveBefore = set.haveWarningRegistration();
		    final EventRegistration rslt =
			set.setExpirationWarningListener(listener,
							 minWarning, handback);

		    if (haveBefore != set.haveWarningRegistration()) {
			// Ether we had a registration before and we
			// don't now or do now and not before, ether
			// way wack the expiration manager so it can
			// schedule the right task.
			expMgr.reschedule(set);
		    }

		    return rslt;
		} catch (IOException e) {
		    // This means the listener could not be serialized,
		    // re-throw as a IllegalArgumentException 
		    throw new IllegalArgumentException("Passed a listener " +
			      "that could not be serialized");
		}
	    }
	} finally {
	    store.releaseMuatorLock();
	}    
    }

    /**
     * Remote a set if its expiration time has been reached.
     */
    void expireIfTime(LeaseSet set) {
	try {
	    store.acquireMuatorLock();

	    synchronized (set) {
		if (setLeasePolicy.ensureCurrent(set)) {
		    // Someone must have renewed the lease...don't expire
		    return;
		}

		removeSet(set);
	    }
	} finally {
	    store.releaseMuatorLock();
	}
    }

    /**
     * Schedule the sending of an expiration warning event.
     * This could be a method on the set it self but this keeps all
     * of the high level synchronization logic in one file.     
     */
    void sendWarningEvent(LeaseSet set) {
	// We don't need to acquire the store lock because we 
	// won't be mutating any persistant state.
	// We will be reading state of the set so we do need to
	// sync on it to ensure we get a constant view
	synchronized (set) {
	    if (!setLeasePolicy.ensureCurrent(set)) {
		// Must have just been canceled or expired, don't send event
		return;
	    }

	    set.sendWarningEvent();
	}
    }

    // Inherit java doc from super type
    public EventRegistration setRenewalFailureListener(
				 UUID		     id,
				 RemoteEventListener listener, 
				 MarshalledObject    handback)
	throws ThrowThis
    {
	final LeaseSet set = getSet(id);

	try {
	    store.acquireMuatorLock();	    
	    synchronized (set) {
		ensureCurrent(set);

		try {
		    return set.setRenewalFailureListener(listener,
							 handback);
		} catch (IOException e) {
		    // This means the listener could not be serlized,
		    // re-throw as a IllegalArgumentException 
		    throw new IllegalArgumentException("Passed a listener " +
			      "that could not be serialized");
		}
	    }
	} finally {
	    store.releaseMuatorLock();
	}    
    }

    /**
     * Handle failures to renew a lease by removing the lease from its set
     * and if needed schedule sending an event.
     * @param clw The wrapped client lease for the lease that could not
     *            be removed.  <code>clw.isDeformed</code> must be
     *            <code>false</code>
     */
    void renewalFailure(ClientLeaseWrapper clw) {
	final LeaseSet set = clw.getLeaseSet(setTable);
	if (set == null) 
	    // set must have just been removed, no state to update, no
	    // events to send, just return
	    return;

	try {
	    store.acquireMuatorLock();	    

	    synchronized (set) {
		final PrintWriter writer = debugRenewalFailure;

		if (writer != null) 
		    writer.println("Failed renewal on " + clw);

		if (!setLeasePolicy.ensureCurrent(set)) {
		    // expired, no state to update, no
		    // events to send, just return
		    return;
		}

		set.renewalFailure(clw);

		synchronized (leaseTable) {			
		    leaseTable.remove(clw);
		}

		// Note, we don't need to remove the lease from the
		// LRM because it has already been removed.
	    }
	} finally {
	    store.releaseMuatorLock();
	}    	
    }


    /**
     * The implemention of <code>SendMonitor</code> we use to track 
     * event delevery threads.  Each set gets its own object.
     */
    private class SendMonitorImpl implements SendMonitor {
	/** Set this is the monitor for */
	final private LeaseSet set;

	/**
	 * Simple constructor
	 * @param set The set this monitor is assocated with
	 */
	private SendMonitorImpl(LeaseSet set) {
	    this.set = set;
	}

	// Methods needed to meet contract of SendMonitor
	
	// Inherit java doc from super type
	public void definiteException(EventType           type,
				      RemoteEventListener listener,
				      RemoteEvent         ev,
				      long                registrationNumber,
				      Throwable           t)
	{
	    // This may be more locking than we need (especially
	    // locking on the set ) since EventType objects already
	    // perform a lot of locking, but this discipline is
	    // constant with the rest of Norm and is there for less
	    // likely to lead to bugs down the road.
	    try {
		store.acquireMuatorLock();	    

		if (!setLeasePolicy.ensureCurrent(set))
		    // Set is dead, don't bother updating its state
		    return;

		synchronized (set) {
		    set.definiteException(type, ev, registrationNumber);
		}
	    } finally {
		store.releaseMuatorLock();
	    }    
	}
	
	// Inherit java doc from super type
	public boolean isCurrent() {
	    synchronized (set) {
		return setLeasePolicy.ensureCurrent(set);
	    }
	}
    }


    /**
     * Method used to remove membeship expired leases from the server.
     * Assumes that they have already been removed from the set and 
     * leaseTable is locked.
     * @param deadLeases An iterator with the leases that have to be 
     *                   removed
     */
    private void removeClientLeases(Iterator deadLeases) {
	while(deadLeases.hasNext()) {
	    ClientLeaseWrapper clw = (ClientLeaseWrapper)deadLeases.next();

	    // Remove from lrm
	    try {
		lrm.remove(clw);
	    } catch (UnknownLeaseException e) {
		// This can happen if there was some problem renewing
		// the lease, or its LRM expiration just ran out.
		// Since we are remove the lease any way ignore.
	    }

	    // Remove the lease from ether the leaseTable or the 
	    // deformedLease list
	    leaseTable.remove(clw);
	}
    }

    /**
     * Throw a NoSuchObjectException, wrapped in a ThrowThis if
     * the passed set has expired. Assumes set is locked
     */
    private void ensureCurrent(LeaseSet set) throws ThrowThis {
	if (!setLeasePolicy.ensureCurrent(set)) {
	    throw new ThrowThis(new NoSuchObjectException("Set has expired"));
	}
    }

    /** 
     * Return the set with the specified id, or throw a
     * NoSuchObjectException, wrapped in a ThrowThis if the set can't be found
     */
    private LeaseSet getSet(UUID id) throws ThrowThis {
	final LeaseSet rslt = (LeaseSet)setTable.get(id);
	if (rslt == null) 
	    throw new ThrowThis(new NoSuchObjectException("Can't find set"));
	return rslt;
    }


    /////////////////////////////////////////
    // Methods defined in LeaseRenewalService

    // Inherit java doc from super type
    public LeaseRenewalSet createLeaseRenewalSet(long leaseDuration) {
	final UUID newID      = setIDFactory.newUUID();
	final LeaseSet newSet = new LeaseSet(newID, generator, store, this);
	
	Lease newLease;

	try {
	    newLease = setLeasePolicy.leaseFor(newSet, leaseDuration);
	} catch (LeaseDeniedException e) {
	    // This will never happen because we never use a policy that
	    // denies granting (or renewing) a lease complain bitterly
	    // and throw a runtime exception
	    System.err.println("NormServerImpl.createLeaseRenewalSet:" +
		"Got LeaseDeniedException creating lease - this should not " +
		"happen!");
	    throw new InternalNormException("Error creating lease", e);
	}

	final SetProxy rslt = new SetProxy(ourStub, newID, newLease);

	try {
	    store.acquireMuatorLock();
	    setTable.put(newID, newSet);
	    final Object u = new CreateLeaseSet(newSet);
	    store.update(u);
	} finally {
	    store.releaseMuatorLock();
	}

	return rslt;
    }

    //////////////////////////////
    // Callbacks used by LeaseSet

    /**
     * Method used by <code>LeaseSet</code> when it needs to cons up
     * a SetProxy with an up to-date Lease.  Assumes the appropriate 
     * locks have been obtained.  */
    SetProxy newSetProxy(LeaseSet set) {
	Lease l;
	try {
	    l = leaseFactory.newLease(set.getCookie(), this, 
				      set.getExpiration());
	} catch (LeaseDeniedException e) {
	    // This will never happen because we never use a policy that
	    // denies granting (or renewing) a lease complain bitterly
	    // and throw a runtime exception
	    System.err.println("NormServerImpl.createLeaseRenewalSet:" +
		"Got LeaseDeniedException creating lease - this should not " +
		"happen!");
	    throw new InternalNormException("Error creating lease", e);
	}

	return new SetProxy(this, set.getUUID(), l);
    }

    /**
     *  Create a new <code>SendMonitorImpl</code>
     */
    // $$$ This feels a bit like a kludge, one of the places where this
    // is called back is in LeaseSet's constructor! So the constructor is
    // leting a refence to the contructed object escape before the object
    // is done.  There a couple altertnitves but I am not sure they are
    // beter, the best one is proalby to defer creation of the set's 
    // EventType object until first use and grab the necessary SendMonitor
    // then, but then we will need all sorts of null guards...
    SendMonitor newSendMonitor(LeaseSet set) {
	return new SendMonitorImpl(set);
    }

    //////////////////////////////////////////
    // Thread to persist client lease renewals
    private class RenewLogThread extends Thread {
	/** Create a daemon thread */
	private RenewLogThread() {
	    super("log renewals thread");
	    setDaemon(true);
	}

	public void run() {
	    while (true) {
		try {
		    ClientLeaseWrapper clw;
		    synchronized (renewedList) {
			// If there is an item on the list pull it off for
			// processing, otherwise wait and try again
			if (renewedList.isEmpty()) {
			    try {
				renewedList.wait();
				continue; // go back to top of loop
			    } catch (InterruptedException e) {
				// someone wants us dead
				return;
			    }
			} else {
			    clw = (ClientLeaseWrapper)renewedList.remove(0);
			}			
		    }

		    final PrintWriter writer = debugRenewals;

		    if (writer != null) 
			writer.println("Loggin renewal of " + clw + " at " +
				       System.currentTimeMillis());

		    // A lease was renewed, log the new state
		    final LeaseSet set = clw.getLeaseSet(setTable);
		    if (set == null) 
			// set must have just been removed, no state
			// to update, go to next item in list
			continue;
		    
		    try {	
			store.acquireMuatorLock();  
			synchronized (set) {
			    if (!setLeasePolicy.ensureCurrent(set)) {
				// expired, no state to update, go to next item
				// in list
				continue;
			    }
			    
			    clw.clearRenewed();
			    // Small window here where the lease can
			    // be renewed, it's renewed flag is re-set
			    // and because it was cleared the clw ends
			    // up on renewed list, we then come back
			    // to this thread and get log the new
			    // state, and come back later and log that
			    // state again.  Since this is just
			    // slightly wasteful, not incorrect this
			    // is ok. [Reversing these two lines of
			    // course would be wrong...]
			    set.logRenewal(clw);	
			}
		    } finally {
			store.releaseMuatorLock();
		    }

		    // Give other threads a chance to run
		    Thread.yield();
		} catch (RuntimeException e) {
		    System.err.println("Runtime exception in RenewLogThread" +
				       "...restarting");
		    e.printStackTrace();
		}
	    }
	}
    }


    //////////////////////////////
    // Methods defined in Landlord

    // Inherit java doc from super type
    public long renew(Object cookie, long extension) 
	throws UnknownLeaseException, LeaseDeniedException
    {
	final LeaseSet set = (LeaseSet)setTable.get(cookie);
	if (set == null) 
	    throw new UnknownLeaseException("No lease for cookie:" + cookie);
	try {
	    store.acquireMuatorLock();	    

	    synchronized (set) {
		if (!setLeasePolicy.ensureCurrent(set)) {
		    // Lease has expired, don't renew
		    throw new 
			UnknownLeaseException("Lease has already expired");
		}
		// No one can expire the lease, so it is safe to update.
		// $$$ Might be better to make an extra call to currentTime
		// and calculate the new duration just be for returning
		final long granted = setLeasePolicy.renew(set, extension);

		// Log update 
		final Object u = new LeaseSet.ChangeSetExpiration(set);
		store.update(u);
		return granted;
	    }
	} finally {
	    store.releaseMuatorLock();
	}
    }

    // Inherit java doc from super type
    public void cancel(Object cookie) throws UnknownLeaseException {
	final LeaseSet set = (LeaseSet)setTable.get(cookie);
	if (set == null) 
	    throw new UnknownLeaseException("No lease for cookie:" + cookie);

	try {
	    store.acquireMuatorLock();
	    
	    synchronized (set) {
		if (!setLeasePolicy.ensureCurrent(set)) {
		    //Someone else beat us to it, just return
		    return;
		}

		removeSet(set);
	    }
	} finally {
	    store.releaseMuatorLock();
	}
    }

    /**
     * Do the heavy lifting on removing a set, assumes the locks on the
     * set and store have been aquired.
     */
    private void removeSet(LeaseSet set) {
	setTable.remove(set.getCookie());	
	// set.destroy will persist the removal of the set, and 
	// change the set's expiration to -1 which will cause any
	// other operations on the set to throw NoSuchObjectException
	// or UnknownLeaseException (they will all lock on the set
	// and ensure current before doing anything substave)
	final Set leases = set.destroy(); 
	synchronized (leaseTable) {
	    removeClientLeases(leases.iterator());
	}
    }

    // Inherit java doc from super type
    public RenewResults renewAll(Object[] cookies, long[] extensions) {
	return LandlordUtil.renewAll(this, cookies, extensions);
    }

    // Inherit java doc from super type
    public Map cancelAll(Object[] cookies) {
	return LandlordUtil.cancelAll(this, cookies);	    
    }

    ////////////////////////////////////////////////
    // Methods and classes needed by PersistantStore

    /**
     * Called by <code>PersistentStore</code> after every update to give
     * server a chance to trigger a snapshot
     * @param updateCount Number of updates since last snapshot
     */
    void updatePerformed(int updateCount) {
	// First check to see if the size of the log is larger than a 
	// minimum threshold, this keeps snapshots from happening when 
	// the state of the server is very small (like initially when
	// the snapshot size is zero)
	if (updateCount >= logToSnapshotThresh) {
	    // Compare the size of log to what next snapshot would be
	    final int snapshotSize = setTable.size() + leaseTable.size();
	    if ((float)updateCount >= snapshotWt*((float)snapshotSize)) {
		// Both conditions meet, trigger snapshot
		snapshotter.takeSnapshot();
	    }
	}
    }

    /**
     * Perform the 3rd stage of log recovery, restoring the various pieces of 
     * transient state (populating the LRM, leaseTable, restoring various 
     * transient fields).
     */
    void restoreTransientState() {
	final long now = System.currentTimeMillis();

	for (Iterator i = setTable.values().iterator(); i.hasNext(); ) {

	    final LeaseSet set = (LeaseSet)i.next();

	    // Has this set expired?
	    if (now > set.getExpiration()) {
		// yes it has, remove it and move on
		i.remove();
		// This is all we have to do because none of the other
		// book keeping has been done yet for this set
		continue;
	    }

	    final Iterator leases = set.restoreTransientState(generator, store, 
							      this, now);
	
	    // Go through all the leases in the set and add them to 
	    // the right tables.
	    while (leases.hasNext()) {
		// We sync here so all the updates will be flushed to
		// memory before the the LRM's threads start working on
		// this lease
		ClientLeaseWrapper clw;
		synchronized (this) {
		    clw = (ClientLeaseWrapper)leases.next();

		    // Let the clw recover its state	    
		    clw.recoverTransient(renewedList, idGen);

		    // Put the lease in the leaseTable
		    leaseTable.put(clw);
		}

		// Note: there is no race condition here because
		// failureNotifer and renewedList will buffer any
		// renewals/renewal failure until the rest of the server
		// is ready to handle them
		//
		// Note: in set.restoreTransientState we dropped any
		// client leases who's membership expirations were
		// before `now', however, it is possible that since we
		// set now more leases have reached their
		// membershipExpiration in which case the LRM will
		// silently drop them shortly after we add them, this
		// is ok, they will get cleaned up by the next
		// snapshot and all the other methods are smart enough
		// not be confuses by by membership expired leases.
		final Throwable lt = clw.getLastFailure();

		if ((lt == null) || 
		    (ThrowableConstants.retryable(lt) == 
		     ThrowableConstants.INDEFINITE)) 
		{
		    lrm.renewUntil(clw, clw.getMembershipExpiration(), 
				   clw.getRenewDuration(), failureNotifer);
		} else {
		    // This lease has already suffred a definite failure,
		    // don't let the LRM renew it.  Note we will only get
		    // here if we logged a renewal failure, but could not
		    // log the LeaseRenewalEvent from the LRM
		    failureNotifer.notify(new LeaseRenewalEvent(lrm,
			clw, clw.getMembershipExpiration(), lt));
		}
	    }
	}
    }

    /**
     * Implementation of <code>LogHandler</code> used by NormServerImpl
     */
    private class OurLogHandler extends LogHandler {
	// Inherit java doc from super type
	// Snapshot format
	//     Generator we use to create EventType objects
	//     The number of LeaseSet objects in the setTable
	//     All of the LeaseSet objects
	public void snapshot(OutputStream out) throws Exception {
	    final long now = System.currentTimeMillis();
	    final ObjectOutputStream oostream = new ObjectOutputStream(out);

	    oostream.writeInt(CURRENT_LOG_VERSION);
	    oostream.writeObject(generator);
	    oostream.writeInt(setTable.size());
	    final Collection sets = setTable.values();
	    for (Iterator i = sets.iterator(); i.hasNext();) {
		final LeaseSet set = (LeaseSet)i.next();

		// Grab lock on set so any concurrent getLeases() calls
		// will not be corrupted
		synchronized (set) { 

		    // Take this opportunity to clean up any stale
		    // data
		    final Iterator deadLeases = set.cleanup(now);
		    if (deadLeases != null) {
			synchronized (leaseTable) {
			    removeClientLeases(deadLeases);
			}
		    }

		    oostream.writeObject(set);
		}
	    }
	    oostream.flush();

	    // $$$ We are missing an optimization here, as write each
	    // set we could check to see if any of the client leases
	    // in that set have been renewed, but not logged and clear
	    // them from the renewedList (this assumes that the
	    // renewal changes the persisted state of the wrapper, not
	    // the processing by logRenewal()).  This is not a bug
	    // since any clw we pull off the renewedList and persist
	    // after this point will have state at lease as uptodate
	    // as this snapshot.
	}

	// Inherit java doc from super type
	public void recover(InputStream in) throws Exception {
	    final ObjectInputStream oistream = new ObjectInputStream(in);

	    int version;
	    version = oistream.readInt();

	    if (version != CURRENT_LOG_VERSION) {
		    throw new CorruptedStoreException("Incompatible version " +
		        "ID in log, looking for " + CURRENT_LOG_VERSION +
			", got " + version);

	    }

	    generator = (EventTypeGenerator)oistream.readObject();
	    final int size = oistream.readInt();
	    setTable = Collections.synchronizedMap(new java.util.HashMap(size));
	    for (int i = 0; i < size; i++) {
		final LeaseSet set = (LeaseSet)oistream.readObject();
		setTable.put(set.getUUID(), set);
	    }
	}

	// Inherit java doc from super type
	public void applyUpdate(Object update) throws Exception {
	    final LoggedOperation op = (LoggedOperation)update;
	    op.apply(setTable);
	}
    }

    /**
     * Return a string summarizing the inventory of the server 
     */
    private String inventory() {
	return leaseTable.size() + " client leases (" + 
	    leaseTable.deformedCount() + " deformed), " + setTable.size() +
	    " sets.";
    }

    /**
     * Thread that performs the actual snapshots, done in a separate thread
     * so it will not hang up in progress remote calls
     */
    private class SnapshotThread extends Thread {
	/** Create a daemon thread */
	private SnapshotThread() {
	    super("snapshot thread");
	    setDaemon(true);
	}

	/** Signal this thread that is should take a snaps */
	private synchronized void takeSnapshot() {
	    notifyAll();
	}

	public void run() {
	    while (!isInterrupted()) {
		synchronized (this) {
		    try {
			wait();
		    } catch (InterruptedException e) {
			return;
		    }
		}

		try {
		    final PrintWriter writer = debugSnapshot;
		    if (writer != null) 
			writer.println("Taking snapshot:" + inventory());

		    store.snapshot();

		    if (writer != null) {
			writer.println("Snapshot done:" + inventory());
		    }
		} catch (InterruptedIOException e) {
		    return;
		} catch (Exception e) {
		    if (e instanceof LogException &&
			((LogException)e).detail instanceof
			InterruptedIOException)
			return;
		    /* $$$
		     * if taking the snapshot fails for any reason,
		     * then one of the following must be done:
		     *   -- output the problem to a file and exit
		     *   -- output the problem to a file and continue
		     *   -- set an "I have a problem" attribute and
		     *      then send a notification
		     * this issue will be addressed at a later time
		     */
		    e.printStackTrace();
		}
	    }
	}
    }

    ///////////////////////////////////
    // Methods defined in Administrable

    // Inherit java doc from super type
    public Object getAdmin() {
	return this;
    }

    ///////////////////////////////
    // Methods defined in JoinAdmin

    // Inherit java doc from super type
    public Entry[] getLookupAttributes() {
	return joinState.getAttributes();
    }

    // Inherit java doc from super type
    public void addLookupAttributes(Entry[] attrSets) {
	joinState.addAttributes(attrSets, true);
    }

    // Inherit java doc from super type
    public void modifyLookupAttributes(Entry[] attrSetTemplates, 
				       Entry[] attrSets) 
    {
	joinState.modifyAttributes(attrSetTemplates, attrSets, true);
    }
  
    // Inherit java doc from super type
    public String[] getLookupGroups() {
	return joinState.getGroups();
    }

    // Inherit java doc from super type
    public void addLookupGroups(String[] groups) {
	joinState.addGroups(groups);
    }

    // Inherit java doc from super type
    public void removeLookupGroups(String[] groups) {
	joinState.removeGroups(groups);
    }

    // Inherit java doc from super type
    public void setLookupGroups(String[] groups) {
	joinState.setGroups(groups);
    }

    // Inherit java doc from super type
    public LookupLocator[] getLookupLocators() {
	return joinState.getLocators();
    }

    // Inherit java doc from super type
    public void addLookupLocators(LookupLocator[] locators) {
	joinState.addLocators(locators);
    }

    // Inherit java doc from super type
    public void removeLookupLocators(LookupLocator[] locators) {
	joinState.removeLocators(locators);
    }

    // Inherit java doc from super type
    public void setLookupLocators(LookupLocator[] locators) {
	joinState.setLocators(locators);
    }

    //////////////////////////////////
    // Methods defined in DestroyAdmin

    // Inherit java doc from super type
    public void destroy() {
	Exception failed = null;

	try {
	    ActivationGroupID grp = ActivationGroup.currentGroupID();
	    ActivationSystem sys = ActivationGroup.getSystem();
	    sys.unregisterGroup(grp);
	} catch (Exception t) {
	    System.err.println("While unexporting and/or unregistering " +
			       "with RMID :");
	    t.printStackTrace(System.err);
	    System.err.println("destroy continuing");
	}

	try {
	    store.destroy();
	} catch (Exception t) {
	    System.err.println("While destroying persistent store:");
	    t.printStackTrace(System.err);
	    System.err.println("destroy continuing");
	}

	killVM(false);	// go away
    }

    // inherit doc comment
    public long killVM() {
	return killVM(true);
    }

    /**
     * Sleep time for <code>killVM</code>.
     *
     * @see killVM
     */
    private final static long KILL_DELAY = 10;

    /**
     * Kill the VM after a sutable delay.  Return a long
     * indicating at lease how long clients should wait before
     * calling to the service
     * @param log If true dump a message to System.err with the
     * the current time before calling exit.
     */
    private long killVM(final boolean log) {
	/*
	 * We have to do this in a separate thread or the remote
	 * system will always get a RemoteException when the server
	 * to which it is talking (this one) dies in the middle of
	 * the call.  So we start up a new thread that sleeps for a
	 * hundredth of a second and then shuts down the VM.  This lets
	 * the thread running shutDown return before the denouement.
	 */
	new Thread() {
	    public void run() {
		try {
		    sleep(KILL_DELAY);
		} catch (InterruptedException e) {
		    ;	// oh, well
		}
		if (log) 
		    System.err.println("Killing VM at " + 
				       System.currentTimeMillis());
		System.exit(0);
	    }
	}.start();
	return KILL_DELAY * 100;
    }

    //////////////////////////////////
    // Methods defined in DestroyAdmin

    // Inherit java doc from super type
    public synchronized void setSnapshotWeight(float weight) {
	snapshotWt = weight;
    }

    // Inherit java doc from super type
    public synchronized float getSnapshotWeight() {
	return snapshotWt;
    }

    // Inherit java doc from super type
    public synchronized void setLogToSnapshotThreshold(int threshold) {
	logToSnapshotThresh = threshold;
    }

    // Inherit java doc from super type
    public synchronized int getLogToSnapshotThreshold() {
	return logToSnapshotThresh;
    }


    ////////////////////////////////////
    // Server setup and activation code

    // Activation

    /**
     * If for some reason we can't activate we dump a message to the log 
     * and unregister with activation.
     * @param m message to dump to <code>System.err</code>
     * @param e If non-<code>null</code> this exception is dumped to
     * <code>System.err</code> as well 
     */
    private void alertAndUnregister(String m, Throwable e) {
	System.err.println("Fatal error during activation:" + m);
	if (e != null)
	    e.printStackTrace();
	
	Exception failed = null;

	System.err.println("Attempting to unregister with activation");
	try {
	    ActivationGroupID grp = ActivationGroup.currentGroupID();
	    ActivationSystem sys = ActivationGroup.getSystem();
	    sys.unregisterGroup(grp);
	} catch (ActivationException ee) {
	    failed = ee;
	} catch (RemoteException ee) {
	    failed = ee;
	}

	if (failed != null) {
	    System.err.println("Failed in unregister attempt");
	    failed.printStackTrace();
	}
	
	System.err.println("Succeeded in unregistering");
    }

    /**
     * Activation constructor
     *
     * @param activationID activation ID passed in by the activation daemon.
     * @param data state data needed to re-activate a Norm server
     */
    public NormServerImpl(ActivationID activationID, MarshalledObject data) 
	throws IOException, ClassNotFoundException
    {
	boolean runtimeThrown = false;
	System.err.println("Norm is activating");
	this.activationID = activationID;

	try {
	    try {
		ourStub = 
		    (NormServer)Activatable.exportObject(this, activationID, 0);
	    } catch (RemoteException e) {
		 alertAndUnregister("Failure exporting object", e);
		 throw e;
	    }

	    // Get Log Dir Name
	    try {
		logDir = (String)data.get();
	    } catch (IOException e) {
		alertAndUnregister("Failure recoving activation data", e);
		throw e;
	    } catch (ClassNotFoundException e) {
		alertAndUnregister("Failure recoving activation data", e);
		throw e;	  
	    } catch (ClassCastException e) {
		alertAndUnregister("Failure recoving activation data", e);
		runtimeThrown = true;
		throw e;
	    }

	    // We use some of these during the recovery process
	    failureNotifer = new FailureNotifer(this);
	    renewLogger = new RenewLogThread();
	    snapshotter = new SnapshotThread();

	    try {
		store = new PersistentStore(logDir, new OurLogHandler(), this);
		// Creating the store completes the first two stages of 
		// log recovery (reading the snapshot and the updates)
		// Perform the last stage here of restoring transient state
		restoreTransientState();

		final PrintWriter writer = debugSnapshot;
		if (writer != null) 
		    writer.println("Log recovered:" + inventory());

	    } catch (StoreConflictException e) {
		alertAndUnregister("Another Norm Server is using this " +
		    "log (" + logDir + ")", e);
		runtimeThrown = true;
		throw new IllegalStateException("Trying to start a Norm " +
		    "server with a log owned by another server");
	    } catch (CorruptedStoreException e) {
		alertAndUnregister("Log corrupted, can't recover ", e);
		runtimeThrown = true;
		throw new IllegalStateException("Corrupted log");
	    } catch (StoreException e) {
		alertAndUnregister("Can't recover log", e);
		runtimeThrown = true;
		throw new IllegalStateException("Corrupted log");
	    }

	    // Create new baseline snapshot
	    try {
		final PrintWriter writer = debugSnapshot;
		if (writer != null) 
		    writer.println("Taking new baseline snapshot:"
				   + inventory());

		store.snapshot();

		if (writer != null) {
		    writer.println("Baseline snapshot done:" + inventory());
		}
	    } catch (IOException e) {
		alertAndUnregister("Can't create new baseline snapshot", e);
		throw e;
	    }

	    try {
		joinState = new JoinState(this, lrm);
		store.addSubStore(joinState);
	    } catch (StoreException e) {
		alertAndUnregister("Can't create JoinState", e);
		runtimeThrown = true;
		throw new IllegalStateException("Can't create JoinState");
	    }

	    // Make sure we have a ServiceInfo and BasicServiceType
	    // attribute
	    // $$$ Should we really being doing replacment here?
	    // If this is an inplace upgrade we proably want to register
	    // new attributes (maybe check to see if the current ones
	    // need to be replaced?)
	    try {
		final Entry[] currentAttrs = joinState.getAttributes();
		boolean haveServiceInfo = false;
		boolean haveBasicServiceType = false;

		for (int i=0; i<currentAttrs.length; i++) {
		    if (currentAttrs[i] instanceof ServiceInfo)
			haveServiceInfo = true;
		    else if (currentAttrs[i] instanceof BasicServiceType)
			haveBasicServiceType = true;
		}

		Entry[] toAdd = null;
		if (!haveServiceInfo && !haveBasicServiceType)
		    toAdd = new Entry[2];
		else if (!haveServiceInfo || !haveBasicServiceType)
		    toAdd = new Entry[1];
		
		if (toAdd != null) {
		    int i = 0;
		    if (!haveServiceInfo) {
			toAdd[i] = new ServiceInfo("Lease Renewal Service", 
		            "Sun Microsystems, Inc.",  
			    "Sun Microsystems, Inc.", 
			    VersionConstants.SERVER_VERSION,
			    "", "");
			i++;
		    } 

		    if (!haveBasicServiceType) {
			toAdd[i] = 
			    new BasicServiceType("Lease Renewal Service");
			i++;
		    }
			
		    joinState.addAttributes(toAdd, false);
		}
	    } catch (Exception e) {
 		System.err.println("Problem setting up service controlled " +
				   "attributes");
		e.printStackTrace();
		System.err.println("Ignoring and continuing");
	    }

	} catch (RuntimeException e) {
	    if (runtimeThrown == false) { 
		alertAndUnregister("Unexpected exception during activation",
				   e);
	    }

	    throw e;
	} catch (Error e) {
	    alertAndUnregister("Unexpected exception during activation", e);
	    throw e;
	}	  

	try {
	    leaseFactory = new CodebaseLeaseFactory();
	} catch (Exception e) {
	    alertAndUnregister("Can't create CodebaseLeaseFactory", e);
	    runtimeThrown = true;
	    throw new 
		IllegalStateException("Can't create CodebaseLeaseFactory");
	}

	setLeasePolicy = 
	    new LeaseDurationPolicy(
	        1000 * 3600 * 2, // At most grant 2 hour leases
		1000 * 3600 * 1, // default lease is one hour
		ourStub, expMgr, leaseFactory);

	// $$$ By rights this code should be in
	// restoreTransientState(), however we can't have an independent
	// thread running around changing persistant state util we get
	// to this point (I think the only real issue is the baseline
	// snapshot) and once we place a set in the expMgr it the
	// underlying wakeup queue will start running which can cause
	// calls to expireIfTime() (and sendWarningEvent() though those
	// should not be a problem since they don't log anything).
	//
	// I would prefer to ether modify WakeupQueue() to have a "start now"
	// method (eqv to how we create failureNotifer above but call 
	// start() bellow) or be able to hold an exclusive snapshot
	// lock until the initial snapshot is done.  Ether should allow
	// this code to be moved into restoreTransientState

	for (Iterator i = setTable.values().iterator(); i.hasNext(); ) {
	    final LeaseSet set = (LeaseSet)i.next();
	    synchronized (set) {
		expMgr.schedule(set);
	    }
	}

	failureNotifer.start();
	renewLogger.start();
	snapshotter.start();

	System.err.println("Norm is activated");
    }
}
