/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.norm;

import java.io.Serializable;

import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.MarshalledObject;

import java.util.List;
import java.util.LinkedList;

import net.jini.core.lease.Lease;
import net.jini.core.event.EventRegistration;
import net.jini.core.event.RemoteEventListener;
import net.jini.lease.LeaseRenewalSet;
import net.jini.lease.LeaseUnmarshalException;

import com.sun.jini.proxy.UUID;
import com.sun.jini.proxy.ThrowThis;

/**
 * Client side proxy for Norm's lease renwal sets.  Used an object of
 * type NormServer to communicate back to server.
 */
class SetProxy implements LeaseRenewalSet, Serializable {
    /** 
     * ID of the set this object is a proxy for
     * @serial
     */
    final private UUID setID;

    /** 
     * Lease for this set
     * @serial
     */
    final private Lease ourLease;
     
    /**
     * Remote ref to server we belong to 
     * @serial
     */
    final private NormServer server;


    /**
     * Simple constructor
     * @param server object proxy should call back to
     * @param id ID of set this proxy is for
     * @param lease on for set this is proxy is for
     */
    SetProxy(NormServer server, UUID id, Lease lease) {
	this.server = server;
	setID = id;
	ourLease = lease;
    }

    // Inherit java doc from super type
    public void renewFor(Lease leaseToRenew, long membershipDuration)
	throws RemoteException 
    {
	renewFor(leaseToRenew, membershipDuration, Lease.FOREVER);
    }

    // Inherit java doc from super type
    public void renewFor(Lease leaseToRenew, long membershipDuration,
			 long  renewDuration)
	throws RemoteException 
    {
	// Do some client side error checking
	if (leaseToRenew == null) 
	    throw new NullPointerException("LeaseRenewalSet.renewFor:Must " +
					   "pass a non-null lease");

	if ((membershipDuration != Lease.FOREVER) &&
	    (renewDuration == Lease.ANY))
	    throw new IllegalArgumentException(
	        "LeaseRenewalSet.renewFor:renewDuration can only be " +
		"Lease.ANY if membershipDuration is Lease.FOREVER");

	if (!(renewDuration == Lease.ANY || renewDuration == Lease.FOREVER ||
	      renewDuration > 0))
	    throw new IllegalArgumentException(
	        "LeaseRenewalSet.renewFor:renewDuration can only be " +
		"Lease.ANY, Lease.FOREVER, or positive");

	// It would be nice to do the check here on leaseToRenew being
	// granted by this set, however, because we are trying to keep
	// LandlordLease out of the norm.jar we would then also have
	// to keep SetProxy out of norm.jar which could be a real
	// hassle, so for now we will do the check on the server side.

	try {
	    server.renewFor(setID, leaseToRenew, membershipDuration, 
			    renewDuration);
	} catch (ThrowThis e) {
	    // The server wants to throw some remote exception
	    e.throwRemoteException();
	}
    }

    // Inherit java doc from super type
    public Lease remove(Lease leaseToRemove) throws RemoteException {
	// Do some client side error checking
	if (leaseToRemove == null) 
	    throw new NullPointerException("LeaseRenewalSet.remove:Must " +
	        "pass a non-null lease");

	try {
	    return server.remove(setID, leaseToRemove);
	} catch (ThrowThis e) {
	    // The server wants to throw some remote exception
	    e.throwRemoteException();
	}

	// Need this because compiler does not know that
	// throwRemoteException is not going to return
	return null;
    }

    // Inherit java doc from super type
    public Lease[] getLeases() throws LeaseUnmarshalException, RemoteException {
	try {
	    MarshalledObject mls[] = server.getLeases(setID);

	    if (mls == null || mls.length == 0) 
		return new Lease[0];
	    
	    final List leases = new LinkedList();
	    final List problems = new LinkedList();
	    final List exceptions = new LinkedList();
	    for (int i=0; i<mls.length; i++) {
		Lease l;
		final MarshalledObject ml = mls[i];
		try {
		    l = (Lease)ml.get();
		    leases.add(l);
		} catch (Throwable t) {
		    problems.add(ml);
		    exceptions.add(t);
		}		
	    }

	    final Lease[] rslt =
		(Lease[])leases.toArray(new Lease[leases.size()]);

	    if (problems.isEmpty())
		return rslt;
	    
	    final MarshalledObject sml[] =
		(MarshalledObject[])problems.toArray(
                    new MarshalledObject[problems.size()]);
	    final Throwable es[] =
		(Throwable[])exceptions.toArray(
		    new Throwable[exceptions.size()]);

	    throw new LeaseUnmarshalException("Problem unmarshalling lease(s)",
		rslt, sml, es);
	    
	} catch (ThrowThis e) {
	    // The server wants to throw some remote exception
	    e.throwRemoteException();
	}

	// Need this because compiler does not know that
	// throwRemoteException is not going to return
	return null;
    }

    // Inherit java doc from super type
    public EventRegistration setExpirationWarningListener(
	                         RemoteEventListener listener, 
				 long                minWarning, 
				 MarshalledObject    handback)
	throws RemoteException    
    {
	// Do some client side error checking
	if (listener == null) 
	    throw new NullPointerException(
	        "LeaseRenewalSet.setExpirationWarningListener:Must " +
		"pass a non-null listener");

	if (minWarning < 0)
	    throw new IllegalArgumentException(
	        "LeaseRenewalSet.setExpirationWarningListener:minWarning " +
		"must be positive");

	try {
	    return server.setExpirationWarningListener(setID,
	        listener, minWarning, handback);
	} catch (ThrowThis e) {
	    // The server wants to throw some remote exception
	    e.throwRemoteException();
	}

	// Need this because compiler does not know that
	// throwRemoteException is not going to return
	return null;
    }

    // Inherit java doc from super type
    public void clearExpirationWarningListener() throws RemoteException {
	try {
	    server.setExpirationWarningListener(setID, null, 
						NormServer.NO_LISTENER, null);
	} catch (ThrowThis e) {
	    // The server wants to throw some remote exception
	    e.throwRemoteException();
	}
    }

    // Inherit java doc from super type
    public EventRegistration setRenewalFailureListener(
				 RemoteEventListener listener, 
				 MarshalledObject    handback)
	throws RemoteException
    {
	// Do some client side error checking
	if (listener == null) 
	    throw new NullPointerException(
	        "LeaseRenewalSet.setRenewalFailureListener:Must " +
		"pass a non-null listener");

	try {
	    return server.setRenewalFailureListener(setID, listener, handback);
	} catch (ThrowThis e) {
	    // The server wants to throw some remote exception
	    e.throwRemoteException();
	}

	// Need this because compiler does not know that
	// throwRemoteException is not going to return
	return null;
    }


    // Inherit java doc from super type
    public void clearRenewalFailureListener() throws RemoteException {
	try {
	    server.setRenewalFailureListener(setID, null, null);
	} catch (ThrowThis e) {
	    // The server wants to throw some remote exception
	    e.throwRemoteException();
	}
    }

    // Inherit java doc from super type
    public Lease getRenewalSetLease() {
	return ourLease;
    }

    // Inherit java doc from super type
    public int hashCode() {
	return setID.hashCode();
    }

    // Inherit java doc from super type
    public boolean equals(Object obj) {
	if (!(obj instanceof SetProxy))
	    return false;

	final SetProxy that = (SetProxy)obj;

	return (setID.equals(that.setID) && server.equals(that.server));
    }
}
