/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.norm.event;

import java.io.IOException;
import java.rmi.MarshalledObject;
import java.util.List;

import net.jini.core.event.RemoteEvent;
import net.jini.core.event.RemoteEventListener;
import net.jini.core.event.UnknownEventException;

import com.sun.jini.thread.TaskManager;
import com.sun.jini.thread.RetryTask;
import com.sun.jini.constants.ThrowableConstants;

/**
 * Representation of an event type the supports a single registrant.
 * The registrant, event ID, and event sequence number information
 * portions of this object are preserved when serialized.  The send
 * monitor, and currently scheduled event sends are not.
 */
// $$$ I am not sure if this class does too much locking, or too little.
// It does too little in that it does not do any locking during serialization
// It does too much in that Norms current locking discipline means that for
// the most part this classes does not have to worry about multiple threads
// of control entering at one time.
// Before we make this a general utility we have to re-think the locking
// strategy.
public class EventType implements java.io.Serializable {
    /**
     * Listener registered for events of this type.  Stored in 
     * marshaled form so object the <code>EventType</code> object can
     * be recovered even if the listener's codebase is not available.
     * If this field is <code>null</code>, <code>listener</code> and 
     * <code>handback</code> will be also.  
     * @serial
     */
    private MarshalledObject marshaledListener;

    /** Transient cache of listener in unmarshaled form */
    private transient RemoteEventListener listener;

    /**
     * Handback object associated with current listener.
     * @serial
     */
    private MarshalledObject handback;

    /** 
     * Sequence number of the current listener/handback pair, incremented
     * every time a new listener is set (even if the objects 
     * are equivalent)
     */
    private long registrationNumber;

    /**
     * Last event sequence number we used
     * @serial
     */
    private long lastSeqNum;

    /**
     * Our event ID
     * @serial
     */
    private long evID;

    /** 
     * Object we check with to ensure leases have not expired and notify
     * when we get a definite exception during an event send attempt	
     * If this field is <code>null</code> generator will be also
     */
    private transient SendMonitor monitor;

    /** 
     * Event type generator that created us. If this field is
     * <code>null</code> monitor will be also.
     */
    private transient EventTypeGenerator generator;

    /**
     * Simple constructor.  Initially the last sequence number is set to 0.
     * @param generator EventTypeGenerator that is creating this event type
     * @param monitor   Object which is to monitor the sending of events
     *                  of this type.
     * @param evID      Event ID of this event type
     * @param listener  The listener events of this type should go to
     * @param handback  The object that should be passed to listener
     *                  as part of the event.     
     * @throws IOException if listener can not be serialized 
     */
    EventType(EventTypeGenerator generator, SendMonitor monitor, long evID,
	      RemoteEventListener listener, MarshalledObject handback)
        throws IOException
    {
	if (generator == null)
	    throw new NullPointerException("EventType(): Must create event " +
                "type objects with a non-null generator");

	if (monitor == null)
	    throw new NullPointerException("EventType(): Must create event " +
                "type objects with a non-null monitor");

	this.generator = generator;
	this.monitor = monitor;
	this.evID = evID;
	setLastSequenceNumber(0);
	setListener(listener, handback);
    }    

    /** Utility method to null out listener */
    private void clearListener() {
	this.listener = null;
	this.handback = null;
	marshaledListener = null;	    
    }
       

    /**
     * (Re)set the listener for this type of event.  Any pending
     * events that have not yet been sent will be sent to the new
     * listener, passing the new handback.  Setting the listener to
     * <code>null</code> will cancel the sending of all pending
     * events.  
     * @param listener The listener events of this type should go to
     * @param handback The object that should be passed to listener
     *                 as part of the event.     
     * @throws IOException if listener can not be serialized */
    public synchronized void setListener(RemoteEventListener listener, 
					 MarshalledObject    handback)
        throws IOException
    {
	registrationNumber++;
	
	if (listener == null) {
	    clearListener();
	} else {	    
	    marshaledListener = new MarshalledObject(listener);
	    this.listener = listener;
	    this.handback = handback;
	}
    }

    /**
     * Returns <code>true</code> if there is a listener registered for this 
     * event type.
     */
    public boolean haveListener() {
	return marshaledListener != null;
    }

    /**
     * Convince method to get the listener.
     * @return The listener or <code>null</code> if the listener can't be
     * unpacked or there is no listener.
     */
    private RemoteEventListener getListener() {
	if (!haveListener()) 
	    return null;

	if (listener != null) 
	    return listener;

	// There is a listener, but it is not unpacked yet, try to unpack
	try {
	    listener = (RemoteEventListener)marshaledListener.get();
	} catch (IOException e) {
	    // swallow and return null bellow

	    // $$$ is this really the right thing to do?
	    // we proably really have a corrupted marshaledListener here
	} catch (ClassNotFoundException e) {
	    // swallow and return null bellow
	}

	// If get() threw an exception listener is still null
	return listener;
    }

    /**
     * Atomically clear the current registration if its sequence
     * number matches the passed in sequence number.  If the
     * replacement occurs it will have the same effect as calling
     * <code>setListener(null, null)</code>.
     * <p>
     * Can be used by code that needs to remove event registrations in
     * response to exceptions thrown during event delivery without
     * risking clobbering new registrations.
     * @param oldSequenceNumber Sequence number of the 
     *        registration that had a problem 
     * @return <code>true</code> if the state of the object was
     * changed and <code>false</code> otherwise.
     * @see EventType#setListener 
     */
    public synchronized boolean clearListenerIfSequenceMatch(
        long oldSequenceNumber) 
    {
	if (oldSequenceNumber == registrationNumber) {
	    clearListener();
	    return true;
	}

	return false;
    }

    /**
     * Set the object's notion of the last sequence number.  The next event
     * scheduled to be sent will have a sequence number one greater than
     * the value past to this call.
     * <p>
     * Note: this method is not synchronized.
     * @param seqNum value for the last sequence number
     */
    public void setLastSequenceNumber(long seqNum) {
	lastSeqNum = seqNum;
    }

    /**
     * Return the sequence number of the last event that was scheduled
     * to be sent.  Intended primarily for creating
     * <code>EventRegistration</code> objects and the like.
     */
    public synchronized long getLastSequenceNumber () {
	return lastSeqNum;
    }
    
    /**
     * Return the <code>long</code> that was uniquely associated with this
     * object when it was created.
     */
    public long getEventID() {
	return evID;
    }

    /**
     * Schedule the sending of an event.  This event will be sent to
     * the currently registered listener.  If the listener changes
     * before the the event is successfully sent the event will be sent
     * to the new listener.  If the currently listener is
     * <code>null</code> this call will have no affect.
     * @param factory An object that will be used to create the
     * <code>Event</code> object when necessary.
     * @return The sequence number assigned to the event, if the
     * current listener is <code>null</code> this will be the same
     * value as returned by the last call to this method.
     * @throws IllegalStateException if this method is called
     *         after the object has be deserialized and before
     *         <code>restoreTransientState</code> has been called
     * @see EventType#restoreTransientState
     */
    public synchronized long sendEvent(EventFactory factory) {
	if (generator == null) {
	    // Have not had our state restored, complain
	    throw new IllegalStateException("EventType.sendEvent:" +
	        "called before state was fully restored");
	}

	// If we don't have a listener we don't do anything
	if (!haveListener())
	    return lastSeqNum;
	
	final TaskManager mgr = generator.getTaskManager();
	mgr.add(new SendTask(mgr, factory, ++lastSeqNum));
	
	return lastSeqNum;
    }

    /**
     * Restore the transient state of this object after recovering it
     * from a serialization stream.  Nether argument can be
     * <code>null</code>.
     * <p>
     * Note: this method is not synchronized.  
     * @param generator The <code>EventTypeGenerator</code> that was used
     *                  to create this EventType object originally.
     * @param monitor   The object that monitors the progress of events
     *                  set by this object.  */
    public void restoreTransientState(EventTypeGenerator generator,  
                                      SendMonitor        monitor)
    {
	if (generator == null)
	    throw new NullPointerException("EventType.restoreTransientState:" +
	        "Must called with a non-null generator");

	if (monitor == null)
	    throw new NullPointerException("EventType.restoreTransientState:" +
	        "Must called with a non-null monitor");

	this.generator = generator;
	this.monitor   = monitor;
	generator.recoverEventID(evID);
    }

    /**
     * Sub-class of <code>RetryTask</code> used by <code>EventType</code>
     * to send events
     */
    private class SendTask extends RetryTask {
	/** Max time we are willing to let a send attempt to go on for */
	final static private long MAX_TIME = 1000 * 60 * 60 * 24; //~ 1 Day
	
	/** Factory used to create the <code>RemoteEvent</code> to be sent */
	final private EventFactory eventFactory;

	/** Sequence number the event should have */
	final private long seqNum;

	/** Cached event */
	private RemoteEvent event;
	
	/** 
	 * Registration seqence number of the listener/handback pair
	 * event was built for
	 */
	private long eventForRegistrationNumber = -1;

	/**
	 * Simple constructor
	 * @param taskManager <code>TaskManager</code> this task is to be
	 *                    put into.
	 * @param eventFactory <code>EventFactory</code> that will be used
	 *                     to create the event to be sent.
	 * @param seqNum      The sequence number of the event
	 */
	private SendTask(TaskManager taskManager, EventFactory eventFactory,
			 long seqNum)
	{
	    super(taskManager);
	    this.eventFactory = eventFactory;
	    this.seqNum = seqNum;
	}

	// Inherit java doc from super type
	public boolean tryOnce() {
	    final long now = System.currentTimeMillis();
	    if (now - startTime() > MAX_TIME)
		return true;	// we have been trying too long, stop here

	    if (!EventType.this.monitor.isCurrent())
		return true;	// lease gone, stop here

	    // Local copies of listener and handback so they won't
	    // be clobbered by setListener calls
	    RemoteEventListener listener;
	    MarshalledObject handback;
	    long registrationNumber;
	    
	    synchronized (EventType.this) {
		if (!EventType.this.haveListener())
		    return true; // No currently registered listener, stop here

		listener = EventType.this.getListener();
		if (listener == null)
		    return false; // There is a listener, but we can't unpack it
		                  // schedule a re-try latter

		handback = EventType.this.handback;
		registrationNumber = EventType.this.registrationNumber;
	    }

	    // If the handback has changed we need to create a new
	    // event object (we aproximate this test by checking the
	    // registrationNumber

	    if (event == null || 
		eventForRegistrationNumber != registrationNumber) 
	    {
		event = eventFactory.createEvent(EventType.this.evID, seqNum,
						 handback);
		eventForRegistrationNumber = registrationNumber;
	    }

	    // Try sending 
	    try {
		listener.notify(event);
		return true;
	    } catch (Throwable t) {
		// Classify the exception using ThrowableConstants, if
		// it is a bad object or uncategorized (which must be
		// a UnknownEventException) drop the event
		// registration, if it is a bad invocation mark the
		// try is done (since re-sending this event won't work), 
		// but don't drop the registration.  If indefinite return 
		// false so a retry will be scheduled.
		final int cat = ThrowableConstants.retryable(t);
		if (cat == ThrowableConstants.INDEFINITE) {
		    return false;
		} else if (cat == ThrowableConstants.BAD_INVOCATION) {
		    return true;
		} else {
		    EventType.this.monitor.definiteException(EventType.this,
		        listener, event, registrationNumber, t);
		    return true;
		}
	    }
	    
	}

	// Inherit java doc from super type
	public boolean runAfter(List tasks, int size) {
	    // We don't need to run these tasks in any particular order
	    return false;
	}
    }
}
