/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.norm.event;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.rmi.MarshalledObject;
import net.jini.core.event.RemoteEventListener;
import com.sun.jini.thread.TaskManager;

/**
 * Factory class for <code>EventType</code> objects.  All
 * <code>EventType</code> objects created by the same generator (or
 * associated with the same generator by a
 * <code>EventType.restoreTransientState<code> call) will use the same
 * thread pool to manage their event send threads.
 *
 * @see EventType 
 * @see EventType.restoreTransientState
 */
public class EventTypeGenerator implements java.io.Serializable {
    /**
     * Next event ID
     * @serial
     */
    private long nextEvID = 1;

    /**
     * Task manager used to send events
     */
    private transient TaskManager taskManager = new TaskManager();

    /**
     * Create a new <code>EventType</code> object.  The event ID for
     * this type will be generated by this call.
     * <p>
     * @param listener the listener initially associated with the
     *                 event type. May be <code>null</code>.
     * @param handback The handback object initially associated with
     *                 the event type. May be <code>null</code>.
     * @param monitor  Object to callback when an event sending
     *        attempt fails with a definite exception and to 
     *        ensure that the lease on the event is still current.
     *        May not be <code>null</code>
     * @return The new <code>EventType</code> object.
     * @throws IOException if listener can not be serialized 
     */
    public synchronized EventType newEventType(RemoteEventListener listener, 
					       MarshalledObject    handback,
					       SendMonitor         monitor)
	throws IOException
    {
	EventType rslt = new EventType(this, monitor, nextEvID, 
				       listener, handback);
	// Do this afterward so we don't use up an ID if 
	// we can't create the new object
	nextEvID++;
	return rslt;
    }

    /**
     * Create a new <code>EventType</code> object specify the 
     * event id it should have.
     * <p>
     * @param listener the listener initially associated with the
     *                 event type. May be <code>null</code>.
     * @param handback The handback object initially associated with
     *                 the event type. May be <code>null</code>.
     * @param eventID  The event ID of this type.
     * @param monitor  Object to callback when an event sending
     *        attempt fails with a definite exception and to 
     *        ensure that the lease on the event is still current.
     *        May not be <code>null</code>
     * @return The new <code>EventType</code> object.
     * @throws IOException if listener can not be serialized 
     */
    public EventType newEventType(RemoteEventListener listener, 
				  MarshalledObject    handback,	  
				  SendMonitor         monitor,
				  long                eventID)
	throws IOException
    {
	return new EventType(this, monitor, eventID, listener, handback);
    }

    /**
     * Called by event types during transient state recovery to ensure
     * the generator knows about there event ID
     * <p>
     * Note: this method is not synchronized.  
     * @param evID Event ID of recovered <code>EventType</code> object
     */
    void recoverEventID(long evID) {
	if (evID >= nextEvID)
	    nextEvID = evID + 1;
    }

    /**
     * Return the task manager that <code>EventType</code> objects created
     * by this generator should use to send their events.
     */
    TaskManager getTaskManager() {
	return taskManager;
    }

    /**
     * Override <code>readObject</code> to create a <code>TaskManager</code>
     * @see java.io.ObjectInputStream#defaultReadObject docs
     */
    private void readObject(ObjectInputStream in)
	throws IOException, ClassNotFoundException
    {
	// fill in the object from the stream 
	in.defaultReadObject();

	taskManager = new TaskManager();
    }
}
