/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.norm.lookup;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;

import java.rmi.MarshalledObject;

import java.util.List;

import net.jini.core.discovery.LookupLocator;
import net.jini.core.lookup.ServiceID;
import net.jini.core.entry.Entry;

import net.jini.discovery.DiscoveryGroupManagement;
import net.jini.discovery.LookupDiscoveryManager;
import net.jini.lookup.ServiceIDListener;
import net.jini.lookup.JoinManager;
import net.jini.lease.LeaseRenewalManager;

import com.sun.jini.reliableLog.LogHandler;
import com.sun.jini.reliableLog.ReliableLog;

/**
 * Utility class that combines <code>JoinManager</code> with persistence
 */
public class JoinState extends LogHandler implements SubStore, 
    ServiceIDListener 
{
    /** Our join manager */
    private JoinManager joinMgr;

    /**
     * Lease renewal manager (if any) that the client wants our
     * <code>JoinManager</code> to use.
     */
    final private LeaseRenewalManager lrm;

    /**
     * Service we are registering with lookup services
     */
    final Object service;

    /**
     * The <code>LookupDiscoveryManager</code> we are using to find lookups
     */
    private LookupDiscoveryManager ldm;

    /**
     * Pass through to get our serviceID from the log recovery method
     * to the code that creates the <code>JoinManager</code>
     */
    private ServiceID serviceID;

    /**
     * Pass through to get attribute array from the log recovery method 
     * to the code that creates the <code>JoinManager</code>
     */
    private Entry[] attributes;

    /**
     * Pass through to get lookup locator array from the log
     * recovery method to the code that creates the
     * <code>JoinManager</code> */
    private LookupLocator locators[];

    /**
     * Pass through to get lookup group array from the log
     * recovery method to the code that creates the
     * <code>JoinManager</code> 
     */
    private String groups[] = DiscoveryGroupManagement.NO_GROUPS;
    
    /** Log we are using to persist our state to disk */
    private ReliableLog log;

    /**
     * Simple constructor
     * @param service The object to register with lookup
     * @param lrm A <code>LeaseRenewalManager</code> to pass to the
     *        <code>JoinManager</code>.  May be <code>null</code>
     */
    public JoinState(Object service, LeaseRenewalManager lrm) {
	this.service = service;
	this.lrm = lrm;
    }

    ////////////////////////////////////////////////
    // Methods needed to meet the SubStore interface

    // Inherit JavaDoc from super-type
    public String subDirectory() {
	return "JoinState";
    }

    // Inherit JavaDoc from super-type
    public void setDirectory(File dir) throws IOException {
	try {
	    log = new ReliableLog(dir.getCanonicalPath(), this);
	    log.recover();
	} catch (IOException e) {
	    System.err.println("JoinState:Problem recovering/creating log");
	    e.printStackTrace();
	    throw new IOException("JoinState: log " +
				  "is corrupted:" + e.getLocalizedMessage());
	}

	// Create LookupDiscoveryManager
	try {
	    ldm = new LookupDiscoveryManager(groups, locators, null);
	} catch (IOException e) {
	    System.err.println("JoinState:Problem starting discovery");
	    e.printStackTrace();
	    throw new IOException("JoinState: Problem starting discovery:" +
				  e.getLocalizedMessage());    
	}

	// Create JoinManager
	try {
	    if (serviceID == null)
		joinMgr = new JoinManager(service, attributes, this, ldm, lrm);
	    else
		joinMgr = new JoinManager(service, attributes, serviceID,
					  ldm, lrm);
	} catch (IOException e) {
	    System.err.println("JoinState:Problem starting JoinManager");
	    e.printStackTrace();
	    throw new IOException("JoinState: Problem starting JoinManager:" +
				  e.getLocalizedMessage());    
	}
    
	// For now we are treating the state of the
	// JoinManager/LookupDiscoveryManager as truth for our
	// log, now that we have a lookup manager, force it into
	// sync with our log
	try {
	    synchronized (log) {
		log.snapshot();
	    }
	} catch (IOException e) {
	    System.err.println("JoinState:Problem creating initial snapshot");
	    e.printStackTrace();
	    System.err.println("Ignoring and going on");
	}
    }

    // Inherit JavaDoc from super-type
    public void prepareDestroy() {
	try {
	    if (log != null)
		log.close();
	} catch (IOException e) {
	    System.err.println("JoinState:Problem closing log during destroy");
	    e.printStackTrace();
	    System.err.println("Ignoring and going on");
	}

	if (ldm != null)
	    ldm.terminate();

	if (joinMgr != null)
	    joinMgr.terminate();
    }

    //////////////////////////////////////////////////
    // Methods needed to meet the LogHandler interface
	
    // Inherit doc comment from super interface
    public void snapshot(OutputStream out) throws IOException {
	ObjectOutputStream oostream = new ObjectOutputStream(out);
	oostream.writeObject(serviceID);
	writeAttributes(joinMgr.getAttributes(), oostream);
	oostream.writeObject(ldm.getGroups());
	oostream.writeObject(ldm.getLocators());
	oostream.flush();
    }

    // Inherit doc comment from super interface
    public void recover(InputStream in) throws Exception {
	ObjectInputStream oistream = new ObjectInputStream(in);
	serviceID  = (ServiceID)oistream.readObject();
	attributes = readAttributes(oistream);
	groups     = (String[])oistream.readObject();
	locators   = (LookupLocator[])oistream.readObject();
    }
    
    /**
     * This method always throws
     * <code>UnsupportedOperationException</code> since
     * <code>FileJoinAdminState</code> should never update a
     * log.
     */
    public void applyUpdate(Object update) throws Exception {
	throw new UnsupportedOperationException(
                "JoinState:Recovering log update this should not happen");
    }

    /**
     * Utility method to write out an array of entities to an
     * <code>ObjectOutputStream</code>.  Can be recovered by a call
     * to <code>readAttributes()</code>
     * <p>
     * Packages each attribute in its own <code>MarshalledObject</code> so 
     * a bad codebase on an attribute class will not corrupt the whole array.
     */
    static private void writeAttributes(Entry[] attributes,
					ObjectOutputStream out) 
	throws IOException
    {
	// Need to package each attribute in its own marshaled object,
	// this makes sure that the attribute's code base is preserved
	// and when we unpack to discard attributes who's codebase
	// has been lost without throwing away those we can still deal with. 
	
	out.writeInt(attributes.length);
	for (int i=0; i<attributes.length; i++) 
	    out.writeObject(new MarshalledObject(attributes[i]));
    }

    /**
     * Utility method to read in an array of entities from a
     * <code>ObjectInputStream</code>.  Array should have been written 
     * by a call to <code>writeAttributes()</code>
     * <p>
     * Will try and recover as many attributes as possible.
     * Attributes which can't be recovered won't be returned but they
     * will remain in the log (though they will be lost when the next
     * snapshot is taken).
     */
    static private Entry[] readAttributes(ObjectInputStream in) 
	throws IOException, ClassNotFoundException
    {
	final List entries = new java.util.LinkedList();
	final int objectCount = in.readInt();
	for (int i=0; i<objectCount; i++) {
	    try {
		MarshalledObject mo = (MarshalledObject)in.readObject();
		entries.add(mo.get());
	    } catch (IOException e) {
		System.err.println("Exception while recovering attribute:" +
				   e.getMessage());
	    } catch (ClassNotFoundException e) {
		System.err.println("Exception while recovering attribute:" +
				   e.getMessage());
	    }
	}
	return (Entry[])entries.toArray(new Entry[0]);
    }

    /**
     * Used by all the methods that change persistent state to
     * commit the change to disk
     */
    private void takeSnapshot() throws IOException {
	synchronized (log) {
	    log.snapshot();
	}
    }

    // ************************************************************
    // Methods required by ServiceIDListener

    // Inherit doc comment from super interface
    public void serviceIDNotify(ServiceID serviceID) {
	this.serviceID = serviceID;
	
	try {
	    takeSnapshot();
	} catch (IOException e) {
	    System.err.println("JoinState:Can't persist service id");
	    e.printStackTrace();
	    System.err.println("...continuing...may not recover " +
			       "service id after restart");
	}
    }

    /////////////////////////////////////////////////////////////////
    // Effectively a clone of the JoinAdmin interface, used by
    // the client to implement JoinAdmin and to set its attributes

    /**
     * Get the list of groups to join.  An empty array means the service
     * joins no groups (as opposed to "all" groups).
     *
     * @return an array of groups to join. An empty array means the service
     *         joins no groups (as opposed to "all" groups).
     */
    public String[] getGroups() {
	return ldm.getGroups();
    }

    /**
     * Add new groups to the set to join.  Lookup services in the new
     * groups will be discovered and joined.
     *
     * @param groups groups to join
     */
    public void addGroups(String[] groups) {
 	try {
	    ldm.addGroups(groups);
	} catch (IOException e) {
	    throw new RuntimeException("Could not change groups:" + 
				       e.getLocalizedMessage());
	}

 	try {
	    takeSnapshot();
	} catch (IOException e) {
	    throw new RuntimeException("Could not log change:"+ 
				       e.getLocalizedMessage());
	}
    }

    /**
     * Remove groups from the set to join.  Leases are cancelled at lookup
     * services that are not members of any of the remaining groups.
     *
     * @param groups groups to leave
     */
    public void removeGroups(String[] groups) {
	ldm.removeGroups(groups);

	try {
	    takeSnapshot();
	} catch (IOException e) {
	    throw new RuntimeException("Could not log change:" +
				       e.getLocalizedMessage());
	}
    }

    /**
     * Replace the list of groups to join with a new list.  Leases are
     * cancelled at lookup services that are not members of any of the
     * new groups.  Lookup services in the new groups will be discovered
     * and joined.
     *
     * @param groups groups to join
     */
    public void setGroups(String[] groups) {
	try {
	    ldm.setGroups(groups);
	} catch (IOException e) {
	    throw new RuntimeException("Could not change groups:" +
				       e.getLocalizedMessage());
	}

	try {
	    takeSnapshot();
	} catch (IOException e) {
	    throw new RuntimeException("Could not log change:" +
				       e.getLocalizedMessage());
	}
    }

    /** 
     * Get the list of locators of specific lookup services to join. 
     *
     * @return the list of locators of specific lookup services to join
     */    
    public LookupLocator[] getLocators() {
	return ldm.getLocators();
    }

    /**
     * Add locators for specific new lookup services to join.  The new
     * lookup services will be discovered and joined.
     *
     * @param locators locators of specific lookup services to join
     */
    public void addLocators(LookupLocator[] locators) {
	ldm.addLocators(locators);
	try {
	    takeSnapshot();
	} catch (IOException e) {
	    throw new RuntimeException("Could not log change:" +
				       e.getLocalizedMessage());
	}
    }

    /**
     * Remove locators for specific lookup services from the set to join.
     * Any leases held at the lookup services are cancelled.
     *
     * @param locators locators of specific lookup services to leave
     */   
    public void removeLocators(LookupLocator[] locators) {
	ldm.removeLocators(locators);
	try {
	    takeSnapshot();
	} catch (IOException e) {
	    throw new RuntimeException("Could not log change:" +
				       e.getLocalizedMessage());
	}
    }

   /**
     * Replace the list of locators of specific lookup services to join
     * with a new list.  Leases are cancelled at lookup services that were
     * in the old list but are not in the new list.  Any new lookup services
     * will be discovered and joined.
     *
     * @param locators locators of specific lookup services to join
     */
    public void setLocators(LookupLocator[] locators) {
	ldm.setLocators(locators);
	try {
	    takeSnapshot();
	} catch (IOException e) {
	    throw new RuntimeException("Could not log change:" +
				       e.getLocalizedMessage());
	}
    }

    /** 
     * Get the current attribute sets for the service. 
     * 
     * @return the current attribute sets for the service
     * @throws java.rmi.RemoteException
     */
    public Entry[] getAttributes() {
	return joinMgr.getAttributes();
    }


    /** 
     * Add attribute sets for the service.  The resulting set will be used
     * for all future joins.  The attribute sets are also added to all 
     * currently-joined lookup services.
     *
     * @param attrSets the attribute sets to add
     * @param checkSC  <code>boolean<\code> flag indicating whether the
     *                 elements of the set of attributes to add should be
     *                 checked to determine if they are service controlled
     * @throws java.lang.SecurityException this exception occurs when
     *         the <code>checkSC<\code> parameter is <code>true<\code>, and
     *         at least one of the attributes to be added is an instance
     *         of the <code>ServiceControlled<\code> marker interface.
     */
    public void addAttributes(Entry[] attrSets, boolean checkSC) {
	joinMgr.addAttributes(attrSets, checkSC);
	try {
	    takeSnapshot();
	} catch (IOException e) {
	    throw new RuntimeException("Could not log change:" +
				       e.getLocalizedMessage());
	}
    }

    /**  
     * Modify the current attribute sets, using the same semantics as
     * ServiceRegistration.modifyAttributes.  The resulting set will be used
     * for all future joins.  The same modifications are also made to all 
     * currently-joined lookup services.
     *
     * @param attrSetTemplates the templates for matching attribute sets
     * @param attrSets the modifications to make to matching sets
     * @param checkSC  <code>boolean<\code> flag indicating whether the
     *                 elements of the set of attributes to add should be
     *                 checked to determine if they are service controlled
     * @throws java.lang.SecurityException this exception occurs when
     *         the <code>checkSC<\code> parameter is <code>true<\code>, and
     *         at least one of the attributes to be added is an instance
     *         of the <code>ServiceControlled<\code> marker interface.
         *     
     * @see net.jini.core.lookup.ServiceRegistration#modifyAttributes
     */
    public void modifyAttributes(Entry[] attrSetTemplates,
				 Entry[] attrSets,
				 boolean checkSC)
    {
	joinMgr.modifyAttributes(attrSetTemplates, attrSets, checkSC);

	try {
	    takeSnapshot();
	} catch (IOException e) {
	    throw new RuntimeException("Could not log change:" +
				       e.getLocalizedMessage());
	}
    }
}
