/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import net.jini.space.InternalSpaceException;

import java.util.Observer;
import java.util.Observable;
import java.util.Properties;
import java.util.Enumeration;
import java.util.ArrayList;
import java.util.List;
import java.util.Arrays;

import java.io.*;
import java.net.*;

import java.rmi.ConnectException;
import java.rmi.MarshalledObject;


/**
 * This class is used by a <code>FrontEndSpace</code> to control 
 * its <code>BackEndSpace</code>.  It starts the back end process,
 * coordinates its startup with the front end so they don't access the
 * database at the same time, and restarts the back end if it crashes.
 *
 * @see FrontEndSpace
 * @see FrontEndSpace#driver
 * @see FrontEndSpace#startBackEnd
 * @see BackEndSpace
 * @see BackEndSpace#main */
class BackEndDriver extends Thread implements Observer {
    /** The process object for the current back end. */
    private Process		proc;

    /** Flag denoting whether process termination is intentional. */
    private boolean		procDestroyed = false;

    /** Object used to control access to proc reference */
    private final Object	procLock = new Object();


    /** The name of the space. */
    private String		name;

    /** The path to the space's store. */
    private String		path;

    /** The input stream from the back end. */
    private BufferedReader	in;

    /** The output stream to the back end. */
    private PrintWriter		out;

    /**
     * Is the space ready for the front end to use (that is, is the
     * back end finished applying its logs?
     */
    private boolean		isReady = false;
    
    /**
     * True if it is safe for the back-end to access the database.
     *  This field is only changed from methods called by the front end.
     */
    private boolean             backendDBAccessOk = true;

    /**
     * The server socket for communication with the back end.  We use
     * a socket instead of using <code>System.in</code> and
     * <code>System.out</code> because rmid will close
     * <code>System.in</code> and, due to a JDK 1.2 bug (bugid
     * 4180429), this will cause the back end to sometimes not have a
     * valid <code>System.in</code>.  The network socket doesn't have
     * that problem.
     */
    private ServerSocket	sSock;

    private final static PrintWriter beDbg =
	BasicSpace.debug.getWriter("backEnd");

    /**
     * Create a <code>BackEndDriver</code> for the given name and path.
     */
    BackEndDriver(String name, String path) {
	this.name = name;
	this.path = path;
    }

    /**
     * Kill the back-end process associated with the last
     * incarnation of this <code>FrontEndSpace</code> if it is still 
     * around
     */
    private void killExisting() {
	FileInputStream inStream = null;
	ObjectInputStream oinStream = null;

	final File file = new File(path, 
				   BackEndSpace.BACK_END_CONTROL_FILE_NAME);

	// If there is no such file there is no active back-end
	if (!file.exists())
	    return;

	try {
	    // Look for BackEndControl object
	    inStream = new FileInputStream(file);
	    oinStream = new ObjectInputStream(inStream);

	    final MarshalledObject mo = 
		(MarshalledObject)oinStream.readObject();
	    final BackEndControl bec = (BackEndControl)mo.get();

	    // Now kill the old back-end
	    bec.finishAndTerminate();
	} catch (ConnectException e) {
	    // This proably means there was an old back-end process but now
	    // it is gone, ignore
	} catch (IOException e) {
	    // This proably should not happen, send a note to err and
	    // press on
	    System.err.println("IOException while killing old back end " + 
	       "process:" + e.getLocalizedMessage() + "...continuing");
	} catch (ClassNotFoundException e) {
	    // This is odd since it should be in our classpath send a
	    // note to err and press on
	    System.err.println("ClassNotFoundException while killing old " +
	       "back end proces:" + e.getLocalizedMessage() + "...continuing");
	} finally {
	    // Close the object stream if we opened it, this will close
	    // the file stream too, if we never opened the object stream
	    // close the file stream explicitly
	    try {
		if (oinStream != null) 
		    oinStream.close();
		else if (inStream != null)
		    inStream.close();
	    } catch (IOException e) {
		// ignore
	    }
	}
    }

    /**
     * Loop starting up the back end.  The first time the back end is
     * started is different -- at that point the front end isn't
     * producing any logs, and so all logs should be consumed by the
     * back end.  If the back end goes down, its restart should not
     * consume the most recent file since the (now active) front end
     * may still be writing to it.
     */
    public void run() {
	Runtime runtime = Runtime.getRuntime();

	try {
	    sSock = new ServerSocket(0);
	} catch (IOException e) {
	    e.printStackTrace();
	    System.err.println("cannot start the back end");
	    return;
	}

	// -a means "read all logs"
	String[] command = buildCommand(new String[] { "-a" } );
	boolean first = true;		// first time starting things

	// Kill the back-end process associated with the last
	// incarnation of this FrontEndSpace if it is still around
	killExisting();

	try {
            procLoop:
	    while (!procDestroyed) {

		isReady = false;

		// Now that we have locked out the front end from accessing the 
		// database, wait for the backendDBAccessOk to become true.
		synchronized (this) {
		    while (!backendDBAccessOk)
			wait();

		    try {
                        synchronized(procLock) {
	                    if (!procDestroyed) { 
		                if (proc != null) // if alive, kill it
		                    destroyBackEnd(false);       
			        proc = runtime.exec(command);	// start process
                            }
                            else  // If we are terminating, then don't bother
                                break procLoop;
                        }
		    } catch (IOException e) {
			e.printStackTrace();
			throw new InternalSpaceException(
			    "can't run back end", e);
		    }
		}	    

		try {
		    connectStreams();		// connect streams together

		    // get the first MSG line
		    String line;
		    while ((line = in.readLine()) != null) {
			if (line.startsWith(BackEndSpace.MSG))
			    break;
		    }

		    // it better be READY_MSG
		    if (line != null && line.equals(BackEndSpace.READY_MSG)) {
			if (beDbg != null) {
			    beDbg.println("BackEndDriver: got "
				+ BackEndSpace.READY_MSG);
			}
			synchronized (this) {
			    isReady = true;
			    notifyAll();
			}
		    } else {
			throw new InternalSpaceException("Back end space said "
			    + (line == null ? "EOF" : line)
			    + ", not " + BackEndSpace.READY_MSG);
		    }
		    // Kick off thread to consume data coming in on in
		    // so backend does not block if it prints to System.out
		    connectToBitBucket(in);
		
		    // now wait for the process to die (it really never should)
		    int status = proc.waitFor();
                    synchronized(procLock) {
                        proc = null;
                    }

		    out = null;
		    in = null;
		    if (beDbg != null)
			beDbg.println("back end exited with status " + status);
		} catch (IOException e) {
		    if (beDbg != null)
			e.printStackTrace(beDbg);
		}

		// only the first iteration uses -a: otherwise no extra options
		if (first) {
		    command = buildCommand(null);
		    first = false;
		}
	    }
	} catch (InterruptedException e) {
	    return;	// we've been asked to go away
	}
    }

    /**
     * Build up a command array for the back end, adding in the given options
     * as parameters to the back end (not the VM).
     */
    private String[] buildCommand(String[] options) {
	List cmdline = new ArrayList();		// the command line as its built

	// Which JVM sould we use and were can we find it
	String javaProg = System.getProperty("com.sun.jini.outrigger.java");
	if (javaProg == null) {
	    // Go with the default
	    String javaHome = null;

	    try {
		javaHome = System.getProperty("java.home");
	    } catch (Exception e) {
		System.err.println("Received the following exception " +
				   "when trying to read java.home property");
		e.printStackTrace();
		System.err.println("Using default of `java' to spawn JVM for " +
				   "backend");
	    }

	    if (javaHome != null) {
		javaProg = javaHome + File.separator + "bin"
		    + File.separator + "java";
	    } else {
		javaProg = "java";
	    }
	}

	cmdline.add(javaProg);
	cmdline.add("-Xbootclasspath:"
		    + System.getProperty("sun.boot.class.path"));

	/*
	 * Add in the net.jini properties (most likely set on the
	 * command line).  We also add in java.compiler since it should
	 * be the same in front and back ends.
	 */
	Properties props = System.getProperties();
	Enumeration keys = props.keys();
	while (keys.hasMoreElements()) {
	    String key = (String) keys.nextElement();
	    if (key.startsWith("net.jini.") || 
		key.startsWith("com.sun.jini.") || 
		key.equals("java.compiler") || key.equals("java.class.path"))
		cmdline.add("-D" + key + "=" + props.get(key));
	}

	// Option to add to back-end VM
	String extraOptions = 
	    System.getProperty("com.sun.jini.outrigger.backend.vm.options");
	if (extraOptions != null) 
	    cmdline.add(extraOptions);

	// the class name whose main to use
	cmdline.add(BackEndSpace.CLASS_NAME);

	// add in any options
	if (options != null)
	    cmdline.addAll(Arrays.asList(options));

	// add in the operands

	// The BackEnd is always on our host (for now) 
	cmdline.add("127.0.0.1");
	cmdline.add(String.valueOf(sSock.getLocalPort()));
	cmdline.add(name);
	cmdline.add(path);

	// now convert it into an array for exec
	String[] cmd = new String[cmdline.size()];
	System.arraycopy(cmdline.toArray(), 0, cmd, 0, cmd.length);
	if (beDbg != null) {
	    beDbg.print("command line:");
	    for (int i = 0; i < cmd.length; i++) {
		beDbg.print(" ");
		beDbg.print(cmd[i]);
	    }
	    beDbg.println();
	}

	return cmd;
    }

    /**
     * Connect the streams of the back end child process to the
     * streams we use to communicate with them.
     */
    private void connectStreams() throws IOException {
	if (beDbg == null) {
	    // no debug output, just connect its output to my stream and its
	    // error to my error
	    connect(new InputStreamReader(proc.getErrorStream()),
		new Writer[] { new PrintWriter(System.err, true) } );
	    in = new BufferedReader(
			    new InputStreamReader(proc.getInputStream()));
	} else {
	    // connect its output to both logging and my stream and its error
	    // to both logging and my error
	    beDbg.println("BackEndDriver: Connecting streams");
	    connect(new InputStreamReader(proc.getErrorStream()),
		new Writer[] { beDbg } );
	    PipedReader pin = new PipedReader();
	    in = new BufferedReader(pin);
	    Writer pout;
	    try {
		pout = new PrintWriter(new PipedWriter(pin), true);
	    } catch (IOException e) {			// should never happen
		throw new InternalSpaceException("PipedReader closed?", e);
	    }
	    connect(new InputStreamReader(proc.getInputStream()),
		new Writer[] { pout, beDbg } );
	}

	// We do this after connecting the stream so we will still 
	// get debugging output even if the BackEnd never connects
	Socket sock = sSock.accept();
	out = new PrintWriter(sock.getOutputStream(), true);
    }

    /**
     * Set up a thread that consumes the input of <code>in</code> and
     * writes it to the stream <code>out</code>.
     */
    private void connect(final Reader in, final Writer[] out) {
	new Thread() {
	    public void run() {
		try {
		    char[] buf = new char[512];
		    int len;
		    while ((len = in.read(buf)) != -1) {
			for (int i = 0; i < out.length; i++) {
			    out[i].write(buf, 0, len);
			    out[i].flush();
			}
		    }
		} catch (IOException e) {
		    return;
		}
	    }
	}.start();
    }

    /**
     * Set up a thread that consumes the input of <code>in</code> and
     * throws it away.
     */
    private void connectToBitBucket(final Reader in) {
	new Thread() {
	    public void run() {
		try {
		    char[] buf = new char[512];
		    while (in.read(buf) != -1) {
			// Throw away the data
		    }
		} catch (IOException e) {
		    return;
		}
	    }
	}.start();
    }

    /**
     * Wait until the back-end is done updating the database and lockout 
     * back-end
     */
    synchronized void waitUntilReady() {
	while (!isReady) {
	    try {
		if (beDbg != null)
		    beDbg.println("BackEndDriver: waitUntilReady waiting");
		wait();
	    } catch (InterruptedException e) {
		continue;	// try again
	    }
	}

	// Not safe for the back-end to access the db
	backendDBAccessOk = false;

	if (beDbg != null)
	    beDbg.println("BackEndDriver: waitUntilReady returns");
    }

    /**
     * Tell the <code>BackEndSpace</code> that it is safe for it to 
     * access the database again
     */
    synchronized void dbAccessFinished() {
	backendDBAccessOk = true;
	notifyAll();
    }

    /** Kill the back-end process.
     *  This method is called (externally) from
     *  <code>FrontEndSpace</code> and is then forwarded 
     *  to the internal version.
     */
    void destroyBackEnd() throws InterruptedException {
        destroyBackEnd(true);
    }

    /** 
     *  Kill the back-end process. 
     *  This method is called (internally) from 
     *  <code>destroyBackEnd()</code> and <code>run()</code>.    
     *  <code>destroyBackEnd()</code> calls it with true, meaning 
     *  that we are terminating the final time. <code>run()</code> 
     *  calls it with false, meaning that the process is being  
     *  cleaned up before another run.
     */
    private void destroyBackEnd(boolean terminating) 
            throws InterruptedException 
    {
        synchronized(procLock) {
            if(terminating)
                procDestroyed = true;
            if(proc == null)
                return;
	    proc.destroy();
	    proc.waitFor();
        }
    }

    /**
     * This is our <code>Observer</code> code: when we observe a change
     * of log files from the <code>LogOutputStream</code> object we
     * tell the back end to check for new logs.
     */
    public synchronized void update(Observable source, Object arg) {
	if (out != null)
	    out.println(BackEndSpace.LOG_COMPLETE_MSG + " " + arg);
    }
}
