/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import net.jini.space.*;
import net.jini.core.lease.*;
import net.jini.core.transaction.*;
import net.jini.core.transaction.server.*;
import net.jini.core.event.EventRegistration;
import com.sun.jini.mahout.StorableObject;
import com.sun.jini.mahout.binder.Binder;
import com.sun.jini.lease.landlord.*;
import com.sun.jini.system.*;

import java.rmi.Remote;
import java.rmi.MarshalledObject;
import java.rmi.RemoteException;
import java.rmi.NoSuchObjectException;
import java.rmi.server.UnicastRemoteObject;
import java.io.*;
import java.net.*;
import java.util.*;

import COM.odi.ObjectStore;
import COM.odi.util.OSHashtable;
import COM.odi.util.OSVector;

/**
 * This is the space that is the back end of the two-server model.
 * It sits between the front end and the database consuming logs 
 * when they are complete, or during activation.
 *
 * @see BackEndDriver
 * @see FrontEndSpace
 */
public class BackEndSpace extends BasicSpace implements BackEndControl {

    /**
     * This class extends <code>ODIStore</code> to add roots that the
     * back end needs.
     */
    class BackEndStore extends ODIStore {
	BackEndStore(String name, String path) {
	    super(BackEndSpace.this, name, path, ObjectStore.RETAIN_HOLLOW);
	}

	void getFields() {
	    super.getFields();
	    if (pendingTxns == null) {
		pendingTxns = (OSHashtable)
		    getRoot(PENDING_TXNS_ROOT, hashFactory);
	    }
	}

	void resetFields() {
	    super.resetFields();
	    pendingTxns = null;
	}
    }

    /**
     * Object to process the operations from a log file.
     *
     * @see LogProcessor
     */
    private transient LogProcessor	processor = new LogProcessor();

    /**
     * The pending transactions (if any) between log files.  Operations
     * performed under a transaction in one log file for which the
     * commit or abort is in another log file are stored in this table
     * between log files.
     */
    private transient OSHashtable	pendingTxns = null;

    /**
     * The interval (in numbers of log files consumed) between runs of the
     * GC on the database.
     */
     private transient int gcInterval =
	    Integer.getInteger("com.sun.jini.outrigger.backend.gcInterval",
						DEFAULT_GC_INTERVAL).intValue();
    /**
     * The count of logs we've read.
     */
    private transient int		logcount;

    /**
     * Our back end store
     */
    private transient BackEndStore	store;

    /**
     * The base name for the log files.
     */
    private transient String		logFileBase;

    /**
     * The directory our log lives in
     */
    private transient String            logDirectory;

    /**
     * The factory that is used to create (if necessary) a hash table
     * for the uncompleted transactions list.
     */
    private static ObjectFactory	hashFactory = new ObjectFactory() {
	public Object newObject() {
	    return new OSHashtable();
	}
    };

    /** The header that marks messages between the back and front ends. */
    static final String		MSG = "-- ";
    /** Message that says "the database is ready for the front end to use." */
    static final String		READY_MSG = MSG + "DB ready";
    /** Message that says "the front end has completed a log." */
    static final String		LOG_COMPLETE_MSG = MSG + "log completed: ";

    /** The name of the root that stores pending transactions. */
    static final String			PENDING_TXNS_ROOT = "PendingTxns";

    /** This class's name. */
    static final String		CLASS_NAME = BackEndSpace.class.getName();

    /** Name of the file we place the BackEndControl object in */
    static final String  BACK_END_CONTROL_FILE_NAME = "BackEndControlObject";

    /** The default interval (in number of logs) between GC'ing the database. */
    private static final int DEFAULT_GC_INTERVAL = 5;

    /** The name of the root that stores the name of the last log file read. */
    private static final String		LAST_LOG_ROOT = "LastLog";

    private static final PrintWriter	logDebug = LogFile.logDebug;
    private static final PrintWriter	backDebug = debug.getWriter("backEnd");

    /**
     * Thread that will kill the back-end process if necessary 
     * @serial
     */
    private final Terminator terminatorThread = new Terminator();

    /**
     * This is the class that receives the messages from the
     * LogInputFile.  We need a separate class because we would get
     * into infinite recursion if <code>BackEndSpace</code> overrode
     * the methods it inherits from <code>BasicSpace</code>: if
     * <code>LogInputFile</code> would invoke (say) <code>writeOp</code>,
     * our implementation would need to invoke the superclass's
     * implementation to actually apply the write, and the superclass's
     * implementation would invoke <code>writeOp</code> to log the
     * write, getting back into our code again.
     */
    private class LogProcessor implements LogOps {
	// inherit doc comment
	public void bootOp(long time) {
	    if (logDebug != null)
		logDebug.println("bootOp(" + new Date(time) + ")");
	    cleanUpAfterReboot();
	}

	// inherit doc comment
	public void writeOp(EntryRep entry, Long trID) {
	    if (logDebug != null)
		logDebug.println("writeOp(" + entry + ", " + trID + ")");
	    if (trID != null) {
		pendingTxn(trID).addWrite(entry);
	    } else {
		try {
		    write(entry, null, 0);
		} catch (RemoteException e) {
		    // I say this because when we're done it will be true.
		    System.out.println("RemoteException, but tr is null!");
		    e.printStackTrace();
		} catch (TransactionException e) {
		    System.out.println("TransactionException, but tr is null!");
		    e.printStackTrace();
		    System.exit(-2);
		}
	    }
	}

	// inherit doc comment
	public void takeOp(long id, Long trID) {
	    if (logDebug != null)
		logDebug.println("takeOp(" + id + ", " + trID + ")");
	    if (trID != null)
		pendingTxn(trID).addTake(id);
	    else
		contents.cancel(new Long(id));
	}

	// inherit doc comment
	public void registerOp(EntryRep tmpl, NotifyChit chit) {
	    if (logDebug != null)
		logDebug.println("notifyOp(" + tmpl + ", " + chit + ")");
	    register(tmpl, chit);
	}

	// inherit doc comment
	public void notifiedOp(long chitID, long newSeq) {
	    // If we don't know about this chit, it is proably an event 
	    // registration under a transaction
	    if (null == templates.getLeasedResource(new Long(chitID)))
		return;
	    if (logDebug != null)
		logDebug.println("notifiedOp(" + chitID + ", " + newSeq + ")");
	    changeSeq(chitID, newSeq);
	}

	// inherit doc comment
	public void renewOp(Long cookie, long expiration) {
	    if (logDebug != null)
		logDebug.println("renewOp(" + cookie + ", " + expiration + ")");
	    try {
		renew(cookie, expiration);
	    } catch (LeaseDeniedException e) {
		throw new InternalSpaceException("renewing lease", e);
	    } catch (UnknownLeaseException e) {
		// May have been an event registration under a transaction 
	    }
	}

	// inherit doc comment
	public void cancelOp(Long cookie) {
	    if (logDebug != null)
		logDebug.println("cancelOp(" + cookie + ")");
	    try {
		cancel(cookie);
	    } catch (UnknownLeaseException e) {
		;	// okay -- might be transaction
	    }
	    removePendingTxn(cookie);
	}

	// inherit doc comment
	public void prepareOp(Long trID, StorableObject tr) {
	    if (logDebug != null)
		logDebug.println("prepareOp(" + tr + ")");
	    PendingTxn pt = pendingTxn(trID);
	    pt.transaction(tr);
	    pt.state(PREPARED);
	}

	// inherit doc comment
	public void commitOp(Long trID) {
	    if (logDebug != null)
		logDebug.println("commitOp(" + trID + ")");
	    PendingTxn pt = pendingTxn(trID);
	    pt.commit(processor);
	    removePendingTxn(trID);
	}

	/**
	 * Calculate when it is appropriate to perform persistent
	 * garbage collection.
	 */
	private boolean doStoreGC() {
	    if ((++logcount % gcInterval)== 0)
		return true;

	    return false;
	}

	/**
	 * Consume the log files that exist.  If <code>all</code> is
	 * <code>true</code>, all found log files will be processed.
	 * If <code>log</code> is <code>false</code>, then all but the
	 * most recent will be processed; this will prevent the back
	 * end from reading the log file that is currently being
	 * produced by the front end.
	 */
	private void consumeLogs(boolean all) {
	    Iterator it;
	    try {
		it = LogInputFile.logs(logFileBase, all);
	    } catch (IOException e) {
		throw new InternalSpaceException("couldn't open logs");
	    }
	    while (it.hasNext()) {
		LogInputFile log = (LogInputFile) it.next();
		if (logDebug != null) {
		    logDebug.println();
		    logDebug.println(log + ":");
		}
		if (log == null)		// file already consumed
		    continue;
		try {
		    startTxn();
		    LastLog lg = (LastLog) store.getRoot(LAST_LOG_ROOT);
		    String logFile = log.toString();
		    if (lg == null || !lg.sameAs(logFile))
		        log.consume(this);
		    store.setRoot(LAST_LOG_ROOT, new LastLog(logFile));
		} finally {
		    commitTxn();
		}
		log.finished();	// must happen after commit

		if (logDebug != null)
		    logDebug.println("");
		if (doStoreGC()) {
		    BasicSpace.dumpMemory(dbgMemory, 
			"BackEndSpace::consumeLogs() Before Reap & GC");

		    // Open a transaction so we can reap dead entries 
		    // and event templates
		    try {
			startTxn();

			if (!disableReaping) {
			    if(dbgMemory != null)
				dbgMemory.println("BackEndSpace::consumeLogs() - "
				+ "Reaping nodes ...");
			    contents.reap();
			    templates.reap();
                        }
			else {
			    if(dbgMemory != null)
				dbgMemory.println("BackEndSpace::consumeLogs() - "
				+ "NOT Reaping nodes ...");
			}

		    } finally {
			commitTxn();
		    }	

		    store.gc();

		    doDebugGC(dbgMemory);

		    BasicSpace.dumpMemory(dbgMemory, 
		        "BackEndSpace::consumeLogs() After Reap & GC");
		}
	    }
	}
    }

    /**
     * This class remembers which log file was the last to be
     * successfully consumed.  If the recovery mechanism reopens this
     * file, then it will skip its contents -- this indicates a crash
     * happened after the contents were committed to the DB but before
     * the file was unlinked.  
     */
    private static class LastLog {
	private String	logFile;
	private long	timeStamp;

	LastLog(String path) {
	    logFile = path;
	    timeStamp = new File(logFile).lastModified();
	}

	boolean sameAs(String otherPath) {
	    if (!logFile.equals(otherPath))
		return false;
	    return (new File(otherPath).lastModified() == timeStamp);
	}
    }

    /**
     * This object is used to handle lease issues in the back end,
     * which are different from those in the front end.  Leases have
     * already been allocated by the front end -- we are only recording
     * those decisions.
     */
    private static class BackEndLeasePolicy implements LeasePolicy {
	/**
	 * Return <code>null</code>.  The front end has already defined
	 * the expiration on the object in question, so this method is
	 * invoked by the implementation at conventional points, but
	 * there is nothing to do.
	 */
	public Lease leaseFor(LeasedResource res, long duration) {
	    return null;
	}

	/**
	 * In this case, <code>duration</code> is really
	 * <code>expiration</code>, which has been pre-calculated.  We
	 * do this because it allows us to leverage the logic in
	 * <code>BasicSpace.renew</code> that finds the correct
	 * resource to renew by passing the expiration as the
	 * duration.
	 *
	 * @see cancel
	 */
	public long renew(LeasedResource res, long expires) {
	    res.setExpiration(expires);
	    return expires;
	}

	// inherit doc comment
	public boolean ensureCurrent(LeasedResource res) {
	    // We always return true because the back-end should NEVER
	    // reap a resource unless the front end tells it to do
	    // so.  ensureCurrent is used by the renew logic to reject
	    // renewals on already expired resources, however, in the 
	    // case of the back-end it may just be that it did not get the
	    // renewal notice until after the expiration passed, the
	    // renew must still happen, because the client thinks it did.
	    return true;
	}
    }

    /**
     * Create a new <code>BackEndSpace</code> with the given
     * <code>name</code> and <code>path</code>.
     */
    public BackEndSpace(String name, String path) throws IOException {
	logDirectory = path;
	store = new BackEndStore(name, path);
	logFileBase = (new File(path, name)).getAbsolutePath();
	initialize(store);

	// The policy has no state so we can use the same one for both.
	entryLeasePolicy = new BackEndLeasePolicy();
	eventLeasePolicy = entryLeasePolicy; 

	setupBackEndControl();
    }

    /**
     * Overridden <code>BasicSpace</code> method that throws
     * <code>UnsupportedOperationException</code> 
     * since the <code>BackEndSpace</code> doesn't advertise its
     * availability.
     */
    Binder createBinder()
    {
        throw new UnsupportedOperationException(
	    "BackEndSpace::createBinder is unsupported");
    }


    /**
     * Always return false, the only time the back-end should mark
     * something as expired is when the front end tells it to.
     */
    boolean expired(LeasedResource resource, long now) {
	return false;
    }

    /**
     * Overridden to do nothing -- the back end leaves all the runtime stuff
     * up to the front end.
     */
    protected void startRuntime() {
    }

    /**
     * Return the pending transaction description for the given
     * transaction, creating the object and adding it to the table if
     * necessary.
     */
    private PendingTxn pendingTxn(Long trID) {
	PendingTxn pt = (PendingTxn) pendingTxns.get(trID);
	if (pt == null) {
	    pt = new PendingTxn(trID);
	    pendingTxns.put(trID, pt);
	}
	return pt;
    }

    /**
     * Remove a pending transaction from the table.  If it isn't there,
     * this call is harmless.
     */
    private void removePendingTxn(Long trID) {
	pendingTxns.remove(trID); // if it fails, it wasn't there to remove
    }

    /**
     * Clean up after a reboot of the system, aborting all transactions
     * that were still active at the crash.
     */
    private void cleanUpAfterReboot() {
	// get a list of pending transactions that have been aborted by a boot
	Enumeration e = pendingTxns.elements();
	Vector remove = new Vector();			// things to remove
	while (e.hasMoreElements()) {
	    PendingTxn pt = (PendingTxn) e.nextElement();
	    if (pt.state() == COMMITTED) {
		System.err.println("A COMMITTED transaction in pending???");
		System.err.println("\t" + pt);
	    } else if (pt.state() != PREPARED) {
		remove.addElement(pt.id());
	    }
	}

	// Now remove them.  We've done this in two passes to ensure that
	// OSHashtable's enumeration won't fail on removing during iteration
	e = remove.elements();
	while (e.hasMoreElements())
	    pendingTxns.remove(e.nextElement());
    }

    //******************************************************************
    // Methods and classes associated with implementing BackEndControl

    /**
     * Do the necessary work to setup BackEndControl implementation
     */
    private void setupBackEndControl() throws IOException {
	terminatorThread.start();

	// Write BackEndControl object into log
	FileOutputStream ostream = null;
	ObjectOutputStream oostream = null;
	try {
	    final File file = new File(logDirectory, 
				       BACK_END_CONTROL_FILE_NAME);
	    final MarshalledObject us = new MarshalledObject(this);
	    ostream = new FileOutputStream(file);
	    oostream = new ObjectOutputStream(ostream);
	    oostream.writeObject(us);
	} finally {
	    // Close the object stream if we opened it, this will close
	    // the file stream too, if we never opened the object stream
	    // close the file stream explicitly
	    try {
		if (oostream != null) 
		    oostream.close();
		else if (ostream != null)
		    ostream.close();
	    } catch (IOException e) {
		// Swallow the exception, if the main try block threw
		// an Exception we don't want to mask it, if the only
		// problem was the close we don't want to kill the 
		// whole object
	    }
	}
    }

    // Inherit doc comment
    public void finishAndTerminate() {
	// Tell the terminator thread to close the database
	terminatorThread.stopProcessing();

	// Wait for the database to close
	terminatorThread.waitOnDBClose();
    }


    /**
     * Latch class used internally by the terminator thread.  Calls of
     * <code>waitOnSet</code> will block until <code>set</code> is
     * called.  Once <code>set</code> is called current in any active
     * calls to <code>waitOnSet</code> will return and future calls to
     * <code>waitOnSet</code> will return immediately
     */
    static private class Latch {
	private boolean set = false;

	/**
	 * Set the latch, this will let <code>waitForSet()</code> to return
	 */
	synchronized void set() {
	    set = true;
	    notifyAll();
	}

	/**
	 * Wait until some one calls <code>set()</code>
	 */
	void waitOnSet() {
	    while (!set) {
		synchronized (this) {
		    try {
			wait();
		    } catch (InterruptedException e) {
			// Just going to retry if set is not true
			// and if it is we should return
		    }
		} 
	    }
	}
    }

    /**
     * Thread that blocks until someone calls
     * <code>stopProcessing()</code>, and then closes the database,
     * un-exports the <code>BackEndControl</code>, and forces the
     * process to exit.  Will grab and hold <code>dbAccessLock</code>
     * before closing the database. Once the database is closed
     * <code>waitOnDBClose()</code> will return.  Will not force the
     * process to exit before all remote calls have completed.
     */
    private class Terminator extends Thread { 
	// Fields used to coordinate the actions of the main loop, the
	// finishAndTerminate() method, and the the termination thread

	/**
	 * Terminate thread blocks on this flag, until
	 * <code>stopProcessing()</code> is called.
	 */
	final private Latch stopProcessing = new Latch();
    
	/**
	 * Flag used to coordinate closing of the database with the 
	 * return of the <code>waitOnDBClose()</code> calls.
	 */
	final private Latch dbClosed = new Latch();

	/** 
	 * Must be held to open or access database, grabbed by terminate
	 * thread to prevent database from being re-open once closed.
	 */
	final Object dbAccessLock = new Object();

	/**
	 * Create a new Terminator thread
	 */
	private Terminator() {
	    super("Back-end Terminator");
	    this.setDaemon(true);	// thread shouldn't keep the VM alive
	}

	/**
	 * Call this method to terminate log processing and start the
	 * process of killing the back-end
	 */
	void stopProcessing() {
	    stopProcessing.set();
	}

	/**
	 * Wait for the database being permanently close
	 */
	void waitOnDBClose() {
	    dbClosed.waitOnSet();
	}

	/**
	 * Wait around until someone sets stopProcessing and then close
	 * the database and exit the process
	 */
	public void run() {
	    // Wait until we need to stop processing
	    stopProcessing.waitOnSet();

	    // Keep anyone from opening database
	    synchronized (dbAccessLock) {

		store.close(false);

		// Let finishAndTerminate() return
		dbClosed.set();

		// Wait for finishAndTerminate() to return
		try {
		    while (!UnicastRemoteObject.unexportObject(
		        BackEndSpace.this, false));
		} catch (NoSuchObjectException e) {
		    // If the object is not exported the call must be done
		}
		// Now that the caller has gotten the message we can return
		System.exit(-3);
	    }	    
	}
    }

    /**
     * Main program, intended to be used by the front end to start the back end up.
     * Usage:
     * <pre>
     *     java ... com.sun.jini.outrigger.BackEndSpace [-a] name path
     * </pre>
     * The <code>-a</code> option is specified when the front end is
     * first starting to tell the back end to consume all the logs it
     * can find.  Subsequent restarts (after the front end is up and
     * running) don't specify <code>-a</code> so the back end will skip
     * the most recent log, since the front end may still be writing to
     * it.
     *
     * @see BackEndDriver
     */
    public static void main(String[] args) {
	BackEndSpace space = null;

	try {
	    // parse the command line
	    POSIXCommandLine line = new POSIXCommandLine(CLASS_NAME, args);
	    boolean startWithAll = line.getBoolean('a');
	    String[] operands = line.getOperands();
	    String host = operands[0];
	    String port = operands[1];
	    String name = operands[2];
	    String path = operands[3];

	    if (backDebug != null) {
		backDebug.println("BackEndSpace.main"
		    + ", startWithAll = " + startWithAll
		    + ", host = " + host + ", port = " + port
		    + ", name = " + name + ", path = " + path);
	    }

	    // create the space
	    space = new BackEndSpace(name, path);

	    synchronized (space.terminatorThread.dbAccessLock) {
		space.processor.consumeLogs(startWithAll);
		space.store.closeAndReset(false); // let the front end at it
	    }

	    System.out.println(READY_MSG);	// tell the driver we're done

	    doDebugGC(dbgMemory);
            BasicSpace.dumpMemory(dbgMemory, "BackEndSpace::main() Inital Memory");

	    Socket inSock = new Socket(host, Integer.parseInt(port));
	    // input from the front end
	    BufferedReader in = new BufferedReader(
		new InputStreamReader(inSock.getInputStream()));
	    for (;;) {
		String str = in.readLine();	// get a message
		if (backDebug != null)
		    backDebug.println("input line: " + str);
		if (str == null) {		// the parent died, so should we
		    space.finishAndTerminate();
		    break;
		} else if (str.startsWith(LOG_COMPLETE_MSG)) {
		    synchronized (space.terminatorThread.dbAccessLock) {
			space.store.openDB(ObjectStore.OPEN_UPDATE);
			space.processor.consumeLogs(false);// leave latest alone
		    }
		} else {
		    System.err.println("Unexpected BackEndSpace input: " + str);
		}
	    }
	} catch (IOException e) {
	    if (backDebug != null)
		e.printStackTrace(backDebug);
	    else
		e.printStackTrace();

	    if (space != null) 
		space.finishAndTerminate();
	    else
		System.exit(-3);
	} catch (CommandLine.BadInvocationException e) {
	    System.exit(-1);
	} catch (CommandLine.HelpOnlyException e) {
	    System.exit(0);
	} catch (Throwable t) {
	    // force exit even if other non-daemon threads exist
	    try {
		if (backDebug != null) {
		    t.printStackTrace(backDebug);
		    backDebug.flush();
		} else {
		    t.printStackTrace();
		    System.err.flush();
		}
	    } finally {
		System.exit(1);
	    }
	}
    }
}
