/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import net.jini.space.*;
import net.jini.core.entry.*;
import net.jini.core.transaction.*;
import net.jini.core.transaction.server.*;
import net.jini.core.event.*;
import net.jini.core.lease.*;
import com.sun.jini.mahout.*;
import com.sun.jini.lease.landlord.*;
import com.sun.jini.thread.*;
import com.sun.jini.debug.*;
import com.sun.jini.constants.*;

import java.util.*;
import java.lang.reflect.*;
import java.io.*;
import java.security.SecureRandom;

import java.rmi.*;
import java.rmi.activation.*;
import java.rmi.server.*;
import java.rmi.registry.*;

import java.net.URL;

import COM.odi.util.OSVector;
import COM.odi.util.OSHashtable;

import com.sun.jini.outrigger.TemplateHolderSet.AddRecord;

import com.sun.jini.mahout.binder.Binder;
import com.sun.jini.mahout.binder.DirectoryAdmin;

import net.jini.core.discovery.LookupLocator;
import net.jini.lookup.entry.ServiceInfo;
import net.jini.lookup.entry.ServiceType;


/**
 * A basic implementation of a JavaSpaces<sup><font size=-2>TM</font></sup> 
 * service. This abstract class is designed for use by both transient and 
 * persistent implementations, which subclass <code>BasicSpace</code> to
 * provide the storage-related calls.
 * <p>
 * Storage requires setup, transactions, and transient copies.  The
 * <code>setupStore</code> abstract method lets the subclass do
 * whatever it needs to set up its storage state.  The methods
 * <code>startTxn</code> and <code>endTxn</code> start and end
 * transactions whenever the code touches variables that are not
 * <code>transient</code>.  The method <code>transientOf</code> returns
 * a transient version of whatever object it was passed.  This is used
 * when the space needs a reference to a copy that will be valid after
 * a transaction ends, such as to be able to return it to the client.
 * <p>
 * <code>BasicSpace</code> maintains a list of types of entries it has
 * seen and their subtypes using a <code>TypeTree</code> object.
 * Each type of entry has an <code>EntryHolder</code> that is
 * stored in the <code>EntryHolderSet</code> object named
 * <code>contents</code>.
 * <p>
 * On <code>write</code>, the written entry's class and superclass are
 * added to the known <code>types</code>, and its <code>EntryRep</code>
 * is added to the space's <code>contents</code>.
 * <p>
 * On <code>read</code>, the <code>find</code> method searches through
 * the entries of its type and subtypes, asking each entry holder if it
 * has an entry that matches the template.  If a match is found, the
 * matching <code>EntryRep</code> is returned.  If none of the
 * appropriate holders has a match, it will return <code>null</code>.
 * <p>
 * On <code>take</code> we also use <code>find</code> with a
 * <code>boolean</code> that says to remove the entry that matches.
 * <p>
 * Notification requires a separate <code>Notifier</code> queue and
 * thread.  When an entry is written, a reference to it is added to the
 * queue of "unexamined entries".  The notifier thread pulls entries
 * off the queue and checks them against registered notification
 * templates.  When it has found all matches for the template, the
 * <code>Notifier</code> thread adds the notifications for this write
 * to its list of undelivered notifications, which it periodically
 * attempts to deliver.
 * <p>
 * On <code>notify</code>, the template is added to the
 * <code>TemplateHolderSet</code> named <code>template</code>.  This
 * stores <code>TemplateHolder</code> objects for each known type.
 * <p>
 * In this implementation, <code>EntryRep</code> ID's are approximate
 * time stamps.
 *
 * @see #setupStore
 * @see #startTxn
 * @see #endTxn
 * @see #transientOf
 * @see EntryRep
 * @see TypeTree
 * @see EntryHolderSet
 * @see TemplateHolderSet
 * @see Notifier
 *
 * @author Ken Arnold
 */
public abstract class BasicSpace 
    implements OutriggerSpace, TimeConstants, LogOps, LocalLandlord
{
    static final String 	DEBUG_PROP = "com.sun.jini.outrigger.debug";
    static final Debug 		debug = new Debug(DEBUG_PROP);

    static final PrintWriter	dbgOps    = debug.getWriter("ops");
    static final PrintWriter	dbgLeases = debug.getWriter("leases");
    static final PrintWriter	dbgTxns   = debug.getWriter("txns");
    static final PrintWriter	dbgMemory = debug.getWriter("memory");
    static final PrintWriter	dbgStartup = debug.getWriter("startup");

    private final static boolean	DEBUG = false;

    /**
     * List of <code>com.sun.jini.outrigger.EntryHolder</code>s for 
     * each specific type.
     *
     * @serial
     */
    protected EntryHolderSet	contents;

    //!! The list current is never GC'ed.  Maybe weak refs to the class?
    /**
     * A list of known subclasses of each class of entry.
     * @serial
     */
    protected TypeTree		types;

    /**
     * Templates for which someone has registered interest
     *
     * @serial
     */
    protected TemplateHolderSet	templates;

    /**
     * The transactions in which this space is a participant.
     *
     * @serial
     */
    protected Hashtable		txns = new Hashtable();

    /**
     * The transactions on this space that are prepared from a previous
     * run, but whose managers were uncontactable at boot.
     *
     * @serial
     */
    private OSHashtable		brokenTxns;

    /**
     * The crash count.  This must be different each boot that forgets
     * (implicitly aborts) pending transactions on shutdown.  A subclass
     * could override this setting if it did *not* forget operations on
     * crashes, but right now they all do so we just set it here.
     *
     * @serial
     */
    protected final long	crashCount = System.currentTimeMillis();

    /**
     * The reaper thread for expired notifications
     *
     * @serial
     */
    private Thread		expireThread;

    /**
     * The reaper thread for removed entries. 
     * Otherwise the entries are not garbage collected after removal
     * since their "prev" links are never traversed which has the side
     * effect of removing the associated object references.
     *
     * @serial
     */
    private Thread		entryReaperThread;

    /**
     * Flag to determine whether to disable entry reaping.
     */
    protected boolean disableReaping =
        Boolean.getBoolean(
            "com.sun.jini.outrigger.basicspace.disableReaping");

    /**
     * Background entry reaping interval (in milliseconds).
     * Default is 5 minutes.
     */
    private long reapingInterval =
        Long.getLong(
            "com.sun.jini.outrigger.basicspace.reapingInterval",
	    5 * MINUTES).longValue();

    /**
     * The Store object for this space
     *
     * @serial
     */
    private Store		store;

    /**
     * The notification object
     *
     * @serial
     */
    private Notifier		notifier;

    /**
     * The object that knows if a <code>write</code> is in progress.
     *
     * @serial
     */
    private InProgress		writing = new InProgress("BasicSpace.write");

    /**
     * The last entry written -- updated by recentWrite()
     *
     * @serial
     */
    private WrittenEntry 	lastWritten = WrittenEntry.startNode();

    /**
     * The monitor for ongoing transactions.
     */
    private transient TxnMonitor	txnMonitor;

    /**
     * The proxy for this space.
     *
     * @serial
     */
    private SpaceProxy	spaceProxy = null;

    /**
     * The admin proxy for this space.
     *
     * @serial
     */
    private AdminProxy	adminProxy = null;

    /**
     * The default name for a basic space bound in the RMI registry, or 
     * the name attribute used for a basic space registered with a Jini 
     * Lookup service
     */
    public static final String	DEFAULT_NAME = "JavaSpace";

    /**
     * The property string that is used to override the default
     * setting for the name of the space. If set, its value will
     * override the <code>DEFAULT_NAME</code> setting. Otherwise
     * the name will default to <code>DEFAULT_NAME</code>.
     */
    public static final String	SPACE_NAME_PROP = 
	"com.sun.jini.outrigger.spaceName";

    /**
     * Policy used to create and renew leases on entries
     *
     * @see com.sun.jini.lease.landlord.LandlordLease
     * @serial
     */
    protected LeasePolicy entryLeasePolicy = 
        new LeaseDurationPolicy(Long.MAX_VALUE,	// Allow VERY long leases
				1 * DAYS,       // default lease is one day 
				this, null, null);

    /** Policy used to create and renew leases on events */
    private final static long MAX_EVENT_LEASE =
	Long.getLong(
	    "com.sun.jini.outrigger.maxEventLease", 1 * HOURS).longValue();
    private final static long DEFAULT_EVENT_LEASE =
	Long.getLong(
	    "com.sun.jini.outrigger.defaultEventLease", 1 * HOURS).longValue();

    /**
     * Policy used to create and renew leases on event registrations
     *
     * @see com.sun.jini.lease.landlord.LandlordLease
     * @serial
     */
    protected LeasePolicy eventLeasePolicy =
        new LeaseDurationPolicy(MAX_EVENT_LEASE,
				DEFAULT_EVENT_LEASE,
				this, null, null);
    /**
     * Object we use to lock <code>binder</code> and <code>admi</code>n.
     */
    transient private Object adminLock = new Object();

    /**
     * Flag which is set upon a destroy request. If the this flag
     * flag is true, then binder creation is disabled.
     * This variable is guarded by <code>adminLock</code>.
     */
    private boolean destroyInProgress = false;

    /**
     * @serial
     */
    private Binder binder;

    /**
     * @serial
     */
    private JavaSpaceAdmin admin;

    /**
     * Our IDs have two 32 bit parts.  One part is a sequence number,
     * the other is a random number generated from a secure seed.
     * this field holds the next number in the sequence.
     * <p>
     * Access to this field is guarded by <code>idGen</code>'s lock */
    private static int nextSeqID = 0;

    /** 
     * Our IDs have two 32 bit parts.  One part is a sequence number,
     * the other is a random number generated from a secure seed.
     * This is the generator for the random part of our IDs. 
     * <p>
     * The lock on this object should be obtained before using it or 
     * <code>nextSeqID</code>.
     */
    private static final SecureRandom idGen = new SecureRandom();

    /** 
     * The buffer for generating IDs. 
     * <p>
     * Access to this field is guarded by <code>idGen</code>'s lock
     */
    private static final byte[] idGenBuf = new byte[4];

    //!! JWFM: Ugly static that is set to "this" in BasicSpace
    // constructor works because we never create more than one space
    // in a given VM.  Need to give SimpleEntryHolder and
    // TemplateHolder access to cancelOp() and expired() so
    // back-end won't expire entries on their own and the front end
    // will tell the back-end about reaped resources
    static BasicSpace staticSpace = null;

    /**
     * Create a new (transient) <code>BasicSpace()</code>. This method
     * exports <code>this</code> object as well.
     */
    BasicSpace() throws RemoteException {

	spaceProxy = new SpaceProxy(
	    (OutriggerSpace)UnicastRemoteObject.exportObject(this));

        setSpaceRef(this);

        doDebugGC(dbgMemory);
	dumpMemory(dbgMemory, "BasicSpace::BasicSpace(String) - Memory");
    }
    
    /**
     * Create a new (activatable) <code>BasicSpace</code>.  This method
     * also exports <code>this</code> object.
     */
    BasicSpace(ActivationID activationID, MarshalledObject data) 
	throws RemoteException, IOException
    {
        spaceProxy = new SpaceProxy((OutriggerSpace)
	    Activatable.exportObject(this,activationID, 0));

	setSpaceRef(this);

        doDebugGC(dbgMemory);
        dumpMemory(dbgMemory, 
	    "BasicSpace::BasicSpace(ActivationID, MarshalledObject)" 
	    + " - Memory");
    }

    /**
     * Set the (static) space reference to the provided instance
     * reference.  Note that this method can only be successfully
     * called once per VM. 
     */
    private synchronized static void setSpaceRef(BasicSpace ref) {
	if (staticSpace != null)
	    throw new IllegalStateException("Can't have two " +
	        "Outrigger Spaces in the same VM");
	staticSpace = ref;
    }

    /**
     * Each specific subclass is responsible for invoking
     * <CODE>initialize</CODE> when its store is ready for it.
     */
    protected void initialize(Store store) throws RemoteException, IOException {
	initState(store);
    }

    /**
     * The <code>initState</code> method's purpose is to encapsulate the subset
     * of <code>BasicSpace</code> initialization common to "activated 
     * construction" and regular construction.
     */
    private void initState(Store store) throws RemoteException {
	this.store = store;
        store.setupStore();

	if (DEBUG)
	    writing.debug(true);

	startRuntime();
    }

    /**
     * Start up the runtime processes (notification, reaping, etc.).
     * This is separated out and protected so that a subclass can prevent
     * any runtime processing.  Specifically, the BackEndSpace should
     * be doing none of this work.
     */
    protected void startRuntime() {
	notifier = new Notifier(this);

	startThreads();
    }


    /**
     * Return this space's proxy object.
     */
    public SpaceProxy getSpaceProxy() throws RemoteException {
	return spaceProxy;
    }

    /**
     * Set up whatever <code>com.sun.jini.outrigger.Store</code> we're using.
     */
    public void setupStore() {
	store.setupStore();
    }

    /**
     * Start up a new transaction for this space.
     */
    public void startTxn() {
	store.startTxn();
    }

    /**
     * Commit the current transaction if <code>commit</code> is true,
     * otherwise abort it.  It is harmless to invoke this if no
     * transaction is in progress -- this allows you to put a
     * <code>commitTxn</code> or <code>abortTxn</code> in a
     * <code>finally</code> clause without worry that it may get
     * invoked before the <code>startTxn</code> in the <code>try</code>
     * block happens.
     *
     */
    public void endTxn(boolean commit) {
	store.endTxn(commit);
    }

    /**
     * Return a reference to a version of this object usable in a
     * transient space. For example, as a return value of a remote method.
     * The returned reference may be the same object reference as the original 
     * if no change is needed.
     */
    public Object transientOf(Object orig) {
	return store.transientOf(orig);
    }

    /**
     * Commit the current transaction, i.e., invoke <code>endTxn(true)</code>.
     *
     * @see #endTxn
     */
    void commitTxn() {
	endTxn(true);
    }

    /**
     * Abort the current transaction, i.e., invoke <code>endTxn(false)</code>.
     *
     * @see #endTxn
     */
    void abortTxn() {
	endTxn(false);
    }

    /**
     * Log a server boot.  This implementation does nothing.
     */
    public void bootOp(long time) { }

    /**
     * Log a "write" operation.  This implementation does nothing.
     */
    public void writeOp(EntryRep rep, Long trID) { }

    /**
     * Log a "take" operation.  This implementation does nothing.
     */
    public void takeOp(long repID, Long trID) { }

    /**
     * Log a "notify" operation.  This implementation does nothing.
     */
    public void registerOp(EntryRep tmpl, NotifyChit chit) { }

    /**
     * Log a "notified" operation.  This implementation does nothing.
     */
    public void notifiedOp(long chitID, long newSeq) { }

    /**
     * Log a "renew" operation.  This implementation does nothing.
     */
    public void renewOp(Long cookie, long expiration) { }

    /**
     * Log a "cancel" operation, including a transaction abort.
     * This implementation does nothing.
     */
    public void cancelOp(Long cookie) { }

    /**
     * Log a transaction prepare.  This implementation does nothing.
     */
    public void prepareOp(Long trID, StorableObject tr) { }

    /**
     * Log a transaction commit.  This implementation does nothing.
     */
    public void commitOp(Long trID) { }

    // purposefully inherit doc comment from supertype
    public Lease write(EntryRep rep, Transaction tr, long lease)
	throws TransactionException, RemoteException
    {
	if (dbgOps != null)
	    dbgOps.println("BasicSpace: write: just inside");
	boolean invokedStart = false;
	try {
	    try {
		writing.start();	// say that we've started a write
		invokedStart = true;
	    } catch (InterruptedException e) {
		throw new InternalSpaceException("write interrupted");
	    }

	    if (dbgOps != null)
		dbgOps.println("BasicSpace: write: setID");
	    setID(rep);

	    if (dbgOps != null)
		dbgOps.println("BasicSpace: write: startTxn");
	    startTxn();

	    if (false) {
		ServerTransaction str = serverTransaction(tr);
		if (dbgOps != null)
		    dbgOps.println("BasicSpace: write: str= " + str);
		if (dbgOps != null)
		    dbgOps.println("BasicSpace: write: adding an action to "
			+ "transaction(" + str.id + ")");
	    }

	    Txn txn = enterTxn(tr);

	    if (dbgOps != null)
		dbgOps.println("writing " + rep + " (txn = " + txn + ")");

	    Lease entryLease = addWrittenRep(rep, txn, lease);

	    /*
	     * Log this addition.
	     */
	    writeOp(rep, idFor(tr));

	    /*
	     * Remember this entry as recently written.  This list is used
	     * for notifications and for ongoing matches.
	     */
	    recentWrite(rep, txn);

	    return entryLease;
        } finally {
	    commitTxn();
	    if (invokedStart)
		writing.stop();	// say that we've stopped a write
	}
    }

    /**
     * Add the written rep to the relevant data structures.  If the
     * <code>lease</code> is not zero, a lease will be returned;
     * otherwise the method will return <code>null</code>.  This is the
     * common code between <code>write</code> and the code in
     * <code>FrontEndSpace</code> that handles prepared but not
     * committed written <code>EntryRep</code> objects.
     *
     * @see #write
     * @see FrontEndSpace#addPreparedWrite
     */
    Lease addWrittenRep(EntryRep rep, Txn txn, long lease) {
	if (dbgOps != null)
	    dbgOps.println("BasicSpace: addWrittenRep");
	/*
	 * Since we're adding it, update the type information.  Extra
	 * type information isn't very critical, but having an object
	 * in the system which cannot be matched by all its
	 * superclasses is Just Plain Wrong.  We do this first in case
	 * of a crash between the two operations (We could do it all
	 * in one synchronized wad, but that seems unnecessary,
	 * considering the small overhead of extra type info that will
	 * likely reappear as the entry that failed to be successfully
	 * added due to the crash gets added again later).
	 */
	types.addTypes(rep);

	/*
	 * Now create the lease for this entry.  NOTE: this must be
	 * done before the add().  Otherwise there is a hole
	 * between the add and the time leaseFor() is invoked
	 * wherein an entry has been added to the holder with an
	 * expiration of zero, i.e., it has already expired.
	 * During this window, anyone scanning the list will see
	 * the "expired" entry and remove it, killing it.
	 * (Alas, and yet so young!).
	 */
	Lease entryLease = null;
	try {
	    if (lease != 0)
		entryLease = entryLeasePolicy.leaseFor(rep, lease);
	} catch (LeaseDeniedException e) {
	    // This should never happen, we should not be using a
	    // policy that could thrown LeaseDenied on lease
	    // creation, re-throw as InternalSpaceException
	    throw new 
		InternalSpaceException("BasicSpace:entryLeasePolicy." +
		    "leaseFor threw LeaseDeniedException", e);
	}

	/*
	 * Now we can add the entry, safe in the knowledge that it can be
	 * found by any of its superclasses
	 */
	contents.holderFor(rep).add(rep, txn);

	return entryLease;
    }

    /**
     * Take the actions necessary for side effects of a recently
     * written entry.  Specifically, set up any pending notifications
     * from the writing of <code>entry</code> and remember it as the
     * new head of the <CODE>lastWritten</CODE> entry list.  This does
     * its work as quickly as possible and lets the actual effects
     * happen in other threads.
     */
    void recentWrite(EntryRep rep, TransactableMgr txn) {
	//!! notifier probably should use lastWritten instead of its own queue
	if (notifier != null)
	    notifier.newEntry((EntryRep) transientOf(rep), txn);
	synchronized (this) {
	    lastWritten = lastWritten.add(rep);
	}
    }

    /**
     * Set up any pending notifications from the abort that
     * happened on the existing <code>entry</code>.  This does its
     * work as quickly as possible and lets the <code>Notifier</code>
     * analyze this in another thread.
     */
    void abortOnEntry(EntryRep rep) {
	if (notifier != null)
	    notifier.abortForEntry((EntryRep) transientOf(rep));
    }

    /**
     * Tell the notifier that the given chit has an exception to send
     */
    void exceptionToSend(NotifyChit chit) {
	if (notifier != null)
	    notifier.exceptionToSend((NotifyChit) transientOf(chit));
    }

    // purposefully inherit doc comment from supertype
    public EventRegistration
	notify(EntryRep tmpl, Transaction tr, RemoteEventListener listener,
	       long leaseTime, MarshalledObject handback)
	throws TransactionException, RemoteException
    {
	AddRecord ar = notifyImpl(tmpl, tr, listener, leaseTime, handback);
	if (ar == null)
	    return null;
	else
	    return ar.reg;
    }

    /**
     * The actual implementation of <code>notify</code>, which returns a
     * container class holding both the <code>EventRegistration</code>
     * object and the chit.  This is used internally.
     */
    private TemplateHolderSet.AddRecord
	notifyImpl(EntryRep tmpl, Transaction tr, RemoteEventListener listener,
	           long leaseTime, MarshalledObject handback)
	throws TransactionException, RemoteException
    {
	if (dbgOps != null)
	    dbgOps.println("BasicSpace: notify");

	if (listener == null)
	    throw new NullPointerException("Passed null listener for event " +
					   "registration");

	// very much like write, but into a different list
	TemplateHolderSet.AddRecord ar = null;
	try {
	    startTxn();

	    /*
	     * Setup up data for this template.
	     */
	    tmpl = setupTmpl(tmpl);

	    /**
	     * Now we add the template since it can be found.
	     */
            ServerTransaction str = serverTransaction(tr);

	    Txn txn = enterTxn(tr);
	    // add() invokes registerOp on this
	    ar = templates.add(tmpl, listener, txn, nextID(), eventLeasePolicy,
			       leaseTime, handback, spaceProxy, this, tr);
        } finally {
	    commitTxn();
	}
	return ar;
    }

    private EntryRep setupTmpl(EntryRep tmpl) {
	/*
	 * If the notification is for a null template, use the special
	 * matchAnyEntryRep.  When the notification system adds a
	 * record for this EntryRep, any time any incoming EntryRep
	 * enters the space, the notification system will
	 * successfully match it and send the notification.
	 */
	if (tmpl == null)
	    tmpl = EntryRep.matchAnyEntryRep();


	/*
	 * This entry may be a new type if the notificaiton request
	 * arrives before any entry of the type is written.
	 */
	types.addTypes(tmpl);

	return tmpl;
    }

    /**
     * Add a pre-created registration to the list of notifications.  We
     * don't share code with notify because it is too different -- several
     * objects and IDs are allocated in notify, but the register method 
     * requires none to be allocated.
     *
     * @see #notify
     */
    protected void register(EntryRep tmpl, NotifyChit chit) {
	if (dbgOps != null)
	    dbgOps.println("BasicSpace: register " + tmpl + ", " + chit);
	tmpl = setupTmpl(tmpl);
	templates.add(tmpl, chit);
    }

    /**
     * Change the sequence number for the given chit <code>idNum</code> to 
     * be the given value <code>newSeq</code> (if the current value is less 
     * than the specified one).
     * Sequence numbers only increase -- a lesser number here means that
     * the calls are out of order, which is ignored as being harmless.
     * Chit ID's are used because the template holder can map a chit ID
     * to a chit, and then to that chit's handle, which is what stores the
     * sequence ID.  No mapping exists from the event directly to the handle.
     */
    protected void changeSeq(long idNum, long newSeq) {
	if (dbgOps != null)
	    dbgOps.println("changeSeq(" + idNum + ", " + newSeq + ")");
	templates.changeSeq(idNum, newSeq);
    }

    /**
     * Return the list of chits that need to be notified because of the
     * given entry.  We use <code>transientOf</code> because the actual
     * notification is done outside a transaction.
     *
     * @see Notifier
     */
    Vector chitsFor(EntryRep rep, TransactableMgr txn, boolean proxyOnly) {
	// a list of templates that match this entry
	Vector v = templates.allMatches(rep, txn, proxyOnly);
	return (Vector) transientOf(v);
    }

    // purposefully inherit doc comment from supertype
    public void cancel(Object cookie) throws UnknownLeaseException {
	try {
	    startTxn();
	    if (DEBUG)
		System.out.println("cancel(" + cookie + ")");
	    if (!contents.cancel(cookie) && !templates.cancel(cookie))
		throw new UnknownLeaseException();
	    cancelOp((Long) cookie);
	} finally {
	    commitTxn();
	}
    }

    // purposefully inherit doc comment from supertype
    public long renew(Object cookie, long extension) 
        throws UnknownLeaseException, LeaseDeniedException
    {
	LeasedResource  resource;
	LeasePolicy     policy;
	
	if (null != (resource = contents.getLeasedResource(cookie))) 
	    policy = entryLeasePolicy;
	else if (null != (resource = templates.getLeasedResource(cookie)))
	    policy = eventLeasePolicy;
	else
	    throw new UnknownLeaseException("Not lease for cookie:" + cookie);

	// We check to see if the lease has expired twice so we avoid 
	// unnecessary synchronization 
	if (!policy.ensureCurrent(resource)) {
	    // Lease has expired, don't renew
	    throw new UnknownLeaseException("Lease has already expired");
	}
	    
	synchronized (resource) {
	    if (!policy.ensureCurrent(resource)) {
		// Lease has expired, don't renew
		throw new LeaseDeniedException("Lease has already expired");
	    }

	    // No one can expire the lease, so it is safe to update.

	    try {
		startTxn();
		long granted = policy.renew(resource, extension);
		renewOp((Long) cookie, resource.getExpiration());
		if (dbgLeases != null) {
		    dbgLeases.println("renew(" + cookie + ", " + 
				      extension + ") returns " + granted);
		}
		return granted;
	    } finally {
		commitTxn();
	    }
	}
    }

    /**
     * Atomic test and set of expiration status of the resource.  If
     * the resource is expired returns true and if necessary set the
     * expiration to 0.
     */
    boolean expired(LeasedResource resource, long now) {
	if (resource.getExpiration() > now) 
	    return false;

	// Else may be expired, grab the resource so no one can renew
	// it, check to make sure it is still expired, and set its
	// expiration to 0 so everyone else will definitely think it
	// is gone
	synchronized (resource) {
	    if (resource.getExpiration() > now) 
		return false;

	    // Really is gone...
	    // This tells everyone it is gone, even if it is still in some
	    // table or local reference.
	    resource.setExpiration(0);
	    return true;
	}
    }

    // purposefully inherit doc comment
    public Landlord.RenewResults renewAll(Object[] cookies, long[] extensions) {
	try {
	    if (dbgLeases != null)
		dbgLeases.println("renewAll: " + cookies.length + " leases");

	    startTxn();
	    return LandlordUtil.renewAll(this, cookies, extensions);
	} finally {
	    commitTxn();
	}
    }

    // purposefully inherit doc comment
    public Map cancelAll(Object[] cookies) {
	try {
	    startTxn();
	    return LandlordUtil.cancelAll(this, cookies);	    
	} finally {
	    commitTxn();
	}
    }

    // purposefully inherit doc comment from supertype
    public EntryRep
	getMatch(EntryRep tmpl, Transaction tr, long timeout, boolean takeIt,
		 boolean ifExists)
	throws TransactionException, TransactionConflictException,
	       RemoteException
    {
	return (EntryRep)
	    getMatch(tmpl, tr, timeout, takeIt, ifExists, null, 0);
    }

    // purposefully inherit doc comment from supertype
    public Object
        getMatch(EntryRep tmpl, Transaction tr, long timeout, boolean takeIt,
                 boolean ifExists, RemoteEventListener listener, long id)
        throws TransactionException, TransactionConflictException,
               RemoteException
    {
        try {
            startTxn();
            return getMatchImpl(tmpl, tr, timeout, takeIt, ifExists,
                                    listener, id);
        } finally {
            commitTxn();
        }
    }

    /**
     * The current call number to getMatchImpl, for debug purposes only.
     */
    private transient int getMatchCallCnt = 0;

    /**
     * Implement the various <code>getMatch</code> calls in common code.
     * This is long and hairy, and probably needs to be split into separate
     * calls, but it hasn't been yet.
     */
    private Object
        getMatchImpl(EntryRep tmpl, Transaction tr, long timeout,
	    boolean takeIt, boolean ifExists, RemoteEventListener listener,
	    long id)
	throws TransactionException, TransactionConflictException,
	    RemoteException
    {
	int thisCall = 0;	// used in debug code, otherwise left as zero
	if (dbgOps != null) {
	    synchronized (this) {
		thisCall = getMatchCallCnt++;
	    }
	}

	// after this time, can stop looking
	long now = System.currentTimeMillis();
	long endTime;				// when should we stop?
	if (Long.MAX_VALUE - timeout <= now)	// if timeout takes us too far
	    endTime = Long.MAX_VALUE;		// ... go as long as possible
	else					// otherwise
	    endTime = now + timeout;		// ... go up to then

        Txn txn = enterTxn(tr);

        if (dbgOps != null) {
            dbgOps.println(thisCall + ":getMatch(" + tmpl + ", " + txn
		+ ", " + timeout + ", " + (takeIt ? "take" : "read") + ", "
                + (ifExists ? "" : "!") + "ifExists)");
        }

	/*
	 * This plugs a hole -- if an entry is written while getMatch
	 * is scanning the list, it may arrive after getMatch has
	 * passed that point in the list.  So getMatch first sets up
	 * the notify, so such a notification may arrive *before*
	 * getMatch returns.  We assume that the long held in the
	 * handback object "id" will uniquely identify the event to the
	 * listener who has not yet gotten the event registration ID
	 * back (see SpaceProxy's listener code).
	 */
	WrittenEntry writtenEntry = lastWritten;
	if (dbgOps != null)
	    dbgOps.println(thisCall + ":writtenEntry = " + writtenEntry);

	EntryRep rep = null;
	TransactionConflictException conflicts = null;

	/*
	 * First we do the straight search
	 */
	try {
	    rep = find(tmpl, txn, takeIt);
	    if (dbgOps != null) {
		dbgOps.println(thisCall + ":getMatch: did find() "
		    + "got rep = " + rep);
	    }
	    if (rep != null) {	// found it
		if (takeIt)
		    takeOp(rep.id, idFor(tr));
		return rep;
	    }
	} catch (TransactionConflictException e) {
	    conflicts = e;
	}

	if (dbgOps != null) {
	    dbgOps.println(thisCall + ":nothing found, " +
		(conflicts == null ? "null" : conflicts.txns().size() + "") +
		" conflicts, scanning, endTime = " + endTime);
	}


	/*
	 * If we are under an internal probe, then we are re-checking
	 * status for a prior blocking call.
	 * Note: INTERNAL_PROBE is a negative value, so endTime 
	 * calculations are not going to be valid past this point.
	 * Therefore, we need to exit from this block with some sort
	 * of answer.
	 */
	if (timeout == INTERNAL_PROBE) {
	    if (dbgOps != null) {
		dbgOps.println(thisCall + ":virtual timeout " 
	    	    + "conflicts=" + conflicts.txns().size() + ", "
		    + "ifExists=" + ifExists);
	    }
	    if (conflicts != null && ifExists) {
		monitor(conflicts.txns());
		throw conflicts;
	    } else {
		return null;
	    }
	}

	/*
	 * Now we look through entries written since we started.
	 */
	boolean blockedWrite = false;	// have we blocked further writes?
	boolean timedOut = false;	// did we time out?
	long time = System.currentTimeMillis();
	if (time >= endTime) {
	    if (dbgOps != null) {
		dbgOps.println(thisCall + ":out of time (" + time + ") "
		    + "before scanning");
	    }
	    if (conflicts != null) 
		monitor(conflicts.txns());
	    
	    return null;
	}

	try {
	    do {
		if (dbgOps != null) {
		    dbgOps.println(thisCall + ":checking " + writtenEntry
			+ " (" + writtenEntry.timestamp() + ")");
		}
		if (writtenEntry.timestamp() > endTime)	{
		    timedOut = true;			// hit something later
		    if (dbgOps != null)
			dbgOps.println(thisCall + ":timed out");
		    break;				// than we care about
		}

		EntryRep wer = writtenEntry.rep();	// null at head of list
		if (wer != null) {
		    try {
			if (dbgOps != null)
			    dbgOps.println(thisCall + ":trying " + wer);
			EntryHolder holder = contents.holderFor(wer);
			
			if (dbgOps != null) {
			    dbgOps.println(thisCall + ":getMatch: " +
						"holder = " + holder);
			    dbgOps.println("doing holder.take: tmpl = " + tmpl +
						" wer = " + wer);
			}

			rep = holder.match(tmpl, wer, txn, takeIt, time,
					   conflicts);
			if (rep != null) {
			    if (dbgOps != null)
				dbgOps.println(thisCall + ":found it!");
			    if (takeIt)
				takeOp(rep.id, idFor(tr));
			    return rep;
			}
		    } catch (TransactionConflictException e) {
			if (dbgOps != null)
			    dbgOps.println(thisCall + ":conflicts");
			if (conflicts == null)
			    conflicts = e;
			// otherwise match() already did the adding
		    }
		    if (dbgOps != null)
			dbgOps.println(thisCall + ":no good");
		}

		/*
		 * If we are almost at the end of the list we have to block
		 * further writes from proceeding so that something can't
		 * get added to the end of the list until after we have
		 * set up for notification.  We actually might get to the
		 * end of the list twice if a new entry is added after
		 * the null test but before block() is invoked.  That's
		 * okay -- it means we've just blocked writes a tad early,
		 * but we have to make sure we don't block writes twice.
		 */
		if (writtenEntry.next() == null && !blockedWrite) {
		    blockedWrite = true;
		    try {
			if (dbgOps != null)
			    dbgOps.println(thisCall + ":blocking writes");
			writing.block();
		    } catch (InterruptedException e) {
			throw
			    new InternalSpaceException("getMatch interrupted");
		    }
		}
		if (dbgOps != null)
		    dbgOps.println(thisCall + ": getMatch: writtenEntry" +
			"writtenEntry = " + writtenEntry);	
		writtenEntry = writtenEntry.next();
	    } while (writtenEntry != null);
	    if (dbgOps != null)
		dbgOps.println(thisCall + ":nothing found in loop");

	    /*
	     * Nothing found
	     */

	    if (timedOut) {
		if (conflicts != null)
		    monitor(conflicts.txns());
		return null;
	    }

	    /*
	     * If this is an ifExists search and there were no
	     * conflicts, then there is nothing, so just return null.
	     * Any writes overlapping here logically happen after this
	     * call.  If there is only a conflict, and if there is no
	     * listener to handle future arrivals or changes of entry
	     * states for a non-conflicted future invocation, then we
	     * should throw TransactionConflictException to indicate
	     * the situation.
	     */
	    if (ifExists) {
		if (conflicts == null)		// a real answer
		    return null;
		else if (listener == null) {	// no listener, report conflict
		    monitor(conflicts.txns());	// watch the conflicting txns
		    throw conflicts;
		}
	    }

	    /*
	     * This is either an ifExists search with a conflict found, or
	     * a wait-for-it search with nothing or a conflict found, so set
	     * up the notification if there is a listener.
	     */
	    if (listener != null) {
		if (dbgOps != null)
		    dbgOps.println(thisCall + ":setting up notify");
		try {
		    // In the case of a null template, if a match was found,
		    // but the template is null, the notification must
		    // must be set to what was found and
		    // not for the null template.

		    TemplateHolderSet.AddRecord ar =
			notifyImpl(tmpl, tr, listener, timeout,
				   new MarshalledObject(new Long(id)));
		    EventRegistration reg = ar.reg;
		    // During the notify() the txn might have changed state
		    if (txn != null && txn.getState() != ACTIVE) {
			try {
			    reg.getLease().cancel();
			} catch (UnknownLeaseException e) {
			    ;	// that's okay
			}
			throw new CannotJoinException("not active");
		    }
		    if (conflicts != null)	// watch the conflicting txns
			monitor(ar.chit, conflicts.txns());
		    return reg;
		} catch (IOException e) {
		    throw new InternalSpaceException("cannot marshall id");
		}
	    }
	    if (dbgOps != null)
		dbgOps.println(thisCall + ":returning null");
	    return null;
	} finally {
	    // now we can unblock writing
	    if (blockedWrite) {
		if (dbgOps != null)
		    dbgOps.println(thisCall + ":unblocking writing");
		writing.unblock();
	    }
	}
    }

    /**
     * Make sure the transactions listed here are monitored for as long
     * as the given leased resource exists.  The lease would usually
     * govern the notification registration.
     */
    void monitor(LeasedResource res, Collection toMonitor) {
	if (TxnMonitor.dbg != null) {
	    TxnMonitor.dbg.println("monitor " + toMonitor.size() + " txns, res "
		+ res);
	    Iterator it = toMonitor.iterator();
	    while (it.hasNext())
		TxnMonitor.dbg.println("\t" + it.next());
	}
	monitor().add(res, toMonitor);
    }

    /**
     * Make sure the transactions listed here are monitored for
     * a reasonable amount of time since they recently caused a conflict,
     * although for a non-leased event.
     */
    void monitor(Collection toMonitor) {
	if (TxnMonitor.dbg != null) {
	    TxnMonitor.dbg.println("monitor " + toMonitor.size()
		+ " toMonitor, no lease");
	    Iterator it = toMonitor.iterator();
	    while (it.hasNext())
		TxnMonitor.dbg.println("\t" + it.next());
	}
	monitor().add(toMonitor);
    }

    /**
     * Return the monitor, allocating it if necessary.
     */
    private TxnMonitor monitor() {
	if (txnMonitor == null)
	    txnMonitor = new TxnMonitor(this);
	return txnMonitor;
    }

    /**
     * Debug method: Dump out the bucket for the holder for the given
     * rep.
     */
    void dump(String name, EntryRep rep) {
	dump((SimpleEntryHolder) contents.holderFor(rep), name, rep);
    }

    /**
     * Debug method:  Dump out the bucket for the given holder, for an
     * operation using the given rep.
     */
    static void dump(SimpleEntryHolder holder, String name, EntryRep rep) {
	try {
	    holder.dump(name + " " + rep.entry());
	} catch (Exception e) {
	    e.printStackTrace();
	}
    }

    /**
     * Find an entry that matches the given template.  This simply uses
     * the other find after retrieving the class for the template.
     * This is the starting point for the recursion down the known
     * subtypes of the template type -- the one where the type of the
     * entry in the template is also the type to search.
     */
    private EntryRep find(EntryRep tmplRep, Txn txn, boolean takeIt)
	throws TransactionException, TransactionConflictException
    {
	try {
	    startTxn();
	    if (tmplRep == null)
		tmplRep = EntryRep.matchAnyEntryRep();

	    //!! This is a simple mechanism, but heavy.  Need better locking
	        EntryRep ret =
		    find(tmplRep, tmplRep.classFor(), txn, takeIt, null);

	    return (EntryRep) transientOf(ret);
	} finally {
	    commitTxn();
	}
    }

    /**
     * Find an entry that is at least <code>whichClass</code> that
     * matches the template <code>tmplRep</code> (at least the
     * template's type, and matches in values provided).
     * <code>whichClass</code> is at least the type of the template.
     * If <code>takeIt</code> is <code>true</code> the matching entry
     * is removed.
     */
    private EntryRep
	find(EntryRep tmplRep, String whichClass, Txn txn, boolean takeIt,
	     TransactionConflictException conflicts)
	throws TransactionException, TransactionConflictException
    {
	if (dbgOps != null) {
	    dbgOps.println("BasicSpace: find2");
	    dbgOps.println(whichClass + ": finding");
	}

	/*
	 * The iterator returned by subTypes includes both this class
	 * and all subtypes.
	 */
	Iterator subtypes = types.subTypes(whichClass);
	String className = null;
	EntryRep result = null;
	boolean foundConflicts = false;
	EntryHolder holder = null;

	if (dbgOps != null)
	    dbgOps.println("BasicSpace: find: subtypes = " + subtypes);

	while (subtypes.hasNext()) {
	    className = (String) subtypes.next();
	    if (dbgOps != null)
		dbgOps.println("BasicSpace: find: className = "+className);

	    holder = contents.holderFor(className);
	    try {
	        result = holder.hasMatch(tmplRep, txn, takeIt, conflicts);
	        if (result != null) {
		    if (dbgOps != null)
		        dbgOps.println("Found " + result);
		    return result;
	        }
	    } catch (TransactionConflictException e) {
		/*
		 * Only thrown if conflicts was previously null -- otherwise
		 * hasMatch has merged in the conflicting txns into the
		 * exception it was passed.  So here, conflicts == null
		 */
		conflicts = e;
		foundConflicts = true;
	    }
	}


	if (dbgOps != null) {
	    dbgOps.println(whichClass + ": not found");
	    dbgOps.println("BasicSpace: find2: conflicts = " +
		(conflicts != null));
	}

	// if this was the first conflict found, throw the exception
	if (foundConflicts)
	    throw conflicts;

	// no result found
	return null;
    }


    /**
     * Set up the running threads needed for this space.
     */
    private void startThreads() {
	/*
	 * This thread reaps expired entries from the space.  Many expired
	 * entries are caught while doing other work in the space, but this
	 * thread will ensure that every expired entry is eventually purged,
	 * even if other work doesn't encounter it.
	 */
	expireThread = new Thread("expireThread") {
	    public void run() {
		try {
		    for (;;) {
			try {
			    startTxn();
			    templates.reapExpired();
			} finally {
			    commitTxn();
			}
			sleep(5 * MINUTES);
		    }
		} catch (InterruptedException e) {
		    return;
		}
	    }
	};
	// lower priority than default
	expireThread.setPriority(expireThread.getPriority() - 1);
	expireThread.setDaemon(true);
	expireThread.start();

        if (!disableReaping) {
	    if(dbgMemory != null) 
		dbgMemory.println("entryReaperThread::reapingInterval = " 
		    + reapingInterval);

            entryReaperThread = new Thread("entryReaperThread") {
                public void run() {
        	    try {
        	        for (;;) {
        		    sleep(reapingInterval);
        		    templates.reap();
        		    contents.reap();
        		}
       		    } catch (InterruptedException e) {
       		        return;
       		    }
       	        }
       	    };
    	    // lower priority than default, set to daemon and start.
    	    entryReaperThread.setPriority(entryReaperThread.getPriority() - 1);
    	    entryReaperThread.setDaemon(true);
    	    entryReaperThread.start();
	}
	else {
	    if(dbgMemory != null) 
	        dbgMemory.println("entryReaperThread is disabled.");
	}

    }

    /**
     * Set the ID for the given rep.
     */
    private void setID(EntryRep rep) {
	if (rep.id == EntryRep.ID_UNSET)
	    rep.id(nextID());
    }

    /**
     * Generate a new ID.  Part of the ID is from a
     * <code>SecureRandom</code> object so that it is next to
     * impossible to forge an ID.
     */
    static long nextID() {
	long id;
	synchronized (idGen) {
	    do {
		id = 0;
		idGen.nextBytes(idGenBuf);
		
		// Set the high 32 bits to the random number
		for (int i = 0; i < 4; i++)
		    id = (id << 8) | (idGenBuf[i] & 0xFF);
		id = id << 32;

		// Set the low 32 bits to the next sequence number
		id = id | (nextSeqID & 0xFFFFFFFF);

	    } while (id == 0 || id == EntryRep.ID_UNSET);   // skip flag values

	    // Bump the sequence number while we still have the lock on idGen
	    nextSeqID++;
	}
	return id;
    }

    // ------------------------------------------------------------
    //			Admin stuff
    // ------------------------------------------------------------

    /**
     * Return a proxy that implements that <code>JavaSpaceAdmin</code>
     * interface.
     *
     * @see JavaSpaceAdmin
     */
    public Object getAdmin() {
        synchronized (adminLock) {
            if (admin == null) 
                admin = new AdminProxy(this, getBinder().getDirectoryAdmin());
            return admin;
        }
    }

    /**
     * Utility method to get the binder, creating if necessary
     * assumes adminLock is held. If a destroy call has occurred, though,
     * just return <code>null</code>.
     */
    private Binder getBinder() {
	if (destroyInProgress) 
	    return null;
        if (binder == null) {
            // We need to get the binder
            try {
                binder = createBinder();
            } catch (Exception e) {
                throw new InternalSpaceException(
                        "Trouble creating binder admin", e);
            }
        }
        return binder;
    }

    /**
     * Default implemention of createBinder, just creates a 
     * transient binder. Subclasses will override this method in order
     * to produce the appropriate return value.
     */
    Binder createBinder() 
        throws IOException, ClassNotFoundException, ActivationException
    {
        return new com.sun.jini.mahout.binder.LookupBinder(
            attributesFor(), spaceProxy);
    }

    // purposefully inherit doc comment
    // Implementation of the OutriggerAdmin interface
    public JavaSpace space() throws RemoteException {
	return getSpaceProxy();
    }

    // purposefully inherit doc comment
    public RemoteIter contents(EntryRep tmpl, Transaction tr)
	throws TransactionException, RemoteException
    {
	//Note: Nothing special needs to be done for the
	//      BasicSpace iterator since each invocation
	//      of this method creates a new BasicSpaceIterator
	//      UnicastRemoteObject to handle enumerating the
	//      BasicSpace's contents.

	if (DEBUG) {
	    System.out.println("Admin: contents");
	    try {
		PrintWriter out = new PrintWriter(
		    new FileOutputStream("/tmp/restartBasicSpaceAdmin.txt",
			true));
		out.println("BasicSpaceAdmin restarted at: "
		    + System.currentTimeMillis());
		out.flush();
		out.close();
	    } catch (Exception e) {
		;	// just a debug exception
	    }
	}


	try {
	    startTxn();
	    ServerTransaction str = serverTransaction(tr);

	    return new RemoteIterImpl(tmpl, enterTxn(tr));
	} finally {
	    commitTxn();
	}
    }

    /**
     * Sleep time for <code>killVM</code>.
     *
     * @see killVM
     */
    private final static long KILL_DELAY = 10;

    // inherit doc comment
    public long killVM() {
	/*
	 * We have to do this in a separate thread or the remote
	 * system will always get a RemoteException when the server to
	 * which it is talking (this one) dies in the middle of the
	 * call.  So we start up a new thread that sleeps for
	 * KILL_DELAY milliseconds and then shuts down the VM.  This
	 * lets the thread running killVM return before the
	 * denoument.
	 */
	new Thread() {
	    public void run() {
		try {
		    sleep(KILL_DELAY);
		} catch (InterruptedException e) {
		    ;	// oh, well
		}
		System.exit(0);
	    }
	}.start();
	// We have the client sleep longer than KILL_DELAY so we have a buffer
	// for variegates of thread scheduling, note 100x is an empirically
	// arrived at multiplier.
	return KILL_DELAY * 100;
    }

    /**
     * Destroy this space, exiting when finished.  This unregisters
     * itself and related objects, and then destroys the persistent
     * state.
     */
    public void destroy() {

	// Lock out binder creation attempts
        synchronized (adminLock) {
	    if (!destroyInProgress)
	        destroyInProgress = true;
	}

	Exception failed = null;
	try {
	    ActivationGroupID grp = ActivationGroup.currentGroupID();
	    if (grp == null)		// must be Unicast export
		UnicastRemoteObject.unexportObject(this, true);
	    else {
		ActivationSystem sys = ActivationGroup.getSystem();
		sys.unregisterGroup(grp);
	    }
	} catch (Exception t) {
	    System.err.println("destroy unregistering:");
	    t.printStackTrace(System.err);
	    System.err.println("destroy continuing");
	}

	try {
	    // binder creation should be locked out at this point
	    // so we don't need the adminLock here.
	    if (binder != null)
	        binder.destroy();
	} catch (Exception t) {
	    System.err.println("Binder destruction:");
	    t.printStackTrace(System.err);
	    System.err.println("destroy continuing");	    
	}

	try {
	    if (store != null)
		store.destroy();// if persistent, recursively destroys entire
	                        // dir
	} catch (Exception t) {
	    System.err.println("Store destruction:");
	    t.printStackTrace(System.err);
	    System.err.println("destroy continuing");	    
	}
	if (dbgMemory != null) {
            dumpMemory(dbgMemory, "BasicSpace::destroy() - Before Reap");
            if(!disableReaping) {
                dbgMemory.println("BasicSpace::destroy() - Reaping entries");
                templates.reap();
                contents.reap();
            }
            else {
                dbgMemory.println("BasicSpace::destroy() - Reaping not run");
            }

            doDebugGC(dbgMemory);
            dumpMemory(dbgMemory, "BasicSpace::destroy() - After Reap");
        }

	killVM();	// go away
    }

    /**
     * <code>Admin</code> objects are connected to particular server objects
     * to implement the <code>JavaSpaceAdmin</code> interface for the server.
     *
     * @see JavaSpaceAdmin
     */
    private static class AdminProxy implements JavaSpaceAdmin, 
	KillVMAdmin, Serializable 
    {
	static final long serialVersionUID = -5918335755497933326L;

        /** Reference to the actual remote admin object. */
        private OutriggerAdmin          admin;

	/** Reference to the object that implements our JoinAdmin interface */
	private DirectoryAdmin          dirAdmin;
 
	private static final boolean DEBUG = false;

	/**
	 * Create an <code>AdminProxy</code> for the given remote admin
	 * objects.
	 */
	AdminProxy(OutriggerAdmin admin, DirectoryAdmin dirAdmin) {
	    this.admin = admin;
	    this.dirAdmin = dirAdmin;
	}

	// inherit doc comment
	public JavaSpace space() throws RemoteException {
	    return admin.space();
	}

	// inherit doc comment
	public AdminIterator contents(Entry tmpl, Transaction tr)
	    throws TransactionException, RemoteException
	{
	    return contents(tmpl, tr, 128);
	}

	// inherit doc comment
	public AdminIterator contents(Entry tmpl, Transaction tr, 
				      int fetchSize)
	    throws TransactionException, RemoteException
	{
	    return new BasicSpaceIterator(
	        admin.contents(SpaceProxy.repFor(tmpl), tr), fetchSize);
	}

	// inherit doc comment
	public long killVM() throws RemoteException {
	    return admin.killVM();
	}

	// inherit doc comment
	public void destroy() throws RemoteException {
	    admin.destroy();
	}

        // --------------------------------------------------
        //              RegistryAdmin
        // --------------------------------------------------
	// inherit doc comment
	public void enableRegistryBinding(String name, int port) 
	    throws IOException 
	{
	    dirAdmin.enableRegistryBinding(name, port);
	}

        // --------------------------------------------------
        //              JoinAdmin
        // --------------------------------------------------
	// inherit doc comment
	public Entry[] getLookupAttributes() throws RemoteException {
	    return dirAdmin.getLookupAttributes();
	}

	// inherit doc comment
	public void addLookupAttributes(Entry[] attrSets) 
	    throws RemoteException 
	{
	    dirAdmin.addLookupAttributes(attrSets);
	}

	// inherit doc comment
	public void modifyLookupAttributes(Entry[] attrSetTemplates, 
					   Entry[] attrSets)
	    throws RemoteException
	{
	    dirAdmin.modifyLookupAttributes(attrSetTemplates, attrSets);
	}

	// inherit doc comment
	public String[] getLookupGroups() throws RemoteException {
	    return dirAdmin.getLookupGroups();
	}

	// inherit doc comment
	public void addLookupGroups(String[] groups) throws RemoteException {
	    dirAdmin.addLookupGroups(groups);
	}

	// inherit doc comment
	public void removeLookupGroups(String[] groups) throws RemoteException {
	    dirAdmin.removeLookupGroups(groups);
	}

	// inherit doc comment
	public void setLookupGroups(String[] groups) throws RemoteException {
	    dirAdmin.setLookupGroups(groups);
	}

	// inherit doc comment
	public LookupLocator[] getLookupLocators() throws RemoteException {
	    return dirAdmin.getLookupLocators();
	}

	// inherit doc comment
	public void addLookupLocators(LookupLocator[] locators)
	    throws RemoteException
	{
	    dirAdmin.addLookupLocators(locators);
	}

	// inherit doc comment
	public void removeLookupLocators(LookupLocator[] locators) 
	    throws RemoteException
	{
	    dirAdmin.removeLookupLocators(locators);
	}

	// inherit doc comment
	public void setLookupLocators(LookupLocator[] locators) 
	    throws RemoteException
	{
	    dirAdmin.setLookupLocators(locators);
	}
    }

    /**
     * An iterator that returns all the reps in the space that are of
     * at least the given class.  It works through the contents of each
     * <code>EntryHolder</code> in turn, exhausting each holder's enumerated
     * contents.
     *
     * @see EntryHolder#contents
     */
    class AllReps implements RepEnum {
	RepEnum		curEnum;	// current unexhausted enum (or null)
	Stack		toDo;		// classes left to do
	Txn		txn;		// txn under which this is done

	private final boolean DEBUG = false;

	/** Create a new <code>AllReps</code> object for the given class. */
	AllReps(String classFor, Txn txn) {
	    toDo = new Stack();
	    this.txn = txn;
	    setup(classFor);		// start with this class
	}

	/**
	 * Set up the enumerator for the given class, adding any of its
	 * subclasses to the <code>toDo</code> stack.  Only direct subclasses
	 * are added -- the rest will be picked up when the subclasses are
	 * set up.
	 */
	private void setup(String classFor) {
	    if (DEBUG)
		System.out.println("setup(" + classFor + ")");

	    if (classFor == null)
		return;

	    // if we want all classes
	    if (classFor.equals(EntryRep.matchAnyClassName())) {
		Enumeration e = contents.allHolders();
		while (e.hasMoreElements()) {	// put all holders in "To Do"
		    SimpleEntryHolder h = (SimpleEntryHolder) e.nextElement();
		    toDo.push(h.className);	
		}
	    } else {
	        // add subclasses to the toDo stack
	        OSVector subclasses = types.classVector(classFor);
	        if (DEBUG) {
		    int cnt = (subclasses == null ? 0 : subclasses.size());
		    debug(this, cnt + " subclasses");
	        }

	        if (subclasses != null) {
	            Enumeration e = subclasses.elements();
	            while (e.hasMoreElements())
		        toDo.push(e.nextElement());
		}

		// add the class itself to stack
	        toDo.push(classFor);
	    }

	    if (!toDo.isEmpty())		//no holders means nothing to do
	        curEnum = enumFor((String)toDo.pop());
	}

	/** 
	 * Return the <code>RepEnum</code> for the given class.
	 */
	private RepEnum enumFor(String classFor) {
	    EntryHolder holder = contents.holderFor(classFor);
	    return holder.contents(txn);
	}

	// purposefully inherit doc comment
	public EntryRep nextRep() {
	    /*
	     * We loop to handle the case where the list of elements for
	     * a type is empty, which means we have to proceed immediately
	     * onward to the next type.
	     */
	    for (;;) {
		if (curEnum == null) {			// already exhausted all
		    if (DEBUG)
			debug(this, "already exhausted");
		    return null;
		}

		EntryRep rep = curEnum.nextRep();	// get next rep
		if (DEBUG) {
		    System.out.println("BasicSpace.AllReps: nextRep: rep= "
			+ rep);
		}
		if (rep != null) {			// if there is one
		    if (DEBUG)
			debug(this, "got one");
		    return rep;				// ... just return it
		}

		// current enum is exhausted, try the next
		if (toDo.isEmpty()) {			// if no more to try
		    if (DEBUG)
			debug(this, "exhausted");
		    curEnum = null;			// exhausted all
		    return null;
		}

		if (DEBUG)
		    debug(this, "moving to next class");


		curEnum = enumFor((String)toDo.pop());	// setup next class
	    }
	}
    }

    /**
     * Implementation of the remote iteration interface.
     */
    private class RemoteIterImpl extends UnicastRemoteObject
	implements RemoteIter
    {
	EntryRep	tmpl;	// the template to match
	RepEnum		enum;	// enum for all potential matches
	boolean		closed;	// has it been closed?

	private final boolean DEBUG = false;

	/** Last batch of <code>EntryRep</code> objects sent to the client. */
	private EntryRep lastBatch[] = null;

	/** ID of last entry in <code>lastBatch</code>. */
	private Object lastId = null;

	/**
	 * Create a <code>RemoteIterImpl</code> object to return all
	 * objects that match the given template.
	 */
	RemoteIterImpl(EntryRep tmpl, Txn txn) throws RemoteException {
	    if (tmpl == null)
		tmpl = EntryRep.matchAnyEntryRep();	      // get all
	    this.tmpl = tmpl;
	    try {
		startTxn();
		enum = new AllReps(tmpl.classFor(), txn);
		if (DEBUG) {
		    System.out.println("BasicSpace.RemoteIterImpl: "
			+ "RemoteIterImpl" + " enum= " + enum);
		}
	    } catch (Exception e) {
		System.out.print("All reps failed:");
		e.printStackTrace(System.out);
	    } finally {
		commitTxn();
	    }
	}

	/**
	 * Utility to set <code>lastBatch</code> and <code>lastId</code>.
	 */
	private void rememberLast(EntryRep[] newLast) {
	    lastBatch = newLast;
	    if (newLast == null) 
		lastId = null;
	    else
		lastId = lastBatch[lastBatch.length-1].getCookie();
	}
	
	// inherit doc comment
	public EntryRep[] nextReps(int max, Object id) {
	    if (closed && id != null && lastId == null)
		// They never got the null
		return null;

	    assertOpen();
	    if (DEBUG)
		debug(this, "nextReps(" + max + "," + id + ")");

	    if (id != null && lastId == null) 
		throw new InternalSpaceException("First call to " +
		    "RemoteIter.next() should have id == null");

	    // Did the client get the result of the last call?
	    if (id != null && lastId != null && !id.equals(lastId)) 
		// No, must have been a RemoteException on the 
		// return, re-send the last batch of entries
		return lastBatch;
		
	    // Otherwise the last call must have gotten through, move forward.

	    if (enum == null) {			       // already exhausted all
		close();			       // consider it closed
		return null;
	    }

	    int limit = Math.min(max, 512);	       // no more than 512
	    EntryRep[] reps = new EntryRep[limit];     // assume we'll fill it
	    if (DEBUG)
		debug(this, "limit = " + limit);
	    try {
		startTxn();

		// look for entries in reps that are at least the right type
		int i = 0;

		while (i < reps.length) {
		    reps[i] = enum.nextRep();

		    if (reps[i] == null) {		// reached the end
			enum = null;	// let the GC work -- it may be some
					// time before this object is GCed
			if (DEBUG)
			    debug(this, "Ran out at " + i);
			if (i == 0) {	// found nothing on this round
			    close();
			    return null;
			}
			EntryRep[] r = new EntryRep[i];
			System.arraycopy(reps, 0, r, 0, r.length);
			rememberLast(r);
			return r;
		    }

		    // if this is a match, add it to the list
		    if (tmpl.matches(reps[i])) {
			reps[i] = (EntryRep) transientOf(reps[i]);
			i++;
		    }
		}
		if (DEBUG)
		    debug(this, "Returning full array");

		rememberLast(reps);
		return reps;	// filled up the array -- return it
	    } finally {
		commitTxn();
	    }
	}

	/**
	 * Delete the entry of the given class and id.  We use class
	 * to get us to the proper <code>EntryHolder</code> efficiently.
	 */
	public void delete(Object id) {
	    assertOpen();
	    try {
		cancel(id);
	    } catch (UnknownLeaseException e) {
		// doesn't matter if cancel fails, since that means
		// the entry is already gone.
	    }
	}

	/**
	 * Close operations on this iterator.
	 */
	public void close() {
	    closed = true;
	    enum = null;
	    tmpl = null;
	    rememberLast(null);
	}

	/**
	 * Throw IllegalStateException if operations happen on closed
	 * iterator.
	 */
	private void assertOpen() throws IllegalStateException {
	    if (closed) {
		if (DEBUG)
		    debug(this, "operation on closed RemoteIterImpl");
		throw new IllegalStateException("closed AdminIterator");
	    }
	}
    }

    /**
     * The iterator that is returned to the client.
     * <code>BasicSpaceIterator</code> uses a <code>RemoteIter</code>
     * reference to talk to the server.  The <code>RemoteIter</code>
     * object is its partner on the server side.
     */
    static class BasicSpaceIterator implements AdminIterator {
	private RemoteIter iter;	// the remote partner

	/** Last set of reps we got from the server */
	private EntryRep[] reps;

	/** 
	 * Index of the next entry in rep to return.  If <code>delete()</code>
	 * we will call <code>iter.delete()</code> next - 1
	 */
	private int next = -1;	

	/**
	 * How many entries to ask for each time we go to the server
	 */
	private int fetchSize;

	// ID of last entry we got from server
	private Object lastId = null;

	/**
	 * Create client side iterator proxy.
	 * @param iter Server side interface for iterator
	 * @param fetchSize Number of entries to ask for when it goes to the
	 *                  server
	 */
	BasicSpaceIterator(RemoteIter iter, int fetchSize) {
	    this.iter = iter;
	    this.fetchSize = fetchSize;
	}

	// purposefully inherit doc comment from supertype
	public Entry next() throws UnusableEntryException, RemoteException {
	    assertOpen();
	    
	    if (next < 0 || next >= reps.length) {
		// Need to get more Entries
		reps = iter.nextReps(fetchSize, lastId);

		if (reps == null) {
		    // Finished
		    close();
		    return null;
		}		

		lastId = reps[reps.length-1].getCookie();

		next = 0;
	    }

	    // This may throw UnusableEntryException, but that's
	    // the right thing
	    return reps[next++].entry();
	}

	// purposefully inherit doc comment from supertype
	public void delete() throws RemoteException {
	    if (next < 0) 
		throw new IllegalStateException("AdminIterator:Can't call " +
		    "delete before calling next()");

	    assertOpen();
	    iter.delete(reps[next-1].getCookie());
	}

	// purposefully inherit doc comment from supertype
	public void close() throws RemoteException {
	    if (iter != null) {
		// always set to null, and then try to use the remote method,
		// which is actually optional
		RemoteIter it = iter;
		iter = null;
		reps = null;
		it.close();
	    }
	}

	/**
	 * Throw <code>IllegalStateException</code> if this iterator
	 * has been closed; otherwise just return.
	 */
	private void assertOpen() throws IllegalStateException {
	    if (iter == null) {
		System.out.println("operation on closed BasicSpaceIterator");
		throw new IllegalStateException("closed AdminIterator");
	    }
	}
    }

    // ------------------------------------------------------------
    //			Transaction stuff
    // ------------------------------------------------------------

    /**
     * This method takes a transactional semantic object
     * <code>baseTr</code> (the transaction object passed to us by the
     * client) and retrieves the associated <code>Txn</code> object
     * (the internal representation for that transaction) from the
     * space's records.  If no <code>Txn</code> object is associated
     * with the given semantic object, a new <code>Txn</code> object is
     * created, associated and recorded.
     */
    Txn enterTxn(Transaction baseTr)
        throws TransactionException, RemoteException
    {
	if (dbgTxns != null)
	    dbgTxns.println("BasicSpace: enterTxn(Transaction)");

	if (baseTr == null)
	    return null;

	ServerTransaction tr = serverTransaction(baseTr);
	if (tr.isNested())
	    throw new CannotNestException("subtransactions not supported");
        try {
	    Txn txn = getTxn(tr);
	    if (dbgTxns != null) {
		dbgTxns.println("txn.getState() = " +
		    TxnConstants.getName(txn.getState()));
	    }
	    if (txn.getState() != ACTIVE)
		throw new CannotJoinException("transaction is not active");
	    return txn;
        } catch (UnknownTransactionException ute) {
	    // if the txn is unknown, to the manager, this throws an exception
	    tr.join((TransactionParticipant) this, crashCount);

            Txn txn = new Txn(tr);
            txns.put(tr, txn);
            return txn;
        }
    }

    /**
     * Look in the table for the <code>Txn</code> object for the given
     * manager/id pair.  If there is one, return it.  If there isn't
     * one, throw <code>UnknownTransactionException</code>.
     */
    private Txn getTxn(TransactionManager mgr, long id)
	throws UnknownTransactionException
    {
	return getTxn(new ServerTransaction(mgr, id));
    }

    /**
     * Look in the table for the <code>Txn</code> object for the given
     * transaction identifier.  If there is one, return it.  If there
     * isn't one, throw <code>UnknownTransactionException</code>.
     */
    private Txn getTxn(Transaction baseTr) throws UnknownTransactionException {
	Txn txn = null;		// object to return

	txn = (Txn) txns.get(baseTr);
	if (txn == null)
	    txn = checkForBrokenTxn(baseTr);
	if (dbgTxns != null)
            dbgTxns.println("BasicSpace: getTxn got Txn=" + txn);

        if (txn == null)
            throw new UnknownTransactionException("unknown transaction");
        return txn;
    }

    /**
     * Remove the given <code>Txn</code> object from the
     * <code>txns</code> table.
     */
    private void removeTxn(Txn txn) {
	txns.remove(txn.transaction(this));
    }

    /**
     * Check to see if the given <code>Transaction</code> object represents
     * a broken transaction, that is, one that used to be known but couldn't
     * be recreated on boot.  If it does, fix up the tables so that it is
     * now properly known.  If it doesn't, return <code>null</code>.  If
     * we can't tell, throw <code>RemoteException</code> so that the
     * invoker will know that it should try again later when maybe it can
     * be resolved.
     */
    private Txn checkForBrokenTxn(Transaction baseTr)
	throws UnknownTransactionException
    {
	if (brokenTxns == null)		// none are broken, so nothing to do
	    return null;

	synchronized (brokenTxns) {	// no changes allowed while we work
	    ServerTransaction tr = serverTransaction(baseTr);
	    Long idObj = new Long(tr.id);
	    OSVector objs = (OSVector) brokenTxns.get(idObj);
	    if (objs == null)	// nothing stored under this id
		return null;
	    /*
	     * Loop through the exceptions.  If any one of them is this
	     * one, return it.  If none of them are, and any throw an
	     * exception, it might be the one we want, but still broken.
	     * If none of them are, but none throw remote exception, then
	     * everything was cleared out as unbroken, so return null.
	     */
	    RemoteException rex = null;
	    for (int i = 0; i < objs.size(); i++) {
		try {
		    Txn txn =
			checkBrokenTxn(objs, i, tr);
		    if (txn != null)
			return txn;
		    else
			i--;	// re-examine this slot: checkBroken changed it
		} catch (RemoteException e) {
		    ;		// this one is still bad, whatever one it is
		}
	    }
	    return null;
	}
    }

    /**
     * Check to see if this particular transaction object that was
     * broken is now fixed.  If so, we clean it out of the table.  Then
     * we see if it is the one we were looking for.  If so, we return
     * the <code>Txn</code> object.  Otherwise we return <code>null</code>.
     * Note that anyone calling this must expect that the tables will change
     * due to the operation of this method -- if this method does not throw
     * an exception, it has removed the element from the table.
     *
     * @throws RemoteException
     *		<code>trObj.get()</code> threw <code>RemoteException</code>
     *
     * @see #checkForBrokenTxn
     */
    private Txn checkBrokenTxn(OSVector objs, int pos, ServerTransaction that)
	throws RemoteException
    {
	Txn txn = (Txn) objs.elementAt(pos);

	// this call will generate a RemoteException if the txn is still broken
	ServerTransaction tr =
	    (ServerTransaction) txn.storableTransaction().get();

	/*
	 * We got it without an exception, so it's not broken -- remove it
	 * from the table of broken transactions, cleaning up the structure.
	 * The code in checkForBrokenTxn is written so this is harmless.
	 */
	objs.removeElementAt(pos);		// remove it from array
	if (objs.size() == 0) {			// if the whole table is empty
	    brokenTxns.remove(new Long(tr.id));	// remove it
	    if (brokenTxns.size() == 0)		// last one out
		brokenTxns = null;
	}
	txns.put(tr, txn);			// add to known transactions

	if (tr.equals(that))			// if this was the desired one
	    return txn;				// ... return it
	else					// otherwise
	    return null;			// ... return null
    }

    /**
     * Add this Transaction/Txn pair to the list of broken transactions.
     */
    void addBrokenTxn(StorableObject trObj, Txn txn, Long id) {
	if (brokenTxns == null) {
	    synchronized (this) {
		if (brokenTxns == null)	// might have been created
		    brokenTxns = new OSHashtable();
	    }
	}
	synchronized (brokenTxns) {
	    OSVector objs = (OSVector) brokenTxns.get(id);
	    if (objs == null) {
		objs = new OSVector();
		brokenTxns.put(id, objs);
	    }
	    objs.addElement(txn);
	}
    }


    /**
     * We assume that each <code>Transaction</code> object sent to the
     * space is actually the <code>ServerTransaction</code> subtype.
     * This method does the downcast, and turns an error into the
     * appropriate <code>UnknownTransactionException</code> exception.
     */
    ServerTransaction serverTransaction(Transaction baseTr)
        throws UnknownTransactionException
    {
        try {
            return (ServerTransaction) baseTr;
        } catch (ClassCastException e) {
            throw new
		UnknownTransactionException("unexpected transaction type");
        }
    }



    //---------------------------------------------------
    // TransactionParticipant methods
    //---------------------------------------------------

    // inherit doc comment
    public int prepare(TransactionManager mgr, long id)
	throws UnknownTransactionException
    {
	//The Space will make sure that for each action
	//under the given transaction, a record is
	//flushed to the log.

	if (dbgTxns != null)
	    dbgTxns.println("BasicSpace: P prepare");
	try {
	    startTxn();
	    Txn txn = getTxn(mgr, id);
	    prepareOp(idFor(txn.transaction(this)), txn.storableTransaction());
	    int result = txn.prepare(this);
	    if (result == NOTCHANGED || result == ABORTED)
		removeTxn(txn);
	    return result;
	} finally {
	    commitTxn();
	}
    }

    // inherit doc comment
    public void commit(TransactionManager mgr, long id)
	throws UnknownTransactionException
    {
	//The changes logged since the last sync point
	//will now be rolled forward.

        if (dbgTxns != null)
            dbgTxns.println("BasicSpace: P commit");

	try {
	    startTxn();
	    Txn txn = getTxn(mgr, id);
	    commitOp(idFor(txn.transaction(this)));
	    txn.commit(this);
	    removeTxn(txn);
	} catch (TransactionException e) {
	    ;	// must have already committed, so just return
	} finally {
	    commitTxn();
	}
    }

    // inherit doc comment
    public void abort(TransactionManager mgr, long id)
	throws UnknownTransactionException
    {
	cancelOp(new Long(id));

        if (dbgTxns != null)
            dbgTxns.println("BasicSpace: P abort");

        try {
	    startTxn();
            //!! Multi-participants not supported
            //!! No early return when not synchronous
            Txn txn = getTxn(mgr, id);
            txn.abort(this);
	    removeTxn(txn);
        } catch (TransactionException e) {
            ;   // must have already aborted, so just return
        } finally {
	    commitTxn();
        }
    }

    // inherit doc comment
    public int prepareAndCommit(TransactionManager mgr, long id)
	throws UnknownTransactionException
    {
	int result = prepare(mgr, id);
	if (result == PREPARED) {
	    commit(mgr, id);
	    result = COMMITTED;
	}

	return result;
    }

    /**
     * Return the ID for the given transaction.
     */
    private static Long idFor(ServerTransaction tr) {
	return (tr == null ? null : new Long(tr.id));
    }

    /**
     * Return the ID for the given transaction.
     */
    private static Long idFor(Transaction tr) {
	return idFor((ServerTransaction) tr);
    }

    // ------------------------------------------------------------
    //			Debug stuff
    // ------------------------------------------------------------

    /**
     * Print out a debug description of <code>obj</code>, followed by
     * the given <code>str</code>.
     */
    private final void debug(Object obj, String str) {
	String name = obj.getClass().getName();
	int dollar = name.indexOf('$');
	if (dollar > 0)
	    name = name.substring(dollar + 1);
	System.out.print(name);
	System.out.print(':');
	System.out.println(str);
    }

    /**
     * Create the service-owned attributes for an Outrigger server.
     */
    protected static Entry[] attributesFor() {
	final Entry info = new ServiceInfo("JavaSpace", 
	    "Sun Microsystems, Inc.", "Sun Microsystems, Inc.",
	    com.sun.jini.constants.VersionConstants.SERVER_VERSION, "", "");
	
	final Entry type = 
	    new com.sun.jini.lookup.entry.BasicServiceType("JavaSpace");

	return new Entry[]{info, type};
    }

    /**
     * Utility method that prints formatted memory usage statistics.
     */
    static void dumpMemory(PrintWriter w, String msg) {
	if(w != null) {
            final Runtime rt = Runtime.getRuntime();
            final long totalMem = rt.totalMemory();
            final long freeMem = rt.freeMemory();
            w.println(msg + "\n <<Memory Stats>>"
		+ " Used:" + (totalMem-freeMem)
		+ " Free:" + freeMem 
		+ " Total:" + totalMem 
		);
	}
    }

    /**
     * Utility method that invokes the Runtime.gc()
     */
    static void doDebugGC(PrintWriter w) {
	if(w != null) {
            w.println("BasicSpace::doDebugGC() ...");
	    Runtime.getRuntime().gc();
	}
    }

}
