/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import java.io.IOException;
import java.net.MalformedURLException;
import net.jini.core.discovery.LookupLocator;
import net.jini.discovery.LookupDiscovery;
import net.jini.discovery.DiscoveryEvent;
import net.jini.discovery.DiscoveryListener;
import net.jini.core.lookup.ServiceRegistrar;
import com.sun.jini.mahout.Locator;
import com.sun.jini.mahout.ParseUtils;


/**
 * A <code>DiscoveryLocator</code> implements the methods needed in
 * order to locate a Jini Lookup service.  Provides a blocking interface
 * in place of <code>LookupDiscovery</code> callback-based interface.
 *
 * @author Ken Arnold
 * @author Gary Holness
 *
 * @see Locator 
 * @see net.jini.discovery.LookupDiscovery
 */
public class DiscoveryLocator extends Locator {
    /** Set if we are using unicast discovery */
    private LookupLocator locator;

    /** 
     * Value of the <code>com.sun.jini.lookup.groups</code> property if set, 
     * otherwise set to {""}.
     */       
    private String groups[] = new String[] {""};

    /**
     * How long to wait for a lookup server to show-up
     */
    private long timeOut;

    /**
     * Create a <code>DiscoveryLocator</code>, once
     * <code>locate</code> is called, that will wait forever for a 
     * lookup to respond.
     * <p>
     * How it will find lookup servers is determined by the values of
     * the <code>com.sun.jini.lookup.groups</code> and
     * <code>com.sun.jini.lookup.locator</code> properties.
     * <ul>
     * <li>The <code>com.sun.jini.lookup.locator</code> must be set
     * to a "jini" URL (jini://host[:port]). If it it set properly,
     * unicast discovery will be used.
     * <p>
     * <li>If the <code>com.sun.jini.lookup.groups</code> is set,
     * multicast discovery will be performed, using the provided list.  
     * The value of this property must be a list of one or more strings
     * separated by commas or white space. If there is any white space
     * in the list the entire list must be enclosed in double quotes.
     * The string "public" will converted to "" (the empty string).
     * If the list consists only of the string "none", the list will be
     * set to the empty list, effectively disabling discovery.
     * <p>
     * </ul>
     * The <code>com.sun.jini.lookup.locator</code> has priority over
     * the <code>com.sun.jini.lookup.groups</code> property.  If
     * neither property is set (or both are set improperly) it will 
     * default to using multicast discovery and the public lookup group.
     */
    public DiscoveryLocator() {
	this(Long.MAX_VALUE);
    }

    /**
     * Create a <code>DiscoveryLocator</code> that will wait for a given
     * timeout (in milliseconds) for a lookup server to respond. See the 
     * no-arg constructor for details on how the lookup service is found.
     * @see #DiscoveryLocator
     */
    public DiscoveryLocator(long timeOut) {
	this.timeOut = timeOut;
	
	final String locatorProp = 
	    System.getProperty("com.sun.jini.lookup.locator");

	if (locatorProp != null) {
	    // Use unicast
	    try {
		locator = new LookupLocator(locatorProp);
		return;
	    } catch (MalformedURLException e) {
		System.err.println("com.sun.jini.lookup.locator was given " +
		     "a malformed URL:" + locatorProp + " trying multicast " +
		     "discovery protocol");
	    }
	}

	// Either the property was not set, or it was but it was bad,
	// either way setup for multicast discovery
	locator = null;	

	// If we are using groups, then check the
	// com.sun.jini.lookup.groups property
	final String groupsProp = 
	    System.getProperty("com.sun.jini.lookup.groups");

	if (groupsProp != null) {
	    try {
		groups = ParseUtils.extractGroups(groupsProp);
	    } catch (IllegalArgumentException e) {
		System.err.println("com.sun.jini.lookup.groups was given " +
		    "a malformed group list, will use the group \"\"");
	    }
	}
    }

    /** 
     * Predicate that returns true if the timeout period for
     * <code>locate</code> has elapsed.  Should be passed the time when
     * the call was started.
     */
    private boolean timeOutP(long start) {
	final long limit = start + timeOut;
	if (limit < 0) 
	    return false;

	if (limit < System.currentTimeMillis())
	    return true;
	
	return false;
    }

    /**
     * Finds a lookup service based on the value of the locator field.
     */
    private Object locatorLocate() {
	ServiceRegistrar tmpobj = null;
	IOException lastException =  null;
	int tryCount = 0;

	if (locator == null)
	    return null;

	final long start = System.currentTimeMillis();

	while (true) {
	    try {
		tryCount++;
		tmpobj = locator.getRegistrar();
		if (tmpobj != null)
		    return tmpobj;
	    } catch (IOException e) {
		lastException = e;
		// Catch the re-try bellow
	    } catch (Exception e) {
		System.out.println("DiscoveryLocator: find: " + 
				   e.getMessage());
		e.printStackTrace();
	    }

	    if (tryCount % 10 == 0) {
		System.err.println("Having trouble finding lookup host:" +
				   locator.getHost() + ":" + 
				   locator.getPort());
		if (lastException != null) {
		    lastException.printStackTrace();
		    lastException = null;
		}
	    }

	    try {
		Thread.sleep(1000);
	    } catch (InterruptedException e) {
	    }

	    if (timeOutP(start)) 
		return null;
	}
    } 

    /**
     * Find a lookup using multicast discovery and the passed group list.
     */
    private Object discoveryLocate(String[] grps) {
	OurDiscoveryListener dl;
	try {
	    dl = new OurDiscoveryListener(grps);
	} catch (IOException e) {
	    System.out.println("Failure setting up discovery:");	    
	    e.printStackTrace();
	    throw new RuntimeException(
	         "IOException in LookupDiscovery setup:"+ e.getMessage());
	}

	try {
	    final long start = System.currentTimeMillis();
	    while (true) {
		try {
		    // Is it already done?
		    if (dl.registrar != null) 
			return dl.registrar;

		    // No, wait
		    synchronized(dl) {		       
			dl.wait(10000);
		    }

		    // Did we timeout or succeed?
		    if (dl.registrar != null)
			// Success!
			return dl.registrar;
		
		    // Failure, complain bitterly and try again
		    System.err.println("Having trouble finding a lookup " +
				       "server in the group(s):");
		    for (int i=0; i<grps.length; i++)
			System.err.println("\t" + grps[i]);
		    System.err.println("...trying again");
		    
		    if (timeOutP(start)) 
			return null;

		} catch (InterruptedException e) {
		    // Ignore
		}
	    }
	} finally {
	    dl.terminate();
	}
    }		    

    /** 
     * A <code>DiscoveryListener</code>, creates its own
     * <code>LookupDiscovery</code> object.  When it finds a lookup
     * server it sticks the proxy in the <code>registrar</code> field
     * and calls <code>notifyAll</code> on itself.
     * <p>
     * Client should call <code>terminate</code> when done.
     */
    private class OurDiscoveryListener implements DiscoveryListener {
	ServiceRegistrar registrar = null;
	private LookupDiscovery discoverer;
	
	/**
	 * Create a <code>OurDiscoveryListener</code> that will look for
	 * a lookup server in the indicated groups
	 */
	private OurDiscoveryListener(String[] grps) throws IOException {
	    discoverer = new LookupDiscovery(grps);
	    discoverer.addDiscoveryListener(this);
	}

	// purposefully inherit doc comment from supertype
	public void discovered(DiscoveryEvent e) {
	    if (registrar != null)
		return;
	    registrar = e.getRegistrars()[0];
	    discoverer.removeDiscoveryListener(this);	    
	    synchronized(this) {
		notifyAll();
	    }
	}

	// purposefully inherit doc comment from supertype
	public void discarded(DiscoveryEvent e) {};

	/**
	 * Kill the discovery thread
	 */
	private void terminate() {
	    discoverer.terminate();
	}
    }

    /**
     * Find a <code>ServiceRegistrar</code>.  This method returns null
     * if no Jini Lookup service can be found.  The semantics are currently 
     * to return the first instance of a lookup service found.  See the
     * constructors for how it finds the lookup server.
     * @see #DiscoveryLocator
     */
    public Object locate() {
	if (locator == null) 
	    return discoveryLocate(groups);
	else
	    return locatorLocate();
    }

    /**
     * Find a <code>ServiceRegistrar</code>.  This method returns null
     * if no Jini Lookup service can be found.  Will use the passed group 
     * list and multicast discovery independent of the values of the
     * <code>com.sun.jini.lookup.locator</code> and
     * <code>com.sun.jini.lookup.groups</code> properties.
     */
    public Object locate(String[] userGroups) {
	return discoveryLocate(userGroups);
    }
}
