/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import com.sun.jini.collection.FastList;
import com.sun.jini.lease.landlord.LeasedResource;

/**
 * This object holds an annotated reference to an
 * <code>EntryRep</code> object.  Currently there is one annotation,
 * which is a hash code for the object that can be used as a
 * quick-reject comparison when scanning through the list.  The handle
 * holds a hash code that is based on the bytes that encode the first
 * <i>N</i> fields, where <i>N</i> is the number of fields in the
 * entry up to a maximum (currently this maximum is 16 (64 bits
 * divided by 4 bits/field), so that 4 is the minimum number of bits
 * per field in the hash).
 * <p>
 * When comparing, the template's own hash is calculated, and also a
 * mask that masks out the hash codes of wildcard fields.  A template
 * will match an entry only if the entry's EntryHandle hash masked with
 * the template's wildcard mask is the same as the template's hash.
 * <p>
 * Care must be taken since the template may be a supertype of the type
 * being searched.  This is why the number of fields in the static
 * methods is passed as an argument, not simply taken from the entry in
 * question.  When a template's hash is being created, its hash value
 * is calculated as if it were of the class being searched, with the
 * subclass's field count.  Any extra fields are assumed to be
 * wildcards.  This means that the template's hash must be recalculated
 * for each subclass it is compared against, but this only happens once
 * per known subclass, and so is probably not onerous.  
 */
// We do not store this data on the EntryRep object itself because it
// is not really part of the client<->JavaSpaces(TM) service protocol -- 
// some implementations of EntryHolder may not choose to use this
// mechanism.  It does add an extra object per EntryRep object in
// those that *do* use it, and so we may want to re-examine this in the
// future.

class EntryHandle extends FastList.Node implements LeaseDesc {
    private EntryRep	rep;		// the rep this handle manages
    private long	hash;		// the content hash for the rep

    /** Print debug messages? */
    private static final boolean DEBUG = false;

    /**
     * Create a new handle, calculating the hash for the object.
     */
    EntryHandle(EntryRep rep) {
	this.rep = rep;
	hash = (rep != null ? hashFor(rep, rep.numFields()) : -1);
    }

    /**
     * Create a new node that is equivalent to that node
     */
    EntryHandle(EntryHandle that) {
	this.rep = that.rep;
	this.hash = that.hash;
    }

    /**
     * Return the handle's <code>EntryRep</code> object.
     */
    EntryRep rep() {
	return rep;
    }

    // inherit doc comment
    public String classFor() {
	return rep.classFor();
    }

    // inherit doc comment
    public LeasedResource getLeasedResource() {
	return rep();
    }

    /**
     * Return this handle's content hash.
     */
    long hash() {
	return hash;
    }

    /**
     * Calculate the hash for a particular entry, assuming the given number
     * of fields.
     *
     * @see #hashFor(EntryRep,int,EntryHandleHashDesc)
     */
    static long hashFor(EntryRep rep, int numFields) {
	return hashFor(rep, numFields, null);
    }

    /**
     * Calculate the hash for a particular entry, assuming the given
     * number of fields, filling in the fields of <code>desc</code>
     * with the relevant values.  <code>desc</code> may be
     * <code>null</code>.  <code>numFields</code> must be >= the number
     * of fields in the <code>rep</code> object (this is not
     * checked).
     *
     * @see #hashFor(EntryRep,int)
     * @see #descFor(EntryRep,int)
     * @see EntryHandleHashDesc
     */
    private static long
	hashFor(EntryRep rep, int numFields, EntryHandleHashDesc hashDesc)
    {
	if (rep == null || numFields == 0)
	    return 0;

	int bitsPerField = Math.max(64 / numFields, 4);	// at least 4 bits
	int fieldsInHash = 64 / bitsPerField;		// max fields used
	long mask =					// per-field bit mask
		    0xffffffffffffffffL >>> (64 - bitsPerField);
	long hash = 0;					// current hash value

	// field counts will be different if rep is a template of a superclass
	long endField = Math.min(fieldsInHash, rep.numFields());

	// set the appropriate rep of the overall hash for the field's hash
	for (int i = 0; i < endField; i++)
	    hash |= (hashForField(rep, i) & mask) << (i * bitsPerField);

	// If someone wants to remember these results, fill 'em in
	if (hashDesc != null) {
	    hashDesc.bitsPerField = bitsPerField;
	    hashDesc.fieldsInHash = fieldsInHash;
	    hashDesc.mask = mask;
	}

	return hash;
    }

    /**
     * Return the template description -- mask and hash.
     *
     * @see EntryHandleTmplDesc
     */
    static EntryHandleTmplDesc descFor(EntryRep tmpl, int numFields) {
	EntryHandleHashDesc hashDesc = new EntryHandleHashDesc();
	EntryHandleTmplDesc tmplDesc = new EntryHandleTmplDesc();

	// Get the hash and the related useful information
	tmplDesc.hash = hashFor(tmpl, numFields, hashDesc);

	// Create the mask to mask away wildcard fields
	for (int i = 0; i < hashDesc.fieldsInHash; i++) {
	    // If this field is one we have a value for, set bits in the mask
	    if (i < tmpl.numFields() && tmpl.value(i) != null)
		tmplDesc.mask |= (hashDesc.mask << (i * hashDesc.bitsPerField));
	}

	// Ensure that the non-value fields are masked out
	tmplDesc.hash &= tmplDesc.mask;

	return tmplDesc;
    }

    /**
     * Return the hash value for a given field, which is then merged in
     * as part of the overall hash for the entry.  The last 32 bytes of
     * the field value are used (or fewer if there are fewer).
     *
     * @see #hashFor(EntryRep,int,EntryHandleHashDesc)
     */
    static long hashForField(EntryRep rep, int field) {
	java.rmi.MarshalledObject v = rep.value(field);
	if (v == null)				// for templates, it's just zero
	    return 0;
	else
	    return v.hashCode();
    }

    public String toString() {
	return "0x" + Long.toHexString(hash) + " [" + rep + "]";
    }

    /**
     * Return <code>true</code> if the operation <code>op</code> under
     * the given transaction (represented by the transaction's manager)
     * can be performed on the object represented by this handle.  This
     * method is overridden by transaction-specific handle types.  For
     * <code>EntryHandle</code> all operations can be performed on an
     * entry that is not in a transaction.
     */
    boolean canPerform(TransactableMgr mgr, int op) {
	if (DEBUG)
	    System.out.println("EntryHandle: canPerform");
	return true;	// all operations are legal on a non-transacted entry
    }

    /**
     * Return <code>true</code> if the given transaction is already
     * known to the entry this handle represents.  This method is
     * overridden by transaction-specific handle types.  For
     * <code>EntryHandle</code>, only the <code>null</code> manager (no
     * transaction) is known.
     */
    boolean knownMgr(TransactableMgr mgr) {
	return (mgr == null);
    }

    /**
     * Return <code>true</code> if the entry this handle represents is
     * being managed within any transaction.  This method is overridden
     * by transaction-specific handle types.  For
     * <code>EntryHandle</code> the answer if <code>false</code>.
     */
    boolean managed() {
	return false;
    }

    /**
     * Add into the collection any transactions that are known to this
     * handle.  This method is overridden by transaction-specific types.
     * For <code>EntryHandle</code>, nothing is added.
     */
    void addTxns(java.util.Collection collection) {
	return;		// nothing to add.
    }

    /**
     * Wait until any transaction managing this entry is stable,
     * i.e., it is not in the middle of preparing, committing, or
     * aborting.  This method is overridden by transaction-specific
     * handle types.  For <code>EntryHandle</code> this returns
     * immediately.
     */
    void waitUntilStable() {
	return;
    }
}
