/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import net.jini.core.entry.*;
import com.sun.jini.lease.landlord.LeasedResource;

import java.util.ArrayList;
import java.lang.reflect.*;
import java.io.*;
import java.rmi.MarshalledObject;
import java.rmi.server.RMIClassLoader;

/**
 * An <code>EntryRep</code> object contains a packaged
 * <code>Entry</code> object for communication between the client and a
 * <code>JavaSpace</code>.
 *
 * @see JavaSpace
 * @see Entry
 */
class EntryRep implements java.io.Serializable, LeasedResource {
    static final long serialVersionUID = 918820158005999990L;

    // These fields are package accessible since this is an internal workhorse
    // for the package.  The number of accessor methods would be large, and
    // would, for the most part, encompass read/write access to each field
    // anyway.

    MarshalledObject[]	values;		// marshalled fields (null == wildcard)
    String[]		superclasses;	// classes IDs of the superclasses
    String		className;	// the class ID of the entry
    String		codebase;	// the codebase for this entry class
    long		id;		// space-relative storage id
    long		expires;	// expiration time

    /**
     * This object represents the passing of a <code>null</code> parameter as a
     * template, which is designed to match any entry.  When a <code>null</code>
     * is passed, it is replaced with this rep, which is then handled specially
     * in a few relevant places.
     */
    private static final EntryRep matchAnyRep = new EntryRep(
	new Entry() {
	    // keeps tests happy
	    static final long serialVersionUID = -4244768995726274609L;
	}, false);

    /**
     * The realClass object is transient because we neither need nor want
     * it reconstituted on the other side.  All we want is to be able to
     * recreate it on the receiving client side.  If it were not transient,
     * not only would an unnecessary object creation occur, but it might
     * force the download of the actual class to the server.
     */
    private transient Class realClass;	// real class of the contained object

    static final long			ID_UNSET = 0xffffffffffffffffL;

    private static final boolean	DEBUG = false;

    /**
     * Set this entry's generic data to be shared with the <code>other</code>
     * object.  Those fields that are object references that will be the same
     * for all objects of the same type are shared this way.
     * <p>
     * Note that <code>codebase</code> is <em>not</em> shared.  If it were,
     * then the failure of one codebase could make all entries inaccessible.
     * Each entry is usable insofar as the codebase under which it was
     * written is usable.
     */
    void shareWith(EntryRep other) {
	className = other.className;
	superclasses = other.superclasses;
    }

    /**
     * Create a serialized form of the entry.  If <code>validate</code> is
     * <code>true</code>, basic sanity checks are done on the class to
     * ensure that it meets the requirements to be an <code>Entry</code>.
     * <code>validate</code> is <code>false</code> only when creating the
     * stand-in object for "match any", which is never actually marshalled
     * on the wire and so which doesn't need to be "proper".
     */
    private EntryRep(Entry entry, boolean validate) {
	int fnum = 0;
	Field[] fields = null;
	try {
	    id = ID_UNSET;

	    realClass = entry.getClass();
	    if (validate)
		ensureValidClass(realClass);
	    className = UID(realClass);
	    codebase = RMIClassLoader.getClassAnnotation(realClass);

	    /*
	     * Build up the per-field and superclass information through
	     * the reflection API.  These variables are declared here so
	     * they can be used in the catch clause.
	     */
	    fields = realClass.getFields();
	    int numFields = fields.length;

	    // collect the usable field values in vals[0..nvals-1]
	    MarshalledObject[] vals = new MarshalledObject[numFields];
	    int nvals = 0;

	    for (fnum = 0; fnum < fields.length; fnum++) {
		if (!usableField(fields[fnum]))
		    continue;

		Object fieldValue = fields[fnum].get(entry);
		if (fieldValue == null)
		    vals[nvals] = null;
		else
		    vals[nvals] = new MarshalledObject(fieldValue);
		nvals++;
	    }

	    // copy the vals with the correct length
	    this.values = new MarshalledObject[nvals];
	    System.arraycopy(vals, 0, this.values, 0, nvals);

	    /*
	     * Loop through the supertypes, making a list of all
	     * superclasses.
	     */
	    ArrayList sclasses = new ArrayList();
	    for (Class c = realClass.getSuperclass();
		 c != Object.class;
		 c = c.getSuperclass())
	    {
		try {
		    sclasses.add(UID(c));
		} catch (ClassCastException cce) {
		    break;	// not Serializable
		}
	    }
	    superclasses = new String[sclasses.size()];
	    System.arraycopy(sclasses.toArray(), 0,
			     superclasses, 0, superclasses.length);
	} catch (IllegalAccessException e) {
	    // must have happened in field loop
	    throw new IllegalArgumentException(
		"Couldn't access field " + fields[fnum]);
	} catch (IOException e) {
	    e.printStackTrace();
	    throw new InternalError("Internal Error: " + e);
	} catch (NullPointerException e) {
	    e.printStackTrace();
	    System.out.println("EntryRep: NullPointerException");
	}
    }

    /**
     * Create a serialized form of the entry with our object's
     * relevant fields set.
     */
    public EntryRep(Entry entry) {
	this(entry, true);
    }

    /** Used to look up no-arg constructors.  */
    private static Class[] noArg = new Class[0];

    /**
     * Ensure that the entry class is valid, that is, that it has appropriate
     * access.  If not, throw <code>IllegalArgumentException</code>.
     */
    private static void ensureValidClass(Class c) {
	boolean ctorOK = false;
	try {
	    if (!Modifier.isPublic(c.getModifiers())) {
		throw new IllegalArgumentException(
		    "entry class " + c.getName() + " not public");
	    }
	    Constructor ctor = c.getConstructor(noArg);
	    ctorOK = Modifier.isPublic(ctor.getModifiers());
	} catch (NoSuchMethodException e) {
	    ctorOK = false;
	} catch (SecurityException e) {
	    ctorOK = false;
	}
	if (!ctorOK) {
	    throw new IllegalArgumentException("entry class " + c.getName() +
		" needs public no-arg constructor");
	}
    }

    /**
     * The <code>EntryRep</code> that marks a ``match any'' request.
     * This is used to represent a <code>null</code> template.
     */
    static EntryRep matchAnyEntryRep() {
	return matchAnyRep;
    }

    /**
     * Return <code>true</code> if the given rep is that ``match any''
     * <code>EntryRep</code>.
     */
    static boolean isMatchAny(EntryRep rep) {
	return matchAnyRep.equals(rep);
    }

    /**
     * Return the class name that is used by the ``match any'' EntryRep
     */
    static String matchAnyClassName() {
	return matchAnyRep.classFor();
    }

    /**
     * Return an <code>Entry</code> object built out of this
     * <code>EntryRep</code> This is used by the client-side proxy to
     * convert the <code>EntryRep</code> it gets from the space server
     * into the actual <code>Entry</code> object it represents.
     *
     * @throws UnusableEntryException
     *		    One or more fields in the entry cannot be
     *		    deserialized, or the class for the entry type
     *		    itself cannot be deserialized.
     */
    Entry entry() throws UnusableEntryException {
	ObjectInputStream objIn = null;
	try {
	    ArrayList badFields = null;
	    ArrayList except = null;

	    realClass = RMIClassLoader.loadClass(codebase, className);
	    Entry entryObj = (Entry) realClass.newInstance();

	    Field[] fields = realClass.getFields();

	    /*
	     * Loop through the fields, ensuring no primitives and
	     * checking for wildcards.
	     */
	    int nvals = 0;		// index into this.values[]
	    for (int i = 0; i < fields.length; i++) {
		Throwable nested = null;
		try {
		    if (!usableField(fields[i]))
			continue;
		    
		    MarshalledObject val = values[nvals++];
		    Object value = (val == null ? null : val.get());
		    fields[i].set(entryObj, value);

		} catch (ArrayIndexOutOfBoundsException e) {
		    // We ran off the end of the values array
		    // which means that we encountered a usable field which
		    // isn't in values[].  It would be confusing to
		    // nest an ArrayIndexOutOfBoundsException, so we'll
		    // record a more descriptive one.
		    nested = new IncompatibleClassChangeError("Field " +
			    fields[i].getName() + " was added to the class " +
			    fields[i].getDeclaringClass().getName() +
			    " after EntryRep construction").fillInStackTrace();
		} catch (IllegalAccessException e) {
		    nested = e;
		} catch (ClassNotFoundException e) {
		    nested = e;
		} catch (java.rmi.RemoteException e) {
		    nested = e;
		}
		if (nested != null) {	// some problem occurred
		    if (badFields == null) {
			badFields = new ArrayList(fields.length);
			except = new ArrayList(fields.length);
		    }
		    badFields.add(fields[i].getName());
		    except.add(nested);
		}
	    }

	    // See if any fields have vanished from the class
	    if (nvals < values.length) {
		throw new UnusableEntryException(
			entryObj,		// should this be null?
			null,			// array of bad-field names
			new Throwable[] {	// array of exceptions
			    new IncompatibleClassChangeError(
				    "A usable field has been removed from " +
				    entryObj.getClass().getName() +
				    " since this EntryRep was created")
			});
	    }

	    // if there were any bad fields, throw the exception
	    if (badFields != null) {
		String[] bf =
		    (String[]) badFields.toArray(new String[badFields.size()]);
		Throwable[] ex =
		    (Throwable[]) except.toArray(new Throwable[bf.length]);
		throw new UnusableEntryException(entryObj, bf, ex);
	    }

	    // everything fine, return the entry
	    return entryObj;
	} catch (InstantiationException e) {
	    /*
	     * If this happens outside a per-field deserialization then
	     * this is a complete failure  The per-field ones are caught
	     * inside the per-field loop.
	     */
	    throw new UnusableEntryException(e);
	} catch (ClassNotFoundException e) {
	    throw new UnusableEntryException(e);	// see above
	} catch (IllegalAccessException e) {
	    throw new UnusableEntryException(e);	// see above
	} catch (RuntimeException e) {
	    throw new UnusableEntryException(e);	// see above
	} catch (IOException ioerr) {
	    ioerr.printStackTrace();
	    throw new InternalError("Internal Error: " + ioerr);
	}
    }

    // inherit doc comment
    public int hashCode() {
	return (int) id ^ className.hashCode();
    }

    /**
     * To be equal, the other object must by an <code>EntryRep</code> for
     * an object of the same class with the same values for each field.
     * This is <em>not</em> a template match -- see <code>matches</code>.
     *
     * @see matches
     */
    public boolean equals(Object o) {
	if (!(o instanceof EntryRep))
	    return false;

	EntryRep other = (EntryRep) o;

	if (!className.equals(other.className))
	    return false;

	if (values.length != other.values.length)
	    return false;

	// faster than checking the bytes themselves
	for (int i = 0; i < values.length; i++) {
	    if ((values[i] == null) != (other.values[i] != null))
		return false;
	}

	// the most expensive test we save for last.  If it was cheaper,
	// it would replace almost all of the above.
	for (int i = 0; i < values.length; i++) {
	    if (values[i] != null && !values[i].equals(other.values[i]))
		return false;
	}

	return true;
    }

    /**
     * Return the unique ID for this class (which is simply it's name).
     */
    static String UID(Class c) {
	return c.getName();
    }

    /**
     * Return <code>true</code> if the field is to be used for the
     * entry.  That is, return <code>true</code> if the field isn't
     * <code>transient</code>, <code>static</code>, or <code>final</code>.
     * @throws IllegalArgumentException
     *			The field is not <code>transient</code>,
     *			<code>static</code>, or <code>final</code>, but
     *			is primitive and hence not a proper field for
     *			an <code>Entry</code>.
     */
    private boolean usableField(Field field) {
	// ignore anything that isn't a public non-static mutable field
	final int ignoreMods =
	    (Modifier.TRANSIENT | Modifier.STATIC | Modifier.FINAL);

	if ((field.getModifiers() & ignoreMods) != 0)
	    return false;

	// if it isn't ignorable, it has to be an object of some kind
	if (field.getType().isPrimitive()) {
	    throw new IllegalArgumentException(
		"primitive fields not allowed in an Entry");
	}

	return true;
    }

    /**
     * Return the ID.
     */
    long id() {
	return id;
    }

    /**
     * Set the ID to <code>newID</code>.
     */
    public void id(long newID) {
	id = newID;
    }

    /**
     * Return the <code>MarshalledObject</code> for the given field.
     */
    public MarshalledObject value(int fieldNum) {
	return values[fieldNum];
    }

    /**
     * Return the number of fields in this kind of entry.
     */
    public int numFields() {
	return values.length;
    }

    /**
     * Return the class name for this entry.
     */
    public String classFor() {
	return className;
    }

    /**
     * Return the array names of superclasses of this entry type.
     */
    public String[] superclasses() {
	return superclasses;
    }

    /**
     * See if the other object matches the template object this
     * represents.  (Note that even though "this" is a template, it may
     * have no wildcards -- a template can have all values.)
     */
    boolean matches(EntryRep other) {
	/*
	 * We use the fact that this is the template in several ways in
	 * the method implementation.  For instance, in this next loop,
	 * we know that the real object must be at least my type, which
	 * means (a) the field types already match, and (b) it has at
	 * least as many fields as the this does.
	 */

	//Note: If this object is the MatchAny template then 
	//      return true (all entries match MatchAny)
	if (EntryRep.isMatchAny(this))
	    return true;

	if (DEBUG)
	    System.out.println("matches: " + values.length + " fields");
	for (int f = 0; f < values.length; f++) {
	    if (values[f] == null) {		// skip wildcards
		if (DEBUG)
		    System.out.println("\tfield "+ f + " is a wildcard");
		continue;
	    }
	    if (!values[f].equals(other.values[f])) {
		if (DEBUG)
		    System.out.println("\field t"+ f + " doesn't match");
		return false;
	    }
	}
	if (DEBUG)
	    System.out.println("return true");
	return true;				// no mismatches, so must be OK
    }

    public String toString() {
	return ("EntryRep[" + className + "]");
    }

    /**
     * Return <code>true</code> if this entry represents an object that
     * is at least the type of the <code>otherClass</code>.
     */
    boolean isAtLeastA(String otherClass) {
        if (className.equals(matchAnyClassName()))
	    return true;

	if (className.equals(otherClass))
	    return true;
	for (int i = 0; i < superclasses.length; i++)
	    if (superclasses[i].equals(otherClass))
		return true;
	return false;
    }

    // Methods required by LeasedResource
    // inherit doc comment from LeasedResource
    public void setExpiration(long newExpiration) {
	expires = newExpiration;
    }

    // inherit doc comment from LeasedResource
    public long getExpiration() {
	return expires;
    }

    // inherit doc comment from LeasedResource
    // We use the Rep ID as the cookie
    public Object getCookie() {
	return new Long(id);
    }
}
