/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import net.jini.space.*;
import net.jini.core.entry.*;
import net.jini.core.transaction.Transaction;
import net.jini.core.event.EventRegistration;
import com.sun.jini.mahout.*;
import com.sun.jini.mahout.binder.Binder;
import com.sun.jini.mahout.binder.BinderFactory;
import com.sun.jini.mahout.binder.BinderSetup;
import com.sun.jini.system.FileSystem;

import java.io.*;
import java.rmi.RemoteException;
import java.rmi.MarshalledObject;
import java.rmi.activation.ActivationException;
import java.rmi.activation.ActivationID;
import java.rmi.activation.ActivationGroup;
import java.rmi.activation.ActivationGroupID;
import java.util.Collection;
import java.util.ArrayList;
import java.util.Enumeration;

import COM.odi.ObjectStore;
import COM.odi.util.OSHashtable;
import COM.odi.util.OSVector;

/**
 * This is the front end of the two-server model.  It operates in transient
 * memory, logging all relevant calls into a log that is later consumed
 * by the back-end space.
 *
 * @see BackEndSpace
 */
public class FrontEndSpace extends TransientSpace {
    /**
     * The log we are writing to.
     */
    private transient LogOutputFile	log;

    /**
     * The path for the directory we use for persistent storage.
     */
    private transient String		path;

    /**
     * The driver for the back end.
     */
    private transient BackEndDriver	driver;

    /**
     * The thread (if any) doing the initialization.
     */
    private transient Thread		initThread;

    /**
     * Our store as a <code>FrontEndStore</code> (saves repeated casting).
     * @serial
     */
    private FrontEndStore		feStore;

    private static final PrintWriter beDbg =
	BasicSpace.debug.getWriter("backEnd");

    private static final PrintWriter initDbg =
	BasicSpace.debug.getWriter("init");

    /**
     * Create a new <code>FrontEndSpace</code> when activated -- this
     * is the ``activation constructor''.  This is the only constructor
     * needed, since the front end is created by registering it as an
     * activatable object, not by invoking any constructor.
     */
    public FrontEndSpace(ActivationID activationID, MarshalledObject data)
	throws RemoteException, IOException, ClassNotFoundException
    {
	super(activationID, data);

	// Persistent data path
	path = (String)data.get();

	// Check path for validity and create directory, if necessary (i.e.
	// the first time we are activated).
	FileSystem.ensureDir(path);
	
	//!! This is unfortunate -- the initialization must happen after we've
	//!! set path in this constructor.  The real solution is to separate
	//!! construction from state initialization -- right now this only
	//!! works because FrontEndSpace is the only subclass of
	//!! TransientSpace.  -arnold
	initialize(newStore());

        System.out.println("BasicSpace: Front end activated");
    }


    // inherit doc comment
    public void bootOp(long time) {
	log.bootOp(time);
    }

    // inherit doc comment
    public void writeOp(EntryRep rep, Long trID) {
	log.writeOp(rep, trID);
    }

    // inherit doc comment
    public void takeOp(long repID, Long trID) {
	log.takeOp(repID, trID);
    }

    // inherit doc comment
    public void registerOp(EntryRep tmpl, NotifyChit chit) {
	// Don't want to log event registration done under transactions
	if (chit instanceof TransactableChit) 
	    return;
	log.registerOp(tmpl, chit);
    }

    // inherit doc comment
    public void notifiedOp(long chitID, long newSequence) {
	log.notifiedOp(chitID, newSequence);
    }

    // inherit doc comment
    public void renewOp(Long cookie, long expiration) {
	log.renewOp(cookie, expiration);
    }

    // inherit doc comment
    public void cancelOp(Long cookie) {
	log.cancelOp(cookie);
    }

    // inherit doc comment
    public void prepareOp(Long trID, StorableObject tr) {
	log.prepareOp(trID, tr);
    }

    // inherit doc comment
    public void commitOp(Long trID) {
	log.commitOp(trID);
    }

    /**
     * Override <code>TransientSpace.newStore</code> to return a new
     * <code>FrontEndStore</code> object.
     */
    Store newStore() {
	feStore = new FrontEndStore();
	return feStore;
    }

    /**
     * Return a proxy that implements the <code>JavaSpaceAdmin</code>
     * interface after making sure that initialization has happened.
     *
     * @see JavaSpaceAdmin
     */
    public Object getAdmin() {
	assertReady();			// make sure we're ready for business
	return super.getAdmin();
    }

    /**
     * Assert that the front end is ready to work, waiting if necessary.
     */
    private void assertReady() {
	feStore.assertReady();
    }

    /**
     * Destroy this space.
     */
    /*
     * We first make sure that things are
     * copacetic by letting initialization complete.  With more subtle
     * thinking we could make this happen during initialization, but
     * that would be complicated, and probably rarely valuable, since
     * destruction is rare, and happening to destroy during
     * initialization would be rarer, and such a mechanism (being
     * subtle) would have subtle bugs.
     */
    public void destroy() {
	try {
	    assertReady();
	} catch (Exception t) {
	    System.err.println("destroy assertReady:");
	    t.printStackTrace(System.err);
	    System.err.println("destroy continuing");
	    // we continue: at least destroy whatever super.destroy *can*
	}
	super.destroy();
    }

    /**
     * Handle storage for the <code>FrontEndSpace</code>.  This is just
     * like transient storage, except that <code>setupStore</code> has
     * to be different -- it must read its initial state from the
     * database.
     */
    private class FrontEndStore extends TransientStore {
	private ODIStore	odiStore;	// the store we use to read from
	private boolean		inODI;		// in the middle of using ODI?

	FrontEndStore() {
	    super(FrontEndSpace.this);
	}

	/**
	 * Initialize our store from the database.
	 */
	public void setupStore() {
	    initThread = new Thread("FrontEndSpaceInitialize") {
		public void run() {
		    long start = 0;
		    if (initDbg != null) {
			start = System.currentTimeMillis();
			initDbg.println("FrontEndStore:run():start @ " + start);
		    }

		    startupBackEnd();		// waits until BackEnd is ready

		    if (beDbg != null)
			beDbg.println("FrontEndSpace:backend started, reading");
		    if (initDbg != null) {
			initDbg.println("FrontEndSpace:backend started, "
			    + "reading @ " + System.currentTimeMillis());
		    }
		    // Thread has its own getName() method
		    odiStore = new ODIStore(FrontEndSpace.this, DEFAULT_NAME,
			path, ObjectStore.RETAIN_UPDATE);
		    inODI = true;
		    odiStore.setupStore();

		    if (initDbg != null) {
			initDbg.println("FrontEndSpace:backend read @ "
			    + System.currentTimeMillis());
		    }
		    OSHashtable pTxns;
		    try {
			odiStore.startTxn();
			ObjectStore.deepFetch(contents);
			ObjectStore.deepFetch(templates);
			ObjectStore.deepFetch(types);
			ObjectStore.deepFetch(txns);
			pTxns = (OSHashtable)
			    odiStore.getRoot(BackEndSpace.PENDING_TXNS_ROOT);
			ObjectStore.deepFetch(pTxns);
		    } finally {
			odiStore.endTxn(true);
			inODI = false;
		    }
		    if (beDbg != null)
			beDbg.println("FrontEndSpace:finished init from DB");

		    if (initDbg != null) {
			initDbg.println("FrontEndSpace:finished init from DB"
			    + " @ " + System.currentTimeMillis());
		    }

		    odiStore.close(true);
		    driver.dbAccessFinished();

		    // setup up the prepared-but-not-committed transactions
		    // We do this after the DB close since it adds data to
		    // the stuff read from 
		    preparedTxns(pTxns);

		    // create the log file (we're ready for it now)
		    try {
			log = new LogOutputFile(
			    new File(path, DEFAULT_NAME).getAbsolutePath());
		    } catch (IOException e) {
			e.printStackTrace();
			System.err.println("FrontEnd: log creation failed");
			throw new InternalSpaceException(
			    "FrontEnd: log creation failed", e);
		    }
		    log.observable().addObserver(driver);
		    bootOp(System.currentTimeMillis());

		    if (initDbg != null) {
			initDbg.println("FrontEndSpace:allocating ID @ "
			    + System.currentTimeMillis());
		    }

		    /*
		     * With SecureRandom, the first ID requires
		     * generation of a secure seed, which can take
		     * several seconds.  We do it here so it doesn't
		     * affect the first call's time.  (I tried doing
		     * this in a separate thread so some of the startup
		     * would occur during the roundtrip back the
		     * client, but it didn't help much and this is
		     * simpler.)
		     */
		    nextID();

		    if (initDbg != null) {
			long end = System.currentTimeMillis();
			initDbg.println("FrontEndSpace:finished init @ "
			    + end + ", elapsed " + (end - start));
		    }
		}
	    };
	    initThread.start();
	}

	/**
	 * Assert that the store is ready for use.
	 */
	synchronized void assertReady() {
	    try {
		if (initThread != null) {
		    long start = 0;
		    if (initDbg != null) {
			start = System.currentTimeMillis();
			initDbg.println("FrontEndSpace:assertReady joining @ "
			    + start);
		    }
		    initThread.join();
		    if (initDbg != null) {
			long end = System.currentTimeMillis();
			initDbg.println("FrontEndSpace:assertReady joined  @ "
			    + end + ", elapsed " + (end - start));
		    }
		    initThread = null;
		}
	    } catch (InterruptedException e) {
		e.printStackTrace();
		throw new InternalSpaceException("initialization interrupted",
						 e);
	    }
	}

	/**
	 * If we're using ODI, ask our <code>ODIStore</code> to
	 * <code>startTxn</code>; otherwise do what
	 * <code>TransientStore</code> does.
	 */
	public void startTxn() {
	    // if we're not in the intializer thread, better make sure things
	    // are ready
	    if (initThread != null && Thread.currentThread() != initThread)
		assertReady();

	    if (inODI)
		odiStore.startTxn();
	    else
		super.startTxn();
	}

	/**
	 * If we're using ODI, ask our <code>ODIStore</code> to
	 * <code>endTxn</code>; otherwise do what
	 * <code>TransientStore</code> does.
	 */
	public void endTxn(boolean commit) {
	    if (inODI)
		odiStore.endTxn(commit);
	    else
		super.endTxn(commit);
	}

	/**
	 * Destroy the back end and the database.
	 */
	public void destroy() {
	    // Important to destroy the back-end before destroying the
	    // database, which also deletes all of the files, so that
	    // no one has any file handles left open...on the win32
	    // and MVFS files systems a file with open handles can not
	    // be deleted.
	    try {
		if (driver != null) {
		    driver.interrupt();
		    driver.destroyBackEnd();
		}
	    } catch (Exception e) {
		System.err.println("destroying back end:");
		e.printStackTrace();
		System.err.println("destroy continuing");
	    }

	    try {
		if (log != null)
		    log.close();
	    } catch (Exception e) {
		System.err.println("destroying log:");
		e.printStackTrace();
		System.err.println("destroy continuing");
	    }

	    try {
		if (odiStore != null)
		    odiStore.destroy();
	    } catch (Exception e) {
		System.err.println("destroying odiStore:");
		e.printStackTrace();
		System.err.println("destroy continuing");
	    }

	    super.destroy();
	}
    };

    /**
     * Start the back end.  This is done via the <code>driver</code>
     * object, which this method creates.
     *
     * @see driver
     */
    void startupBackEnd() {
	driver = new BackEndDriver(DEFAULT_NAME, path);
	driver.start();					// start it running
	driver.waitUntilReady();
    }

    /**
     * Make the modifications necessary to handle any pending
     * transactions that have been prepared but not yet committed when
     * the back end is finished reading its logs.  We make sure to monitor
     * all the pending transactions.
     *
     * @see #setupPreparedTxn
     */
    void preparedTxns(OSHashtable txnIDs) {
	if (dbgTxns != null)
	    dbgTxns.println("looking for prepared transactions");
	Enumeration e = txnIDs.elements();
	Collection preparedTxns = new ArrayList();
	while (e.hasMoreElements()) {
	    PendingTxn pt = (PendingTxn) e.nextElement();
	    if (pt.state() == PREPARED)
		preparedTxns.add(setupPreparedTxn(pt));
	}
	if (preparedTxns.size() > 0)
	    monitor(preparedTxns);
    }

    /**
     * Make the modifications necessary to handle this pending
     * transaction.  This requires setting up the
     * <code>TxnEntryHandle</code> objects and the like for all
     * operations under the transaction.  If the original transaction
     * object cannot be deserialized, the manager is unavailable and so
     * the transaction is broken.  This is handled by adding it to the
     * table of broken transactions handled in
     * <code>BasicSpace</code>.
     *
     * @see #preparedTxns
     * @see BasicSpace#getTxn
     */
    Txn setupPreparedTxn(PendingTxn pt) {
	if (dbgTxns != null)
	    dbgTxns.println("setting up prepared Txn " + pt.id());

	OSVector ops = pt.ops();

	// First create the Txn object that represents this transaction
	Txn txn;
	try {
	    Transaction tr = pt.transaction();
            txn = new Txn(tr);
	    txns.put(tr, txn);
	} catch (RemoteException e) {
	    // couldn't unpack transaction object -- it's broken
	    StorableObject trObj = pt.storableTransaction();
	    txn = new Txn(trObj);
	    addBrokenTxn(trObj, txn, pt.id());
	    if (dbgTxns != null)
		dbgTxns.println(pt.id() + " is broken");
	}
	if (dbgTxns != null)
	    dbgTxns.println("re-preparing transaction " + pt.id());

	// now make its ops represented in the space
	for (int i = 0; i < ops.size(); i++) {
	    PendingTxn.PendingOp op = (PendingTxn.PendingOp) ops.get(i);
	    if (op instanceof PendingTxn.WriteOp)
		addPreparedWrite((PendingTxn.WriteOp) op, txn);
	    else if (op instanceof PendingTxn.TakeOp)
		addPreparedTake((PendingTxn.TakeOp) op, txn);
	    else {
		throw new InternalSpaceException("unknown operation type: " +
		    op.getClass().getName());
	    }
	}
	txn.forceState(PREPARED);
	return txn;
    }

    /**
     * Make the modifications necessary to remember a pending write.
     *
     * @see BasicSpace#addWrittenRep
     */
    private void addPreparedWrite(PendingTxn.WriteOp op, Txn txn) {
	if (dbgTxns != null)
	    dbgTxns.println("adding prepared write " + op);
	addWrittenRep(op.rep, txn, 0);
    }

    /**
     * Make the modifications necessary to remember a pending take.
     */
    private void addPreparedTake(PendingTxn.TakeOp op, Txn txn) {
	if (dbgTxns != null)
	    dbgTxns.println("adding prepared take " + op);
	EntryRep rep = (EntryRep) contents.getLeasedResource(new Long(op.id));
	EntryHolder holder = contents.holderFor(rep);
	try {
	    // use the rep as its own template -- guarantee match
	    if (holder.match(rep, rep, txn, true, 0, null) == null) {
		throw new InternalSpaceException(
		    "match not found while re-preparing take");
	    }
	} catch (TransactionConflictException e) {
	    throw new InternalSpaceException(
		"conflict while re-preparing take", e);
	}
    }
    /**
     * Look to see if we have already created a binder, if not 
     * create one and log it to file system
     */
    Binder createBinder() 
        throws IOException, ClassNotFoundException, ActivationException
    {
        // have we already created a binder in a previous 
        // VM?
        final File logFile = new File(path, DEFAULT_NAME + ".binder");
        BinderFactory factory;

        if (logFile.exists()) {
            // Get binder factory from log
            final FileInputStream fis = new FileInputStream(logFile);
            final ObjectInputStream ois = new ObjectInputStream(fis);
            factory = (BinderFactory)ois.readObject();
            ois.close();
        } else {
            // Create the binder builder, get the factory from it 
            // and write it to the file system
            final ActivationGroupID gid = ActivationGroup.currentGroupID();
            final BinderSetup builder =
                new com.sun.jini.mahout.binder.LookupBinderSetup(
                    null, path, null, gid, attributesFor());
            builder.setService(getSpaceProxy());

            final FileOutputStream fos = new FileOutputStream(logFile);
            final ObjectOutputStream oos = new ObjectOutputStream(fos);
            factory = builder.getBinderFactory();

            oos.writeObject(factory);
            oos.close();
        }
        
        return factory.get();
    }


}
