/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import java.io.*;
import java.util.*;

/**
 * The base class for the logging file classes.  This class provides
 * the common functionality, but you should not instantiate it.
 * <p>
 * (Note -- I use <code>protected</code> here as an advisory notice.
 * Clearly, since this is package code, all classes in the package have
 * access, but fields marked <code>protected</code> are expected to be
 * used only by subclasses.  Use good taste.)
 *
 * @author Ken Arnold
 *
 * @see LogOutputFile
 * @see LogInputFile
 */
class LogFile {
    /**
     * The directory in which the log files live.
     */
    protected File baseDir;

    /**
     * The base part of the file name (e.g., <code>"log."</code> for
     * <code>"log.0"</code>, <code>"log.1"</code>, ...)
     */
    protected String baseFile;

    /**
     * The current path to the log (e.g., <code>"/var/adm/log.3"</code>)
     */
    protected File currentPath;

    /**
     * The version of the log stream (the highest one known).
     */
    protected static final byte LOG_VERSION = 0;

    /** A log entry that records a <code>write</code>. */
    protected static final byte BOOT_OP		= 1;
    /** A log entry that records a <code>write</code>. */
    protected static final byte WRITE_OP	= 2;
    /** A log entry that records a <code>take</code>. */
    protected static final byte TAKE_OP		= 3;
    /** A log entry that records a <code>notify</code>. */
    protected static final byte REGISTER_OP	= 4;
    /** A log entry that records a <code>notify</code>. */
    protected static final byte RENEW_OP	= 5;
    /** A log entry that records a notification and new sequence number. */
    protected static final byte NOTIFIED_OP	= 6;
    /** A log entry that records a <code>cancel</code>. */
    protected static final byte CANCEL_OP	= 7;
    /** A log entry that records a transaction <code>prepare</code>. */
    protected static final byte PREPARE_OP	= 8;
    /** A log entry that records a transaction <code>commit</code>.
     * (Abort is logged as a "cancel".) */
    protected static final byte COMMIT_OP	= 9;

    static final PrintWriter logDebug = BasicSpace.debug.getWriter("logs");

    /**
     * Create a log file with the given base directory, base file name
     * within the directory, and current path.  This is intended only for
     * when you are sure of the exact values -- no modifications are
     * made to the values.
     */
    protected LogFile(File baseDir, String baseFile, File currentPath) {
	this.baseDir = baseDir;
	this.baseFile = baseFile;
	this.currentPath = currentPath;
    }

    /**
     * Create a log file from the given template.  If
     * <code>basePath</code> has a directory component, it is used as
     * the base directory.  Otherwise the base directory is
     * <code>"."</code>.  If <code>basePath</code> names a directory,
     * the base name will be <code>""</code>.  Otherwise the file
     * component is used as the base, with a "." added at the end if it
     * is not already present.
     */
    protected LogFile(String basePath) throws IOException {
	baseDir = new File(basePath);
	if (baseDir.isDirectory()) {
	    baseFile = "";
	} else {
	    baseFile = baseDir.getName();
	    String pname = baseDir.getParent();
	    if (pname == null)
		pname = ".";
	    baseDir = new File(pname);
	    if (baseFile.charAt(baseFile.length() - 1) != '.')
		baseFile += ".";
	}
    }

    /**
     * Fill in a list of existing matching log files, oldest to newest,
     * returning the highest number used as a suffix, or -1 if
     * no files were found.  If two files have the same time, they are
     * sorted by the numeric value of the suffix.
     */
    int existingLogs(Collection files) {
	if (logDebug != null)
	    logDebug.println("scanning " + baseDir + " for " + baseFile);

	String[] inDir = baseDir.list();		// directory contents
	TreeMap found = new TreeMap();			// the files we've found
	int highest = -1;				// largest # suffix seen

    fileLoop:
	for (int f = 0; f < inDir.length; f++) {

	    String name = inDir[f];

	    if (logDebug != null)
		logDebug.println("    checking " + name);

	    if (!name.startsWith(baseFile))		// is it one of ours?
		continue;

	    // ensure that there is a numerical suffix
	    int num;
	    try {
		num = Integer.parseInt(name.substring(baseFile.length()));
		if (num > highest)			// keep track of highest
		    highest = num;
	    } catch (NumberFormatException e) {
		continue fileLoop;			// can't be one of ours
	    }

	    found.put(new Integer(num), new File(baseDir, name));
	}

	files.addAll(found.values());
	if (logDebug != null) {
	    logDebug.println("returning " + files.size() + " files");
	    Iterator it = files.iterator();
	    while (it.hasNext())
		logDebug.println(it.next());
	}

	return highest;
    }

    /**
     * Destroy all log files associated with this stream.
     */
    void destroy() throws IOException {
	if (logDebug != null)
	    logDebug.println("destroy");

	ArrayList files = new ArrayList();
	final ArrayList errors = new ArrayList();
	existingLogs(files);
	for (int i = 0; i < files.size(); i++) {
	    File log = (File) files.get(i);
	    try {
		if (!log.delete())
		    errors.add(new IOException("couldn't delete " + log));
	    } catch (SecurityException e) {
		errors.add(e);
	    }
	}
	if (errors.size() == 0)
	    return;
	throw new IOException("Removing logs") {
	    // keeps tests happy
	    static final long serialVersionUID = 5708956268216175935L;

	    public void printStackTrace(PrintStream out) {
		super.printStackTrace(out);
		for (int i = 0; i < errors.size(); i++) {
		    out.println("nested exception " + i + ":");
		    ((Exception) errors.get(i)).printStackTrace(out);
		}
	    }
	    public void printStackTrace(PrintWriter out) {
		super.printStackTrace(out);
		for (int i = 0; i < errors.size(); i++) {
		    out.println("nested exception " + i + ":");
		    ((Exception) errors.get(i)).printStackTrace(out);
		}
	    }
	};
    }

    public String toString() {
	return currentPath.toString();
    }
}
