/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import net.jini.space.*;
import net.jini.core.transaction.Transaction;
import net.jini.core.transaction.server.ServerTransaction;
import net.jini.core.event.EventRegistration;
import com.sun.jini.mahout.StorableObject;

import java.util.*;
import java.rmi.MarshalledObject;

import java.io.*;

/**
 * A class to help you read log files created by <code>LogOutputFile</code>.
 *
 * @author Ken Arnold
 *
 * @see LogOutputFile
 */
class LogInputFile extends LogFile {
    private FileInputStream fileIn;	// the actual file input stream
    private ObjectInputStream in;	// read objects from fileIn
    private HashMap txns;		// txns known in current file

    /**
     * Return an <code>Iterator</code> that will loop through all
     * the logs that match the given <code>basePath</code> pattern,
     * interpreted as described in the matching <code>LogStream</code>
     * constructor.  If <code>returnAll</code> is <code>false</code>,
     * the most recent file will be left off the list.  This would be
     * the proper value for an ongoing poll looking for completed log
     * files.  You would specify <code>true</code> during recovery,
     * when all existing logs should be committed because no new ones
     * are currently being created
     *
     * @see java.util.Iterator
     * @see LogStream#LogStream(String)
     */
    static Iterator logs(String basePath, boolean returnAll)
	throws IOException
    {
	LogFile lf = new LogFile(basePath); // an object to represent the path
	ArrayList inDir = new ArrayList();
	lf.existingLogs(inDir);

	// strip off most recent if we're not trying to read them all
	if (!returnAll && inDir.size() > 0)
	    inDir.remove(inDir.size() - 1);

	return new LogInputFileIterator(inDir, lf);
    }

    /**
     * The implementation of <code>Iterator</code> returned by
     * <code>LogInputStream.logs</code>.  The <code>next</code> method
     * occasionally returns <code>null</code>.
     *
     * @see LogInputFileIterator#next
     */
    private static class LogInputFileIterator implements Iterator {
	private LogFile		baseLogFile;
	private Iterator	fileList;

	/**
	 * Create a new <code>LogInputFileIterator</code> object
	 * for the given list.
	 */
	LogInputFileIterator(Collection files, LogFile baseLogFile) {
	    this.baseLogFile = baseLogFile;
	    fileList = files.iterator();
	}

	public boolean hasNext() {
	    return fileList.hasNext();
	}

	/**
	 * Return the next <code>File</code> object, or
	 * <code>null</code>.  You will get <code>null</code> when the
	 * file existed at the time of listing, but no longer exists
	 * when the iterator gets to it.  For example, if a process is
	 * consuming all completed logs, the listing might find a log,
	 * but that process may have consumed and removed it by the
	 * time you invoke <code>next</code>, so you will get a
	 * <code>null</code>.
	 */
	public Object next() {
	    File file = (File) fileList.next();
	    try {
		return new LogInputFile(baseLogFile, file);
	    } catch (IOException e) {
		file.delete();	// file is malformed -- remove it
		return null;	// can't throw any reasonable exception,
				// so signal the problem with a null
	    }
	}

	/**
	 * Remove the <code>File</code> object returned by the iterator
	 * from the list.  This does <em>not</em> remove the file
	 * itself.
	 */
	public void remove() {
	    fileList.remove();
	}
    }

    /**
     * Create a new <code>LogInputFile</code>.
     * <p>
     * <b>Note:</b> Don't invoke this.  This is needed by the
     * enumeration returned by <code>logs</code>, which is how you
     * should be getting <code>LogInputFile</code> objects.  When
     * nested classes arrive, this constructor can be properly
     * protected.
     * @see logs
     */
    LogInputFile(LogFile desc, File path) throws IOException {
	super(desc.baseDir, desc.baseFile, path);
	fileIn = new FileInputStream(path);
	in = new ObjectInputStream(fileIn);
	int fileVer = in.readByte();
	txns = new HashMap();
	if (fileVer != LOG_VERSION) {
	    System.err.println("unsupported log version: " + fileVer);
	    System.exit(-1);
	}
    }

    /**
     * Consume the input file, invoking the appropriate operations on
     * the given object.
     */
    synchronized void consume(LogOps opOn) {
	try {
	    EntryRep rep;
	    long id;
	    Object obj;
	    Long cookie;
	    for (;;) {
		int op = in.readByte();
		switch (op) {
		  case -1:
		    return;	// got to EOF without error

		  case BOOT_OP:
		    long time = in.readLong();
		    opOn.bootOp(time);
		    break;

		  case WRITE_OP:
		    rep = (EntryRep) in.readObject();
		    cookie = (Long) in.readObject();
		    opOn.writeOp(rep, cookie);
		    break;

		  case TAKE_OP:
		    id = in.readLong();
		    cookie = (Long) in.readObject();
		    opOn.takeOp(id, cookie);
		    break;

		  case REGISTER_OP:
		    rep = (EntryRep) in.readObject();
		    NotifyChit chit = (NotifyChit) in.readObject();
		    opOn.registerOp(rep, chit);
		    break;

		  case NOTIFIED_OP:
		    id = in.readLong();
		    long newSeq = in.readLong();
		    opOn.notifiedOp(id, newSeq);
		    break;

		  case RENEW_OP:
		    cookie = new Long(in.readLong());
		    long expires = in.readLong();
		    opOn.renewOp(cookie, expires);
		    break;

		  case CANCEL_OP:
		    cookie = new Long(in.readLong());
		    opOn.cancelOp(cookie);
		    break;

		  case PREPARE_OP:
		    cookie = (Long) in.readObject();
		    StorableObject tr = (StorableObject) in.readObject();
		    opOn.prepareOp(cookie, tr);
		    break;

		  case COMMIT_OP:
		    cookie = (Long) in.readObject();
		    opOn.commitOp(cookie);
		    break;

		  default:
		    if (logDebug != null) {
			logDebug.println("Invalid operation " + op + " in " +
			    " log " + currentPath + ": stop consumption");
		    }
		    return;
		}
	    }
	} catch (EOFException e) {
	    if (logDebug != null) {
		logDebug.println(e + " in log " + currentPath
		    + ": stop consumption");
		e.printStackTrace(logDebug);
	    }
	    ;	// that's that
	} catch (IOException e) {
	    e.printStackTrace();
	    throw new InternalSpaceException("unexpected IO error in log", e);
	} catch (ClassNotFoundException e) {
	    e.printStackTrace();
	    throw new InternalSpaceException("unexpected class?", e);
	}
    }

    /**
     * This log has been successfully drained, and committed -- it can be
     * removed.
     *
     * @see close
     */
    void finished() {
	try {
	    close();
	} catch (IOException e) {
	    ; // don't care
	}
	if (logDebug == null)
	    currentPath.delete();
	else {
	    String bakFile = currentPath.toString() + "BAK";
	    logDebug.println("finished with " + currentPath);
	    currentPath.renameTo(new File(bakFile));
	}
    }

    /**
     * Close the log, but don't remove it.
     */
    void close() throws IOException {
	in.close();
    }
}
