/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import net.jini.core.transaction.Transaction;
import net.jini.core.transaction.server.ServerTransaction;
import net.jini.core.event.EventRegistration;
import net.jini.space.InternalSpaceException;
import com.sun.jini.mahout.StorableObject;

import java.io.*;
import java.util.*;
import java.rmi.MarshalledObject;

/**
 * A class to write a log file, to be read later by
 * <code>LogInputFile</code>.  Each operation on the file is forced to
 * disk, so when the operation logging function returns, the data is
 * committed to the log in a recoverable way.
 * <p>
 * <code>LogOutputFile</code> cannot extend <code>Observable</code>
 * because it must extend <code>LogFile</code> (clearly
 * <code>Observable</code> should have been an interface).  It acts as
 * an <code>Observable</code> by having a method that returns its
 * "observable part", which is an object that reports observable
 * events.  Right now the only observable event is the switching to a
 * new physical file when the current one becomes full.
 *
 * @author Ken Arnold
 *
 * @see LogInputFile
 * @see java.util.Observable
 */
class LogOutputFile extends LogFile implements LogOps {
    private FileOutputStream fileOut;		// the actual file output
    private ObjectOutputStream out;		// objects written to fileOut
    private int suffix;				// the current suffix number
    private int opCnt;				// number of ops on current file
    private int maxOps;				// max ops to allow in file
    private HashMap txns;			// txns known in current file
    private Observable observable;		// handle Observer/Observable

    private static final int MAX_OPS = getMaxOps();// maximum ops in file
    private static final boolean DEBUG = false;

    /**
     * Create a <code>LogOutputFile</code> object that will stream output
     * to a series of files described by <code>basePath</code>, as interpreted
     * by the relevant <code>LogFile</code> constructor.  When the file becomes
     * full (the maximum number of operations is reached), the file is closed
     * and a new file with the next highest suffix is created.  The
     * <code>Observable</code> notification for this event passes a
     * <code>File</code> argument for the filled file as the argument
     * to <code>Observer</code>.
     *
     * @see LogStream#LogStream(String)
     * @see maxOps()
     * @see maxOps(int)
     * @see observable()
     */
    LogOutputFile(String basePath) throws IOException {
	super(basePath);
	ArrayList inDir = new ArrayList();
	suffix = existingLogs(inDir);
	maxOps = MAX_OPS;
	txns = new HashMap();
	nextPath();
    }

    /**
     * Return an <code>Observable</code> object that represents this object
     * in the Observer/Observable pattern.
     *
     * @see java.util.Observer
     */
    Observable observable() {
	if (observable == null) {		// defer allocation until needed
	    observable = new Observable() {	// we only use this if changed
		public void notifyObservers() {
		    setChanged();
		    super.notifyObservers();
		}
		public void notifyObservers(Object arg) {
		    setChanged();
		    super.notifyObservers(arg);
		}
	    };
	}
	return observable;
    }

    /**
     * Return the maximum number of operations allowed on this file.
     */
    int maxOps() {
	return maxOps;
    }

    /**
     * Set the maximum number of operations allowed on this file.
     */
    void maxOps(int newMax) {
	maxOps = newMax;
    }

    /**
     * Switch this over to the next path in the list
     */
    private void nextPath() throws IOException {
	File oldPath = null;
	if (out != null) {
	    out.flush();		// push all the bytes towards the file
	    fileOut.getFD().sync();	// force the bytes to disk
	    out.close();		// close the stream
	    oldPath = currentPath;
	}

	suffix++;			// go to next suffix
	currentPath = new File(baseDir, baseFile + suffix);
	if (DEBUG) {
	    System.out.println("starting log " + currentPath +
			       " after " + opCnt + " operations");
	}
	fileOut = new FileOutputStream(currentPath);
	out = new ObjectOutputStream(fileOut);
	out.write(LOG_VERSION);
	opCnt = 0;
	txns.clear();			// forget previous txns

	/*
	 * Tell anyone about the old path.  This is done after the
	 * new one is created so that the old path can be known not
	 * to be the newest (because something newer is there).
	 */
	if (observable != null && oldPath != null)
	    observable.notifyObservers(oldPath);
    }

    /**
     * Log a server boot.
     */
    public synchronized void bootOp(long time) {
	try {
	    out.writeByte(BOOT_OP);
	    out.writeLong(time);
	    flush();
	} catch (IOException e) {
	    failed(e);
	}
    }

    /**
     * Log a <code>write</code> operation.
     */
    public synchronized void writeOp(EntryRep entry, Long trID) {
	try {
	    out.writeByte(WRITE_OP);
	    out.writeObject(entry);
	    out.writeObject(trID);
	    flush();
	} catch (IOException e) {
	    failed(e);
	}
    }

    /**
     * Log a <code>take</code> operation.
     */
    public synchronized void takeOp(long id, Long trID) {
	try {
	    out.writeByte(TAKE_OP);
	    out.writeLong(id);
	    out.writeObject(trID);
	    flush();
	} catch (IOException e) {
	    failed(e);
	}
    }

    /**
     * Log a <code>notify</code> operation.
     */
    public synchronized void registerOp(EntryRep tmpl, NotifyChit chit) {
	try {
	    out.writeByte(REGISTER_OP);
	    out.writeObject(tmpl);
	    out.writeObject(chit);
	    flush();
	} catch (IOException e) {
	    failed(e);
	}
    }

    /**
     * Log a notification, changing the sequence number.
     */
    public synchronized void notifiedOp(long chitID, long newSeq) {
	try {
	    out.writeByte(NOTIFIED_OP);
	    out.writeLong(chitID);
	    out.writeLong(newSeq);
	    flush();
	} catch (IOException e) {
	    failed(e);
	}
    }

    /**
     * Log a <code>renew</code> operation.
     */
    public synchronized void renewOp(Long id, long expiration) {
	try {
	    out.writeByte(RENEW_OP);
	    out.writeLong(id.longValue());
	    out.writeLong(expiration);
	    flush();
	} catch (IOException e) {
	    failed(e);
	}
    }

    /**
     * Log a <code>cancel</code> operation.
     */
    public synchronized void cancelOp(Long id) {
	try {
	    out.writeByte(CANCEL_OP);
	    out.writeLong(id.longValue());
	    flush();
	} catch (IOException e) {
	    failed(e);
	}
    }

    /**
     * Log a transaction prepare.
     */
    public synchronized void prepareOp(Long trID, StorableObject tr) {
	try {
	    out.writeByte(PREPARE_OP);
	    out.writeObject(trID);
	    out.writeObject(tr);
	    flush();
	} catch (IOException e) {
	    failed(e);
	}
    }

    /**
     * Log a transaction commit.
     */
    public synchronized void commitOp(Long trID) {
	try {
	    out.writeByte(COMMIT_OP);
	    out.writeObject(trID);
	    flush();
	} catch (IOException e) {
	    failed(e);
	}
    }

    /**
     * Flush the current output after an operation.  If the number of operations
     * is exceeded, shift over to the next path.
     */
    private synchronized void flush() throws IOException {
	out.flush();
	fileOut.getFD().sync();
	if (++opCnt >= maxOps)
	    nextPath();
	else
	    out.reset();	// not critical to flush this
    }

    private void failed(Exception e) throws InternalSpaceException {
	System.err.print("logging: ");
	e.printStackTrace();
	throw new InternalSpaceException("logging failed", e);
    }

    private static int getMaxOps() {
	int ops = 1000;
	try {
	    String prop = System.getProperty("com.sun.jini.outrigger.maxOps");
	    if (logDebug != null)
		logDebug.println("maxOps property = " + prop);
	    if (prop != null)
		ops = Integer.parseInt(prop);
	} catch (SecurityException e) {
	    ;	// return the original if we can't access properties
	} catch (NumberFormatException e) {
	    ;	// return the original if the number is ugly
	}
	if (logDebug != null)
	    logDebug.println("maxOps = " + ops);
	return ops;
    }

    /**
     * A test program.  Usage:
     *
     * <pre>
     *	    java LogOutputFile basePath opCnt [maxOpsPerFile]
     * </pre>
     */
    public static void main(String[] args) throws Throwable {
	LogOutputFile out = new LogOutputFile(args[0]);

	if (args.length > 2)
	    out.maxOps(Integer.parseInt(args[2]));

	for (int i = Integer.parseInt(args[1]); i > 0; i--)
	    out.takeOp(i, null);
    }

    /**
     * Close the log, but don't remove it.
     */
    void close() throws IOException {
	out.close();
    }
}
