/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import net.jini.space.*;
import net.jini.core.event.*;
import net.jini.core.lease.*;
import net.jini.core.transaction.*;
import com.sun.jini.collection.FastList;
import com.sun.jini.lease.landlord.LeasedResource;
import com.sun.jini.thread.TaskManager;
import com.sun.jini.thread.RetryTask;

import java.rmi.*;
import java.util.*;

/**
 * The notifier thread.  This thread is responsible for notifying
 * objects for which interest has been registered.  It operates in
 * transient space as much as possible.  Pending notifications will be
 * lost when the server goes down, but registrations of interest
 * survive across server crashes for persistent servers.
 *
 * @see JavaSpace#notify
 * @see BasicSpace#notify
 * @see NotifyChit
 */
class Notifier extends Thread implements com.sun.jini.constants.TimeConstants {
    private FastList	possibles;	// list of entries waiting to be checked
    private TaskManager	pending;	// pending notifications
    private BasicSpace	space;		// the space I work for

    private final static long	PAUSE_TIME = 5000;	// time between checks
    private final static int	MAX_ATTEMPTS = 10;	// max times to retry

    private final static java.io.PrintWriter dbgNotify =
	BasicSpace.debug.getWriter("notify");

    /**
     * Create a notifier connected to the given <code>space</code>.
     */
    Notifier(BasicSpace space) {
	super("Notifier");
	this.space = space;
	pending = new TaskManager();
	possibles = new FastList();
	start();
    }

    /**
     * Add a new entry to the list of entries to consider for matches.
     */
    void newEntry(EntryRep entry, TransactableMgr txn) {
	possibles.add(new WriteNode(entry, txn));
	doNotify();
    }

    /**
     * Add a new entry to the list of entries to consider for matches
     * because of a previous transaction conflict that might just have
     * been cleared because of an abort.
     */
    void abortForEntry(EntryRep entry) {
	possibles.add(new WriteNode(entry, true));
	doNotify();
    }

    /**
     * Perform a <code>notifyAll</code> on this object after synchronizing.
     */
    private void doNotify() {
	synchronized (this) {
	    notifyAll();
	}
    }

    /**
     * Objects of this class represent newly written nodes in the space.
     *
     * @see #newEntry
     */
    private static class WriteNode extends FastList.Node {
	EntryRep	entry;
	TransactableMgr	txn;
	boolean		proxyOnly;

	WriteNode(EntryRep entry, TransactableMgr txn) {
	    this.entry = entry;
	    this.txn = txn;
	}

	WriteNode(EntryRep entry, boolean proxyOnly) {
	    this.entry = entry;
	    this.proxyOnly = proxyOnly;
	}
    }

    /**
     * Send an exception to the given chit that is waiting on a blocked
     * call.
     */
    void exceptionToSend(NotifyChit chit) {
	if (dbgNotify != null)
	    dbgNotify.println("exceptionToSend: " + chit);
	chit.forceNotify();	// make sure the exception gets sent
	pending.add(new NotifyTask(chit));
	doNotify();
    }

    /**
     * Return the space to which this notifier is attached
     *
     * @see Notifier(BasicSpace)
     */
    BasicSpace space() {
	return space;
    }

    /**
     * Loop looking for new entries to check for, and notifications to
     * be sent.
     */
    public void run() {
	for (;;) {
	    checkPossibles();	// check entries for matches
	}
    }

    /**
     * Check the list of newly-written entries to see if notifications
     * are required.  This is done here in our thread so that the thread
     * that wrote the entry doesn't have to block waiting for this check.
     */
    private void checkPossibles() {
	// wait for something to do, if necessary
	if (possibles.head() == null) {
	    synchronized (this) {
		try {
		    //!! when nothing is pending, and there are no possibles,
		    //!! sleep much longer?
		    wait(PAUSE_TIME);
		} catch (InterruptedException e) {
		    return;		// shouldn't happen
		}
	    }
	}

	WriteNode n = (WriteNode) possibles.head();
	if (n == null)		// don't start transaction if nothing to do
	    return;

	try {
	    space.startTxn();
	    for (; n != null; n = (WriteNode) n.next()) {
		possibles.remove(n);
		// get chit list
		Vector chits = space.chitsFor(n.entry, n.txn, n.proxyOnly);
		if (dbgNotify != null) {
		    dbgNotify.println("entry " + n.entry + ", txn " + n.txn
			+ " got " + (chits == null ? -1 : chits.size()));
		}
		if (chits != null && chits.size() != 0) {   // if anyone cares
		    logMatches(chits);			    // remember 'em
		    for (int i = 0; i < chits.size(); i++) {// and add 'em
			NotifyChit chit = (NotifyChit) chits.elementAt(i);
			pending.add(new NotifyTask(chit));
		    }
		}
	    }
	} finally {
	    space.commitTxn();
	}
    }

    /**
     * Log out all the matches shown in this list of chits.  We only need
     * to log one for each handle, not one for each chit, because the sequence
     * number is per-handle.
     */
    private void logMatches(Vector chits) {
	NotifyChit chit;
	TemplateHandle handle;
	if (chits.size() == 1) {
	    // one chit is a simpler, faster case
	    chit = (NotifyChit) chits.elementAt(0);
	    handle = chit.handle();
	    space.notifiedOp(chit.id(), handle.numMatches());
	} else {
	    // only log once per handle
	    Set known = new HashSet();
	    Enumeration e = chits.elements();
	    while (e.hasMoreElements()) {
		chit = (NotifyChit) e.nextElement();
		handle = chit.handle();
		if (!known.contains(handle)) {
		    known.add(handle);
		    space.notifiedOp(chit.id(), handle.numMatches());
		}
	    }
	}
    }

    /*
     * Static stuff for Pending (can't put it in the class, unfortunately).
     */
				// 1 day =hrs  mins secs milliseconds
    private static final long	MAX_TIME = 1 * DAYS;
    private static final long	delays[] = {
				    1 * SECONDS, 5 * SECONDS,
				    10 * SECONDS, 60 * SECONDS, 60 * SECONDS
				};

    static {
	/*
	 * Make the delays the amount of time since the start -- it
	 * is easier to declare the intervals, but the elapsed time is
	 * more <i>useful</i>.
	 */
	for (int i = 1; i < delays.length; i++)
	    delays[i] += delays[i - 1];
    }

    /**
     * A task that represent a notification of matching a particular
     * template under a given transaction.
     */
    private class NotifyTask extends RetryTask {
	private NotifyChit	chit;		// our target

	/**
	 * Create an object to represent this list of chits needing
	 * notification.
	 */
	NotifyTask(NotifyChit chit) {
	    super(Notifier.this.pending);
	    this.chit = chit;
	}

	/**
	 * Try to notify the target.  Return <code>true</code> if the
	 * notification was successful.
	 * <p>
	 * We know that we are the only one dealing with the given chit
	 * because <code>runAfter</code> makes sure of it.
	 */
	public boolean tryOnce() {
	    long curTime = System.currentTimeMillis();
	    if (curTime - startTime() > MAX_TIME)
		return true;	// just stop here, we are declaring "success"

	    BasicSpace space = space();
	    boolean successful = true;	// notification successful?
	    try {
		TemplateHandle handle = chit.handle();

		long seqNum = handle.numMatches();
		if (dbgNotify != null)
		    dbgNotify.println(this + ": seqNum = " + seqNum);
		if (!chit.shouldNotify(seqNum)) {
		    if (dbgNotify != null)
			dbgNotify.println(this + ": skipping, already done");
		    return true;	// my work is done already
		}

		if (dbgNotify != null)
		    dbgNotify.println(this + ": trying");
		RemoteEventListener listener = getListener(chit);

		boolean wanted = true;
		if (space.expired(chit, curTime)) {
		    notWanted(chit);
		    wanted = false;
		}

		if (wanted) {
		    RemoteEvent ev = new RemoteEvent(
			space.getSpaceProxy(), handle.id(),
			seqNum, chit.handback()
		    );
		    listener.notify(ev);
		    try {
			chit.notified(seqNum);
		    } catch (Exception e) {
			System.err.println("notifier exception:");
			e.printStackTrace(System.err);
		    }
		}
	    } catch (NoSuchObjectException e) {
		// the object has gone away
		notWanted(chit);
		// this is still "successful" -- we know to stop sending this
	    } catch (UnknownEventException e) {
		// they didn't want to know about this, so stop them getting
		// future notifications, too.
		notWanted(chit);
		// this is still "successful" -- we know to stop sending this
	    } catch (RuntimeException e) {
		// The assumption here is that the client is hosed, and won't 
		// get better, so cancel the lease.
		notWanted(chit);
		// this is still "successful" -- we know to stop sending this
	    } catch (RemoteException e) {
		successful = false;	// didn't get there, try again
	    }

	    if (dbgNotify != null) {
		dbgNotify.println(this + ": " +
		    (successful ? "successful" : "failed"));
	    }

	    if (successful && chit.sendException())	// is sending exception
		chit.setExpiration(0);			// sending is expiration

	    if (!successful && attempt() > MAX_ATTEMPTS)
		return true;		// as successful as we're going to be

	    return successful;
	}

	private void notWanted(LeasedResource resource) {
	    try {
		space.cancel(resource.getCookie());
	    } catch (UnknownLeaseException err) {
		; // if we're already unregistered, that's cool too
	    }
	}

	private RemoteEventListener getListener(NotifyChit chit)
	    throws RemoteException
	{
	    try {
		return chit.listener();
	    } catch (COM.odi.NoTransactionInProgressException e) {
		BasicSpace space = space();
		try {
		    space.startTxn();
		    return chit.listener();
		} finally {
		    space.commitTxn();
		}
	    }
	}

	/**
	 * We must run after any running notification of the same chit
	 * since it will be manipulating the same chit information.  If
	 * we fail to do this, we might unnecessarily notify twice.
	 */
	public boolean runAfter(java.util.List list, int max) {
	    for (int i = 0; i < max; i++) {
		Object task = list.get(i);
		if (task instanceof NotifyTask) {
		    NotifyTask nt = (NotifyTask) task;
		    if (nt.chit == chit)
			return true;
		}
	    }
	    return false;
	}
    }
}
