/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import net.jini.core.event.*;
import com.sun.jini.mahout.StorableObject;

import java.rmi.RemoteException;
import java.rmi.MarshalledObject;

import java.io.*;

import com.sun.jini.lease.landlord.LeasedResource;

import java.rmi.MarshalledObject;

/**
 * An object that represents a registration of interest.  When a match is
 * found, this chit will be used to designate an event listener and the
 * cookie for the registration.  That cookie is also used when
 * registration is cancelled.
 * <p>
 * This gets much more complicated than you might think because of the
 * ODI persistence model.  The <code>RemoteEventListener</code> reference must
 * be <code>transient</code> -- I cannot make <code>RemoteStub</code>
 * persistence-capable since I don't own it.  But must store
 * registrations in the database (I allow notifications to be lost
 * across server crashes, but I cannot allow registrations to be
 * forgotten). So they are stored as a MarshalledObject, with the
 * listener field re-initialized from it as necessary.
 *
 * @author Ken Arnold
 *
 * @see BasicSpace#notify
 * @see BasicSpace#cancel
 */
public class NotifyChit implements LeaseDesc, LeasedResource, Externalizable {
    private Long		cookie;		// cancellation cookie
    private long		expires = 0;	// when this expires;
    private MarshalledObject	handback;	// what to hand back
    private StorableObject	listener;	// the listener ref
    private TemplateHandle	handle;		// handle holding this
    private transient long	lastNotified;	// last value used in a notify

    /*
     * Note: lastNotified is transient because it does not need to
     * preservered between server runs.  It is designed to help
     * compress pending notifications, and pending notifications are
     * lost when the server goes down.  The first time a server boots,
     * the first notification generated for a chit should be sent, so
     * there is no need to compress it out.  This is also why
     * lastNotified is not logged (i.e., there is no method in LogOps
     * for it).
     */

    // stamp this so osjcfp'ed version has same ID
    private static final long serialVersionUID = -866459651906238985L;

    /** DO NOT USE: Used by the object output stream. */
    public NotifyChit() { }

    /**
     * Create a notify chit for the given <code>listener</code>,
     * <code>cookie</code> pair.  Leaves the expiration set to 0.
     */
    NotifyChit(RemoteEventListener listener, long cookie, 
	MarshalledObject handback)
    {
	// can't catch exceptions, so we pass null and assign afterwards
	this((StorableObject) null, cookie, handback);
	try {
	    this.cookie = new Long(cookie);
	    this.listener = new StorableObject(listener);
	} catch (RemoteException e) {
	    // This should never happen -- the listener is currently in
	    // hand and should therefore be valid at this point for all
	    // uses that don't involve remote contact.
	    System.err.println("Cannot setup listener");
	    e.printStackTrace();
	}
    }

    /**
     * Create a notify chit for the given <code>listener</code>,
     * <code>cookie</code> pair.
     */
    NotifyChit(StorableObject listener, long cookie, MarshalledObject handback)
    {
	this.cookie = new Long(cookie);
	this.handback = handback;
	this.listener = listener;
    }

    /**
     * Return the Chit's ID (it's cookie value as a long).
     */
    long id() {
	return cookie.longValue();
    }

    /**
     * Return the listener.  Deserialize the bytes if we don't have an
     * actual reference in hand.
     *
     * @exception java.rmi.RemoteException
     *		Problems re-establishing connection with listener.
     */
    RemoteEventListener listener() throws RemoteException {
	try {
	    return (RemoteEventListener) listener.get();
	} catch (RemoteException e) {
	    throw e;			// we should raise these
	}
    }

    // inherit doc comment from LeaseDesc
    public String classFor() {
	return handle.classFor();
    }

    // inherit doc comment
    public LeasedResource getLeasedResource() {
	return this;
    }

    /**
     * Return the handback object.
     */
    MarshalledObject handback() {
	return handback;
    }

    /**
     * Set the handback object.
     */
    void handback(MarshalledObject obj) {
	handback = obj;
    }

    /**
     * Return the holding handle.
     */
    TemplateHandle handle() {
	return handle;
    }

    /**
     * Return <code>true</code> if the given sequence number should be sent
     */
    synchronized boolean shouldNotify(long seqNum) {
	return (seqNum > lastNotified);
    }

    /**
     * Ensure that the next notification gets out.
     */
    synchronized void forceNotify() {
	lastNotified = -1;
    }

    /**
     * Tell this chit that the given sequence number has been
     * successfully sent.  
     */
    synchronized void notified(long seqNum) {
	if (seqNum > lastNotified)
	    lastNotified = seqNum;
    }

    /**
     * Set the holding handle.
     */
    void handle(TemplateHandle handle) {
	this.handle = handle;
    }

    /**
     * Return <code>true</code> if this notification can see things the happen
     * inside the given transaction.
     */
    boolean canSee(TransactableMgr writeTxn) {
	return (writeTxn == null);
    }

    /**
     * Return <code>true</code> if this chit is trying to send an exception.
     */
    boolean sendException() {
	return false;		// normal chits don't do this.
    }

    // inherit doc comment
    public int hashCode() {
	return (int) cookie.longValue();
    }

    // inherit doc comment
    public boolean equals(Object o) {
	try {
	    return cookie.equals(((NotifyChit) o).cookie);
	} catch (ClassCastException e) {
	    return false;
	}
    }

    // Methods required by LeasedResource
    // inherit doc comment from LeasedResource
    public void setExpiration(long newExpiration) {
	expires = newExpiration;
    }

    // inherit doc comment from LeasedResource
    public Object getCookie() {
	return cookie;
    }

    // inherit doc comment from LeasedResource
    public long getExpiration() {
	return expires;
    }

    /**
     * @serialData
     */
    // inherit doc comment
    public void writeExternal(ObjectOutput out) throws IOException {
	/*
	 * "handle" is not written because when a chit is written to a
	 * log file, we don't want to serialize it's handle and all
	 * associated data.  See LogOutputFile and LogInputFile, where
	 * this assumption is made.
	 *
	 */
	out.writeLong(cookie.longValue());
	out.writeLong(expires);
	out.writeObject(handback);
	out.writeObject(listener);
    }

    // inherit doc comment
    public void readExternal(ObjectInput in)
	throws IOException, ClassNotFoundException
    {
	cookie = new Long(in.readLong());
	expires = in.readLong();
	handback = (MarshalledObject) in.readObject();
	listener = (StorableObject) in.readObject();
    }
}
