/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import COM.odi.*;
import COM.odi.util.OSHashtable;
import COM.odi.util.OSVector;

import java.util.*;
import java.io.*;

/**
 * This class implements a view of a store using an ODI PSE database.
 * This takes a very simplistic view since we only need to use ODI
 * in a simple way:
 * <ul>
 * <li> We use only one session -- a global one.
 * <li> Only one thread can access the database at a time.
 * </ul>
 *
 * @see BasicSpace
 */
class ODIStore implements Store {
    private BasicSpace		space;		// the space we're helping
    private String		name;		// the space's name
    private String		path;		// path for persistent store
    private String		dbname;		// name of the DB itself
    private Database		db;		// the database
    private COM.odi.Transaction	odiTxn;		// current ODI txn
    private Thread		dbThread;	// db lock owner
    private int			txnNesting;	// nesting depth
    private Segment		segment;	// the one segment
    private int			commitMode;	// parameter to commit()
    private Session 		session;	// our session
    private ODIStoreDB		dbRef;		// ref to our root manager

    /** The properties that govern the ODI garbage collector. */
    private final static Properties	GC_PROPS;

    static {
	GC_PROPS = new Properties(System.getProperties());
	GC_PROPS.put("COM.odi.ProGCKey", "StartGCMarkStart");
    }

    private static final PrintWriter dbgODI =
					BasicSpace.debug.getWriter("odi");
    private static final PrintWriter dbgStacks =
					BasicSpace.debug.getWriter("odiStacks");

    /**
     * Create an <code>ODIStore</code> object for the given
     * <code>space</code>, <code>name</code>, <code>path</code>, and
     * <code>commitMode</code>.  The <code>commitMode</code> will be
     * passed to ODI's <code>commit</code> method at the end of each
     * transaction.
     */
    ODIStore(BasicSpace space, String name, String path, int commitMode) {
	this.space = space;
	this.name = name;
	this.path = path;
	this.commitMode = commitMode;
    }

    /**
     * Ensure that the object is deep-fetched and then return it.
     */
    public Object transientOf(Object obj) {
	if (obj == null)
	    return null;

	/*
	 * Ensure that the contents of a simple Vector are themselves
	 * properly transient.  This is for methods that return
	 * Vectors, which are not themselves persistent but may
	 * contain persistent objects.  Vector is the only container
	 * class we do this for because Vector is the only container
	 * class we store persistent objects in 
	 */
	if (obj instanceof Vector) {
	    Vector v = (Vector) obj;
	    for (int i = v.size() - 1; i >= 0; i--)
		v.setElementAt(transientOf(v.elementAt(i)), i);
	    return v;
	}

	ObjectStore.migrate(obj, segment, false);
	ObjectStore.deepFetch(obj);
	return obj;
    }

    /**
     * Create the session object and set the session field.
     */
    private void createSession() {
        // We want to disable ODI's cross VM database locking feature
        // because we do our own locking (activation won't allow
        // multiple concurrent VM's for the same activatable object,
        // and we won't create multiple spaces pointing at the same
        // log) and ODI's locking keeps us from automaticly recovering
        // from crashes
        Properties props = new Properties(System.getProperties());
        props.setProperty("COM.odi.useDatabaseLocking", "false");
	session = Session.createGlobal(null, props);
    }	

    /**
     * Set up the ODI persistent store.
     */
    public void setupStore() {
	createSession();

        dbname = new File(path, name + ".odb").toString();
        if (dbgODI != null)
            dbgODI.println("dbname = " + dbname);

	openDB(ObjectStore.OPEN_UPDATE);
	startTxn();
        ODIStoreDB.setupDB(space.DEFAULT_NAME, dbRef);
	endTxn(true);
    }

    /**
     * Refresh the database root, stored fields and the
     * migration segment.
     */
    private void refreshDB() {
        // start the ODIStoreDB scheme going for this DB
	startTxn();
	getFields();
        segment = Segment.of(dbRef);
	endTxn(true);
    }

    /**
     * Open the database.  If the database is being re-opened
     * after a persistent garbage collection, the database
     * can be non-null, but still closed.  This situation
     * is also taken into account.
     */
    void openDB(int openMode) {
	//If you re-open the database, you must
	//refresh it

	if (session == null)
	    createSession();

	if ((db != null) && (!db.isOpen())) {
	    System.out.println("re-opening DB: " + dbname);
	    db.open(openMode);
	}

	if (db == null) {         //skip if already open
	    System.out.println("opening DB: " + dbname);
	    try {
		db = Database.open(dbname, openMode);
	    } catch (COM.odi.DatabaseNotFoundException e) {
		db = Database.create(dbname, 0666);	// make it if necessary
	    }
	}

	refreshDB();
    }

    /**
     * Start an ODI txn.
     */
    public void startTxn() {
	getDBLock();			// wait for the DB to be ours
    }

    /**
     * End an ODI txn.
     */
    public void endTxn(boolean commit) {
	/*
	 * If commitTxn is part of a finally and the startTxn never
	 * happened, someone else might have the lock, or nobody
	 * might.
	 */
	if (odiTxn != null && Thread.currentThread() == dbThread)
	    releaseDBLock(commit);
    }

    /**
     * Remove the database.
     */
    public void destroy() {
	try {
	    if (db != null)
		db.destroy();
        } catch (Exception e) {
            System.err.println("destroying db:");
            e.printStackTrace();
            System.err.println("destroy continuing");
        }

	try {
            com.sun.jini.system.FileSystem.destroy(new File(path), true);
        } catch (Exception e) {
            System.err.println("destroying " + path + ":");
            e.printStackTrace();
            System.err.println("destroy continuing");
        }
    }

    /**
     * Get the fields of our space from the database root.  We only
     * need to do any actual work if we haven't ever gotten them
     * before, since the commit leaves behind references that are
     * still valid in the next txn.  In addition, make sure that
     * the fields are not stale.  If stale, then force a refresh.
     */
    void getFields() {
	boolean refresh = false;

	if (space.contents == null || space.templates == null ||
	    space.types == null)
	{
            if (dbgODI != null)
                dbgODI.println("refreshing db");
 
            dbRef = ODIStoreDB.refresh(db);
 
            // don't set name -- it's readonly so can't change
            space.contents  = dbRef.contents;
            space.templates = dbRef.templates;
            space.types     = dbRef.types;
	}
    }


    /**
     * Get the DB lock.  All access to the database is single-threaded
     * to avoid complications.  This is inefficient in some cases, but
     * always safe.
     *
     * @see releaseDBLock
     */
    private void getDBLock() {
	try {
	    Thread t = Thread.currentThread();
	    synchronized (db) {
		if (dbgODI != null) {
		    dbgODI.println("Getting db lock ");
		    if (dbgStacks != null)
			new Throwable().printStackTrace(dbgStacks);
		}
		if (dbThread == t)	// already have it in this thread
		    txnNesting++;
		else {
		    while (dbThread != null) {
			if (dbgODI != null)
			    dbgODI.println("dbThread = " + dbThread.getName());
			db.wait();
		    }
		    dbThread = t;
		    txnNesting = 0;
		    odiTxn =
			COM.odi.Transaction.begin(ObjectStore.UPDATE);
		}

		if (dbgODI != null) {
		    dbgODI.println("dbThread set to " + t.getName() + 
				   ", nesting = " + txnNesting);
		}
	    }
	} catch (InterruptedException e) {
	    System.err.println("Unexpected Interruption");
	    e.printStackTrace(System.err);
	    throw new IllegalMonitorStateException("Unexpected Interruption");
	}
    }

    /**
     * Release the DB lock (previously acquired by <code>getDBLock</code>).
     *
     * @see getDBLock
     */
    private void releaseDBLock(boolean commit) {
	Thread t = Thread.currentThread();
	synchronized (db) {
	    if (dbgODI != null) {
		dbgODI.println("releasing db lock held by " +
		    dbThread.getName() + ", nesting = " + txnNesting);
		if (dbgStacks != null)
		    new Throwable().printStackTrace(dbgStacks);
	    }

	    if (dbThread != t) {
		throw new IllegalArgumentException(
		    "releasing DB lock when not held by this thread");
	    }

	    if (txnNesting > 0) {
		--txnNesting;
		if (dbgODI != null)
		    dbgODI.println("nesting now " + txnNesting);
		return;
	    }

	    if (commit) {
		if (dbgODI != null) {
		    dbgODI.println("odiTxn.commit(" + commitMode + ")");

		    String msg = null;
		    switch(commitMode) {
		        case ObjectStore.RETAIN_STALE:
			  msg = "ObjectStore.RETAIN_STALE";
			  break;
		        case ObjectStore.RETAIN_READONLY:
			  msg = "ObjectStore.RETAIN_READONLY";
			  break;
		        case ObjectStore.RETAIN_UPDATE:
			  msg = "ObjectStore.RETAIN_UPDATE";
			  break;
		        case ObjectStore.RETAIN_HOLLOW:
			  msg = "ObjectStore.RETAIN_HOLLOW";
			  break;
			default:
			  msg = "UNKNOWN";
		    }
		    dbgODI.println("odiTxn.commit(" + msg + ")");
		}

		odiTxn.commit(commitMode);
	    } else {
		if (dbgODI != null)
		    dbgODI.println("odiTxn.abort()");
		odiTxn.abort();
	    }
	    odiTxn = null;
	    dbThread = null;
	    db.notifyAll();
	}
    }

    /**
     * Look up the named root.  If the root doesn't exist or is
     * <code>null</code>, create a new object of the given class
     * and assign it to the root.
     */
    Object getRoot(String name, ObjectFactory factory) {
	Object rootval = null;
	try {
	    rootval = db.getRoot(name);
	    if (dbgODI != null) {
		dbgODI.print("ODIStoreDB:refresh:root " + name + " found");
		if (rootval == null)
		    dbgODI.print(", null");
		dbgODI.println();
	    }
	} catch (COM.odi.DatabaseRootNotFoundException e) {
	    db.createRoot(name, null);
	    if (dbgODI != null)
		dbgODI.println("creating " + name);
	}

	// allocate it if necessary (root didn't exist or was null)
	if (rootval == null) {
	    try {
		rootval = factory.newObject();
		db.setRoot(name, rootval);
	    } catch (Exception e) {
		throw new IllegalArgumentException("bad class: " + factory
		    + ": " + e);
	    }
	}

	return rootval;
    }

    /**
     * Return the root for the given <code>name</code>.  If the root is
     * unknown, returns <code>null</code>.
     */
    Object getRoot(String name) {
	Object rootval = null;
	try {
	    rootval = db.getRoot(name);
	    if (dbgODI != null) {
		dbgODI.print("ODIStoreDB:refresh:root " + name + " found");
		if (rootval == null)
		    dbgODI.print(", null");
		dbgODI.println();
	    }
	    return rootval;
	} catch (COM.odi.DatabaseRootNotFoundException e) {
	    if (dbgODI != null) {
		dbgODI.println("ODIStoreDB:refresh:root " + name
		    + " not found");
	    }
	    return null;
	}
    }

    /**
     * Set the root for the given <code>name</code> to given
     * <code>value</code>.  If the root does not exist, create it.
     */
    void setRoot(String name, Object value) {
	try {
	    startTxn();
	    db.setRoot(name, value);
	} catch (COM.odi.DatabaseRootNotFoundException e) {
	    db.createRoot(name, value);
	    if (dbgODI != null)
		dbgODI.println("creating " + name);
	} finally {
	    endTxn(true);
	}
    }

    /**
     * Reset the fields in the Space reachable from the
     * database root to null. This is done to force a
     * refresh of these fields from a newly obtained
     * database root.
     */
    void resetFields() {
	    space.contents = null;
	    space.types = null;
	    space.templates = null;
    }

    /**
     * Closes the database and forces fields to be
     * refreshed from a newly obtained database root.
     *
     * @param retain Indicates whether objects obtained
     *		     from the database will remain behind
     *		     as transient object in memory.
     */
    void closeAndReset(boolean retain) {
	//since the database is nulled by
	//the close method, we impose the
	//nulling of the fields in the space
	//in a finally block.

	try {
	    close(retain);
	} finally {
	    resetFields();
	}
    }

    /**
     * Close the database.  The <code>retain</code> parameter is passed
     * to the database's <code>close</code> method (<code>true</code>
     * means retain any objects transient -- otherwise they are simply
     * stale).
     */
    void close(boolean retain) {
	if (db == null)
	    return;

	try {
	    db.close(retain);
	if (dbgODI != null)
	    dbgODI.println("db " + name + " closed");
	} finally {
	    db = null;
	}
	session.terminate();
	session = null;
    }

    /**
     * Run one pass of the database garbage collector, passing
     * <code>GC_PROPS</code> as the properties parameter.  This
     * requires closing an reopening the database, since the GC runs
     * only on closed database.
     */
    void gc() {
	db.close();
	Properties results = db.GC(GC_PROPS);
	if (results == null)
	    System.err.println("*** couldn't GC, ignoring");
	else if (dbgODI != null) {
	    Object unref = results.getProperty("Unreferenced Objects");
	    if (unref != null)
		dbgODI.println("Database GC'ed " + unref + " object(s)");
	}
	resetFields();
	openDB(ObjectStore.OPEN_UPDATE);
    }
}
