/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import com.sun.jini.mahout.ExecUtilBase;
import java.io.IOException;

/**
 * A utility class for starting an Outrigger server.
 */
public class OutriggerExecUtil extends ExecUtilBase {
    /** Boolean that is true if we are spawning an activatable space */
    private boolean activatable;

    /** Enum for preferred persistent implementation of Outrigger */
    public static final int PERSISTENT = 0;

    /** Enum for preferred transient implementation of Outrigger */
    public static final int TRANSIENT = 1;

    /**
     * The fully specified path of the executable jar file for starting 
     * Outrigger
     */
    String outriggerJarFile = null;
    String outriggerTransientJarFile = null;

    /** 
     * Change codebase, classpath, and/or policyFilePath the
     * appropriate <code>com.sun.jini.exec.outrigger.*</code> property has
     * been set.
     *
     * @param type        Which implementation of Outrigger to spawn.
     */
    private void constructorUtil(int type) {

	final String outriggerCodebase = 
	    System.getProperty("com.sun.jini.exec.outrigger.codebase");

	if (codebase == null && outriggerCodebase == null)
	    throw new IllegalStateException("OutriggerExecUtil:Must specify " +
	         "a codebase using com.sun.jini.exec.outrigger.codebase or " +
		 "com.sun.jini.exec.codebase");
	
	if (outriggerCodebase != null)
	    codebase = outriggerCodebase;


	outriggerJarFile = 
	    System.getProperty("com.sun.jini.exec.outrigger.class.path");
	System.out.println("JAR FILE 1 ==> " + outriggerJarFile);

	outriggerTransientJarFile = 
	    System.getProperty(
		   "com.sun.jini.exec.outrigger.transient.class.path");
	System.out.println("JAR FILE 2 ==> " + outriggerTransientJarFile);

	if (outriggerJarFile == null && type == PERSISTENT)
	    throw new IllegalStateException("OutriggerExecUtil:Must specify " +
	         "a classpath using com.sun.jini.exec.outrigger.class.path." +
		 "Note: currently the class path is really the name of the\n" +
		 "      executable jar file (outrigger.jar)");
	
	if (outriggerTransientJarFile == null && type == TRANSIENT)
	    throw new IllegalStateException("OutriggerExecUtil:Must specify " +
	         "a classpath using com.sun.jini.exec.outrigger.transient." +
		 "class.path." +
		 "Note: currently the class path is really the name of the\n" +
		 "      executable jar file (outrigger.jar)");
	
	final String outriggerPolicy = 
	    System.getProperty("com.sun.jini.exec.outrigger.policy");

	if (policyFilePath == null && outriggerPolicy == null)
	    throw new IllegalStateException("OutriggerExecUtil:Must specify " +
	         "a policy file using com.sun.jini.exec.outrigger.policy or " +
		 "com.sun.jini.exec.policy");
	
	if (outriggerPolicy != null)
	    policyFilePath = outriggerPolicy;
    }

    /** 
     * Return name of the class to spawn based on one of the values
     * <code>PERSISTENT</code> or <code>TRANSIENT</code>.
     */
    private static String spaceClassName(int enum) {
	switch (enum) {
	  case TRANSIENT:
	    return "com.sun.jini.outrigger.StartTransientOutrigger";
	  case PERSISTENT:
	    return "com.sun.jini.outrigger.StartPersistentOutrigger";
	  default:
	    throw new IllegalArgumentException("Type of space must be " +
	         "PERSISTENT(" + PERSISTENT + ") or TRANSIENT(" + TRANSIENT +
		 ")");
	}
    }

    /**
     * Append <code>cmd</code> with appropriate string to set the
     * space's name to <code>name</code>.  <code>name</code> may be
     * <code>null</code>.
     */
    private void addName(String name) {
	if (name == null)
	    return;

	cmd += "-Dcom.sun.jini.outrigger.spaceName=" + name + " ";
    }

    /**
     * Set <code>cmd</code>.
     * @param useRegistry If <code>true</code> setup the service to bind into
     *                    RMI registry.  Otherwise use the Jini Lookup service.
     * @param port        Port to find the RMI registry on. Ignored if 
     *                    <code>useRegistry</code> is <code>false</code>.
     * @param groups      Lookup groups the service should bind into. Ignored
     *                    if <code>useRegistry</code> is <code>true</code>.
     * @param type        Which implementation of Outrigger to spawn.
     * @param name        Name to register the service under.
     */
    private void buildCmd(boolean useRegistry, int port, String groups,
			  int type, String name)
    {
	if (type == TRANSIENT) {
	    activatable = false; 

	    cmd += "-Djava.security.policy=" + policyFilePath  + " ";

	    if (useRegistry) {
		cmd += "-Dcom.sun.jini.use.registry=yes ";
		if (port >= 0)
		    cmd += "-Dcom.sun.jini.rmiRegistryPort=" + port + " ";
	    }

	    cmd += "-Djava.class.path=" + classpath + " ";
	    cmd += "-Djava.rmi.server.codebase=" + codebase + " ";

 	    addName(name);

	    String jarFile = (type == PERSISTENT) ? outriggerJarFile :
		                                    outriggerTransientJarFile;

	    cmd += "-jar " + jarFile;
	    cmd += " " + groups;
	} else {
	    activatable = true; 

	    cmd += "-Djava.security.policy=" + policyFilePath  + " ";

	    if (useRegistry) {
		cmd += "-Dcom.sun.jini.use.registry=yes ";
		if (port >= 0)
		    cmd += "-Dcom.sun.jini.rmiRegistryPort=" + port + " ";
	    }

	    cmd += "-Djava.class.path=" + classpath + " ";

 	    addName(name);

	    cmd += "-jar " + outriggerJarFile + " " + codebase + " " + 
		policyFilePath + " " + persistencePath + " " + groups;
	}

    }

    // purposefully inherit doc comment from supertype
    public void waitForSetup() throws InterruptedException {
	if (activatable)
	    super.waitForSetup();
	// else just return
    }

    /**
     * Build an <code>OutriggerExecUtil</code> for a space that will
     * bind itself into Jini Lookup services.
     *
     * @param type   Type of space to start
     * @param groups Lookup group list
     * @param name   The space's name
     */
    public OutriggerExecUtil(int type, String groups, String name) 
	throws IOException 
    {
	super("Outrigger");
	constructorUtil(type);
	buildCmd(false, -1, groups, type, name);
    }

    /**
     * Build an <code>OutriggerExecUtil</code> for a space that will
     * bind itself into an RMI registry.
     *
     * @param type   Type of space to start
     * @param port   Registry port to bind to 
     * @param name   The space's name
     */
    public OutriggerExecUtil(int type, int port, String name) 
	throws IOException 
    {
	super("Outrigger");
	constructorUtil(type);
	buildCmd(true, port, "none", type, name);
    }
}
