/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import net.jini.core.transaction.*;
import net.jini.core.transaction.server.*;
import net.jini.core.event.*;
import net.jini.core.lease.*;

import com.sun.jini.lease.landlord.Landlord;

import java.rmi.*;

/**
 * This interface is the private wire protocol to the Outrigger
 * implementations of JavaSpaces<sup><font size=-2>TM</font></sup> 
 * technology.
 *
 * @author Ken Arnold
 *
 * @see EntryRep
 */
public interface OutriggerSpace extends TransactionParticipant, Landlord, 
    OutriggerAdmin, Remote 
{
    /**
     * Timeout value used to indicate that a read/take query is
     * is a retry attempt from a pending/blocking call on the 
     * SpaceProxy. Logically, this timeout value is equivalent
     * to NO_WAIT except for the special case where 
     * <code>ifExists</code> is true in a call to 
     * <code>getMatch</code>.  
     *
     * @see #getMatch
     */
    static final long INTERNAL_PROBE = -1;

    /**
     * Write a new entry into the space.
     *
     * @exception TransactionException A transaction error occurred
     */
    Lease write(EntryRep entry, Transaction txn, long lease)
	throws TransactionException, RemoteException;

    /**
     * Read or take any matching entry from the space, returning
     * <code>null</code> if no match is found.  Matching is
     * done against <code>tmpl</code> with blank fields being wildcards
     * ("match anything") other fields being values ("match exactly on
     * the serialized form").  If a match is found, an EntryRep for it
     * is returned.  If no match is found and <code>listener</code> is
     * <code>null</code>, <code>null</code> is returned.  If no match
     * is found and <code>listener</code> is not <code>null</code>, an
     * <code>EventRegistration</code> object for the notification is
     * returned. If <code>ifExists</code> is true and an potential 
     * match is currently under another transaction, then a 
     * TransactionConflictException is thrown indicating this
     * condition.
     *
     * Note: The special timeout value of <code>INTERNAL_PROBE</code>
     * is used to determine the return semantics of the last case 
     * described above. If the last case occurs with a NO_WAIT timeout
     * (or a timeout value that expires after performing the entry search)
     * then <code>null</code> is returned.  If the last case occurs with
     * a timeout of <code>INTERNAL_PROBE</code> then a 
     * <code>TransactionConflictException</code> is thrown.
     * The assumption is that a NO_WAIT call just needs a "yes/no" answer 
     * while the <code>INTERNAL_PROBE</code> call is querying the 
     * system state and can still wait for a definitive answer.
     *
     * @exception TransactionException A transaction error occurred
     * @exception TransactionConflictException A transaction conflict exists
     */
    Object getMatch(EntryRep tmpl, Transaction txn, long timeout,
		    boolean read, boolean ifExists,
		    RemoteEventListener listener, long id)
	throws TransactionException, TransactionConflictException,
	       RemoteException;

    /**
     * This is <code>getMatch</code> where <code>listener</code> is
     * <code>null</code>, so the return type is guaranteed to be
     * EntryRep.
     *
     * @exception TransactionException A transaction error occurred
     */
    EntryRep getMatch(EntryRep tmpl, Transaction txn, long timeout,
		      boolean read, boolean ifExists)
	throws TransactionException, TransactionConflictException,
	       RemoteException;

    /**
     * When entries are written that match this template notify the
     * given <code>listener</code>.  A resulting cookie is returned.
     * Matching is done as for <code>read</code>.
     *
     * @see #cancel
     * @see net.jini.core.event.EventRegistration
     */
    EventRegistration
	notify(EntryRep tmpl, Transaction txn, RemoteEventListener listener,
	       long lease, MarshalledObject handback)
	throws TransactionException, RemoteException;

    /**
     * Renew lease on an entry or event registration.
     *
     * @exception UnknownLeaseException
     *	    The cookie does not denote an existing event registration or 
     *      entry in this space.
     * @exception LeaseDeniedException
     *      The space does not wish to renew the lease at this time.
     * @see com.sun.jini.lease.landlord.Landlord
     * @see #notify
     * @see #cancel
     */
    long renew(Object cookie, long extension)
	throws LeaseDeniedException, UnknownLeaseException, RemoteException;

    /**
     * Cancel the registration or entry identified by the
     * <code>cookie</code> returned by <code>Entry.notify</code>.
     *
     * @exception UnknownLeaseException
     *	    The cookie does not denote an existing event registration or 
     *      entry in this space.
     *
     * @see com.sun.jini.lease.landlord.Landlord
     * @see #notify
     * @see #renew */
    void cancel(Object cookie)
	throws UnknownLeaseException, RemoteException;

    /**
     * Return the a <code>SpaceProxy</code> associated with
     * this outrigger server.
     */
    SpaceProxy getSpaceProxy() throws RemoteException;

    /**
     * Return the admin proxy for this space.
     */
    Object getAdmin() throws RemoteException;
}
