/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import net.jini.core.transaction.Transaction;
import net.jini.core.transaction.server.TransactionConstants;
import com.sun.jini.mahout.StorableObject;

import java.rmi.RemoteException;

import COM.odi.util.OSVector;

/**
 * This object represents a pending transaction in a <code>BackEndSpace</code>.
 * As operations are performed under the transaction, they are logged into
 * this object.  When the transaction is committed, each operation is
 * committed into the DB.
 */
class PendingTxn implements TransactionConstants {
    /** An superclass for the objects that represent pending operations. */
    static abstract class PendingOp {
	/**
	 * Commit the operation by invoking the relevant method on the
	 * <code>processor</code> object.
	 */
	abstract void commit(LogOps processor);
    };

    /** An object that represents a pending write. */
    static class WriteOp extends PendingOp {
	EntryRep rep;
	WriteOp(EntryRep rep) { this.rep = rep; }
	void commit(LogOps processor) { processor.writeOp(rep, null); }
    };

    /** An object that represents a pending take. */
    static class TakeOp extends PendingOp {
	long id;
	TakeOp(long id) { this.id = id; }
	void commit(LogOps processor) { processor.takeOp(id, null); }
    }

    private Long		id;	// the transaction ID
    private int			state;	// current state
    private OSVector		ops;	// list of pending operations

    /*
     * The transaction is stored as a StorableObject so that it can be
     * reliably transmitted through the log and stored in the database,
     * even if the transaction manager is not currently available.  Otherwise
     * there is a potential for a RemoteException when deserializing the
     * object, which would leave us with unusable data.  This way we can
     * defer any use of it to whenever we need to.
     */
    private StorableObject	tr;	// the ServerTransaction object itself

    /**
     * Create a new <code>PendingTxn</code> for the given <code>id</code>.
     */
    PendingTxn(Long id) {
	this.id = id;
	state = ACTIVE;
	ops = new OSVector();
    }

    /** Add a new pending <code>write</code> operation. */
    void addWrite(EntryRep rep) { ops.addElement(new WriteOp(rep));	}
    /** Add a new pending <code>take</code> operation. */
    void addTake (long     id)  { ops.addElement(new TakeOp(id));	}

    /**
     * Return the current state of the transaction as known by
     * this object.
     */
    int state() {
	return state;
    }

    /**
     * Set the state of the transaction.
     */
    void state(int newState) {
	state = newState;
    }

    /**
     * Return the transaction id.
     */
    Long id() {
	return id;
    }

    /**
     * Return the list of operations stored in this transaction.  This
     * list should be treated as read-only.
     */
    OSVector ops() {
	return ops;
    }

    /**
     * Return the <code>Transaction</code> object.  This is <code>null</code>
     * if the transaction was never set, or is the result of invoking
     * <code>get</code> on the <code>StorableObject</code>.
     */
    Transaction transaction() throws RemoteException {
	if (tr == null)
	    return null;
	return (Transaction) tr.get();
    }

    /**
     * Set the <code>Transaction</code> object.
     */
    void transaction(StorableObject transaction) {
	tr = transaction;
    }

    /**
     * Return the <code>StorableObject</code> that contains the
     * <code>Transaction</code> object.
     */
    StorableObject storableTransaction() {
	return tr;
    }

    /**
     * Commit all the operations by invoking the relevant method
     * on the <code>processor</code> object.
     */
    void commit(LogOps processor) {
	java.util.Enumeration e = ops.elements();
	while (e.hasMoreElements())
	    ((PendingOp) e.nextElement()).commit(processor);
	state(COMMITTED);
    }
}
