/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import net.jini.space.*;
import net.jini.core.entry.*;
import com.sun.jini.collection.FastList;

import COM.odi.util.OSHashtable;

/**
 * A simple implementation of <code>EntryHolder</code>, which simply
 * stores the entries in a list.  This works fine for small holders.
 * More complicated ones that use hashing techniques and key lookup for
 * common values will handle larger sets.
 *
 * @author Ken Arnold
 *
 * @see EntryHandle
 */
class SimpleEntryHolder implements EntryHolder {
    static final long serialVersionUID = -9124787531852336772L;

    int numFields = -1;			// # of fields in this kind of entry
    String className;			// name of the class this holds
    boolean hasContents = false;	// any contents yet?
    FastList contents = new FastList();	// the list that holds the handles
    // Map of LeaseDescs indexed by the cookie of the underlying 
    // LeasedResource.
    //!! Could be final, but 1.2 compiler bug (id 4147610) makes it 
    // impossible
    protected OSHashtable idMap;

    private static final java.io.PrintWriter dbgHolder =
					BasicSpace.debug.getWriter("holders");

    /**
     * Create a new <code>SimpleEntryHolder</code> with the shared
     * <code>idMap</code>, and which will hold classes of the given
     * <code>className</code>.  The <code>idMap</code> is shared with
     * <code>EntryHolderSet</code> so that there is one table that can
     * map ID to <code>EntryRep</code>
     */
    SimpleEntryHolder(OSHashtable idMap, String className) {
	this.idMap = idMap;
	this.className = className;
    }

    // inherit doc comment from superclass
    public EntryRep
	hasMatch(EntryRep tmpl, TransactableMgr txn, boolean takeIt,
		 TransactionConflictException conflict)
	throws TransactionConflictException
    {
	if (dbgHolder != null)
	    dbgHolder.println("SimpleEntryHolder: hasMatch");
	if (numFields < 0)	// nothing has ever been added to this list
	    return null;

	int op = (takeIt ? TransactableMgr.TAKE : TransactableMgr.READ);

	EntryHandleTmplDesc desc = EntryHandle.descFor(tmpl, numFields);

	boolean foundConflict = false;
	long startTime = System.currentTimeMillis();
	for (EntryHandle handle = (EntryHandle) contents.head();
	     handle != null;
	     handle = (EntryHandle) handle.next())
	{
	    if (dbgHolder != null)
	        dbgHolder.println("SimpleEntryHolder: hasMatch: handle = " +
				   handle);
	    // Quick reject -- see the if handle mask is incompatible
	    if ((handle.hash() & desc.mask) != desc.hash)
		continue;

	    try {
		EntryRep rep = handle.rep();
		EntryRep result =
		    match(tmpl, rep, handle, txn, takeIt, startTime, conflict);
		if (result != null)
		    return result;
	    } catch (TransactionConflictException e) {
		// this happens on the first conflict only -- if conflict is
		// not null, the conflicting transactions have are simply added
		// to the list in the exception
		foundConflict = true;
		if (conflict == null)
		    conflict = e;	// already contains conflicting txns
	    }
	}

	if (foundConflict)	// if *we* found a conflict, throw it
	    throw conflict;
	return null;
    }

    /**
     * Debug method:  Dump out the state of this holder, printing out
     * the name of the dump first.
     */
    void dump(String name) {
	try {
	    System.out.println(name);
	    for (EntryHandle handle = (EntryHandle) contents.head();
		 handle != null;
		 handle = (EntryHandle) handle.next())
	    {
		EntryRep rep = handle.rep();
		System.out.println("    " + rep + ", " + rep.entry());
	    }
	} catch (UnusableEntryException e) {
	    e.printStackTrace();
	}
    }

    /**
     * If the template <code>tmpl</code> matches it, return the passed-in
     * <code>rep</code>.  The operation is performed under the transaction
     * <code>txn</code>.  If <code>takeIt</code> is <code>true</code>, this
     * is part of a take operation.  The <code>time</code> is the blocking
     * time.  The <code>conflict</code> parameter represents any 
     * <code>TransactionConflictException</code> that occurred during the
     * search of which this is a part; if there was none, <code>conflict</code>
     * will be <code>null</code>.
     */
    public EntryRep match(EntryRep tmpl, EntryRep rep, TransactableMgr txn,
			  boolean takeIt, long time,
			  TransactionConflictException conflict)
	throws TransactionConflictException
    {
        if (tmpl == null)
	    tmpl = EntryRep.matchAnyEntryRep();

	if (!hasContents)	// if I never inserted entries,
	    return null;	// .. then nothing was ever in here

	// ensure class match
	if (!rep.isAtLeastA(tmpl.classFor()))
	    return null;

	EntryHandle handle = handleFor(rep);
	if (dbgHolder != null)
	    dbgHolder.println("SimpleEntryHolder: take: handle = " + handle);
	if (handle == null)	// already removed
	    return null;

	return match(tmpl, rep, handle, txn, takeIt, time, conflict);
    }

    /**
     * Perform the actual match, which requires, in addition to the
     * public match, the <code>handle</code> for the entry in question.
     */
    private EntryRep
	match(EntryRep tmpl, EntryRep rep, EntryHandle handle,
	      TransactableMgr txn, boolean takeIt, long time,
	      TransactionConflictException conflict)
	throws TransactionConflictException
    {
	// tmpl == rep in FrontEndSpace.addPreparedTake, so optimize it
	if (tmpl != rep && !tmpl.matches(rep))
	    return null;
	synchronized (handle) {
	    // get rid of stale entries
	    if (isExpired(time, handle))
		return null;
	    handle.waitUntilStable();
	    if (handle.removed())    // oh, well -- someone got it first
		return null;
	    int op = (takeIt ? TransactableMgr.TAKE : TransactableMgr.READ);
	    if (!handle.canPerform(txn, op)) {
		// if can't operate due to a txn conflict
		if (dbgHolder != null) {
		    dbgHolder.println("match, but can't perform " +
			op + "; handle.knownMgr(txn) == " +
			handle.knownMgr(txn));
		}

		if (conflict != null) {
		    handle.addTxns(conflict.txns());
		    return null;
		} else {
		    conflict = new TransactionConflictException();
		    handle.addTxns(conflict.txns());
		    throw conflict;
		}
	    }
	    /*
	     * If this manager (txn) is known to the existing
	     * handle, then txn is a transaction which can
	     * operate directly on the existing handle.  If so,
	     * then we can operate on it directly.  Note that
	     * if txn is null and the handle is an EntryHandle
	     * (not a TxnEntryHandle), then txn is "known" in
	     * this sense.
	     *
	     * Else, this operation is new relative to the
	     * known transaction (possibly null) and txn.
	     */
	    if (handle.knownMgr(txn)) {	// added in this txn if taking
		if (dbgHolder != null)
		    dbgHolder.println("already known");
		// Do we need to take this entry? If we do try removal
		// (&& short-circuits if we don't want to take), if it
		// has already been removed return null (can't take
		// something twice).
		if (takeIt && !remove(handle))
		    return null;
	    } else {
		/*
		 * Since this txn is not known to the handle, txn
		 * must be non-null.  If the handle is already a
		 * TxnEntryHandle, then we should just add txn to
		 * the list of managers working with this object
		 * (it must, by the way, be a read operation).
		 *
		 * Else, we must create a TxnEntryHandle for this
		 * EntryHandle, and then add this txn to that new
		 * handle.  The TxnEntryHandle will keep a reference
		 * to the original EntryHandle so that it can be
		 * replaced in the list when the transaction completes
		 * if necessary.
		 */
		TxnEntryHandle txnHandle;
		if (handle instanceof TxnEntryHandle) {
		    if (dbgHolder != null)
			dbgHolder.println("already a TxnEntryHandle");
		    txnHandle = ((TxnEntryHandle) handle);
		} else {
		    if (dbgHolder != null)
			dbgHolder.println("creating TxnEntryHandle");
		    txnHandle =
			new TxnEntryHandle(handle, txn, this, op);
		    // insert the new handle (referring to the old one)
		    replace(handle, txnHandle, txn);
		}
		txnHandle.add(txn, op);	// add manager to handle's list
		txn.add(txnHandle);	// add handle to mgr's list
	    }
	    return rep;
	}
    }

    /**
     * Return <code>true</code> if the entry held by the given
     * <code>handle</code> has expired by the time in <code>now</code>.
     */
    private boolean isExpired(long now, EntryHandle handle) {
	if (BasicSpace.staticSpace.expired(handle.rep(), now)) {
	    if (dbgHolder != null) {
		dbgHolder.println("expired " + handle.rep().id() + " at " + 
		    handle.rep().getExpiration() +  " (now " + now + ")");
	    }
	    remove(handle);
	    // Log its removal
	    BasicSpace.staticSpace.cancelOp((Long)handle.rep().getCookie());
	    return true;
	} 
	return false;
    }
	    

    // inherit doc comment from superclass
    public Object add(EntryRep rep, TransactableMgr txn) {
	// if not in txn, the write is already committed, so add it in now
	return add(rep, txn, true);
    }

    /**
     * Add the given <code>rep</code> to this holder, managed by the
     * specified <code>txn</code>.  If <code>addToMap</code> is
     * <code>true</code>, the allocated <code>EntryHandle</code> will
     * be added to the hash map for this holder.  The <code>Object</code>
     * returned is the allocated handle.
     */
    Object add(EntryRep rep, TransactableMgr txn, boolean addToMap) {
	if (!hasContents) {	// no lock needed--setting it twice is harmless
	    hasContents = true;
	}

	EntryHandle handle = newHandle(rep, txn);
	contents.add(handle);

	if (addToMap) {
	    idMap.put(rep.getCookie(), handle);
	}

	if (numFields < 0)	// no lock needed--setting it twice is harmless
	    numFields = rep.numFields();

	EntryHandle head = (EntryHandle) contents.head();
	if (head != null && head != handle)
	    rep.shareWith(head.rep());
	return handle;
    }

    // We only call this method when we own <code>replaceMe<code>'s lock, 
    // this combined with checking to make sure the a handle has not be
    // removed after we acquire the lock but before we use the handle
    // for real makes this operation safe and
    void replace(EntryHandle replaceMe, EntryHandle with, 
		 TransactableMgr mgr) 
    {
	// We require that that both handles are for the same
	// underlying EntryRep, otherwise such a replacement could
	// change the mapping of cookies to EnteryReps.
	if (replaceMe.rep() != with.rep())
	    throw new InternalSpaceException("SimpleEntryHolder.replace:" +
	        "trying to replace one handle with another that has a " +
		"different entry rep");

	contents.addAfter(replaceMe, with);
	contents.remove(replaceMe);
	idMap.put(with.rep().getCookie(), with);
    }

    /**
     * Override if you need to return a subtype of <code>EntryHandle</code>.
     */
    EntryHandle newHandle(EntryRep rep, TransactableMgr txn) {
	if (txn == null)
	    return new EntryHandle(rep);
	else {
	    TxnEntryHandle h =
		new TxnEntryHandle(rep, txn, this, TransactableMgr.WRITE);
	    txn.add(h);
	    return h;
	}
    }

    // inherit doc comment from superclass
    public RepEnum contents(TransactableMgr mgr) {
	return new SimpleRepEnum(mgr);
    }

    /**
     * The class that implements <code>RepEnum</code> for this class.
     */
    private class SimpleRepEnum implements RepEnum {
	/** The last node we saw */
	private EntryHandle handle;
	private TransactableMgr mgr;
	private long startTime;
	private final java.io.PrintWriter dbgEnum =
					    BasicSpace.debug.getWriter("enums");
	int count = 0;

	SimpleRepEnum(TransactableMgr mgr) {
	    if (contents == null) //!! work around ODI bug with inner classes
		COM.odi.ObjectStore.fetch(SimpleEntryHolder.this);
	    this.mgr = mgr;
	    handle = (EntryHandle) contents.head();
	    startTime = System.currentTimeMillis();
	}

	// inherit doc comment from superclass
	public EntryRep nextRep() {
	    if (dbgEnum != null) {
	        dbgEnum.println("SimpleEntryHolder.SimpleRepEnum: nextRep:" +
		    " just inside");
	    }


	    //skip over handles which are either removed or unable
	    //to perform a READ operation

	    while (handle != null &&
		(!handle.canPerform(mgr, TransactableMgr.READ) ||
		 isExpired(startTime, handle) || handle.removed()))
	    {
		handle = (EntryHandle) handle.next();

		if (dbgEnum != null) {
		    dbgEnum.println("SimpleEntryHolder.SimpleRepEnum: " +
			" nextRep: handle= " + handle + " handle.next= " +
			handle.next);
		}
	    }

	    if (handle == null)
		return null;

	    EntryHandle h = handle;
	    handle = (EntryHandle) h.next();
	    return h.rep();
	}
    }

    /**
     * Remove the given handle from this holder and the <code>idMap</code>.
     * If the handle isn't in this holder, this does nothing.
     */
    boolean remove(EntryHandle h) {
	boolean ok = contents.remove(h);
	if (ok)
	    idMap.remove(h.rep().getCookie());
	return ok;
    }

    /**
     * Return the handle for the given <code>EntryRep</code> object.  
     * This is done via lookup in the <code>idMap</code>.  The 
     * <code>idMap</code> is doing double duty here.  The 
     * <code>LeaseDesc</code> associated with an <code>EntryRep</code> 
     * is also the rep's <code>EntryHandle</code>.
     */
    private EntryHandle handleFor(EntryRep rep) {
	return (EntryHandle) idMap.get(rep.getCookie());
    }

    // inherit doc comment from superclass
    public boolean cancel(LeaseDesc desc) {
	EntryHandle h = (EntryHandle)desc;
	if (h == null)
	    return false;
	remove(h);
	return true;
    }


    // inherit doc comment from superclass
    public void reap() {
	contents.reap();
    }

}
