/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import net.jini.space.*;
import net.jini.core.entry.*;
import net.jini.core.event.*;
import net.jini.core.lease.*;
import net.jini.core.transaction.*;
import com.sun.jini.debug.Debug;
import net.jini.admin.Administrable;

import java.rmi.*;
import java.rmi.server.*;
import java.util.*;
import java.io.*;

/**
 * This class is the client-side proxy for the <code>BasicSpace</code>
 * implementation of a JavaSpaces<sup><font size=-2>TM</font></sup> service. 
 * <code>BasicSpace</code> implements the <code>OutriggerSpace</code> 
 * interface, and each <code>SpaceProxy</code> object holds a reference 
 * to the remote OutriggerSpace server it represents to the client. The 
 * client makes calls from the <code>JavaSpace</code> interface, which 
 * the <code>SpaceProxy</code> translates into appropriate
 * <code>OutriggerSpace</code> calls to the <code>BasicSpace</code>
 * server.
 *
 * @see OutriggerSpace
 * @see BasicSpace
 * @author Ken Arnold
 */
class SpaceProxy implements JavaSpace, Administrable, Serializable {
    static final long serialVersionUID = 8346504944797488306L;

    /**
     * The remote space this proxy works with.
     */
    private OutriggerSpace space;

    /**
     * The hashtable of <code>Answer</code> objects from blocking read
     * and take calls.  This is only used/allocated on the client side.
     *
     * @see #getMatch
     * @see SpaceProxy.Answer
     */
    private transient Hashtable results;

    /**
     * The listener for this proxy.
     *
     * @see #getMatch
     */
    private transient Listener listener;

    // we have our own copy so we don't have to load BasicSpace into the client
    private static final String	DEBUG_PROP = "com.sun.jini.outrigger.debug";
    private static final Debug 	debug = newDebug();

    static final PrintWriter	dbgProxy = 
	(debug != null)?debug.getWriter("proxy"):null;

    // --------------------------------------------------
    //		Construction
    // --------------------------------------------------

    /**
     * Create a new <code>SpaceProxy</code> for the given space.
     */
    SpaceProxy(OutriggerSpace space) {
	this.space = space;
    }

    /**
     * Allocate a new <code>Debug</code> object based on properties. 
     * Will return <code>null</code> if accessing the VM properties results
     * is a <code>SecurityException</code>.
     */
    private static Debug newDebug() {
	try {
	    return new Debug(DEBUG_PROP);
	} catch (SecurityException e) {
	    // Ignore, most people will nether use the debug property, nor 
	    // give the proxy permission to read properties. 
	}
	return null;
    }

    // inherit doc comment
    public boolean equals(Object other) {
	if (other instanceof SpaceProxy)
	    return space.equals(((SpaceProxy) other).space);
	return false;
    }

    // inherit doc comment
    public int hashCode() {
	return space.hashCode();
    }

    /**
     * Read this object back.  This exists so that the <code>results</code> 
     * hashtable can be allocated -- it is not transmitted.
     */
    private void readObject(ObjectInputStream in)
	throws IOException, ClassNotFoundException
    {
	in.defaultReadObject();
	results = new Hashtable();
    }

    // --------------------------------------------------
    //		JavaSpace method implementations
    // --------------------------------------------------

    // inherit doc comment
    public Lease write(Entry entry, Transaction txn, long lease)
	throws TransactionException, RemoteException
    {
	if (entry == null)
	    throw new NullPointerException("Cannot write null Entry");
	return space.write(repFor(entry), txn, lease);
    }

    // inherit doc comment, use internal routine for common code
    public Entry read(Entry tmpl, Transaction txn, long timeout)
	throws UnusableEntryException, TransactionException,
	       InterruptedException, RemoteException
    {
						 // take?, ifExists?
	return getMatch(repFor(tmpl), txn, timeout, false, false);
    }

    // inherit doc comment, use internal routine for common code
    public Entry readIfExists(Entry tmpl, Transaction txn, long timeout)
	throws UnusableEntryException, TransactionException,
	       InterruptedException, RemoteException
    {
						 // take?, ifExists?
	return getMatch(repFor(tmpl), txn, timeout, false, true);
    }

    // inherit doc comment, use internal routine for common code
    public Entry take(Entry tmpl, Transaction txn, long timeout)
	throws UnusableEntryException, TransactionException,
	       InterruptedException, RemoteException
    {
						 // take?, ifExists?
	return getMatch(repFor(tmpl), txn, timeout, true,  false);
    }

    // inherit doc comment, use internal routine for common code
    public Entry takeIfExists(Entry tmpl, Transaction txn, long timeout)
	throws UnusableEntryException, TransactionException,
	       InterruptedException, RemoteException
    {
						 // take?, ifExists?
	return getMatch(repFor(tmpl), txn, timeout, true,  true);
    }

    // inherit doc comment
    public Entry snapshot(Entry entry) {
	if (entry == null)
	    return null;
	else
	    return new SnapshotRep(entry);
    }

    // inherit doc comment
    public EventRegistration
	notify(Entry tmpl, Transaction txn, RemoteEventListener listener,
	       long lease, MarshalledObject handback)
	throws TransactionException, RemoteException
    {
	if (listener == null)
	    throw new NullPointerException("Passed null listener for event " +
					   "registration");
	return space.notify(repFor(tmpl), txn, listener, lease, handback);
    }

    // --------------------------------------------------
    //          Administrable methods implementation
    // --------------------------------------------------
    // inherit doc comment
    public Object getAdmin() throws RemoteException {
	return space.getAdmin();
    }

    // --------------------------------------------------
    //		Private implementation
    // --------------------------------------------------

    /**
     * Get a match from the space.  This is common to all the read and
     * take methods because the all basically do the same thing, with
     * some behavioral differences easily represented as flags.
     */
    private Entry
	getMatch(EntryRep tmpl, Transaction txn, long timeout, boolean takeIt,
		 boolean ifExists)
	throws UnusableEntryException, TransactionException, RemoteException,
	       InterruptedException
    {
	if (dbgProxy != null)
	    dbgProxy.println("SpaceProxy: getMatch");

        if (timeout < 0) 
	    throw new IllegalArgumentException(
		"Timeout values must be non-negative");

	long start = System.currentTimeMillis();// remember start time
	if (dbgProxy != null)
	    dbgProxy.println("timeout =\t" + timeout);

	// setup values used in blocking for values
	RemoteEventListener myListener;
	Answer answer;
	long answerID;
	if (timeout == NO_WAIT) {
	    myListener = null;
	    answer = null;
	    answerID = 0;
	} else {
	    // initialize listener if not already done
	    if (listener == null)
		createListener();
	    myListener = listener;
	    answer = new Answer(); // adds to results map
	    answerID = answer.id();
	}

	// see if there already is a match

	try {
	    EntryRep rep;
	    EventRegistration ev;

	    Object obj;
	    try {
		obj = space.getMatch(tmpl, txn, timeout, takeIt, ifExists,
				     myListener, answerID);
	    } catch (TransactionConflictException e) {
		if (myListener == null)
		    return null;
		throw new InternalSpaceException(
		    "TransactionConflictException leaked through");
	    }
	    rep =(obj instanceof EventRegistration ? null : (EntryRep) obj);
	    ev = (obj instanceof EntryRep ? null : (EventRegistration) obj);
	    if (dbgProxy != null)
		dbgProxy.println("first getMatch returned " + obj);

	    // hey, it worked!  Or worked as much as we need, anyway
	    if (rep != null || timeout == NO_WAIT)
		return entryFrom(rep);

	    if (rep == null && ifExists && ev == null)
		return null;

	    if (ev != null)			// if a registration happened
		answer.register(ev);		// ... remember it

	    long roundtrip = -1;		// the roundtrip time.
	    Entry result;
	    boolean conflict;			// only conflicting match found

	    /*
	     * Loop, trying repeatedly to satisfy the request.  We loop
	     * because of race conditions -- a matching entry might be 
	     * written, but taken before we get a chance.
	     */
	    do {
		conflict = false;	// haven't found a conflict this time

		// how long should I be interested?
		long now = System.currentTimeMillis();
		long elapsed = now - start;
		long remain = timeout - elapsed;

		// time already ran out
		if (remain <= 0)
		    return null;

		// basic dumb value for round trip
		if (roundtrip < 0)
		    roundtrip = elapsed;

		// figure out how long to wait
		Lease lease = answer.event.getLease();
		long expiresIn = lease.getExpiration() - now;
		long waitFor;
		if (remain <= expiresIn)
		    waitFor = remain;
		else {
		    long renewTime = expiresIn - 5 * roundtrip;
		    waitFor = Math.min(remain, renewTime);
		}

		if (dbgProxy != null) {
		    dbgProxy.println("elapsed =\t" + elapsed);
		    dbgProxy.println("remain =\t" + remain);
		    dbgProxy.println("roundtrip =\t" + roundtrip);
		    dbgProxy.println("expiresIn =\t" + expiresIn);
		    dbgProxy.println("waitFor =\t" + waitFor);
		}

		// if remaining time is short, just wait the requisite time
		// and try again.
		if (waitFor <= roundtrip) {
		    answer.unregister();		// stop listening
		    Thread.sleep(remain);
		    // return whatever is found
		    try {
			return entryFrom(
			    space.getMatch(tmpl, txn, NO_WAIT, takeIt, ifExists)
			);
		    } catch (TransactionConflictException e) {
			return null;
		    }
		}

		synchronized (answer) {
		    // notification may have come just before
		    if (!answer.notified) {
			if (dbgProxy != null)
			    dbgProxy.println("waiting");
			answer.wait(waitFor); 
		    }

		    if (dbgProxy != null)
			dbgProxy.println("notified = " + answer.notified);

		    if (answer.notified) {	// got an answer: try it
			// if the answer was to throw something, throw it
			if (answer.toThrow != null) {
			    Exception exc = answer.toThrow;
			    if (dbgProxy != null) {
				dbgProxy.println("throwing " + exc);
				new Throwable().printStackTrace(dbgProxy);
			    }
			    if (exc instanceof TransactionException)
				throw (TransactionException) exc;
			    if (exc instanceof RuntimeException)
				throw (RuntimeException) exc;
			    if (exc instanceof UnusableEntryException)
				throw (UnusableEntryException) exc;
			    if (exc instanceof InterruptedException)
				throw (InterruptedException) exc;
			    if (exc instanceof RemoteException)
				throw (RemoteException) exc;
			    throw new UnexpectedException("toThrow type", exc);
			}
			try {
			    /*
			     * Attempt to get a match using a "probe" timeout.
			     * @see  com.sun.jini.outrigger.OutriggerSpace#getMatch
			     */
			    rep = space.getMatch(tmpl, txn, 
						 OutriggerSpace.INTERNAL_PROBE, 
				                 takeIt, ifExists);
			    result = entryFrom(rep);
			} catch (TransactionConflictException e) {
			    conflict = true;
			    result = null;
			}
			answer.notified = false;
		    } else if (waitFor < remain) {  // no answer, must renew
			if (dbgProxy != null)
			    dbgProxy.println("renewing lease");
			try {
			    lease.renew(remain);
			} catch (LeaseException e) {
			    if (dbgProxy != null)
				dbgProxy.println("renewing lease failed: " + e);
			    return null;	    // time ran out
			}
			result = null;
		    } else {			    // no answer, end of time
			if (dbgProxy != null)
			    dbgProxy.println("time ran out, return null");
			return null;
		    }
		}

		// skip out of we've got a clear answer on an ifExists
		if (result == null && ifExists && !conflict)
		    return null;

	    } while (result == null);		// repeat until we get a result
	    return result;
	} finally {
	    // ALWAYS unregister (removes answer from results)
	    if (answer != null)
		answer.unregister();
	}
    }

    /**
     * The ID used for the next Answer object.  This is protected from
     * concurrent access by <code>Answer.class</code> -- if the Java
     * programming language would only allow static members of inner
     * classes, this would be one.
     *
     * @see #Answer 
     */
    private static long nextAnswerID = 1;

    /**
     * <code>Answer</code> objects represent notifications of matching
     * entries.  Each object has an ID that represents a particular
     * registration.  The fields of this class are directly accessed
     * by <code>SpaceProxy.getMatch</code>.  Each object is created
     * and discarded by <code>getMatch</code>, and is briefly used by
     * <code>Listener.notify</code>.
     */
    private class Answer {
	Long		  myID;		// my ID for which event I'm waiting on
	boolean		  notified;	// have I been notified?
	EventRegistration event;	// the event registration record
	Exception	  toThrow;	// to be thrown (if not null)

	/**
	 * Create a new <code>Answer</code> object, assigning it its ID.
	 * The new object is added to the <code>results</code> map: call
	 * <code>unregister</code> to remove it.
	 */
	Answer() {
	    notified = false;
	    event = null;
	    synchronized (Answer.class) {
		myID = new Long(nextAnswerID++);
	    }
	    results.put(myID, this);
	}

	public String toString() {
	    return ("Answer: " + myID + "/" + event + ", " +
		    "notified = " + notified);
	}

	/**
	 * Return this object's current ID.
	 */
	long id() {
	    return myID.longValue();
	}

	/**
	 * Record a registration of interest in a given event.  This is
	 * called with the return value of the notification call on the
	 * space.  It sets the appropriate fields of the <code>Answer</code>
	 * object for this thread.
	 */
	void register(EventRegistration event) {
	    if (dbgProxy != null)
		dbgProxy.println("register(" + event + ")");
	    this.event = event;
	}

	/**
	 * Unregister interest in events, if any, associated with this object,
	 * cancelling interest in the event and removing the relevant
	 * entries from the <code>results</code> table.
	 */
	void unregister() {
	    if (dbgProxy != null)
		dbgProxy.println("unregister(), event is " + event);
	    results.remove(myID);
	    if (event != null) {
		try {
		    event.getLease().cancel();
		} catch (UnknownLeaseException e) {
		    ;	// that's okay
		} catch (RemoteException e) {
		    ;	// we tried -- we'll catch it on the next notification
		}
		event = null;
	    }
	}
    }
    /**
     * Each proxy has one <code>RemoteEventListener</code> object that is
     * registered as the interested party for matches.  This method
     * creates the object if this is the first time one is needed.
     */
    private synchronized void createListener() throws RemoteException {
	if (listener == null)
	    listener = new Listener(this);
    }

    /**
     * The marker interface implemented by our listener.  This is a
     * marker interface because the server wants to know about these
     * internal registrations and treat them separately.
     */
    interface ProxyListener extends RemoteEventListener {
    }

    /**
     * The <code>Listener</code> is used by the <code>SpaceProxy</code>
     * to be the object that is notified when a matching entry is written
     * into the space.
     */
    class Listener extends UnicastRemoteObject implements ProxyListener {
	/** The object I'm a proxy for. */
	private SpaceProxy spaceProxy;

	/**
	 * Create a new <code>Listener</code> for the given proxy.
	 */
	Listener(SpaceProxy p) throws RemoteException {
	    spaceProxy = p;
	}

	/**
	 * Called when a matching entry is written to the space.
	 */
	public void notify(RemoteEvent ev)
	    throws UnknownEventException
	{
	    Long evIDObj;
	    Exception toThrow = null;
	    try {
		Object payload = ev.getRegistrationObject().get();
		if (payload instanceof Long)
		    evIDObj = (Long) payload;
		else {
		    ThrowThis th = (ThrowThis) payload;
		    evIDObj = th.id();
		    toThrow = th.toThrow();
		}
		if (SpaceProxy.dbgProxy != null)
		    SpaceProxy.dbgProxy.println("notify(" + evIDObj + ")");
	    } catch (ClassNotFoundException e) {
		throw new InternalSpaceException("decoding long?", e);
	    } catch (java.io.IOException e) {
		throw new InternalSpaceException("decoding long?", e);
	    }
	    Answer answer = (Answer) spaceProxy.results.get(evIDObj);
	    long evID = evIDObj.longValue();
	    // if the event ID is out of date, or if for some reason it is
	    // talking about a different event that we have registered, we
	    // don't know about it anymore
	    if (answer == null || answer.id() != evID) {
		if (dbgProxy != null) {
		    dbgProxy.print("bad notify: ");
		    if (answer == null) {
			dbgProxy.print("answer == null");
		    } else if (answer.id() != evID) {
			dbgProxy.print("answer.id() {" + answer.id() + "}"
			    + " != evID {" + evID + "}");
		    } else {
			dbgProxy.print("WHY???");
		    }
		    dbgProxy.println();
		}
		throw new UnknownEventException();
	    }
	    // Let our thread know that we have an answer for it.
	    synchronized (answer) {
		answer.toThrow = toThrow;
		answer.notified = true;
		answer.notifyAll();
	    }
	    return;
	}
    }

    /**
     * Return an <code>EntryRep</code> object for the given <code>Entry</code>.
     */
    static EntryRep repFor(Entry entry) {
	if (entry == null)
	    return null;
	if (entry instanceof SnapshotRep)	// snapshots are pre-calculated
	    return ((SnapshotRep) entry).rep();
	return new EntryRep(entry);
    }

    /**
     * Return an entry generated from the given rep.
     */
    static Entry entryFrom(EntryRep rep) throws UnusableEntryException {
	if (rep == null)
	    return null;
	return rep.entry();
    }
}
