/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import com.sun.jini.debug.Debug;
import com.sun.jini.mahout.ServiceConfig;
import com.sun.jini.start.ServiceStarter;

import net.jini.space.JavaSpace;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.rmi.activation.ActivationException;
import java.rmi.activation.ActivationGroup;
import java.rmi.activation.ActivationGroupDesc;
import java.rmi.activation.ActivationGroupDesc.CommandEnvironment;
import java.rmi.activation.ActivationGroupID;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Properties;
import java.util.StringTokenizer;

//////////
// !! JWFM Work around for Manifest-bootclasspath bug (id 4164467) in 
// JDK1.2-beta-4
import java.util.jar.JarFile;
import java.util.jar.Manifest;
import java.util.jar.Attributes;

/**
 * This class contains the public methods and command line interface for
 * creating an activatable instance of the outrigger implementation of the
 * JavaSpace service. This class empoloys the utility class 
 * <code>com.sun.jini.start.ServiceStarter</code> to start the desired
 * service. That utility class requires that this class define a
 * <code>public static create</code> method which performs all functions
 * that are specific to the desired service, and which can be invoked by the
 * the utility class.
 *
 * @see com.sun.jini.start.ServiceStarter
 */
public class StartPersistentOutrigger {

    /**
     * Utility for outputting debug messages.
     */
    private static final Debug debug = new Debug(BasicSpace.DEBUG_PROP);

    /**
     * Print writer for startup related messages.
     */
    private static final PrintWriter dbgStartup = debug.getWriter("startup");

    /** 
     * The fully-qualified path.classname of this class, which will be
     * passed to the <code>com.sun.jini.start.ServiceStarter</code>
     * utility class
     */
    private static String starterClass = 
	StartPersistentOutrigger.class.getName();

    /** 
     * The fully-qualified path.classname of the class that defines the
     * implementation of the backend server of the service that is to be
     * started
     */
    private static String implClass = FrontEndSpace.class.getName();

    /** 
     * The name used to identify the particular resource bundle from which
     * usage/error message information may be retrieved for display
     * (helpful with internationalization)
     */
    private static String resourcename = "service";

    /**
     * Command line interface for starting an instance of the service
     * referenced in the <code>implClass</code> field
     * <p>
     *
     * @param args <code>String</code> array containing the command line
     *             arguments
     *
     * @see com.sun.jini.start.ServiceStarter
     */
    public static void main(String[] args) throws Throwable {
        JavaSpace proxy = (JavaSpace) 
	    ServiceStarter.create(args, starterClass, implClass, resourcename);
	ServiceConfig.config(proxy, "Outrigger",
	    BasicSpace.SPACE_NAME_PROP, BasicSpace.DEFAULT_NAME);
    }//end main

    /**
     * This method defines how the service referenced in the
     * <code>implClass</code> field of this class is to be started.
     * This method will be invoked through reflection by the utility
     * class <code>com.sun.jini.start.ServiceStarter</code>. This method
     * performs all functions that require knowledge of, or access to,
     * the service-specific information; information that the utility
     * class cannot easily obtain or exploit while attempting to start
     * the desired service.
     * <p>
     * This method returns a proxy that provides client-side access to
     * the backend server of the started service.
     *
     * @param serverStub a <code>Remote</code> reference to the backend server
     *                   of the started service
     * 
     * @return an instance of the public interface to the desired service
     *         referenced in the <code>implClass</code> field of this class;
     *         that is, a proxy that provides client-side access to the
     *         backend server of the started service
     * 
     * @throws java.rmi.activation.ActivationException this is a general
     *         exception thrown by the activation system to indicate a
     *         a number of exceptional conditions that prevent the
     *         activation system from fulfilling a particular request.
     * 
     * @throws java.io.IOException 
     *
     * @see com.sun.jini.start.ServiceStarter
     */
    public static JavaSpace create(Remote serverStub)
        throws ActivationException, IOException
    {
        if (dbgStartup != null) 
            dbgStartup.println("StartPersistentOutrigger::create(stub=" + 
		serverStub + ")");
	
	return new SpaceProxy((OutriggerSpace)serverStub);
    }//end create


    /**
     * Creates an activation group suitable for an outrigger service to 
     * be registered with the activation system. This method is called
     * from the <code>ServiceStarter</code> framework prior to the call to 
     * <code>create</code>.
     *
     * @param classpath  the classpath for the service to be registered with
     *                   the activation system
     * @param codebase   the codebase for the service to be registered with
     *                   the activation system
     * @param policyFile the security policy file to apply to the service
     *                   to be registered with the activation system
     * @param properties if not <code>null</code>, this argument contains
     *                   the properties for the activation group in which
     *                   the service will be registered
     * @param serverJVM  if not <code>null</code>, the contents of this
     *                   argument represents a path and command that allows
     *                   parties that invoke this method to specify the
     *                   JVM implementation to use for the execution of the
     *                   service's backend server. The value of this argument
     *                   must be an absolute path to an executable entity and
     *                   that executable entity must accept all of the
     *                   standard JVM command-line options (for example,
     *                   '-D<property>, -classpath <classpath>', and so on),
     *                   and directly or indirectly start a JVM. Note that
     *                   although the use of this argument is not necessary
     *                   in most situations, it is useful for testing the
     *                   service on alternate JVM implementations. 
     * @param options    if not <code>null</code>, this argument contains
     *                   the command options for the JVM in which the
     *                   service's backend server executes
     * 
     * @throws java.rmi.activation.ActivationException this is a general
     *         exception thrown by the activation system to indicate a
     *         a number of exceptional conditions that prevent the
     *         activation system from fulfilling a particular request.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the activation 
     *         system's VM and the VM in which this class is instantiated.
     *         When this exception does occur, the ActivationGroup may
     *         or may not have been created successfully.
     */
    public static ActivationGroupID createGroup(final String classpath,
        final String codebase, final String policyFile, 
        final Properties properties, final String serverJVM, 
        final String[] options) 
        throws ActivationException, RemoteException, IOException
    {
        if (dbgStartup != null) { 
            dbgStartup.print("StartPersistentOutrigger::createGroup(\n" +
                "\tclasspath=" + classpath + "\n" +
                "\tcodebase=" + codebase + "\n" +
                "\tpolicyFile=" + policyFile + "\n" +
                "\tproperties=" + properties + "\n" +
                "\tserverJVM=" + serverJVM + "\n" +
                "\toptions=");
            if (options != null) {
                for (int i=0; i < options.length; i++) {
                    dbgStartup.print(options[i]);
                    if (i != (options.length - 1))
                        dbgStartup.print(",");
		    else
                        dbgStartup.print("\n");
		}
	    } else {
                dbgStartup.print("null\n");
	    }
            dbgStartup.print("\n");
	}

	Properties props = null;

	if (properties == null)
	    props = new Properties();
	else
	    props = (Properties)properties.clone();

	props.put("java.security.policy", policyFile);
	props.put("java.rmi.server.codebase", codebase);

	//////////
	// !! JWFM Work around for Manifest-bootclasspath bug (id 4164467) in
	// JDK1.2-beta-4
	StringTokenizer st = new StringTokenizer(classpath, File.pathSeparator);
	String bootPath = "-Xbootclasspath/p:";
	while (st.hasMoreTokens()) {
	    final String path = st.nextToken();

	    if (path.endsWith(".jar")) {
		final int i = path.lastIndexOf('/');
		final String dir = 
		    path.substring(0, i + 1).replace('/',  File.separatorChar);
		final String jar = path.substring(i+1);
		final ArrayList jfs = new ArrayList(1);
		addJar(jar, jfs, dir);
		for (Iterator j = jfs.iterator(); j.hasNext();)
		    bootPath = bootPath + File.pathSeparator + j.next();
	    } else {
		bootPath = bootPath + File.pathSeparator + path;
	    }
	}
	// End Workaround

	String argv[] = null;
	if (options == null) {
	    argv = new String[]{bootPath};
	} else {
	    String[] nargv = new String[options.length + 1];
	    System.arraycopy(options, 0, nargv, 0, options.length);
	    nargv[options.length] = bootPath;
	    argv = nargv;
	}

        if (dbgStartup != null) { 
            dbgStartup.println("StartPersistentOutrigger::createGroup() - " +
                "Processed options: ");
            for (int i=0; i < argv.length; i++) {
                dbgStartup.println("\t" + argv[i]);
	    }
            dbgStartup.println("StartPersistentOutrigger::createGroup() - " +
                "Processed properties: ");
            dbgStartup.println(props.toString());
	}

	CommandEnvironment cmd = 
	    new CommandEnvironment(serverJVM, argv);
	return ActivationGroup.getSystem().registerGroup(
	           new ActivationGroupDesc(props, cmd));
    }        

    //////////
    // !! JWFM Work around for Manifest-bootclasspath bug (id 4164467) in 
    // JDK1.2-beta-4
    /**
     * Place the path to the passed jar file and recursively any classpaths
     * in the jar file's manifest in the passed array list.
     * @param jar The jar file to work on, should only be the file part
     *            of the path
     * @param jfs The array list to return the results in
     * @param dir The directory the jar file is in
     */
    private static void addJar(String jar, ArrayList jfs, String dir)
	throws IOException
    {
	JarFile jf = new JarFile(dir + jar);
	jfs.add(jf.getName());
	Manifest man = jf.getManifest();
	if (man == null)
	    return;
	Attributes attrs = man.getMainAttributes();
	if (attrs == null)
	    return;
	String val = attrs.getValue(Attributes.Name.CLASS_PATH);
	if (val == null)
	    return;
	int i = jar.lastIndexOf(File.separatorChar);
	dir = dir + jar.substring(0, i + 1).replace('/', File.separatorChar);
	StringTokenizer st = new StringTokenizer(val);
	while (st.hasMoreTokens()) {
	    addJar(st.nextToken(), jfs, dir);
	}
    }
}
