/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import com.sun.jini.debug.Debug;
import com.sun.jini.mahout.ServiceConfig;
import com.sun.jini.start.ServiceStarter;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.rmi.RMISecurityManager;

import net.jini.space.JavaSpace;

/**
 * This class contains the public methods and command line interface for
 * creating a transient instance of the outrigger implementation of the
 * JavaSpace service. This class empoloys the utility class 
 * <code>com.sun.jini.start.ServiceStarter</code> to start the desired
 * service. That utility class requires that this class define a
 * <code>public static create</code> method which performs all functions
 * that are specific to the desired service, and which can be invoked by the
 * the utility class.
 *
 * @see com.sun.jini.start.ServiceStarter
 */
public class StartTransientOutrigger {

    /**
     * Utility for outputting debug messages.
     */
    private static final Debug debug = new Debug(BasicSpace.DEBUG_PROP);

    /**
     * Print writer for startup related messages.
     */
    private static final PrintWriter dbgStartup = debug.getWriter("startup");

    /** 
     * The fully-qualified path.classname of this class, which will be
     * passed to the <code>com.sun.jini.start.ServiceStarter</code>
     * utility class
     */
    private static String starterClass = 
	StartTransientOutrigger.class.getName();

    /** 
     * The fully-qualified path.classname of the class that defines the
     * implementation of the backend server of the service that is to be
     * started
     */
    private static String implClass = TransientSpace.class.getName();

    /** 
     * The name used to identify the particular resource bundle from which
     * usage/error message information may be retrieved for display
     * (helpful with internationalization)
     */
    private static String resourcename = "service";

    /**
     * Command line interface for starting a transient instance
     * of the desired service.
     * <p>
     *
     * @param args <code>String</code> array containing the command line
     *             arguments
     *
     * @see com.sun.jini.start.ServiceStarter
     */
    public static void main(String[] args) throws Throwable {
	// Prepend dir arg to args in order to passing
	// service starter parsing requirements
	String[] fakeArgs = new String[args.length + 1];
	File tmp = null;
	tmp = File.createTempFile("Outrigger", null);
	tmp.delete();
	fakeArgs[0] = tmp.getAbsolutePath();
	System.arraycopy(args, 0, fakeArgs, 1, args.length);
	// create Transient will also set the security manager
        JavaSpace proxy = (JavaSpace) 
	    ServiceStarter.createTransient(
	        fakeArgs, starterClass, 
	        System.getProperty("java.class.path"),
	        System.getProperty("java.rmi.server.codebase"),
                resourcename);
        ServiceConfig.config(proxy, "Outrigger", 
	    BasicSpace.SPACE_NAME_PROP, BasicSpace.DEFAULT_NAME);
    }//end main

    /**
     * This method defines how the service referenced in the
     * <code>implClass</code> field of this class is to be started.
     * This method will be invoked through reflection by the utility
     * class <code>com.sun.jini.start.ServiceStarter</code>. This method
     * performs all functions that require knowledge of, or access to,
     * the service-specific information; information that the utility
     * class cannot easily obtain or exploit while attempting to start
     * the desired service.
     * <p>
     * This method returns a proxy that provides client-side access to
     * the backend server of the started service.
     *
     * @param logDirname <code>String</code> containing the full path and
     *                   directory name in which the state of the service to
     *                   be created will be recorded for persistence.
     *                   This is ignored by the transient instance.
     *
     * @return an instance of the public interface to the service that is
     *         to be started by this method; that is, a proxy that provides
     *         client-side access to the backend server of the started service
     * 
     * @throws java.io.IOException 
     *
     * @see com.sun.jini.start.ServiceStarter
     */
    public static JavaSpace create(String logDirname) throws IOException
    {
        // Service Starter sets up a security manager but just in case
        // someone calls this method directly.
        if(System.getSecurityManager() == null) {
            System.setSecurityManager(new RMISecurityManager());
        }

        if (dbgStartup != null) 
            dbgStartup.println("StartTransientOutrigger::create(log=" + 
		logDirname + ")");
        final TransientSpace ts = new TransientSpace();

	return ts.getSpaceProxy();
    }//end create
}
