/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import COM.odi.util.OSVector;

/**
 * A handle to a stored template, used for notification checks.
 * Because the handle hash codes and masks are relative to the number
 * of fields, this stores the <code>EntryHandleTmplDesc</code> for each
 * field count against which this template has been compared.  We do
 * this as a vector because the number of fields is likely to be
 * relatively small (we do not expect entries with thousands of
 * fields), and the lookup speed we gain is important.  This handle
 * stores the list of chits that have registered with equivalent
 * templates.
 *
 * @author Ken Arnold
 *
 * @see TemplateHolder
 * @see TemplateHolderSet
 */
class TemplateHandle extends EntryHandle {
    private long	numMatches;	// count of times this matched something
    private OSVector	descs;		// the descriptions for each field count
    private OSVector	chits;		// the list of registered interest chits
    private boolean	onAbort;	// notify if the txn is aborted

    /**
     * Create a handle for the template <code>tmpl</code>.
     */
    TemplateHandle(EntryRep tmpl) {
	super(tmpl);
	descs = new OSVector();
	chits = new OSVector();
    }

    /**
     * Return the description for the given field count.
     */
    //!! Since the mask/hash algorithm tops out at a certain number of fields,
    //!! we could avoid some overhead by topping out at the same count.
    EntryHandleTmplDesc descFor(int numFields) {
	Object o;
	if (numFields < descs.size() &&
	    (o = descs.elementAt(numFields)) != null)
	{
	    return (EntryHandleTmplDesc) o;
	}
	EntryHandleTmplDesc desc = descFor(rep(), numFields);
	descs.setSize(numFields + 1);
	descs.setElementAt(desc, numFields);
	return desc;
    }

    /**
     * Return <code>true</code> if this template matches the given entry,
     * bumping the match count.
     */
    boolean matches(EntryRep entry) {
	return rep().matches(entry);
    }

    /**
     * Return <code>true</code> if this template matches the given entry,
     * bumping the match count.
     */
    synchronized void usedMatch() {
	numMatches++;
    }

    /**
     * Return the event ID for this handle.
     */
    public long id() {
	return rep().id;
    }

    /**
     * Return the number of matches on this handle.
     */
    long numMatches() {
	return numMatches;
    }

    /**
     * Set the sequence ID to the given value if that value is less than
     * the current value.  Otherwise the call has no effect.
     */
    synchronized void changeSeq(long newSeq) {
	if (numMatches < newSeq)
	    numMatches = newSeq;
    }

    /**
     * Set the event ID for this handle.
     */
    public void id(long id) {
	if (id() != 0)
	    throw new IllegalArgumentException("id already set");
	rep().id(id);
    }

    /**
     * Add a chit to this handle.
     */
    void add(NotifyChit chit) {
	chit.handle(this);
	chits.addElement(chit);
    }

    /**
     * Return the list of registered interest chits this template.
     */
    OSVector chits() {
	return chits;
    }

    /**
     * Set the "notify on abort" flag.  If <code>true</code>, then an
     * abort of the transaction should cause a notification to the listener.
     */
    void notifyOnAbort(boolean nv) {
	onAbort = nv;
    }

    /**
     * Return the "notify on abort" flag.
     * @see #notifyOnAbort(boolean)
     */
    boolean notifyOnAbort() {
	return onAbort;
    }

    /**
     * Return the transaction we are part of.  <code>TemplateHandle</code>
     * is never in a transaction, so it's implementation of the method
     * always returns <code>null</code>.
     */
    TransactableMgr txn() {
	return null;
    }
}
