/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import COM.odi.util.OSVector;
import java.util.Vector;
import java.util.Enumeration;
import net.jini.core.lease.Lease;

/**
 * A holder of templates of a given type.
 *
 * @author Ken Arnold
 *
 * @see TemplateHolderSet
 * @see TemplateHolderHandle
 */
class TemplateHolder extends SimpleEntryHolder {
    static final long serialVersionUID = 8812450886813060000L;

    private static final java.io.PrintWriter dbgEvents =
	BasicSpace.debug.getWriter("events");

    /**
     * Create a new TemplateHolder that shares the given
     * <code>idMap</code>, and which will hold templates of the type
     * named in <code>className</code>.
     */
    TemplateHolder(COM.odi.util.OSHashtable idMap, String className) {
	super(idMap, className);
    }

    /**
     * Invoked by <code>SimpleEntryHolder</code> to create a new
     * handle: we override this to return a
     * <code>TemplateHandle</code>.
     *
     * @see SimpleEntryHolder#add
     */
    EntryHandle newHandle(EntryRep tmpl, TransactableMgr txnMgr) {
	if (txnMgr == null)
	    return new TemplateHandle(tmpl);
	else
	    return new TxnTemplateHandle(tmpl, txnMgr);
    }

    /**
     * Add the template to the list of known templates.  In this case,
     * we only add if there isn't already a match.  This need not be
     * precise -- if parallel checks end up causing two adds that's not
     * critical.  The merging is more of an optimization, not a correctness
     * requirement.
     */
    public Object add(EntryRep tmpl, TransactableMgr txn) {
	for (TemplateHandle handle = (TemplateHandle) contents.head();
	     handle != null;
	     handle = (TemplateHandle) handle.next())
	{
	    if (tmpl.equals(handle.rep())) {
		if (handle.txn() != txn) {
		    if (dbgEvents != null)
			dbgEvents.println("txns don't match, continuing");
		    continue;
		}
		synchronized (handle) {
		    if (!handle.removed()) {
			if (dbgEvents != null)
			    dbgEvents.println("found match using txn " + txn);
			return handle;
		    }
		}
	    }
	}
	if (dbgEvents != null)
	    dbgEvents.println("adding: no match using txn " + txn);
	return super.add(tmpl, txn, false);
    }

    /**
     * Search for all matches of the given <code>rep</code>, adding
     * all the chits registered to a matching template into the
     * <code>results</code> vector.
     */
    public synchronized void allMatches(EntryRep rep, TransactableMgr writeTxn,
	boolean proxyOnly, Vector results)
    {
	if (dbgEvents != null) {
	    dbgEvents.println();
	    dbgEvents.print("TemplateHolder: allMatches, ");
	    dbgEvents.println("numFields = " + numFields);
	}

	if (numFields < 0)	// nothing has ever been added to this list
	    return;

	long entryHash = EntryHandle.hashFor(rep, rep.numFields());
	if (dbgEvents != null) {
	    dbgEvents.print("rep hash = 0x" + Long.toHexString(entryHash));
	    dbgEvents.println(", writeTxn = " + writeTxn);
	}
	for (TemplateHandle handle = (TemplateHandle) contents.head();
	     handle != null;
	     handle = (TemplateHandle) handle.next())
	{
	    // See the if handle mask is incompatible
	    EntryHandleTmplDesc desc = handle.descFor(rep.numFields());
	    if (dbgEvents != null) {
		dbgEvents.print("template hash = 0x" +
				Long.toHexString(desc.hash));
		dbgEvents.print(", mask = 0x" + Long.toHexString(desc.mask));
		dbgEvents.println(", txn = " + handle.txn());
	    }
	    if ((entryHash & desc.mask) != desc.hash)
		continue;

	    /*
	     * It is compatible, now test for a match.  We *don't* synchronize
	     * on the node here because if we mistakenly notify about something
	     * that is being removed that is acceptable.
	     */
	    if (dbgEvents != null) {
		EntryRep inList = handle.rep();
		dbgEvents.println("inList.matches(rep) == " +
		    inList.matches(rep));
	    }

	    if (handle.matches(rep)) {
		if (handle.removed())	// oh, well -- they don't care anymore
		    continue;
		handle.waitUntilStable();
		int addCnt = 0;
		// copy in all the chits
		final OSVector chits = handle.chits();
		synchronized (chits) {
		    Enumeration e = handle.chits().elements();
		    while (e.hasMoreElements()) {
			try {
			    NotifyChit chit = (NotifyChit) e.nextElement();
			    if (proxyOnly) {
				Object list = chit.listener();
				if (!(list instanceof 
				      SpaceProxy.ProxyListener))
				    continue;
			    }
			    if (chit.canSee(writeTxn)) {
				results.addElement(chit);
				addCnt++;
			    }
			} catch (java.rmi.RemoteException re) {
			    // couldn't get the listener, we'll just skip it
			    continue;
			}
		    }
		}
		if (dbgEvents != null) {
		    dbgEvents.println("found template match, added " + addCnt
			+ " chits");
		}
		if (addCnt > 0)
		    handle.usedMatch();
	    }
	}
    }

    // inherit doc comment
    public boolean cancel(LeaseDesc desc) {
	NotifyChit chit = (NotifyChit)desc;
	Object     cookie = chit.getCookie();
	boolean    foundInHandle = false;

	// Remove the chit from its handle, and then the handle from
	// the ourselves if it is empty
	final TemplateHandle handle = chit.handle();
	final OSVector chits = handle.chits();
	synchronized (chits) {
	    for (int i = 0; i < chits.size(); i++) {
		final NotifyChit comp = (NotifyChit)chits.elementAt(i);
		if (cookie.equals(comp.getCookie())) {
		    chits.removeElementAt(i);
		    foundInHandle = true;
		    if (chits.size() == 0)
			remove(handle);
		    break;
		}
	    }
	}
	
	boolean foundInMap =  null != (idMap.remove(chit.getCookie()));
	return foundInMap || foundInHandle;
    }

    /**
     * Search the holder for chits that expired on or before the given time.
     */
    void reapExpired(long after) {
	// Do a search though all the chits in this Holder, removing
	// ones that have expired.
	for (TemplateHandle handle = (TemplateHandle) contents.head();
	     handle != null;
	     handle = (TemplateHandle) handle.next())
	{
	    final OSVector chits = handle.chits();
	    synchronized (chits) {
		for (int i = 0; i < chits.size(); i++) {
		    NotifyChit chit = (NotifyChit)chits.elementAt(i);
		    // if expired, remove it
		    if (BasicSpace.staticSpace.expired(chit, after)) {
			if (dbgEvents != null) 
			    dbgEvents.println("Expiring chit " + 
			        chit.getCookie() + " it expired at " +
                                chit.getExpiration() + " now is " + after);
			chits.removeElementAt(i);
			idMap.remove(chit.getCookie());
			BasicSpace.staticSpace.cancelOp((Long)
							chit.getCookie());
			i--;	// examine this index next time in loop
		    }
		}
		// If after walking over the handle's chits the chits
		// vector is now empty, remove handle
		if (chits.size() == 0)
		    remove(handle);
	    }
	}
    }

    /**
     * Override remove method from <code>SimpleEntryHolder</code>.
     * <code>SimpleEntryHolder</code> stores <code>EntryRep</code>
     * objects in the <code>idMap</code>; we store chits, which don't
     * have a 1:1 mapping with with <code>TemplateHandles</code>, so
     * the manipulation of the <code>idMap</code> in
     * <code>remove</code> is wrong for us.
     */
    boolean remove(EntryHandle h) {
	return contents.remove(h);
    }
}
