/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import net.jini.core.event.*;
import net.jini.core.lease.*;
import com.sun.jini.lease.landlord.LeasePolicy;
import com.sun.jini.mahout.StorableObject;
import net.jini.space.InternalSpaceException;

import java.rmi.MarshalledObject;
import java.util.Vector;
import java.util.Enumeration;

/**
 * A set of <code>TemplateHolder</code> objects, indexible by template class.
 *
 * @author Ken Arnold
 *
 * @see TemplateHolder
 */
class TemplateHolderSet extends EntryHolderSet {
    /** The thread driving reaping expired registrations. */
    private transient Thread reaperThread = null;

    private static final boolean DEBUG = false;

    /**
     * Return a new holder.  This is invoked by
     * <code>EntryHolderSet</code>, and overridden here to make the
     * holders <code>TemplateHolder</code> objects.
     */
    EntryHolder newHolder(COM.odi.util.OSHashtable idMap, String className) {
	return new TemplateHolder(idMap, className);
    }

    /**
     * Return a vector that has all the chits that are registered for
     * templates that match the given entry.
     */
    Vector allMatches(EntryRep rep, TransactableMgr txn, boolean proxyOnly) {
	if (DEBUG)
	    System.out.println("TemplateHolderSet: allMatches");
	TemplateHolder holder;
	Vector results = new Vector();

	holder = (TemplateHolder) holderFor(rep.classFor());
	holder.allMatches(rep, txn, proxyOnly, results);
	int last = results.size();
	if (DEBUG)
	    System.out.println("allMatches: " + rep.classFor() + ": " + last);
	// now add in matches with any templates of supertypes of the entry
	for (int i = 0; i < rep.superclasses.length; i++) {
	    holder = (TemplateHolder) holderFor(rep.superclasses[i]);
	    holder.allMatches(rep, txn, proxyOnly, results);
	    if (DEBUG) {
		System.out.println("allMatches: " + rep.classFor() + ": " +
		    (last - results.size()));
		last = results.size();
	    }
	}

	holder = (TemplateHolder) holders.get(EntryRep.matchAnyClassName());
	if (holder!= null)
	    holder.allMatches(rep, txn, proxyOnly, results);

	if (DEBUG)
	    System.out.println("allMatches: " + results.size() + " total");
	return results;
    }

    /**
     * Container for the two values that add() must return
     */
    static class AddRecord {
	EventRegistration	reg;
	NotifyChit		chit;

	AddRecord(EventRegistration reg, NotifyChit chit) {
	    this.reg = reg;
	    this.chit = chit;
	}
    };

    /**
     * Add a new registration of interest for the template
     * <code>tmpl</code>, to notify the given <code>listener</code>.
     * This method allocates the registration cookie and returns it to
     * the invoker, along with the chit itself.
     */
    AddRecord add(EntryRep tmpl, RemoteEventListener listener,
	TransactableMgr txn, long resourceCookie, LeasePolicy leasePolicy, 
	long leaseTime, MarshalledObject handback,
	net.jini.space.JavaSpace space, LogOps logger,
	net.jini.core.transaction.Transaction tr)
    {
	if (DEBUG)
	    System.out.println("TemplateHolderSet: add");
	TemplateHandle handle = handleFor(tmpl, txn);

	NotifyChit chit = (txn == null ?
	    new NotifyChit(listener, resourceCookie, handback) :
	    new TransactableChit(txn, listener, resourceCookie, handback)
	);

	// Create the lease (also sets the chit's expiration)
	Lease lease = null;
	try {
	    lease = leasePolicy.leaseFor(chit, leaseTime);
	    // System.out.println("lease for " + leaseTime + " = " + lease);
	} catch (LeaseDeniedException e) {
	    // This should never happen, we should not be using a
	    // policy that could thrown LeaseDenied on lease
	    // creation, re-throw as InternalSpaceException
	    throw new InternalSpaceException("BasicSpace:TemplateHolderSet:" +
		"leasePolicy.leaseFor threw LeaseDeniedException", e);
	}

	/*
	 * Remember the sequence start before we add it, since after we
	 * add it notifications are possible.  If we waited until later
	 * to get this number, a notification might happen after the add
	 * and before we got the number, leading to a notification with a
	 * lower sequence number than the EventRegistration's starting
	 * value.  Note: It could happen that an event fires after we
	 * get the start but before the add takes effect.  Then we could
	 * return a seq number that is less than [the next event that will
	 * be generated minus one].  This is acceptable because it does
	 * not violate the spec (which says that operations are not ordered
	 * relative to each other) and it does not change which number is
	 * first sent to the listener (which would be that of the next event
	 * in any case).
	 */
	long seqStart = handle.numMatches();

	add(handle, chit);	// add it into the tables

	if (txn == null)	// log it (but txns are aborted by crash)
	    logger.registerOp(tmpl, chit);
	else			// make it known to the txn
	    txn.add((Transactable) chit);

	EventRegistration reg =
	    new EventRegistration(tmpl.id(), space, lease, seqStart);
	return new AddRecord(reg, chit);
    }

    /**
     * Add a registration for the template <code>tmpl</code> under the
     * given <code>chit</code>.
     */
    void add(EntryRep tmpl, NotifyChit chit) {
	TemplateHandle handle = handleFor(tmpl, null);
	add(handle, chit);
    }

    /**
     * Return the handle for matches of the given template under the 
     * specified transaction.
     */
    private TemplateHandle handleFor(EntryRep tmpl, TransactableMgr txn) {
	TemplateHolder holder = (TemplateHolder) holderFor(tmpl);
	// we don't add the handle to the idMap -- we add the chits
	return (TemplateHandle) holder.add(tmpl, txn);
    }

    /**
     * Private implementation of add for the handle and chit pair.
     */
    private void add(TemplateHandle handle, NotifyChit chit) {
	EntryRep tmpl = handle.rep();
	if (tmpl.id() == tmpl.ID_UNSET) {
	    synchronized (this) {		// protect this id field
		if (tmpl.id() == tmpl.ID_UNSET)	// in case someone beat us to it
		    tmpl.id(BasicSpace.nextID());
	    }
	}

	// Now that the chit is all set up we can add it to our tables
	idMap.put(chit.getCookie(), chit);
	handle.add(chit);
    }

    /**
     * This maps a chit's ID to its handle, and then updates the ID of
     * that handle to the given value, ignoring values less than the
     * current one.
     */
    void changeSeq(long id, long newSeq) {
	NotifyChit chit = (NotifyChit) getLeasedResource(new Long(id));
	TemplateHandle handle = chit.handle();
	handle.changeSeq(newSeq);
    }

    /**
     * Reap expired chits.
     */
    void reapExpired() {
	long time = System.currentTimeMillis();
	TemplateHolder[] snapshot = makeSnapshot();
	for (int i = 0; i < snapshot.length; i++)
	    snapshot[i].reapExpired(time);
    }

    /**
     * Return a snapshot of the holders so we can release the lock safely.
     */
    private TemplateHolder[] makeSnapshot() {
	synchronized (holders) {
	    TemplateHolder[] snapshot = new TemplateHolder[holders.size()];
	    Enumeration e = holders.elements();
	    for (int i = 0; i < snapshot.length; i++)
		snapshot[i] = (TemplateHolder) e.nextElement();
	    return snapshot;
	}
    }
}
