/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import net.jini.space.*;
import net.jini.core.event.RemoteEventListener;
import net.jini.core.transaction.*;

import java.rmi.MarshalledObject;

/**
 * This class is for chits that were registered under a transaction,
 * and therefore vanish when the transaction is committed.  This means
 * that <code>prepare</code> returns <code>READONLY</code>, and so
 * <code>commit</code> will never be called.
 *
 * @see NotifyChit
 */

class TransactableChit extends NotifyChit implements Transactable {
    static final long serialVersionUID = -3536516340557928230L;

    /** The manager of this chit (needed in <code>canSee</code>). */
    private TransactableMgr	mgr;
    private boolean		sendException;
    private boolean		forProxy;

    private static java.io.PrintWriter	dbgTxns = BasicSpace.dbgTxns;

    private static final Exception TRANSACTION_EXCEPTION =
	new TransactionException("completed while operation in progress");

    /**
     * Create a new <code>TransactableChit</code> managed by mgr, with
     * the given <code>listener</code>, <code>cookie</code>, and
     * <code>handback</code>.
     */
    TransactableChit(TransactableMgr mgr, RemoteEventListener listener,
	long cookie, MarshalledObject handback)
    {
	super(listener, cookie, handback);
	this.mgr = mgr;
	forProxy = (listener instanceof SpaceProxy.ProxyListener);
    }

    /**
     * Always returns <code>NOTCHANGED</code>.  If this represents a
     * blocking <code>read</code> or <code>take</code>, this first sets
     * up notification of the client to cause an appropriate
     * <code>TransactionException</code> to be thrown by the blocking
     * call.
     */
    //!! Why is mgr passed to these calls when we've got one?  -arnold
    public int prepare(TransactableMgr mgr, BasicSpace space) {
	// if it isn't talking to one of our own
	if (!forProxy)
	    setExpiration(0);	// just expire it
	else {
	    // set up the exception to blocked calls.
	    try {
		Long num = (Long) handback().get();
		ThrowThis toThrow = new ThrowThis(num, TRANSACTION_EXCEPTION);
		handback(new MarshalledObject(toThrow));
	    } catch (java.io.IOException e) {
		throw new InternalSpaceException("can't create ThrowThis", e);
	    } catch (ClassNotFoundException e) {
		throw new InternalSpaceException("can't create ThrowThis", e);
	    }
	    sendException = true;
	    if (dbgTxns != null)
		dbgTxns.println("TransactableChit.prepare: exceptionToSend");
	    space.exceptionToSend(this);
	}

	// This prevents calls to commit()
	return net.jini.core.transaction.server.TransactionConstants.NOTCHANGED;
    }

    /**
     * This should never be called, since <code>prepare</code> returns
     * <code>NOTCHANGED</code>.
     */
    public void commit(TransactableMgr mgr, BasicSpace space) {
	throw new InternalSpaceException("committing a NotifyChit!");
    }

    /**
     * Equivalent, in fact, to <code>prepare</code>.
     */
    public void abort(TransactableMgr mgr, BasicSpace space) {
	prepare(mgr, space);	// same thing, actually
    }

    /**
     * Return <code>true</code> if the writing transaction is
     * <code>null</code>, or if it is my transaction, too.
     */
    boolean canSee(TransactableMgr writeTxn) {
	return (writeTxn == null || writeTxn == mgr);
    }

    /**
     * Return <code>true</code> if this object is now set to send
     * an exception.
     *
     * @see #prepare
     */
    boolean sendException() {
	return sendException;
    }
}
