/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import net.jini.space.*;
import net.jini.core.transaction.*;
import net.jini.core.transaction.server.*;
import com.sun.jini.mahout.StorableObject;
import com.sun.jini.constants.TxnConstants;

import java.rmi.RemoteException;
import java.util.*;

import COM.odi.util.OSVector;

/**
 * This class represents a space's state in a single transaction.
 */
class Txn implements TransactableMgr, TransactionConstants, java.io.Serializable
{
    static final long serialVersionUID = 6306972778849218437L;

    private int				state;		// our current state
    private boolean			changed;	// set by prepare
    private StorableObject		tr;		// our transaction
    private OSVector			txnables;	// members
    private transient TxnMonitorTask	monitorTask;	// monitor for this txn

    private static final Transactable	noop = new TxnNoop();

    static final java.io.PrintWriter	dbgTxns   = BasicSpace.dbgTxns;

    /**
     * Create a new <code>Txn</code> that represents our state in the
     * given <code>Transaction</code>.
     */
    Txn(Transaction tr) throws RemoteException {
	this.tr = new StorableObject(tr);
	construct();

	if (dbgTxns != null)
	    dbgTxns.println("Txn: state = " + TxnConstants.getName(state));
    }

    /**
     * Create a new <code>Txn</code> that represents our state in a
     * transaction whose only known data is its ID.  We pass in the
     * broken <code>StorableObject</code> as the transaction.
     *
     * @see BasicSpace#addBrokenTxn
     */
    Txn(StorableObject trObj) {
	this.tr = trObj;
	construct();
    }

    /**
     * Common code for constructors.
     */
    private void construct() {
	state = ACTIVE;
	txnables = new OSVector();
    }

    /**
     * Return our local view of the current state.
     */
    int getState() {
	return state;
    }

    /**
     * Force the state into the given one.
     */
    void forceState(int newState) {
	state = newState;
    }

    /**
     * Prepare for transaction commit.
     */
    synchronized int prepare(BasicSpace space) {
	if (dbgTxns != null) {
	    dbgTxns.println("Txn: prepare: state = " + 
		TxnConstants.getName(state));
	}

	BitSet unchanged = null;
	switch (state) {
	  case ABORTED:				// previously aborted
	    return ABORTED;

	  case COMMITTED:			// previously committed
	    throw new IllegalStateException();	// "cannot happen"

	  case NOTCHANGED:			// previously voted NOTCHANGED
	  case PREPARED:			// previously voted PREPARED
	    return state;			// they are idempotent, and
	                                        // and we have nothing to do
	                                        // so return

	  case ACTIVE:				// currently active
	    changed = false;			// did this txn change anything?
	    state = VOTING;

	    if (dbgTxns != null)
	        dbgTxns.println("Txn: prepare: preparing");

	    // loop through Transactable members of this Txn
	    Enumeration e = txnables.elements();
	    for (int i = 0; e.hasMoreElements(); i++) {
		int prepState =			// get this member's vote
		    ((Transactable) e.nextElement()).prepare(this, space);
		if (dbgTxns != null) {
		    dbgTxns.println("Transactable(" + (i+1) + "):" +
			" state= " + TxnConstants.getName(prepState));
		}
		switch (prepState) {
		  case PREPARED:		// has prepared state
		    changed = true;		// this means a change
		    continue;

		  case ABORTED:			// has to abort
		    try {
			abort(space);		// abort this txn (does cleanup)
		    } catch (CannotAbortException ex) {
			// "cannot happen"
			throw new IllegalStateException(
			    "cannot abort? \"" + ex + '"');
		    }
		    return ABORTED;		// vote aborted

		  case NOTCHANGED:		// no change
		    if (unchanged == null)	// create the set if needed
			unchanged = new BitSet();
		    unchanged.set(i);		// mark this as unchanged
		    continue;

		  default:			// huh?
		    throw new
			InternalSpaceException("prepare said " + prepState);
		}
	    }
	    state = (changed ? PREPARED : NOTCHANGED);
	    // have to watch this since it's now holding permanent resources
	    if (changed)
		space.monitor(Collections.nCopies(1, this));
	    notifyAll();
	    break;
	  default:
	    throw new IllegalStateException("unknown Txn state: " + state);
	}

	// Remove the unchanged things from the set
	if (changed && unchanged != null) {
	    for (int i = unchanged.size() - 1; i >= 0; --i)
		if (unchanged.get(i))
		    txnables.setElementAt(noop, i);
	}

	return state;
    }

    /**
     * Abort the transaction.  This must be callable from <code>prepare</code>
     * because if a <code>Transactable</code> votes <code>ABORT</code>, this
     * method is called to make that happen.
     */
    synchronized void abort(BasicSpace space) throws CannotAbortException {
	if (dbgTxns != null) {
	    dbgTxns.println("Txn: abort: state = " +
		TxnConstants.getName(state));
	}

	switch (state) {
	  case ABORTED:		// already aborted
	  case NOTCHANGED:	// nothing to abort
	    break;

	  case COMMITTED:	// "cannot happen"
	    throw new CannotAbortException("aborting a committed txn");

	  case ACTIVE:
	  case PREPARED:
	    state = VOTING;	// I'm in the middle of voting to abort
	    if (dbgTxns != null)
	        dbgTxns.println("Txn aborting");

	    Enumeration e = txnables.elements();
	    while (e.hasMoreElements())
		((Transactable) e.nextElement()).abort(this, space);
	    state = ABORTED;
	    notifyAll();
	    cleanup();
	    break;

	  default:
	    throw new IllegalStateException("unknown Txn state: " + state);
	}
    }

    /**
     * Having prepared, roll the changes forward.
     */
    synchronized void commit(BasicSpace space) {
	//!! Need to involve mgr here
	if (dbgTxns != null) {
	    dbgTxns.println("Txn: commit: state = " +
		TxnConstants.getName(state));
	}

	switch (state) {
	  case ABORTED:		// "cannot happen" stuff
	  case ACTIVE:
	  case NOTCHANGED:
	    throw new IllegalStateException("committing "
		+ TxnConstants.getName(state) + " txn");

	  case COMMITTED:	// previous committed, that's okay
	    return;

	  case PREPARED:	// voted PREPARED, time to finish up
	    Enumeration e = txnables.elements();
	    if (dbgTxns != null)
	        dbgTxns.println("Txn committing");
	    while (e.hasMoreElements())
		((Transactable) e.nextElement()).commit(this, space);
	    state = COMMITTED;
	    notifyAll();
	    cleanup();
	    return;

	  default:
	    throw new IllegalStateException("unknown Txn state: " + state);
	}
    }

    /**
     * Make sure this transaction is currently active, after invoking
     * <code>waitUntilStable</code>.
     *
     * @throws TransactionException The transaction isn't active.
     *
     * @see waitUntilStable
     */
    void checkActive() throws TransactionException {
	waitUntilStable();
	if (state != ACTIVE)
	    throw new TransactionException("Transaction no longer active");
    }

    // inherit doc comment
    public Transactable add(Transactable t) {
	txnables.addElement(t);
	return t;
    }

    /**
     * Return the storable version of this transaction.
     */
    StorableObject storableTransaction() {
	return tr;
    }

    // inherit doc comment
    public ServerTransaction transaction(BasicSpace space) {
	try {
	    return (ServerTransaction) tr.get();
	} catch (RemoteException e) {
	    System.err.println("Couldn't contact manager, aborting");
	    try {
		abort(space);
	    } catch (CannotAbortException ex) {
		; // must already *be* aborted or some such
	    }
	    return null;
	}
    }

    // inherit doc comment
    public void waitUntilStable() {
	if (dbgTxns != null)
	    dbgTxns.println("Txn: waitUntilStable");
	try {
	    while (state == VOTING)
		wait();
	} catch (InterruptedException e) {
	    throw new InternalSpaceException("unexpected Interruption", e);
	}
    }

    /**
     * Return the monitor task for this object.
     */
    TxnMonitorTask monitorTask() {
	return monitorTask;
    }

    /**
     * Set the monitor task for this object.
     */
    void monitorTask(TxnMonitorTask task) {
	monitorTask = task;
    }

    /**
     * Clean up any state when the transaction is finished.
     */
    private void cleanup() {
	if (monitorTask != null)
	    monitorTask.cancel();	// stop doing this
    }
}
