/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import net.jini.space.InternalSpaceException;
import net.jini.core.transaction.*;

import java.util.HashSet;
import java.util.Iterator;

/**
 * This is a kind of <code>EntryHandle</code> that is used to represent
 * <code>EntryRep</code> objects that are currently involved in some
 * transaction.  Each handle has one or more managers that are governing
 * the transaction.
 */
class TxnEntryHandle extends EntryHandle implements Transactable {
    /**
     * The list of known managers.  In order to keep things small in the
     * common case that there is only one known manager, <code>mgrs</code>
     * is managed as a ``list'' in one of two states -- it is either a
     * direct reference to the only manager for this handle, or a reference
     * to an <code>HashSet</code> with entries for each associated manager.
     */
    private Object		mgrs;

    /**
     * The current state of the handle, such as <code>READ</code> or
     * <code>TAKE</code>.
     */
    private int			state;

    /**
     * The holder this transacted entry handle is contained in.
     */
    private SimpleEntryHolder	holder;

    /**
     * Print out debug messages?
     */
    static final java.io.PrintWriter dbgHandles =
				    BasicSpace.debug.getWriter("txnHandles");

    /**
     * Create a new TxnEntryHandle for the given <code>rep</code>.  The
     * handle will represent the operation contained in <code>state</code>
     * under the transaction managed by <code>mgr</code>.  The
     * <code>holder</code> is the <code>SimpleEntryHolder</code> that
     * will contain the created handle; this is needed so that
     * <code>commit</code> and <code>abort</code> can manipulate the
     * contents.
     */
    TxnEntryHandle(EntryRep rep, TransactableMgr mgr, SimpleEntryHolder holder,
	int state)
    {
	super(rep);
	this.mgrs = mgr;
	this.state = state;
	this.holder = holder;
	if (dbgHandles != null)
	    dbgHandles.println("TxnEntryHandle: TxnEntryHandle: " +
	                       "new container for actions done " +
			       "under a transaction");
    }

    /**
     * Create a new TxnEntryHandle to replace the given <code>handle</code>.
     * The new handle will represent the operation contained in
     * <code>state</code> under the transaction managed by <code>mgr</code>.
     * The <code>holder</code> is the <code>SimpleEntryHolder</code> that
     * will contain the created handle; this is needed so that
     * <code>commit</code> and <code>abort</code> can manipulate the
     * contents.
     */
    TxnEntryHandle(EntryHandle handle, TransactableMgr mgr,
	SimpleEntryHolder holder, int state)
    {
	super(handle);
	this.mgrs = mgr;
	this.state = state;
	this.holder = holder;
        if (dbgHandles != null)
            dbgHandles.println("TxnEntryHandle: TxnEntryHandle: " +
                               "new action done under a transaction" +
				"on existing Entry");
    }

    // inherit doc comment
    public int prepare(TransactableMgr mgr, BasicSpace space) {
        //The WRITE and TAKE make entries to the transactional
        //log during the prepare phase.

	if (dbgHandles != null)
	    dbgHandles.println("TxnEntryHandle: prepare: state= " +
				TransactableMgr.stateNames[state]);
	if (state == TransactableMgr.READ) {
	    replaceWithNormal(mgr);
	    return NOTCHANGED;
	}

	return PREPARED;
    }

    // inherit doc comment
    public void abort(TransactableMgr mgr, BasicSpace space) {
	if (dbgHandles != null)
	    dbgHandles.println("TxnEntryHandle: abort: state= " +
				TransactableMgr.stateNames[state]);
	if (state == TransactableMgr.READ || state == TransactableMgr.TAKE) {
	    replaceWithNormal(mgr);
	    space.abortOnEntry(rep());
	} else {
	    holder.remove(this);
	}
    }

    // inherit doc comment
    public void commit(TransactableMgr mgr, BasicSpace space) {
	//The TAKE TxnEntryHandle removes itself from the
	//SimpleEntryHolder's list upon commit.
	//The WRITE and READ TxnEntryHandles replace themselves
	//with regular EntryHandles upon commit.

	if (dbgHandles != null)
	    dbgHandles.println("TxnEntryHandle: commit: state= " +
				TransactableMgr.stateNames[state]);
	// have to synchronize in case someone is about to take this one
	synchronized (this) {
	    // put it back as a normal handle
	    switch (state) {
	      case TransactableMgr.WRITE:
		//!! assumption -- single level transactions:
		//!! we pass null where we should pass the parent txn
		if (removed())
		    break;

		space.recentWrite(rep(), null);
		// FALLTHROUGH

	      case TransactableMgr.READ:
		replaceWithNormal(mgr);
		break;

	      case TransactableMgr.TAKE:
		holder.remove(this);
		break;

	      default:
		throw new InternalSpaceException("unexpected state in "
		    + "TxnEntryHandle.commit(): " + state);
	    }
	}
    }

    /**
     * Replace this handle in the list with the ``normal'' (original)
     * non-transacted handle or an equivalent.
     */
    private synchronized void replaceWithNormal(TransactableMgr mgr) {
	//!! Note assumption of single-level transactions.
	//   The "null" in remove() should be the parent's transaction,
	//   or null if no parent.  We assume no parent, and so just pass
	//   null.
	if (onlyMgr(mgr))
	    holder.replace(this, new EntryHandle(this), null);
	else
	    removeMgr(mgr);
    }

    /**
     * Add <code>mgr</code> to the list of known managers, setting the
     * state of this handle to <code>op</code>.
     */
    synchronized void add(TransactableMgr mgr, int op) {
	if (mgr == mgrs)
	    return;	// already the only one known

	if (mgrs instanceof TransactableMgr) {
	    Object origMgr = mgrs;
	    mgrs = new HashSet();
	    ((HashSet) mgrs).add(origMgr);
	}

	// if mgr is in already this is harmless, and checking to prevent
	// a redundant addition is more expensive
	((HashSet) mgrs).add(mgr);
	monitor(mgr);
	state = op;
    }

    /**
     * If we need to, add this manager to the list of transactions that
     * need to be monitored because of conflicts over this entry.  Any
     * existing blocking txn is sufficient.
     *
     * @see TxnMonitorTask#addSibling
     */
    private void monitor(TransactableMgr mgr) {
	Txn txn = (Txn) mgr;
	Iterator it = ((HashSet) mgrs).iterator();
	while (it.hasNext()) {
	    Txn otherTxn = (Txn) it.next();
	    if (otherTxn != mgr && otherTxn.monitorTask() != null) {
		otherTxn.monitorTask().addSibling(txn);
		return;
	    }
	}
    }

    /**
     * Remove the given mgr from the list of known managers.
     */
    private synchronized void removeMgr(TransactableMgr mgr) {
	HashSet tab = (HashSet) mgrs;
	tab.remove(mgr);
    }

    /**
     * Returns <code>true</code> if the operation <code>op</code> under
     * the transaction manager by <code>mgr</code> is legal on this
     * entry given the operation already performed on this entry under
     * other transactions.  It is legal to:
     * <ul>
     * <li>Read an entry that has been read in any other transaction,
     *     or that has been written under the same transaction (that is,
     *     if <code>mgr</code> is the same as the writing transaction).
     * <li>Take an entry that was written under the same transaction, or
     *     which was read <em>only</em> under the same transaction (that is,
     *     no other active transactions have also read it).
     * </ul>
     * All other operations are not legal, so <code>canPerform</code>
     * otherwise returns <code>false</code>.
     */
    boolean canPerform(TransactableMgr mgr, int op) {
	//Note: If two transactions are performed
	//	within the same TransactableMgr
	//	(see Txn.java), that means its in
	//	the same Transaction (see Transaction.java).
	//      Operations under the same Transaction scope
	//	are the same as if under a null Transaction
	//	i.e. all operations are visible.
	//
	//	Semantics need to be added for multi-level
	//	transactions.

	if (dbgHandles != null) {
	    dbgHandles.println("TxnEntryHandle: canPerform(" + mgr + ", " + op
		+ "): state = " + TransactableMgr.stateNames[state]);
	}

	switch (state) {
	  case TransactableMgr.READ:
	    if (op == TransactableMgr.READ)
		return true;
	    return onlyMgr(mgr); // can only modify if I'm the only participant
	  case TransactableMgr.WRITE:
	    if ( (op == TransactableMgr.READ) ||
		 (op == TransactableMgr.TAKE) )
		return onlyMgr(mgr);
	    return false;	  // can't read unless I'm the writer
	  case TransactableMgr.TAKE:
	    return false;	  // can't take anything taken any time
	}
	return false;
    }

    /**
     * Return <code>true</code> if <code>mgr</code> is one of the managers
     * known to be managing this entry.
     */
    boolean knownMgr(TransactableMgr mgr) {
	if (mgr == null)
	    return false;
	if (mgr == mgrs)
	    return true;
	if (mgrs instanceof HashSet)
	    return ((HashSet) mgrs).contains(mgr);
	return false;
    }

    /**
     * Return <code>true</code> if the given manager is the only one
     * managing this handle.
     */
    private boolean onlyMgr(TransactableMgr mgr) {
	if (mgr == null)
	    return false;
	if (mgr == mgrs)
	    return true;
	if (mgrs instanceof HashSet) {
	    HashSet tab = (HashSet) mgrs;
	    return (tab.size() == 1 && tab.contains(mgr));
	}
	return false;
    }

    /**
     * Add all the managers of this transaction to the given collection.
     */
    void addTxns(java.util.Collection collection) {
	if (mgrs == null)
	    return;
	else if (mgrs instanceof HashSet)
	    collection.addAll((HashSet) mgrs);
	else
	    collection.add(mgrs);
    }

    /**
     * Return <code>true</code>: every <code>TxnEntryHandle</code> is
     * managed by at least one manager.
     */
    boolean managed() {
	return true;
    }

    /*
     * Wait until all the transactions involved with this are ready
     * to proceed, that is, they are not in the middle of changing
     * state.
     */
    void waitUntilReady() throws TransactionException {
	if (mgrs instanceof TransactableMgr)
	    ((TransactableMgr) mgrs).waitUntilStable();
	else {
	    Iterator e = ((HashSet) mgrs).iterator();
	    while (e.hasNext())
		((TransactableMgr) e.next()).waitUntilStable();
	}
    }
}
