/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import net.jini.core.lease.*;
import net.jini.core.transaction.*;
import net.jini.core.transaction.server.*;

import com.sun.jini.thread.TaskManager;
import com.sun.jini.thread.RetryTask;
import com.sun.jini.collection.FastList;
import com.sun.jini.lease.landlord.LeasedResource;

import java.util.Collection;
import java.util.Iterator;

/**
 * This class provides a driver for monitoring the state of transactions
 * that have blocked progress of other operations recently.  It creates
 * tasks that monitor each transaction by intermittently querying the
 * transaction's state.  If it finds that the transaction has aborted,
 * it makes sure that the local space aborts the transaction, too, so
 * that operations will cease to be blocked by the transaction.
 *
 * @see TxnMonitorTask
 * @see BasicSpace#monitor
 */
class TxnMonitor implements Runnable {
    /**
     * Each <code>ToMonitor</code> object represents a need to monitor
     * the given transactions, possibly under a lease.
     *
     * @see #pending
     */
    private static class ToMonitor extends FastList.Node {
	LeasedResource	resource;	// resource governing interest in txns
	Collection	txns;		// the transactions to monitor

	ToMonitor(LeasedResource resource, Collection txns) {
	    this.resource = resource;
	    this.txns = txns;
	}
    }

    /**
     * This list is used to contain requests to monitor interfering
     * transactions.  We use a list like this so that the <code>getMatch</code>
     * request that detected the conflict doesn't have to wait for all
     * the setup before returning -- it just puts the data on this list
     * and the <code>TxnMonitor</code> pulls it off using its own thread.
     *
     * @see #ToMonitor
     * @see BasicSpace#getMatch
     */
    private FastList pending = new FastList();

    /**
     * The manager for <code>TxnMonitorTask</code> objects.
     */
    private TaskManager taskManager = new TaskManager();

    /**
     * The space we belong to.  Needed for aborts.
     */
    private BasicSpace	space;

    static final java.io.PrintWriter
	dbg = BasicSpace.debug.getWriter("txnMonitor");

    /**
     * Create a new TxnMonitor.
     */
    TxnMonitor(BasicSpace space) {
	this.space = space;
	new Thread(this, "TxnMonitor").start();
    }

    /**
     * Return the space we're part of.
     */
    BasicSpace space() {
	return space;
    }

    /**
     * Add a set of <code>transactions</code> to be monitored under the
     * given leased <code>resource</code>.
     */
    void add(LeasedResource resource, Collection transactions) {
	pending.add(new ToMonitor(resource, transactions));
	synchronized (this) {
	    notifyAll();
	}
    }

    /**
     * Add a set of <code>transactions</code> to be monitored under no
     * lease.
     */
    void add(Collection transactions) {
	pending.add(new ToMonitor(null, transactions));
	synchronized (this) {
	    notifyAll();
	}
    }

    /**
     * Take pending monitor requests off the queue, creating the
     * required <code>TxnMonitorTask</code> objects and scheduling them.
     */
    public void run() {
	try {
	    ToMonitor tm;
	    for (;;)  {
		synchronized (this) {
		    while ((tm = (ToMonitor) pending.head()) == null)
			wait();
		}
		if (dbg != null)
		    dbg.println("for resource " + tm.resource + ":");
		pending.remove(tm);
		Iterator it = tm.txns.iterator();
		while (it.hasNext()) {
		    Txn txn = (Txn) it.next();
		    TxnMonitorTask task = taskFor(txn);
		    task.add(tm.resource);
		}
	    }
	} catch (InterruptedException e) {
	    return;
	}
    }

    /**
     * Return the monitor task for this transaction, creating it if
     * necessary.
     */
    private TxnMonitorTask taskFor(Txn txn) {
	TxnMonitorTask task = txn.monitorTask();
	if (task == null) {
	    if (dbg != null)
		dbg.println("creating task for " + txn);
	    task = new TxnMonitorTask(txn, this, taskManager);
	    txn.monitorTask(task);
	    taskManager.add(task);	// add it after we've set it in the txn
	}
	return task;
    }
}
