/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import net.jini.core.lease.*;
import net.jini.core.transaction.*;
import net.jini.core.transaction.server.*;
import com.sun.jini.thread.RetryTask;
import com.sun.jini.thread.TaskManager;
import com.sun.jini.lease.landlord.LeasedResource;
import com.sun.jini.constants.TxnConstants;

import java.lang.ref.WeakReference;
import java.util.Map;
import java.util.WeakHashMap;
import java.util.Iterator;
import java.util.Collection;
import java.util.Collections;
import java.rmi.*;

/**
 * A task that will try to validate the state of a transaction.  This
 * uses weak references a good deal to let the other parts of the system
 * be GC'ed as necessary.
 * <p>
 * The retry mechanism is subtle, so bear with me.  The purpose is
 * to ensure that if any activity is being blocked by a given
 * transaction, that transaction will be tested at some point in
 * the future (if necessary, i.e., if it still is thought to be
 * active).  We assume it to be rare that a transactions that the
 * space thinks is active is, in fact, aborted, so the algorithm is
 * designed to guarantee the detection without a lot of overhead,
 * specifically without a lot of RMI calls.
 * <p>
 * Each task has three values: a <code>nextQuery</code> time, a
 * <code>mustQuery</code> boolean that force the next query to
 * be made, and <code>deltaT</code>, the time at which the
 * following query will be scheduled.  When the task is awakened
 * at its <code>nextQuery</code> time, it checks to see if it must
 * make an actual query to the transaction manager, which it will
 * do if either <code>mustQuery</code> is <code>true</code>, or
 * if any of the related leases expire in the future.  Whether or
 * not an actual query is made, <code>deltaT</code> is added to
 * <code>nextQuery</code> to get the <code>nextQuery</code> time,
 * <code>deltaT</code> is doubled, and <code>mustQuery</code>
 * boolean is set to <code>false</code>.
 * <p>
 * There are two kinds of requests that a with which transaction
 * can cause a conflict -- those under long leases (such as
 * blocking reads and takes) and those that are under short leases
 * (such as reads and takes with zero-length timeouts).  We will
 * treat them separately at several points of the algorithm.  A
 * short lease is any lease whose expiration time is sooner than
 * the <code>nextQuery</code> time.  Any other lease is a long
 * lease.  If a short query arrives, <code>mustQuery</code> is
 * set to <code>true</code>.
 * <p>
 * The result is that any time a transaction causes a conflict, if
 * the lease has not expired by the time of the
 * <code>nextQuery</code> there will be a query.  There will also
 * be a query if any conflict occurred on a short lease.
 * <p>
 * The first time a transaction causes a conflict, we schedule a
 * time in the future at which we will query its status.  We do not
 * query right away because often a transaction will complete on
 * its own before we get to that time, making the check
 * unnecessary.  An instant query is, therefore, unnecessarily
 * aggressive, since giving an initial grace time will usually mean
 * no query is made at all.  So if the first conflict occurs at
 * <i>T</i><sub>0</sub>, the <code>nextQuery</code> value will be
 * <i>T</i><sub>0</sub><code>+INITIAL_GRACE</code>, the boolean
 * will be <code>true</code> to force that query to happen, and
 * <code>deltaT</code> will be set to <code>INITIAL_GRACE</code>.
 * @see TxnMonitor
 */
class TxnMonitorTask extends RetryTask
    implements TransactionConstants, com.sun.jini.constants.TimeConstants
{
    private Txn		txn;		// transaction being monitored
    private TxnMonitor	monitor;	// the monitor we were made by
    private Map		resources;	// all the leased resources for this
    private int		failCnt;	// count of RemoteExceptions

    private long	nextQuery;	// next query time
    private boolean	mustQuery;	// next query must actually try
    private long	deltaT;		// next value added to nextQuery

    /**
     * The initial grace period before the first query.
     */
    private static final long	INITIAL_GRACE = 15 * SECONDS;

    /**
     * The retry time when we have an encountered an exception
     */
    private static final long	BETWEEN_EXCEPTIONS = 15 * SECONDS;

    /**
     * The largest value that <code>deltaT</code> will reach.
     */
    private static final long	MAX_DELTA_T = 1 * HOURS;

    /**
     * The maximum number of failures allowed in a row before we simply
     * give up on the transaction and consider it aborted.
     */
    private static final int	MAX_FAILURES = 3;

    /**
     * Create a new TxnMonitorTask.
     */
    TxnMonitorTask(Txn txn, TxnMonitor monitor, TaskManager manager) {
	super(manager);
	this.txn = txn;
	this.monitor = monitor;
	nextQuery = startTime();	// retryTime will add INITIAL_GRACE
	deltaT = INITIAL_GRACE;
	mustQuery = true;
    }

    /**
     * Return the time of the next query, bumping <code>deltaT</code> as
     * necessary for the next iteration.  If the transaction has voted
     * <code>PREPARED</code> or the manager has been giving us a
     * <code>RemoteException</code>, we should retry on short times;
     * otherwise we back off quickly.
     */
    public long retryTime() {
	if (failCnt == 0 && txn.getState() != PREPARED) {	// no failures
	    if (TxnMonitor.dbg != null)
		TxnMonitor.dbg.println(this + " retryTime adds " + deltaT);
	    nextQuery += deltaT;
	    if (deltaT < MAX_DELTA_T)
		deltaT = Math.min(deltaT * 2, MAX_DELTA_T);
	} else {
	    if (TxnMonitor.dbg != null) {
		TxnMonitor.dbg.println(this + " retryTime adds "
		    + BETWEEN_EXCEPTIONS + " (for " +
		    (failCnt != 0 ? "failure" : "PREPARED") + ")");
	    }
	    nextQuery += BETWEEN_EXCEPTIONS;
	}
	return nextQuery;
    }

    /**
     * We can run in parallel with any task, so just return <CODE>false</CODE>.
     */
    public boolean runAfter(java.util.List tasks, int size) {
	return false;
    }

    /**
     * Add a ``sibling'' transaction, one that is now blocking progress
     * on one of the same entries.  For example, if a client is blocked
     * on a <code>read</code>, another transaction can read the same
     * entry, thereby also blocking that same client.  This means that
     * the transaction for the second <code>read</code> must be
     * watched, too.  The list of leased resources for the second transaction
     * might be less that the list of those in this transaction, but
     * the process of figuring out the subset is too expensive, since
     * we have tried to make the checking process itself cheap,
     * anyway.  So we add all resources this task is currently monitoring
     * to the task monitoring the second transaction.  If there are
     * no resources, then the blocking occurred because of a short lease
     * or all the leases have expired, in which case the second transaction
     * isn't blocking the way of anything currently, so this method does
     * nothing.
     * <p>
     * Of course, in order to avoid blocking the thread that is calling
     * this (which is trying to perform a <code>read</code>, after
     * all), we simply add each lease in this task to the monitor's
     * queue.
     *
     * @see TxnEntryHandle#monitor
     */
    //!! Would it be worth the overhead to make TxnEntryHandle.monitor
    //!! search for the transaction with the smallest set of leases?  -arnold
    synchronized void addSibling(Txn txn) {
	if (resources == null || resources.size() == 0)
	    return;
	Collection sibling = Collections.nCopies(1, txn);
	Iterator it = resources.keySet().iterator();
	while (it.hasNext()) {
	    LeasedResource resource = (LeasedResource) it.next();
	    if (resource != null)	// from a weak map, so might be null
		monitor.add(resource, sibling);
	}
    }

    /**
     * Try to see if this transcation should be aborted.  This returns
     * <code>true</code> (don't repeat the task) if it knows that the
     * transaction is no longer interesting to anyone.
     */
    public boolean tryOnce() {
	if (TxnMonitor.dbg != null) {
	    TxnMonitor.dbg.println(this + " attempt " + attempt());
	    TxnMonitor.dbg.println(this + " mustQuery = " + mustQuery);
	}
	/*
	 * The first time we do nothing, since RetryTask invokes run first,
	 * but we want to wait a bit before testing the transaction.
	 */
	if (attempt() == 0)
	    return false;

	if (TxnMonitor.dbg != null) {
	    TxnMonitor.dbg.println(this + " txn.getState() = "
		+ TxnConstants.getName(txn.getState()));
	}

	// not active or prepared == no longer blocking
	int txnState = txn.getState();
	if (txnState != ACTIVE && txnState != PREPARED)
	    return true;

	// if we're prepared, test every time -- this shouldn't take long
	mustQuery |= (txnState == PREPARED);

	/*
	 * Go through the resources to see if we can find one still active
	 * that cares.  Must be synchronized since we test, then clear --
	 * another thread that set the flag between the test and clear
	 * would have its requirements lost.
	 */
	synchronized (this) {
	    if (!mustQuery) {		// then try resources
		if (resources == null)	// no resources, so nobody wants it
		    return false;	// try again next time

		Iterator it = resources.keySet().iterator();
		boolean foundNeed = false;
		if (TxnMonitor.dbg != null)
		    TxnMonitor.dbg.println(this + " nextQuery = " + nextQuery);
		while (it.hasNext()) {
		    LeasedResource resource = (LeasedResource) it.next();
		    if (resource == null)	// gone -- the map will reap it
			continue;
		    if (TxnMonitor.dbg != null) {
			TxnMonitor.dbg.println(this
			    + " resource.getExpiration() = "
			    + resource.getExpiration());
		    }
		    if (resource.getExpiration() < nextQuery)
			it.remove();	// expired, so we don't care about it
		    else {
			foundNeed = true;
			break;
		    }
		}
		if (TxnMonitor.dbg != null)
		    TxnMonitor.dbg.println(this + " foundNeed = " + foundNeed);
		if (!foundNeed)		// nobody wants it
		    return false;	// try again next time
	    }
	    mustQuery = false;		// clear it for next time
	}

	/*
	 * Now we know (a) the transaction itself is alive, and (b) some
	 * lease still cares.  Make sure it's still active as far as the
	 * it knows, and if it is, then ask the manager about it.
	 */
	ServerTransaction tr = txn.transaction(monitor.space());
	if (TxnMonitor.dbg != null)
	    TxnMonitor.dbg.println(this + " tr = " + tr);
	if (tr == null)			// if txn now thinks it has been aborted
	    return true;

	int trState;
	try {
	    trState = tr.getState();
	} catch (TransactionException e) {
	    if (TxnMonitor.dbg != null)
		TxnMonitor.dbg.println(this + " TransactionException");
	    trState = ABORTED;
	} catch (NoSuchObjectException e) {
	    if (TxnMonitor.dbg != null)
		TxnMonitor.dbg.println(this + " NoSuchObjectException");
	    trState = ABORTED;
	} catch (RemoteException e) {
	    if (TxnMonitor.dbg != null) {
		TxnMonitor.dbg.println(this + " RemoteException, failCnt = "
		    + failCnt);
	    }
	    if (txn.getState() != PREPARED && ++failCnt >= MAX_FAILURES)
		trState = ABORTED;	// too many failures
	    else {
		mustQuery = true;	// keep on trying
		return false;		// try again next time
	    }
	}
	if (TxnMonitor.dbg != null)
	    TxnMonitor.dbg.println(this + " trState = " + trState);
	failCnt = 0;			// reset failures -- we got a response

	/*
	 * If the two states aren't the same, the state changed and
	 * we need to account for that locally here by calling the
	 * method that would make the change (the one we should have
	 * gotten.  (We use the external forms of abort, commit, etc.,
	 * because they are what the manager would call, and therefore
	 * these calls will always do exactly what the incoming manager
	 * calls would have done.  I don't want this to be fragile by
	 * bypassing those calls and going straight to the Txn object's
	 * calls, which might skip something important in the BasicSpace
	 * calls).
	 */
	if (trState != txnState) {
	    if (TxnMonitor.dbg != null) {
		TxnMonitor.dbg.println(this + "mgr state [" +
		    TxnConstants.getName(trState) + "] != local state [" +
		    TxnConstants.getName(txnState) + "]");
	    }
	    try {
		switch (trState) {
		  case ABORTED:
		    monitor.space().abort(tr.mgr, tr.id);
		    return true;

		  case COMMITTED:
		    monitor.space().commit(tr.mgr, tr.id);
		    return true;
		}
	    } catch (UnknownTransactionException e) {
		// we must somehow have already gotten the abort() or
		// commit(), and have therefore forgotten about the transaction,
		// while this code was executing
		return true;
	    }
	    // we can't fake anything else -- the manager will have to call us
	}

	if (TxnMonitor.dbg != null)
	    TxnMonitor.dbg.println(this + " return false");
	return false;			// now we know so nothing more to do
    }

    /**
     * Add in a resource.  The lease may already be in, in which case it is
     * ignored, or it may be null, in which case it was a non-leased probe
     * that was blocked and we simply set <code>mustQuery</code> to
     * <code>true</code>.
     */
    synchronized void add(LeasedResource resource) {
	if (resource == null || resource.getExpiration() <= nextQuery) {
	    if (TxnMonitor.dbg != null)
		TxnMonitor.dbg.println("adding resource to task -- SHORT");
	    mustQuery = true;
	} else {
	    if (TxnMonitor.dbg != null)
		TxnMonitor.dbg.println("adding resource to task -- LONG");
	    if (resources == null)
		resources = new WeakHashMap();	// we use it like a WeakHashSet
	    resources.put(resource, resource);
	}
    }
}
