/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Random;
import java.util.Vector;
import java.util.NoSuchElementException;

import COM.odi.util.OSVector;
import COM.odi.util.OSHashtable;

/**
 * A type tree for entries.  It maintains, for each class, a list of
 * known subclasses so that we can walk down the relevant subportion of
 * the subtype tree for a template, looking for matching entries.  This
 * list of subtypes is not current garbage collected -- if a subtype
 * was once written, it's subtype entry will never be removed from this
 * tree.  All operations are done via class name.
 *
 * @author Ken Arnold
 * @author Gary Holness
 *
 * @see BasicSpace
 */
class TypeTree {
    /** For each type, a vector of known subtypes */
    private OSHashtable subclasses = new OSHashtable();

    /**
     * A generator used to randomize the order of iterator returns
     * @see #RandomizedIterator
     */
    static final private Random numgen = new Random();

    /**
     * Name of the root bucket of the type tree
     */
    static final private String ROOT = 
	net.jini.core.entry.Entry.class.getName();

    private final static boolean DEBUG = false;

    /**
     * Return the vector of subclasses for the given class.
     */
    OSVector classVector(String whichClass) {
	return (OSVector) subclasses.get(whichClass);
    }

    /**
     * An iterator that will walk through a list of known types.
     *
     * @see #RandomizedIterator
     */
    abstract class TypeTreeIterator implements Iterator {
	protected int cursor;		// the current position in the list
	protected Object[] typearray;	// the list of types as an array

	// inherit doc comment
        public boolean hasNext() {
	    if (DEBUG)
	        System.out.println("TypeTreeIterator: cursor = " + cursor +
					    " # types = " + typearray.length);
            if (cursor < typearray.length) {
		if (DEBUG)
		    System.out.println("TypeTreeIterator: hasNext: TRUE");
                return true;
	    }
 
	    if (DEBUG)
	        System.out.println("TypeTreeIterator: hasNext: FALSE");

            return false;
        }
 
	// inherit doc comment
        public Object next() throws NoSuchElementException {
            Object val = null;
 
            if (cursor >= typearray.length)
                throw new NoSuchElementException("TypeTreeIterator: next");
 
            try {
                val = typearray[cursor];
                cursor++;
            } catch (ArrayIndexOutOfBoundsException e) {
                throw new NoSuchElementException("TypeTreeIterator: next" +
                                                        e.getMessage());
            }

	    if (DEBUG)
	        System.out.println("TypeTree: next: val = " + val);
 
            return val;
        }
 
	/**
	 * Unimplemented operations
	 * @throws UnsupportedOperationException Always
	 */
        public void remove() throws UnsupportedOperationException {
            throw new UnsupportedOperationException(
		"TypeTreeIterator: remove not supported");
        }
    }

    /**
     * This class implements a randomized iterator over the
     * <code>TypeTree</code>.  Given a <code>className</code>, it
     * maintains a randomized list of subtypes for the given
     * <code>className</code>, including the class itself.
     */
    class RandomizedIterator extends TypeTreeIterator {
	/**
	 * Create a new <code>RandomizedIterator</code> for the given
	 * class.
	 */
	RandomizedIterator(String className) {
	    super();
	    init(className);
	}


	/*
	 * Traverse the given type tree and add to the list all the
	 * subtypes encountered within.
	 */
	private void walkTree(Collection children, Collection list) {
	    if (children != null) {
		list.addAll(children);
	        Object[] kids = children.toArray();
		for (int i = 0; i< kids.length; i++) {
		    walkTree(collectionFor(classVector((String)kids[i])), list);
		}
	    }
	}


	/**
	 * Creates a <code>Collection</code> from the given
	 * <code>OSVector</code>.  This method will go away when
	 * <code>the</code> Collection class is supported in the ODI
	 * PSEPro stuff.
	 */
	private Collection collectionFor(OSVector vec) {
	    Vector tmp = null;
	    if (vec != null) {
		tmp = new Vector();
		Enumeration e = vec.elements();
		while(e.hasMoreElements())
		    tmp.addElement(e.nextElement());
	    }

	    return (Collection)tmp;
	}

	/**
	 * Set up this iterator to walk over the subtypes of this class,
	 * including the class itself.  It then randomizes the list.
	 */
	private void init(String className) {
            Collection types = new ArrayList();

	    if (className.equals(EntryRep.matchAnyClassName())) {
		// handle "match any" specially" -- search from ROOT
		// Simplification suggested by 
		// Lutz Birkhahn <lutz.birkhahn@GMX.DE>
		className = ROOT;
	    } else {
		// add this class
		types.add(className);
	    }

	    // add all subclasses
	    walkTree(collectionFor(classVector(className)), types);

	    // Convert it to an array and then randomize
	    if (DEBUG) {
	        System.out.println("RandomizedIterator: init: # types = " +
								types.size());
	    }

	    typearray = types.toArray();
	    int randnum = 0;
	    Object tmpobj = null;

	    for (int i = 0; i < typearray.length; i++) {
		randnum = numgen.nextInt(typearray.length - i);
		tmpobj = typearray[i];
		typearray[i] = typearray[randnum];
		typearray[randnum] = tmpobj;
	    }


	    if (DEBUG) {
	        for (int i = 0; i < typearray.length; i++) {
	            System.out.println("RandomizedIterator: init: " +
			    "array[" + i + "] = " + typearray[i]);
		}
	    }
	}

    }

    /**
     * Return an iterator over the subtypes of the given class
     * (including the type itself).  This implementation always returns
     * an iterator that randomizes the order of the classes returned.
     * In other words, it returns the names of all classes that are
     * instances of the class that named, in a random ordering.
     */
    Iterator subTypes(String className) {
	return new RandomizedIterator(className);
    }

    /**
     * Update the type tree with the given bits.  This will traverse the
     * given EntryRep's list of superclasses, retrieve the subclass list
     * at each list item and update it with the given EntryRep's type.
     *
     *  SupClass List
     *   |
     *   V
     *   SupC1-->Sub1OfSupC1--Sub2OfSupC1...SubNOfSupC1--EntryRep
     *   |
     *   |
     *   SupC2-->Sub1OfSupC2--Sub2OfSupC2...SubNOfSupC2--EntryRep
     *   .
     *   .
     *   .
     *   SupCN-->Sub1OfSupCN--Sub2OfSupCN...SubNOfSupCN--EntryRep
     */
    void addTypes(EntryRep bits) {
	String classFor = bits.classFor();
	String[] superclasses = bits.superclasses();
	if (DEBUG)
	    System.out.print("Adding superclasses of " + classFor + ':');

	//The given EntryRep will add its className to the
	//subtype list of all its supertypes.

	String prevClass = classFor;
	for (int i = 0; i < superclasses.length; i++) {
	    if (DEBUG)
		System.out.print(" " + superclasses[i]);
	    if (!addKnown(superclasses[i], prevClass)) {
		if (DEBUG)
		    System.out.print("...");
		return;
	    }
	    prevClass = superclasses[i];
	}

	// If we are here prevClass must have java.Object as its
	// direct superclass (we don't store "java.Object" in
	// EntryRep.superclasses since that would be redundant) and
	// prevClass is not already in the the tree.  Place it in the
	// "net.jini.core.entry.Entry" bucket so it does not get lost
	// if it does not have any sub-classes.
	//
	// Fix suggested by Lutz Birkhahn <lutz.birkhahn@GMX.DE>
	
	if (DEBUG)
	    System.out.print(" " + ROOT);
	if (!addKnown(ROOT, prevClass)) {
	    if (DEBUG)
		System.out.print("...");
	    return;
	}
	
	if (DEBUG)
	    System.out.println();
    }

    /**
     * Add the subclass to the list of known subclasses of this superclass.
     */
    private boolean addKnown(String superclass, String subclass) {
	OSVector v;

	synchronized (subclasses) {
	    v = classVector(superclass);
	    if (v == null) {
		v = new OSVector();
		subclasses.put(superclass, v);
	    }
	}

	synchronized (v) {
	    if (v.contains(subclass))
		return false;
	    if (DEBUG)
		System.out.print('+');
	    v.addElement(subclass);
	}
	return true;
    }

    /**
     * Return an <code>Enumeration</code> of known types.
     */
    java.util.Enumeration knownTypes() {
	return subclasses.elements();
    }
}
