/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.outrigger;

/**
 * This class is used to create a singly-linked list of written nodes
 * for notification and searches.  When an entry is written, it is
 * added to this list.  As each thread interested in recently-written
 * entries works, it has a reference to the <code>WrittenEntry</code>
 * object that mark the last entry the thread examined.  When it is
 * done with that entry, it proceeds to the next.  As each of these
 * references progresses, the <code>WrittenEntry</code> nodes in the
 * list get left behind to be collected as garbage.
 * <p>
 * The list initially starts with the return value of <code>startNode</code>,
 * and nodes are added using <code>add</code>.  The constructor is private to
 * enforce this usage.
 *
 * @see BasicSpace#write
 * @see BasicSpace#getMatch(EntryRep,Transaction,long,boolean,boolean,RemoteEventListener,long)
 * @see Notifier
 *
 * @author Ken Arnold
 */
class WrittenEntry {
    /** The time at which this entry was written. */
    private long		timestamp;

    /** The next node in the list. */
    private WrittenEntry	next;

    /** The EntryRep this node refers to. */
    private EntryRep		rep;

    /**
     * Create a new time-stamped entry for the given EntryRep.  The comment
     * for the class explains why this is private.
     */
    private WrittenEntry(EntryRep rep) {
	this.rep = rep;
	timestamp = System.currentTimeMillis();
    }

    /**
     * Return the node which is the head of the list.  It is time-stamped,
     * and has an <code>EntryRep</code> of <code>null</code>.
     */
    static WrittenEntry startNode() {
	return new WrittenEntry(null);
    }

    /**
     * Add a new entry after this one.  If this entry has already had a node
     * added (possibly by another thread), <code>add</code> will skip ahead to
     * the end of the list.
     */
    synchronized WrittenEntry add(EntryRep rep) {
	WrittenEntry entry = this;
	for (entry = this; entry.next != null; entry = entry.next)
	    continue;
	return (entry.next = new WrittenEntry(rep));
    }

    /**
     * Return the time stamp when this entry was written.
     */
    long timestamp() {
	return timestamp;
    }

    /**
     * Return the EntryRep of this entry.  If this is the first node (the one
     * returned by <code>startNode</code>), the EntryRep is <code>null</code>.
     *
     * @see #startNode
     */
    EntryRep rep() {
	return rep;
    }

    /**
     * Return the next object in the list.
     */
    WrittenEntry next() {
	return next;
    }

    public String toString() {
	return
	    (rep + " (" + timestamp + ") ->" + (next == null ? " null" : ""));
    }
}
