/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.reggie;

import java.util.HashMap;
import java.lang.ref.WeakReference;
import java.lang.ref.ReferenceQueue;

/**
 * Maps EntryClass and ServiceType instances to canonical instances, so
 * that equals and isAssignableFrom methods work correctly.  Two instances
 * are considered equivalent if they have the same class name.
 */
class ClassResolver {

    /** Map from String (classname) to WeakRep(ServiceType) */
    private final HashMap serviceMap = new HashMap();
    /** Reference queue for WeakReps of serviceMap */
    private final ReferenceQueue serviceRefQueue = new ReferenceQueue();
    /** Map from String (classname) to WeakRep(EntryClass) */
    private final HashMap entryMap = new HashMap();
    /** Reference queue for WeakReps of entryMap */
    private final ReferenceQueue entryRefQueue = new ReferenceQueue();

    /** Simple constructor */
    public ClassResolver() {}

    /** Returns canonical instance of a ServiceType. */
    public ServiceType resolve(ServiceType stype) {
	if (stype == null)
	    return null;
	synchronized (serviceMap) {
	    /* remove any queued refs */
	    while (true) {
		WeakRep ref = (WeakRep)serviceRefQueue.poll();
		if (ref == null)
		    break;
		serviceMap.remove(ref.name);
	    }
	    WeakRep ref = (WeakRep)serviceMap.get(stype.getName());
	    if (ref != null) {
		ServiceType rstype = (ServiceType)ref.get();
		if (rstype != null)
		    return rstype;
	    }
	    resolve(stype.getInterfaces());
	    stype.canonical(resolve(stype.getSuperclass()));
	    serviceMap.put(stype.getName(), new WeakRep(stype.getName(),
							stype,
							serviceRefQueue));
	    return stype;
	}
    }

    /**
     * Replaces the elements of the ServiceType array with their
     * canonical instances.
     */
    public void resolve(ServiceType[] stypes) {
	if (stypes != null) {
	    for (int i = stypes.length; --i >= 0; ) {
		stypes[i] = resolve(stypes[i]);
	    }
	}
    }

    /** Returns canonical instance of an EntryClass. */
    public EntryClass resolve(EntryClass eclass) {
	if (eclass == null)
	    return null;
	synchronized (entryMap) {
	    /* remove any queued refs */
	    while (true) {
		WeakRep ref = (WeakRep)entryRefQueue.poll();
		if (ref == null)
		    break;
		entryMap.remove(ref.name);
	    }
	    WeakRep ref = (WeakRep)entryMap.get(eclass.getName());
	    if (ref != null) {
		EntryClass reclass = (EntryClass)ref.get();
		if (reclass != null) {
		    if (reclass.getNumFields() != eclass.getNumFields())
			throw new IncompatibleClassChangeError(
					eclass.getName() +
					": different number of public fields");
		    return reclass;
		}
	    }
	    eclass.canonical(resolve(eclass.getSuperclass()));
	    entryMap.put(eclass.getName(), new WeakRep(eclass.getName(),
						       eclass,
						       entryRefQueue));
	    return eclass;
	}
    }

    /**
     * Replace all embedded descriptors in an Item with their
     * canonical instances.
     */
    public void resolve(Item item) {
	item.serviceType = resolve(item.serviceType);
	resolve(item.attributeSets);
    }

    /**
     * Replace all embedded descriptors in a Template with their
     * canonical instances.
     */
    public void resolve(Template tmpl) {
	resolve(tmpl.serviceTypes);
	resolve(tmpl.attributeSetTemplates);
    }

    /**
     * Replace all embedded descriptors in an EntryRep array with their
     * canonical instances.
     */
    public void resolve(EntryRep[] attrSets) {
	if (attrSets != null) {
	    for (int i = attrSets.length; --i >= 0; ) {
		attrSets[i].eclass = resolve(attrSets[i].eclass);
	    }
	}
    }

    /**
     * Replace all embedded descriptors in an EntryRep array with their
     * canonical instances, allowing null EntryReps.
     */
    public void resolveWithNulls(EntryRep[] attrSets) {
	if (attrSets != null) {
	    for (int i = attrSets.length; --i >= 0; ) {
		if (attrSets[i] != null)
		    attrSets[i].eclass = resolve(attrSets[i].eclass);
	    }
	}
    }

    /**
     * A weak ref extended with the key for the map.
     */
    private static class WeakRep extends WeakReference {
	/** The key for the map */
	public String name;

	/** Always create a queued ref. */
	public WeakRep(String name, Object referent, ReferenceQueue refQueue)
	{
	    super(referent, refQueue);
	    this.name = name;
	}
    }
}
