/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.reggie;

import java.rmi.server.RMIClassLoader;
import java.io.IOException;

/**
 * A ServiceType is a descriptor for a class, packaged up for
 * transmission between client-side proxies and the registrar server.
 * Instances are never visible to clients, they are private to the
 * communication between the proxies and the server.
 * <p>
 * Although this class could have been done as a subclass of EntryClass,
 * it isn't because the two descriptor spaces are maintained independently
 * (i.e., equality is not defined).  If ServiceType was a subclass, then
 * we couldn't use typechecking to ensure that instances don't get mixed.
 * What we really want is implementation inheritance without type
 * inheritance, but the Java programming language doesn't provide a way to 
 * do that (delegation is not a good choice here).
 * <p>
 * This class only has a bare minimum of methods, to minimize
 * the amount of code downloaded into clients.
 * <p>
 * Equality is based on object identity, on the assumption that
 * ClassResolver is always used and that equality is only interesting
 * in the registrar.
 *
 * @see ClassMapper
 * @see ClassResolver
 */
class ServiceType implements java.io.Serializable {

    /**
     * Class name
     *
     * @serial
     */
    protected String name;
    /**
     * Descriptor for the superclass
     *
     * @serial
     */
    protected ServiceType superclass;
    /**
     * Descriptor for the interfaces.  As a special case, interfaces is
     * null for the descriptor for java.lang.Object, and non-null otherwise.
     * This avoids carrying a boolean isInterface around.
     *
     * @serial
     */
    protected ServiceType[] interfaces;
    /**
     * An instance containing only name, no supertype info.
     * This is only used on the registrar side, to minimize the amount
     * of info transmitted back to clients.
     */
    protected transient ServiceType replacement;
    /** The concrete classes. */
    protected transient ServiceType[] concreteClasses;

    private static final long serialVersionUID = 9194732528513491231L;
    private static final ServiceType[] empty = {};

    /** Should only be called by ClassMapper */
    public ServiceType(Class clazz,
		       ServiceType superclass,
		       ServiceType[] interfaces)
    {
	name = clazz.getName();
	this.superclass = superclass;
	if (clazz != Object.class)
	    this.interfaces = interfaces;
    }

    /**
     * Constructor used for creating replacement instances,
     * containing only name.
     */
    private ServiceType(ServiceType stype) {
	name = stype.name;
    }

    /** Return the class name */
    public String getName() {
	return name;
    }

    /** Return the superclass descriptor */
    public ServiceType getSuperclass() {
	return superclass;
    }

    /** Return the interfaces.  The array is not a copy; do not modify it. */
    public ServiceType[] getInterfaces() {
	if (interfaces != null)
	    return interfaces;
	return empty;
    }

    /** Return the replacement, if any, containing only name and rep. */
    public ServiceType getReplacement() {
	return replacement;
    }

    /**
     * This is really only needed in the registrar, but it's very
     * convenient to have here.
     * @see Class#isAssignableFrom
     */
    public boolean isAssignableFrom(ServiceType cls) {
	if (this == cls)
	    return true;
	if (isInterface()) {
	    if (isAssignableFrom(cls.interfaces))
		return true;
	    for (ServiceType sup = cls.superclass;
		 sup != null && sup.interfaces != null;
		 sup = sup.superclass)
	    {
		if (isAssignableFrom(sup.interfaces))
		    return true;
	    }
	} else {
	    for (ServiceType sup = cls.superclass;
		 sup != null;
		 sup = sup.superclass)
	    {
		if (this == sup)
		    return true;
	    }
	}
	return false;
    }

    /** Test if this isAssignableFrom any of the given interface types. */
    private boolean isAssignableFrom(ServiceType[] ifaces)
    {
	for (int i = ifaces.length; --i >= 0; ) {
	    if (this == ifaces[i] || isAssignableFrom(ifaces[i].interfaces))
		return true;
	}
	return false;
    }

    /** @see Class#isInterface */
    public boolean isInterface() {
	return (superclass == null && interfaces != null);
    }

    /** Converts the descriptor to a Class instance, loading from codebase */
    public Class toClass(String codebase)
	throws IOException, ClassNotFoundException
    {
	return RMIClassLoader.loadClass(codebase, name);
    }

    /**
     * Sets this descriptor to be the canonical one.  Replaces the
     * superclass with the canonical superclass, and constructs the
     * replacement object.
     *
     * This should only be called by ClassResolver.
     */
    public void canonical(ServiceType superclass) {
	this.superclass = superclass;
	replacement = new ServiceType(this);
	concreteClasses = empty;
    }

    /**
     * Returns the set of all ServiceTypes T such that there is at least one
     * registered service with serviceType equal to T and for which
     * this.isAssignableFrom(T).  This is only used on the registrar side.
     */
    public ServiceType[] getConcreteClasses() {
	return concreteClasses;
    }

    /** Sets the concrete classes. */
    public void setConcreteClasses(ServiceType[] concreteClasses) {
	if (concreteClasses.length == 0)
	    concreteClasses = empty;
	this.concreteClasses = concreteClasses;
    }
}
