/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.start;

import java.io.File;

import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;

import java.security.SecureClassLoader;

import java.util.ArrayList;
import java.util.StringTokenizer;

/** 
 * This class provides utilities that can be useful when creating and
 * manipulating class loaders.
 */
public class ClassLoaderUtil {

    /** Flags used by the static utility methods of this class that are
     *  involved with class loaders. For more information, refer to the
     *  documentation for the methods in which these flags are used.
     */
    public static final int PARENT_DEFAULT       = 0;
    public static final int PARENT_EXTENSION     = (PARENT_DEFAULT-1);
    public static final int PARENT_SYSTEM_PARENT = (PARENT_DEFAULT-2);

    /** Special class loader that causes the RMI sub-system to annotate
     *  classes loaded with this class loader with the given codebase
     *  rather than with the usual search path urls. A class loader such
     *  as this helps facilitate the separation of classpath and codebase.
     *  <p>
     *  To help understand the utility of this class, recall the following
     *  characteristics of the RMI sub-system with respect to class loaders: 
     * <p>
     *   <ul>
     *     <li>  when the RMI sub-system marshals an object, it invokes
     *           the <code>getURLs</code> method on the class loader
     *           associated with the object being marshalled, retrieving
     *           the urls making up that class loader's search path. It
     *           then uses those urls to annotate the serialization stream.
     *     <li>  when the RMI sub-systsem loads a class, it creates a new
     *           instance of <code>URLClassLoader</code> that has a search
     *           path equal to the urls contained in the codebase annotation
     *           of the class being loaded; and which has as its parent
     *           class loader, the current context class loader. Note that
     *           the RMI sub-system only creates a new class loader if it has
     *           not already created a class loader with a search path that
     *           allows it to load the desired class (recall that RMI caches
     *           the class loaders it creates, and so it certainly may
     *           have already created a class loader with the "appropriate"
     *           search path).
     *  </ul>
     * <p>
     * Since the RMI sub-system operates as described above, this class 
     * can be used to cause RMI, when marshalling classes whose class loader
     * is an instance of this class, to annotate those classes with a
     * codebase that is different than the search path of the class loader.
     * Thus, when the marshalled class is un-marshalled, a new class loader
     * will be created with that annotated codebase as search path.
     */
    static class CodebaseClassLoader extends URLClassLoader {
        private URL[] codebase;
        /** Constructs an instance of this class having the given parent.
         *
         * @param searchPath <code>URL</code> array in which each element
         *                   corresponds to a component of the search path
         *                   for the new class loader; that is, the array
         *                   elements are the locations from which the new
         *                   class loader will load requested classes.
         * @param parent     the class loader to set as the parent of the
         *                   new <code>ClassLoader</code> instance
         * @param codebase   <code>URL</code> array in which each element
         *                   corresponds to a codebase location with which
         *                   the RMI sub-system will annotate classes.
         */
        CodebaseClassLoader(URL[] searchPath,
                            ClassLoader parent,
                            URL[] codebase)
        {
            super(searchPath,parent);
            this.codebase = codebase;
        }//end constructor

        /** Constructs an instance of this class with default parent
         * (typically the system class loader).
         *
         * @param searchPath <code>URL</code> array in which each element
         *                   corresponds to a component of the search path
         *                   for the new class loader; that is, the array
         *                   elements are the locations from which the new
         *                   class loader will load requested classes.
         * @param codebase   <code>URL</code> array in which each element
         *                   corresponds to a codebase location with which
         *                   the RMI sub-system will annotate classes.
         */
        CodebaseClassLoader(URL[] searchPath, URL[] codebase) {
            super(searchPath);
            this.codebase = codebase;
        }//end constructor

        /* Override getURLs so that RMI will use the codebase to annotate 
         * the classes loaded with this class loader, rather than the urls
         */
        public URL[] getURLs() {
            return codebase;
        }
    }//end class CodebaseClassLoader

    /**
     * Method that creates a <code>URLClassLoader</code> that will be
     * configured to load requested classes from the locations indicated
     * by the <code>searchPath</code> parameter; and which will, when
     * the RMI sub-system marshals classes that are loaded through the
     * object returned by this method, cause the resulting serialization
     * stream to be annotated with the contents of the <code>codebase</code>
     * parameter.
     * <p>
     * The structure of the delegation tree in which the returned class loader
     * will reside depends on the value of the <code>nParentsToSkip</code>
     * parameter. That parameter provides control over the structure (that
     * is, the ancestoral relationships) of the delegation tree. For
     * details on the possible values of this parameter, and the resulting
     * delegation tree structure that results, refer to the method
     * <code>getURLClassLosder</code> in this class.
     *
     * @param searchPath     <code>String</code> containing the location(s)
     *                       from which the new class loader will load
     *                       requested classes. This parameter represents
     *                       a classpath; that is, path/filenames separated
     *                       by path separators.
     * @param codebase       <code>String</code> containing the url(s)
     *                       with which the RMI sub-system will annotate
     *                       classes. This parameter represents a codebase;
     *                       that is, a set of urls separated by spaces.
     * @param nParentsToSkip <code>int</code> value indicating the structure
     *                       of the delegation tree with which to create the
     *                       desired class loader. This parameter takes one
     *                       of the following values: 
     *                       <ul> <li> StartUtil.PARENT_DEFAULT 
     *                            <li> StartUtil.PARENT_EXTENSION 
     *                            <li> StartUtil.PARENT_SYSTEM_PARENT 
     *                            <li> any positive integer value </ul>
     *
     * @see com.sun.jini.start.ClassLoaderUtil#getURLClassLoader
     * @see com.sun.jini.start.ClassLoaderUtil#CodebaseClassLoader
     */
    public static URLClassLoader getCodebaseClassLoader(String searchPath,
                                                        String codebase,
                                                        int    nParentsToSkip)
                                                  throws MalformedURLException
    {
        URL[] searchPathURLs = getClasspathURLs(searchPath);
        URL[] codebaseURLs   = getCodebaseURLs(codebase);
        ArrayList loaderList = null;
        switch(nParentsToSkip) {
            case PARENT_DEFAULT:
                return (new CodebaseClassLoader(searchPathURLs,codebaseURLs));
            case PARENT_EXTENSION: //explicitly use the extesion cl as parent
                loaderList = getCurClassLoaderTree();
                int n = loaderList.size()-2;// index where extension should be
                ClassLoader extClassLoader = (ClassLoader)(loaderList.get(n));
                return ( new CodebaseClassLoader(searchPathURLs,
                                                 extClassLoader,
                                                 codebaseURLs) );
            case PARENT_SYSTEM_PARENT: //use parent of system cl as parent
                ClassLoader sysCL = ClassLoader.getSystemClassLoader();
                ClassLoader sysParentCL = sysCL.getParent();
                return ( new CodebaseClassLoader(searchPathURLs,
                                                 sysParentCL,
                                                 codebaseURLs) );
            default: //skip the requested number of parents in the tree
                loaderList = getCurClassLoaderTree();
                for(int i=0;i<nParentsToSkip;i++) {
                    loaderList.remove(0);
                }
                ClassLoader parentClassLoader = (ClassLoader)loaderList.get(0);
                return ( new CodebaseClassLoader(searchPathURLs,
                                                 parentClassLoader,
                                                 codebaseURLs) );
        }//end case
    }//end getCodebaseClassLoader

    /**
     * Method that creates a <code>URLClassLoader</code> that will be 
     * configured to load requested classes from the locations indicated
     * by the <code>searchPath</code> parameter (which may be a classpath
     * or a codebase, as indicated by the <code>isClasspath</code>parameter);
     * and which will have a delegation tree whose structure is determined
     * by the <code>nParentsToSkip</code> parameter.
     * <p>
     * How this method treats the <code>searchPath</code> parameter is
     * determined by the <code>isClasspath</code>parameter. If that parameter
     * is <code>true</code>, this method will create the desired class loader
     * using file urls. If the value of parameter is <code>false</code>,
     * this method will constuct the class loader using default urls.
     * <p>
     * The structure of the delegation tree in which the returned class loader
     * will reside depends on the value of the <code>nParentsToSkip</code>
     * parameter. That parameter provides control over the structure (that is,
     * the ancestoral relationships) of the delegation tree. The possible
     * values, and the resulting delegation tree structure that results,
     * are as follows:
     * <p>
     *   <ul>
     *     <li> StartUtil.PARENT_DEFAULT - creates a class loader with 
     *          delegation tree with the default structure; that is, the
     *          parent of the new class loader will be the current 
     *          "system class loader" (often referred to as the "application
     *          class loader". The resulting delegation tree will be of the
     *          form:
     *      'bootCL--extensionCL--parentCL0-- ... --parentCLn--systemCL--newCL'
     *     <li> StartUtil.PARENT_EXTENSION - creates a class loader with 
     *          delegation tree of the form:
     *                   'bootCL--extensionCL--newCL';
     *          that is, there are no other class loaders between the
     *          "extension class loader" and the class loader created by
     *          this method.
     *     <li> StartUtil.PARENT_SYSTEM_PARENT - creates a class loader whose
     *          parent class loader is the parent loader of the current
     *          system class loader. Thus, the new class loader will belong
     *          to a tree of the form:
     *          'bootCL--extensionCL--parentCL0-- ... --parentCLn--newCL',
     *          where the "system class loader" has effectively been 
     *          "skipped" in the tree. Typically, the system class loader
     *          is the default delegation parent of any new ClassLoader
     *          instances. Thus, the system class loader typically resides
     *          in a delegation tree of the form: 
     *          'bootCL--extensionCL--parentCL0-- ... --parentCLn--systemCL'
     *          Thus, the 'systemCL' component of the tree is "skipped" when
     *          the nParentsToSkip parameter is equal to PARENT_SYSTEM_PARENT.
     *     <li> Any positive value - creates a class loader with delegation
     *          tree of the form: 
     *          'bootCL--extensionCL--parentCL0-- ... --parentCLi--newCL',
     *          where the current tree is "walked", skipping the desired
     *          number of parent class loaders in the current tree; where
     *          the current tree is of the form: 
     *   'bootCL--extensionCL--parentCL0-...-parentCLi-...-parentCLn--systemCL'
     *  </ul>
     * <p>
     * @param searchPath     <code>String</code> containing the location(s)
     *                       from which the new class loader will load
     *                       requested classes. This parameter represents
     *                       either a classpath (path/filenames separated by
     *                       path separators), or a codebase (urls separated
     *                       by spaces).
     * @param isClasspath    <code>boolean</code> value indicating how this
     *                       method should treat the <code>searchPath</code>
     *                       parameter: as a classpath or a codebase.
     * @param nParentsToSkip <code>int</code> value indicating the structure
     *                       of the delegation tree with which to create the
     *                       desired class loader. This parameter takes one
     *                       of the following values: 
     *                       <ul> <li> StartUtil.PARENT_DEFAULT 
     *                            <li> StartUtil.PARENT_EXTENSION 
     *                            <li> StartUtil.PARENT_SYSTEM_PARENT 
     *                            <li> any positive integer value </ul>
     *
     * @return an instance of <code>URLClassLoader</code> with characteristics
     *         as indicated by the values of the input parameters of this
     *         method.
     *                       
     * @throws java.net.MalformedURLException 
     */
    public static URLClassLoader getURLClassLoader(String  searchPath,
                                                   boolean isClasspath,
                                                   int     nParentsToSkip)
                                                  throws MalformedURLException
    {
        URL[] urls = ( (isClasspath) ?  getClasspathURLs(searchPath)
                                      : getCodebaseURLs(searchPath)  );
        ArrayList loaderList = null;
        switch(nParentsToSkip) {
            case PARENT_DEFAULT:
                return ( new URLClassLoader(urls) );
            case PARENT_EXTENSION: //explicitly use the extesion cl as parent
                loaderList = getCurClassLoaderTree();
                int n = loaderList.size()-2;// index where extension should be
                ClassLoader extClassLoader = (ClassLoader)(loaderList.get(n));
                return ( new URLClassLoader(urls,extClassLoader) );
            case PARENT_SYSTEM_PARENT: //use parent of system cl as parent
                /* Retrieve the default delegation parent for new ClassLoader
                 * instances (typically, this is the class loader that was
                 * used to start the application). 
                 */
                ClassLoader sysClassLoader
                                        = ClassLoader.getSystemClassLoader();
                /* Retrieve the parent of the class loader just retrieved */
                ClassLoader sysParentClassLoader = sysClassLoader.getParent();
                /* Return a new class loader in which the default delegation
                 * parent is removed (or "skipped") in the tree in favor of the
                 * default parent's own parent loader.
                 */
                return ( new URLClassLoader(urls,sysParentClassLoader) );
            default: //skip the requested number of parents in the tree
                loaderList = getCurClassLoaderTree();
                for(int i=0;i<nParentsToSkip;i++) {
                    loaderList.remove(0);
                }
                ClassLoader parentClassLoader = (ClassLoader)loaderList.get(0);
                return ( new URLClassLoader(urls,parentClassLoader) );
        }//end case

    }//end getURLClassLoader

    /**
     * Method that creates a <code>URLClassLoader</code> that will be 
     * configured to load requested classes from the locations indicated
     * by the <code>searchPath</code> parameter (which may be a classpath
     * or a codebase, as indicated by the <code>isClasspath</code>parameter);
     * and which will have a default delegation tree.
     *
     * @param searchPath     <code>String</code> containing the location(s)
     *                       from which the new class loader will load
     *                       requested classes. This parameter represents
     *                       either a classpath (path/filenames separated by
     *                       path separators), or a codebase (urls separated
     *                       by spaces).
     * @param isClasspath    <code>boolean</code> value indicating how this
     *                       method should treat the <code>searchPath</code>
     *                       parameter: as a classpath or a codebase.
     *
     * @return an instance of <code>URLClassLoader</code> with characteristics
     *         as indicated by the values of the input parameters of this
     *         method.
     *                       
     * @throws java.net.MalformedURLException 
     */
    public static URLClassLoader getURLClassLoader(String  searchPath,
                                                   boolean isClasspath)
                                                  throws MalformedURLException
    {
        return getURLClassLoader(searchPath,isClasspath,PARENT_DEFAULT);
    }//end getURLClassLoader

    /**
     * Method that creates a <code>URLClassLoader</code> that will be 
     * configured to load requested classes from the locations indicated
     * by the <code>searchPath</code> parameter (which is assumed to be
     * a codebase); and which will have a default delegation tree.
     *
     * @param searchPath     <code>String</code> containing the location(s)
     *                       from which the new class loader will load
     *                       requested classes. This parameter represents
     *                       is assumed to be a codebase (urls separated
     *                       by spaces).
     *
     * @return an instance of <code>URLClassLoader</code> with characteristics
     *         as indicated by the values of the input parameters of this
     *         method.
     *                       
     * @throws java.net.MalformedURLException 
     */
    public static URLClassLoader getURLClassLoader(String searchPath)
                                                  throws MalformedURLException
    {
        return getURLClassLoader(searchPath,false,PARENT_DEFAULT);
    }//end getURLClassLoader


    /**
     * Utility method that converts the components of a <code>String</code>
     * representing a classpath into file urls; returning an array in which
     * each element of the array is an instance of <code>URL</code>.
     *
     * @param classpath <code>String</code> containing components separated
     *                  by path separators that represent the components
     *                  making up a classpath
     *
     * @return an array containing instances of <code>URL</code> in which
     *         each element of the array corresponds to one of the components
     *         in the <code>classpath</code> parameter
     *                       
     * @throws java.net.MalformedURLException 
     */
    public static URL[] getClasspathURLs(String classpath)
                                                 throws MalformedURLException
    {
        StringTokenizer st = new StringTokenizer(classpath,File.pathSeparator);
        URL[] urls = new URL[st.countTokens()];
        for (int i=0; st.hasMoreTokens(); i++) {
            urls[i] = new File(st.nextToken()).toURL();
        }
        return urls;
    }//end getClasspathURLs

    /**
     * Utility method that constructs from the given array of <code>URL</code>
     * instances, a single space-separated <code>String</code>. The 
     * <code>String</code> returned by this method can be used for things
     * such as the  <code>location</code> argument when constructing an
     * instance of <code>ActivationDesc</code>.
     *
     * @param urls <code>URL</code> array containing containing the elements
     *             from which the return value will be constructed
     *
     * @return a single space-separated <code>String</code>. The 
     *         <code>String</code> returned by this method can be used
     *         for things such as the  <code>location</code> argument
     *         when constructing an instance of <code>ActivationDesc</code>
     */
    public static String URLsToString(URL[] urls) {
        if ( (urls == null) || (urls.length == 0) ) return new String("");
        StringBuffer buffer = new StringBuffer(urls[0].toString());
        for(int i=1;i<urls.length;i++) {
            buffer.append(" "+urls[i].toString());
        }
        return  buffer.toString();
    }//end URLsToString

    /**
     * Utility method that converts the components of a <code>String</code>
     * representing a codebase into standard urls; returning an array in which
     * each element of the array is an instance of <code>URL</code>.
     *
     * @param codebase  <code>String</code> containing components separated
     *                  by spaces in which each component is in url format
     *                  (http, ftp, etc)
     *
     * @return an array containing instances of <code>URL</code> in which
     *         each element of the array corresponds to one of the components
     *         in the <code>codebase</code> parameter
     *                       
     * @throws java.net.MalformedURLException 
     */
    public static URL[] getCodebaseURLs(String codebase)
                                                 throws MalformedURLException
    {
        StringTokenizer st = new StringTokenizer(codebase);
        URL[] urls = new URL[st.countTokens()];
        for (int i=0; st.hasMoreTokens(); i++) {
            urls[i] = new URL(st.nextToken());
        }
        return urls;
    }//end getCodebaseURLs

    /**
     * Utility method that retrieves the components making up the class loader
     * delegation tree for the current context class loader and returns each
     * in an <code>ArrayList</code>.
     *
     * @return an <code>ArrayList</code> instance in which each element of the
     *         list is one of the components making up the current delegation
     *         tree.
     */
    public static ArrayList getCurClassLoaderTree() {
        Thread curThread = Thread.currentThread();
        ClassLoader curClassLoader = curThread.getContextClassLoader();
        return getClassLoaderTree(curClassLoader);
    }//end getCurClassLoaderTree

    /**
     * Utility method that retrieves the components making up the class loader
     * delegation tree for the given <code>classloader</code>parameter and
     * returns each in an <code>ArrayList</code>.
     *
     * @param classloader <code>ClassLoader</code> instance whose delegation
     *                    tree is to be retrieved and returned
     *
     * @return an <code>ArrayList</code> instance in which each element of the
     *         list is one of the components making up the delegation tree
     *         of the given class loader.
     */
    public static ArrayList getClassLoaderTree(ClassLoader classloader) {
        ArrayList loaderList = new ArrayList();
        ClassLoader curClassLoader = classloader;
        loaderList.add(curClassLoader);
        ClassLoader parentClassLoader = curClassLoader.getParent();
        loaderList.add(parentClassLoader);
        while(parentClassLoader != null) {
            parentClassLoader = parentClassLoader.getParent();
            loaderList.add(parentClassLoader);
        }//end loop
        return loaderList;
    }//end getClassLoaderTree

    /**
     * Utility method that diplays the class loader delegation tree for
     * the current context class loader. For each class loader in the tree,
     * this method displays the locations from which that class loader
     * will retrieve and load requested classes.
     * <p>
     * This method can be useful when debugging problems related to the
     * receipt of exceptions such as <code>ClassNotFoundException</code>.
     */
    public static void displayCurClassLoaderTree() {
        Thread curThread = Thread.currentThread();
        ClassLoader curClassLoader = curThread.getContextClassLoader();
        displayClassLoaderTree(curClassLoader);
    }//end displayCurClassLoaderTree

    /**
     * Utility method that diplays the class loader delegation tree for
     * the given class loader. For each class loader in the tree, this
     * method displays the locations from which that class loader will
     * retrieve and load requested classes.
     * <p>
     * This method can be useful when debugging problems related to the
     * receipt of exceptions such as <code>ClassNotFoundException</code>.
     *
     * @param classloader <code>ClassLoader</code> instance whose delegation
     *                    tree is to be displayed
     */
    public static void displayClassLoaderTree(ClassLoader classloader) {
        ArrayList loaderList = getClassLoaderTree(classloader);
        System.out.println("");
        System.out.println
                         ("ClassLoader Tree has "+loaderList.size()+" levels");
        System.out.println("  cl0 -- boot ClassLoader");
        System.out.println("   |");
        System.out.println("  cl1 -- extension ClassLoader");
        for(int i=((loaderList.size()-2)-1),k=2;i>=0;i--,k++) {
            System.out.println("   |");
            ClassLoader curClassLoader = (ClassLoader)loaderList.get(i);
            System.out.print("  cl"+k+" -- ClassLoader "
                                   +curClassLoader+": ");
            if(curClassLoader instanceof URLClassLoader) {
                URL[] urls = ((URLClassLoader)(curClassLoader)).getURLs();
                if(urls != null) {
                    System.out.print(urls[0]);
                    for(int j=1;j<urls.length;j++){
                        System.out.print(", "+urls[j]);
                    }
                } else {//urls == null
                    System.out.print("null search path");
                }//endif
            } else {
                if(curClassLoader instanceof SecureClassLoader) {
                    System.out.print("is instance of SecureClassLoader");
                } else {
                    System.out.print("is unknown ClassLoader type");
                }
            }//endif
            System.out.println("");
        }//end loop
            System.out.println("");
    }//end displayClassLoaderTree

}//end class ClassLoaderUtil
