/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.start;

import com.sun.jini.start.StartUtil;

import java.net.ConnectException;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.Socket;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.net.URL;
import java.io.BufferedReader;
import java.io.DataOutputStream;
import java.io.File;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.InterruptedIOException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.StringTokenizer;

/** Class which can be used to determines if an HTTP server daemon is
 *  currently up, running, and has access to a desired jar file.
 */
public class HTTPDStatus {

    public  static final  int DOWN          = 0;
    public  static final  int UP            = 1;
    public  static final  int UP_BUT_NO_JAR = 2;

    private static String defProtocol = "no_protocol";
    private static String defHost     = getLocalHost();
    private static int    defPort     = 8080;
    private static String defJarFile  = "reggie-dl.jar";
    private static int    timeout     = 1*1000; // wait 1 second for data

    /**
     * Command line interface executing a determination of whether or not
     * an HTTP server daemon is up, running, and has access to a desired
     * jar file. This method parses the command line, determines if the
     * appropriate valid arguments have been input, and then invokes a
     * method which determines the status of the HTTP server referenced
     * by the input arguments.
     *
     * @param args <code>String</code> array containing the command line
     *             arguments
     */
    public static void main(String[] argv) {
        String host    = defHost;
        int    port    = defPort;
        String jarFile = defJarFile;

        if(argv.length > 3) {
            System.err.println("HTTPDUp <hostname> <port> <jarFilename>");
            return;
        }//endif
        // assume the first integer is the port
        int portIndx = -1;
        for(int i=0;i<argv.length;i++) {
            try {
                int tmpPort = Integer.parseInt(argv[i]);
                port = tmpPort;
                portIndx = i;
                break;
            } catch(NumberFormatException e) { }
        }//endloop
        // assume the first String ending in ".jar" is the jar file to download
        int jarIndx = -1;
        for(int i=0;i<argv.length;i++) {
            if(argv[i].endsWith(".jar")) {
                jarFile = argv[i];
                jarIndx = i;
                break;
            }
        }//endloop
        // assume the first argument that is not a port or jar file is a host
        for(int i=0;i<argv.length;i++) {
            if( (portIndx == i) || (jarIndx == i) ) continue;
            host = argv[i];
            break;
        }//endloop
	int status = httpdStatus(host,port,jarFile);
        System.out.println(status);
    }//end main

    /**
     * This method determines the current status of the HTTP server(s)
     * satisfying the configuration reflected in the  <code>codebase</code>
     * parameter.
     * <p>
     * This method uses the host name and port number referenced in the
     * <code>codebase</code> parameter to poll for the existence of a running
     * HTTP server with access to the jar file also referenced in the
     * <code>codebase</code> parameter.
     *
     * @param codebase <code>String</code> containing the codebase to poll
     *                 for the existence of a running HTTP server with
     *                 access to the jar file referenced in this parameter
     * 
     * @return <code>int</code> value equal to one of the public constants
     *         defined in this class (DOWN,UP,UP_BUT_NO_JAR) that indicates
     *         the current status of the HTTP server satisfying the
     *         configuration reflected in the  <code>codebase</code> parameter
     */
    public static int httpdStatus(String codebase) {
        /* Step through each codebase in the string, validating each */
        String codebaseHost    = defHost;
        int    codebasePort    = defPort;
        String codebaseJarFile = defJarFile;
        StringTokenizer st = new StringTokenizer(codebase," \"");
        for (int i = 0; st.hasMoreTokens(); i++) {
            String url = st.nextToken();
            try {
                URL codebaseUrl = new URL(url);
                codebaseHost = codebaseUrl.getHost();
                int tmpPort  = codebaseUrl.getPort();
                if(tmpPort != -1) codebasePort = tmpPort;
                codebaseJarFile = codebaseUrl.getFile();
            } catch(MalformedURLException e) {
                try {
                    codebaseHost = InetAddress.getLocalHost().getHostName();
                } catch(UnknownHostException e1) { }
            }
            int status = httpdStatus(codebaseHost,
                                     codebasePort,
                                     codebaseJarFile);
            if(status != UP) return status;
        }//endloop
        return UP;
    }//end httpdStatus

    /**
     * Like the version of this method that takes only a <code>codebase</code>
     * parameter and returns a status value, this method also determines the
     * current status of the HTTP server(s) satisfying the configuration
     * reflected in the <code>codebase</code> parameter. But rather than
     * returning a status value, this method will test each component of the
     * <code>codebase</code> parameter and display a warning message if
     * it is determined that a potential problem exists.
     *
     * @param codebase  <code>String</code> containing the codebase to poll
     *                  for the existence of a running HTTP server with
     *                  access to the jar file referenced in this parameter
     * @param startUtil an instance of the <code>StartUtil</code> class,
     *                  configured with the appropriate resource bundle
     *                  information, which is used as the mechanism through
     *                  which warning message(s) will be displayed
     */
    public static void httpdStatus(String codebase,StartUtil startUtil) {
        /* Step through each codebase in the string, validating each */
        String codebaseHost     = defHost;
        int    codebasePort     = defPort;
        String codebaseJarFile  = defJarFile;
        StringTokenizer st = new StringTokenizer(codebase," \"");
        for (int i = 0; st.hasMoreTokens(); i++) {
            String curProtocol = defProtocol;
            String url = st.nextToken();
            try {
                URL codebaseUrl = new URL(url);
                curProtocol = codebaseUrl.getProtocol();
                codebaseHost = codebaseUrl.getHost();
                int tmpPort  = codebaseUrl.getPort();
                if(tmpPort != -1) codebasePort = tmpPort;
                codebaseJarFile = codebaseUrl.getFile();
            } catch(MalformedURLException e) {
                try {
                    codebaseHost = InetAddress.getLocalHost().getHostName();
                } catch(UnknownHostException e1) { }
            }
            String urlStr = "("+url+")";
            if( ("http").equals(curProtocol) || ("ftp").equals(curProtocol) ) {
                int status = httpdStatus(codebaseHost,
                                         codebasePort,
                                         codebaseJarFile);
                if (status == HTTPDStatus.DOWN) {
                    startUtil.print("httpserver.down", urlStr);
                } else if(status == HTTPDStatus.UP_BUT_NO_JAR) {
	            startUtil.print("httpserver.nojar", urlStr);
                }//endif
            } else {
                if(("file").equals(curProtocol)) {
                    startUtil.print("fileurl.warning", urlStr);
                } else {
                    startUtil.print("nonstandardurl.warning", urlStr);
                }
            }
        }//endloop
    }//end httpdStatus

    /**
     * This method determines the current status of the HTTP server
     * satisfying the configuration reflected in the <code>host</code>,
     * <code>port</code>, and <code>jarFile</code> parameters.
     * This method uses the <code>host</code> and <code>port</code>
     * parameters to poll for the existence of a running HTTP server
     * with access to the jar file referenced in the <code>jarFile</code>
     * parameter.
     *
     * @param host    <code>String</code> containing the name of the host on
     *                which to poll for the desired HTTP server
     * @param port    <code>int</code> value representing the port number of
     *                the host on which to poll for the desired HTTP server
     * @param jarFile <code>String</code> containing the full path and file
     *                name of a jar file. If this method determines that an
     *                HTTP server is running on the indicated <code>host</code>
     *                and is listening on the indicated <code>port</code>
     *                number, then a determination will be made as to whether
     *                that HTTP server can access (can serve) the jar file
     *                referenced by the contents of this parameter.
     * 
     * @return <code>int</code> value equal to one of the public constants
     *         defined in this class (DOWN,UP,UP_BUT_NO_JAR) that indicates
     *         the current status of the HTTP server satisfying the
     *         configuration reflected in the  <code>codebase</code> parameter
     */
    public static int httpdStatus(String host, int port, String jarFile) {
        Socket s = openSocket(host,port);
        if(s != null) {
            writeGetToSocket(s,jarFile);
            String[] headerArray = getHeader(s);
            boolean httpdOK = replyOK(headerArray);
            drainSocket(s);
            closeSocket(s);
            if(httpdOK) {
                return UP;
            } else { // could connect, but couldn't retrieve jar file
                return UP_BUT_NO_JAR;
            }
        } else { // couldn't even connect ==> down
            return DOWN;
        }
    }//end httpdStatus

    /**
     * This method is an alternate implementation of the 3-arg version of
     * the <code>httpdStatus</code> method of this class. This method is
     * much simpler and more general than its alternate form. This method
     * is presented as the alternate form of <code>httpdStatus</code>
     * because, unfortunately, when the HTTP server is running on Windows,
     * it causes the error 'socket write error <code=10053>'; which is
     * annoying, but seems to be harmless. Also, this method only determines
     * if the HTTP server is up or down; it doesn't determine the status
     * of the server's ability to access the jar file.
     *
     * @param host    <code>String</code> containing the name of the host on
     *                which to poll for the desired HTTP server
     * @param port    <code>int</code> value representing the port number of
     *                the host on which to poll for the desired HTTP server
     * @param jarFile <code>String</code> containing the full path and file
     *                name of a jar file. If this method determines that an
     *                HTTP server is running on the indicated <code>host</code,
     *                and is listening on the indicated <code>port</code>
     *                number, then a determination will be made as to whether
     *                that HTTP server can access (can serve) the jar file
     *                referenced by the contents of this parameter.
     * 
     * @return <code>int</code> value equal to one of the public constants
     *         defined in this class (DOWN or UP) that indicates the
     *         tcurrent status of the HTTP server satisfying the cnfiguration
     *         reflected in the  <code>codebase</code> parameter
     */
    public static int httpdStatusAlt(String host, int port, String jarFile) {
        try {
            URL serverURL = new URL("http",host,port,"/"+jarFile);
            Object content = serverURL.getContent();
            if(content != null) {
                return UP;
            } else {
                return DOWN;
            }
        } catch(Exception e) {
            return DOWN;
        }
    }//end httpdStatusAlt

    /**
     * Returns the name of the host on which this method is being executed.
     * If that name cannot be obtained, the <code>String</code> "localhost"
     * will be returned.
     * 
     * @return <code>String</code> containing the name of the host on which
     *         this method is currently executing; or "localhost" if the
     *         actual name of the host cannot be obtained.
     */
    private static String getLocalHost() {
        String host = "localhost";
        try {
            host = InetAddress.getLocalHost().getHostName();
        } catch(UnknownHostException e) {
            System.err.println("\nOn hostname retrieval -- "+e.toString());
        }
        return host;
    }//end getLocalHost

    /**
     * Constructs and returns a stream socket to the destination corresponding
     * to the <code>host</code> and <code>port</code> parameters.
     * 
     * @param host    <code>String</code> containing the non-<code>null</code>
     *                name of the destination's host
     * @param port    <code>int</code> value (0 <= port <= 65535) containing
     *                the destination's port number
     *
     * @return <code>Socket</code> instance referencing a stream socket to the
     *         destination corresponding to the <code>host</code> and 
     *         <code>port</code> parameters
     */
    private static Socket openSocket(String host, int port) {
        Socket s = null;
        try {
            s = new Socket(host,port);
        } catch(Exception e) { }
        return s;
    }//end openSocket

    /**
     * Writes a retrieval request (an HTTP "GET" request) to the given socket
     * for the data reflected in the <code>request</code> parameter.
     * 
     * @param s       the <code>Socket</code> on which to write the request
     * @param request <code>String</code> referencing the data to request
     */
    private static void writeGetToSocket(Socket s, String request) {
        if( (s == null) || (request == null) ) return;
        try {
            DataOutputStream out = new DataOutputStream(s.getOutputStream());
            out.writeBytes("GET /"+request+"\r\n\r\n");
            out.flush();
        } catch (Exception e) {
            System.err.println("\nOn socket write -- "+e.toString());
        }
    }//end writeGetToSocket

    /**
     * Reads from the given socket, the header data sent by an HTTP server
     * as a result of a "GET" request; and returns that header data in 
     * a <code>String</code> array.
     * 
     * @param s the <code>Socket</code> from which to retrieve the header data
     *
     * @return <code>String</code> array that contains the header data 
     *         retrieved from the given socket
     */
    private static String[] getHeader(Socket s) {
        if(s == null) return new String[0];
        String headerStr = null;
        try {
            InputStream cReader  = s.getInputStream();
            StringBuffer buffer  = new StringBuffer();
            while(true) {
                int ch = cReader.read();
                if(ch < 0) {
                    break;
                } else {
                    buffer.append((char)ch);
                    int len = buffer.length();
                    if(len > 3) {
                        /* Test for end of header */
                        if(    (buffer.charAt(len-4) == '\r')
                            && (buffer.charAt(len-3) == '\n') 
                            && (buffer.charAt(len-2) == '\r') 
                            && (buffer.charAt(len-1) == '\n') )
                        {
                            headerStr = buffer.toString();
                            break; 
                        }
                    }//endif(len > 3)
                }//endif(ch<0)
            }//end while
        } catch (InterruptedIOException e) {
        } catch (IOException e) {
            System.err.println("\nWhile getting header -- "+e.toString());
        }//end try
        ArrayList headerList = new ArrayList();
        StringTokenizer st = new StringTokenizer(headerStr,"\r\n");
        while(st.hasMoreTokens()) {
            headerList.add(st.nextToken());
        }//endloop
        return ((String[])(headerList).toArray(new String[headerList.size()]));
    }//end getHeader

    /**
     * Examines the status field of the given header data, and returns
     * <code>true</code> if the contents of that field indicate that the
     * request that generated the header is valid and acceptable to the
     * server; and <code>false</code> otherwise.
     * 
     * @param header <code>String</code> containing the header data to examine
     *
     * @return <code>true</code> if the contents of given data indicate that
     *         the request that generated the header is valid and acceptable
     *         to the server; and <code>false</code> otherwise
     */
    private static boolean replyOK(String[] header) {
        boolean responseOK = false;
        if((header == null) || (header.length == 0)) return responseOK;
        if(header[0].startsWith("HTTP") && (header[0].indexOf("200 OK")>=0)){
            responseOK = true;
        }
        return responseOK;
    }//end readHeader

    /**
     * Examines the "Content-Length:" field of the given header data, and
     * returns the number of bytes on the socket to which the given header
     * refers.
     * 
     * @param header <code>String</code> containing the header data to examine
     *
     * @return <code>int</code> value that indicates the number of bytes on
     *         the socket to which the given header refers
     */
    private static int getNBytesToRead(String[] header) {
        int nBytes = -1; // number of bytes in body as indicated in header
        if((header == null) || (header.length < 2)) return nBytes;
        if(header[1].startsWith("Content-Length:")){
            StringTokenizer st = new StringTokenizer(header[1]);
            while(st.hasMoreTokens()) {
                String token = st.nextToken();
                try {
                    nBytes = Integer.parseInt(token);
                } catch(NumberFormatException e) { }
            }
        }
        return nBytes;
    }//end getNBytesToRead

    /**
     * Reads and discards all data currently on the given socket.
     * 
     * @param s the <code>Socket</code> from which to read data
     */
    private static void drainSocket(Socket s) {
        if(s == null) return;
        try {
            InputStream cReader = s.getInputStream();
            for(int ch = cReader.read(); ch >= 0; ch = cReader.read()) { }
        } catch (InterruptedIOException e) {
        } catch (IOException e) {
            System.err.println("\nWhile draining the socket -- "+e.toString());
        }
    }//end drainSocket

    /**
     * Reads and discards the next <code>nBytes</code> of data on the given
     * socket.
     * 
     * @param s      the <code>Socket</code> from which to read data
     * @param nBytes the specific number of bytes to read and discard
     */
    private static void drainSocket(Socket s, int nBytes) {
        if( (s == null) || (nBytes <= 0) ) return;
        try {
            InputStream cReader = s.getInputStream();
            int ch;
            int nBytesRead = 0;
            for(int i=0;i<nBytes;i++) {
                try {
                    ch = cReader.read();
                    if(ch == -1) {
                        break;
                    } else {
                        nBytesRead++;
                    }
                } catch (InterruptedIOException e) {
                    System.err.println("\nSocket read time out: "+nBytesRead
                                       +" bytes read before timeout occurred");
                }
            }//end loop
        } catch (IOException e) {
            System.err.println("\nWhile draining the socket -- "+e.toString());
        }
    }//end drainSocket

    /**
     * Changes the time-out period of the given socket to the given value
     * in milliseconds.
     * 
     * @param s         the <code>Socket</code> whose time-out period is to be
     *                  changed
     * @param timeoutMs the new value of the given socket's time-out period
     *                  in milliseconds
     */
    private static void setSocketTimeout(Socket s, int timeoutMs) {
        if(s == null) return;
        int t = (timeoutMs < 0 ? 0 : timeoutMs);
        try {
            s.setSoTimeout(t);
        } catch(IOException e) {
            System.err.println("\nOn setting socket timeout -- "+e.toString());
        }
    }//end setSocketTimeout

    /**
     * Changes the linger-on-close time-out period of the given socket to the
     * given value in milliseconds.
     * 
     * @param s         the <code>Socket</code> whose linger-on-close time-out
     *                  period is to be changed
     * @param timeoutMs a non-negative number specifying the maximum number of
     *                  milliseconds that <code>close</code> will block until
     *                  the given socket is forcibly closed. If this value is
     *                  zero, <code>close</code> will return immediately.
     */
    private static void setSocketLingerOn(Socket s, int timeoutMs) {
        if(s == null) return;
        int t = (timeoutMs < 0 ? 0 : timeoutMs);
        try {
            s.setSoLinger(true,t);
        } catch(IOException e) {
            System.err.println("\nOn setting socket linger on -- "
                               +e.toString());
        }
    }//end setSocketLingerOn

    /**
     * Disables the linger-on-close option of the given socket.
     * 
     * @param s the <code>Socket</code> whose linger-on-close option is to
     *          be disabled
     */
    private static void setSocketLingerOff(Socket s) {
        if(s == null) return;
        try {
            s.setSoLinger(false,0);
        } catch(IOException e) {
            System.err.println("\nOn setting socket linger on -- "
                               +e.toString());
        }
    }//end setSocketLingerOff

    /**
     * Closes the given socket.
     * 
     * @param s the <code>Socket</code> to close
     */
    private static void closeSocket(Socket s) {
        if(s == null) return;
        try {
            s.close();
        } catch(IOException e) {
            System.err.println("\nOn closing socket -- "+e.toString());
        }
    }//end closeSocket

    /**
     * Prints to standard err, configuration information of the given socket.
     * 
     * @param s the <code>Socket</code> whose information will be displayed
     */
    private static void displaySocketInfo(Socket s) {
        if(s == null) return;
        System.err.println("Socket Information -- ");
        System.err.println("  Destination:    "
                             +"host/address - "+s.getInetAddress()
                             +", port - "+
                             +s.getPort());
        System.err.println("  Source:         "
                             +"host/address - "+s.getLocalAddress()
                             +", port - "+
                             +s.getLocalPort());
        try {
            System.err.println("  Socket Timeout: "+s.getSoTimeout());
            int linger = s.getSoLinger();
            if(linger < 0) {
                System.err.println("  Socket Linger:  disabled");
            } else {
                System.err.println("  Socket Linger:  "+linger);
            }
            boolean usingNagle = s.getTcpNoDelay();
            if(usingNagle == true) {
                System.err.println("  Socket is using Nagle's algorithm");
            } else {
                System.err.println("  Socket is NOT using Nagle's algorithm");
            }
        } catch(SocketException e) { }
        System.err.println("\n");
    }//end displaySocketInfo

    /**
     * Simple utility method used to impose a delay, blocking processing for
     * the requested number of milliseconds.
     * 
     * @param t the number of milliseconds to delay processing
     */
    private static void delay(int t) {
        if(t <= 0) return;
        try {Thread.sleep(t);} catch (InterruptedException e) { }
    }//end delay

}//end class HTTPDStatus
