/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.start;

import com.sun.jini.admin.DestroyAdmin;
import com.sun.jini.start.StartUtil;
import com.sun.jini.start.ClassLoaderUtil;

import net.jini.admin.Administrable;
import net.jini.admin.JoinAdmin;

import net.jini.lookup.DiscoveryAdmin;

import net.jini.core.discovery.LookupLocator;
import net.jini.core.lookup.ServiceRegistrar;

import java.io.File;
import java.io.IOException;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;

import java.rmi.activation.Activatable;
import java.rmi.activation.ActivationDesc;
import java.rmi.activation.ActivationException;
import java.rmi.activation.ActivationGroup;
import java.rmi.activation.ActivationGroupDesc;
import java.rmi.activation.ActivationGroupDesc.CommandEnvironment;
import java.rmi.activation.ActivationGroupID;
import java.rmi.MarshalledObject;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.RMISecurityManager;
import java.rmi.server.RemoteObject;

import java.util.Properties;
import java.util.StringTokenizer;

/** 
 * This class provides static methods that can be used to start an
 * activatable service.
 */
public class ServiceStarter {

    /** Data structure used by some of the <code>create</code> methods to
     *  return multiple values (such as the service reference and the
     *  <code>ActivationGroupID</code> of the service).
     */
    public static class Created {
        /** The activation group id of the service */
        public final ActivationGroupID gid;
        /** The reference to the proxy of the created service */
	public final Object proxy;
        /** Constructs an instance of this class.
         * @param gid   activation group id of the created service
         * @param proxy reference to the proxy of the created service
         */
        public Created(ActivationGroupID gid, Object proxy) {
            this.gid   = gid;
            this.proxy = proxy;
        }
    }

    /**
     * This method creates an activatable instance of a service's
     * implementation by parsing a set of arguments retrieved from the 
     * command line, and then invoking a static creation method on a 
     * "service starting class" that is supplied by the entity that
     * invokes this method. The class supplied to this method must
     * know how to instantiate the proxy that provides client-side
     * access to the service instance.
     * <p>
     * This method provides a mechanism where a service may be started
     * through a non-generic, custom-defined, class that knows the
     * specific start requirements of the desired service. The 
     * arguments input (through the <code>args</code> parameter) 
     * must satisfy the argument list format for starting activatable
     * services as specified by the <code>StartUtil</code> class.
     * <p>
     * Note that when the service is destroyed (through the
     * <code>com.sun.jini.admin.DestroyAdmin</code> interface provided
     * by the service's implementation), it will also destroy the 
     * activation group to which the service belongs, regardless of
     * whether or not there are other activatable objects still registered
     * in that activation group. This is due to the following reason:
     * <p>
     * When the <code> ActivationDesc</code> used to register the desired
     * service is ultimately created (through the invocation of this method),
     * the <code>location</code> parameter to that class' constructor is
     * <i>not</i> set. If the value of that parameter were to be set, then
     * both the import codebase (the classpath) and the export codebase of
     * the service to be registered would be identical. This means that the
     * activation system would use that single location to both retrieve
     * classes when activating the server-side of the service, and to
     * annotate (with the RMI codebase) all remote classes that are
     * downloaded to the client-side.
     * <p>
     * Because it is desirable to allow for different values in the import
     * and export codebases, this class wishes to distinguish those 
     * parameters for the activation system. To do this, this class must
     * communicate each codebase's value to the activation system through
     * a different mechanism than the <code>location</code> parameter of
     * the constructor of <code>ActivationDesc</code>. That alternate
     * mechanism consists of creating an activation group descriptor
     * (<code>ActivationGroupDesc</code>) in which the properties are
     * over-ridden to include the desired import codebase as the new
     * classpath property ("-cp <import codebase>"); and the export 
     * codebase as the new <code>java.rmi.server.codebase</code> property.
     * The activation descriptor (<code>ActivationDesc</code>) to be used
     * in the service's registration with the activation system is then
     * created using the same group ID (<code>ActivationGroupID</code>)
     * as that generated by the creation of the activation group descriptor
     * with the desired import and export codebases.
     * <p>
     * Because the activation group that is created and registered with the
     * activation system is useful only to entities that have the same
     * import and export codebases as those registered with the group,
     * (and because such entities wishing to register with that activation
     * group would also have to obtain the activation group ID from the
     * service registered by this class), it is not expected that when
     * the service is destroyed through the <code>DestroyAdmin</code> that
     * any other entities will be registered in that activation group. Being
     * the only entity registered in the activation group means that when
     * it is destroyed, it is also the last entity registered in the group.
     * Thus, upon destroying the service, the activation group created here
     * is also ultimately unregistered (removed) from the activation system
     * because once that service is destroyed, there is no way to access
     * the activation group ID for future registrations (or even for 
     * destroying).
     *
     * @param args             <code>String</code> array containing the
     *                         arguments (typically retrieved from the command
     *                         line) to use when starting the service. The
     *                         arguments contained in this parameter must
     *                         satisfy the argument list format specified by
     *                         the <code>StartUtil</code> class. 
     * @param starterClassName the fully-qualified path.classname of the
     *                         class that defines the public static
     *                         <code>create</code> method (and possibly an
     *                         optional public static <code>createGroup</code>
     *                         method that this method invokes through
     *                         reflection in order to start a desired service
     * @param implClassName    the fully-qualified path.classname of the
     *                         class that defines the implementation of the
     *                         backend server of the service that is to be
     *                         started by this method
     * @param resourcename     the name used to identify a particular 
     *                         resource bundle from which usage/error
     *                         message information may be retrieved for
     *                         display (helpful with internationalization).
     *                         Two common values for this parameter are:
     *                         "lookup" and "service".
     * 
     * @return an instance of the service's interface which is a proxy that
     *         provides client-side access to the created service
     *
     * @see com.sun.jini.start.StartUtil
     */
    public static Object create(String[] args,
                                String   starterClassName,
                                String   implClassName,
                                String   resourcename) 
    {
        /* Parse the input args */
        String resource = ((resourcename == null) ? "service" : resourcename);
        StartUtil startUtil = new StartUtil(resource);
        ParsedArgs parsedArgs = startUtil.parseActivatableArgs(args);
        if(parsedArgs == null) {
            startUtil.printUsage();
            return null;
        }
        return create(parsedArgs,starterClassName,implClassName,resourcename);
    }//end create

    /**
     * This method creates an activatable instance of a service's
     * implementation by invoking a static creation method on a 
     * "service starting class" that is supplied by the entity that
     * invokes this method. The class supplied to this method must
     * know how to instantiate the proxy that provides client-side
     * access to the service instance. 
     * <p>
     * This version of the <code>create</code> method is intended as a
     * programmatic API. Unlike the version of <code>create</code> that
     * takes a <code>String</code> array containing items entered on the
     * command line, this version expects that such arguments have
     * already been parsed and will be passed in through an instance of
     * the data structure class <code>ParsedArgs</code>.
     * <p>
     * Note that, as with the other version the <code>create</code>
     * method, when the service is destroyed, it will also destroy the 
     * activation group to which the service belongs, regardless of
     * whether or not there are other activatable objects still registered
     * in that activation group.
     *
     * @param parsedArgs       instance of <code>ParsedArgs</code> which is a
     *                         data structure containing the arguments to use
     *                         when starting the service. 
     * @param starterClassName the fully-qualified path.classname of the
     *                         class that defines the public static
     *                         <code>create</code> method (and possibly an
     *                         optional public static <code>createGroup</code>
     *                         method that this method invokes through
     *                         reflection in order to start a desired service
     * @param implClassName    the fully-qualified path.classname of the
     *                         class that defines the implementation of the
     *                         backend server of the service that is to be
     *                         started by this method
     * @param resourcename     the name used to identify a particular 
     *                         resource bundle from which usage/error
     *                         message information may be retrieved for
     *                         display (helpful with internationalization).
     *                         Two common values for this parameter are:
     *                         "lookup" and "service".
     * 
     * @return an instance of the service's interface which is a proxy that
     *         provides client-side access to the created service
     *
     * @see com.sun.jini.start.ParsedArgs
     * @see com.sun.jini.start.StartUtil
     */
    public static Object create(ParsedArgs parsedArgs,
                                String     starterClassName,
                                String     implClassName,
                                String     resourcename) 
    {
        String resource = ((resourcename == null) ? "service" : resourcename);
        String classpath = System.getProperty("java.class.path");
        try {
            return (doCreate(parsedArgs,starterClassName,implClassName,
                             classpath,new StartUtil(resource))).proxy;
        } catch (Exception e) {
            throw (RuntimeException)e;
        }
    }//end create

    /**
     * This method creates an activatable instance of a service's
     * implementation by invoking a static creation method on a 
     * "service starting class" that is supplied by the entity that
     * invokes this method. The class supplied to this method must
     * know how to instantiate the proxy that provides client-side
     * access to the service instance. 
     * <p>
     * Through the data structure returned by this version of the
     * <code>create</code> method, entities are given access not only
     * to a proxy to the created service, but to the service's
     * <code>ActivationGroupID</code> as well.
     *
     * @param codebase         the RMI codebase with which to annotate all
     *                         remote classes that are exported by the 
     *                         service to be started
     * @param policyFile       the path and name of the policy file that
     *                         applies to the JVM in which the service to
     *                         be started will run
     * @param logDir           the directory in which the state of the
     *                         service to be started will be recorded for
     *                         persistence
     * @param groups           if the service to be started is a lookup
     *                         service, this parameter contains the names
     *                         of the groups to which that lookup service
     *                         is to be a member; otherwise, this parameter
     *                         contains the names of the groups whose members
     *                         are the lookup services the service to be
     *                         started should discover and join
     * @param locators         if the service to be started is a lookup
     *                         service, this parameter typically will not be
     *                         when starting that lookup service; otherwise,
     *                         this parameter contains references to the
     *                         specific lookup services the service to be
     *                         started should discover and join
     * @param javaProgram      specifies which JVM implementation should be
     *                         used for the backend server of the service to
     *                         be started. When not <code>null</code>, the
     *                         value of this parameter is an absolute path
     *                         to an executable that accepts all the standard
     *                         JVM command-line options (for example, -D,
     *                         -classpath, and so on), and which directly or
     *                         indirectly starts a JVM. The value contained in
     *                         this parameter is useful for testing the 
     *                         service to be started on alternate JVM
     *                         implementations
     * @param properties       specifies the system properties with which to
     *                         start the desired service
     * @param options          specifies the command options (for example -cp
     *                         <classpath>) to use when starting the desired
     *                         service
     * @param starterClassName the fully-qualified path.classname of the
     *                         class that defines the public static
     *                         <code>create</code> method (and possibly an
     *                         optional public static <code>createGroup</code>
     *                         method that this method invokes through
     *                         reflection in order to start a desired service
     * @param implClassName    the fully-qualified path.classname of the
     *                         class that defines the implementation of the
     *                         backend server of the service that is to be
     *                         started by this method
     * @param classpath        the classpath from which to load the various
     *                         classes needed to start the service (this may
     *                         be different from the classpath from which the
     *                         the current class was loaded)
     * @param resourcename     the name used to identify a particular 
     *                         resource bundle from which usage/error
     *                         message information may be retrieved for
     *                         display (helpful with internationalization).
     *                         Two common values for this parameter are:
     *                         "lookup" and "service".
     * 
     * @return an instance of the static inner class <code>Created</code>,
     *         which contains the service's <code>ActivationGroupID</code>
     *         and an instance of the service's interface which is a proxy
     *         that provides client-side access to the created service
     *
     * @throws java.rmi.activation.ActivationException this is a general
     *         exception thrown by the activation system to indicate a
     *         a number of exceptional conditions that prevent the
     *         activation system from fulfilling a particular request.
     * 
     * @throws java.io.IOException 
     * 
     * @see com.sun.jini.start.ParsedArgs
     */
    public static Created create(String          codebase,
                                 String          policyFile,
                                 String          logDir,
                                 String[]        groups,
                                 LookupLocator[] locators,
                                 String          javaProgram,
                                 Properties      properties,
                                 String[]        options,
                                 String          starterClassName,
                                 String          implClassName,
                                 String          classpath,
                                 String          resourcename) 
                                       throws ActivationException, IOException
    {
        String resource = ((resourcename == null) ? "service" : resourcename);
        ParsedArgs parsedArgs = new ParsedArgs(codebase,policyFile,logDir,
                                               groups,locators,javaProgram,
                                               properties,options);
        return doCreate(parsedArgs,starterClassName,implClassName,
                         classpath, new StartUtil(resource));
    }//end create

    /**
     * This method creates an activatable instance of a service's
     * implementation by invoking a static creation method on a 
     * "service starting class" that is supplied by the entity that
     * invokes this method. The class supplied to this method must
     * know how to instantiate the proxy that provides client-side
     * access to the service instance. 
     * <p>
     * Through the data structure returned by this version of the
     * <code>create</code> method, entities are given access not only
     * to a proxy to the created service, but to the service's
     * <code>ActivationGroupID</code> as well. 
     * <p>
     * This version of the <code>create</code> method, provided as a
     * convenience, is intended to be used in situations where both the
     * service's proxy and its <code>ActivationGroupID</code> are desired,
     * but where there is no need (or no desire) to start the service using
     * an alternate JVM implementation, or to employ properties and options
     * that differ from the default values.
     * 
     *
     * @param codebase         the RMI codebase with which to annotate all
     *                         remote classes that are exported by the 
     *                         service to be started
     * @param policyFile       the path and name of the policy file that
     *                         applies to the JVM in which the service to
     *                         be started will run
     * @param logDir           the directory in which the state of the
     *                         service to be started will be recorded for
     *                         persistence
     * @param groups           if the service to be started is a lookup
     *                         service, this parameter contains the names
     *                         of the groups to which that lookup service
     *                         is to be a member; otherwise, this parameter
     *                         contains the names of the groups whose members
     *                         are the lookup services the service to be
     *                         started should discover and join
     * @param locators         if the service to be started is a lookup
     *                         service, this parameter typically will not be
     *                         when starting that lookup service; otherwise,
     *                         this parameter contains references to the
     *                         specific lookup services the service to be
     *                         started should discover and join
     * @param starterClassName the fully-qualified path.classname of the
     *                         class that defines the public static
     *                         <code>create</code> method (and possibly an
     *                         optional public static <code>createGroup</code>
     *                         method that this method invokes through
     *                         reflection in order to start a desired service
     * @param implClassName    the fully-qualified path.classname of the
     *                         class that defines the implementation of the
     *                         backend server of the service that is to be
     *                         started by this method
     * @param classpath        the classpath from which to load the various
     *                         classes needed to start the service (this may
     *                         be different from the classpath from which the
     *                         the current class was loaded)
     * @param resourcename     the name used to identify a particular 
     *                         resource bundle from which usage/error
     *                         message information may be retrieved for
     *                         display (helpful with internationalization).
     *                         Two common values for this parameter are:
     *                         "lookup" and "service".
     * 
     * @return an instance of the static inner class <code>Created</code>,
     *         which contains the service's <code>ActivationGroupID</code>
     *         and an instance of the service's interface which is a proxy
     *         that provides client-side access to the created service
     *
     * @throws java.rmi.activation.ActivationException this is a general
     *         exception thrown by the activation system to indicate a
     *         a number of exceptional conditions that prevent the
     *         activation system from fulfilling a particular request.
     * 
     * @throws java.io.IOException 
     * 
     * @see com.sun.jini.start.ParsedArgs
     */
    public static Created create(String          codebase,
                                 String          policyFile,
                                 String          logDir,
                                 String[]        groups,
                                 LookupLocator[] locators,
                                 String          starterClassName,
                                 String          implClassName,
                                 String          classpath,
                                 String          resourcename) 
                                       throws ActivationException, IOException
    {
        return create(codebase,policyFile,logDir,groups,locators,
                      null,null,null,
                      starterClassName,implClassName,classpath,resourcename);
    }//end create

    /**
     * This method creates a non-activatable and non-persistent (transient)
     * instance of a service's implementation by parsing a set of arguments
     * retrieved from the command line, and then invoking a static creation
     * method on a "transient service starting class" that is supplied by
     * the entity that invokes this method. The class supplied to this
     * method must know how to instantiate the proxy that provides
     * client-side access to service instance.
     * <p>
     * This method provides a mechanism where a service may be started
     * through a non-generic,custom-defined, class that knows the
     * specific start requirements of the desired transient service. The 
     * arguments input (through the <code>args</code> parameter) 
     * must satisfy the argument list format for starting transient
     * services as specified by the <code>StartUtil</code> class.
     * <p>
     * Although "transient" means that the service is not persisted for
     * recovery from system outages and crashes, the transient version of
     * services started by this method typically records its state to files 
     * located in the directory entered on the command line. This feature
     * is particularly convenient for debugging.
     *
     * @param args             <code>String</code> array containing the
     *                         arguments (typically retrieved from the command
     *                         line) to use when starting the service. The
     *                         arguments contained in this parameter must
     *                         satisfy the argument list format specified by
     *                         the <code>StartUtil</code> class. 
     * @param starterClassName the fully-qualified path.classname of the
     *                         class that defines a public static
     *                         <code>create</code> method that this method
     *                         invokes through reflection in order to start
     *                         the transient version of the desired service
     * @param classpath        the classpath from which to load the various
     *                         classes needed to start the service (this may
     *                         be different from the classpath from which the
     *                         the current class was loaded)
     * @param codebase         the RMI codebase with which to annotate all
     *                         remote classes that are exported by the 
     *                         service to be started
     * @param resourcename     the name used to identify a particular 
     *                         resource bundle from which usage/error
     *                         message information may be retrieved for
     *                         display (helpful with internationalization)
     * 
     * @return an instance of the service's interface which is a proxy that
     *         provides client-side access to the created service
     * 
     * @throws java.io.IOException 
     */
    public static Object createTransient(String[] args,
                                         String   starterClassName,
                                         String   classpath,
                                         String   codebase,
                                         String   resourcename) 
    {
        /* Parse the input args */
        StartUtil startUtil = new StartUtil(resourcename);
        ParsedArgs parsedArgs = startUtil.parseTransientArgs(args);
        if(parsedArgs == null) {
            startUtil.printTransientUsage();
            return null;
        }
        try {
            return doCreateTransient(parsedArgs,starterClassName,
                                     classpath,codebase,startUtil);
	} catch (Exception e) {
            throw (RuntimeException)e;
	}
    }//end createTransient

    /**
     * Administratively destroys the service referenced by the input
     * parameter. The service input to this method must implement
     * both <code>net.jini.admin.Administrable</code> and the
     * <code>com.sun.jini.admin.DestroyAdmin</code> interfaces
     * in order for this method to successfully destroy the service.
     *
     * @param service reference to the service to destroy
     *
     * @return <code>true</code> if the service's destroy method was invoked
     *         successfully; <code>false</code> otherwise.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         service's backend. When this exception does occur, the
     *         service may or may not have been successfully destroyed.
     * 
     * @see net.jini.admin.Administrable
     * @see net.jini.admin.Administrable#getAdmin
     * @see com.sun.jini.admin.DestroyAdmin
     * @see com.sun.jini.admin.DestroyAdmin#destroy
     */
    public static boolean destroy(Object service) throws RemoteException {
        /* First, test that the service implements both of the appropriate
         * administration interfaces
         */
        DestroyAdmin destroyAdmin = null;
        if( !(service instanceof Administrable) )  return false;
        Object admin = ((Administrable)service).getAdmin();
        if( !(admin instanceof DestroyAdmin) )     return false;
        destroyAdmin = (DestroyAdmin)admin;
        /* Destroy the service */
        destroyAdmin.destroy();
        return true;
    }//end destroy

    /**
     * Tests for the existence of the ActivationSystem (RMID). This method
     * will always make at least one attempt to verify the existence of a
     * a running ActivationSystem. 
     *
     * @param n <code>int</code> value representing the number of additional
     *          attempts (beyond the initial attempt) to wait for the
     *          activation system to come up. This values translates 
     *          directly to the number of seconds to wait for the
     *          activation system.
     *
     * @return <code>true</code> if the activation system is up and running;
     *         <code>false</code> otherwise
     */
    public static boolean activationUp(int n) {
        /* First attempt */
        try {
            ActivationGroup.getSystem();
            return true;
        } catch (ActivationException e) { }
        /* Make a new attempt every second for n seconds */
        for(int i=0; i<n; i++) {
            try {
                ActivationGroup.getSystem();
                return true;
            } catch (ActivationException e) { }
            try {
                Thread.sleep(1000); // wait 1 second
            } catch (InterruptedException e) { }
        }
        return false;
    }//end activationUp

    /**
     * Administratively replaces the current set of groups in which the lookup
     * service (referenced by the <code>proxy</code> parameter) is a member
     * with a new set of member groups represented by the group names 
     * contained in the <code>groups</code> parameter.
     *
     * @param proxy   instance of the proxy to the lookup service whose
     *                current member groups are to be replaced
     * @param groups  String array whose elements are the names of the 
     *                member groups with which to replace the current
     *                member groups of the lookup service. If 
     *                <code>null</code> is input, no attempt will be
     *                made to replace the lookup service's current
     *                set of member groups. 
     *
     * @return <code>true</code> if the lookup service's set of member
     *         groups was successfully replaced; <code>false</code> otherwise.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the proxy and the
     *         service's backend. When this exception does occur, the
     *         set of groups to which the lookup service is a member may or
     *         may not have been successfully replaced.
     * 
     * @see net.jini.admin.Administrable
     * @see net.jini.admin.Administrable#getAdmin
     * @see net.jini.lookup.DiscoveryAdmin
     * @see net.jini.lookup.DiscoveryAdmin#setMemberGroups
     */
    public static boolean setLookupMemberGroups(ServiceRegistrar proxy,
                                                String[] groups)
                                                        throws RemoteException
    {
        if( (proxy == null) || (groups == null) ) return false;
        /* First, test that the lookup service implements both of the
         * appropriate administration interfaces
         */
        DiscoveryAdmin discoveryAdmin = null;
        if( !(proxy instanceof Administrable) ) return false;
        Object admin = ((Administrable)proxy).getAdmin();
        if( !(admin instanceof DiscoveryAdmin) ) return false;
        discoveryAdmin = (DiscoveryAdmin)admin;
        /* Set the member groups for the lookup service */
        discoveryAdmin.setMemberGroups(groups);
        return true;
    }//end setLookupMemberGroups

    /**
     * Administratively replaces - with the respective sets represented by
     * the <code>groups</code> and <code>locators</code> parameters - the
     * set of groups and locators the service referenced by the
     * <code>proxy</code> parameter is currently configured to join.
     *
     * @param proxy    instance of the proxy to the service whose current
     *                 sets of groups and locators to join are to be replaced
     * @param groups   <code>String</code> array whose elements are the
     *                 names of the groups with which to replace the
     *                 service's current set of groups to join
     * @param locators array containing instances of the class
     *                 <code>net.jini.core.discovery.LookupLocator</code>,
     *                 each element corresponding to a specific lookup
     *                 service the service should join, that will replace
     *                 the set of locators with which the service is
     *                 currently configured. If <code>null</code> is
     *                 input, no attempt will be made to replace the
     *                 set of locators.
     *
     * @return <code>true</code> if this method is successful in replacing
     *         the appropriate join sets; <code>false</code> otherwise.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the proxy and the
     *         service's backend. When this exception does occur, the
     *         set of groups the service should join may or may not have
     *         been successfully replaced.
     * 
     * @see net.jini.core.discovery.LookupLocator
     * @see net.jini.admin.Administrable
     * @see net.jini.admin.Administrable#getAdmin
     * @see net.jini.admin.JoinAdmin
     * @see net.jini.lookup.DiscoveryAdmin#setLookupGroups
     */
    public static boolean setGroupsAndLocators(Object proxy,
                                               String[] groups,
                                               LookupLocator[] locators)
                                                        throws RemoteException
    {
        if(proxy == null) return false;
        /* First, test that the service implements both of the appropriate
         * administration interfaces
         */
        JoinAdmin joinAdmin = null;
        if( !(proxy instanceof Administrable) ) return false;
        Object admin = ((Administrable)proxy).getAdmin();
        if( !(admin instanceof JoinAdmin) ) return false;
        joinAdmin = (JoinAdmin)admin;
        /* Set the groups and locators to join */
        joinAdmin.setLookupGroups(groups);
        if( (locators != null) && (locators.length > 0) ) {
            joinAdmin.setLookupLocators(locators);
        }
        return true;
    }//end setGroupsAndLocators

    /**
     * Creates an activation group suitable for a service to be registered
     * with the activation system.
     *
     * @param classpath  the classpath for the service to be registered with
     *                   the activation system
     * @param codebase   the codebase for the service to be registered with
     *                   the activation system
     * @param policyFile the security policy file to apply to the service
     *                   to be registered with the activation system
     * @param properties if not <code>null</code>, this argument contains
     *                   the properties for the activation group in which
     *                   the service will be registered
     * @param serverJVM  if not <code>null</code>, the contents of this
     *                   argument represents a path and command that allows
     *                   parties that invoke this method to specify the
     *                   JVM implementation to use for the execution of the
     *                   service's backend server. The value of this argument
     *                   must be an absolute path to an executable entity and
     *                   that executable entity must accept all of the
     *                   standard JVM command-line options (for example,
     *                   '-D<property>, -classpath <classpath>', and so on),
     *                   and directly or indirectly start a JVM. Note that
     *                   although the use of this argument is not necessary
     *                   in most situations, it is useful for testing the
     *                   service on alternate JVM implementations. 
     * @param options    if not <code>null</code>, this argument contains
     *                   the command options for the JVM in which the
     *                   service's backend server executes
     * 
     * @throws java.rmi.activation.ActivationException this is a general
     *         exception thrown by the activation system to indicate a
     *         a number of exceptional conditions that prevent the
     *         activation system from fulfilling a particular request.
     * 
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the activation 
     *         system's VM and the VM in which this class is instantiated.
     *         When this exception does occur, the ActivationGroup may
     *         or may not have been created successfully.
     */
    private static ActivationGroupID createGroup(String     classpath,
                                                 String     codebase,
                                                 String     policyFile,
                                                 Properties properties,
                                                 String     serverJVM,
                                                 String[]   options)
                                    throws ActivationException, RemoteException
    {
        if (properties == null) {
            properties = new Properties();
        } else {
            properties = (Properties)properties.clone();
        }
        properties.put("java.security.policy", policyFile);
        properties.put("java.rmi.server.codebase", codebase);
        if (options == null) {
            options = new String[]{"-cp", classpath};
        } else {
            String[] nOptions = new String[options.length + 2];
            System.arraycopy(options, 0, nOptions, 0, options.length);
            nOptions[options.length] = "-cp";
            nOptions[options.length + 1] = classpath;
            options = nOptions;
        }
        CommandEnvironment cmdToExecute
                                  = new CommandEnvironment(serverJVM,options);
        return ActivationGroup.getSystem().registerGroup
                           (new ActivationGroupDesc(properties,cmdToExecute));
    }//end createGroup

    /**
     * Private method which does most of the work involved with starting an
     * activatable service as advertised by the public <code>create</code>
     * methods of this class.
     * <p>
     * Before attempting to register the desired service with the activation
     * system, this method analyzes the input arguments for illegal or invalid
     * values, determines if the activation system is running, and determines
     * if a class server that can serve the appropriate classes is running.
     * This method then creates an <code>ActivationGroupID</code>, invokes
     * the <code>create</code> method defined by the class referenced by
     * the <code>starterClassName</code> parameter, and then sets the
     * appropriate group and/or locator sets for the service.
     *
     * @param parsedArgs       instance of <code>ParsedArgs</code> which is a
     *                         data structure containing the arguments to use
     *                         when starting the service. 
     * @param starterClassName the fully-qualified path.classname of the
     *                         class that defines the public static
     *                         <code>create</code> method (and possibly an
     *                         optional public static <code>createGroup</code>
     *                         method that this method invokes through
     *                         reflection in order to start a desired service
     * @param implClassName    the fully-qualified path.classname of the
     *                         class that defines the implementation of the
     *                         backend server of the service that is to be
     *                         started by this method
     * @param classpath        the classpath from which to load the various
     *                         classes needed to start the service (this may
     *                         be different from the classpath from which the
     *                         the current class was loaded)
     * @param startUtil        an instance of the <code>StartUtil</code> class,
     *                         configured with the appropriate resource bundle
     *                         information, which is used as a mechanism
     *                         through which usage/error messages can be
     *                         displayed
     *
     * @return an instance of the service's interface which is a proxy that
     *         provides client-side access to the created service
     *
     * @throws java.rmi.activation.ActivationException this is a general
     *         exception thrown by the activation system to indicate a
     *         a number of exceptional conditions that prevent the
     *         activation system from fulfilling a particular request.
     * 
     * @throws java.io.IOException 
     * 
     * @see com.sun.jini.start.ParsedArgs
     * @see com.sun.jini.start.StartUtil
     */
    private static Created doCreate(ParsedArgs parsedArgs,
                                    String     starterClassName,
                                    String     implClassName,
                                    String     classpath,
                                    StartUtil  startUtil)
                                       throws ActivationException, IOException
    {
        /* Do not proceed unless the required parameters are set */
        String codebase = parsedArgs.getCodebase();
        if(codebase == null) {
            startUtil.print("codebase.null", null);
            return null;
        }//endif
        if(parsedArgs.getPolicyFile() == null) {
            startUtil.print("policy.null", null);
            return null;
        }//endif
        if(parsedArgs.getLogDir() == null) {
            startUtil.print("logdir.null", null);
            return null;
        }//endif
        /* Do not proceed unless the Activation daemon (RMID) is up */
        if (!activationUp(10)) {
            startUtil.print("activation.notUp", null);
            return null;
        }//endif
        /* If there are problems with any of the components of the RMI
         * codebase parameter, display a warning, but allow the process
         * to proceed.
         */
        HTTPDStatus.httpdStatus(codebase,startUtil);
        /* Do not proceed unless the starter class can be loaded from the
         * desired classpath
         */
        URLClassLoader newClassLoader = null;
        Class starterClass = null;
        try {
            /* Note that the class loader that is created here is created
             * with the following characteristics:
             *   -- parent class loader equal to the "extenstion class loader"
             *   -- search path equal to the classpath
             *   -- getURLs() method that returns the codebase (for annotation)
             * Doing so will result in a class loader belonging to a delegation
             * tree that has the following form:
             *             'bootCL--extensionCL--classpathCL'
             * That is, there will be no other class loaders between the
             * extension class loader and the new class loader. This is 
             * done to guarantee that when creating the desired service,
             * the classes needed to create that service are loaded from
             * only the classpath that is input to this method; not from
             * the "system class loader" which may also have access to
             * the necessary classes. Loading the classes from two different
             * class loaders may result in delegation problems (class loader
             * "interference"). Such problems typically manifest themselves
             * through exceptions such as ClassNotFoundException.
             */
             newClassLoader = ClassLoaderUtil.getCodebaseClassLoader
                                            (classpath,codebase,
                                             ClassLoaderUtil.PARENT_EXTENSION);
            starterClass   = newClassLoader.loadClass(starterClassName);
        } catch (Exception e) {
            startUtil.print("starterclass.loadproblem", starterClassName);
	    e.printStackTrace(System.out);
            throw (RuntimeException)e;
        }
        /* In order to invoke the method(s) of the starter class, the
         * ContextClassLoader of the current thread must be set to the
         * new ClassLoader; otherwise an IllegalAccessException will be
         * thrown because the classpath of the old ClassLoader will be
         * searched for the definition of the object on which either
         * invocation is made.
         *
         * After invoking the method(s) of the starter class, the current
         * thread's ContextClassLoader is set back to its original ClassLoader
         */
        Thread curThread = Thread.currentThread();
        ClassLoader oldClassLoader = curThread.getContextClassLoader();
        curThread.setContextClassLoader(newClassLoader);
        ActivationGroupID gid = null;
        Method starterMethod  = null;
        Object proxy          = null;
        try {//invocation block
            /* Create an ActivationGroupID for the desired service. If the
             * starter class provides its own version of the createGroup()
             * method, then invoke that method; otherwise, invoke the version
             * of createGroup() provided in this class
             */
            try {//createGroup() block
                Class[] argTypes = { String.class,      //classpath
                                     String.class,      //codebase
                                     String.class,      //policyFile
                                     Properties.class,  //properties
                                     String.class,      //javaProgram cmdPath
                                     String[].class };  //options (argv)
                starterMethod = starterClass.getMethod("createGroup",argTypes);
                /* A version of the createGroup() method with the appropriate
                 * arguments is provided by the starter class. Invoke that
                 * method with the current values in the parsed arguments
                 */
                Object[] methodArgs = { classpath,
                                        codebase,
                                        parsedArgs.getPolicyFile(),
                                        parsedArgs.getProperties(),
                                        parsedArgs.getJavaProgram(),
                                        parsedArgs.getOptions() };
                /* Since method is static, use null for 1st parameter */
                gid = (ActivationGroupID)starterMethod.invoke(null,methodArgs);
            } catch (NoSuchMethodException e) {
                /* Starter class does not provide a createGroup() method, so
                 * use the one provided in this class
                 */
                gid = createGroup(classpath,
                                  codebase,
                                  parsedArgs.getPolicyFile(),
                                  parsedArgs.getProperties(), 
                                  parsedArgs.getJavaProgram(), 
                                  parsedArgs.getOptions());
            } //end try: createGroup() block
            /* Create the activation descriptor for the desired service */
            MarshalledObject params
                              = new MarshalledObject(parsedArgs.getLogDir());
            ActivationDesc desc = new ActivationDesc(gid, implClassName,
                                                     null, params, true);
            /* Register the desired service with the activation system */
            Remote serverStub = Activatable.register(desc);
            /* Invoke the create() method provided by the starter class */
            if(serverStub != null) {
                Class[] argTypes = {Remote.class};
                starterMethod = starterClass.getMethod("create",argTypes);
                Object[] methodArgs = { serverStub };
                proxy = starterMethod.invoke(null,methodArgs);
            }//endif
            /* Unlike the case above when the proxy was created using a 
             * class loader having the extension class loader as its
             * parent, it is important to return an object that is loaded 
             * from a class loader that has the codebase (input to this
             * method) in its search path, and whose parent class loader
             * is the original "system class loader".
             * Thus, a new instance of the proxy just created must be created
             * with a class loader of the form:
             *
             * 'bootCL--extCL--parentCL0--...--parentCLn--systemCL--codebaseCL'
             *
             * (Note that the "system class loader", sometimes referred to
             * as the "application class loader", is typically the class
             * loader used to start the application; that is, the class 
             * loader that was most likely used to load the class from which
             * this method is invoked. It is also the default delegation
             * parent used when creating new instances of ClassLoader.)
             *
             * Returning the proxy created above, rather than returning a 
             * new proxy created using the class loader just described,
             * can result in class loading problems that may only occur
             * sometime after the service has been started. Such problems
             * can occur when interactions with the service cause class
             * loader "mismatches"; some classes being loaded by a class
             * loader created by the RMI sub-system, some not.
             * 
             * In order return an instance of the desired object with the
             * appropriate class loader, the following is done:
             *
             *   -- the current ContextClassLoader is re-set to the original
             *      "system class loader"
             *   -- the original proxy, created with the classpath class
             *      loader, is marshalled and then un-marshalled; that is,
             *             (new MarshalledObject(proxy).get())
             *
             *      The marshalling process causes the RMI sub-system to
             *      call the getURLs() method of the original proxy's class
             *      loader. That method returns the codebase -- not the
             *      classpath -- since the class loader of the original proxy
             *      is an instacne of CodebaseClassLoader. The RMI sub-system
             *      then uses the output of the getURLs() method to annotate
             *      the serialization stream created by the marshalling
             *      process.
             *
             *      The un-marshalling process causes the RMI sub-system to
             *      create a new URLClassLoader with parent equal to the
             *      current context class loader (the "system class loader"),
             *      and with search path equal to the codebase with which the
             *      serialization stream was annotated (which was the 
             *      codebase of the original proxy's class loader).
             *   -- member groups, or join groups and join locators are set
             *   -- the result of the marshalling/un-marshalling process is
             *
             *      Note that the object (proxy) that is returned is
             *      associated with a class loader that has the codebase
             *      input to this method as its search path; which is as
             *      described above.
             */
            curThread.setContextClassLoader(oldClassLoader);
            proxy = (new MarshalledObject(proxy)).get();
            if(proxy instanceof ServiceRegistrar) {
                /* Set the groups the new lookup service should belong to */
                try {
                    setLookupMemberGroups((ServiceRegistrar)proxy,
                                          parsedArgs.getGroups());
                } catch(Exception e) {
                    startUtil.print(".membergroups.problem", null);
                    e.printStackTrace(System.out);
                    throw (RuntimeException)e;
                }
            } else {
                /* Set the groups and locators the new service should join */
                try {
                    setGroupsAndLocators(proxy,
                                         parsedArgs.getGroups(),
                                         parsedArgs.getLocators());
                } catch(Exception e) {
                    startUtil.print("joingroupslocs.problem", null);
                    e.printStackTrace(System.out);
                    throw (RuntimeException)e;
                }
            }
            return new Created(gid,proxy);
        } catch(Exception e) {
            handleReflectionException(e,starterClassName,"create",startUtil);
            try {
                ActivationGroup.getSystem().unregisterGroup(gid);
            } catch (Exception ee) { }
            if (e instanceof ActivationException) {
                e.printStackTrace(System.out);
                throw (ActivationException)e;
            } else if (e instanceof IOException) {
                e.printStackTrace(System.out);
                throw (IOException)e;
            }
            e.printStackTrace(System.out);
            throw (RuntimeException)e;
        } finally {
            /* Re-set the current thread's ClassLoader */
            curThread.setContextClassLoader(oldClassLoader);
        }//end invocation block
    }//end doCreate

    /**
     * Private method which does most of the work involved with starting
     * the transient version of a service as advertised by the public
     * <code>createTransient</code> method of this class.
     * <p>
     * Before attempting to create the desired service, this method analyzes
     * the input arguments for illegal or invalid values, and determines if
     * a class server that can serve the appropriate classes is running.
     * This method then sets an <code>RMISecurityManager</code>, constructs
     * an instance of the service implementation, invokes the
     * <code>create</code> method defined by the class referenced by
     * the <code>starterClassName</code> parameter, and then sets the
     * appropriate group and/or locator sets for the service.
     *
     * @param parsedArgs       instance of <code>ParsedArgs</code> which is a
     *                         data structure containing the arguments to use
     *                         when starting the service. 
     * @param starterClassName the fully-qualified path.classname of the
     *                         class that defines a public static
     *                         <code>create</code> method that this method
     *                         invokes through reflection in order to start
     *                         the transient version of the desired service
     * @param classpath        the classpath from which to load the various
     *                         classes needed to start the service (this may
     *                         be different from the classpath from which the
     *                         the current class was loaded)
     * @param codebase         the RMI codebase with which to annotate all
     *                         remote classes that are exported by the 
     *                         service to be started
     * @param startUtil        an instance of the <code>StartUtil</code> class,
     *                         configured with the appropriate resource bundle
     *                         information, which is used as a mechanism
     *                         through which usage/error messages can be
     *                         displayed
     *
     * @return an instance of the service's interface which is a proxy that
     *         provides client-side access to the created service
     *
     * @throws java.io.IOException 
     * 
     * @see com.sun.jini.start.ParsedArgs
     * @see com.sun.jini.start.StartUtil
     */
    private static Object doCreateTransient(ParsedArgs parsedArgs,
                                            String     starterClassName,
                                            String     classpath,
                                            String     codebase,
                                            StartUtil  startUtil)
                                                           throws IOException
    {
        /* Do not proceed unless the required parameters are set */
        if(classpath == null) {
            classpath = System.getProperty("java.class.path");
        }
        if(codebase == null) {
            codebase = System.getProperty("java.rmi.server.codebase");
        }
        String policyFile = System.getProperty("java.security.policy");
        String logDir     = parsedArgs.getLogDir();
        if(codebase == null) {
            startUtil.print("codebase.null", null);
            return null;
        }//endif
        if(policyFile == null) {
            startUtil.print("policy.null", null);
            return null;
        }//endif
        if(logDir == null) {
            startUtil.print("logdir.null", null);
            return null;
        }//endif
        /* If there are problems with any of the components of the RMI
         * codebase parameter, display a warning, but allow the process
         * to proceed.
         */
        HTTPDStatus.httpdStatus(codebase,startUtil);
        /* Do not proceed unless the starter class can be loaded from the
         * desired classpath
         */
        URLClassLoader newClassLoader = null;
        Class starterClass = null;
        try {
            /* Create a class loader with the following characteristics:
             *   -- parent class loader equal to the "extenstion class loader"
             *   -- search path equal to the classpath
             *   -- getURLs() method that returns the codebase (for annotation)
             * Doing so will result in a class loader belonging to a delegation
             * tree that has the following form:
             */
            newClassLoader = ClassLoaderUtil.getCodebaseClassLoader
                                            (classpath,codebase,
                                             ClassLoaderUtil.PARENT_EXTENSION);
            starterClass   = newClassLoader.loadClass(starterClassName);
        } catch (Exception e) {
            startUtil.print("starterclass.loadproblem", starterClassName);
	    e.printStackTrace(System.out);
            throw (RuntimeException)e;
        }
        /* An RMISecurityManager must be set */
        System.setSecurityManager(new RMISecurityManager());
        /* In order to invoke the method(s) of the starter class and to
         * construct an instance of the transient version of the desired
         * service, the ContextClassLoader of the current thread must be
         * set to the new ClassLoader; otherwise an IllegalAccessException
         * will be thrown because the classpath of the old ClassLoader will
         * be searched for the definition of the object on which either
         * invocation is made.
         *
         * After invoking the method(s) of the starter class and constructing
         * the transient version of the desired service, the current thread's
         * ContextClassLoader is set back to its original ClassLoader
         */
        Thread curThread = Thread.currentThread();
        ClassLoader oldClassLoader = curThread.getContextClassLoader();
        curThread.setContextClassLoader(newClassLoader);
        Object proxy = null;
        try {//invocation block
            /* Invoke the create() method provided by the starter class */
            try {//starter method invocation
                Class[]  argTypes    = {String.class};
                Object[] argVals     = {logDir};
                Method starterMethod = starterClass.getMethod("create",
                                                              argTypes);
                proxy = starterMethod.invoke(null,argVals);
            } catch(Exception e) {
                handleReflectionException(e,starterClassName,"create",
                                          startUtil);
                e.printStackTrace(System.out);
                throw (RuntimeException)e;
            }//end try: starter method invocation
            /* Create a new instance of the proxy just created using a class
             * loader of the form:
             *
             * 'bootCL--extCL--parentCL0--...--parentCLn--systemCL--codebaseCL'
             *
             * In order return an instance of the desired object with the
             * appropriate class loader, the following is done:
             *
             *   -- the current ContextClassLoader is re-set to the original
             *      "system class loader"
             *   -- the original proxy, created with the classpath class
             *      loader, is marshalled and then un-marshalled; that is,
             *             (new MarshalledObject(proxy).get())
             *   -- member groups, or join groups and join locators are set
             *   -- the result of the marshalling/un-marshalling process is
             */
            curThread.setContextClassLoader(oldClassLoader);
            proxy = ( (new MarshalledObject(proxy)).get() );
            /* Handle groups and locators */
            if(proxy instanceof ServiceRegistrar) {
                /* Set the groups the new lookup service should belong to */
                try {
                    setLookupMemberGroups((ServiceRegistrar)proxy,
                                          parsedArgs.getGroups());
                } catch(Exception e) {
                    startUtil.print(".membergroups.problem", null);
                    e.printStackTrace(System.out);
                }
            } else {
                /* Set the groups and locators the new service should join */
                try {
                    setGroupsAndLocators(proxy,
                                         parsedArgs.getGroups(),
                                         parsedArgs.getLocators());
                } catch(Exception e) {
                    startUtil.print("joingroupslocs.problem", null);
                    e.printStackTrace(System.out);
                }
            }
        } catch(Exception e) {
            handleReflectionException(e,starterClassName,"create",startUtil);
	    e.printStackTrace(System.out);
            throw (RuntimeException)e;
        } finally {
            /* Re-set the current thread's ClassLoader */
            curThread.setContextClassLoader(oldClassLoader);
        }//end invocation block
        return proxy;
    }//end doCreateTransient

    /**
     * Method which allows the <code>create</code> method of this class
     * to group most of the reflection-related exception handling and
     * reporting in a single convenient location. This method displays
     * the appropriate usage/error message based on the exception that
     * is input.
     *
     * @param throwable  instance of the exception to use to determine
     *                   which message to display
     * @param classname  <code>String</code> containing the fully-qualified
     *                   path.classname of the service requested to be started
     *                   by this class
     * @param methodname <code>String</code> containing the name of the
     *                   method in which a failure occurred while attempting
     *                   an invocation through reflection
     * @param startUtil  an instance of the <code>StartUtil</code> class,
     *                   configured with the appropriate resource bundle
     *                   information, which is used as the mechanism through
     *                   which the appropriate message is displayed
     *
     * @see com.sun.start.StartUtil
     */
    private static void handleReflectionException(Throwable throwable,
                                                  String    classname,
                                                  String    methodname,
                                                  StartUtil startUtil)
    {
        startUtil.print("reflectinvoke.problem", classname+"."+methodname);
        if (throwable instanceof IllegalArgumentException) {
            startUtil.print("reflectinvoke.illegalarg",null);
        } else if(throwable instanceof IllegalAccessException) {
            startUtil.print("reflectinvoke.illegalaccess",null);
        } else if(throwable instanceof InvocationTargetException) {
            startUtil.print("reflectinvoke.targetexception",null);
        } else if(throwable instanceof NoSuchMethodException) {
            startUtil.print("reflect.nosuchmethod",null);
        }
    }//end handleReflectionException

}//end class ServiceStarter
