/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package net.jini.discovery;

import net.jini.core.discovery.LookupLocator;
import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.net.DatagramPacket;
import java.net.InetAddress;
import net.jini.core.lookup.ServiceID;

/**
 * Encapsulate the details of unmarshaling an incoming multicast
 * announcement packet.
 *
 * @see OutgoingMulticastAnnouncement
 * @author Bryan O'Sullivan
 */
public class IncomingMulticastAnnouncement {
    /**
     * The ServiceID that has been announced.
     */
    protected ServiceID serviceID;
    /**
     * The LookupLocator that has been announced.
     */
    protected LookupLocator locator;
    /**
     * The groups of which the announcing lookup service is a member.
     */
    protected String[] groups;
    /**
     * Current version of the multicast announcement protocol.
     */
    protected final int protoVersion = 1;

    /**
     * Construct a new object, initialized by unmarshaling the
     * contents of a multicast announcement packet.
     *
     * @param p the packet to unmarshal
     * @exception IOException a problem occurred in unmarshaling the
     * packet
     */
    public IncomingMulticastAnnouncement(DatagramPacket p) throws IOException {
	ByteArrayInputStream bs = new ByteArrayInputStream(p.getData());
	DataInputStream ds = new DataInputStream(bs);
	int proto = ds.readInt();
	if (proto != protoVersion)
	    throw new IOException("unsupported protocol version: " + proto);
	String host = ds.readUTF();
	int port = ds.readInt();
	if (port <= 0 || port >= 65536)
	    throw new IOException("port number out of range: " + port);
	locator = new LookupLocator(host, port);
	serviceID = new ServiceID(ds);
	int groupCount = ds.readInt();
	// We know that the minimal length of a string written with
	// DataOutput.writeUTF is two bytes (for the empty string), so
	// we make some attempt to see if the count can plausibly fit
	// into this packet.
	if (groupCount < 0 || groupCount > bs.available() / 2)
	    throw new IOException("group count invalid: " + groupCount);
	groups = new String[groupCount];
	for (int i = 0; i < groups.length; i++) {
	    groups[i] = ds.readUTF();
	}
    }

    /**
     * Return the ServiceID of the announcing lookup service.
     */
    public ServiceID getServiceID() {
	return serviceID;
    }

    /**
     * Return a LookupLocator for performing unicast discovery of the
     * announcing lookup service.
     */
    public LookupLocator getLocator() {
	return locator;
    }

    /**
     * Return the groups of which the announcing lookup service is a
     * member.
     */
    public String[] getGroups() {
	return groups;
    }

    public int hashCode() {
	return serviceID.hashCode();
    }

    /**
     * Two announcements are equal if they have the same service id.
     */
    public boolean equals(Object o) {
	return o instanceof IncomingMulticastAnnouncement &&
	    ((IncomingMulticastAnnouncement) o).serviceID.equals(serviceID);
    }
}
