/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package net.jini.discovery;

import net.jini.core.discovery.LookupLocator;

import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.net.DatagramPacket;
import java.net.MulticastSocket;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.rmi.RemoteException;
import java.security.AccessControlContext;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.security.PrivilegedActionException;
import java.security.PrivilegedExceptionAction;
import java.util.Collection;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import net.jini.core.lookup.ServiceID;
import net.jini.core.lookup.ServiceRegistrar;

/**
 * This class allows one to find instances of the lookup
 * service that are nearby on the network, using group names.
 * It provides a simple unified front end to the multicast request
 * and announcement protocols.
 *
 * @see net.jini.core.lookup.ServiceRegistrar
 * @see DiscoveryChangeListener
 * @see DiscoveryManagement
 * @see DiscoveryGroupManagement
 * @see DiscoveryListener
 * @see DiscoveryEvent
 * @see DiscoveryPermission
 * @author Bryan O'Sullivan
 */
public class LookupDiscovery implements DiscoveryManagement,
                                        DiscoveryGroupManagement
{
    /**
     * Convenience constant used to request that attempts be made to
     * discover all lookup services that are within range, and which
     * belong to any group. Must define this constant here as well a in
     * <code>DiscoveryGroupManagement</code> for compatibility with
     * earlier releases.
     */
    public static final String[] ALL_GROUPS 
                                       = DiscoveryGroupManagement.ALL_GROUPS;
    /**
     * Convenience constant used to request that discovery by group
     * membership be halted (or not started, if the group discovery
     * mechanism is simply being instantiated). Must define this constant
     * here as well a in <code>DiscoveryGroupManagement</code> for
     * compatibility with earlier releases.
     */
    public static final String[] NO_GROUPS 
                                        = DiscoveryGroupManagement.NO_GROUPS;
    /** Becomes true if the terminate method is called. */
    private boolean terminated = false;
    /** The set of listeners for discovery events. */
    private ArrayList listeners = new ArrayList(1);
    /** The groups to discover. Empty set -- NO_GROUPS, null -- ALL_GROUPS */
    private Set groups = null;
    /** Map from ServiceID to IncomingUnicastResponse. */
    private Map registrars = new HashMap(11);
    /** Unicast discoveries that are currently pending. */
    private Set pendingDiscoveries = new HashSet(11);
    /** Thread that handles pending notifications. */
    private Notifier notifierThread;
    /** Notifications to be sent to listeners. */
    private LinkedList pendingNotifies = new LinkedList();
    /** Thread that handles pending unicast discoveries. */
    private UnicastDiscoverer discovererThread;
    /** Thread that handles incoming multicast announcements. */
    private AnnouncementListener announceeThread;
    /** Collection of Thread objects representing multicast request handles. */
    private Collection requestors = new LinkedList();
    /** Thread that manages incoming multicast responses. Created and 
     *  destroyed on demand.
     */
    private ResponseListener respondeeThread = null;
    /** Access control context to use for listener callbacks. */
    private final AccessControlContext acc = AccessController.getContext();
    /** Map from ServiceID to multicast announcement time stamps */
    private HashMap regTimes = new HashMap(11);
    /** Thread that discards registrars when multicast announcements stop. */
    private AnnouncementTimerThread announcementTimerThread;
    /** Constants used to tell the notifierThread the type of event to send */
    private static final int DISCOVERED = 0;
    private static final int DISCARDED  = 1;
    private static final int CHANGED    = 2;

    /**
     * This thread listens for multicast announcements from lookup
     * services.  If it determines that we haven't yet discovered a
     * given service, it marks that service for unicast discovery
     * later.
     */
    private class AnnouncementListener extends Thread {
	/** Multicast socket for receiving packets */
	private MulticastSocket sock;
	
	/** Create a daemon thread */
	public AnnouncementListener() throws IOException {
	    super("multicast discovery announcement listener");
	    setDaemon(true);
	    sock = new MulticastSocket(Constants.discoveryPort);
	    sock.joinGroup(Constants.getAnnouncementAddress());
	    setInterface(sock);
	}

	/** True if thread has been interrupted */
	private boolean interrupted = false;

	/* This is a workaround for Thread.interrupt not working on
	 * MulticastSocket.receive on all platforms.
	 */
	public void interrupt() {
	    interrupted = true;
	    sock.close();
	}

	public boolean isInterrupted() {
	    return interrupted;
	}

	public void run() {
	    byte[] buf = new byte[512];
	    DatagramPacket pkt = new DatagramPacket(buf, buf.length);

	    while (!isInterrupted()) {
		try {
		    pkt.setLength(buf.length);
		    try {
			sock.receive(pkt);
		    } catch (NullPointerException e) {
			break; // workaround for bug 4190513
		    }
		    IncomingMulticastAnnouncement ann =
			new IncomingMulticastAnnouncement(pkt);
                    /* If the registrars map contains the service ID of the
                     * registrar that sent the current announcement then that
                     * registrar has already been discovered.
                     * 
                     * Determine if the set of groups to which that 
                     * already-discovered registrar belongs have been replaced
                     * with a set containing none of the desired groups; if
                     * yes, then discard the registrar and wait for the next
                     * multicast announcement to arrive.
                     * 
                     * If the registrar that sent the current announcement
                     * has not already been discovered, then check to see if
                     * any of the group(s) in which the registrar is a member
                     * are in the set of desired groups to discover. If yes,
                     * then queue the registrar for unicast discovery; if no,
                     * then wait for the next multicast announcement to arrive.
                     */
                    ServiceID srvcID = ann.getServiceID();
		    synchronized (registrars) {
                        IncomingUnicastResponse resp =
                              (IncomingUnicastResponse)registrars.get(srvcID);
                        if(resp != null) { // already in discovered set
                            /* time stamp the current announcement */
                            regTimes.put
                                 (srvcID,new Long(System.currentTimeMillis()));
                            /* queue a potential discarded or changed event */
                            synchronized (pendingDiscoveries) {
                                if(    (pendingDiscoveries.add(ann))
                                    && (discovererThread == null) )
                                {
                                    discovererThread = new UnicastDiscoverer();
                                    discovererThread.start();
                                }//endif
                            }//end sync
                            continue;
                        } else { // not in discovered set, are we interested?
                            if( !groupsOverlap(ann.getGroups()) ) continue;
                        }//endif
		    }//end sync

		    // Ensure that we're not already trying to
		    // discover this lookup service.
		    synchronized (pendingDiscoveries) {
			if (pendingDiscoveries.add(ann.getLocator()) &&
			    discovererThread == null)
			{
			    discovererThread = new UnicastDiscoverer();
			    discovererThread.start();
			}
		    }
		} catch (InterruptedIOException e) {
		    break;
		} catch (Exception e) {
		    printException(e);
		    // ignore
		}
	    }
	    sock.close();
	    sock = null;
	}
    }

    /**
     * This thread waits for newly-announced lookup services to be
     * added to its queue, and snarfs them via unicast discovery. <p>
     *
     * Why handle unicast discovery asynchronously?  It can take a while
     * for unicast discovery to fail if a lookup service disappears in
     * between making an announcement/connect and discovery starting, and we
     * prefer not to miss multicast announcements or unicast connects.
     */
    private class UnicastDiscoverer extends Thread {
	/** Create a daemon thread */
	public UnicastDiscoverer() {
	    super("unicast discovery");
	    setDaemon(true);
	}

	/** True if thread has been interrupted */
	private boolean interrupted = false;

	/**
	 * We can't normally distinguish between Thread.interrupt() and
	 * timeout from setSoTimeout (see bug 4158021), so we have to
	 * invent our own way to distinguish.
	 */
	public void interrupt() {
	    interrupted = true;
	    super.interrupt();
	}

	public boolean isInterrupted() {
	    return interrupted;
	}

	public void run() {
	    int timeout = 60 * 1000;
	    try {
		timeout = Integer.getInteger("net.jini.discovery.timeout",
					     timeout).intValue();
	    } catch (SecurityException e) {
	    }
	    while (!isInterrupted()) {
		Object req = null;
		try {
		    // If the set of lookups to discover is non-empty,
		    // then find some element from the set, else go away.
		    boolean done = false; // bug 4084783/4187594
		    synchronized (pendingDiscoveries) {
			if (pendingDiscoveries.isEmpty()) {
			    discovererThread = null;
			    done = true; // bug 4084783/4187594
			    return;
			}
			req = pendingDiscoveries.iterator().next();
		    }
		    Socket sock = null;
                    IncomingMulticastAnnouncement announcement = null;
                    IncomingUnicastResponse response = null;
		    if (req instanceof Socket) {
			sock = (Socket)req;
		    } else if(req instanceof LookupLocator) {
			sock = new Socket(((LookupLocator)req).getHost(),
					  ((LookupLocator)req).getPort());
                    } else if(req instanceof IncomingMulticastAnnouncement) {
                        announcement = (IncomingMulticastAnnouncement)req;
                    } else if(req instanceof IncomingUnicastResponse) { 
                        response = (IncomingUnicastResponse)req;
                    }
                    if(sock != null) { // do unicast exchange
                        IncomingUnicastResponse resp;
                        try {
                            sock.setSoTimeout(timeout);
                            OutputStream os = sock.getOutputStream();
                            InputStream is = sock.getInputStream();
                            OutgoingUnicastRequest.marshal(os);
                            resp = new IncomingUnicastResponse(is);
                        } finally {
                            try {
                                sock.close();
                            } catch (IOException e) { /* ignore */ }
                        }
                        maybeAddNewRegistrar(resp);
                    } else if(announcement != null) { // group changes
                        ServiceID srvcID = announcement.getServiceID();
		        synchronized (registrars) {
                            IncomingUnicastResponse resp =
                              (IncomingUnicastResponse)registrars.get(srvcID);
                            if(resp != null) {
                                maybeSendEvent(resp, announcement.getGroups());
                            }
                        }//end sync
                    } else if(response != null) { // reachability
		        synchronized (registrars) {
                            maybeSendEvent(response, null);
                        }//end sync
                    }//endif
		} catch (InterruptedIOException e) {
		    break;
		} catch (Exception e) {
		    printException(e);
		    // ignore
		} finally {
		    if (req != null) {
			// Remove the lookup service from the set
			// of those awaiting discovery, regardless
			// of whether or not we were successful in
			// discovering it.
			synchronized (pendingDiscoveries) {
			    pendingDiscoveries.remove(req);
			}
		    }
		}
	    }
	    synchronized (pendingDiscoveries) {
		discovererThread = null;
		for (Iterator iter = pendingDiscoveries.iterator();
		     iter.hasNext(); )
		{
		    Object req = iter.next();
		    iter.remove();
		    if (req instanceof Socket) {
			try {
			    ((Socket)req).close();
			} catch (IOException e) {
			    // ignore
			}
		    }
		}
	    }
	}
    }

    /**
     * This thread listens for incoming responses to multicast
     * requests, and then queues them for UnicastDiscoverer to process.
     */
    private class ResponseListener extends Thread {
	/** Server socket for accepting connections */
	public ServerSocket serv;
	
	/** Create a daemon thread */
	public ResponseListener() throws IOException {
	    super("multicast discovery response listener");
	    setDaemon(true);
	    serv = new ServerSocket(0);
	}

	/** True if thread has been interrupted */
	private boolean interrupted = false;

	/* This is a workaround for Thread.interrupt not working on
	 * ServerSocket.accept on all platforms.  ServerSocket.close
	 * can't be used as a workaround, because it also doesn't work
	 * on all platforms.
	 */
	public void interrupt() {
	    interrupted = true;
	    try {
		(new Socket(InetAddress.getLocalHost(), getPort())).close();
	    } catch (IOException e) {
	    }
	}

	public boolean isInterrupted() {
	    return interrupted;
	}

	public void run() {
	    while (!isInterrupted()) {
		try {
		    Socket sock = serv.accept();
		    if (isInterrupted()) {
			try {
			    sock.close();
			} catch (IOException e) {
			}
			break;
		    }
		    synchronized (pendingDiscoveries) {
			pendingDiscoveries.add(sock);
			if (discovererThread == null) {
			    discovererThread = new UnicastDiscoverer();
			    discovererThread.start();
			}
		    }
		} catch (InterruptedIOException e) {
		    break;
		} catch (Exception e) {
		    printException(e);
		    // ignore with impunity
		}
	    }
	    try {
		serv.close();
	    } catch (IOException e) {
		printException(e);
		// ignore
	    }
	}

	/** Return the local port of the socket */
	public int getPort() {
	    return serv.getLocalPort();
	}
    }

    /**
     * This thread sends out multicast requests periodically for
     * a limited period of time.
     */
    private class Requestor extends Thread {
	/** Multicast socket for sending packets */
	private MulticastSocket sock;
	/** Unicast response port */
	private int responsePort;
	/** Groups to request */
	private String[] groups;

	/** Create a daemon thread */
	public Requestor(String[] groups, int port) throws IOException {
	    super("multicast discovery request");
	    setDaemon(true);
	    sock = new MulticastSocket(Constants.discoveryPort);
	    int ttl = 15;
	    try {
		ttl = Integer.getInteger("net.jini.discovery.ttl",
					 ttl).intValue();
	    } catch (SecurityException e) {
	    }
	    sock.setTimeToLive(ttl);
	    setInterface(sock);
	    responsePort = port;
	    this.groups = groups == null ? new String[0] : groups;
	}

	/**
	 * It would be better to not sleep at the end, and have the
	 * ResponseListener use setSoTimeout on the server socket and
	 * go away by itself when its been idle for a while with no
	 * requestors, but using setSoTimeout in that way can cause
	 * random hangs on the Solaris(TM) operating system.
	 */
	public void run() {
	    int count; // bug 4084783/4187594
	    try {
		for (count = 7; --count >= 0 && !isInterrupted(); ) {
		    try {
			DatagramPacket[] reqs = 
			    OutgoingMulticastRequest.marshal(responsePort,
							     groups,
							     getServiceIDs());
			for (int i = 0; i < reqs.length; i++) {
			    try {
				sock.send(reqs[i]);
			    } catch (InterruptedIOException e) {
				return;
			    } catch (Exception e) {
				printException(e);
				// relatively safe to ignore
			    }
			}
		    } catch (InterruptedIOException e) {
			break;
		    } catch (Exception e) {
			printException(e);
			// @@@ what to do here?
		    }
		    Thread.sleep(count > 0 ? 5000 : 2 * 60 * 1000);
		}
	    } catch (InterruptedException e) {
		printException(e);
		// die gracefully
	    } finally {
		synchronized (requestors) {
		    requestors.remove(Thread.currentThread());
		    if (respondeeThread != null && requestors.isEmpty()) {
			respondeeThread.interrupt();
			respondeeThread = null;
		    }
		}
		sock.close();
	    }
	}
    }

    /** Simple holder for notify task data */
    private static class NotifyTask {
	/** The listeners to notify */
	public final ArrayList listeners;
	/** Map of discovered registrars to groups in which each is a member */
	public final Map groupsMap;
	/** The type of event to send: DISCOVERED, DISCARDED, CHANGED */
	public final int eventType;
	public NotifyTask(ArrayList listeners, Map groupsMap, int eventType) {
	    this.listeners = listeners;
	    this.groupsMap = groupsMap;
	    this.eventType = eventType;
	}
    }

    /** Task thread for calling listeners with notifications */
    private class Notifier extends Thread {

	/** Create a daemon thread */
	public Notifier() {
	    super("event listener notification");
	    setDaemon(true);
	}

	public void run() {
	    while (true) {
		final NotifyTask task;
		synchronized (pendingNotifies) {
		    if (pendingNotifies.isEmpty()) {
			notifierThread = null;
			return;
		    }
		    task = (NotifyTask)pendingNotifies.removeFirst();
		}
		AccessController.doPrivileged(new PrivilegedAction() {
		    public Object run() {
			for (Iterator iter = task.listeners.iterator();
			     iter.hasNext(); )
			{
			    DiscoveryListener l =
                                             (DiscoveryListener)iter.next();
                            /* always send discovered and discarded events */
                            if(     (task.eventType == CHANGED)
                                && !(l instanceof DiscoveryChangeListener) )
                            {
                                continue;
                            }
			    DiscoveryEvent e =
				new DiscoveryEvent
                                        ( LookupDiscovery.this,
                                          deepCopy((HashMap)task.groupsMap) );
                            switch(task.eventType) {
                                case DISCOVERED:
                                    l.discovered(e);
                                    break;
                                case DISCARDED:
                                    l.discarded(e);
                                    break;
                                case CHANGED:
                                    ((DiscoveryChangeListener)l).changed(e);
                                    break;
                            }
			}
			return null;
		    }
		}, acc);
	    }
	}
    }

    /** Print the exception stack trace if debugging is enabled. */
    private static void printException(Exception e) {
	try {
	    if (System.getProperty("net.jini.discovery.debug") != null) {
		e.printStackTrace();
	    }
	} catch (SecurityException ee) {
	}
    }

    /**
     * Set the network interface to use for multicast, if the default is
     * overridden.
     */
    private static void setInterface(MulticastSocket sock) throws IOException {
	try {
	    String iface = System.getProperty("net.jini.discovery.interface");
	    if (iface != null)
		sock.setInterface(InetAddress.getByName(iface));
	} catch (SecurityException e) {
	}
    }

    /**
     * If the given lookup service is not already in the set we have
     * discovered, we add it here, and notify listeners as
     * appropriate.
     *
     * @param resp the response to add
     */
    private void maybeAddNewRegistrar(IncomingUnicastResponse resp) {
	synchronized (registrars) {
	    if(groupsOverlap(resp.groups) &&
	      !resp.equals(registrars.put(resp.registrar.getServiceID(),resp)))
	    {
                /* Time stamp the service ID. The first time stamp associated
                 * with the current service ID occurs here. All other time
                 * stamps for that service ID will occur when multicast
                 * announcements for that service ID arrive (in the
                 * AnnouncementListener thread).
                 *
                 * Note that if the time stamp for the service ID were
                 * initialized upon the arrival of the first announcement,
                 * rather than here when it is first discovered, the
                 * AnnouncementTimerThread would not be able to detect the
                 * termination of announcements for the case where the 
                 * termination happens to occur between the time the lookup 
                 * is first discovered here, and the time the first
                 * announcement was supposed to have arrived. This can
                 * happen because a multicast request from the client can
                 * cause the lookup to be discovered before the first
                 * announcement arrives.
                 */
                regTimes.put(resp.registrar.getServiceID(),
                             new Long(System.currentTimeMillis()) );
                if(!listeners.isEmpty()) {
		    addNotify((ArrayList)listeners.clone(),
                              mapRegToGroups(resp.registrar,resp.groups),
                              DISCOVERED);
                }
	    }
	}
    }

    /**
     * Add a notification task to the pending queue, and start a
     * notifierThread if one isn't running.
     */
    private void addNotify(ArrayList notifies, Map groupsMap, int eventType) {
	synchronized (pendingNotifies) {
	    pendingNotifies.addLast(new NotifyTask(notifies,
                                                   groupsMap,
                                                   eventType));
	    if (notifierThread == null) {
		AccessController.doPrivileged(new PrivilegedAction() {
		    public Object run() {
			notifierThread = new Notifier();
			notifierThread.start();
			return null;
		    }
		});
	    }
	}
    }

    /**
     * Indicate whether any of the groups named in possibilities
     * matches any of the groups we're interested in discovering.
     *
     * @param possibilities - the set of groups we may be interested
     * (must not be null)
     */
    private boolean groupsOverlap(String[] possibilities) {
	// Match if we're interested in any group, or if we're
	// interested in none and there are no possibilities.
	if (groups == null)
	    return true;
	for (int i = 0; i < possibilities.length; i++) {
	    if (groups.contains(possibilities[i]))
		return true;
	}
	return false;
    }

    /**
     * Construct a new lookup discovery object, set to discover the
     * given set of groups.  The set is represented as an array of
     * strings.  This array may be empty, which is taken as the empty
     * set, and discovery is not performed.  The reference passed in
     * may be null, which is taken as no set, and in which case
     * discovery of all reachable lookup services is performed.
     * Otherwise, the array contains the names of groups to discover.
     * The caller must have DiscoveryPermission for each group (or
     * for all groups, if the array is null).
     *
     * @param groups the set of group names to discover (null for no
     * set, empty for no discovery)
     * @see #NO_GROUPS
     * @see #ALL_GROUPS
     * @see #setGroups
     * @see DiscoveryPermission
     * @exception IOException an error occurred in starting discovery
     */
    public LookupDiscovery(String[] groups) throws IOException {
	checkGroups(groups);
	if (groups != null) {
	    this.groups = new HashSet(groups.length * 2);
	    for (int i = 0; i < groups.length; i++) {
		this.groups.add(groups[i]);
	    }
	}
	try {
	    AccessController.doPrivileged(new PrivilegedExceptionAction() {
		public Object run() throws Exception {
		    announceeThread = new AnnouncementListener();
                    announcementTimerThread = new AnnouncementTimerThread();
		    return null;
		}
	    });
	} catch (PrivilegedActionException e) {
	    throw (IOException)e.getException();
	}
	if (this.groups == null || !this.groups.isEmpty())
	    requestGroups(this.groups);
	announceeThread.start();
	announcementTimerThread.start();
    }

    /**
     * Check that the caller has discovery permission for each group.
     */
    private static void checkGroups(String[] groups) {
	SecurityManager sm = System.getSecurityManager();
	if (sm == null)
	    return;
	if (groups != null) {
	    for (int i = 0; i < groups.length; i++) {
		sm.checkPermission(new DiscoveryPermission(groups[i]));
	    }
	} else {
	    sm.checkPermission(new DiscoveryPermission("*"));
	}
    }

    /**
     * Interrupt all currently-running threads.
     */
    private void nukeThreads() {
	AccessController.doPrivileged(new PrivilegedAction() {
	    public Object run() {
                announcementTimerThread.interrupt();
		synchronized (requestors) {
		    for (Iterator iter = requestors.iterator();
			 iter.hasNext(); )
		    {
			Thread t = (Thread) iter.next();
			t.interrupt();
		    }
		    if (respondeeThread != null)
			respondeeThread.interrupt();
		}
		announceeThread.interrupt();
		synchronized (pendingDiscoveries) {
		    if (discovererThread != null)
			discovererThread.interrupt();
		}
		return null;
	    }
	});
    }
    
    /**
     * Register a listener as interested in receiving DiscoveryEvent
     * notifications.
     *
     * @param l the listener to register
     * @see DiscoveryEvent
     */
    public void addDiscoveryListener(DiscoveryListener l) {
	synchronized (registrars) {
	    if (listeners.indexOf(l) >= 0) return; //already have this listener
	    listeners.add(l);
	    if (registrars.isEmpty()) return;//nothing to send the new listener
            HashMap groupsMap = new HashMap(registrars.size());
	    Iterator iter = registrars.values().iterator();
	    for (int i = 0; iter.hasNext(); i++) {
                IncomingUnicastResponse resp = 
                                         (IncomingUnicastResponse)iter.next();
                groupsMap.put(resp.registrar,resp.groups);
	    }
	    ArrayList list = new ArrayList(1);
	    list.add(l);
	    addNotify(list, groupsMap, DISCOVERED);
	}
    }

    /**
     * Indicate that a listener is no longer interested in receiving
     * DiscoveryEvent notifications.
     *
     * @param l the listener to unregister
     */
    public void removeDiscoveryListener(DiscoveryListener l) {
	synchronized (registrars) {
	    listeners.remove(l);
	}
    }

    /**
     * Returns an array of instances of <code>ServiceRegistrar</code>, each
     * corresponding to a proxy to one of the currently discovered lookup
     * services. For each invocation of this method, a new array is returned.
     *
     * @return array of instances of <code>ServiceRegistrar</code>, each
     *         corresponding to a proxy to one of the currently discovered
     *         lookup services
     *
     * @see net.jini.core.lookup.ServiceRegistrar
     * @see net.jini.DiscoveryManagement#removeDiscoveryListener
     */
    public ServiceRegistrar[] getRegistrars() {
        synchronized (registrars) {
            if (registrars.isEmpty()) {
                return new ServiceRegistrar[0];
            }
            Iterator iter = registrars.values().iterator();
            ServiceRegistrar[] regs = new ServiceRegistrar[registrars.size()];
            for (int i=0;iter.hasNext();i++) {
                regs[i] = ((IncomingUnicastResponse)iter.next()).registrar;
            }
            return regs;
        }
    }

    /**
     * Discard a registrar from the set of registrars already
     * discovered.  This does not prevent that registrar from being
     * rediscovered; it is intended to be used to clear unreachable
     * entries from the set. <p>
     *
     * If the registrar has been discovered using this LookupDiscovery
     * object, each listener registered with this object will have its
     * discarded method called with the given registrar as parameter.
     *
     * @param reg the registrar to discard
     * @see DiscoveryListener#discarded
     */
    public void discard(ServiceRegistrar reg) {
        synchronized (registrars) {
            sendDiscarded(reg,null);
        }//end sync
    }//end discard

    /**
     * Return the set of group names this LookupDiscovery instance is
     * trying to discover.  If this method returns the empty array,
     * that value is guaranteed to be referentially equal to
     * LookupDiscovery.NO_GROUPS.
     *
     * @return the set of groups to be discovered (null for all, empty
     * for no discovery)
     * @see #NO_GROUPS
     * @see #ALL_GROUPS
     */
    public String[] getGroups() {
	synchronized (registrars) {
	    if (groups == null)
		return ALL_GROUPS;
	    if (groups.isEmpty())
		return NO_GROUPS;
	    return collectionToStrings(groups);
	}
    }

    /**
     * Return the set of service IDs we have discovered to date.
     */
    private ServiceID[] getServiceIDs() {
	synchronized (registrars) {
	    return (ServiceID[])
		registrars.keySet().toArray(new ServiceID[registrars.size()]);
	}
    }

    /**
     * Change the set of groups to be discovered to correspond to the
     * given set.  The set is represented as an array of strings.
     * This array may be empty, which is taken as the empty set, and
     * discovery is not performed.  The reference passed in may be
     * null, which is taken as no set, and in which case discovery of
     * all reachable lookup services is performed.  Otherwise, the
     * array contains the names of groups to discover.
     * The caller must have DiscoveryPermission for each group (or
     * for all groups, if the array is null).
     *
     * @param newGroups the new set of groups to discover (null for
     * all, empty array for no discovery)
     * @exception IOException an error occurred when starting
     * multicast discovery
     * @exception IllegalStateException this method was called after
     * the terminate method was called
     * @see #LookupDiscovery
     * @see #ALL_GROUPS
     * @see #NO_GROUPS
     * @see DiscoveryPermission
     */
    public void setGroups(String[] newGroups) throws IOException {
	boolean maybeDiscard = false;
	Set newGrps = null;
	if (newGroups != null) {	
	    newGrps = new HashSet(newGroups.length * 2);
	    for (int i = 0; i < newGroups.length; i++) {
		newGrps.add(newGroups[i]);
	    }
	}
	synchronized (registrars) {
	    if (terminated)
		throw new IllegalStateException("discovery terminated");
	    if (newGroups == null) {
		if (groups != null) {
		    groups = null;
		    requestGroups(null);
		}
		return;
	    }
	    if (groups == null) {
		groups = new HashSet(11);
		maybeDiscard = true;
	    }
	    Set toAdd = new HashSet(newGrps);
	    toAdd.removeAll(groups);
	    // Figure out which groups to get rid of.  We start off
	    // with the full set for which we are already listening,
	    // and eliminate any that are in both the new set and the
	    // current set.
	    Collection toRemove = new HashSet(groups);
	    toRemove.removeAll(newGrps);
	    // Add new groups before we remove any old groups, because
	    // removeGroups will start a new round of multicast requests
	    // if the set of groups becomes empty, and we don't want it
	    // to do so without reason.
	    groups.addAll(toAdd);
	    if (!toRemove.isEmpty())
		maybeDiscard |= removeGroupsInt(collectionToStrings(toRemove));
	    if (!toAdd.isEmpty())
		requestGroups(toAdd);
	}
	if (maybeDiscard)
	    maybeDiscardRegistrars();
    }

    /**
     * Check the registrars we've previously discovered, and discard
     * all that are no longer members of any of our groups.
     */
    private void maybeDiscardRegistrars() {
	synchronized (registrars) {
            HashMap groupsMap = new HashMap(registrars.size());
	    for(Iterator iter=registrars.values().iterator();iter.hasNext(); ){
		IncomingUnicastResponse ent =
                                         (IncomingUnicastResponse)iter.next();
		if(!groupsOverlap(ent.groups)) { // not interested anymore
                    groupsMap.put(ent.registrar,ent.groups);
                    regTimes.remove((ent.registrar).getServiceID());
		    iter.remove(); // remove (srvcID,response) mapping
		}
	    }
            if( !groupsMap.isEmpty() && !listeners.isEmpty() ) {
		addNotify((ArrayList)listeners.clone(), groupsMap, DISCARDED);
	    }
	}
    }

    /**
     * Start multicast requesting a given set of groups.
     */
    private void requestGroups(final Collection req) throws IOException {
	try {
	    AccessController.doPrivileged(new PrivilegedExceptionAction() {
		public Object run() throws Exception {
		    Thread t;
		    synchronized (requestors) {
			if (respondeeThread == null) {
			    respondeeThread = new ResponseListener();
			    respondeeThread.start();
			}
			t = new Requestor(collectionToStrings(req),
					  respondeeThread.getPort());
			requestors.add(t);
		    }
		    t.start();
		    return null;
		}
	    });
	} catch (PrivilegedActionException e) {
	    throw (IOException)e.getException();
	}
    }

    /**
     * Add a set of groups to the set to be discovered.
     * The caller must have DiscoveryPermission for each group.
     *
     * @param newGroups the groups to add
     * @exception IOException the multicast request protocol failed to start
     * @exception IllegalStateException this method was called after
     * the terminate method was called
     * @exception UnsupportedOperationException there is no set of
     * groups to add to
     * @see DiscoveryPermission
     */
    public void addGroups(String[] newGroups) throws IOException {
	checkGroups(newGroups);
	synchronized (registrars) {
	    if (terminated)
		throw new IllegalStateException("discovery terminated");
	    if (groups == null)
		throw new UnsupportedOperationException(
					  "can't add to \"any groups\"");
	    Collection req = new ArrayList(newGroups.length);
	    for (int i = 0; i < newGroups.length; i++) {
		if (groups.add(newGroups[i]))
		    req.add(newGroups[i]);
	    }
	    if (!req.isEmpty())
		requestGroups(req);
	}
    }

    /**
     * Remove a set of groups from the set to be discovered.
     *
     * @param oldGroups groups to remove
     * @exception IllegalStateException this method was called after
     * the terminate method was called
     * @exception UnsupportedOperationException there is no set of
     * groups from which to remove
     */
    public void removeGroups(String[] oldGroups) {
	boolean maybeDiscard;
	synchronized (registrars) {
	    if (terminated)
		throw new IllegalStateException("discovery terminated");
	    if (groups == null)
		throw new UnsupportedOperationException(
					   "can't remove from \"any groups\"");
	    maybeDiscard = removeGroupsInt(oldGroups);
	}
	if (maybeDiscard)
	    maybeDiscardRegistrars();
    }

    /**
     * Remove the specified groups, and return true if any were
     * actually removed.
     */
    private boolean removeGroupsInt(String[] oldGroups) {
	boolean removed = false;
	for (int i = 0; i < oldGroups.length; i++) {
	    removed |= groups.remove(oldGroups[i]);
	}
	return removed;
    }

    /**
     * Terminate the discovery process.
     */
    public void terminate() {
	synchronized (registrars) {
	    if (terminated)
		return;
	    terminated = true;
	}
	nukeThreads();
    }

    public void finalize() {
	terminate();
    }

    /** Convert a collection to an array of strings. */
    private static final String[] collectionToStrings(Collection c) {
	return c == null ? null : (String[]) c.toArray(new String[c.size()]);
    }

    /** After a possible change in the member groups of the 
     *  <code>ServiceRegistrar</code> corresponding to the given 
     *  <code>IncomingUnicastResponse</code> parameter, this method
     *  determines whether or not the registrar's member groups have
     *  changed in such a way that either a changed event or a discarded
     *  event is warranted.
     *  <p>
     *  Note that even if the contents of the new set of groups initially 
     *  indicate that the corresponding registrar is a candidate for
     *  a discarded or a changed event, further analysis must be performed.
     *  This is because there is no guarantee that the new set of member
     *  groups have not been "split" across the multicast announcements
     *  sent by the lookup service; and so there is no guarantee that the
     *  contents of the new group set actually reflect a change that warrants
     *  an event. To guarantee that the new group set accurately reflects
     *  the registrar's member groups, this method makes a remote call to
     *  the registrar to retrieve its actual member groups.
     *  <p>
     *  There is one situation where it is not necessary to query the
     *  registrar for its current member groups. That situation is 
     *  when the set of groups input to the <code>newGroups</code> parameter
     *  is equivalent to NO_GROUPS. If that new group set is equivalent
     *  to NO_GROUPS, it is guaranteed that the registrar's member groups
     *  have not been split across the multicast announcements.
     *
     * @param response  instance of <code>IncomingUnicastResponse</code> 
     *                  corresponding to the registrar whose current and
     *                  previous member groups are to be compared
     * @param newGroups <code>String</code> array containing the new
     *                  member groups of the registrar corresponding to the 
     *                  <code>response</code> parameter (just after a
     *                  possible change)
     */
    private void maybeSendEvent(IncomingUnicastResponse response,
                                String[]                newGroups)
    {
        ServiceRegistrar reg = response.getRegistrar();
        String[] oldGroups   = response.getGroups();
        boolean getActual    = true;
        if(newGroups == null) { // newGroups null means get actual groups now
            newGroups = getActualGroups(reg);
            if(newGroups == null) return; // if null, then it was discarded
            getActual = false;
        }//endif

        if(groupSetsEqual(oldGroups,newGroups)) return;

        String[] actualGroups = newGroups;
        if( getActual && (newGroups.length > 0) ) {
            actualGroups = getActualGroups(reg);
            if(actualGroups == null) return; // null ==> was already discarded
        }//endif
        notifyOnGroupChange(reg, oldGroups, actualGroups);
    }//end maybeSendEvent

    /** After a possible change in the member groups of the given
     *  <code>ServiceRegistrar</code> parameter, this method compares
     *  the registrar's original set of member groups to its new set
     *  of member groups.
     *  <p>
     *  If the criteria shown below is satisfied, either a discarded event
     *  or a changed event will be sent to any registered listeners. The
     *  criteria is based on whether the old and new groups are equal,
     *  and whether one or more elements of the new group set also belong
     *  to the set of groups to discover (the new groups are "still of
     *  interest"). The criteria is as follows:
     *  <p>
     *  if (old groups and new groups)
     *  <p><ul>
     *       <li> (not equal but stillInterested) --> send a changed event
     *       <li> (!stillInterested)              --> send a discarded event
     *    </ul>
     *  <p>
     *
     * @param reg       instance of <code>ServiceRegistrar</code> 
     *                  corresponding to the registrar whose current and
     *                  previous member groups are to be compared; and 
     *                  whose corresponding service ID is used as the key
     *                  into the various data structures that contain
     *                  pertinent information about that registrar
     * @param oldGroups <code>String</code> array containing the member
     *                  groups of the <code>reg</code> parameter prior to
     *                  being changed
     * @param oldGroups <code>String</code> array containing the current
     *                  member groups of the <code>reg</code> parameter
     *                  (just after a possible change)
     */
    private void notifyOnGroupChange(ServiceRegistrar reg,
                                     String[]         oldGroups,
                                     String[]         newGroups)
    {
        boolean equal           = groupSetsEqual(oldGroups,newGroups);
        boolean stillInterested = groupsOverlap(newGroups);
        if(!equal && stillInterested) {
            sendChanged(reg,newGroups);
        } else if(!stillInterested) {
            sendDiscarded(reg,newGroups);
        }//endif
    }//end notifyOnGroupChange

    /** Convenience method that sends a discarded event containing only
     *  one registrar to all registered listeners. This method must be
     *  called from within a block that is synchronized on the registrars
     *  map.
     * 
     *  @param reg       instance of <code>ServiceRegistrar</code> 
     *                   corresponding to the registrar to include in the
     *                   event
     *  @param curGroups <code>String</code> array containing the current
     *                   member groups of the registrar referenced by the 
     *                   <code>reg</code> parameter
     */
    private void sendDiscarded(ServiceRegistrar reg, String[] curGroups) {
        ServiceID srvcID = reg.getServiceID();
        if(curGroups == null) { // discard request is from external source
            IncomingUnicastResponse resp =
                              (IncomingUnicastResponse)registrars.get(srvcID);
            if(resp == null) return;
            curGroups = resp.getGroups();
        }
        if( registrars.remove(srvcID) != null ) { 
            regTimes.remove(srvcID);
            if( !listeners.isEmpty() ) {
                addNotify((ArrayList)listeners.clone(),
                           mapRegToGroups(reg,curGroups), DISCARDED);
            }//endif
        }//endif
    }//end sendDiscarded

    /** Convenience method that sends a changed event containing only
     *  one registrar to all registered listeners that are interested in
     *  such events. This method must be clled from within a block that
     *  is synchronized on the registrars map.
     * 
     *  @param reg       instance of <code>ServiceRegistrar</code> 
     *                   corresponding to the registrar to include in the
     *                   event
     *  @param curGroups <code>String</code> array containing the current
     *                   member groups of the registrar referenced by the 
     *                   <code>reg</code> parameter
     */
    private void sendChanged(ServiceRegistrar reg, String[] curGroups) {
        /* replace old groups with new; prevents repeated changed events */
        IncomingUnicastResponse resp = 
                   (IncomingUnicastResponse)registrars.get(reg.getServiceID());
        resp.groups = (String[])curGroups.clone();
        if( !listeners.isEmpty() ) {
            addNotify((ArrayList)listeners.clone(), 
                       mapRegToGroups(reg,curGroups), CHANGED);
        }//endif
    }//end sendChanged

    /** Creates and returns a deep copy of the input parameter. This method
     *  assumes the input map is a HashMap of the registrar-to-groups mapping;
     *  and returns a clone not only of the map, but of each key-value pair
     *  contained in the mapping.
     *
     * @param groupsMap mapping from a set of registrars to the member groups
     *                  of each registrar 
     * 
     *  @return clone of the input map, and of each key-value pair contained
     *          in the input map
     */
    private Map deepCopy(HashMap groupsMap) {
        /* clone the input HashMap */
        HashMap newMap = (HashMap)(groupsMap.clone());
        /* clone the values of each mapping in place */
        Set eSet = newMap.entrySet();
        for(Iterator itr = eSet.iterator(); itr.hasNext(); ) {
            Map.Entry pair = (Map.Entry)itr.next();
            /* only need to clone the value of the order pair */
            pair.setValue( ((String[])pair.getValue()).clone() );
        }
        return newMap;
    }//end deepCopy

    /** Determines if two sets of registrar member groups have identical
     *  contents. Assumes there are no duplicates, and the sets can never
     *  be null.
     *
     * @param groupSet0    <code>String</code> array containing the group
     *                     names from the first set used in the comparison
     * @param groupSet1    <code>String</code> array containing the group
     *                     names from the second set used in the comparison
     * 
     *  @return <code>true</code> if the contents of each set is identical;
     *          <code>false</code> otherwise
     */
    private static boolean groupSetsEqual(String[] groupSet0,
                                          String[] groupSet1)
    {
        if(groupSet0.length != groupSet1.length) return false;
        /* every element of one set contained in the other set? */
        iLoop:
        for(int i=0;i<groupSet0.length;i++) {
            for(int j=0;j<groupSet1.length;j++) {
                if( groupSet0[i].equals(groupSet1[j]) ) {
                    continue iLoop;
                }
            }
            return false;
        }
        return true;
    }//end groupSetsEqual

    /** This method retrieves from the given <code>ServiceRegistrar</code>,
     *  the current groups in which that registrar is a member. If the
     *  registrar is un-reachable, then this method will discard the
     *  registrar.
     *
     * @param ServiceRegistrar instance of <code>ServiceRegistrar</code> 
     *                         referencing the registrar whose member 
     *                         groups are to be retrieved and returned
     * 
     *  @return <code>String</code> array containing the current member groups
     *          of the registrar referenced by the <code>reg</code> parameter
     */
    private String[] getActualGroups(ServiceRegistrar reg) {
        try {
            return reg.getGroups();
        } catch(RemoteException e) { // reg unreachable
            discard(reg);
            return null;
        }//end try
    }//end getActualGroups

    /** Convenience method that creates and returns a mapping of a single
     *  <code>ServiceRegistrar</code> instance to a set of groups.
     * 
     *  @param reg       instance of <code>ServiceRegistrar</code> 
     *                   corresponding to the registrar to use as the key
     *                   to the mapping
     *  @param curGroups <code>String</code> array containing the current
     *                   member groups of the registrar referenced by the 
     *                   <code>reg</code> parameter; and which is used
     *                   as the value of the mapping
     *
     *   @return <code>Map</code> instance containing a single mapping from
     *           a given registrar to its current member groups
     */
    private Map mapRegToGroups(ServiceRegistrar reg, String[] curGroups) {
        HashMap groupsMap = new HashMap(1);
        groupsMap.put(reg,curGroups);
        return groupsMap;
    }//end mapRegToGroups

    /**
     * This thread monitors the mulitcast announcements sent from the
     * lookup service(s) that have already been discovered by this class,
     * looking for indications that those announcements have terminated.
     * 
     * The data structure used to map the discovered lookup services to
     * the time of arrival of the most recent multicast announcement from
     * each such lookup service is examined at regular intervals; dependent
     * on the system property <code>net.jini.discovery.announce</code>.
     *
     * If the difference between the current time and the last time of
     * arrival for any announcement exceeds a predetermined threshold, the
     * corresponding lookup is polled for its current set of member groups.
     * If that lookup service is unreachable, or if it is reachable but its
     * member groups have been replaced, the lookup service is discarded.
     */
    private class AnnouncementTimerThread extends Thread {
        /* Default number of milliseconds in interval between announcements */
        private static final int INTERVAL    = 2 * 60 * 1000;
        /* Default interval sum to exceed for declaring announcements stopped*/
        private static final int N_INTERVALS = 3;
        /** Create a daemon thread */
        public AnnouncementTimerThread() {
            super("multicast announcement timer");
            setDaemon(true);
        }
        public synchronized void run() {
            int interval = INTERVAL;
            try {
                 interval = Integer.getInteger("net.jini.discovery.announce",
                                               interval).intValue();
            } catch (SecurityException e) { }
            int timeThreshold = N_INTERVALS*interval;
            try {
                while(!isInterrupted()) {
                    wait(interval);
                    long curTime = System.currentTimeMillis();
                    synchronized (registrars) {
                        /* can't modify regTimes while iterating over it, 
                         * so clone it
                         */
                        HashMap regTimesClone = (HashMap)(regTimes.clone());
                        Set eSet = regTimesClone.entrySet();
                        for(Iterator itr = eSet.iterator(); itr.hasNext(); ) {
                            Map.Entry pair   = (Map.Entry)itr.next();
                            ServiceID srvcID = (ServiceID)pair.getKey();
                            long tStamp = ((Long)pair.getValue()).longValue();
                            long deltaT = curTime - tStamp;
                            if(deltaT > timeThreshold) {
                                /* announcements stopped, queue reachability
                                 * test and potential discarded event
                                 */
                                IncomingUnicastResponse resp =
                               (IncomingUnicastResponse)registrars.get(srvcID);
                                synchronized (pendingDiscoveries) {
                                    if(    (pendingDiscoveries.add(resp))
                                        && (discovererThread == null) )
                                    {
                                        discovererThread = 
                                                      new UnicastDiscoverer();
                                        discovererThread.start();
                                    }//endif
                                }//end sync
                            }//end if
                        }//end loop (itr)
                    }//end sync
                }//end loop (!isInterrupted)
            } catch (InterruptedException e) { }
        }//end run
    }//end class AnnouncementTimerThread
}
