/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */

package net.jini.discovery;

import net.jini.core.discovery.LookupLocator;
import net.jini.core.lookup.ServiceRegistrar;

import java.io.IOException;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

/** 
 * This class is a helper utility class that organizes and manages all
 * discovery-related activities on behalf of a client or service. Rather
 * than providing its own facility for coordinating and maintaining
 * all of the necessary state information related to group names,
 * <code>LookupLocator</code> objects, and <code>DiscoveryListener</code>
 * objects, clients and services can employ this class to provide those
 * facilities on their behalf.
 *
 * @see net.jini.DiscoveryManagement
 * @see net.jini.DiscoveryGroupManagement
 * @see net.jini.DiscoveryLocatorManagement
 * 
 * @author Zane Pan
 */
public class LookupDiscoveryManager implements DiscoveryManagement,
                                               DiscoveryGroupManagement,
                                               DiscoveryLocatorManagement
{
    /** Constants used to indicate the type of discovery used */
    final public static int FROM_GROUP = 1;
    final public static int FROM_LOCATOR = 2;
    /** Constants used to indicate the type of event to send */
    private static final int DISCOVERED = 0;
    private static final int DISCARDED  = 1;
    private static final int CHANGED    = 2;

    private final ArrayList listeners = new ArrayList(1);
    private final ArrayList discoveredSet = new ArrayList();
    private LookupLocatorDiscovery locatorDisc;
    private LookupDiscovery lookupDisc;
    private DiscMgrListener locatorlistener = new DiscMgrListener
                                                               (FROM_LOCATOR);
    private DiscMgrListener grouplistener = new DiscMgrListener(FROM_GROUP);

    private final class ProxyReg  {
	public ServiceRegistrar proxy;
        public String[] memberGroups;
	private int from;
	private boolean bDiscarded = false;
        public ProxyReg(ServiceRegistrar proxy,
                        String[] memberGroups,
                        int from)
        {
            if(proxy == null) {
	        throw new IllegalArgumentException("proxy cannot be null");
            }
	    this.proxy        = proxy;
	    this.memberGroups = memberGroups;
	    this.from         = from;
	}	    

	public boolean equals(Object obj) {
	    if (obj instanceof ProxyReg){
		return proxy.equals(((ProxyReg)obj).proxy);
	    } else return false;
	}

	public int hashCode() {
	    return proxy.hashCode();
	}
	
	public synchronized void addFrom(int from) {
	    this.from |= from;
	}

	public synchronized boolean removeFrom(int from) {
	    if( from == 0 ) {
	        throw new RuntimeException
                                   ("call to removeFrom with bad argument");
            }
	    this.from &= ~from;
	    return (this.from == 0);
	}

	public synchronized int getFrom() {
	    return this.from;
	}

	public void discard() {
	    synchronized(this) {
		bDiscarded = true;
	    }
            /* Give group discovery priority to avoid race condition. Locator
             * discard will eventually occur as a result of 1st group discard.
             */
	    if((from & FROM_GROUP) != 0) {
                lookupDisc.discard(proxy);
	    } else if((from & FROM_LOCATOR) != 0) {
                locatorDisc.discard(proxy);
            }
	}

	public synchronized boolean isDiscarded() {
	    return bDiscarded;
	}

        public synchronized  String[] getMemberGroups() {
            return memberGroups;
        }

        public synchronized  void setMemberGroups(String[] newMemberGroups) {
            memberGroups = newMemberGroups;
        }
    }//end class ProxyReg

    private final class DiscMgrListener implements DiscoveryChangeListener {
	private int from;
	public DiscMgrListener(int from) {
	    this.from = from;
	}

	public void discovered(DiscoveryEvent e) {
            ServiceRegistrar[] proxys = (ServiceRegistrar[])e.getRegistrars();
            Map groupsMap = e.getGroups();
            HashMap discoveredGroupsMap = new HashMap(proxys.length);
            HashMap changedGroupsMap    = new HashMap(proxys.length);
	    for(int i=0; i<proxys.length; i++) {
		ProxyReg reg = findReg(proxys[i]);
		if(reg == null) { // newly discovered
                    reg = new ProxyReg(proxys[i],
                                       (String[])(groupsMap.get(proxys[i])),
                                       from);
		    addDiscoveredSet(reg);
                    discoveredGroupsMap.put(proxys[i],
                                            groupsMap.get(proxys[i]));
		} else { // already in set, send changed event?
                    if(from == FROM_LOCATOR) {
                        reg.addFrom(from); // update bit, send no event
                    } else { // (from == FROM_GROUP)
                        String[] oldGroups = reg.getMemberGroups();
                        String[] newGroups = (String[])groupsMap.get
                                                                 (reg.proxy);
                        if( groupSetsEqual(oldGroups,newGroups) ) {
                            reg.addFrom(from); // update bit, send no event
                        } else {
                            reg.setMemberGroups(newGroups);// update the groups
                            changedGroupsMap.put(reg.proxy,newGroups);
                        }//endif
                    }//endif
                }//endif
	    }//end loop
            notifyListener(discoveredGroupsMap, DISCOVERED);
            notifyListener(changedGroupsMap, CHANGED);
	}//end discovered

	public void discarded(DiscoveryEvent e) {
	    ServiceRegistrar[] proxys = (ServiceRegistrar[])e.getRegistrars();
            Map groupsMap = e.getGroups();
            HashMap discardedGroupsMap = new HashMap(proxys.length);
            HashMap changedGroupsMap   = new HashMap(proxys.length);
	    for(int i=0; i<proxys.length; i++) {
                ProxyReg reg = findReg(proxys[i]);
		if(reg != null) {
                    boolean commDiscard = true; //comm or noInterest discard?
                    String[] newGroups = (String[])groupsMap.get(reg.proxy);
                    if(from == FROM_GROUP) {
                        /* If we reach this point, we know that the current
                         * instance of this listener is a group discovery
                         * listener. Because this listener is a group
                         * discovery listener, one might think that the
                         * current reg's discovery bit must be set to either
                         * group discovery alone or both group and locator
                         * discovery; never to only locator discovery. But
                         * calling setGroups or removeGroups can result in
                         * a discarded event and the bit being cleared of
                         * group discovery, leaving only locator discovery.
                         * Because of this, the current reg's discovery bit
                         * is examined, and if it is set to only locator
                         * discovery, the current discarded event is ignored.
                         */
                        if(reg.getFrom() == FROM_LOCATOR)  continue;

                        String[] oldGroups = reg.getMemberGroups();
                        if( !groupSetsEqual(oldGroups,newGroups) ) {
                            commDiscard = false;
                        }
                    }//endif
                    if(commDiscard) {
                        if(removeDiscoveredSet(reg)) {
                            discardedGroupsMap.put(reg.proxy,newGroups);
                        } else {
                            reg.discard();
                        }
                    } else { //"no interest" discard sent by LookupDiscovery
                        reg.setMemberGroups(newGroups); // update reg's groups
                        changedGroupsMap.put(reg.proxy,newGroups);
                    }//endif
		}//endif
	    }//end loop
            notifyListener(discardedGroupsMap, DISCARDED);
            notifyListener(changedGroupsMap, CHANGED);
	}//end discarded

        public void changed(DiscoveryEvent e) {
            /* update the groups of each changed registrar */
            ServiceRegistrar[] proxys = (ServiceRegistrar[])e.getRegistrars();
            Map groupsMap = e.getGroups();
            HashMap changedGroupsMap = new HashMap(proxys.length);
	    for(int i=0; i<proxys.length; i++) {
		ProxyReg reg = findReg(proxys[i]);
		if(reg != null) {
                    String[] newGroups = (String[])groupsMap.get(reg.proxy);
                    reg.setMemberGroups(newGroups);
                    changedGroupsMap.put(reg.proxy,newGroups);
                }//endif
	    }//end loop
            notifyListener(changedGroupsMap, CHANGED);
        }//end changed

	private void addDiscoveredSet(ProxyReg reg) {
	    synchronized(discoveredSet) {
		discoveredSet.add(reg);
	    }
	}

	private boolean removeDiscoveredSet(ProxyReg reg) {
	    boolean bret = reg.removeFrom(from);
	    if(bret) {
		synchronized(discoveredSet) {
		    discoveredSet.remove(reg);
		}
	    }
	    return bret;
	}
    }

    /** 
     * Constructs an instance of this class that will organize and manage 
     * all discovery-related activities on behalf of the client or service
     * that instantiates this class.
     * <p>
     * If <code>null</code> (<code>DiscoveryGroupManagement.ALL_GROUPS</code>)
     * is input to the <code>groups</code> parameter, then attempts will be
     * made via group discovery to discover all lookup services located within
     * range of the entity that constructs this class. If the empty array
     * (<code>DiscoveryGroupManagement.NO_GROUPS</code>) is input to that
     * parameter, no group discovery will be performed until the set of
     * groups to discover is populated.
     * <p>
     * If an empty array or a <code>null</code> reference is input to the
     * <code>locators</code> parameter, no locator discovery will be performed
     * until the set of locators to discover is populated.
     *
     * @param groups   <code>String</code> array, none of whose elements may
     *                 be <code>null</code>, consisting of the names of the
     *                 groups whose members are lookup services the client
     *                 or service wishes to discover.
     * @param locators array of instances of <code>LookupLocator</code>, none
     *                 of whose elements may be <code>null</code>, and in
     *                 which each element corresponds to a specific lookup
     *                 service the client or service wishes to discover via
     *                 locator discovery.
     * @param listener a reference to <code>DiscoveryListener</code> object
     *                 that will be notified when a targeted lookup service
     *                 is discovered or discarded.
     *
     * @throws java.io.IOException because construction of this class may
     *         initiate the discovery process, which can throw an
     *         <code>IOException</code> when socket allocation occurs.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         either one or more of the elements of the <code>groups</code>
     *         parameter is <code>null</code>, or one or more elements of
     *         the <code>locators</code> parameter is null.
     *
     * @see net.jini.core.discovery.LookupLocator
     * @see net.jini.discovery.DiscoveryListener
     */
    public LookupDiscoveryManager(String[] groups,  
				  LookupLocator[] locators,
				  DiscoveryListener listener)
                                                           throws IOException
    {
        /* Configure for group discovery */
 	lookupDisc = new LookupDiscovery(groups);
	lookupDisc.addDiscoveryListener(grouplistener);
        /* Configure for locator discovery */
	locatorDisc = new LookupLocatorDiscovery(locators);
	locatorDisc.addDiscoveryListener(locatorlistener);
        /* Initiate the discovery process */
	if(listener != null)
	    listeners.add(listener);
    }

    /**
     * Returns an array consisting of the elements of the managed set
     * of locators; that is, instances of <code>LookupLocator</code> in
     * which each instance corresponds to a specific lookup service to
     * discover. The returned set will include both the set of 
     * <code>LookupLocator</code>s corresponding to lookup services 
     * that have already been discovered as well as the set of those
     * that have not yet been discovered. If the managed set of locators
     * is empty, this method will return the empty array. This method
     * returns a new array upon each invocation.
     *
     * @return <code>LookupLocator</code> array consisting of the elements
     *         of the managed set of locators
     *
     * @see net.jini.DiscoveryLocatorManagement#getLocators
     */
    public LookupLocator[] getLocators(){
	return locatorDisc.getLocators();
    }

    /**
     * Adds a set of locators to the managed set of locators. Elements in the
     * input set that duplicate (using the <code>LookupLocator.equals</code>
     * method) elements already in the managed set will be ignored. If the
     * empty array is input, the managed set of locators will not change.
     *
     * @param locators <code>LookupLocator</code> array consisting of the
     *                 locators to add to the managed set.
     * 
     * @throws java.lang.UnsupportedOperationException this exception 
     *         occurs when there is no managed set of locators to augment.
     *         That is, the current managed set of locators is
     *         <code>null</code>.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         either <code>null</code> is input to the <code>locators</code>
     *         parameter, or one or more of the elements of the
     *         <code>locators</code> parameter is <code>null</code>.
     *
     * @see net.jini.DiscoveryLocatorManagement#addLocators
     */
    public void addLocators(LookupLocator[] locators){
	locatorDisc.addLocators(locators);
    }

    /**
     * Deletes a set of locators from the managed set of locators, and discards
     * any already-discovered lookup service that corresponds to a deleted
     * locator. For any lookup service that is discarded as a result of an
     * invocation of this method, a discard notification is sent; and that
     * lookup service will not be eligible for re-discovery (assuming it is
     * not currently eligible for discovery through other means, such as
     * group discovery).
     * <p>
     * If the empty array is input, this method takes no action.
     *
     * @param locators <code>LookupLocator</code> array consisting of the
     *                 locators that will be removed from the managed set.
     * 
     * @throws java.lang.UnsupportedOperationException this exception 
     *         occurs when there is no managed set of locators from which
     *         remove elements.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         <code>null</code> is input to the <code>locators</code>
     *         parameter.
     *
     * @see net.jini.DiscoveryLocatorManagement#removeLocators
     */
    public void removeLocators(LookupLocator[] locators) {
	locatorDisc.removeLocators(locators);
    }

    /**
     * Replaces all of the locators in the managed set with locators from
     * a new set, and discards any already-discovered lookup service that
     * corresponds to a locator that is removed from the managed set
     * as a result of an invocation of this method. For any such lookup
     * service that is discarded, a discard notification is sent; and that
     * lookup service will not be eligible for re-discovery (assuming it is
     * not currently eligible for discovery through other means, such as
     * group discovery).
     * <p>
     * If the empty array is input, locator discovery will cease until this
     * method is invoked with an input parameter that is non-<code>null</code>
     * and non-empty.
     *
     * @param locators <code>LookupLocator</code> array consisting of the 
     *                 locators that will replace the current locators in the
     *                 managed set.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         <code>null</code> is input to the <code>locators</code>
     *         parameter.
     *
     * @see net.jini.DiscoveryLocatorManagement#setLocators
     */
    public void setLocators(LookupLocator[] locators) {
	locatorDisc.setLocators(locators);
    }
    
    /** 
     * Returns an array consisting of the elements of the managed set
     * of groups; that is, the names of the groups whose members are the
     * lookup services to discover. If the managed set of groups is empty,
     * this method will return the empty array. If there is no managed set
     * of groups, then null is returned; indicating that all groups are to
     * be discovered. This method returns a new array upon each invocation.
     *
     * @return <code>String</code> array consisting of the elements of the
     *         managed set of groups
     *
     * @see net.jini.DiscoveryGroupManagement#getGroups
     */
    public String[] getGroups() {
	return lookupDisc.getGroups();
    }

    /**   
     * Adds a set of group names to the managed set of groups. Elements in
     * the input set that duplicate elements already in the managed set
     * will be ignored. Once a new name is added to the managed set,
     * attempts will be made to discover all (as yet) undiscovered lookup
     * services that are members of the group having that name. If the empty
     * array (<code>DiscoveryGroupManagement.NO_GROUPS</code>) is input, the
     * managed set of groups will not change.
     *
     * Note that any entity that invokes this method must have
     * <code>DiscoveryPermission</code> on each of the groups in the
     * new set, otherwise a <code>SecurityException</code> will be
     * propagated through this method.
     *
     * @param groups <code>String</code> array consisting of the group names
     *               to add to the managed set.
     *
     * @throws java.io.IOException because an invocation of this method may
     *         result in the re-initiation of the discovery process, which can
     *         throw an <code>IOException</code> when socket allocation occurs.
     * 
     * @throws java.lang.UnsupportedOperationException this exception 
     *         occurs when there is no managed set of groups to augment.
     *         That is, the current managed set of groups is <code>null</code>.
     *         If the managed set of groups is <code>null</code>, all groups
     *         are being discovered; thus, requesting that a set of groups be
     *         added to the set of all groups makes no sense.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         either <code>null</code> is input to the <code>groups</code>
     *         parameter, or one or more of the elements of the
     *         <code>groups</code> parameter is <code>null</code>. If a
     *         <code>null</code> is input, then the entity is effectively
     *         requesting that "all groups" be added to the current managed
     *         set of groups; which is not allowed. (Note that if the entity
     *         wishes to change the managed set of groups from a finite set
     *         of names to "all groups", the <code>setGroups</code> method
     *         should be invoked with <code>null</code> input.)
     *
     * @see net.jini.DiscoveryGroupManagement#addGroups
     */
    public void addGroups(String[] groups) throws IOException {
	lookupDisc.addGroups(groups);
    }

    /**   
     * Deletes a set of group names from the managed set of groups. If the
     * empty array (<code>DiscoveryGroupManagement.NO_GROUPS</code>) is input,
     * this method takes no action.
     *
     * @param groups <code>String</code> array consisting of the group names
     *               that will be removed from the managed set.
     *
     * @throws java.lang.UnsupportedOperationException this exception 
     *         occurs when there is no managed set of groups from which to
     *         remove elements.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         <code>null</code> is input to the <code>groups</code>
     *         parameter.
     *
     * @see net.jini.DiscoveryGroupManagement#removeGroups
     */
    public void removeGroups(String[] groups) {
        if (groups == null) {
            throw new NullPointerException
                                ("can't remove null from groups to discover");
        }
        String[] curGroupsToDiscover = lookupDisc.getGroups();
        if (curGroupsToDiscover == null) {
            throw new UnsupportedOperationException
                                        ("can't remove from \"any groups\"");
        }
        if (curGroupsToDiscover.length == 0) return;

        HashSet curGroups = new HashSet(curGroupsToDiscover.length);
        for(int i=0;i<curGroupsToDiscover.length;i++) {
            curGroups.add(curGroupsToDiscover[i]);
        }
        boolean removed = false;
        for(int i=0;i<groups.length;i++) {
            removed |= curGroups.remove(groups[i]);
        }
        if(!removed) return; // nothing removed

        String[] newGroupsToDiscover = (String[])(curGroups).toArray
                                                (new String[curGroups.size()]);
        maybeDiscardOnGroupChangeRequest(newGroupsToDiscover);
	lookupDisc.removeGroups(groups); //generates discards that get ignored
    }

    /**   
     * Replaces all of the group names in the managed set with names from
     * a new set. Once a new group name has been placed in the managed
     * set, if there are lookup services belonging to that group that have
     * already been discovered, no event will be sent to the entity's
     * listener for those particular lookup services. Attempts to discover
     * all (as yet) undiscovered lookup services belonging to that group
     * will continue to be made.
     * <p>
     * If null (<code>DiscoveryGroupManagement.ALL_GROUPS</code>) is input
     * to this method, then attempts will be made to discover all (as yet)
     * undiscovered lookup services that are within range, and which
     * are members of any group. If the empty array
     * (<code>DiscoveryGroupManagement.NO_GROUPS</code>) is input, then
     * group discovery will cease until this method is invoked with an
     * input parameter that is non-<code>null</code> and non-empty.
     *
     * Note that any entity that invokes this method must have
     * <code>DiscoveryPermission</code> on each of the groups in the
     * new set, otherwise a <code>SecurityException</code> will be
     * propagated through this method.
     *
     * @param groups <code>String</code> array consisting of the group
     *               names that will replace the current names in the
     *               managed set.
     *
     * @throws java.io.IOException because an invocation of this method may
     *         result in the re-initiation of the discovery process, which can
     *         throw an <code>IOException</code> when socket allocation occurs.
     *
     * @see net.jini.DiscoveryGroupManagement#setGroups
     */
    public void setGroups(String[] groups) throws IOException {
        maybeDiscardOnGroupChangeRequest(groups);
        lookupDisc.setGroups(groups); //generates discards that get ignored
    }

    /**
     * Adds an instance of <code>DiscoveryListener</code> to the set of
     * objects listening for discovery events. Once the listener is
     * registered, it will be notified of all lookup services discovered
     * to date, and will then be notified as new lookup services are
     * discovered or existing lookup services are discarded.
     * <p>
     * If <code>null</code> is input, this method takes no action. If the
     * listener input to this method duplicates (using the <code>equals</code>
     * method) another element in the current set of listeners, no action
     * is taken.
     *
     * @param listener an instance of <code>DiscoveryListener</code>
     *                 corresponding to the listener to add to the set of
     *                 listeners.
     *
     * @see net.jini.DiscoveryListener
     * @see net.jini.DiscoveryManagement#addDiscoveryListener
     */
    public void addDiscoveryListener(DiscoveryListener listener) {
        HashMap groupsMap = new HashMap(discoveredSet.size());
	synchronized(listeners) {
            if(!listeners.contains(listener)) {
                listeners.add(listener);
            }
            if(discoveredSet.isEmpty()) return;
	    for(int i=0; i< discoveredSet.size(); i++) {
                ProxyReg reg = (ProxyReg)discoveredSet.get(i);
                groupsMap.put(reg.proxy,reg.getMemberGroups());
	    }
	}
        notifyListener(listener, groupsMap, DISCOVERED);
    }

    /**
     * Removes a listener from the set of objects listening for discovery
     * events. If the listener object input to this method does not exist
     * in the set of listeners, then this method will take no action.
     *
     * @param listener an instance of <code>DiscoveryListener</code>
     *                 corresponding to the listener to remove from the set
     *                 of listeners.
     *
     * @see net.jini.DiscoveryListener
     * @see net.jini.DiscoveryManagement#removeDiscoveryListener
     */
    public void removeDiscoveryListener(DiscoveryListener listener) {
	synchronized(listeners) {
	    listeners.remove(listener);
	}
    }

    /**
     * Returns an array of instances of <code>ServiceRegistrar</code>, each
     * corresponding to a proxy to one of the currently discovered lookup
     * services. For each invocation of this method, a new array is returned.
     *
     * @return array of instances of <code>ServiceRegistrar</code>, each
     *         corresponding to a proxy to one of the currently discovered
     *         lookup services
     *
     * @see net.jini.core.lookup.ServiceRegistrar
     * @see net.jini.DiscoveryManagement#removeDiscoveryListener
     */
    public ServiceRegistrar[] getRegistrars() {
	ArrayList proxySet = new ArrayList(1);;
	synchronized(discoveredSet) {
	    int k = 0;
	    Iterator iter = discoveredSet.iterator();
	    while(iter.hasNext()) {
		ProxyReg reg = (ProxyReg)iter.next();
		if(!reg.isDiscarded())
		    proxySet.add(reg.proxy);
	    }
	}
	ServiceRegistrar[] ret = new ServiceRegistrar[proxySet.size()];
	proxySet.toArray(ret);
	return ret;
    }

    /**
     * Removes an instance of <code>ServiceRegistrar</code> from the
     * managed set of lookup services, making the corresponding lookup
     * service eligible for re-discovery. This method takes no action if
     * the parameter input to this method is <code>null</code>, or if it
     * does not match (using <code>equals</code>) any of the elements in
     * the managed set.
     *
     * @param proxy the instance of <code>ServiceRegistrar</code> to discard
     *              from the managed set of lookup services
     *
     * @see net.jini.core.lookup.ServiceRegistrar
     * @see net.jini.DiscoveryManagement#discard
     */
    public void discard(ServiceRegistrar proxy) {
	ProxyReg reg = findReg(proxy);
	if(reg != null)
	    reg.discard();
    }

    /**
     * Terminates all threads, ending all discovery processing being
     * performed by the current instance of this class.
     * <p>
     * After this method has been invoked, no new lookup services will
     * be discovered, and the effect of any new operations performed
     * on the current instance of this class are undefined.
     *
     * @see net.jini.DiscoveryManagement#terminate
     */
    public void terminate() {
	synchronized(listeners) {
	    listeners.clear();
	}
	lookupDisc.terminate();
	locatorDisc.terminate();
    }

    /* return where the proxy come from. */
    public int getFrom(ServiceRegistrar proxy) {
	ProxyReg reg = findReg(proxy);
	if(reg != null)
	    return reg.getFrom();
	return 0;
    }

    private ProxyReg findReg(ServiceRegistrar proxy) {
	synchronized(discoveredSet) {
	    Iterator iter = discoveredSet.iterator();
	    while(iter.hasNext()) {
		ProxyReg reg =(ProxyReg)iter.next();
		if(reg.proxy.equals(proxy))
		    return reg;
	    }
	}
	return null;
    }

    private void notifyListener(Map groupsMap, int eventType) {
	if(groupsMap.isEmpty()) return;
	ArrayList notifies;
	synchronized(listeners) {
	    notifies = (ArrayList)listeners.clone();
	}
	Iterator iter = notifies.iterator();
	while(iter.hasNext()) {
	    DiscoveryListener l = (DiscoveryListener)iter.next();
            notifyListener(l, groupsMap, eventType);
	}
    }//end notifyListener
    
    private void notifyListener(DiscoveryListener l,
                                Map groupsMap,
                                int eventType)
    {
        /* always send discovered and discarded events, not always changed */
        if((eventType == CHANGED) && !(l instanceof DiscoveryChangeListener)){
            return;
        }
        DiscoveryEvent evt = new DiscoveryEvent(LookupDiscoveryManager.this,
                                                deepCopy((HashMap)groupsMap) );
        switch(eventType) {
            case DISCOVERED:
                l.discovered(evt);
                break;
            case DISCARDED:
                l.discarded(evt);
                break;
            case CHANGED:
                ((DiscoveryChangeListener)l).changed(evt);
                break;
        }//end switch
    }//end notifyListener

    /** When the set of groups to discover is changed through the invocation
     *  of either <code>setGroups</code> or <code>removeGroups</code>, it
     *  may be necessary to discard one or more of the already-discovered
     *  registrars. This is because those methods may change the groups to 
     *  discover in such a way that one or more of the discovered registrars
     *  are no longer of interest to the client through group discovery.
     *  <p>
     *  For each already-discovered registrar that is an element of the set
     *  of registrars to discover through group discovery, this method
     *  compares the member groups of the registrar to the elements of the
     *  new set of groups to discover resulting from an invocation of
     *  either <code>setGroups</code> or <code>removeGroups</code>. If
     *  the registrar's groups are no longer of interest to the client,
     *  this method discards the registrar from group discovery.
     *  <p>
     *  This method is called by both <code>setGroups</code> and
     *  <code>removeGroups</code>, and is necessary because the normal
     *  discard event resulting from the invocation of either 
     *  <code>setGroups</code> or <code>removeGroups</code> on the
     *  <code>LookupDiscovery</code> utility would be interpreted by
     *  this class' listener as a "communication" discard rather than
     *  as a "no interest" discard. This is because the groups of the
     *  discarded registrars themselves have not changed, only the groups
     *  of interest to the client have changed. 
     *  <p>
     *  Any registrars discarded from group discovery that are also
     *  configured for locator discovery would be discarded from
     *  locator discovery as well; and then would be immediately
     *  re-discovered through locator discovery (because the client has
     *  not changed the locators it is interested in). Thus, for the
     *  situation where the client changes the set of groups to discover,
     *  the discard process must be invoked at the point where the
     *  groups of interest are actually changed. That is, when either
     *  <code>setGroups</code> or <code>removeGroups</code> are called,
     *  the appropriate registrars must be made ineligible for group
     *  discovery, and a discarded must be sent to the client.
     *
     * @param newGroupsToDiscover <code>String</code> array containing the
     *                            new set of groups this class is configured
     *                            to discover
     *                     names from the first set used in the comparison
     */
    public void maybeDiscardOnGroupChangeRequest(String[] newGroupsToDiscover){
        HashMap discardGroupsMap = new HashMap(discoveredSet.size());
        synchronized(discoveredSet) {
            Iterator iter = discoveredSet.iterator();
            for(int i=0;iter.hasNext();i++) {
                ProxyReg reg = (ProxyReg)iter.next();
                int from = reg.getFrom();
                if(    ((from & FROM_GROUP) != 0)
                    && !stillInterested(reg.memberGroups,newGroupsToDiscover) )
                {
                    /* no longer interested, discard it */
                    discardGroupsMap.put(reg.proxy,reg.memberGroups);
                    if(reg.removeFrom(FROM_GROUP)) { // clear the group bit
                        iter.remove();
                    }
                }//endif
            }//end loop
        }//end sync
        notifyListener(discardGroupsMap, DISCARDED);
    }//end maybeDiscardOnGroupChangeRequest


    /** Determines if two sets of registrar member groups have identical
     *  contents. Assumes there are no duplicates, and the sets can never
     *  be null.
     *
     * @param groupSet0    <code>String</code> array containing the group
     *                     names from the first set used in the comparison
     * @param groupSet1    <code>String</code> array containing the group
     *                     names from the second set used in the comparison
     * 
     *  @return <code>true</code> if the contents of each set is identical;
     *          <code>false</code> otherwise
     */
    private static boolean groupSetsEqual(String[] groupSet0,
                                          String[] groupSet1)
    {
        if(groupSet0.length != groupSet1.length) return false;
        /* every element of one set contained in the other set? */
        iLoop:
        for(int i=0;i<groupSet0.length;i++) {
            for(int j=0;j<groupSet1.length;j++) {
                if( groupSet0[i].equals(groupSet1[j]) ) {
                    continue iLoop;
                }
            }
            return false;
        }
        return true;
    }//end groupSetsEqual

    /** Determines if at least one member group of a given registrar is
     *  contained in the given set of desired groups.
     *
     * @param regGroups     <code>String</code> array containing the member
     *                      groups of a given registrar (will never be null)
     * @param desiredGroups <code>String</code> array containing the groups
     *                      to discover (can be null - ALL_GROUPS)
     * 
     *  @return <code>true</code> if at least one element of regGroups is
     *          contained in desiredGroups; <code>false</code> otherwise
     */
    private boolean stillInterested(String[] regGroups,String[] desiredGroups){
        if(desiredGroups == DiscoveryGroupManagement.ALL_GROUPS) return true;
        if(desiredGroups.length == 0) return false;
	for(int i=0;i<regGroups.length;i++) {
	    for (int j=0;j<desiredGroups.length;j++) {
                if(regGroups[i].equals(desiredGroups[j])) return true;
            }
        }
	return false;
    }//end stillInterested

    /** Creates and returns a deep copy of the input parameter. This method
     *  assumes the input map is a HashMap of the registrar-to-groups mapping;
     *  and returns a clone not only of the map, but of each key-value pair
     *  contained in the mapping.
     *
     * @param groupsMap mapping from a set of registrars to the member groups
     *                  of each registrar 
     * 
     *  @return clone of the input map, and of each key-value pair contained
     *          in the input map
     */
    private Map deepCopy(HashMap groupsMap) {
        /* clone the input HashMap */
        HashMap newMap = (HashMap)(groupsMap.clone());
        /* clone the values of each mapping in place */
        Set eSet = newMap.entrySet();
        for(Iterator itr = eSet.iterator(); itr.hasNext(); ) {
            Map.Entry pair = (Map.Entry)itr.next();
            /* only need to clone the value of the order pair */
            pair.setValue( ((String[])pair.getValue()).clone() );
        }
        return newMap;
    }//end deepCopy
}
