/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package net.jini.discovery;

import net.jini.core.discovery.LookupLocator;
import net.jini.core.entry.Entry;
import net.jini.core.lookup.ServiceRegistrar;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import java.net.Socket;
import java.net.SocketException;
import java.net.UnknownHostException;

import java.security.AccessController;
import java.security.PrivilegedAction;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

/**
 * This class encapsulates the functionality required of an entity that
 * wishes to employ the unicast discovery protocol to discover a lookup
 * service. This utility provides an implementation that makes the process
 * of finding specific lookup services much simpler for both services and
 * clients.
 * <p>
 * Because this class participates in only the unicast discovery protocol,
 * and because the unicast discovery protocol imposes no restriction on the
 * physical location of the entity relative to a lookup service, this utility
 * can be used to discover lookup services running on hosts that are located
 * far from, or near to, the host on which the entity is running. This lack
 * of a restriction on location brings with it a requirement that the
 * discovering entity supply this class with specific information about the
 * desired lookup services; namely, the location of the device(s) hosting
 * each lookup service. This information is supplied through an instance
 * of the <code>LookupLocator</code> class.
 *
 * @author Zane Pan
 */
public class LookupLocatorDiscovery implements DiscoveryManagement,
                                               DiscoveryLocatorManagement
{
    private static class LocatorReg implements Comparable {
	public ServiceRegistrar proxy = null;
	public final LookupLocator l;
        public String[] memberGroups = null;
	/** Since cnt only gets modified in the constructor, we therefore
	 *  do not need to synchronize on it.
	 */
	private static final long MINI_RETRY = 15000;
	private static final long[] sleepTime = {0,5*1000, 10*1000, 20*1000,
	                                         30*1000, 60*1000, 
						 2*60*1000, 4*60*1000,
						 8*60*1000, 15*60*1000};
	private static int cnt = 0;
	public long nextTryTime;
	private int numFails = 0;
	private final int id;
	private long time = 0;

	public LocatorReg(LookupLocator l) {
	    id = cnt++;
	    this.l = l;
	    calcNextTryTime();
	}
	
	public void calcNextTryTime() {
	    nextTryTime =  System.currentTimeMillis() + sleepTime[numFails];
	    if(numFails < sleepTime.length -1)
		numFails++;
	}
	
	public boolean equals(Object obj) {
	    return l.equals(((LocatorReg)obj).l);
	}
	
	public int hashCode() {
	    return l.hashCode();
	}
       
	public int compareTo(Object obj) {
	    if (equals(obj)) {
		return 0;
	    }
	    LocatorReg e = (LocatorReg)obj;
	    if(nextTryTime == e.nextTryTime ) {
		if(id < e.id )
		    return 1;
		else
		    return -1;
	    }
	    if (nextTryTime < e.nextTryTime)
		return 1;
	    return -1;
	}

	public boolean tryGetProxy() {
	    if (proxy != null ) {
		throw new IllegalArgumentException
		    ("LookupLocator has been discovered already");

	    }
	    try {
                doUnicastDiscovery(l);
		time =  System.currentTimeMillis();
		return true;
	    } catch (Exception e) {
                printException(e);
		calcNextTryTime();
		return false;
	    }
	}

       /** This method only gets called from discard(). It tries to 
	*  prevent the discard method from being called too frequently. It
	*  measures the time from when discovered() gets called to when
	*  discard() gets called. If the time is less than MINI_RETRY,
	*  we delay our next discover.
	*/
	public void fixupNextTryTime()  {
	    if( (System.currentTimeMillis() - time) > MINI_RETRY ) {
		numFails = 0;
	    }
	}

       /** This method employs the unicast discovery protocol to discover
        *  the registrar having <code>LookupLocator</code> equal to the
        *  value contained in the <code>locator</code> parameter.
	*/
        private void doUnicastDiscovery(LookupLocator locator)
                              throws UnknownHostException, SocketException,
                                     IOException, ClassNotFoundException
        {
            Socket sock = new Socket(locator.getHost(),locator.getPort());
            try {
                IncomingUnicastResponse resp;
                sock.setSoTimeout(unicastSocketTimeout);
                OutputStream os = sock.getOutputStream();
                InputStream is = sock.getInputStream();
                OutgoingUnicastRequest.marshal(os);
                resp = new IncomingUnicastResponse(is);
                proxy        = resp.getRegistrar();
                memberGroups = resp.getGroups();
            } finally {
                try {
                    sock.close();
                } catch (IOException e) { /*ignore*/ }
            }
        }//end doUnicastDiscovery

    }//end class LocatorReg
 
    /**
     * Stores LookupLocators that have not been discovered yet. It is
     * sorted on the nextTryTime. The last key has the least value.
     */ 
    private final TreeMap locatorsMap = new TreeMap();
    /** Stores LookupLocators that have been discovered */
    private final ArrayList discoveredLocators = new ArrayList(1);
    /**
     * Thread to keep trying to find proxies from LookupLocators in
     * the TreeMap. The thread terminates if the TreeMap is empty,
     * and gets created if the TreeMap is not empty.
     */
    private DiscoveryThread discoveryThread = null;
    /** Thread that handles pending notifications. */
    private Notifier notifierThread;
    /** Notifications to be sent to listeners. */
    private LinkedList pendingNotifies = new LinkedList();
    /** Stores DiscoveryListeners **/
    private final ArrayList listeners = new ArrayList(1);

    /** Store the LocatorReg during RMI call */
    private LocatorReg regInRMICall = null;

    /** Time to wait for a response from the lookup service before giving up */
    private static final int unicastSocketTimeout =
	((Integer)AccessController.doPrivileged(new PrivilegedAction() {
	    public Object run() {
		int timeout = 60 * 1000;
		try {
		    return Integer.getInteger("net.jini.discovery.timeout",
					      timeout);
		} catch (SecurityException e) {
		    return new Integer(timeout);
		}
	    }
	})).intValue();

    /**
     * Create a LookupLocatorDiscovery, with an initial array of
     * LookupLocators to be managed.  For each managed LookupLocator,
     * unicast discovery is performed to obtain a ServiceRegistrar
     * proxy for that lookup service.
     * 
     * @param locators the locators to discover
     *
     */
    public LookupLocatorDiscovery(LookupLocator[] locators) {
	synchronized(this) {
	    discoverLocators(locators);
	}
    }

    /**
     * Add a DiscoveryListener to the listener set. The listener's
     * discovered method gets called right way with an array of 
     * ServiceRegistrars that have already been discoved, and will
     * be called in the future whenever additional lookup services
     * are discovered.
     *
     * @param l the new DiscoveryListener to add
     *
     */
    public void addDiscoveryListener(DiscoveryListener l) {
	synchronized(this) {
	    if(listeners.contains(l)) return; //already have this listener
	    listeners.add(l);
	    if(!discoveredLocators.isEmpty()) {
                HashMap groupsMap = new HashMap(discoveredLocators.size());
                Iterator iter = discoveredLocators.iterator();
                for (int i = 0; iter.hasNext(); i++) {
                    LocatorReg reg = (LocatorReg)iter.next();
                    groupsMap.put(reg.proxy,reg.memberGroups);
                }//end loop
		ArrayList list = new ArrayList(1);
		list.add(l);
                addNotify(list, groupsMap, false);
	    }//endif
	}//end sync
    }

    /**
     * Remove a DiscoveryListener from the listener set. It does
     * nothing if the DiscoveryListener does not exist in the 
     * the listener set.
     *
     * @param l the existing DiscoveryListener to remove
     *
     */
    public synchronized void removeDiscoveryListener(DiscoveryListener l) {
	int index = listeners.indexOf(l);
	if(index != -1)
	    listeners.remove(index);
    }

    /**
     * Removes an instance of <code>ServiceRegistrar</code> from the
     * managed set of lookup services, making the corresponding lookup
     * service eligible for re-discovery. This method takes no action if
     * the parameter input to this method is <code>null</code>, or if it
     * does not match (using <code>equals</code>) any of the elements in
     * the managed set.
     *
     * @param proxy the instance of <code>ServiceRegistrar</code> to discard
     *              from the managed set of lookup services
     *
     * @see net.jini.core.lookup.ServiceRegistrar
     * @see net.jini.DiscoveryManagement#discard
     */
    public void discard(ServiceRegistrar proxy) {
	synchronized(this) {
	    LookupLocator lct = findRegFromProxy(proxy);
	    if(lct == null) return;
	    LocatorReg reg = removeDiscoveredLocator(lct);
            HashMap groupsMap = new HashMap(1);
            groupsMap.put(reg.proxy,reg.memberGroups);
	    reg.proxy = null;
            reg.memberGroups = null;
	    reg.fixupNextTryTime();
	    reg.calcNextTryTime();
	    addToMap(reg);
	    if(!listeners.isEmpty()) {
                addNotify((ArrayList)listeners.clone(), groupsMap, true);

	    }
	}
    }

    /**
     * Returns an array consisting of the elements of the managed set
     * of locators; that is, instances of <code>LookupLocator</code> in
     * which each instance corresponds to a specific lookup service to
     * discover. The returned set will include both the set of 
     * <code>LookupLocator</code>s corresponding to lookup services 
     * that have already been discovered as well as the set of those
     * that have not yet been discovered. If the managed set of locators
     * is empty, this method will return the empty array. This method
     * returns a new array upon each invocation.
     *
     * @return <code>LookupLocator</code> array consisting of the elements
     *         of the managed set of locators
     *
     * @see net.jini.DiscoveryLocatorManagement#getLocators
     */
    public synchronized LookupLocator[] getLocators() {
	int size = discoveredLocators.size() + locatorsMap.size();
	if(regInRMICall != null )
	    size +=1;

	LookupLocator[] ret = new LookupLocator[size];
	int k = 0;

	Iterator iter = discoveredLocators.iterator();
	while(iter.hasNext()) 
	    ret[k++] = ((LocatorReg)iter.next()).l;

	iter = locatorsMap.values().iterator();
	while(iter.hasNext())
	    ret[k++] = ((LocatorReg)iter.next()).l;

	if(regInRMICall != null )
	    ret[k] = regInRMICall.l;

	return ret;
    }

    /**
     * Adds a set of locators to the managed set of locators. Elements in the
     * input set that duplicate (using the <code>LookupLocator.equals</code>
     * method) elements already in the managed set will be ignored. If the
     * empty array is input, the managed set of locators will not change.
     *
     * @param locators <code>LookupLocator</code> array consisting of the
     *                 locators to add to the managed set.
     * 
     * @throws java.lang.NullPointerException this exception occurs when
     *         either <code>null</code> is input to the <code>locators</code>
     *         parameter, or one or more of the elements of the
     *         <code>locators</code> parameter is <code>null</code>.
     *
     * @see net.jini.DiscoveryLocatorManagement#addLocators
     */
    public synchronized void addLocators(LookupLocator[] locators) {
        testSetForNull(locators);
	discoverLocators(locators);
    }

    /**
     * Deletes a set of locators from the managed set of locators, and discards
     * any already-discovered lookup service that corresponds to a deleted
     * locator. For any lookup service that is discarded as a result of an
     * invocation of this method, a discard notification is sent; and that
     * lookup service will not be eligible for re-discovery (assuming it is
     * not currently eligible for discovery through other means, such as
     * group discovery).
     * <p>
     * If the empty array is input, this method takes no action.
     *
     * @param locators <code>LookupLocator</code> array consisting of the
     *                 locators that will be removed from the managed set.
     * 
     * @throws java.lang.NullPointerException this exception occurs when
     *         either <code>null</code> is input to the <code>locators</code>
     *         parameter, or one or more of the elements of the
     *         <code>locators</code> parameter is <code>null</code>.
     *
     * @see net.jini.DiscoveryLocatorManagement#removeLocators
     */
    public void removeLocators(LookupLocator[] locators) {
        testSetForNull(locators);
	synchronized(this) {
            HashMap groupsMap = new HashMap(1);
	    for(int i=0; i<locators.length; i++) {
		LocatorReg reg = removeDiscoveredLocator(locators[i]);
		if(reg != null) {
                    groupsMap.put(reg.proxy,reg.memberGroups);
		    continue;
		}
		reg = findReg(locators[i]);
		if(reg != null) {
		    removeLocatorReg(reg);
		    notifyAll();
		}
	    }
	    if(!groupsMap.isEmpty() && !listeners.isEmpty()) {
                addNotify((ArrayList)listeners.clone(), groupsMap, true);
            }
	}
    }

    private void removeLocatorReg(LocatorReg reg) {
	if(reg == regInRMICall )
	    regInRMICall = null;
	else
	    locatorsMap.remove(reg);
    }

     /**
     * Replaces all of the locators in the managed set with locators from
     * a new set, and discards any already-discovered lookup service that
     * corresponds to a locator that is removed from the managed set
     * as a result of an invocation of this method. For any such lookup
     * service that is discarded, a discard notification is sent; and that
     * lookup service will not be eligible for re-discovery (assuming it is
     * not currently eligible for discovery through other means, such as
     * group discovery).
     * <p>
     * If the empty array is input, locator discovery will cease until this
     * method is invoked with an input parameter that is non-<code>null</code>
     * and non-empty.
     *
     * @param locators <code>LookupLocator</code> array consisting of the 
     *                 locators that will replace the current locators in the
     *                 managed set.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         either <code>null</code> is input to the <code>locators</code>
     *         parameter, or one or more of the elements of the
     *         <code>locators</code> parameter is <code>null</code>.
     *
     * @see net.jini.DiscoveryLocatorManagement#setLocators
     */
   public void setLocators(LookupLocator[] locators) {
        testSetForNull(locators);
	synchronized(this) {
            HashMap groupsMap = new HashMap(1);
	    Iterator iter = discoveredLocators.iterator();
	    while(iter.hasNext()) {
		LocatorReg reg = (LocatorReg)iter.next();
		if(!isArrayContains(locators, reg.l)) {
		    iter.remove();
                    groupsMap.put(reg.proxy,reg.memberGroups);
		}
	    }
	    iter = locatorsMap.values().iterator();
	    while(iter.hasNext()) {
		LocatorReg reg = (LocatorReg)iter.next();
		if(!isArrayContains(locators, reg.l)) {
		    iter.remove();
		}
	    }
	    if(regInRMICall != null) {
	       if(!isArrayContains(locators, regInRMICall.l))
		   regInRMICall = null;
	    }
	    discoverLocators(locators);
	    if(!groupsMap.isEmpty() && !listeners.isEmpty()) {
                addNotify((ArrayList)listeners.clone(), groupsMap, true);
            }
	}
    }

    /**
     * Terminates all threads, ending all locator discovery processing being
     * performed by the current instance of this class.
     * <p>
     * After this method has been invoked, no new lookup services will
     * be discovered, and the effect of any new operations performed
     * on the current instance of this class are undefined.
     *
     * @see net.jini.DiscoveryManagement#terminate
     */
    public synchronized void terminate() {
	if(discoveryThread != null )
	    discoveryThread.interrupt();
	regInRMICall = null;
    }

    private ServiceRegistrar[] buildServiceRegistrar() {
	int k = 0;
	ServiceRegistrar[] proxys =
	    new ServiceRegistrar[discoveredLocators.size()];
	Iterator iter = discoveredLocators.iterator();
	while(iter.hasNext()) {
	    LocatorReg reg = (LocatorReg)iter.next();
	    proxys[k++] = reg.proxy;
	}
	return proxys;
    }

    private void discoverLocators(LookupLocator[] lcts) {
	if ( lcts == null)
	    return;
	LookupLocator lct;
	for(int i=0; i<lcts.length; i++) {
	    if(isDiscovered(lcts[i]))
		continue;
	    LocatorReg reg = findReg(lcts[i]);
	    if(reg == null) {
		reg = new LocatorReg(lcts[i]);
		addToMap(reg);
	    }
	}
    }

    private LocatorReg findReg(LookupLocator lct) {
	if(checkRegInRMICall(lct))
	    return regInRMICall;

	Iterator iter = locatorsMap.values().iterator();
	while(iter.hasNext()) {
	    LocatorReg reg = (LocatorReg)iter.next();
	    if (reg.l.equals(lct))
		return reg;
	}
	return null;
    }

    private boolean checkRegInRMICall(LookupLocator lct) {
	if (regInRMICall == null )
	    return false;
	return regInRMICall.l.equals(lct);
    }

    private class DiscoveryThread extends Thread {

	public DiscoveryThread() {
	    super("DiscoveryThread");
	    setDaemon(true);
	}

        public void run() {
	    try {
		while (!isInterrupted()) {
		    synchronized (LookupLocatorDiscovery.this) {
			if(locatorsMap.isEmpty()) {
			    discoveryThread = null;
			    return;
			}
			LocatorReg reg = (LocatorReg)locatorsMap.lastKey();
			long tryTime = reg.nextTryTime; 
			long now = System.currentTimeMillis();
			if(tryTime > now ) {
			    LookupLocatorDiscovery.this.wait(tryTime-now);
			    continue;
			}
			regInRMICall = reg;
			locatorsMap.remove(reg);
		    }
		    regTryGetProxy(regInRMICall);
		}
	    } catch (InterruptedException ex) {
	    }
	}
    }

    private void regTryGetProxy(LocatorReg reg) {
	boolean b = reg.tryGetProxy();
	synchronized (this) {
	    if(regInRMICall == null) return;
	    regInRMICall = null;
	    if(!b) {
		addToMap(reg);
		return;
	    }
	    discoveredLocators.add(reg);
	    if(!listeners.isEmpty())
		addNotify((ArrayList)listeners.clone(),
                          mapRegToGroups(reg.proxy,reg.memberGroups),
                          false);
	}
    }

    private void addToMap(LocatorReg reg) {
	locatorsMap.put(reg, reg);
	if(discoveryThread == null ) {
	   discoveryThread =  new DiscoveryThread();
	   discoveryThread.start();
	} else
	    notifyAll();
    }

    private boolean isDiscovered(LookupLocator lct) {
	Iterator iter = discoveredLocators.iterator();
	while(iter.hasNext()) {
 	    LocatorReg reg = (LocatorReg)iter.next();
	    if(reg.l.equals(lct))
		return true;
	}
	return false;
    }

    private LocatorReg removeDiscoveredLocator(LookupLocator lct) {
	Iterator iter = discoveredLocators.iterator();
	while(iter.hasNext()) {
	    LocatorReg reg = (LocatorReg)iter.next();
	    if(reg.l.equals(lct)) {
		iter.remove();
		return reg;
	    }
	}
	return null;
    }

    /**
     * Add a notification task to the pending queue, and start a
     * notifierThread if one isn't running.
     */
    private void addNotify(ArrayList notifies,
                           Map groupsMap,
			   boolean discard)
    {
	synchronized (pendingNotifies) {
	    pendingNotifies.addLast(new NotifyTask(notifies,
                                                   groupsMap,
                                                   discard));
	    if (notifierThread == null) {
		notifierThread = new Notifier();
		notifierThread.start();
	    }
	}
    }

    /** Simple holder for notify task data */
    private static class NotifyTask {
	/** The listeners to notify */
	public final ArrayList listeners;
	/** Map of discovered registrars to groups in which each is a member */
	public final Map groupsMap;
	/** True if discarded, else discovered */
	public final boolean discard;

	public NotifyTask(ArrayList listeners,
                          Map groupsMap,
			  boolean discard)
	{
	    this.listeners = listeners;
	    this.groupsMap = groupsMap;
	    this.discard   = discard;
	}
    }

    /** Task thread for calling listeners with notifications */
    private class Notifier extends Thread {

	/** Create a daemon thread */
	public Notifier() {
	    super("event notifier");
	    setDaemon(true);
	}

	public void run() {
	    while (true) {
		NotifyTask task;
		synchronized (pendingNotifies) {
		    if (pendingNotifies.isEmpty()) {
			notifierThread = null;
			return;
		    }
		    task = (NotifyTask)pendingNotifies.removeFirst();
		}
		for (Iterator iter = task.listeners.iterator();
		     iter.hasNext(); )
		{
		    DiscoveryListener l = (DiscoveryListener)iter.next();
		    DiscoveryEvent e =
                        new DiscoveryEvent(LookupLocatorDiscovery.this,
                                           deepCopy((HashMap)task.groupsMap) );
		    if (task.discard) {
			l.discarded(e);
		    } else {
			l.discovered(e);
                    }
		}
	    }
	}
    }

    private boolean isArrayContains(Object[] a, Object obj) {
	for(int i=0; i<a.length; i++ ) {
	    if(a[i].equals(obj))
		return true;
	}
	return false;
    }

    private LookupLocator findRegFromProxy(ServiceRegistrar proxy) {
	Iterator iter = discoveredLocators.iterator();
	while(iter.hasNext()) {
	    LocatorReg reg = (LocatorReg)iter.next();
	    if((reg.proxy).equals(proxy ))
		return reg.l;
	}
	return null;    
    }
    
    /* debug function */
    private void printMap () {
	Iterator iter = locatorsMap.values().iterator();
	while(iter.hasNext()) {
	    LocatorReg reg = (LocatorReg)iter.next();	    
	    System.out.println("printMap reg:" + reg.id);
	}
    }

    /**
     * Returns the set of <code>LookupLocator</code> objects representing the
     * desired lookup services that are currently discovered. If no lookup
     * services are currently discovered, this method returns the empty array.
     * This method returns a new array upon each invocation.
     *
     * @return <code>LookupLocator</code> array consisting of the elements
     *         from the managed set of locators that correspond to lookup
     *         services that have already been discovered.
     */
    public synchronized LookupLocator[] getDiscoveredLocators() {
	int size = discoveredLocators.size();
	LookupLocator[] ret = new LookupLocator[size];
	int k = 0;

	Iterator iter = discoveredLocators.iterator();
	while(iter.hasNext()) 
	    ret[k++] = ((LocatorReg)iter.next()).l;
	return ret;
    }

    /**
     * Returns the set of <code>LookupLocator</code> objects representing the
     * desired lookup services that have not yet been discovered. If all
     * of the desired lookup services are currently discovered, this method
     * returns the empty array. This method returns a new array upon each
     * invocation.
     *
     * @return <code>LookupLocator</code> array consisting of the elements
     *         from the managed set of locators that correspond to lookup
     *         services that have not yet been discovered.
     */
    public synchronized LookupLocator[] getUndiscoveredLocators() {
	int size = locatorsMap.size();
	if(regInRMICall != null )
	    size +=1;
	LookupLocator[] ret = new LookupLocator[size];
	int k = 0;

	Iterator iter = locatorsMap.values().iterator();
	while(iter.hasNext())
	    ret[k++] = ((LocatorReg)iter.next()).l;
	if(regInRMICall != null )
	    ret[k] = regInRMICall.l;
	return ret;
    }

    /**
     * Returns an array of instances of <code>ServiceRegistrar</code>, each
     * corresponding to a proxy to one of the currently discovered lookup
     * services. For each invocation of this method, a new array is returned.
     *
     * @return array of instances of <code>ServiceRegistrar</code>, each
     *         corresponding to a proxy to one of the currently discovered
     *         lookup services
     *
     * @see net.jini.core.lookup.ServiceRegistrar
     * @see net.jini.DiscoveryManagement#removeDiscoveryListener
     */
    public ServiceRegistrar[] getRegistrars() {
        synchronized(this) {
            if((discoveredLocators == null) || (discoveredLocators.isEmpty())){
                return new ServiceRegistrar[0];
            }
            return buildServiceRegistrar();
        }
    }

    /**
     * This method is used by the public methods of this class that are
     * specified to throw a <code>NullPointerException</code> when the
     * set of locators is either <code>null</code> or contains one or
     * more <code>null</code> elements; in either case, this method 
     * throws a <code>NullPointerException</code> which should be allowed
     * to propogate outward.
     * 
     * @throws java.lang.NullPointerException this exception occurs when
     *         either <code>null</code> is input to the <code>locatorSet</code>
     *         parameter, or one or more of the elements of the
     *         <code>locatorSet</code> parameter is <code>null</code>.
     */
    private void testSetForNull(LookupLocator[] locatorSet) {
        if(locatorSet == null) {
            throw new NullPointerException("null locator array");
        }
        for(int i=0;i<locatorSet.length;i++) {
            if(locatorSet[i] == null) {
                throw new NullPointerException
                                           ("null element in locator array");
            }
        }
    }

    /** Creates and returns a deep copy of the input parameter. This method
     *  assumes the input map is a HashMap of the registrar-to-groups mapping;
     *  and returns a clone not only of the map, but of each key-value pair
     *  contained in the mapping.
     *
     * @param groupsMap mapping from a set of registrars to the member groups
     *                  of each registrar 
     * 
     *  @return clone of the input map, and of each key-value pair contained
     *          in the input map
     */
    private Map deepCopy(HashMap groupsMap) {
        /* clone the input HashMap */
        HashMap newMap = (HashMap)(groupsMap.clone());
        /* clone the values of each mapping in place */
        Set eSet = newMap.entrySet();
        for(Iterator itr = eSet.iterator(); itr.hasNext(); ) {
            Map.Entry pair = (Map.Entry)itr.next();
            /* only need to clone the value of the order pair */
            pair.setValue( ((String[])pair.getValue()).clone() );
        }
        return newMap;
    }//end deepCopy

    /** Convenience method that creates and returns a mapping of a single
     *  <code>ServiceRegistrar</code> instance to a set of groups.
     * 
     *  @param reg       instance of <code>ServiceRegistrar</code> 
     *                   corresponding to the registrar to use as the key
     *                   to the mapping
     *  @param curGroups <code>String</code> array containing the current
     *                   member groups of the registrar referenced by the 
     *                   <code>reg</code> parameter; and which is used
     *                   as the value of the mapping
     *
     *   @return <code>Map</code> instance containing a single mapping from
     *           a given registrar to its current member groups
     */
    private Map mapRegToGroups(ServiceRegistrar reg, String[] curGroups) {
        HashMap groupsMap = new HashMap(1);
        groupsMap.put(reg,curGroups);
        return groupsMap;
    }//end mapRegToGroups

    /** Print the exception stack trace if debugging is enabled. */
    private static void printException(Exception e) {
        try {
            if (System.getProperty("net.jini.discovery.debug") != null) {
                e.printStackTrace();
            }
        } catch (SecurityException ee) { }
    }//end printException
}
