/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package net.jini.lease;

import java.io.IOException;
import java.rmi.MarshalledObject;
import net.jini.core.lease.Lease;
import net.jini.core.event.RemoteEvent;

/**
 * Basic implementation of <code>RenewalFailureEvent</code> that
 * defers unmarshalling 
 * <p>
 * The <code>Lease</code> that could not be renewed and any
 * accompanying <code>Throwable</code> will not be deserialized until
 * the appropriate accessor is called.
 * 
 * @see LeaseRenewalSet 
 */
public class BasicRenewalFailureEvent extends RenewalFailureEvent {
    private static final long serialVersionUID = -4648837058030641864L;

    /**
     * Throwable, in marshaled form, returned by <code>getThrowable</code>
     * method.  May be null.
     * @serial
     */
    private MarshalledObject marshalledThrowable;

    /**
     * Lease, in marshaled form, returned by <code>getLease</code>
     * method.  
     * @serial 
     */
    private MarshalledObject marshaledLease;

    /**
     * Transient cache of lease returned by <code>getLease</code> method.  
     */
    private transient Lease lease = null;

    /**
     * Transient cache of throwable returned by <code>getThrowable</code> 
     * method.
     */
    private transient Throwable throwable;

    /**
     * Simple constructor.  Note event id is fixed to
     * <code>LeaseRenewalSet.RENEWAL_FAILURE_EVENT_ID</code>
     *
     * @param source the <code>LeaseRenewalSet</code> that generated the event
     * @param seqNum the sequence number of this event
     * @param handback the client handback
     * @param marshaledLease the lease which could not be renewed in
     *        marshaled form.
     * @param throwable the first throwable that was thrown in the last
     *        chain of renewal failures in marshaled form.  May be <code>
     *        null</code> in which case getThrowable() will return <code>
     *        null</code>
     */
    public BasicRenewalFailureEvent(LeaseRenewalSet source, 
				    long seqNum,
				    MarshalledObject handback,
				    MarshalledObject marshaledLease,
				    MarshalledObject marshalledThrowable) 
    {
	super(source, seqNum, handback);
	this.marshalledThrowable = marshalledThrowable;
	this.marshaledLease = marshaledLease;
    }

    /**
     * Returns the lease that could not be renewed.  When the event is
     * deserialized the lease is left in marshaled form.  It is only
     * unmarshalled when this call is made.  If a call to this method
     * fails future calls will attempt to re-unmarshal the lease.
     * Once the lease is successfully unmarshalled it is cached in a
     * transient field so future calls will not result in
     * <code>IOException</code> or
     * <code>ClassNotFoundException</code>.
     *
     * @throws IOException can be thrown if there are difficulties
     * unmarshalling the lease usually this will be some sort class
     * miss-match.
     *
     * @throws ClassNotFoundException can be thrown if there are
     * difficulties unmarshalling lease, usually this will indicate one
     * of the classes associated with the lease's implementation could
     * not be loaded.  
     */
    public Lease getLease() throws IOException, ClassNotFoundException {
	if (lease == null) 
	    lease = (Lease)marshaledLease.get();

	return lease;
    };

    /**
     * Returns the <code>Throwable</code> (if any) that was thrown by
     * the last renewal attempt.
     * <p>
     * When the event is deserialized the <code>Throwable</code> is
     * left in marshaled form.  It is only unmarshalled when this call
     * is made.  If a call to this method fails future calls will
     * attempt to re-unmarshal the <code>Throwable</code>.  Once the
     * <code>Throwable</code> is successfully unmarshalled it is
     * cached in a transient field so future calls will not result in
     * <code>IOException</code> or
     * <code>ClassNotFoundException</code>.
     *
     * @throws IOException can be thrown if there are difficulties
     * unmarshalling the <code>Throwable</code> usually this will be
     * some sort class miss-match.
     *
     * @throws ClassNotFoundException can be thrown if there are
     * difficulties unmarshalling <code>Throwable</code>, usually this
     * will indicate one of the classes associated with the
     * <code>Throwable</code>'s implementation could not be loaded.
     *  */
    public Throwable getThrowable() throws IOException, ClassNotFoundException {
	// Is there a throwable?
	if (marshalledThrowable == null)
	    return null;

	if (throwable == null) {
	    throwable = (Throwable)marshalledThrowable.get();
	}

	return throwable;
    }
}

