/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */

package net.jini.lease;

import net.jini.core.lease.Lease;

/** 
 * The interface that receivers of <code>LeaseRenewalEvent</code>s must
 * implement.
 * <p>
 * With respect to an entity that uses the <code>LeaseRenewalManager</code>
 * to manage leases granted to the entity, this interface defines the
 * mechanism through which an entity receives notification that the 
 * <code>LeaseRenewalManager</code> has failed to renew one of the leases
 * the <code>LeaseRenewalManager</code> is managing for the entity. Such
 * renewal failures typically occur because of one of the following
 * conditions is met:
 * <ul>
 *   <li> After successfully renewing a lease any number of times, and
 *        experiencing no failures, the
 *        <code>LeaseRenewalManager</code> determines - prior to the
 *        next renewal attempt - that the actual expiration time of
 *        the lease has passed; implying that any further attempt to
 *        renew the lease would be fruitless
 *   <li> An indefinite exception occurs during each attempt to renew a
 *        lease from the point that the first such exception occurs
 *        until the point when the <code>LeaseRenewalManager</code>
 *        determines that lease's actual expiration time has passed.
 *   <li> A bad object, bad invocation, or <code>LeaseException</code>
 *        occurs during a lease renewal attempt (collectively referred 
 *        to as definite exceptions).
 * </ul>
 * It is the responsibility of the entity to register with the 
 * <code>LeaseRenewalManager</code>. The object that
 * implements this interface should define the actions to
 * take upon receipt of such notifications. Then, when one of the above 
 * conditions occurs, the <code>LeaseRenewalManager</code> will send an 
 * instance of the <code>LeaseRenewalEvent</code> class to that listener
 * object. Note that prior to sending the event, the 
 * <code>LeaseRenewalManager</code> will remove the affected lease from
 * its managed set of leases.
 * 
 * @see net.jini.core.lease.Lease
 * @see net.jini.lease.LeaseRenewalManager
 * @see net.jini.lease.LeaseRenewalEvent */
public interface LeaseListener extends java.util.EventListener {
    /**
     * Called by the <code>LeaseRenewalManager</code> when it cannot 
     * renew a lease that it is managing, and the lease's desired expiration
     * time has not yet been reached.
     * <p>
     * Note that prior to invoking this method, the
     * <code>LeaseRenewalManager</code> removes the affected lease
     * from the managed set of leases. Note also that because of the
     * re-entrancy guarantee made by the <code>LeaseRenewalManager</code>,
     * new leases can be safely added by this method.
     * <p>
     * @param e  instance of <code>LeaseRenewalEvent</code> containing
     *           information about the lease that the 
     *           <code>LeaseRenewalManager</code> was unable to renew, as
     *           well as information about the condition that made the 
     *           <code>LeaseRenewalManager</code> fail to renew the lease 
     */
    void notify(LeaseRenewalEvent e);
}
