/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */

package net.jini.lease;

import com.sun.jini.thread.TaskManager;
import com.sun.jini.constants.ThrowableConstants;
import java.rmi.NoSuchObjectException;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;
import net.jini.core.lease.*;

/**
 * Provides for the systematic renewal and overall management of a set
 * of leases associated with one or more remote entities on behalf of
 * a local entity.
 * <p>
 * This class removes much of the administrative burden associated
 * with lease renewal.  Clients of the renewal manager simply give
 * their lease's to the manager and the manager renews each lease as
 * necessary to achieve a <em>desired expiration</em> time (which may
 * be latter than the lease's current <em>actual expiration</em>
 * time).  Failures encountered while renewing a lease can optionally
 * be reflected to the client via <code>LeaseRenewalEvent</code>s.
 * <p>
 * Note that this class is not remote. Entities wishing to use this
 * class must create an instance of this class in their own virtual
 * machine to locally manage the leases granted to them.  If the
 * virtual machine that the manager was created in exits or crashes
 * the renewal manager will be destroyed.
 * <p>
 * The LeaseRenewalManager distinguishes between two time values
 * associated with lease expiration: the <em>desired expiration</em>
 * time for the lease, and the <em>actual expiration</em> time granted
 * when the lease is created or last renewed. The desired expiration
 * represents when the client would like the lease to expire. The
 * actual expiration represents when the lease is going to expire if
 * it is not renewed. Both time values are absolute times, not
 * relative time durations. The desired expiration time can be
 * retrieved using the renewal manager's <code>getExpiration</code>
 * method. The actual expiration time of a lease object can be
 * retrieved by invoking the lease's <code>getExpiration</code>
 * method.
 * <p>
 * Each lease in the managed set also has two other associated
 * attributes: a desired <em>renewal duration</em>, and a
 * <em>remaining desired duration</em>. The desired renewal duration
 * is specified (directly or indirectly) when the lease is added to
 * the set. This duration must normally be a positive number; however,
 * it may be <code>Lease.ANY</code> if the lease's desired expiration
 * is <code>Lease.FOREVER</code>. The remaining desired duration is
 * always the desired expiration less the current time.
 * <p>
 * Each time a lease is renewed, the renewal manager will ask for an
 * extension equal to the lease's renewal duration if the renewal
 * duration is:
 * <ul>
 * <li> <code>Lease.ANY</code>, or 
 * <li> less than the remaining desired duration,
 * </ul>
 * otherwise it will ask for an extension equal to the lease's
 * remaining desired duration. 
 * <p>
 * Once a lease lease is given to a lease renewal manager, the manager
 * will continue to renew the lease until one of the following occurs:
 * <ul>
 * <li> The lease's desired or actual expiration time is reached.
 * <li> An explicit removal of the lease from the set is requested via
 *      a <code>cancel</code>, <code>clear</code>, or <code>remove</code>
 *      call on the renewal manager.
 * <li> The renewal manager tries to renew the lease and gets a bad
 *      object exception, bad invocation exception, or
 *      <code>LeaseException</code>.
 * </ul>
 * <p>
 * The methods of this class are appropriately synchronized for
 * concurrent operation. Additionally, this class makes certain
 * guarantees with respect to concurrency. When this class makes a
 * remote call (for example, when requesting the renewal of a lease),
 * any invocations made on the methods of this class will not be
 * blocked. Similarly, this class makes a re-entrancy guarantee with
 * respect to the listener objects registered with this class. Should
 * this class invoke a method on a registered listener (a local call),
 * calls from that method to any other method of this class are
 * guaranteed not to result in a deadlock condition.
 * 
 * @see net.jini.core.lease.Lease
 * @see net.jini.core.lease.LeaseException
 * @see LeaseRenewalEvent 
 *
 * @author Zane Pan */

public class LeaseRenewalManager {
    /** time window in which to look for batchable leases */
    private static final long RENEW_BATCH_TIME_GAP = 1000 * 60 * 5;
    /** assumed round trip time for any lease renewal */
    private static final long RENEW_RTT = 1000 * 10;
    /** maximum number of concurrent lease renewals */
    private static final int  MAX_RENEW_THREAD = 10;

    /** entries for leases that are not actively being renewed */
    private final SortedMap leases = new TreeMap();
    /** entries for leases that are actively being renewed */
    private final List leaseInRenew = new ArrayList(1);
    /** the queuer task */
    private QueuerTask queuer = null;
    /** task manager for queueing and renewing leases */
    TaskManager tasker = new TaskManager(MAX_RENEW_THREAD + 1,
					 1000 * 15, 1.0f);
    /** maintains concurrency constraints when calculating actual renewals */
    private final List calcList = new ArrayList(MAX_RENEW_THREAD);

    private final class RenewTask implements TaskManager.Task {
	/** entries of leases to renew (if multiple, all can be batched) */
	private final List bList;

	/**
	 * Create a collection of entries whose leases can be batch renewed
	 * with the last lease in the map, remove them from the map, and
	 * add them to leaseInRenew.
	 */
	RenewTask() {
	    bList = new ArrayList(1);
	    Entry e = (Entry)leases.lastKey();
	    Map lMap = leases.tailMap(new Entry(e.renew));
	    for (Iterator iter = lMap.values().iterator(); iter.hasNext(); ) {
		Entry be = (Entry)iter.next();
		if (be == e || be.canBatch(e)) {
		    iter.remove();
		    leaseInRenew.add(be);
		    bList.add(be);
		}
	    }
	}

	public void run() {
	    long now = System.currentTimeMillis();
	    List bad = processExpiredLeases(now);
	    if (!bList.isEmpty())
		renewAll(bList, now);
	    if (bad != null)
		tell(bad);
	}

	/** No ordering */
	public boolean runAfter(List tasks, int size) {
	    return false;
	}

	/**
	 * Find any expired leases, remove them from bList and leaseInRenew,
	 * and return any with listeners.
	 */
	private List processExpiredLeases(long now) {
	    List bad = null;
	    synchronized (LeaseRenewalManager.this) {
		for (Iterator iter = bList.iterator(); iter.hasNext(); ) {
		    Entry e = (Entry)iter.next();
		    if (e.endTime <= now) {
			iter.remove();
			if (removeLeaseInRenew(e) && e.listener != null) {
			    if (bad == null)
				bad = new ArrayList(1);
			    bad.add(e);
			}
		    }
		}
	    }
	    return bad;
	}
    }

    private static class Entry implements Comparable {
	/* Since the cnt only gets modified in the constructor, and the
	 * constructor is always called from synchronized code, the cnt
	 * does not need to be synchronized.
	 */
	private static long cnt = 0;

	/** uid */
	public final long id;
	/** the lease */
	public Lease lease;
	/** desired expiration */
	public long expiration;
	/** renew duration */
	public long renewDuration;
	/** the listener, or null */
	public LeaseListener listener;
	/** current expiration */
	public long endTime;
	/** preferred time to renew */
	public long renew;
	/** actual time to renew, given concurrency limitations */
	public long actualRenew;
	/** renewal exception, or null */
	public Throwable ex = null;

	public Entry(Lease lease,
		     long expiration,
		     long renewDuration,
		     LeaseListener listener)
	{
	    this.endTime = lease.getExpiration();
	    this.lease = lease;
	    this.expiration = expiration;
	    this.renewDuration = renewDuration;
	    this.listener = listener;
	    id = cnt++;
	}

	/** create a fake entry for tailMap */
	public Entry(long renew) {
	    this.renew = renew + RENEW_BATCH_TIME_GAP;
	    id = Long.MAX_VALUE;
	}

	/**
	 * If the renewDuration is ANY, return ANY, otherwise return the
	 * minimum of the renewDuration and the time remaining until the
	 * desired expiration.
	 */
	public long getRenewDuration(long now) {
	    if (renewDuration == Lease.ANY)
		return renewDuration;
	    return Math.min(expiration - now, renewDuration);
	}

	/** calculate the renew time for the lease entry */
	public void calcRenew(long now) {
	    endTime = lease.getExpiration();
	    if (expiration <= endTime) {
		renew = Long.MAX_VALUE; // no need for renew anymore
		return;
	    }
	    long delta = endTime - now;
	    if (delta <= RENEW_RTT * 2)
		delta = RENEW_RTT;
	    else if (delta <= RENEW_RTT * 8)
		delta /= 2;
	    else if (delta <= 1000 * 60 * 60 * 24 * 7)
		delta /= 8;
	    else if (delta <= 1000 * 60 * 60 * 24 * 14)
		delta = 1000 * 60 * 60 * 24;
	    else
		delta = 1000 * 60 * 60 * 24 * 3;
	    renew = endTime - delta;
	}

	/** calculate a new renew time due to an indefinite exception */
	public void delayRenew() {
	    long delta = endTime - renew;
	    if (delta <= RENEW_RTT)
		return;
	    else if (delta <= RENEW_RTT * 3)
		 delta = RENEW_RTT;
	    else if (delta <= 1000 * 60 * 60)
		delta /= 3;
	    else if (delta <= 1000 * 60 * 60 * 24)
		delta = 1000 * 60 * 30;
	    else if (delta <= 1000 * 60 * 60 * 24 * 7)
		delta = 1000 * 60 * 60 * 3;
	    else
		delta = 1000 * 60 * 60 * 8;
	    renew += delta;
	}

	/** sort by decreasing renew time, secondary sort by decreasing id */
	public int compareTo(Object obj) {
	    if (this == obj)
		return 0;
	    Entry e = (Entry)obj;
	    if (renew < e.renew || (renew == e.renew && id < e.id))
		return 1;
	    return -1;
	}

	/**
	 * Returns true if this lease can be batched with the (earlier)
	 * renewal of the given lease. Returns true if the two leases can
	 * be batched and a) either the renew duration of this lease is
	 * ANY or the time remaining until the renewal of this lease
	 * (measured at the time e would be renewed) is no more than
	 * RENEW_RTT/2 or the time remaining until the current expiration
	 * of this lease (measured at the time e would be renewed) is no more
	 * than 1/2 of the renew duration, and b) either the renew duration
	 * of e's lease is ANY or e's lease will not be renewed again before
	 * this lease needs to be renewed.  This method must be called with
	 * an entry such that e.renew <= this.renew.
	 */
	public boolean canBatch(Entry e) {
	    return (lease.canBatch(e.lease) &&
		    (renewDuration == Lease.ANY ||
		     renew - e.renew <= RENEW_RTT / 2 ||
		     endTime - e.renew <= renewDuration / 2) &&
		    (e.renewDuration == Lease.ANY ||
		     e.renew > renew - e.renewDuration ||
		     e.renew >= e.expiration - e.renewDuration));
	}
    }

    /**
     * No-argument constructor that creates an instance of this class that
     * initially manages no leases.
     */
    public LeaseRenewalManager() {
    }

    /**
     * Constructs an instance of this class that will initially manage
     * a single lease. Employing this form of the constructor is
     * equivalent to invoking the no-argument form of the constructor
     * followed by an invocation of the three-argument form of the
     * <code>renewUntil</code> method.  See <code>renewUntil</code>
     * for details on the arguments and what exceptions may be thrown
     * by this constructor.
     *
     * @param lease       reference to the initial lease to manage
     * @param desiredExpiration  the desired expiration
     *                    for <code>lease</code>
     * @param listener    reference to the <code>LeaseListener</code> object
     *                    that will receive notifications of any exceptional
     *                    conditions that occur during renewal attempts. If
     *                    <code>null</code> no notifications will
     *                    be sent.
     * @throws java.lang.NullPointerException if <code>lease</code> is 
     * <code>null</code>
     *
     * @see net.jini.lease.LeaseListener
     * @see net.jini.lease.LeaseRenewalManager#renewUntil 
     */
    public LeaseRenewalManager(Lease lease,
                               long desiredExpiration,
			       LeaseListener listener)
    {
	renewUntil(lease, desiredExpiration, listener);
    }

    /**
     * Include a lease in the managed set 
     * until a specified time. 
     * <p>
     * If <code>desiredExpiration</code> is <code>Lease.ANY</code> calling
     * this method is equivalent the following call:
     * <pre>
     *     renewUntil(lease, Lease.FOREVER, Lease.ANY, listener)
     * </pre>
     * otherwise it is equivalent to this call:
     * <pre>
     *     renewUntil(lease, desiredExpiration, Lease.FOREVER, listener)
     * </pre>
     * <p>
     * @param lease The <code>Lease</code> to be managed
     * @param desiredExpiration When the client wants the lease to expire
     *                          in milliseconds since the beginning of the 
     *                          epoch.
     * @param listener    reference to the <code>LeaseListener</code> object
     *                    that will receive notifications of any exceptional
     *                    conditions that occur during renewal attempts. If
     *                    <code>null</code> no notifications will
     *                    be sent.
     * @throws NullPointerException if lease is <code>null</code>
     * @see net.jini.lease.LeaseRenewalManager#renewUntil
     */
    public void renewUntil(Lease lease,
			   long desiredExpiration,
			   LeaseListener listener)
    {
	if (desiredExpiration == Lease.ANY)
	    renewUntil(lease, Lease.FOREVER, Lease.ANY, listener);
	else
	    renewUntil(lease, desiredExpiration, Lease.FOREVER, listener);
    }

    /**
     * Include a lease in the managed set until a
     * specified time and with a specified renewal duration.
     * <p>
     * This method takes as arguments: a reference to the lease to
     * manage, the desired expiration time of the lease, the renewal
     * duration time for the lease, and a reference to the
     * <code>LeaseListener</code> object that will receive notification of
     * exceptional conditions when attempting to renew this lease. The
     * <code>LeaseListener</code> argument may be <code>null</code>.
     * <p>
     * If <code>null</code> is passed as the lease parameter, a
     * <code>NullPointerException</code> will be thrown. If the
     * <code>desiredExpiration</code> parameter is
     * <code>Lease.FOREVER</code>, the <code>renewalDuration</code>
     * parameter may be <code>Lease.ANY</code> or any positive value;
     * otherwise, the <code>renewalDuration</code> parameter must be a
     * positive value. If the <code>renewalDuration</code> parameter
     * does not meet these requirements, an
     * <code>IllegalArgumentException</code> will be thrown.
     * <p>
     * If the lease passed to this method is already in the set of
     * managed leases, the listener object, the desired expiration, and
     * the renewal duration associated with that lease will be replaced
     * with the new listener, desired expiration, and renewal duration.
     * <p>
     * The lease will remain in the set until one of the following occurs:
     * <ul>
     * <li> The lease's desired or actual expiration time is reached.
     * <li> An explicit removal of the lease from the set is requested via
     *      a <code>cancel</code>, <code>clear</code>, or <code>remove</code>
     *      call on the renewal manager.
     * <li> The renewal manager tries to renew the lease and gets a bad
     *      object exception, bad invocation exception, or
     *      <code>LeaseException</code>.
     * </ul>
     * <p>
     * This method will interpret the value of the
     * <code>desiredExpiration</code> parameter as the desired
     * absolute system time after which the lease is no longer
     * valid. This argument provides the ability to indicate an
     * expiration time that extends beyond the actual expiration of
     * the lease. If the value passed for this argument does indeed
     * extend beyond the lease's actual expiration time, then the
     * lease will be systematically renewed at appropriate times until
     * one of the conditions listed above occurs.  If the value is
     * less than or equal to the actual expiration time, nothing will
     * be done to modify the time when the lease actually
     * expires. That is, the lease will not be renewed with an
     * expiration time that is less than the actual expiration time of
     * the lease at the time of the call.
     * <p>
     * If a non-<code>null</code> object reference is passed in as the
     * <code>LeaseListener</code> parameter, it will receive
     * notification of exceptional conditions occurring upon a renewal
     * attempt of the lease. In particular, exceptional conditions
     * include the reception of a <code>LeaseException</code>, bad
     * object exception, or bad invocation exception (collectively
     * these are referred to as <em>definite exceptions</em>) during a
     * renewal attempt or the lease's actual expiration being reached
     * before its desired expiration.
     * <p>
     * If a definite exception occurs during a lease renewal request,
     * the exception will be wrapped in an instance of the
     * <code>LeaseRenewalEvent</code> class and sent to the listener.
     * <p>
     * If an indefinite exception occurs during a renewal request for
     * the lease, renewal requests will continue to be made for that
     * lease until: the lease is renewed successfully, a renewal
     * attempt results in a definite exception, or the lease's actual
     * expiration time has been exceeded. If the lease cannot be
     * successfully renewed before its actual expiration is reached,
     * the exception associated with the most recent renewal attempt
     * will be wrapped in an instance of the
     * <code>LeaseRenewalEvent</code> class and sent to the listener.
     * <p>
     * If the lease's actual expiration is reached before the lease's
     * desired expiration time, and either 1) the last renewal attempt
     * succeeded or 2) there have been no renewal attempts, a
     * <code>LeaseRenewalEvent</code> containing a <code>null</code>
     * exception will be sent to the listener.
     * <p>    
     * @param lease The <code>Lease</code> to be managed
     * @param desiredExpiration When the client wants the lease to expire
     *                          in milliseconds since the beginning of the 
     *                          epoch.
     * @param renewDuration The renewal duration to associate with the lease in 
     *                      milliseconds.
     * @param listener    reference to the <code>LeaseListener</code> object
     *                    that will receive notifications of any exceptional
     *                    conditions that occur during renewal attempts. If
     *                    <code>null</code> no notifications will
     *                    be sent.
     * @throws NullPointerException if lease is <code>null</code>
     * @throws IllegalArgumentException if <code>renewDuration</code> is invalid
     * <p>
     * @see LeaseRenewalEvent
     * @see net.jini.core.lease.LeaseException
     */
    public void renewUntil(Lease lease,
			   long desiredExpiration,
			   long renewDuration,
			   LeaseListener listener)
    {
	validateDuration(renewDuration, desiredExpiration == Lease.FOREVER,
			 "desiredExpiration");
	addLease(lease, desiredExpiration, renewDuration, listener,
		 System.currentTimeMillis());
    }

    /**
     * Include a lease in the managed set for a specified duration.  
     * <p>
     * Calling this method is equivalent the following call:
     * <pre>
     *     renewFor(lease, desiredDuration, Lease.FOREVER, listener)
     * </pre>
     *
     * @param lease       reference to the new lease to manage
     * @param desiredDuration
     *                    The desired duration (relative time) that the 
     *                    caller wants <code>lease</code> to be valid for in
     *                    milliseconds.
     * @param listener    reference to the <code>LeaseListener</code> object
     *                    that will receive notifications of any exceptional
     *                    conditions that occur during renewal attempts. If
     *                    <code>null</code> is input, no notifications will
     *                    be sent.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         the value of the <code>lease</code> parameter is 
     *         <code>null</code>.
     *
     * @see net.jini.lease.LeaseRenewalManager#renewFor 
     */
    public void renewFor(Lease lease, long desiredDuration, 
			 LeaseListener listener) 
    {
	renewFor(lease, desiredDuration, Lease.FOREVER, listener);
    }    

    /**
     * Include a lease in the managed set for a
     * specified duration and with specified renewal duration.
     * <p>
     * The semantics of this method are similar to those of the
     * four-argument form of <code>renewUntil</code>, with
     * <code>desiredDuration</code> + current time being used for the
     * value of the <code>desiredExpiration</code> parameter of
     * <code>renewUntil</code>. The only exception to this is that, in
     * the context of <code>renewFor</code> the value of the
     * <code>renewDuration</code> parameter may only be
     * <code>Lease.ANY</code> if the value of the
     * <code>desiredDuration</code> parameter is <em>exactly</em>
     * <code>Lease.FOREVER.</code>
     * <p>
     * This method tests for arithmetic overflow in the desired
     * expiration time computed from the value of
     * <code>desiredDuration</code> parameter
     * (<code>desiredDuration</code> + current time). Should such
     * overflow be present, a value of <code>Lease.FOREVER</code> is
     * used to represent the lease's desired expiration time.  
     * <p>
     *
     * @param lease       reference to the new lease to manage
     * @param desiredDuration
     *                    The desired duration (relative time) that the 
     *                    caller wants <code>lease</code> to be valid for in
     *                    milliseconds.
     * @param renewDuration The renewal duration to associate with the lease in 
     *                      milliseconds.
     * @param listener    reference to the <code>LeaseListener</code> object
     *                    that will receive notifications of any exceptional
     *                    conditions that occur during renewal attempts. If
     *                    <code>null</code> is input, no notifications will
     *                    be sent.
     *
     * @throws java.lang.NullPointerException this exception occurs when
     *         the value of the <code>lease</code> parameter is 
     *         <code>null</code>.
     * @throws IllegalArgumentException if <code>renewDuration</code> is invalid
     *
     * @see net.jini.lease.LeaseRenewalManager#renewUntil 
     */
    public void renewFor(Lease lease,
			 long desiredDuration,
			 long renewDuration,
			 LeaseListener listener)
    {
	/* Validate before calculating effective desiredExpiration,
	 * if they want a renewDuration of Lease.ANY, desiredDuration has
	 * to be exactly Lease.FOREVER
	 */
	validateDuration(renewDuration, desiredDuration == Lease.FOREVER,
			 "desiredDuration");

	long now = System.currentTimeMillis();
	long desiredExpiration;
	if (desiredDuration < Lease.FOREVER - now) // check overflow.
	    desiredExpiration = now + desiredDuration;
	else
	    desiredExpiration = Lease.FOREVER;
	addLease(lease, desiredExpiration, renewDuration, listener, now);
    }

    /**
     * Error checking function that ensures renewDuration is valid taking
     * into account the whether or not the desired expiration/duration is
     * Lease.FOREVER.  Throws an appropriate IllegalArgumentException if
     * an invalid renewDuration is passed.
     * @param renewDuration Renew duration the clients wants
     * @param isForever     Should be true if client asked for a
     *                      desired expiration/duration of exactly Lease.FOREVER
     * @param name          Name of the desired expiration/duration field, used
     *                      to construct exception
     * @throws IllegalArgumentException if renewDuration is invalid
     */
    private void validateDuration(long renewDuration, boolean isForever, 
				  String name) 
    {
 	if (renewDuration <= 0 && 
	    !(renewDuration == Lease.ANY && isForever))
	{
	    /* A negative renew duration and is not lease.ANY with a
	     * forever desired expiration
	     */
	    if (renewDuration == Lease.ANY) {
		/* Must have been Lease.ANY with a non-FOREVER desired
		 * expiration
		 */
		throw new IllegalArgumentException("A renewDuration of " +
		     "Lease.ANY can only be used with a " + name + " of " +
		     "Lease.FOREVER");
	    } 

	    if (isForever) {
		// Must have been a non-Lease.ANY, non-positive renewDuration
		throw new IllegalArgumentException("When " + name + " is " +
		    "Lease.FOREVER the only valid values for renewDuration " +
		    "are a positive number, Lease.ANY, or Lease.FOREVER");
	    }

	    /* Must be a non-positive renewDuration with a non-Forever 
	     * desired expiration
	     */
	    throw new IllegalArgumentException("When the " + name +
		" is not Lease.FOREVER the only valid values for " +
		"renewDuration are a positive number or Lease.FOREVER");
	}
    }

    private synchronized void addLease(Lease lease,
				       long desiredExpiration,
				       long renewDuration,
				       LeaseListener listener,
				       long now)
    {	    
	Entry e = findEntryDo(lease);
	if (e != null && !removeLeaseInRenew(e))
	    leases.remove(e);
	insertEntry(new Entry(lease, desiredExpiration, renewDuration,
			      listener),
		    now);
	calcActualRenews(now);
    }

    /** calculate the preferred renew time, and put in the map */
    private void insertEntry(Entry e, long now) {
	e.calcRenew(now);
	leases.put(e, e);
    }

    /**
     * Returns the current desired expiration time associated with a
     * particular lease, (not the actual expiration that was granted
     * when the lease was created or last renewed).
     *
     * @param lease  the lease the caller wants the current desired
     *               expiration for.
     *
     * @return a <code>long</code> value corresponding to the current
     *         desired expiration time associated with <code>lease</code>
     *
     * @throws <code>net.jini.core.lease.UnknownLeaseException</code> this
     *         exception occurs when the lease input to this method is not in
     *         the set of managed leases
     *
     * @see net.jini.core.lease.UnknownLeaseException
     */
    public synchronized long getExpiration(Lease lease)
	throws UnknownLeaseException
    {
	return findEntry(lease).expiration;
    }

    /**
     * Replaces the current desired expiration of a given lease from the
     * managed set with a new desired expiration time.
     * <p>
     * Note that an invocation of this method with a lease that is currently
     * a member of the managed set is equivalent to an invocation of the
     * <code>renewUntil</code> method with the lease's current listener
     * input to that method's <code>listener</code> parameter. Specifically,
     * if the value of the <code>expiration</code> parameter is less than
     * or equal to the lease's current desired expiration, this method takes
     * no action.
     *
     * @param lease       the lease whose desired expiration time should
     *                    be replaced
     *
     * @param expiration  <code>long</code> value representing the new
     *                    desired expiration time for the
     *                    <code>lease</code> parameter
     *
     * @throws <code>net.jini.core.lease.UnknownLeaseException</code> this
     *         exception occurs when the lease input to this method is not in
     *         the set of managed leases
     *
     * @see net.jini.lease.LeaseRenewalManager#renewUntil
     * @see net.jini.core.lease.UnknownLeaseException
     */
    public synchronized void setExpiration(Lease lease, long expiration)
	throws UnknownLeaseException
    {
	Entry e = findEntry(lease);
	e.expiration = expiration;
	if (expiration != Lease.FOREVER && e.renewDuration == Lease.ANY)
	    e.renewDuration = Lease.FOREVER;
	if (leaseInRenew.indexOf(e) < 0) {
	    leases.remove(e);
	    long now = System.currentTimeMillis();
	    insertEntry(e, now);
	    calcActualRenews(now);
	}
    }

    /**
     * Removes a given lease from the managed set, and cancels it.
     * <p>
     * Note that even if an exception is thrown as a result of the cancel
     * operation, the lease will still have been removed from the set of
     * leases managed by this class. Additionally, any exception thrown by
     * the <code>cancel</code> method of the lease object itself may also
     * be thrown by this method.
     *
     * @param lease the lease to remove and cancel
     *
     * @throws <code>net.jini.core.lease.UnknownLeaseException</code> this
     *         exception occurs when the lease input to this method is not in
     *         the set of managed leases
     *
     * @throws java.rmi.RemoteException typically, this exception occurs when
     *         there is a communication failure between the client and the
     *         server. When this exception does occur, the lease may or may
     *         not have been successfully cancelled, (but the lease is
     *         guaranteed to have been removed from the managed set).
     *
     * @see net.jini.core.lease.Lease#cancel
     * @see net.jini.core.lease.UnknownLeaseException
     */
    public void cancel(Lease lease)
	throws UnknownLeaseException, RemoteException
    {
	remove(lease);
	lease.cancel();
    }

    /**
     * Removes a given lease from the managed set of leases; but does not
     * cancel the given lease.
     *
     * @param lease the lease to remove from the managed set
     *
     * @throws <code>net.jini.core.lease.UnknownLeaseException</code> this
     *         exception occurs when the lease input to this method is not in
     *         the set of managed leases
     *
     * @see net.jini.core.lease.UnknownLeaseException
     */
    public synchronized void remove(Lease lease) throws UnknownLeaseException {
	Entry e = findEntry(lease);
	if (!removeLeaseInRenew(e))
	    leases.remove(e);
	calcActualRenews();
    }

    /**
     * Removes all leases from the managed set of leases. This method does
     * not request the cancellation of the removed leases.
     */
    public synchronized void clear() {
	leases.clear();
	leaseInRenew.clear();
	calcActualRenews();
    }

    /** calculate the actual renew times, and poke/restart the queuer */
    private void calcActualRenews() {
	calcActualRenews(System.currentTimeMillis());
    }

    /** calculate the actual renew times, and poke/restart the queuer */
    private void calcActualRenews(long now) {
	for (Iterator iter = leases.values().iterator(); iter.hasNext(); ) {
	    Entry e = (Entry)iter.next();
	    e.actualRenew = e.renew;
	    if (e.renew >= e.endTime) {
		if (now >= e.endTime)
		    iter.remove();
		continue;
	    }
	    if (e.endTime <= now && e.listener == null) {
		// lease has expired and no listener, just remove it.
		iter.remove();
		continue;
	    }
	    if (!canBatch(e)) {
		for (Iterator listIter = calcList.iterator();
		     listIter.hasNext(); )
		{
		    if (e.renew >=
			((Entry)listIter.next()).actualRenew - RENEW_RTT)
			break;
		    listIter.remove();
		}
		if (calcList.size() == MAX_RENEW_THREAD) {
		    Entry e1 = (Entry)calcList.remove(0);
		    e.actualRenew = e1.actualRenew - RENEW_RTT;
		}
		calcList.add(e);
	    }
	}
	calcList.clear();
	long newWakeup = wakeupTime();
	if (queuer == null) {
	    if (newWakeup < Long.MAX_VALUE) {
		queuer = new QueuerTask(newWakeup);
		tasker.add(queuer);
	    }
	} else if (newWakeup < queuer.wakeup ||
		   (newWakeup == Long.MAX_VALUE && leaseInRenew.isEmpty()))
	{
	    notifyAll();
	}
    }

    /**
     * Return true if e can be batched with another entry that expires
     * between e.renew - RENEW_BATCH_TIME_GAP and e.renew.
     */
    private boolean canBatch(Entry e) {
	Iterator iter = leases.tailMap(e).values().iterator();
	iter.next(); // skip e itself
	while (iter.hasNext()) {
	    Entry be = (Entry)iter.next();
	    if (e.renew - be.renew > RENEW_BATCH_TIME_GAP)
		break;
	    if (e.canBatch(be))
		return true;
	}
	return false;
    }

    /** find a lease entry, throw exception if not found or expired normally */
    private Entry findEntry(Lease lease) throws UnknownLeaseException {
	Entry e = findEntryDo(lease);
	if (e != null &&
	    (e.renew < e.endTime || System.currentTimeMillis() < e.endTime))
	    return e;
	throw new UnknownLeaseException();
    }

    /** find a lease entry, or null */
    private Entry findEntryDo(Lease lease) {
	Entry e = findLeaseFromIterator(leases.values().iterator(), lease);
	if (e == null)
	    e = findLeaseFromIterator(leaseInRenew.iterator(), lease);
	return e;
    }

    /** find a lease entry, or null */
    private static Entry findLeaseFromIterator(Iterator iter, Lease lease) {
	while (iter.hasNext()) {
	    Entry e = (Entry)iter.next();
	    if (e.lease.equals(lease))
		return e;
	}
	return null;
    }

    /** notify the listener for each lease */
    private void tell(List bad) {
	for (Iterator iter = bad.iterator(); iter.hasNext(); ) {
	    Entry e = (Entry)iter.next();
	    e.listener.notify(new LeaseRenewalEvent(this, e.lease,
						    e.expiration, e.ex));
	}
    }

    /** renew all of the leases (if multiple, all can be batched) */
    private void renewAll(List bList, long now) {
        Map lmeMap = null;
	Throwable t = null;
	List bad = null;

	try {
	    if (bList.size() == 1) {
		Entry e = (Entry)bList.get(0);
		e.lease.renew(e.getRenewDuration(now));
	    } else {
		LeaseMap batchLeaseMap = createBatchLeaseMap(bList, now);
		batchLeaseMap.renewAll();
	    }
	} catch (LeaseMapException ex) {
	    lmeMap = ex.exceptionMap;
	    bad = new ArrayList(lmeMap.size());
	} catch (Throwable ex) {
	    t = ex;
	    bad = new ArrayList(bList.size());  // They may all be bad
	}

	/* For each lease we tried to renew determine the associated
	 * exception (if any), and then ether add the lease back
	 * to leases (if the renewal was successful), schedule a retry
	 * and add back to leases (if the renewal was indefinite), or
	 * drop the lease (by not adding it back to leases) and 
	 * notify any interested listeners.  In any event remove 
	 * lease from the list of leases being renewed.
	 */

	now = System.currentTimeMillis();
	synchronized (this) {
	    for (Iterator iter = bList.iterator(); iter.hasNext(); ) {
		Entry e = (Entry)iter.next();

		if (!removeLeaseInRenew(e))
		    continue;

		// Update the entries exception field 
		if (bad == null) {
		    e.ex = null;
		} else {
		    e.ex = (t != null)?t:(Throwable)lmeMap.get(e.lease);
		}

		if (e.ex == null) {
		    // No problems just put back in list
		    insertEntry(e, now);
		    continue;
		} 

		/* Some sort of problem.  If definite don't put back
		 * into leases and setup to notify the appropriate
		 * listener, if indefinite schedule for a retry and put
		 * back into leases
		 */
		final int cat = ThrowableConstants.retryable(e.ex);
		if (cat == ThrowableConstants.INDEFINITE) {
		    e.delayRenew();
		    leases.put(e, e);		    
		} else if (e.listener != null) { 
		    // Note: For us ThrowableConstants.UNCATEGORIZED == definite
		    bad.add(e);
		}		
	    }
	    calcActualRenews(now);
	}

	if (bad != null)
	    tell(bad);	
    }

    /** create a LeaseMap for batch renewal */
    private static LeaseMap createBatchLeaseMap(List bList, long now) {
	Iterator iter = bList.iterator();
	Entry e = (Entry)iter.next();
	LeaseMap batchLeaseMap =
	    e.lease.createLeaseMap(e.getRenewDuration(now));
	while (iter.hasNext()) {
	    e = (Entry)iter.next();
	    batchLeaseMap.put(e.lease, new Long(e.getRenewDuration(now)));
	}
	return batchLeaseMap;
    }

    /** remove from leaseInRenew, return true if removed */
    private boolean removeLeaseInRenew(Entry e) {
	int index = leaseInRenew.indexOf(e); // avoid iterator cons
	if (index < 0)
	    return false;
	leaseInRenew.remove(index);
	return true;
    }

    /** return the soonest actual renewal time */
    private long wakeupTime() {
	if (leases.isEmpty())
	    return Long.MAX_VALUE;
	return ((Entry)leases.lastKey()).actualRenew;
    }

    private class QueuerTask implements TaskManager.Task {

	/** when to next wake up and queue a new renew task */
	long wakeup;

	QueuerTask(long wakeup) {
	    this.wakeup = wakeup;
	}

	/** No ordering */
	public boolean runAfter(List tasks, int size) {
	    return false;
	}

        public void run() {
	    synchronized (LeaseRenewalManager.this) {
		try {
		    while (true) {
			wakeup = wakeupTime();
			if (wakeup == Long.MAX_VALUE && leaseInRenew.isEmpty())
			    break;
			long delta = wakeup - System.currentTimeMillis();
			if (delta <= 0)
			    tasker.add(new RenewTask());
			else
			    LeaseRenewalManager.this.wait(delta);
		    }
		} catch (InterruptedException ex) {
		}
		queuer = null;
	    }
	}
    }
}
