/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package net.jini.lease;

import net.jini.core.lease.Lease;

import java.rmi.MarshalledObject;

/**
 * Exception thrown when a lease renewal set can't unmarshall one or more
 * leases being returned by a <code>getLeases</code> call.
 * <p>
 * When unmarshalling an instance of <code>MarshalledObject</code>,
 * one of the following checked exceptions is possible: an
 * <code>IOException</code> can occur while deserializing the object
 * from its internal representation; and a
 * <code>ClassNotFoundException</code> can occur if, while
 * deserializing the object from its internal representation, either
 * the class file of the object cannot be found, or the class file of
 * either an interface or a class referenced by the object being
 * deserialized cannot be found.  Typically, a
 * <code>ClassNotFoundException</code> occurs when the codebase from
 * which to retrieve the needed class file is not currently available.
 * <p>
 * This class provides a mechanism that clients of the lease renewal
 * service may use for efficient handling of the exceptions that may
 * occur when unmarshalling elements of a set of marshalled 
 * of <code>Lease</code> objects. When elements in such a set are
 * unmarshalled, the <code>LeaseUnmarshalException</code> class may be
 * used to collect and report pertinent information generated when
 * failure occurs while unmarshalling the elements of the set.
 * <p>
 * The information that may be of interest to entities that receive
 * this exception class is contained in the following fields: a set of
 * <code>Lease</code> objects in which each element is the result of
 * a successful unmarshalling attempt, a set of marshalled
 * <code>Lease</code> objects in which each element could not be
 * successfully unmarshalled, and a set of exceptions
 * (<code>IOException</code>, <code>ClassNotFoundException</code>, or
 * some unchecked exception) in which each element corresponds to one
 * of the unmarshalling failures.
 * <p>
 * Thus, when exceptional conditions occur while unmarshalling a set of 
 * marshalled <code>Lease</code> objects, this class can be used not 
 * only to indicate that an exceptional condition has occurred, but also 
 * to provide information that can be used to perform error handling 
 * activities such as: determining if it is feasible to continue with 
 * processing, reporting errors, attempting recovery, and debugging.
 * <p>
 * Note that this exception class should be used only to report exceptional
 * conditions occurring when unmarshalling a set of marshalled 
 * <code>Lease</code> objects.
 *
 * @see net.jini.core.lease.Lease 
 */
public class LeaseUnmarshalException extends Exception {
    private static final long serialVersionUID = -6736107321698417489L;

    /**
     * Array containing the set of instances of <code>Lease</code> that were
     * successfully unmarshalled during the process in which at least one
     * failure occurred.
     *
     * @serial
     */
    private Lease[] unmarshalledLeases = null;

    /**
     * Array containing the set of <code>Lease</code> instances that could not
     * be unmarshalled. This set should contain at least one element.
     *
     * @serial
     */
    private MarshalledObject[] stillMarshalledLeases = null;

    /**
     * Array containing the set of exceptions that occurred during the
     * unmarshalling process. Each element in this set should be an
     * instance of <code>IOException</code>,
     * <code>ClassNotFoundException</code>, or some unchecked
     * exception. Furthermore, there should be a one-to-one
     * correspondence between each element in this set and each
     * element in the set of still-to-be-unmarshalled
     * <code>Lease</code> instances.  That is, the element of this set
     * corresponding to index i should be an instance of the exception
     * that occurred while attempting to unmarshal the element at
     * index i of stillMarshalledLeases.
     *
     * @serial 
     */
    private Throwable[] exceptions = null;

    /**
     * Constructs a new instance of <code>LeaseUnmarshalException</code>. 
     * With a specified message
     *
     * @param message             The detail message.
     * @param unmarshalledLeases  Array containing the set of instances of
     *                            <code>Lease</code> that were successfully
     *                            unmarshalled.
     *                            
     * @param stillMarshalledLeases
     *                            Array containing the set of marshalled
     *                            <code>Lease</code> instances that could not
     *                            be unmarshalled.
     *                   
     * @param exceptions          Array containing the set of exceptions that
     *                            occurred during the unmarshalling
     *                            process.  Each element in this set
     *                            should be an instance of
     *                            <code>IOException</code>,
     *                            <code>ClassNotFoundException</code>,
     *                            or some unchecked
     *                            exception. Furthermore, there should
     *                            be a one-to-one correspondence
     *                            between each element in this set and
     *                            each element in the
     *                            <code>stillMarshalledLeases</code> parameter.
     * <p>
     *                            That is, the element of this set
     *                            corresponding to index i should be
     *                            an instance of the exception that
     *                            occurred while attempting to
     *                            unmarshal the element at index i of
     *                            the <code>stillMarshalledRegs</code>
     *                            parameter.
     *
     * @throws java.lang.IllegalArgumentException this exception occurs
     *         when the number of elements in the
     *         <code>exceptions</code> parameter is not equal to the
     *         number of elements in <code>the</code> stillMarshalledLeases
     *         parameter.
     */
    public LeaseUnmarshalException(String  message,
				   Lease[] unmarshalledLeases,
				   MarshalledObject[] stillMarshalledLeases,
				   Throwable[] exceptions) 
    {
        super(message);
        /* If the number of exceptions does not equal number of leases
         * that could not be unmarshalled, throw exception
         */
        if(exceptions.length != stillMarshalledLeases.length) {
            throw new IllegalArgumentException
                           ("exceptions.length ("+exceptions.length
                            +") is not equal to stillMarshalledLeases.length"
                            +stillMarshalledLeases.length+")");
        }
        this.unmarshalledLeases = unmarshalledLeases;
        this.stillMarshalledLeases = stillMarshalledLeases;
        this.exceptions = exceptions;
    }

    /**
     * Accessor method that returns an array consisting of instances of 
     * <code>Lease</code>, where each element of the array corresponds to a 
     * successfully unmarshalled object. Note that the same array is
     * returned on each invocation of this method; that is, a copy is
     * not made.
     *
     * @return array of instances of <code>Lease</code>, where each element
     *         corresponds to a successfully unmarshalled object.
     */
    public Lease[] getUnmarshalledLeases() {
        return unmarshalledLeases;
    }

    /**
     * Accessor method that returns an array consisting of instances
     * of <code>MarshalledObject</code>, where each element of the
     * array is a marshalled instance of the <code>Lease</code>
     * interface, and corresponds to an object that could not be
     * successfully unmarshalled. Note that the same array is returned
     * on each invocation of this method; that is, a copy is not made.
     *
     * @return array of marshalled instances of <code>Lease</code>, where
     *         each element corresponds to an object in which failure
     *         occurred while attempting to unmarshal the object.  
     */
    public MarshalledObject[] getStillMarshalledLeases() {
        return stillMarshalledLeases;
    }

    /**
     * Accessor method that returns an array consisting of instances
     * of <code>Throwable</code>, where each element of the array
     * corresponds to one of the exceptions that occurred during the
     * unmarshalling process.  Note that the same array is returned on
     * each invocation of this method; that is, a copy is not made.
     * <p> 
     * Each element in the return set should be an instance of
     * <code>IOException</code>, <code>ClassNotFoundException</code>,
     * or some unchecked exception. Additionally, there should be a
     * one-to-one correspondence between each element in the array
     * returned by this method and the array returned by the
     * <code>getStillMarshalledLeases</code> method. That is, the i-th
     * element of the set returned by this method should be an
     * instance of the exception that occurred while attempting to
     * unmarshal the i-th element of the set returned by the
     * <code>getStillMarshalledLeases</code> method.
     *
     * @return array of instances of <code>Throwable</code>, where each element
     *         corresponds to one of the exceptions that occurred during
     *         the unmarshalling process.  
     */
    public Throwable[] getExceptions() {
        return exceptions;
    }
}

