/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */

package net.jini.lookup;
  
import net.jini.core.lookup.ServiceItem;

/**
 * The LookupCache interface defines the methods provided by the object
 * created and returned by the <code>ServiceDiscoveryManager</code> when
 * a client-like entity invokes the <code>createLookupCache</code> method.
 * It is within that object that discovered service references, which match
 * criteria defined by the entity, are stored. Through this interface the
 * entity may retrieve one or more of the stored service references,
 * register and un-register with the cache's event mechanism, and terminate
 * all of the cache's processing. This interface is used by the
 * <code>ServiceDiscoveryManager</code>.
 *
 * @see ServiceDiscoveryManager
 * @author Zane Pan
 */
public interface LookupCache {
   /** 
     * Finds a <code>ServiceItem</code> object that satisfies the
     * filter. 
     * <p>
     * The service item returned must have been previously discovered to 
     * be both registered with one or more of the lookup services in the 
     * managed set, and to match criteria defined by the entity.
     * <p>
     * The semantics of the <code>filter</code> argument are
     * identical to those of the <code>filter</code> argument specified
     * for a number of the methods defined in the interface of the
     * <code>ServiceDiscoveryManager</code> utility class. This argument is
     * intended to allow an entity to separate its filtering into two
     * steps: an initial filter applied during the discovery phase, and
     * a finer resolution filter applied upon retrieval from the cache.
     * As with the methods of the <code>ServiceDiscoveryManager</code>, if
     * <code>null</code> is the value of this argument, then no additional
     * filtering will be performed.
     *
     * @param filter used for matching <code>ServiceItem</code>s. A null 
     * 		     value means no additional filter should be applied.
     * 		    
     * @return ServiceItem that satisfies the filter, and that was 
     * 			previously discovered to be registered with one
     * 			or more lookup services in the managed set. A 
     * 			<code>null</code> value will be returned if no
     * 			<code>ServiceItem</code> is found that matches the 
     * 			criteria or if the cache is empty.
     * 			
     */
    public ServiceItem lookup(ServiceItemFilter filter);

   /** 
     * Finds an array of <code>ServiceItem</code>s that satisfy the
     * filter.
     * <p>
     * The service item returned must have been previously discovered to 
     * be both registered with one or more of the lookup services in the 
     * managed set, and to match criteria defined by the entity.
     * <p>
     * The semantics of the <code>filter</code> argument are
     * identical to those of the <code>filter</code> argument specified
     * for a number of the methods defined in the interface of the
     * <code>ServiceDiscoveryManager</code> utility class. This argument is
     * intended to allow an entity to separate its filtering into two
     * steps: an initial filter applied during the discovery phase, and
     * a finer resolution filter applied upon retrieval from the cache.
     * As with the methods of the <code>ServiceDiscoveryManager</code>, if
     * <code>null</code> is the value of this argument, then no
     * additional filtering will be performed.
     * 
     * @param filter used for matching <code>ServiceItem</code>s. A null 
     * 		     value means no additional filter should be applied.
     * 		    
     * @param maxMatches maximum number of matches to return. If this 
     *			value is set to <code>Integer.MAX_VALUE</code>
     *			then all elements in the cache that match the
     *			criteria will be returned.
     *
     * @return ServiceItem[] that satisfy the filter, and that were 
     * 			previously discovered to be registered with one
     * 			or more lookup services in the managed set. An 
     * 			empty array will be returned if no
     *                  <code>ServiceItem</code> is found that matches
     * 			the criteria or if the cache is empty.
     *
     * @throws IllegalArgumentException if <code>maxMatches</code> is a
     *			negative number.
     */
    public ServiceItem[] lookup(ServiceItemFilter filter, int maxMatches);

   /** 
     * Registers a <code>ServiceDiscoveryListener</code> object with
     * the event mechanism of a <code>LookupCache</code>. The listener 
     * object will receive a <code>ServiceDiscoveryEvent</code> upon the 
     * discovery, removal, or modification of one of the cache's
     * services. Once a listener is registered, it will be notified of
     * all service references discovered to date, and will be notified as
     * new services are discovered and existing services are modified or
     * discarded. If the parameter value is <code>null</code>, no action 
     * will be taken.
     *
     * @param l the <code>ServiceDiscoveryListener</code> object to
     * 			register.
     */
    public void addListener(ServiceDiscoveryListener l);


   /** 
     * Removes a <code>ServiceDiscoveryListener</code> object from the set  
     * of listeners currently registered with the <code>LookupCache</code>.
     * Once all listeners are removed from the cache's set of listeners, 
     * the cache will send no more <code>ServiceDiscoveryEvent</code> 
     * notifications.
     *
     * @param l the <code>ServiceDiscoveryListener</code> object to remove.
     */
    public void removeListener(ServiceDiscoveryListener l);

   /** 
     * Deletes a service reference from the cache and causes a notification 
     * to be sent to all registered listeners indicating that the service 
     * has been discarded . Note that after the service has been discarded, 
     * there is no guarantee that the service will again become eligible
     * for discovery.
     *
     * @param serviceReference the service reference to
     * 		discard.
     */
    public void discard(Object serviceReference);

   /** 
     * Performs cleanup duties related to the termination of the processing
     * being performed by a particular instance of
     * <code>LookupCache</code>. For that instance, this method cancels
     * all event leases granted by the lookup services that supplied the
     * contents of the cache, and un-exports all remote listener objects
     * registered with those lookup services. The terminate method is
     * typically called when the entity is no longer interested in the
     * contents of the <code>LookupCache</code>.
     */
    void terminate();
}
