/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */

package net.jini.lookup;

import net.jini.core.lookup.ServiceItem;
/**
 * The <code>ServiceDiscoveryListener</code> interface defines the
 * methods used by objects such as a <code>LookupCache</code> to notify
 * an entity that events of interest related to the elements of the cache
 * have occurred. It is the responsibility of the entity wishing to be
 * notified of the occurrence of such events to construct an object
 * that implements the <code>ServiceDiscoveryListener</code> interface
 * and then register that object with the cache's event mechanism. Any
 * implementation of this interface must define the actions to take upon
 * receipt of an event notification. The action taken is dependent on
 * both the application and the particular event that has occurred.
 * <p>
 * Should an exception occur during an invocation of any of the methods
 * defined by this interface, the semantics of how that exception is
 * handled are undefined.
 * <p>
 * Each method defined by this interface must not modify the contents
 * of the <code>ServiceDiscoveryEvent</code> parameter; doing so can
 * result in unpredictable and undesirable effects on future processing
 * by the <code>ServiceDiscoveryManager</code>. It is for this reason
 * that if one of these methods modifies the contents of the parameter,
 * the effects are undefined.
 * <p>
 * This interface makes the following concurrency guarantee. For any
 * given listener object that implements this interface, no two methods
 * (either the same two methods or different methods) defined by the
 * interface can be invoked at the same time by the same cache. For
 * example, the serviceRemoved method must not be invoked while the
 * invocation of another listener's <code>serviceAdded</code> method is
 * in progress.
 * <p>
 * Finally, it should be noted that the intent of the methods of this
 * interface is to allow the recipient of the 
 * <code>ServiceDiscoveryEvent</code> to be informed that a service has been
 * added to, removed from, or modified in the cache. Calls to these methods
 * are synchronous to allow the entity that makes the call (for example, a
 * thread that interacts with the various lookup services of interest) to
 * determine whether or not the call succeeded. However, it is not part of
 * the semantics of the call that the notification return can be delayed
 * while the recipient of the call reacts to the occurrence of the event.
 * Thus, it is highly recommended that implementations of this interface
 * avoid time consuming operations, and return from the method as quickly
 * as possible. For example, one strategy might be to simply note the
 * occurrence of the <code>ServiceDiscoveryEvent</code>, and perform any
 * time consuming event handling asynchronously. 
 *
 * @see LookupCache
 * @see ServiceDiscoveryEvent
 * @author Zane Pan
 */
public interface ServiceDiscoveryListener {
    /** 
     * When the cache receives from one of the managed lookup services,
     * an event signaling the <i>registration</i> of a service of
     * interest for the <i>first time</i> (or for the first time since
     * the service has been discarded), the cache invokes the
     * <code>serviceAdded</code> method on all instances of
     * <code>ServiceDiscoveryListener</code> that are registered with the
     * cache; doing so notifies the entity that a service of interest has
     * been discovered.
     *
     * @param event an instance of <code>ServiceDiscoveryEvent</code> 
     * 			containing references to the service item 
     *			corresponding to the event, including 
     * 			representations of the service's state both 
     *			before and after the event.
     */
    void serviceAdded(ServiceDiscoveryEvent event);

    /** 
     * When the cache receives, from a managed lookup service, an event
     * signaling the removal of a service of interest from the last such
     * lookup service with which it was registered, the cache invokes
     * the <code>serviceRemoved</code> method on all instances of
     * <code>ServiceDiscoveryListener</code> that are registered with
     * the cache; doing so notifies the entity that a service of interest 
     * has been discarded.
     * 
     * @param event a <code>ServiceDiscoveryEvent</code> object 
     * 			containing references to the service item 
     *			corresponding to the event, including 
     *			representations of the service's state both 
     *			before and after the event.
     */
    void serviceRemoved(ServiceDiscoveryEvent event);

    /** 
     * When the cache receives, from a managed lookup service, an event
     * signaling the unique modification of the attributes of a service
     * of interest (across the attribute sets of all references to the
     * service), the cache invokes the <code>serviceChanged</code>
     * method on all instances of <code>ServiceDiscoveryListener</code>
     * that are registered with the cache; doing so notifies the entity
     * that the state of a service of interest has changed.
     * 
     * @param event a <code>ServiceDiscoveryEvent</code> object
     * 			containing references to the service item 
     * 			corresponding to the event, including 
     * 			representations of the service's state both
     * 			before and after the event.
     */
    void serviceChanged(ServiceDiscoveryEvent event);
}
