/*
 * 
 * Copyright 2000 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * The contents of this file are made available under and subject to the Research
 * Use License of the Sun Community Source License, Jini Technology Core Platform
 * Version 1.0 (the "License").  This file has not been tested and may not pass
 * the relevant Technology Compatibility Kit for Internal Deployment Use or
 * Commercial Use.  You must test this file and it must constitute Compliant
 * Covered Code in accordance with Your License prior to any Internal Deployment
 * Use or Commercial Use by You.
 * 
 * Contributor(s): Sun Microsystems, Inc.
 * 
 * Version 1.1Beta
 * 
 */
package com.sun.jini.tool;

import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FilePermission;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.MissingResourceException;
import java.util.ResourceBundle;
import java.util.StringTokenizer;
import java.util.jar.Attributes;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.jar.Manifest;

/** A simple HTTP server, for serving up jar and class files. */
public class ClassServer extends Thread {
    /** Default HTTP port */
    private static int DEFAULT_PORT = 8080;
    /** Default directory to serve files from on non-Windows OS */
    private static String DEFAULT_DIR = "/vob/jive/jars";
    /** Default directory to serve files from on Windows */
    private static String DEFAULT_WIN_DIR = "J:";

    /** Server socket to accept connections on */
    private ServerSocket server;
    /** Directory to serve from */
    private String dir;
    /** Map from String (jar root) to JarFile[] (jar class path) */
    private HashMap map;
    /** Verbosity flag */
    private boolean verbose;
    /** Read permission on dir and all subdirs */
    private FilePermission perm;

    /**
     * Construct a server.  Use the start method to run it.
     *
     * @param port the port to use
     * @param dir the directory to serve files from
     * @param trees true if files within jar files should be served up
     * @param verbose true if downloads should be logged
     */
    public ClassServer(int port, String dir, boolean trees, boolean verbose)
	throws IOException
    {
	if (!dir.endsWith(File.separator))
	    dir = dir + File.separatorChar;
	this.dir = dir;
	this.verbose = verbose;
	server = new ServerSocket(port);
	perm = new FilePermission(dir + '-', "read");
	if (!trees)
	    return;
	map = new HashMap();
	String[] files = new File(dir).list();
	for (int i = 0; i < files.length; i++) {
	    String jar = files[i];
	    if (!jar.endsWith(".jar"))
		continue;
	    ArrayList jfs = new ArrayList(1);
	    addJar(jar, jfs, dir);
	    map.put(jar.substring(0, jar.length() - 4),
		    jfs.toArray(new JarFile[jfs.size()]));
	}
    }

    /** Add transitive Class-Path jars to jfs. */
    private void addJar(String jar, ArrayList jfs, String dir)
	throws IOException
    {
	if (verbose)
	    print("classserver.jar", dir + jar);
	JarFile jf = new JarFile(dir + jar);
	jfs.add(jf);
	Manifest man = jf.getManifest();
	if (man == null)
	    return;
	Attributes attrs = man.getMainAttributes();
	if (attrs == null)
	    return;
	String val = attrs.getValue(Attributes.Name.CLASS_PATH);
	if (val == null)
	    return;
	dir = dir + jar.substring(0, jar.lastIndexOf(File.separatorChar) + 1);
	StringTokenizer st = new StringTokenizer(val);
	while (st.hasMoreTokens()) {
	    addJar(st.nextToken().replace('/', File.separatorChar), jfs, dir);
	}
    }

    /** Just keep looping, spawning a new thread for each incoming request. */
    public void run() {
	try {
	    while (true) {
		new Task(server.accept()).start();
	    }
	} catch (IOException e) {
	    e.printStackTrace();
	}
    }

    /** Simple task thread */
    private class Task extends Thread {
	/** Socket for the incoming request */
	private Socket sock;

	/** Simple constructor */
	public Task(Socket sock) {
	    this.sock = sock;
	}

	/** Read the request and return the initial request line. */
	private String getRequest() throws IOException {
	    BufferedReader in =
		new BufferedReader(new InputStreamReader(
						      sock.getInputStream()));
	    String req = in.readLine();
	    if (req != null) {
		String line;
		do {
		    line = in.readLine();
		} while (line != null && line.length() > 0 &&
			 line.charAt(0) != '\r' && line.charAt(0) != '\n');
	    }
	    return req;
	}

	/** Read specified number of bytes and always close the stream. */
	private byte[] getBytes(InputStream in, long length)
	    throws IOException
	{
	    DataInputStream din = new DataInputStream(in);
	    byte[] bytes = new byte[(int)length];
	    try {
		din.readFully(bytes);
	    } finally {
		din.close();
	    }
	    return bytes;
	}

	/** Return the bytes of the requested file, or null if not found. */
	private byte[] getBytes(String path) throws IOException {
	    if (path.length() == 0 || path.charAt(0) != '/')
		return null;
	    path = path.substring(1);
	    if (map != null) {
		int i = path.indexOf('/');
		if (i > 0) {
		    JarFile[] jfs = (JarFile[])map.get(path.substring(0, i));
		    if (jfs != null) {
			String jpath = path.substring(i + 1);
			for (i = 0; i < jfs.length; i++) {
			    JarEntry je = jfs[i].getJarEntry(jpath);
			    if (je != null)
				return getBytes(jfs[i].getInputStream(je),
						je.getSize());
			}
		    }
		}
	    }
	    File f = new File(dir + path.replace('/', File.separatorChar));
	    if (perm.implies(new FilePermission(f.getPath(), "read"))) {
		try {
		    return getBytes(new FileInputStream(f), f.length());
		} catch (FileNotFoundException e) {
		}
	    }
	    return null;
	}

	/** Process the request */
	public void run() {
	    try {
		DataOutputStream out =
		    new DataOutputStream(sock.getOutputStream());
		String req;
		try {
		    req = getRequest();
		} catch (Exception e) {
		    if (verbose) {
			print("classserver.inputerror",
			      new String[]{sock.getInetAddress().getHostName(),
					   Integer.toString(sock.getPort())});
			e.printStackTrace();
		    }
		    return;
		}
		if (req == null)
		    return;
		if (req.startsWith("SHUTDOWN *")) {
		    if (verbose)
			print("classserver.shutdown",
			      new String[]{sock.getInetAddress().getHostName(),
					   Integer.toString(sock.getPort())});
		    try {
			new ServerSocket(0, 1, sock.getInetAddress());
		    } catch (IOException e) {
			out.writeBytes("HTTP/1.0 403 Forbidden\r\n\r\n");
			out.flush();
			return;
		    }
		    System.exit(0);
		}
		if (req.length() <= 4 || !req.startsWith("GET ")) {
		    if (verbose)
			print("classserver.badrequest",
			      new String[]{req,
					   sock.getInetAddress().getHostName(),
					   Integer.toString(sock.getPort())});
		    out.writeBytes("HTTP/1.0 400 Bad Request\r\n\r\n");
		    out.flush();
		    return;
		}
		String path = req.substring(4);
		int i = path.indexOf(' ');
		if (i > 0)
		    path = path.substring(0, i);
		if (verbose)
		    print("classserver.request",
			  new String[]{path,
				       sock.getInetAddress().getHostName(),
				       Integer.toString(sock.getPort())});
		byte[] bytes;
		try {
		    bytes = getBytes(path);
		} catch (Exception e) {
		    if (verbose)
			e.printStackTrace();
		    out.writeBytes("HTTP/1.0 500 Internal Error\r\n\r\n");
		    out.flush();
		    return;
		}
		if (bytes == null) {
		    if (verbose)
			print("classserver.notfound", path);
		    out.writeBytes("HTTP/1.0 404 Not Found\r\n\r\n");
		    out.flush();
		    return;
		}
		out.writeBytes("HTTP/1.0 200 OK\r\n");
		out.writeBytes("Content-Length: " + bytes.length + "\r\n");
		out.writeBytes("Content-Type: application/java\r\n\r\n");
		out.write(bytes);
		out.flush();
		fileDownloaded(path.substring(1), sock.getInetAddress());
	    } catch (Exception e) {
		if (verbose)
		    e.printStackTrace();
	    } finally {
		try {
		    sock.close();
		} catch (IOException e) {
		}
	    }
	}
    }

    private static ResourceBundle resources;
    private static boolean resinit = false;

    private static String getString(String key) {
	if (!resinit) {
	    try {
		resources = ResourceBundle.getBundle("com.sun.jini.tool.resources.classserver");
		resinit = true;
	    } catch (MissingResourceException e) {
		e.printStackTrace();
	    }
	}
	try {
	    return resources.getString(key);
	} catch (MissingResourceException e) {
	    return null;
	}
    }

    private static void print(String key, String val) {
	String fmt = getString(key);
	if (fmt == null)
	    fmt = "no text found: \"" + key + "\" {0}";
	System.out.println(MessageFormat.format(fmt, new String[]{val}));
    }

    private static void print(String key, String[] vals) {
	String fmt = getString(key);
	if (fmt == null)
	    fmt = "no text found: \"" + key + "\" {0} {1} {2}";
	System.out.println(MessageFormat.format(fmt, vals));
    }

    /**
     * This method provides a way for subclasses to be notified when a
     * file has been completely downloaded.
     * 
     * @param fp The path to the file that was downloaded.
     */
    protected void fileDownloaded(String fp, InetAddress addr) {
    }

    /**
     * Command line interface for creating an HTTP server.
     * The command line options are:
     * <pre>
     * [-port <var>port</var>] [-dir <var>dir</var>] [-trees] [-verbose]
     * </pre>
     * The default port is 8080; the default can be overridden with
     * the -port option.  The default directory on Windows is J:
     * and the default on other systems is /vob/jive/jars; the default
     * can be overridden with the -dir option.  By default, all files
     * under this directory (including all subdirectories) are served
     * up via HTTP.  If the pathname of a file is <var>path</var> relative
     * to the top-level directory, then the file can be downloaded using
     * the URL
     * <pre>
     * http://<var>host</var>:<var>port</var>/<var>path</var>
     * </pre>
     * No caching of directory contents or file contents is performed.
     * If the -verbose option is given, then all attempts to download files
     * are output. <p>
     * 
     * The -trees option can be used to serve up individual files stored
     * within jar files in addition to the files that are served up by
     * default.  If the option is used, the server finds all jar files
     * in the top-level directory (not in subdirectories).  For each
     * jar file, if the name of the jar file is <var>name</var>.jar, then any
     * individual file named <var>file</var> within that jar file (or within
     * the jar or zip files referenced transitively in the Class-Path manifest
     * attribute, can be downloaded using a URL of the form:
     * <pre>
     * http://<var>host</var>:<var>port</var>/<var>name</var>/<var>file</var>
     * </pre>
     * When this option is used, an open file descriptor and cached
     * information is held for each jar file, for the life of the process.
     */
    public static void main(String[] args) {
	int port = DEFAULT_PORT;
	String dir = DEFAULT_DIR;
	if (File.separatorChar == '\\')
	    dir = DEFAULT_WIN_DIR;
	boolean trees = false;
	boolean verbose = false;
	boolean stop = false;
	for (int i = 0; i < args.length ; i++ ) {
	    String arg = args[i];
	    if (arg.equals("-port")) {
		i++;
		port = Integer.parseInt(args[i]);
	    } else if (arg.equals("-dir")) {
		i++;
		dir = args[i];
	    } else if (arg.equals("-verbose")) {
		verbose = true;
	    } else if (arg.equals("-trees")) {
		trees = true;
	    } else if (arg.equals("-stop")) {
		stop = true;
	    } else {
		print("classserver.usage", (String)null);
		return;
	    }
	}
	try {
	    if (stop) {
		Socket sock = new Socket(InetAddress.getLocalHost(), port);
		try {
		    DataOutputStream out =
			new DataOutputStream(sock.getOutputStream());
		    out.writeBytes("SHUTDOWN *\r\n\r\n");
		    out.flush();
		} finally {
		    try {
			sock.close();
		    } catch (IOException e) {
		    }
		}
	    } else {
		new ClassServer(port, dir, trees, verbose).start();
	    }
	} catch (IOException e) {
	    e.printStackTrace();
	}
    }
}
